Package: stochtree
Title: Stochastic Tree Ensembles (XBART and BART) for Supervised
        Learning and Causal Inference
Version: 0.4.5
Authors@R: 
  c(
    person("Drew", "Herren", email = "drewherrenopensource@gmail.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0003-4109-6611")), 
    person("Richard", "Hahn", role = "aut"),
    person("Jared", "Murray", role = "aut"),
    person("Carlos", "Carvalho", role = "aut"),
    person("Jingyu", "He", role = "aut"),
    person("Pedro", "Lima", role = "ctb"),
    person("Entejar", "Alam", role = "ctb"),
    person("stochtree", "contributors", role = c("cph")),
    person("Eigen", "contributors", role = c("cph"), comment = "C++ source uses the Eigen library for matrix operations, see inst/COPYRIGHTS"),
    person("xgboost", "contributors", role = c("cph"), comment = "C++ tree code and related operations include or are inspired by code from the xgboost library, see inst/COPYRIGHTS"),
    person("treelite", "contributors", role = c("cph"), comment = "C++ tree code and related operations include or are inspired by code from the treelite library, see inst/COPYRIGHTS"),
    person("Microsoft", "Corporation", role = c("cph"), comment = "C++ I/O and various project structure code include or are inspired by code from the LightGBM library, which is a copyright of Microsoft, see inst/COPYRIGHTS"),
    person("Niels", "Lohmann", role = c("cph"), comment = "C++ source uses the JSON for Modern C++ library for JSON operations, see inst/COPYRIGHTS"),
    person("Daniel", "Lemire", role = c("cph"), comment = "C++ source uses the fast_double_parser library internally, see inst/COPYRIGHTS"),
    person("Victor", "Zverovich", role = c("cph"), comment = "C++ source uses the fmt library internally, see inst/COPYRIGHTS")
  )
Copyright: Copyright details for stochtree's C++ dependencies, which
        are vendored along with the core stochtree source code, are
        detailed in inst/COPYRIGHTS
Description: Flexible stochastic tree ensemble software. 
    Robust implementations of Bayesian Additive Regression Trees (BART) 
    (Chipman, George, McCulloch (2010) <doi:10.1214/09-AOAS285>) 
    for supervised learning and Bayesian Causal Forests (BCF) 
    (Hahn, Murray, Carvalho (2020) <doi:10.1214/19-BA1195>) 
    for causal inference. Enables model serialization and parallel sampling 
    and provides a low-level interface for custom stochastic forest samplers.
    Includes the grow-from-root algorithm for accelerated forest sampling 
    (He and Hahn (2021) <doi:10.1080/01621459.2021.1942012>), 
    a log-linear leaf model for forest-based heteroskedasticity (Murray (2020) <doi:10.1080/01621459.2020.1813587>), 
    and the cloglog BART model of Alam and Linero (2025) <doi:10.48550/arXiv.2502.00606> for ordinal outcomes.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
LinkingTo: cpp11, BH
Suggests: testthat (>= 3.0.0),
SystemRequirements: C++17
Imports: R6, stats
URL: https://stochtree.ai/, https://github.com/StochasticTree/stochtree
BugReports: https://github.com/StochasticTree/stochtree/issues
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-06-30 14:51:01 UTC; drew
Author: Drew Herren [aut, cre] (ORCID: <https://orcid.org/0000-0003-4109-6611>),
  Richard Hahn [aut],
  Jared Murray [aut],
  Carlos Carvalho [aut],
  Jingyu He [aut],
  Pedro Lima [ctb],
  Entejar Alam [ctb],
  stochtree contributors [cph],
  Eigen contributors [cph] (C++ source uses the Eigen library for matrix
    operations, see inst/COPYRIGHTS),
  xgboost contributors [cph] (C++ tree code and related operations
    include or are inspired by code from the xgboost library, see
    inst/COPYRIGHTS),
  treelite contributors [cph] (C++ tree code and related operations
    include or are inspired by code from the treelite library, see
    inst/COPYRIGHTS),
  Microsoft Corporation [cph] (C++ I/O and various project structure code
    include or are inspired by code from the LightGBM library, which is
    a copyright of Microsoft, see inst/COPYRIGHTS),
  Niels Lohmann [cph] (C++ source uses the JSON for Modern C++ library
    for JSON operations, see inst/COPYRIGHTS),
  Daniel Lemire [cph] (C++ source uses the fast_double_parser library
    internally, see inst/COPYRIGHTS),
  Victor Zverovich [cph] (C++ source uses the fmt library internally, see
    inst/COPYRIGHTS)
Maintainer: Drew Herren <drewherrenopensource@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-30 16:40:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-06-30 17:25:43 UTC; unix
Archs: stochtree.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(extractParameter,bartmodel)
S3method(extractParameter,bcfmodel)
S3method(getRandomEffectSamples,bartmodel)
S3method(getRandomEffectSamples,bcfmodel)
S3method(plot,bartmodel)
S3method(plot,bcfmodel)
S3method(predict,bartmodel)
S3method(predict,bcfmodel)
S3method(print,ForestSamples)
S3method(print,RandomEffectSamples)
S3method(print,bartmodel)
S3method(print,bcfmodel)
S3method(summary,bartmodel)
S3method(summary,bcfmodel)
export(OutcomeModel)
export(bart)
export(bcf)
export(calibrateInverseGammaErrorVariance)
export(computeBARTPosteriorInterval)
export(computeBCFPosteriorInterval)
export(computeContrastBARTModel)
export(computeContrastBCFModel)
export(computeForestLeafIndices)
export(computeForestLeafVariances)
export(computeForestMaxLeafIndex)
export(createBARTModelFromCombinedJson)
export(createBARTModelFromCombinedJsonString)
export(createBARTModelFromJson)
export(createBARTModelFromJsonFile)
export(createBARTModelFromJsonString)
export(createBCFModelFromCombinedJson)
export(createBCFModelFromCombinedJsonString)
export(createBCFModelFromJson)
export(createBCFModelFromJsonFile)
export(createBCFModelFromJsonString)
export(createCppJson)
export(createCppJsonFile)
export(createCppJsonString)
export(createCppRNG)
export(createForest)
export(createForestDataset)
export(createForestModel)
export(createForestModelConfig)
export(createForestSamples)
export(createGlobalModelConfig)
export(createOutcome)
export(createPreprocessorFromJson)
export(createPreprocessorFromJsonString)
export(createRandomEffectSamples)
export(createRandomEffectsDataset)
export(createRandomEffectsModel)
export(createRandomEffectsTracker)
export(extractParameter)
export(getRandomEffectSamples)
export(loadForestContainerCombinedJson)
export(loadForestContainerCombinedJsonString)
export(loadForestContainerJson)
export(loadRandomEffectSamplesCombinedJson)
export(loadRandomEffectSamplesCombinedJsonString)
export(loadRandomEffectSamplesJson)
export(loadScalarJson)
export(loadVectorJson)
export(preprocessPredictionData)
export(preprocessTrainData)
export(resetActiveForest)
export(resetForestModel)
export(resetRandomEffectsModel)
export(resetRandomEffectsTracker)
export(rootResetRandomEffectsModel)
export(rootResetRandomEffectsTracker)
export(sampleBARTPosteriorPredictive)
export(sampleBCFPosteriorPredictive)
export(sampleGlobalErrorVarianceOneIteration)
export(sampleLeafVarianceOneIteration)
export(sample_without_replacement)
export(saveBARTModelToJson)
export(saveBARTModelToJsonFile)
export(saveBARTModelToJsonString)
export(saveBCFModelToJson)
export(saveBCFModelToJsonFile)
export(saveBCFModelToJsonString)
export(savePreprocessorToJson)
export(savePreprocessorToJsonString)
importFrom(R6,R6Class)
importFrom(stats,coef)
importFrom(stats,dnorm)
importFrom(stats,lm)
importFrom(stats,model.matrix)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qgamma)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,resid)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,sigma)
importFrom(stats,var)
useDynLib(stochtree, .registration = TRUE)

# stochtree R package

Software for building stochastic tree ensembles (i.e. BART, XBART) for supervised learning and causal inference.

## Getting started

`stochtree` can be installed from CRAN via

```
install.packages("stochtree")
```

The development version of stochtree can be installed from github via

```
remotes::install_github("StochasticTree/stochtree", ref="r-dev")
```

#' Forest Container Serialization Routines
#' @name ForestSamplesSerialization
#' @description
#' While the [BARTSerialization] and [BCFSerialization] topics focus on JSON serialization / deserialization for
#' entire `bartmodel` and `bcfmodel` objects, this function group provides an interface for a more focused use case:
#' loading a single [ForestSamples] container from a broader BART / BCF model (which may include multiple forests and other parametric terms).
#'
#' `loadForestContainerJson` converts a [CppJson] object representing a BART or BCF model into a [ForestSamples] container
#' by extracting the JSON indexed by a forest label (i.e. `"forest_0"`) and deserializing it into a [ForestSamples] object.
#'
#' Both `loadForestContainerJson` and `loadForestContainerCombinedJson` operate similarly, but on a list of [CppJson] or JSON string
#' representations of BART / BCF models with the same structure.
#'
#' These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#' @returns
#' Each of the functions in this group returns a `ForestSamples` object.
#' @examples
#' X <- matrix(runif(10*100), ncol = 10)
#' y <- -5 + 10*(X[,1] > 0.5) + rnorm(100)
#' bart_model <- bart(X, y, num_gfr=0, num_mcmc=10)
#' bart_json <- saveBARTModelToJson(bart_model)
#' bart_json_string <- saveBARTModelToJsonString(bart_model)
#' bart_json_list <- list(bart_json)
#' bart_json_string_list <- list(bart_json_string)
#' mean_forest <- loadForestContainerJson(bart_json, "forest_0")
#' mean_forest <- loadForestContainerCombinedJson(bart_json_list, "forest_0")
#' mean_forest <- loadForestContainerCombinedJsonString(bart_json_string_list, "forest_0")
#'
NULL
#> NULL

#' Random Effects Container Serialization Routines
#' @name RandomEffectSamplesSerialization
#' @description
#' While the [BARTSerialization] and [BCFSerialization] topics focus on JSON serialization / deserialization for
#' entire `bartmodel` and `bcfmodel` objects, this function group provides an interface for a more focused use case:
#' loading a single [RandomEffectSamples] container from a broader BART / BCF model (which may include forests and other parametric terms).
#'
#' `loadRandomEffectSamplesJson` converts a [CppJson] object representing a BART or BCF model into a [RandomEffectSamples] container
#' by extracting the JSON indexed by an integer label (i.e. `0`) and deserializing it into a [RandomEffectSamples] object.
#'
#' Both `loadRandomEffectSamplesJson` and `loadRandomEffectSamplesCombinedJson` operate similarly, but on a list of [CppJson] or JSON string
#' representations of BART / BCF models with the same structure.
#'
#' These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#' @returns
#' Each of the functions in this group returns a `RandomEffectSamples` object.
#' @examples
#' n <- 100
#' p <- 10
#' X <- matrix(runif(n*p), ncol = p)
#' rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
#' rfx_basis <- rep(1.0, n)
#' y <- (-5 + 10*(X[,1] > 0.5)) + (-2*(rfx_group_ids==1)+2*(rfx_group_ids==2))
#' bart_model <- bart(X_train=X, y_train=y, rfx_group_ids_train=rfx_group_ids,
#'                    rfx_basis_train = rfx_basis, num_gfr=0, num_mcmc=10)
#' bart_json <- saveBARTModelToJson(bart_model)
#' bart_json_string <- saveBARTModelToJsonString(bart_model)
#' bart_json_list <- list(bart_json)
#' bart_json_string_list <- list(bart_json_string)
#' rfx_container <- loadRandomEffectSamplesJson(bart_json, 0)
#' rfx_container <- loadRandomEffectSamplesCombinedJson(bart_json_list, 0)
#' rfx_container <- loadRandomEffectSamplesCombinedJsonString(bart_json_string_list, 0)
#'
NULL
#> NULL

#' @title JSON C++ Object Wrapper
#' @description
#' Wrapper around a C++ `nlohmann::json` object
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

CppJson <- R6::R6Class(
  classname = "CppJson",
  cloneable = FALSE,
  public = list(
    #' @field json_ptr External pointer to a C++ nlohmann::json object
    json_ptr = NULL,

    #' @field num_forests Number of forests in the nlohmann::json object
    num_forests = NULL,

    #' @field forest_labels Names of forest objects in the overall nlohmann::json object
    forest_labels = NULL,

    #' @field num_rfx Number of random effects terms in the nlohman::json object
    num_rfx = NULL,

    #' @field rfx_container_labels Names of rfx container objects in the overall nlohmann::json object
    rfx_container_labels = NULL,

    #' @field rfx_mapper_labels Names of rfx label mapper objects in the overall nlohmann::json object
    rfx_mapper_labels = NULL,

    #' @field rfx_groupid_labels Names of rfx group id objects in the overall nlohmann::json object
    rfx_groupid_labels = NULL,

    #' @description
    #' Create a new CppJson object.
    #' @return A new `CppJson` object.
    initialize = function() {
      self$json_ptr <- init_json_cpp()
      self$num_forests <- 0
      self$forest_labels <- c()
      self$num_rfx <- 0
      self$rfx_container_labels <- c()
      self$rfx_mapper_labels <- c()
      self$rfx_groupid_labels <- c()
    },

    #' @description
    #' Convert a forest container to json and add to the current `CppJson` object
    #' @param forest_samples `ForestSamples` R class
    #' @return None
    add_forest = function(forest_samples) {
      forest_label <- json_add_forest_cpp(
        self$json_ptr,
        forest_samples$forest_container_ptr
      )
      self$num_forests <- self$num_forests + 1
      self$forest_labels <- c(self$forest_labels, forest_label)
    },

    #' @description
    #' Convert a random effects container to json and add to the current `CppJson` object
    #' @param rfx_samples `RandomEffectSamples` R class
    #' @return None
    add_random_effects = function(rfx_samples) {
      rfx_container_label <- json_add_rfx_container_cpp(
        self$json_ptr,
        rfx_samples$rfx_container_ptr
      )
      self$rfx_container_labels <- c(
        self$rfx_container_labels,
        rfx_container_label
      )
      rfx_mapper_label <- json_add_rfx_label_mapper_cpp(
        self$json_ptr,
        rfx_samples$label_mapper_ptr
      )
      self$rfx_mapper_labels <- c(
        self$rfx_mapper_labels,
        rfx_mapper_label
      )
      rfx_groupid_label <- json_add_rfx_groupids_cpp(
        self$json_ptr,
        rfx_samples$training_group_ids
      )
      self$rfx_groupid_labels <- c(
        self$rfx_groupid_labels,
        rfx_groupid_label
      )
      json_increment_rfx_count_cpp(self$json_ptr)
      self$num_rfx <- self$num_rfx + 1
    },

    #' @description
    #' Add a scalar to the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be added to json
    #' @param field_value Numeric value of the field to be added to json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which to place the value
    #' @return None
    add_scalar = function(field_name, field_value, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        json_add_double_cpp(self$json_ptr, field_name, field_value)
      } else {
        json_add_double_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name,
          field_value
        )
      }
    },

    #' @description
    #' Add a scalar to the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be added to json
    #' @param field_value Integer value of the field to be added to json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which to place the value
    #' @return None
    add_integer = function(field_name, field_value, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        json_add_integer_cpp(self$json_ptr, field_name, field_value)
      } else {
        json_add_integer_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name,
          field_value
        )
      }
    },

    #' @description
    #' Add a boolean value to the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be added to json
    #' @param field_value Numeric value of the field to be added to json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which to place the value
    #' @return None
    add_boolean = function(field_name, field_value, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        json_add_bool_cpp(self$json_ptr, field_name, field_value)
      } else {
        json_add_bool_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name,
          field_value
        )
      }
    },

    #' @description
    #' Add a string value to the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be added to json
    #' @param field_value Numeric value of the field to be added to json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which to place the value
    #' @return None
    add_string = function(field_name, field_value, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        json_add_string_cpp(self$json_ptr, field_name, field_value)
      } else {
        json_add_string_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name,
          field_value
        )
      }
    },

    #' @description
    #' Add a vector to the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be added to json
    #' @param field_vector Vector to be stored in json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which to place the value
    #' @return None
    add_vector = function(field_name, field_vector, subfolder_name = NULL) {
      field_vector <- as.numeric(field_vector)
      if (is.null(subfolder_name)) {
        json_add_vector_cpp(self$json_ptr, field_name, field_vector)
      } else {
        json_add_vector_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name,
          field_vector
        )
      }
    },

    #' @description
    #' Add an integer vector to the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be added to json
    #' @param field_vector Vector to be stored in json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which to place the value
    #' @return None
    add_integer_vector = function(
      field_name,
      field_vector,
      subfolder_name = NULL
    ) {
      field_vector <- as.numeric(field_vector)
      if (is.null(subfolder_name)) {
        json_add_integer_vector_cpp(
          self$json_ptr,
          field_name,
          field_vector
        )
      } else {
        json_add_integer_vector_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name,
          field_vector
        )
      }
    },

    #' @description
    #' Add an array to the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be added to json
    #' @param field_vector Character vector to be stored in json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which to place the value
    #' @return None
    add_string_vector = function(
      field_name,
      field_vector,
      subfolder_name = NULL
    ) {
      if (is.null(subfolder_name)) {
        json_add_string_vector_cpp(
          self$json_ptr,
          field_name,
          field_vector
        )
      } else {
        json_add_string_vector_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name,
          field_vector
        )
      }
    },

    #' @description
    #' Add a list of vectors (as an object map of arrays) to the json object under the name "field_name"
    #' @param field_name The name of the field to be added to json
    #' @param field_list List to be stored in json
    #' @return None
    add_list = function(field_name, field_list) {
      stopifnot(sum(!sapply(field_list, is.vector)) == 0)
      list_names <- names(field_list)
      for (i in 1:length(field_list)) {
        vec_name <- list_names[i]
        vec <- field_list[[i]]
        json_add_vector_subfolder_cpp(
          self$json_ptr,
          field_name,
          vec_name,
          vec
        )
      }
    },

    #' @description
    #' Add a list of vectors (as an object map of arrays) to the json object under the name "field_name"
    #' @param field_name The name of the field to be added to json
    #' @param field_list List to be stored in json
    #' @return None
    add_string_list = function(field_name, field_list) {
      stopifnot(sum(!sapply(field_list, is.vector)) == 0)
      list_names <- names(field_list)
      for (i in 1:length(field_list)) {
        vec_name <- list_names[i]
        vec <- field_list[[i]]
        json_add_string_vector_subfolder_cpp(
          self$json_ptr,
          field_name,
          vec_name,
          vec
        )
      }
    },

    #' @description
    #' Retrieve a scalar value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be accessed from json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which the field is stored
    #' @return None
    get_scalar = function(field_name, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        stopifnot(json_contains_field_cpp(self$json_ptr, field_name))
        result <- json_extract_double_cpp(self$json_ptr, field_name)
      } else {
        stopifnot(json_contains_field_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        ))
        result <- json_extract_double_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        )
      }
      return(result)
    },

    #' @description
    #' Retrieve a integer value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be accessed from json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which the field is stored
    #' @return None
    get_integer = function(field_name, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        stopifnot(json_contains_field_cpp(self$json_ptr, field_name))
        result <- json_extract_integer_cpp(self$json_ptr, field_name)
      } else {
        stopifnot(json_contains_field_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        ))
        result <- json_extract_integer_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        )
      }
      return(result)
    },

    #' @description
    #' Retrieve a boolean value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be accessed from json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which the field is stored
    #' @return None
    get_boolean = function(field_name, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        stopifnot(json_contains_field_cpp(self$json_ptr, field_name))
        result <- json_extract_bool_cpp(self$json_ptr, field_name)
      } else {
        stopifnot(json_contains_field_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        ))
        result <- json_extract_bool_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        )
      }
      return(result)
    },

    #' @description
    #' Retrieve a string value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be accessed from json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which the field is stored
    #' @return None
    get_string = function(field_name, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        stopifnot(json_contains_field_cpp(self$json_ptr, field_name))
        result <- json_extract_string_cpp(self$json_ptr, field_name)
      } else {
        stopifnot(json_contains_field_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        ))
        result <- json_extract_string_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        )
      }
      return(result)
    },

    #' @description
    #' Retrieve a vector from the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be accessed from json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which the field is stored
    #' @return None
    get_vector = function(field_name, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        stopifnot(json_contains_field_cpp(self$json_ptr, field_name))
        result <- json_extract_vector_cpp(self$json_ptr, field_name)
      } else {
        stopifnot(json_contains_field_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        ))
        result <- json_extract_vector_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        )
      }
      return(result)
    },

    #' @description
    #' Retrieve an integer vector from the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be accessed from json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which the field is stored
    #' @return None
    get_integer_vector = function(field_name, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        stopifnot(json_contains_field_cpp(self$json_ptr, field_name))
        result <- json_extract_integer_vector_cpp(
          self$json_ptr,
          field_name
        )
      } else {
        stopifnot(json_contains_field_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        ))
        result <- json_extract_integer_vector_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        )
      }
      return(result)
    },

    #' @description
    #' Retrieve a character vector from the json object under the name "field_name" (with optional subfolder "subfolder_name")
    #' @param field_name The name of the field to be accessed from json
    #' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which the field is stored
    #' @return None
    get_string_vector = function(field_name, subfolder_name = NULL) {
      if (is.null(subfolder_name)) {
        stopifnot(json_contains_field_cpp(self$json_ptr, field_name))
        result <- json_extract_string_vector_cpp(
          self$json_ptr,
          field_name
        )
      } else {
        stopifnot(json_contains_field_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        ))
        result <- json_extract_string_vector_subfolder_cpp(
          self$json_ptr,
          subfolder_name,
          field_name
        )
      }
      return(result)
    },

    #' @description
    #' Reconstruct a list of numeric vectors from the json object stored under "field_name"
    #' @param field_name The name of the field to be added to json
    #' @param key_names Vector of names of list elements (each of which is a vector)
    #' @return None
    get_numeric_list = function(field_name, key_names) {
      output <- list()
      for (i in 1:length(key_names)) {
        vec_name <- key_names[i]
        output[[vec_name]] <- json_extract_vector_subfolder_cpp(
          self$json_ptr,
          field_name,
          vec_name
        )
      }
      return(output)
    },

    #' @description
    #' Reconstruct a list of string vectors from the json object stored under "field_name"
    #' @param field_name The name of the field to be added to json
    #' @param key_names Vector of names of list elements (each of which is a vector)
    #' @return None
    get_string_list = function(field_name, key_names) {
      output <- list()
      for (i in 1:length(key_names)) {
        vec_name <- key_names[i]
        output[[vec_name]] <- json_extract_string_vector_subfolder_cpp(
          self$json_ptr,
          field_name,
          vec_name
        )
      }
      return(output)
    },

    #' @description
    #' Convert a JSON object to in-memory string
    #' @return JSON string
    return_json_string = function() {
      return(get_json_string_cpp(self$json_ptr))
    },

    #' @description
    #' Save a json object to file
    #' @param filename String of filepath, must end in ".json"
    #' @return None
    save_file = function(filename) {
      json_save_file_cpp(self$json_ptr, filename)
    },

    #' @description
    #' Load a json object from file
    #' @param filename String of filepath, must end in ".json"
    #' @return None
    load_from_file = function(filename) {
      json_load_file_cpp(self$json_ptr, filename)
    },

    #' @description
    #' Load a json object from string
    #' @param json_string JSON string dump
    #' @return None
    load_from_string = function(json_string) {
      json_load_string_cpp(self$json_ptr, json_string)
    }
  )
)

#' @title Load Forest Samples from JSON
#' @rdname ForestSamplesSerialization
#' @param json_object Object of class `CppJson`
#' @param json_forest_label Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy (must exist in every json object in a list if a list is provided)
#' @export
loadForestContainerJson <- function(json_object, json_forest_label) {
  invisible(output <- ForestSamples$new(0, 1, T))
  output$load_from_json(json_object, json_forest_label)
  return(output)
}

#' @title Combine JSON Model Objects into ForestSamples
#' @rdname ForestSamplesSerialization
#' @param json_object_list List of objects of class `CppJson`
#' @param json_forest_label Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy (must exist in every json object in a list if a list is provided)
#' @export
loadForestContainerCombinedJson <- function(
  json_object_list,
  json_forest_label
) {
  invisible(output <- ForestSamples$new(0, 1, T))
  for (i in 1:length(json_object_list)) {
    json_object <- json_object_list[[i]]
    if (i == 1) {
      output$load_from_json(json_object, json_forest_label)
    } else {
      output$append_from_json(json_object, json_forest_label)
    }
  }
  return(output)
}

#' @title Combine JSON Strings into ForestSamples
#' @rdname ForestSamplesSerialization
#' @param json_string_list List of strings that parse into objects of type `CppJson`
#' @param json_forest_label Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy (must exist in every json object in a list if a list is provided)
#' @export
loadForestContainerCombinedJsonString <- function(
  json_string_list,
  json_forest_label
) {
  invisible(output <- ForestSamples$new(0, 1, T))
  for (i in 1:length(json_string_list)) {
    json_string <- json_string_list[[i]]
    if (i == 1) {
      output$load_from_json_string(json_string, json_forest_label)
    } else {
      output$append_from_json_string(json_string, json_forest_label)
    }
  }
  return(output)
}

#' @title Load Random Effect Samples from JSON
#' @rdname RandomEffectSamplesSerialization
#' @param json_object Object of class `CppJson`
#' @param json_rfx_num Integer index indicating the position of the random effects term to be unpacked (must exist in every json object in a list if a list is provided)
#' @export
loadRandomEffectSamplesJson <- function(json_object, json_rfx_num) {
  json_rfx_container_label <- paste0("random_effect_container_", json_rfx_num)
  json_rfx_mapper_label <- paste0("random_effect_label_mapper_", json_rfx_num)
  json_rfx_groupids_label <- paste0("random_effect_groupids_", json_rfx_num)
  invisible(output <- RandomEffectSamples$new())
  output$load_from_json(
    json_object,
    json_rfx_container_label,
    json_rfx_mapper_label,
    json_rfx_groupids_label
  )
  return(output)
}

#' @title Combine JSON Model Objects into RandomEffectSamples
#' @rdname RandomEffectSamplesSerialization
#' @param json_object_list List of objects of class `CppJson`
#' @param json_rfx_num Integer index indicating the position of the random effects term to be unpacked (must exist in every json object in a list if a list is provided)
#' @export
loadRandomEffectSamplesCombinedJson <- function(
  json_object_list,
  json_rfx_num
) {
  json_rfx_container_label <- paste0("random_effect_container_", json_rfx_num)
  json_rfx_mapper_label <- paste0("random_effect_label_mapper_", json_rfx_num)
  json_rfx_groupids_label <- paste0("random_effect_groupids_", json_rfx_num)
  invisible(output <- RandomEffectSamples$new())
  for (i in 1:length(json_object_list)) {
    json_object <- json_object_list[[i]]
    if (i == 1) {
      output$load_from_json(
        json_object,
        json_rfx_container_label,
        json_rfx_mapper_label,
        json_rfx_groupids_label
      )
    } else {
      output$append_from_json(
        json_object,
        json_rfx_container_label,
        json_rfx_mapper_label,
        json_rfx_groupids_label
      )
    }
  }
  return(output)
}

#' @title Combine JSON Strings into RandomEffectSamples
#' @rdname RandomEffectSamplesSerialization
#' @param json_string_list List of objects of class `CppJson`
#' @param json_rfx_num Integer index indicating the position of the random effects term to be unpacked (must exist in every json object in a list if a list is provided)
#' @export
loadRandomEffectSamplesCombinedJsonString <- function(
  json_string_list,
  json_rfx_num
) {
  json_rfx_container_label <- paste0("random_effect_container_", json_rfx_num)
  json_rfx_mapper_label <- paste0("random_effect_label_mapper_", json_rfx_num)
  json_rfx_groupids_label <- paste0("random_effect_groupids_", json_rfx_num)
  invisible(output <- RandomEffectSamples$new())
  for (i in 1:length(json_string_list)) {
    json_string <- json_string_list[[i]]
    if (i == 1) {
      output$load_from_json_string(
        json_string,
        json_rfx_container_label,
        json_rfx_mapper_label,
        json_rfx_groupids_label
      )
    } else {
      output$append_from_json_string(
        json_string,
        json_rfx_container_label,
        json_rfx_mapper_label,
        json_rfx_groupids_label
      )
    }
  }
  return(output)
}

#' @title Load Vector from JSON
#' @description
#' Load a vector from json
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param json_object Object of class `CppJson`
#' @param json_vector_label Label referring to a particular vector (i.e. "sigma2_global_samples") in the overall json hierarchy
#' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which vector sits
#'
#' @return R vector
#' @export
#'
#' @examples
#' example_vec <- runif(10)
#' example_json <- createCppJson()
#' example_json$add_vector("myvec", example_vec)
#' roundtrip_vec <- loadVectorJson(example_json, "myvec")
loadVectorJson <- function(
  json_object,
  json_vector_label,
  subfolder_name = NULL
) {
  if (is.null(subfolder_name)) {
    output <- json_object$get_vector(json_vector_label)
  } else {
    output <- json_object$get_vector(json_vector_label, subfolder_name)
  }
  return(output)
}

#' @title Load Scalar from JSON
#' @description
#' Load a scalar from json
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param json_object Object of class `CppJson`
#' @param json_scalar_label Label referring to a particular scalar / string value (i.e. "num_samples") in the overall json hierarchy
#' @param subfolder_name (Optional) Name of the subfolder / hierarchy under which vector sits
#'
#' @return R vector
#' @export
#'
#' @examples
#' example_scalar <- 5.4
#' example_json <- createCppJson()
#' example_json$add_scalar("myscalar", example_scalar)
#' roundtrip_scalar <- loadScalarJson(example_json, "myscalar")
loadScalarJson <- function(
  json_object,
  json_scalar_label,
  subfolder_name = NULL
) {
  if (is.null(subfolder_name)) {
    output <- json_object$get_scalar(json_scalar_label)
  } else {
    output <- json_object$get_scalar(json_scalar_label, subfolder_name)
  }
  return(output)
}

#' JSON Creation Routines
#' @name JSONCreation
#' @description
#' The [CppJson] class wraps an external pointer to C++ JSON object for a seamless R serialization interface.
#' This function group provides several utilities for creating [CppJson] objects: from scratch (i.e. initializing an empty JSON object),
#' from a JSON file, or from a string that parses to valid JSON.
#'
#' `createCppJson` creates a new (empty) [CppJson] object. `createCppJsonFile` creates and populates a [CppJson] object from data in a JSON file.
#' `createCppJsonString` creates and populates a [CppJson] object from a JSON string.
#'
#' These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#' @returns
#' Each of the functions in this group returns a `CppJson` object.
#' @examples
#' example_vec <- runif(10)
#' example_json <- createCppJson()
#' example_json$add_vector("myvec", example_vec)
#' tmpjson <- tempfile(fileext = ".json")
#' example_json$save_file(file.path(tmpjson))
#' example_json_roundtrip <- createCppJsonFile(file.path(tmpjson))
#' unlink(tmpjson)
#' example_json_string <- example_json$return_json_string()
#' example_json_roundtrip <- createCppJsonString(example_json_string)
#'
NULL
#> NULL

#' @title Create CppJson Object
#' @rdname JSONCreation
#' @export
createCppJson <- function() {
  return(invisible((CppJson$new())))
}

#' @title Create CppJson Object from File
#' @rdname JSONCreation
#' @param json_filename Name of JSON file to read. Must end in `.json`.
#' @export
createCppJsonFile <- function(json_filename) {
  invisible((output <- CppJson$new()))
  output$load_from_file(json_filename)
  return(output)
}

#' @title Create CppJson Object from String
#' @rdname JSONCreation
#' @param json_string JSON string
#' @export
createCppJsonString <- function(json_string) {
  invisible((output <- CppJson$new()))
  output$load_from_string(json_string)
  return(output)
}

#' Random Effects State Management Routines
#' @name RandomEffectStateManagement
#' @description
#' A forest sampler features two types of state: ephemeral and persistent.
#' Persistent state includes objects like [ForestSamples] and [RandomEffectSamples]
#' which constitute part of the final sampled model.
#' Ephemeral state supports the sampling computations, but is not retained after the sampler finishes.
#'
#' The two primary random-effects-based bits of ephemeral state are the [RandomEffectsModel] and [RandomEffectsTracker] classes,
#' which represent the current state of a random effects model and its corresponding tracking data structures.
#'
#' In a linear sampling loop, this ephemeral state is updated with each iteration of the sampler
#' and any retained forests are copied to a [RandomEffectSamples] object. However, in multi-chain settings,
#' the state of a random effects model must typically be "reset" at the beginning of a new chain. These function enable
#' this process by synchronizing the state of a [RandomEffectsModel] and [RandomEffectsTracker] with a corresponding element
#' of a [RandomEffectSamples] object, or by resetting both to their default (root) state.
#'
#' `resetRandomEffectsModel` resets a `RandomEffectsModel` object based on the parameters indexed by `sample_num` in a `RandomEffectsSamples` object.
#' `resetRandomEffectsTracker` resets a `RandomEffectsTracker` object based on the parameters indexed by `sample_num` in a `RandomEffectsSamples` object.
#'
#' `rootResetRandomEffectsModel` resets a `RandomEffectsModel` object to its "default" state.
#' `rootResetRandomEffectsTracker` resets a `RandomEffectsTracker` object to its "default" state.
#'
#' These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#' @returns
#' All four functions have no return type and operate in-place on the relevant [RandomEffectsModel] or [RandomEffectsTracker] objects
#' @examples
#' n <- 100
#' p <- 10
#' rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
#' rfx_basis <- matrix(rep(1.0, n), ncol=1)
#' rfx_dataset <- createRandomEffectsDataset(rfx_group_ids, rfx_basis)
#' y <- (-2*(rfx_group_ids==1)+2*(rfx_group_ids==2)) + rnorm(n)
#' y_std <- (y-mean(y))/sd(y)
#' outcome <- createOutcome(y_std)
#' rng <- createCppRNG(1234)
#' num_groups <- length(unique(rfx_group_ids))
#' num_components <- ncol(rfx_basis)
#' rfx_model <- createRandomEffectsModel(num_components, num_groups)
#' rfx_tracker <- createRandomEffectsTracker(rfx_group_ids)
#' rfx_samples <- createRandomEffectSamples(num_components, num_groups, rfx_tracker)
#' alpha_init <- rep(1,num_components)
#' xi_init <- matrix(rep(alpha_init, num_groups),num_components,num_groups)
#' sigma_alpha_init <- diag(1,num_components,num_components)
#' sigma_xi_init <- diag(1,num_components,num_components)
#' sigma_xi_shape <- 1
#' sigma_xi_scale <- 1
#' rfx_model$set_working_parameter(alpha_init)
#' rfx_model$set_group_parameters(xi_init)
#' rfx_model$set_working_parameter_cov(sigma_alpha_init)
#' rfx_model$set_group_parameter_cov(sigma_xi_init)
#' rfx_model$set_variance_prior_shape(sigma_xi_shape)
#' rfx_model$set_variance_prior_scale(sigma_xi_scale)
#' for (i in 1:3) {
#'     rfx_model$sample_random_effect(rfx_dataset=rfx_dataset, residual=outcome,
#'                                    rfx_tracker=rfx_tracker, rfx_samples=rfx_samples,
#'                                    keep_sample=TRUE, global_variance=1.0, rng=rng)
#' }
#' resetRandomEffectsModel(rfx_model, rfx_samples, 0, 1.0)
#' resetRandomEffectsTracker(rfx_tracker, rfx_model, rfx_dataset, outcome, rfx_samples)
#' rootResetRandomEffectsModel(rfx_model, alpha_init, xi_init, sigma_alpha_init,
#'                             sigma_xi_init, sigma_xi_shape, sigma_xi_scale)
#' rootResetRandomEffectsTracker(rfx_tracker, rfx_model, rfx_dataset, outcome)
#'
NULL
#> NULL

#' @title Random Effect Container C++ Wrapper
#' @description
#' Class that wraps the "persistent" aspects of a C++ random effects model, including
#' draws of the parameters and a map from the original label indices to the
#' 0-indexed label numbers used to place group samples in memory (i.e. the
#' first label is stored in column 0 of the sample matrix, the second label
#' is store in column 1 of the sample matrix, etc...)
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @description
#' Coordinates various C++ random effects classes and persists those
#' needed for prediction / serialization

RandomEffectSamples <- R6::R6Class(
  classname = "RandomEffectSamples",
  cloneable = FALSE,
  public = list(
    #' @field rfx_container_ptr External pointer to a C++ StochTree::RandomEffectsContainer class
    rfx_container_ptr = NULL,

    #' @field label_mapper_ptr External pointer to a C++ StochTree::LabelMapper class
    label_mapper_ptr = NULL,

    #' @field training_group_ids Unique vector of group IDs that were in the training dataset
    training_group_ids = NULL,

    #' @description
    #' Create a new RandomEffectSamples object.
    #' @return A new `RandomEffectSamples` object.
    initialize = function() {},

    #' @description
    #' Construct RandomEffectSamples object from other "in-session" R objects
    #' @param num_components Number of "components" or bases defining the random effects regression
    #' @param num_groups Number of random effects groups
    #' @param random_effects_tracker Object of type `RandomEffectsTracker`
    #' @return None
    load_in_session = function(
      num_components,
      num_groups,
      random_effects_tracker
    ) {
      # Initialize
      self$rfx_container_ptr <- rfx_container_cpp(
        num_components,
        num_groups
      )
      self$label_mapper_ptr <- rfx_label_mapper_cpp(
        random_effects_tracker$rfx_tracker_ptr
      )
      self$training_group_ids <- rfx_tracker_get_unique_group_ids_cpp(
        random_effects_tracker$rfx_tracker_ptr
      )
    },

    #' @description
    #' Construct RandomEffectSamples object from a json object
    #' @param json_object Object of class `CppJson`
    #' @param json_rfx_container_label Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy
    #' @param json_rfx_mapper_label Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy
    #' @param json_rfx_groupids_label Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy
    #' @return A new `RandomEffectSamples` object.
    load_from_json = function(
      json_object,
      json_rfx_container_label,
      json_rfx_mapper_label,
      json_rfx_groupids_label
    ) {
      self$rfx_container_ptr <- rfx_container_from_json_cpp(
        json_object$json_ptr,
        json_rfx_container_label
      )
      self$label_mapper_ptr <- rfx_label_mapper_from_json_cpp(
        json_object$json_ptr,
        json_rfx_mapper_label
      )
      self$training_group_ids <- rfx_group_ids_from_json_cpp(
        json_object$json_ptr,
        json_rfx_groupids_label
      )
    },

    #' @description
    #' Append random effect draws to `RandomEffectSamples` object from a json object
    #' @param json_object Object of class `CppJson`
    #' @param json_rfx_container_label Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy
    #' @param json_rfx_mapper_label Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy
    #' @param json_rfx_groupids_label Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy
    #' @return None
    append_from_json = function(
      json_object,
      json_rfx_container_label,
      json_rfx_mapper_label,
      json_rfx_groupids_label
    ) {
      rfx_container_append_from_json_cpp(
        self$rfx_container_ptr,
        json_object$json_ptr,
        json_rfx_container_label
      )
    },

    #' @description
    #' Construct RandomEffectSamples object from a json object
    #' @param json_string JSON string which parses into object of class `CppJson`
    #' @param json_rfx_container_label Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy
    #' @param json_rfx_mapper_label Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy
    #' @param json_rfx_groupids_label Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy
    #' @return A new `RandomEffectSamples` object.
    load_from_json_string = function(
      json_string,
      json_rfx_container_label,
      json_rfx_mapper_label,
      json_rfx_groupids_label
    ) {
      self$rfx_container_ptr <- rfx_container_from_json_string_cpp(
        json_string,
        json_rfx_container_label
      )
      self$label_mapper_ptr <- rfx_label_mapper_from_json_string_cpp(
        json_string,
        json_rfx_mapper_label
      )
      self$training_group_ids <- rfx_group_ids_from_json_string_cpp(
        json_string,
        json_rfx_groupids_label
      )
    },

    #' @description
    #' Append random effect draws to `RandomEffectSamples` object from a json object
    #' @param json_string JSON string which parses into object of class `CppJson`
    #' @param json_rfx_container_label Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy
    #' @param json_rfx_mapper_label Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy
    #' @param json_rfx_groupids_label Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy
    #' @return None
    append_from_json_string = function(
      json_string,
      json_rfx_container_label,
      json_rfx_mapper_label,
      json_rfx_groupids_label
    ) {
      # Append RFX objects
      rfx_container_append_from_json_string_cpp(
        self$rfx_container_ptr,
        json_string,
        json_rfx_container_label
      )
    },

    #' @description
    #' Predict random effects for each observation implied by `rfx_group_ids` and `rfx_basis`.
    #' If a random effects model is "intercept-only" the `rfx_basis` will be a vector of ones of size `length(rfx_group_ids)`.
    #' @param rfx_group_ids Indices of random effects groups in a prediction set
    #' @param rfx_basis (Optional) Basis used for random effects prediction
    #' @return Matrix with as many rows as observations provided and as many columns as samples drawn of the model.
    predict = function(rfx_group_ids, rfx_basis = NULL) {
      num_obs = length(rfx_group_ids)
      if (is.null(rfx_basis)) {
        rfx_basis <- matrix(rep(1, num_obs), ncol = 1)
      }
      num_samples = rfx_container_num_samples_cpp(self$rfx_container_ptr)
      num_components = rfx_container_num_components_cpp(
        self$rfx_container_ptr
      )
      num_groups = rfx_container_num_groups_cpp(self$rfx_container_ptr)
      rfx_group_ids_int <- as.integer(rfx_group_ids)
      stopifnot(sum(abs(rfx_group_ids_int - rfx_group_ids)) < 1e-6)
      stopifnot(sum(!(rfx_group_ids %in% self$training_group_ids)) == 0)
      stopifnot(ncol(rfx_basis) == num_components)
      rfx_dataset <- createRandomEffectsDataset(
        rfx_group_ids_int,
        rfx_basis
      )
      output <- rfx_container_predict_cpp(
        self$rfx_container_ptr,
        rfx_dataset$data_ptr,
        self$label_mapper_ptr
      )
      dim(output) <- c(num_obs, num_samples)
      return(output)
    },

    #' @description
    #' Extract the random effects parameters sampled. With the "redundant parameterization"
    #' of Gelman et al (2008), this includes four parameters: alpha (the "working parameter"
    #' shared across every group), xi (the "group parameter" sampled separately for each group),
    #' beta (the product of alpha and xi, which corresponds to the overall group-level random effects),
    #' and sigma (group-independent prior variance for each component of xi).
    #' @return List of arrays. The alpha array has dimension (`num_components`, `num_samples`) and is simply a vector if `num_components = 1`.
    #' The xi and beta arrays have dimension (`num_components`, `num_groups`, `num_samples`) and are simply matrices if `num_components = 1`.
    #' The sigma array has dimension (`num_components`, `num_samples`) and is simply a vector if `num_components = 1`.
    extract_parameter_samples = function() {
      num_samples = rfx_container_num_samples_cpp(self$rfx_container_ptr)
      num_components = rfx_container_num_components_cpp(
        self$rfx_container_ptr
      )
      num_groups = rfx_container_num_groups_cpp(self$rfx_container_ptr)
      beta_samples <- rfx_container_get_beta_cpp(self$rfx_container_ptr)
      xi_samples <- rfx_container_get_xi_cpp(self$rfx_container_ptr)
      alpha_samples <- rfx_container_get_alpha_cpp(self$rfx_container_ptr)
      sigma_samples <- rfx_container_get_sigma_cpp(self$rfx_container_ptr)
      if (num_components == 1) {
        dim(beta_samples) <- c(num_groups, num_samples)
        dim(xi_samples) <- c(num_groups, num_samples)
      } else if (num_components > 1) {
        dim(beta_samples) <- c(num_components, num_groups, num_samples)
        dim(xi_samples) <- c(num_components, num_groups, num_samples)
        dim(alpha_samples) <- c(num_components, num_samples)
        dim(sigma_samples) <- c(num_components, num_samples)
      } else {
        stop(
          "Invalid random effects sample container, num_components is less than 1"
        )
      }

      output = list(
        "beta_samples" = beta_samples,
        "xi_samples" = xi_samples,
        "alpha_samples" = alpha_samples,
        "sigma_samples" = sigma_samples
      )
      return(output)
    },

    #' @description
    #' Modify the `RandomEffectsSamples` object by removing the parameter samples index by `sample_num`.
    #' @param sample_num Index of the RFX sample to be removed
    delete_sample = function(sample_num) {
      rfx_container_delete_sample_cpp(self$rfx_container_ptr, sample_num)
    },

    #' @description
    #' Convert the mapping of group IDs to random effect components indices from C++ to R native format
    #' @return List mapping group ID to random effect components.
    extract_label_mapping = function() {
      keys_and_vals <- rfx_label_mapper_to_list_cpp(self$label_mapper_ptr)
      result <- as.list(keys_and_vals[[2]] + 1)
      setNames(result, keys_and_vals[[1]])
      return(result)
    },

    #' @description
    #' Query the number of samples in the `RandomEffectsSamples` object.
    #' @return Integer number of samples
    num_samples = function() {
      return(rfx_container_num_samples_cpp(self$rfx_container_ptr))
    },

    #' @description
    #' Query the number of components in the `RandomEffectsSamples` object.
    #' @return Integer number of components
    num_components = function() {
      return(rfx_container_num_components_cpp(self$rfx_container_ptr))
    },

    #' @description
    #' Query the number of groups in the `RandomEffectsSamples` object.
    #' @return Integer number of groups
    num_groups = function() {
      return(rfx_container_num_groups_cpp(self$rfx_container_ptr))
    }
  )
)

#' @title Random Effects Tracker C++ Wrapper
#'
#' @description
#' Class that defines a "tracker" for random effects models, most notably
#' storing the data indices available in each group for quicker posterior
#' computation and sampling of random effects terms.
#' The class stores a mapping from every observation to its group index, a mapping
#' from group indices to the training sample observations available in that
#' group, and predictions for each observation.
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

RandomEffectsTracker <- R6::R6Class(
  classname = "RandomEffectsTracker",
  cloneable = FALSE,
  public = list(
    #' @field rfx_tracker_ptr External pointer to a C++ StochTree::RandomEffectsTracker class
    rfx_tracker_ptr = NULL,

    #' @description
    #' Create a new RandomEffectsTracker object.
    #' @param rfx_group_indices Integer indices indicating groups used to define random effects
    #' @return A new `RandomEffectsTracker` object.
    initialize = function(rfx_group_indices) {
      # Initialize
      self$rfx_tracker_ptr <- rfx_tracker_cpp(rfx_group_indices)
    }
  )
)

#' @title Random Effects Model C++ Wrapper
#' @description
#' The core "model" class for sampling random effects.
#' Stores current model state, prior parameters, and procedures for
#' sampling from the conditional posterior of each parameter.
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

RandomEffectsModel <- R6::R6Class(
  classname = "RandomEffectsModel",
  cloneable = FALSE,
  public = list(
    #' @field rfx_model_ptr External pointer to a C++ StochTree::RandomEffectsModel class
    rfx_model_ptr = NULL,

    #' @field num_groups Number of groups in the random effects model
    num_groups = NULL,

    #' @field num_components Number of components (i.e. dimension of basis) in the random effects model
    num_components = NULL,

    #' @description
    #' Create a new RandomEffectsModel object.
    #' @param num_components Number of "components" or bases defining the random effects regression
    #' @param num_groups Number of random effects groups
    #' @return A new `RandomEffectsModel` object.
    initialize = function(num_components, num_groups) {
      # Initialize
      self$rfx_model_ptr <- rfx_model_cpp(num_components, num_groups)
      self$num_components <- num_components
      self$num_groups <- num_groups
    },

    #' @description
    #' Sample from random effects model.
    #' @param rfx_dataset Object of type `RandomEffectsDataset`
    #' @param residual Object of type `Outcome`
    #' @param rfx_tracker Object of type `RandomEffectsTracker`
    #' @param rfx_samples Object of type `RandomEffectSamples`
    #' @param keep_sample Whether sample should be retained in `rfx_samples`. If `FALSE`, the state of `rfx_tracker` will be updated, but the parameter values will not be added to the sample container. Samples are commonly discarded due to burn-in or thinning.
    #' @param global_variance Scalar global variance parameter
    #' @param rng Object of type `CppRNG`
    #' @return None
    sample_random_effect = function(
      rfx_dataset,
      residual,
      rfx_tracker,
      rfx_samples,
      keep_sample,
      global_variance,
      rng
    ) {
      rfx_model_sample_random_effects_cpp(
        self$rfx_model_ptr,
        rfx_dataset$data_ptr,
        residual$data_ptr,
        rfx_tracker$rfx_tracker_ptr,
        rfx_samples$rfx_container_ptr,
        keep_sample,
        global_variance,
        rng$rng_ptr
      )
    },

    #' @description
    #' Predict from (a single sample of a) random effects model.
    #' @param rfx_dataset Object of type `RandomEffectsDataset`
    #' @param rfx_tracker Object of type `RandomEffectsTracker`
    #' @return Vector of predictions with size matching number of observations in rfx_dataset
    predict = function(rfx_dataset, rfx_tracker) {
      pred <- rfx_model_predict_cpp(
        self$rfx_model_ptr,
        rfx_dataset$data_ptr,
        rfx_tracker$rfx_tracker_ptr
      )
      return(pred)
    },

    #' @description
    #' Set value for the "working parameter." This is typically
    #' used for initialization, but could also be used to interrupt
    #' or override the sampler.
    #' @param value Parameter input
    #' @return None
    set_working_parameter = function(value) {
      stopifnot(is.double(value))
      stopifnot(!is.matrix(value))
      stopifnot(length(value) == self$num_components)
      rfx_model_set_working_parameter_cpp(self$rfx_model_ptr, value)
    },

    #' @description
    #' Set value for the "group parameters." This is typically
    #' used for initialization, but could also be used to interrupt
    #' or override the sampler.
    #' @param value Parameter input
    #' @return None
    set_group_parameters = function(value) {
      stopifnot(is.double(value))
      stopifnot(is.matrix(value))
      stopifnot(nrow(value) == self$num_components)
      stopifnot(ncol(value) == self$num_groups)
      rfx_model_set_group_parameters_cpp(self$rfx_model_ptr, value)
    },

    #' @description
    #' Set value for the working parameter covariance. This is typically
    #' used for initialization, but could also be used to interrupt
    #' or override the sampler.
    #' @param value Parameter input
    #' @return None
    set_working_parameter_cov = function(value) {
      stopifnot(is.double(value))
      stopifnot(is.matrix(value))
      stopifnot(nrow(value) == self$num_components)
      stopifnot(ncol(value) == self$num_components)
      rfx_model_set_working_parameter_covariance_cpp(
        self$rfx_model_ptr,
        value
      )
    },

    #' @description
    #' Set value for the group parameter covariance. This is typically
    #' used for initialization, but could also be used to interrupt
    #' or override the sampler.
    #' @param value Parameter input
    #' @return None
    set_group_parameter_cov = function(value) {
      stopifnot(is.double(value))
      stopifnot(is.matrix(value))
      stopifnot(nrow(value) == self$num_components)
      stopifnot(ncol(value) == self$num_components)
      rfx_model_set_group_parameter_covariance_cpp(
        self$rfx_model_ptr,
        value
      )
    },

    #' @description
    #' Set shape parameter for the group parameter variance prior.
    #' @param value Parameter input
    #' @return None
    set_variance_prior_shape = function(value) {
      stopifnot(is.double(value))
      stopifnot(!is.matrix(value))
      stopifnot(length(value) == 1)
      rfx_model_set_variance_prior_shape_cpp(self$rfx_model_ptr, value)
    },

    #' @description
    #' Set shape parameter for the group parameter variance prior.
    #' @param value Parameter input
    #' @return None
    set_variance_prior_scale = function(value) {
      stopifnot(is.double(value))
      stopifnot(!is.matrix(value))
      stopifnot(length(value) == 1)
      rfx_model_set_variance_prior_scale_cpp(self$rfx_model_ptr, value)
    }
  )
)

#' @title Create RandomEffectSamples Object
#' @description
#' Create a `RandomEffectSamples` object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param num_components Number of "components" or bases defining the random effects regression
#' @param num_groups Number of random effects groups
#' @param random_effects_tracker Object of type `RandomEffectsTracker`
#' @return `RandomEffectSamples` object
#' @export
#'
#' @examples
#' n <- 100
#' rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
#' rfx_basis <- matrix(rep(1.0, n), ncol=1)
#' num_groups <- length(unique(rfx_group_ids))
#' num_components <- ncol(rfx_basis)
#' rfx_tracker <- createRandomEffectsTracker(rfx_group_ids)
#' rfx_samples <- createRandomEffectSamples(num_components, num_groups, rfx_tracker)
createRandomEffectSamples <- function(
  num_components,
  num_groups,
  random_effects_tracker
) {
  invisible(output <- RandomEffectSamples$new())
  output$load_in_session(num_components, num_groups, random_effects_tracker)
  return(output)
}

#' @title Create RandomEffectsTracker Object
#' @description
#' Create a `RandomEffectsTracker` object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param rfx_group_indices Integer indices indicating groups used to define random effects
#' @return `RandomEffectsTracker` object
#' @export
#'
#' @examples
#' n <- 100
#' rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
#' rfx_basis <- matrix(rep(1.0, n), ncol=1)
#' num_groups <- length(unique(rfx_group_ids))
#' num_components <- ncol(rfx_basis)
#' rfx_tracker <- createRandomEffectsTracker(rfx_group_ids)
createRandomEffectsTracker <- function(rfx_group_indices) {
  return(invisible((RandomEffectsTracker$new(rfx_group_indices))))
}

#' @title Create RandomEffectsModel Object
#' @description
#' Create a `RandomEffectsModel` object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param num_components Number of "components" or bases defining the random effects regression
#' @param num_groups Number of random effects groups
#' @return `RandomEffectsModel` object
#' @export
#'
#' @examples
#' n <- 100
#' rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
#' rfx_basis <- matrix(rep(1.0, n), ncol=1)
#' num_groups <- length(unique(rfx_group_ids))
#' num_components <- ncol(rfx_basis)
#' rfx_model <- createRandomEffectsModel(num_components, num_groups)
createRandomEffectsModel <- function(num_components, num_groups) {
  return(invisible((RandomEffectsModel$new(num_components, num_groups))))
}

#' @title Reset RandomEffectsModel Object
#' @rdname RandomEffectStateManagement
#' @param rfx_model Object of type `RandomEffectsModel`.
#' @param rfx_samples Object of type `RandomEffectSamples`.
#' @param sample_num Index of sample stored in `rfx_samples` from which to reset the state of a random effects model. Zero-indexed, so resetting based on the first sample would require setting `sample_num = 0`.
#' @param sigma_alpha_init Initial value of the "working parameter" scale parameter.
#' @export
resetRandomEffectsModel <- function(
  rfx_model,
  rfx_samples,
  sample_num,
  sigma_alpha_init
) {
  if (!is.matrix(sigma_alpha_init)) {
    if (!is.double(sigma_alpha_init)) {
      stop("`sigma_alpha_init` must be a numeric scalar or matrix")
    }
    sigma_alpha_init <- as.matrix(sigma_alpha_init)
  }
  reset_rfx_model_cpp(
    rfx_model$rfx_model_ptr,
    rfx_samples$rfx_container_ptr,
    sample_num
  )
  rfx_model$set_working_parameter_cov(sigma_alpha_init)
}

#' @title Reset RandomEffectsTracker Object
#' @rdname RandomEffectStateManagement
#' @param rfx_tracker Object of type `RandomEffectsTracker`.
#' @param rfx_model Object of type `RandomEffectsModel`.
#' @param rfx_dataset Object of type `RandomEffectsDataset`.
#' @param residual Object of type `Outcome`.
#' @param rfx_samples Object of type `RandomEffectSamples`.
#' @export
resetRandomEffectsTracker <- function(
  rfx_tracker,
  rfx_model,
  rfx_dataset,
  residual,
  rfx_samples
) {
  reset_rfx_tracker_cpp(
    rfx_tracker$rfx_tracker_ptr,
    rfx_dataset$data_ptr,
    residual$data_ptr,
    rfx_model$rfx_model_ptr
  )
}

#' @title Reset RandomEffectsModel Object to Default State
#' @rdname RandomEffectStateManagement
#' @param rfx_model Object of type `RandomEffectsModel`.
#' @param alpha_init Initial value of the "working parameter".
#' @param xi_init Initial value of the "group parameters".
#' @param sigma_alpha_init Initial value of the "working parameter" scale parameter.
#' @param sigma_xi_init Initial value of the "group parameters" scale parameter.
#' @param sigma_xi_shape Shape parameter for the inverse gamma variance model on the group parameters.
#' @param sigma_xi_scale Scale parameter for the inverse gamma variance model on the group parameters.
#' @export
rootResetRandomEffectsModel <- function(
  rfx_model,
  alpha_init,
  xi_init,
  sigma_alpha_init,
  sigma_xi_init,
  sigma_xi_shape,
  sigma_xi_scale
) {
  rfx_model$set_working_parameter(alpha_init)
  rfx_model$set_group_parameters(xi_init)
  rfx_model$set_working_parameter_cov(sigma_alpha_init)
  rfx_model$set_group_parameter_cov(sigma_xi_init)
  rfx_model$set_variance_prior_shape(sigma_xi_shape)
  rfx_model$set_variance_prior_scale(sigma_xi_scale)
}

#' @title Reset RandomEffectsTracker Object to Default State
#' @rdname RandomEffectStateManagement
#' @param rfx_tracker Object of type `RandomEffectsTracker`.
#' @param rfx_model Object of type `RandomEffectsModel`.
#' @param rfx_dataset Object of type `RandomEffectsDataset`.
#' @param residual Object of type `Outcome`.
#' @export
rootResetRandomEffectsTracker <- function(
  rfx_tracker,
  rfx_model,
  rfx_dataset,
  residual
) {
  root_reset_rfx_tracker_cpp(
    rfx_tracker$rfx_tracker_ptr,
    rfx_dataset$data_ptr,
    residual$data_ptr,
    rfx_model$rfx_model_ptr
  )
}

#' @title Summarize a RandomEffectSamples Object
#' @description Prints a summary of the RandomEffectSamples object, including number of forests and the underlying model of each forest.
#' @param x RandomEffectSamples object
#' @param ... Additional arguments
#' @export
#' @return RandomEffectSamples object unchanged after printing summary
print.RandomEffectSamples <- function(x, ...) {
  # Construct summary message
  num_samples <- x$num_samples()
  num_components <- x$num_components()
  num_groups <- x$num_groups()
  summary_message <- paste0(
    "Random effects container with ",
    num_samples,
    " posterior draws of a model with ",
    num_components,
    " components and ",
    num_groups,
    " groups."
  )

  # Print the random effects container details
  cat(summary_message, "\n")

  # Return random effects container invisibly
  invisible(x)
}

test_that("MCMC BCF", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) *
    (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) *
    (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) *
    (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # 1 chain, no thinning
  general_param_list <- list(num_chains = 1, keep_every = 1)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 1 chain, no thinning, matrix leaf scale parameter provided
  general_param_list <- list(num_chains = 1, keep_every = 1)
  mu_forest_param_list <- list(sigma2_leaf_init = as.matrix(0.5))
  tau_forest_param_list <- list(sigma2_leaf_init = as.matrix(0.5))
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      prognostic_forest_params = mu_forest_param_list,
      treatment_effect_forest_params = tau_forest_param_list
    )
  )

  # 1 chain, no thinning, scalar leaf scale parameter provided
  general_param_list <- list(num_chains = 1, keep_every = 1)
  mu_forest_param_list <- list(sigma2_leaf_init = 0.5)
  tau_forest_param_list <- list(sigma2_leaf_init = 0.5)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      prognostic_forest_params = mu_forest_param_list,
      treatment_effect_forest_params = tau_forest_param_list
    )
  )

  # 3 chains, no thinning
  general_param_list <- list(num_chains = 3, keep_every = 1)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 1 chain, thinning
  general_param_list <- list(num_chains = 1, keep_every = 5)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 3 chains, thinning
  general_param_list <- list(num_chains = 3, keep_every = 5)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )
})

test_that("GFR BCF", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) *
    (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) *
    (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) *
    (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # 1 chain, no thinning
  general_param_list <- list(num_chains = 1, keep_every = 1)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 3 chains, no thinning
  general_param_list <- list(num_chains = 3, keep_every = 1)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 1 chain, thinning
  general_param_list <- list(num_chains = 1, keep_every = 5)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 3 chains, thinning
  general_param_list <- list(num_chains = 3, keep_every = 5)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check for error when more chains than GFR forests
  general_param_list <- list(num_chains = 11, keep_every = 1)
  expect_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check for error when more chains than GFR forests
  general_param_list <- list(num_chains = 11, keep_every = 5)
  expect_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )
})

test_that("Warmstart BCF", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  # fmt: skip
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  # fmt: skip
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Run a BCF model with only GFR
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bcf_model <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 0,
    general_params = general_param_list
  )

  # Save to JSON string
  bcf_model_json_string <- saveBCFModelToJsonString(bcf_model)

  # Run a new BCF chain from the existing (X)BCF model
  general_param_list <- list(num_chains = 3, keep_every = 5)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      previous_model_json = bcf_model_json_string,
      previous_model_warmstart_sample_num = 10,
      general_params = general_param_list
    )
  )
  expect_warning(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      previous_model_json = bcf_model_json_string,
      previous_model_warmstart_sample_num = 1,
      general_params = general_param_list
    )
  )

  # Generate simulated data with random effects
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  # fmt: skip
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  # fmt: skip
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
  rfx_basis <- rep(1, n)
  rfx_coefs <- c(-5, 5)
  rfx_term <- rfx_coefs[rfx_group_ids] * rfx_basis
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rfx_term + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds]
  rfx_basis_train <- rfx_basis[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Run a BCF model with only GFR
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bcf_model <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    rfx_group_ids_train = rfx_group_ids_train,
    rfx_group_ids_test = rfx_group_ids_test,
    rfx_basis_train = rfx_basis_train,
    rfx_basis_test = rfx_basis_test,
    propensity_test = pi_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 0,
    general_params = general_param_list
  )

  # Save to JSON string
  bcf_model_json_string <- saveBCFModelToJsonString(bcf_model)

  # Run a new BCF chain from the existing (X)BCF model
  general_param_list <- list(num_chains = 3, keep_every = 5)
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      previous_model_json = bcf_model_json_string,
      previous_model_warmstart_sample_num = 10,
      general_params = general_param_list
    )
  )
})

test_that("Multivariate Treatment MCMC BCF", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  # fmt: skip
  pi_X_1 <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  # fmt: skip
  pi_X_2 <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.8) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (0.4) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (0.6) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (0.2))
  pi_X <- cbind(pi_X_1, pi_X_2)
  # fmt: skip
  tau_X_1 <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  # fmt: skip
  tau_X_2 <- (((0 <= X[, 3]) & (0.25 > X[, 3])) * (-0.5) +
    ((0.25 <= X[, 3]) & (0.5 > X[, 3])) * (-1.5) +
    ((0.5 <= X[, 3]) & (0.75 > X[, 3])) * (-1.0) +
    ((0.75 <= X[, 3]) & (1 > X[, 3])) * (0.0))
  tau_X <- cbind(tau_X_1, tau_X_2)
  Z_1 <- as.numeric(rbinom(n, 1, pi_X_1))
  Z_2 <- as.numeric(rbinom(n, 1, pi_X_2))
  Z <- cbind(Z_1, Z_2)
  noise_sd <- 1
  y <- mu_X + rowSums(tau_X * Z) + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds, ]
  Z_train <- Z[train_inds, ]
  pi_test <- pi_X[test_inds, ]
  pi_train <- pi_X[train_inds, ]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds, ]
  tau_train <- tau_X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # 1 chain, no thinning
  general_param_list <- list(
    num_chains = 1,
    keep_every = 1,
    adaptive_coding = F
  )
  expect_no_error({
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
    predict(bcf_model, X = X_test, Z = Z_test, propensity = pi_test)
  })
})

test_that("BCF Predictions", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  # fmt: skip
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  # fmt: skip
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Run a BCF model with only GFR
  general_params <- list(num_chains = 1, keep_every = 1)
  variance_forest_params <- list(num_trees = 50)
  expect_warning(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      general_params = general_params,
      variance_forest_params = variance_forest_params
    )
  )

  # Check that cached predictions agree with results of predict() function
  train_preds <- predict(
    bcf_model,
    X = X_train,
    Z = Z_train,
    propensity = pi_train
  )
  train_preds_mean_cached <- bcf_model$y_hat_train
  train_preds_mean_recomputed <- train_preds$y_hat
  train_preds_variance_cached <- bcf_model$sigma2_x_hat_train
  train_preds_variance_recomputed <- train_preds$variance_forest_predictions

  # Assertion
  expect_equal(train_preds_mean_cached, train_preds_mean_recomputed)
  expect_equal(train_preds_variance_cached, train_preds_variance_recomputed)
})

test_that("Random Effects BCF", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  # fmt: skip
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  # fmt: skip
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
  rfx_basis <- cbind(rep(1, n), runif(n))
  num_rfx_components <- ncol(rfx_basis)
  num_rfx_groups <- length(unique(rfx_group_ids))
  rfx_coefs <- matrix(c(-5, 5, 1, -1), ncol = 2, byrow = T)
  rfx_term <- rowSums(rfx_coefs[rfx_group_ids, ] * rfx_basis)
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rfx_term + rnorm(n, 0, noise_sd)

  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Specify no rfx parameters
  general_param_list <- list()
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_basis_train = rfx_basis_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Specify all rfx parameters as scalars
  rfx_param_list <- list(
    working_parameter_prior_mean = 1.,
    group_parameter_prior_mean = 1.,
    working_parameter_prior_cov = 1.,
    group_parameter_prior_cov = 1.,
    variance_prior_shape = 1,
    variance_prior_scale = 1
  )
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_basis_train = rfx_basis_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      random_effects_params = rfx_param_list
    )
  )

  # Specify all relevant rfx parameters as vectors
  rfx_param_list <- list(
    working_parameter_prior_mean = c(1., 1.),
    group_parameter_prior_mean = c(1., 1.),
    working_parameter_prior_cov = diag(1., 2),
    group_parameter_prior_cov = diag(1., 2),
    variance_prior_shape = 1,
    variance_prior_scale = 1
  )
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_basis_train = rfx_basis_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      random_effects_params = rfx_param_list
    )
  )
})

test_that("BCF internal propensity model works with data frame covariates", {
  skip_on_cran()

  # X_test_raw was incorrectly passed to the internal BART propensity model instead of the preprocessed X_test,
  # causing "undefined columns selected" when X contained factor columns.
  set.seed(42)
  n <- 100
  p <- 4
  X_num <- matrix(runif(n * p), ncol = p)
  # Add a factor column — triggers preprocessTrainData to one-hot encode,
  # changing the column structure and exposing the raw/processed mismatch
  X_cat <- factor(sample(c("a", "b", "c"), n, replace = TRUE))
  X <- data.frame(X_num, cat = X_cat)

  pi_X <- 0.25 + 0.5 * X_num[, 1]
  Z <- rbinom(n, 1, pi_X)
  mu_X <- X_num[, 1] * 2
  tau_X <- X_num[, 2]
  y <- mu_X + tau_X * Z + rnorm(n)

  test_inds <- 1:20
  train_inds <- 21:n
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]
  Z_test <- Z[test_inds]
  y_train <- y[train_inds]

  # No propensity_train provided — triggers internal BART propensity model.
  # Before the fix, this raised "undefined columns selected" because X_test_raw
  # (unprocessed) was passed instead of preprocessed X_test.
  expect_no_error(
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      X_test = X_test,
      Z_test = Z_test,
      num_gfr = 5,
      num_burnin = 0,
      num_mcmc = 5
    )
  )
  expect_true(!is.null(bcf_model$tau_hat_train))
  expect_true(!is.null(bcf_model$tau_hat_test))
})

test_that("BCF JSON serialization roundtrip covers all deserialization paths", {
  skip_on_cran()

  # Generate simulated data
  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  pi_x <- 0.25 + 0.5 * X[, 1]
  mu_x <- pi_x * 5
  tau_x <- X[, 2] * 2
  Z <- rbinom(n, 1, pi_x)
  y <- mu_x + Z * tau_x + rnorm(n, 0, 1)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  pi_test <- pi_x[test_inds]
  pi_train <- pi_x[train_inds]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  y_train <- y[train_inds]

  # Fit model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = 10
  )
  preds_orig <- predict(bcf_model, X_test, Z_test, pi_test)
  y_hat_orig <- rowMeans(preds_orig[["y_hat"]])
  tau_hat_orig <- rowMeans(preds_orig[["tau_hat"]])

  # Path 1: in-memory JSON object
  bcf_json <- saveBCFModelToJson(bcf_model)
  bcf_rt <- createBCFModelFromJson(bcf_json)
  preds_rt <- predict(bcf_rt, X_test, Z_test, pi_test)
  expect_equal(rowMeans(preds_rt[["y_hat"]]), y_hat_orig)
  expect_equal(rowMeans(preds_rt[["tau_hat"]]), tau_hat_orig)

  # Path 2: JSON string
  bcf_json_string <- saveBCFModelToJsonString(bcf_model)
  bcf_rt <- createBCFModelFromJsonString(bcf_json_string)
  preds_rt <- predict(bcf_rt, X_test, Z_test, pi_test)
  expect_equal(rowMeans(preds_rt[["y_hat"]]), y_hat_orig)
  expect_equal(rowMeans(preds_rt[["tau_hat"]]), tau_hat_orig)

  # Path 3: JSON file
  tmpjson <- tempfile(fileext = ".json")
  saveBCFModelToJsonFile(bcf_model, tmpjson)
  bcf_rt <- createBCFModelFromJsonFile(tmpjson)
  unlink(tmpjson)
  preds_rt <- predict(bcf_rt, X_test, Z_test, pi_test)
  expect_equal(rowMeans(preds_rt[["y_hat"]]), y_hat_orig)
  expect_equal(rowMeans(preds_rt[["tau_hat"]]), tau_hat_orig)

  # Path 4: list of in-memory JSON objects (combined)
  bcf_rt <- createBCFModelFromCombinedJson(list(bcf_json))
  preds_rt <- predict(bcf_rt, X_test, Z_test, pi_test)
  expect_equal(rowMeans(preds_rt[["y_hat"]]), y_hat_orig)
  expect_equal(rowMeans(preds_rt[["tau_hat"]]), tau_hat_orig)

  # Path 5: list of JSON strings (combined)
  bcf_rt <- createBCFModelFromCombinedJsonString(list(bcf_json_string))
  preds_rt <- predict(bcf_rt, X_test, Z_test, pi_test)
  expect_equal(rowMeans(preds_rt[["y_hat"]]), y_hat_orig)
  expect_equal(rowMeans(preds_rt[["tau_hat"]]), tau_hat_orig)
})

test_that("BCF factor-valued treatment handling", {
  skip_on_cran()

  # Shared data: binary treatment DGP
  n <- 100
  p <- 5
  set.seed(42)
  X <- matrix(runif(n * p), ncol = p)
  pi_X <- 0.4 + 0.2 * X[, 1]
  Z_numeric <- rbinom(n, 1, pi_X)
  tau_X <- 1 + X[, 2]
  mu_X <- 2 * X[, 3]
  y <- mu_X + tau_X * Z_numeric + rnorm(n, 0, 1)

  # Binary factor treatment: levels "0" and "1"
  # Verify the conversion produces 0/1 values identical to the original
  Z_factor_binary <- factor(Z_numeric)
  expect_equal(levels(Z_factor_binary), c("0", "1"))
  expect_equal(as.integer(Z_factor_binary) - 1L, as.integer(Z_numeric))

  # Factor treatment should run without error and emit an informative message
  expect_message(
    suppressWarnings(bcf(
      X_train = X, y_train = y, Z_train = Z_factor_binary,
      propensity_train = pi_X, num_gfr = 0, num_burnin = 5, num_mcmc = 5
    )),
    regexp = "Z_train is a factor"
  )

  # Logical treatment converted to factor: levels "FALSE" and "TRUE"
  # as.factor(logical) sorts alphabetically: "FALSE" = 0, "TRUE" = 1
  Z_logical <- as.logical(Z_numeric)
  Z_factor_logical <- as.factor(Z_logical)
  expect_equal(levels(Z_factor_logical), c("FALSE", "TRUE"))
  expect_equal(as.integer(Z_factor_logical) - 1L, as.integer(Z_numeric))

  expect_message(
    suppressWarnings(bcf(
      X_train = X, y_train = y, Z_train = Z_factor_logical,
      propensity_train = pi_X, num_gfr = 0, num_burnin = 5, num_mcmc = 5
    )),
    regexp = "Z_train is a factor"
  )

  # Factor treatment with more than 2 levels should error immediately
  Z_factor_categorical <- factor(sample(c("A", "B", "C"), n, replace = TRUE))
  expect_error(
    bcf(
      X_train = X, y_train = y, Z_train = Z_factor_categorical,
      propensity_train = pi_X, num_gfr = 0, num_burnin = 5, num_mcmc = 5
    ),
    regexp = "exactly 2 levels"
  )

  # predict.bcfmodel should also handle factor Z, raising a warning
  suppressMessages(
    bcf_model <- bcf(
      X_train = X, y_train = y, Z_train = Z_numeric,
      propensity_train = pi_X, num_gfr = 0, num_burnin = 5, num_mcmc = 5
    )
  )
  expect_warning(
    predict(bcf_model, X, Z_factor_binary, pi_X),
    regexp = "Z is a factor"
  )
  expect_warning(
    predict(bcf_model, X, Z_factor_logical, pi_X),
    regexp = "Z is a factor"
  )
  expect_error(
    predict(bcf_model, X, Z_factor_categorical, pi_X),
    regexp = "exactly 2 levels"
  )
})

test_that("Warmstart BCF reuses internal propensity model", {
  skip_on_cran()

  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  pi_X <- 0.25 + 0.5 * X[, 1]
  Z <- rbinom(n, 1, pi_X)
  y <- pi_X * 3 + X[, 2] * Z + rnorm(n, 0, 1)
  n_test <- 20
  n_train <- n - n_test
  train_inds <- 1:n_train
  test_inds <- (n_train + 1):n

  X_train <- X[train_inds, ]
  X_test  <- X[test_inds, ]
  Z_train <- Z[train_inds]
  Z_test  <- Z[test_inds]
  y_train <- y[train_inds]

  # Fit first model without propensity — triggers internal propensity BART
  m1 <- bcf(
    X_train = X_train, Z_train = Z_train, y_train = y_train,
    X_test = X_test, Z_test = Z_test,
    num_gfr = 5, num_burnin = 0, num_mcmc = 10
  )
  expect_true(m1$model_params$internal_propensity_model)

  # Propensity predictions from the first model's propensity BART
  pi_train_m1 <- predict(m1$bart_propensity_model, X = X_train, terms = "y_hat", type = "mean")

  # Warm-start second model from first — propensity model should be reused
  m1_json <- saveBCFModelToJsonString(m1)
  m2 <- bcf(
    X_train = X_train, Z_train = Z_train, y_train = y_train,
    X_test = X_test, Z_test = Z_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = 10,
    previous_model_json = m1_json,
    previous_model_warmstart_sample_num = 10L
  )
  expect_true(m2$model_params$internal_propensity_model)

  # Propensity model reused: predictions on train set should be identical
  pi_train_m2 <- predict(m2$bart_propensity_model, X = X_train, terms = "y_hat", type = "mean")
  expect_equal(pi_train_m1, pi_train_m2)

  # Output shapes should be correct
  expect_equal(dim(m2$y_hat_train), c(n_train, 10))
  expect_equal(dim(m2$y_hat_test),  c(n_test, 10))
})

# Tests for multi-chain BART and BCF sampling.
#
# Covers sample-count correctness, GFR warm-start path, chain independence,
# extractParameter dimensions, serialization round-trip, and the
# num_gfr >= num_chains validation.

# ---------------------------------------------------------------------------
# Shared test data helpers
# ---------------------------------------------------------------------------

.make_bart_data <- function() {
  set.seed(42)
  n <- 200; p <- 5
  X <- matrix(runif(n * p), ncol = p)
  y <- 5 * X[, 1] + rnorm(n)
  test_inds  <- sort(sample(1:n, 40))
  train_inds <- setdiff(1:n, test_inds)
  list(
    X_train  = X[train_inds, ],
    X_test   = X[test_inds,  ],
    y_train  = y[train_inds],
    n_train  = length(train_inds),
    n_test   = length(test_inds)
  )
}

.make_bcf_data <- function() {
  set.seed(42)
  n <- 200; p <- 5
  X  <- matrix(runif(n * p), ncol = p)
  pi_X <- 0.25 + 0.5 * X[, 1]
  Z  <- rbinom(n, 1, pi_X)
  y  <- 5 * X[, 1] + 2 * X[, 2] * Z + rnorm(n)
  test_inds  <- sort(sample(1:n, 40))
  train_inds <- setdiff(1:n, test_inds)
  list(
    X_train   = X[train_inds, ],
    X_test    = X[test_inds,  ],
    Z_train   = Z[train_inds],
    Z_test    = Z[test_inds],
    y_train   = y[train_inds],
    pi_train  = pi_X[train_inds],
    pi_test   = pi_X[test_inds],
    n_train   = length(train_inds),
    n_test    = length(test_inds)
  )
}

# ---------------------------------------------------------------------------
# BARTModel multi-chain tests
# ---------------------------------------------------------------------------

test_that("BART multi-chain: sample counts with no GFR", {
  skip_on_cran()
  d <- .make_bart_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected <- n_chains * n_mcmc
  expect_length(m$sigma2_global_samples, expected)
  expect_equal(dim(m$y_hat_train), c(d$n_train, expected))
  expect_equal(dim(m$y_hat_test),  c(d$n_test,  expected))
})

test_that("BART multi-chain: sample counts with GFR warm-start", {
  skip_on_cran()
  d <- .make_bart_data()
  n_chains <- 3; n_mcmc <- 10; n_gfr <- 6
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = n_gfr, num_burnin = 5, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected <- n_chains * n_mcmc
  expect_length(m$sigma2_global_samples, expected)
  expect_equal(dim(m$y_hat_train), c(d$n_train, expected))
})

test_that("BART multi-chain: leaf-scale sample count", {
  skip_on_cran()
  d <- .make_bart_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = n_mcmc,
    general_params = list(
      num_chains = n_chains, num_threads = 1,
      sample_sigma2_global = FALSE
    ),
    mean_forest_params = list(sample_sigma2_leaf = TRUE)
  )
  expect_length(m$sigma2_leaf_samples, n_chains * n_mcmc)
})

test_that("BART multi-chain: chain independence (no GFR)", {
  skip_on_cran()
  d <- .make_bart_data()
  n_mcmc <- 10
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = n_mcmc,
    general_params = list(num_chains = 2, num_threads = 1)
  )
  chain1 <- m$sigma2_global_samples[seq_len(n_mcmc)]
  chain2 <- m$sigma2_global_samples[seq(n_mcmc + 1, 2 * n_mcmc)]
  expect_false(isTRUE(all.equal(chain1, chain2)),
               label = "Chains should produce distinct sigma2 samples")
})

test_that("BART multi-chain: chain independence (with GFR)", {
  skip_on_cran()
  d <- .make_bart_data()
  n_mcmc <- 10; n_gfr <- 4
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = n_gfr, num_burnin = 5, num_mcmc = n_mcmc,
    general_params = list(num_chains = 2, num_threads = 1)
  )
  chain1 <- m$sigma2_global_samples[seq_len(n_mcmc)]
  chain2 <- m$sigma2_global_samples[seq(n_mcmc + 1, 2 * n_mcmc)]
  expect_false(isTRUE(all.equal(chain1, chain2)))
})

test_that("BART multi-chain: extractParameter dimensions", {
  skip_on_cran()
  d <- .make_bart_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected <- n_chains * n_mcmc
  s2 <- extractParameter(m, "sigma2_global")
  expect_length(s2, expected)
  yht <- extractParameter(m, "y_hat_train")
  expect_equal(dim(yht), c(d$n_train, expected))
})

test_that("BART multi-chain: predict() shape and finiteness (no GFR)", {
  skip_on_cran()
  d <- .make_bart_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected_cols <- n_chains * n_mcmc
  result <- predict(m, X = d$X_test, terms = "y_hat")
  expect_equal(dim(result), c(d$n_test, expected_cols))
  expect_true(all(is.finite(result)))
})

test_that("BART multi-chain: predict() shape and finiteness (GFR path)", {
  skip_on_cran()
  d <- .make_bart_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 6, num_burnin = 5, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  result <- predict(m, X = d$X_test, terms = "y_hat")
  expect_equal(dim(result), c(d$n_test, n_chains * n_mcmc))
  expect_true(all(is.finite(result)))
})

test_that("BART multi-chain: num_gfr < num_chains raises an error", {
  skip_on_cran()
  d <- .make_bart_data()
  expect_error(
    bart(
      X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
      num_gfr = 2, num_burnin = 0, num_mcmc = 5,
      general_params = list(num_chains = 4, num_threads = 1)
    )
  )
})

test_that("BART multi-chain: sigma2 samples are finite and positive with GFR", {
  skip_on_cran()
  d <- .make_bart_data()
  m <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 6, num_burnin = 10, num_mcmc = 10,
    general_params = list(num_chains = 3, num_threads = 1)
  )
  expect_true(all(is.finite(m$sigma2_global_samples)))
  expect_true(all(m$sigma2_global_samples > 0))
})

# ---------------------------------------------------------------------------
# BCFModel multi-chain tests
# ---------------------------------------------------------------------------

test_that("BCF multi-chain: sample counts with no GFR", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected <- n_chains * n_mcmc
  expect_length(m$sigma2_global_samples, expected)
  expect_equal(dim(m$tau_hat_train), c(d$n_train, expected))
  expect_equal(dim(m$mu_hat_train),  c(d$n_train, expected))
  expect_equal(dim(m$tau_hat_test),  c(d$n_test,  expected))
})

test_that("BCF multi-chain: sample counts with GFR warm-start", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_chains <- 3; n_mcmc <- 10; n_gfr <- 6
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = n_gfr, num_burnin = 5, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1, adaptive_coding = TRUE)
  )
  expected <- n_chains * n_mcmc
  expect_length(m$sigma2_global_samples, expected)
  expect_equal(dim(m$tau_hat_train), c(d$n_train, expected))
  expect_equal(dim(m$mu_hat_train),  c(d$n_train, expected))
  # BCF-specific scalar parameter arrays
  expect_length(m$b_0_samples, expected)
  expect_length(m$b_1_samples, expected)
  expect_length(m$sigma2_leaf_mu_samples, expected)
})

test_that("BCF multi-chain: chain independence (no GFR)", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_mcmc <- 10
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = n_mcmc,
    general_params = list(num_chains = 2, num_threads = 1)
  )
  chain1 <- m$sigma2_global_samples[seq_len(n_mcmc)]
  chain2 <- m$sigma2_global_samples[seq(n_mcmc + 1, 2 * n_mcmc)]
  expect_false(isTRUE(all.equal(chain1, chain2)),
               label = "BCF chains should produce distinct sigma2 samples")
})

test_that("BCF multi-chain: chain independence (with GFR)", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_mcmc <- 10
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 4, num_burnin = 5, num_mcmc = n_mcmc,
    general_params = list(num_chains = 2, num_threads = 1)
  )
  chain1 <- m$sigma2_global_samples[seq_len(n_mcmc)]
  chain2 <- m$sigma2_global_samples[seq(n_mcmc + 1, 2 * n_mcmc)]
  expect_false(isTRUE(all.equal(chain1, chain2)))
})

test_that("BCF multi-chain: all samples finite with GFR + multiple chains", {
  skip_on_cran()
  # Exercises the tau_0 / adaptive-coding reset logic introduced to prevent
  # residual blowup when transitioning between chains.
  d <- .make_bcf_data()
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 6, num_burnin = 20, num_mcmc = 10,
    general_params = list(num_chains = 3, num_threads = 1, adaptive_coding = TRUE)
  )
  expect_true(all(is.finite(m$sigma2_global_samples)),
              label = "sigma2 samples must be finite (no chain-transition blowup)")
  expect_true(all(m$sigma2_global_samples > 0))
  expect_true(all(is.finite(m$b_0_samples)))
  expect_true(all(is.finite(m$b_1_samples)))
})

test_that("BCF multi-chain: extractParameter dimensions", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected <- n_chains * n_mcmc
  s2 <- extractParameter(m, "sigma2_global")
  expect_length(s2, expected)
  tau_test <- extractParameter(m, "tau_hat_test")
  expect_equal(dim(tau_test), c(d$n_test, expected))
  tau_train <- extractParameter(m, "tau_hat_train")
  expect_equal(dim(tau_train), c(d$n_train, expected))
})

test_that("BCF multi-chain: predict() shape and finiteness for all forest terms (no GFR)", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected_cols <- n_chains * n_mcmc
  for (term in c("y_hat", "cate", "prognostic_function", "mu", "tau")) {
    result <- predict(m, X = d$X_test, Z = d$Z_test, propensity = d$pi_test, terms = term)
    expect_equal(dim(result), c(d$n_test, expected_cols),
                 label = paste0("dim for term='", term, "'"))
    expect_true(all(is.finite(result)),
                label = paste0("finiteness for term='", term, "'"))
  }
})

test_that("BCF multi-chain: predict() shape and finiteness for all forest terms (GFR path)", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_chains <- 3; n_mcmc <- 10; n_gfr <- 6
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = n_gfr, num_burnin = 5, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )
  expected_cols <- n_chains * n_mcmc
  for (term in c("y_hat", "cate", "prognostic_function", "mu", "tau")) {
    result <- predict(m, X = d$X_test, Z = d$Z_test, propensity = d$pi_test, terms = term)
    expect_equal(dim(result), c(d$n_test, expected_cols),
                 label = paste0("dim for term='", term, "'"))
    expect_true(all(is.finite(result)),
                label = paste0("finiteness for term='", term, "'"))
  }
})

test_that("BCF multi-chain: predict() shape and positivity for variance forest term", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_chains <- 3; n_mcmc <- 10
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1),
    variance_forest_params = list(num_trees = 10)
  )
  result <- predict(m, X = d$X_test, Z = d$Z_test, propensity = d$pi_test,
                    terms = "variance_forest")
  expect_equal(dim(result), c(d$n_test, n_chains * n_mcmc))
  expect_true(all(is.finite(result)))
  expect_true(all(result > 0))
})

test_that("BCF multi-chain: num_gfr < num_chains raises an error", {
  skip_on_cran()
  d <- .make_bcf_data()
  expect_error(
    bcf(
      X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
      propensity_train = d$pi_train,
      X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
      num_gfr = 2, num_burnin = 0, num_mcmc = 5,
      general_params = list(num_chains = 4, num_threads = 1)
    )
  )
})

test_that("BCF multi-chain: serialization round-trip preserves predictions", {
  skip_on_cran()
  d <- .make_bcf_data()
  n_chains <- 2; n_mcmc <- 10; n_gfr <- 4
  m <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train,
    X_test = d$X_test, Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = n_gfr, num_burnin = 5, num_mcmc = n_mcmc,
    general_params = list(num_chains = n_chains, num_threads = 1)
  )

  json_str <- saveBCFModelToJsonString(m)
  m2 <- createBCFModelFromJsonString(json_str)

  pred_orig <- predict(
    m, X = d$X_test, Z = d$Z_test, propensity = d$pi_test, terms = "cate"
  )
  pred_rt <- predict(
    m2, X = d$X_test, Z = d$Z_test, propensity = d$pi_test, terms = "cate"
  )
  expect_equal(dim(pred_orig), dim(pred_rt))
  expect_equal(pred_orig, pred_rt)
})

test_that("Random Effects BART with Default Numbering", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  p_w <- 2
  X <- matrix(runif(n * p), ncol = p)
  W <- matrix(runif(n * p_w), ncol = p_w)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5 * W[, 1]) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5 * W[, 1]) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5 * W[, 1]) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5 * W[, 1]))
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
  rfx_basis <- cbind(rep(1, n), runif(n))
  num_rfx_components <- ncol(rfx_basis)
  num_rfx_groups <- length(unique(rfx_group_ids))
  rfx_coefs <- matrix(c(-5, 5, 1, -1), ncol = 2, byrow = T)
  rfx_term <- rowSums(rfx_coefs[rfx_group_ids, ] * rfx_basis)
  noise_sd <- 1
  y <- f_XW + rfx_term + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  W_test <- W[test_inds, ]
  W_train <- W[train_inds, ]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Specify no rfx parameters
  general_param_list <- list()
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      mean_forest_params = mean_forest_param_list
    )
  )

  # Specify all rfx parameters as scalars
  rfx_param_list <- list(
    working_parameter_prior_mean = 1.,
    group_parameter_prior_mean = 1.,
    working_parameter_prior_cov = 1.,
    group_parameter_prior_cov = 1.,
    variance_prior_shape = 1,
    variance_prior_scale = 1
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
  )

  # Specify all relevant rfx parameters as vectors
  rfx_param_list <- list(
    working_parameter_prior_mean = c(1., 1.),
    group_parameter_prior_mean = c(1., 1.),
    working_parameter_prior_cov = diag(1., 2),
    group_parameter_prior_cov = diag(1., 2),
    variance_prior_shape = 1,
    variance_prior_scale = 1
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
  )

  # Specify simpler intercept-only RFX model
  rfx_param_list <- list(
    model_spec = "intercept_only"
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error({
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
    preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
  })
})

test_that("Random Effects BART with Default Numbering", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  p_w <- 2
  X <- matrix(runif(n * p), ncol = p)
  W <- matrix(runif(n * p_w), ncol = p_w)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5 * W[, 1]) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5 * W[, 1]) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5 * W[, 1]) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5 * W[, 1]))
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
  rfx_basis <- cbind(rep(1, n), runif(n))
  num_rfx_components <- ncol(rfx_basis)
  num_rfx_groups <- length(unique(rfx_group_ids))
  rfx_coefs <- matrix(c(-5, 5, 1, -1), ncol = 2, byrow = T)
  rfx_term <- rowSums(rfx_coefs[rfx_group_ids, ] * rfx_basis)
  noise_sd <- 1
  y <- f_XW + rfx_term + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  W_test <- W[test_inds, ]
  W_train <- W[train_inds, ]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Specify no rfx basis directly
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error({
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list
    )
    preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      rfx_basis = rfx_basis_test,
      type = "posterior",
      terms = "rfx"
    )
  })

  # Intercept-only RFX model
  rfx_param_list <- list(
    model_spec = "intercept_only"
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error({
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
    preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
  })
})

test_that("Random Effects BART with Offset Numbering", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  p_w <- 2
  X <- matrix(runif(n * p), ncol = p)
  W <- matrix(runif(n * p_w), ncol = p_w)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5 * W[, 1]) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5 * W[, 1]) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5 * W[, 1]) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5 * W[, 1]))
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE) + 2
  rfx_basis <- cbind(rep(1, n), runif(n))
  num_rfx_components <- ncol(rfx_basis)
  num_rfx_groups <- length(unique(rfx_group_ids))
  rfx_coefs <- matrix(c(-5, 5, 1, -1), ncol = 2, byrow = T)
  rfx_term <- rowSums(rfx_coefs[rfx_group_ids - 2, ] * rfx_basis)
  noise_sd <- 1
  y <- f_XW + rfx_term + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  W_test <- W[test_inds, ]
  W_train <- W[train_inds, ]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Specify no rfx basis directly
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error({
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list
    )
    rfx_preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      rfx_basis = rfx_basis_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      rfx_basis = rfx_basis_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Intercept-only RFX model
  rfx_param_list <- list(
    model_spec = "intercept_only"
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error({
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
    rfx_preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Test print method
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    leaf_basis_train = W_train,
    leaf_basis_test = W_test,
    rfx_group_ids_train = rfx_group_ids_train,
    rfx_group_ids_test = rfx_group_ids_test,
    num_gfr = 0,
    num_burnin = 10,
    num_mcmc = 10,
    mean_forest_params = mean_forest_param_list,
    random_effects_params = rfx_param_list
  )
  rfx_output <- capture.output(print(bart_model$rfx_samples))
  expect_true(any(grepl("Random effects container", rfx_output, fixed = TRUE)))
  expect_true(any(grepl(
    bart_model$rfx_samples$num_samples(),
    rfx_output,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    bart_model$rfx_samples$num_components(),
    rfx_output,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    bart_model$rfx_samples$num_groups(),
    rfx_output,
    fixed = TRUE
  )))
})

test_that("Random Effects BCF with Default Numbering", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  # fmt: skip
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  # fmt: skip
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
  rfx_basis <- cbind(rep(1, n), runif(n))
  num_rfx_components <- ncol(rfx_basis)
  num_rfx_groups <- length(unique(rfx_group_ids))
  rfx_coefs <- matrix(c(-5, 5, 1, -1), ncol = 2, byrow = T)
  rfx_term <- rowSums(rfx_coefs[rfx_group_ids, ] * rfx_basis)
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rfx_term + rnorm(n, 0, noise_sd)

  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit a BCF model with random effects by passing the basis directly
  expect_no_error({
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_basis_train = rfx_basis_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10
    )
    rfx_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      rfx_basis = rfx_basis_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      rfx_basis = rfx_basis_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Fit a BCF model with random effects by specifying an "intercept only" model
  expect_no_error({
    rfx_param_list <- list(
      model_spec = "intercept_only"
    )
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      random_effects_params = rfx_param_list
    )
    rfx_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Fit a BCF model with random effects by specifying an "intercept plus treatment" model
  expect_no_error({
    rfx_param_list <- list(
      model_spec = "intercept_plus_treatment"
    )
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      random_effects_params = rfx_param_list
    )
    rfx_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Test print method
  bcf_model <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    rfx_group_ids_train = rfx_group_ids_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    rfx_group_ids_test = rfx_group_ids_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10,
    random_effects_params = rfx_param_list
  )
  rfx_output <- capture.output(print(bcf_model$rfx_samples))
  expect_true(any(grepl("Random effects container", rfx_output, fixed = TRUE)))
  expect_true(any(grepl(
    bcf_model$rfx_samples$num_samples(),
    rfx_output,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    bcf_model$rfx_samples$num_components(),
    rfx_output,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    bcf_model$rfx_samples$num_groups(),
    rfx_output,
    fixed = TRUE
  )))
})

test_that("Random Effects BCF with Offset Numbering", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  mu_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  # fmt: skip
  pi_X <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (0.2) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (0.4) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (0.6) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (0.8))
  # fmt: skip
  tau_X <- (((0 <= X[, 2]) & (0.25 > X[, 2])) * (0.5) +
    ((0.25 <= X[, 2]) & (0.5 > X[, 2])) * (1.0) +
    ((0.5 <= X[, 2]) & (0.75 > X[, 2])) * (1.5) +
    ((0.75 <= X[, 2]) & (1 > X[, 2])) * (2.0))
  Z <- rbinom(n, 1, pi_X)
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE) + 2
  rfx_basis <- cbind(rep(1, n), runif(n))
  num_rfx_components <- ncol(rfx_basis)
  num_rfx_groups <- length(unique(rfx_group_ids))
  rfx_coefs <- matrix(c(-5, 5, 1, -1), ncol = 2, byrow = T)
  rfx_term <- rowSums(rfx_coefs[rfx_group_ids - 2, ] * rfx_basis)
  noise_sd <- 1
  y <- mu_X + tau_X * Z + rfx_term + rnorm(n, 0, noise_sd)

  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  pi_test <- pi_X[test_inds]
  pi_train <- pi_X[train_inds]
  mu_test <- mu_X[test_inds]
  mu_train <- mu_X[train_inds]
  tau_test <- tau_X[test_inds]
  tau_train <- tau_X[train_inds]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit a BCF model with random effects by passing the basis directly
  expect_no_error({
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_basis_train = rfx_basis_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10
    )
    rfx_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      rfx_basis = rfx_basis_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      rfx_basis = rfx_basis_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Fit a BCF model with random effects by specifying an "intercept only" model
  expect_no_error({
    rfx_param_list <- list(
      model_spec = "intercept_only"
    )
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      random_effects_params = rfx_param_list
    )
    rfx_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Fit a BCF model with random effects by specifying an "intercept plus treatment" model
  expect_no_error({
    rfx_param_list <- list(
      model_spec = "intercept_plus_treatment"
    )
    bcf_model <- bcf(
      X_train = X_train,
      y_train = y_train,
      Z_train = Z_train,
      propensity_train = pi_train,
      rfx_group_ids_train = rfx_group_ids_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      random_effects_params = rfx_param_list
    )
    rfx_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
    yhat_preds <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "y_hat"
    )
  })

  # Test print method
  bcf_model <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    rfx_group_ids_train = rfx_group_ids_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    rfx_group_ids_test = rfx_group_ids_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10,
    random_effects_params = rfx_param_list
  )
  rfx_output <- capture.output(print(bcf_model$rfx_samples))
  expect_true(any(grepl("Random effects container", rfx_output, fixed = TRUE)))
  expect_true(any(grepl(
    bcf_model$rfx_samples$num_samples(),
    rfx_output,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    bcf_model$rfx_samples$num_components(),
    rfx_output,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    bcf_model$rfx_samples$num_groups(),
    rfx_output,
    fixed = TRUE
  )))
})

#' Forest Kernel Computation Routines
#' @name ForestKernelComputation
#' @description
#' Decision tree ensembles can be represented in part by a "kernel" function whose distance metric
#' is based on the extent to which two observations are mapped to the same leaf nodes.
#' This function group offers utilities for evaluating this kernel.
#'
#' `computeForestLeafIndices` computes and return a vector representation of a forest's
#' leaf predictions for every observation in a dataset.
#' The resulting vector has a "row-major" format that can be easily re-represented as
#' as a CSR sparse matrix: elements are organized so that the first `n` elements
#' correspond to leaf predictions for all `n` observations in a dataset for the
#' first tree in an ensemble, the next `n` elements correspond to predictions for
#' the second tree and so on. The "data" for each element corresponds to a uniquely
#' mapped column index that corresponds to a single leaf of a single tree (i.e.
#' if tree 1 has 3 leaves, its column indices range from 0 to 2, and then tree 2's
#' leaf indices begin at 3, etc...).
#'
#' `computeForestLeafVariances` returns each forest's leaf node scale parameters.
#' If leaf scale is not sampled for the forest in question, the function throws an error that the
#' leaf model does not have a stochastic scale parameter.
#'
#' `computeForestMaxLeafIndex` computes and returns the largest possible leaf index computable by `computeForestLeafIndices` for the forests in a designated forest sample container.
#'
#' These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#' @returns
#' `computeForestLeafIndices` returns a vector of size `num_obs * num_trees`, where `num_obs = nrow(covariates)`
#' and `num_trees` is the number of trees in the relevant forest of `model_object`.
#'
#' `computeForestLeafVariances` returns a vector of size `length(forest_inds)` with the leaf scale parameter for each requested forest.
#'
#' `computeForestMaxLeafIndex` returns a vector containing the largest possible leaf index computable by `computeForestLeafIndices` for the forests in a designated forest sample container.
#' @examples
#' X <- matrix(runif(10*100), ncol = 10)
#' y <- -5 + 10*(X[,1] > 0.5) + rnorm(100)
#' bart_model <- bart(X, y, num_gfr=0, num_mcmc=10)
#' leaf_indices <- computeForestLeafIndices(bart_model, X, "mean")
#' leaf_indices <- computeForestLeafIndices(bart_model, X, "mean", 0)
#' leaf_indices <- computeForestLeafIndices(bart_model, X, "mean", c(1,3,9))
#' leaf_variances <- computeForestLeafVariances(bart_model, "mean")
#' leaf_variances <- computeForestLeafVariances(bart_model, "mean", 0)
#' leaf_variances <- computeForestLeafVariances(bart_model, "mean", c(1,3,5))
#' max_leaf_index <- computeForestMaxLeafIndex(bart_model, "mean")
#' max_leaf_index <- computeForestMaxLeafIndex(bart_model, "mean", 0)
#' max_leaf_index <- computeForestMaxLeafIndex(bart_model, "mean", c(1,3,9))
#'
NULL
#> NULL

#' @title Query Forest Leaf Indices
#' @rdname ForestKernelComputation
#' @param model_object Object of type `bartmodel`, `bcfmodel`, or `ForestSamples` corresponding to a BART / BCF model with at least one forest sample, or a low-level `ForestSamples` object.
#' @param covariates Covariates to use for prediction. Must have the same dimensions / column types as the data used to train a forest.
#' @param forest_type Which forest to use from `model_object`.
#' Valid inputs depend on the model type, and whether or not a given forest was sampled in that model.
#'
#'   **1. BART**
#'
#'   - `'mean'`: Extracts leaf indices for the mean forest
#'   - `'variance'`: Extracts leaf indices for the variance forest
#'
#'   **2. BCF**
#'
#'   - `'prognostic'`: Extracts leaf indices for the prognostic forest
#'   - `'treatment'`: Extracts leaf indices for the treatment effect forest
#'   - `'variance'`: Extracts leaf indices for the variance forest
#'
#'   **3. ForestSamples**
#'
#'   - `NULL`: It is not necessary to disambiguate when this function is called directly on a `ForestSamples` object. This is the default value of this
#'
#' @param propensity (Optional) Propensities used for prediction (BCF-only).
#' @param forest_inds (Optional) Indices of the forest sample(s) for which to compute leaf indices. If not provided,
#' this function will return leaf indices for every sample of a forest.
#' This function uses 0-indexing, so the first forest sample corresponds to `forest_num = 0`, and so on.
#' @export
computeForestLeafIndices <- function(
  model_object,
  covariates,
  forest_type = NULL,
  propensity = NULL,
  forest_inds = NULL
) {
  # Extract relevant forest container
  stopifnot(any(c(
    inherits(model_object, "bartmodel"),
    inherits(model_object, "bcfmodel"),
    inherits(model_object, "ForestSamples")
  )))
  model_type <- ifelse(
    inherits(model_object, "bartmodel"),
    "bart",
    ifelse(inherits(model_object, "bcfmodel"), "bcf", "forest_samples")
  )
  if (model_type == "bart") {
    stopifnot(forest_type %in% c("mean", "variance"))
    if (forest_type == "mean") {
      if (!model_object$model_params$include_mean_forest) {
        stop("Mean forest was not sampled in the bart model provided")
      }
      forest_container <- model_object$mean_forests
    } else if (forest_type == "variance") {
      if (!model_object$model_params$include_variance_forest) {
        stop(
          "Variance forest was not sampled in the bart model provided"
        )
      }
      forest_container <- model_object$variance_forests
    }
  } else if (model_type == "bcf") {
    stopifnot(forest_type %in% c("prognostic", "treatment", "variance"))
    if (forest_type == "prognostic") {
      forest_container <- model_object$forests_mu
    } else if (forest_type == "treatment") {
      forest_container <- model_object$forests_tau
    } else if (forest_type == "variance") {
      if (!model_object$model_params$include_variance_forest) {
        stop(
          "Variance forest was not sampled in the bcf model provided"
        )
      }
      forest_container <- model_object$variance_forests
    }
  } else {
    forest_container <- model_object
  }

  # Preprocess covariates
  if ((!is.data.frame(covariates)) && (!is.matrix(covariates))) {
    stop("covariates must be a matrix or dataframe")
  }
  if (model_type %in% c("bart", "bcf")) {
    train_set_metadata <- model_object$train_set_metadata
    covariates_processed <- preprocessPredictionData(
      covariates,
      train_set_metadata
    )
  } else {
    if (!is.matrix(covariates)) {
      stop(
        "covariates must be a matrix since no covariate preprocessor is stored in a `ForestSamples` object provided as `model_object`"
      )
    }
    covariates_processed <- covariates
  }

  # Handle BCF propensity covariate
  if (model_type == "bcf") {
    # Add propensities to covariate set if necessary
    if (model_object$model_params$propensity_covariate != "none") {
      if (is.null(propensity)) {
        if (!model_object$model_params$internal_propensity_model) {
          stop("propensity must be provided for this model")
        }
        # Compute propensity score using the internal bart model
        propensity <- rowMeans(
          predict(
            model_object$bart_propensity_model,
            X = covariates
          )$y_hat
        )
      }
      covariates_processed <- cbind(covariates_processed, propensity)
    }
  }

  # Preprocess forest indices
  num_forests <- forest_container$num_samples()
  if (is.null(forest_inds)) {
    forest_inds <- as.integer(1:num_forests - 1)
  } else {
    stopifnot(all(forest_inds <= num_forests - 1))
    stopifnot(all(forest_inds >= 0))
    forest_inds <- as.integer(forest_inds)
  }

  # Compute leaf indices
  leaf_ind_matrix <- compute_leaf_indices_cpp(
    forest_container$forest_container_ptr,
    covariates_processed,
    forest_inds
  )

  return(leaf_ind_matrix)
}

#' @title Query Forest Leaf Scale Parameters
#' @rdname ForestKernelComputation
#' @param model_object Object of type `bartmodel` or `bcfmodel` corresponding to a BART / BCF model with at least one forest sample
#' @param forest_type Which forest to use from `model_object`.
#' Valid inputs depend on the model type, and whether or not a given forest was sampled in that model.
#'
#'   **1. BART**
#'
#'   - `'mean'`: Extracts leaf indices for the mean forest
#'   - `'variance'`: Extracts leaf indices for the variance forest
#'
#'   **2. BCF**
#'
#'   - `'prognostic'`: Extracts leaf indices for the prognostic forest
#'   - `'treatment'`: Extracts leaf indices for the treatment effect forest
#'   - `'variance'`: Extracts leaf indices for the variance forest
#'
#' @param forest_inds (Optional) Indices of the forest sample(s) for which to compute leaf indices. If not provided,
#' this function will return leaf indices for every sample of a forest.
#' This function uses 0-indexing, so the first forest sample corresponds to `forest_num = 0`, and so on.
#' @export
computeForestLeafVariances <- function(
  model_object,
  forest_type,
  forest_inds = NULL
) {
  # Extract relevant forest container
  stopifnot(any(c(
    inherits(model_object, "bartmodel"),
    inherits(model_object, "bcfmodel")
  )))
  model_type <- ifelse(inherits(model_object, "bartmodel"), "bart", "bcf")
  if (model_type == "bart") {
    stopifnot(forest_type %in% c("mean", "variance"))
    if (forest_type == "mean") {
      if (!model_object$model_params$include_mean_forest) {
        stop("Mean forest was not sampled in the bart model provided")
      }
      if (!model_object$model_params$sample_sigma2_leaf) {
        stop(
          "Leaf scale parameter was not sampled for the mean forest in the bart model provided"
        )
      }
      leaf_scale_vector <- model_object$sigma2_leaf_samples
    } else if (forest_type == "variance") {
      if (!model_object$model_params$include_variance_forest) {
        stop(
          "Variance forest was not sampled in the bart model provided"
        )
      }
      stop(
        "Leaf scale parameter was not sampled for the variance forest in the bart model provided"
      )
    }
  } else {
    stopifnot(forest_type %in% c("prognostic", "treatment", "variance"))
    if (forest_type == "prognostic") {
      if (!model_object$model_params$sample_sigma2_leaf_mu) {
        stop(
          "Leaf scale parameter was not sampled for the prognostic forest in the bcf model provided"
        )
      }
      leaf_scale_vector <- model_object$sigma2_leaf_mu_samples
    } else if (forest_type == "treatment") {
      if (!model_object$model_params$sample_sigma2_leaf_tau) {
        stop(
          "Leaf scale parameter was not sampled for the treatment effect forest in the bcf model provided"
        )
      }
      leaf_scale_vector <- model_object$sigma2_leaf_tau_samples
    } else if (forest_type == "variance") {
      if (!model_object$model_params$include_variance_forest) {
        stop(
          "Variance forest was not sampled in the bcf model provided"
        )
      }
      stop(
        "Leaf scale parameter was not sampled for the variance forest in the bcf model provided"
      )
    }
  }

  # Preprocess forest indices
  num_forests <- model_object$model_params$num_samples
  if (is.null(forest_inds)) {
    forest_inds <- as.integer(1:num_forests)
  } else {
    stopifnot(all(forest_inds <= num_forests - 1))
    stopifnot(all(forest_inds >= 0))
    forest_inds <- as.integer(forest_inds + 1)
  }

  # Gather leaf scale parameters
  leaf_scale_params <- leaf_scale_vector[forest_inds]

  return(leaf_scale_params)
}

#' @title Query Forest Max Leaf Index
#' @rdname ForestKernelComputation
#' @param model_object Object of type `bartmodel`, `bcfmodel`, or `ForestSamples` corresponding to a BART / BCF model with at least one forest sample, or a low-level `ForestSamples` object.
#' @param forest_type Which forest to use from `model_object`.
#' Valid inputs depend on the model type, and whether or not a
#'
#'   **1. BART**
#'
#'   - `'mean'`: Extracts leaf indices for the mean forest
#'   - `'variance'`: Extracts leaf indices for the variance forest
#'
#'   **2. BCF**
#'
#'   - `'prognostic'`: Extracts leaf indices for the prognostic forest
#'   - `'treatment'`: Extracts leaf indices for the treatment effect forest
#'   - `'variance'`: Extracts leaf indices for the variance forest
#'
#'   **3. ForestSamples**
#'
#'   - `NULL`: It is not necessary to disambiguate when this function is called directly on a `ForestSamples` object. This is the default value of this
#'
#' @param forest_inds (Optional) Indices of the forest sample(s) for which to compute max leaf indices. If not provided,
#' this function will return max leaf indices for every sample of a forest.
#' This function uses 0-indexing, so the first forest sample corresponds to `forest_num = 0`, and so on.
#' @export
computeForestMaxLeafIndex <- function(
  model_object,
  forest_type = NULL,
  forest_inds = NULL
) {
  # Extract relevant forest container
  stopifnot(any(c(
    inherits(model_object, "bartmodel"),
    inherits(model_object, "bcfmodel"),
    inherits(model_object, "ForestSamples")
  )))
  model_type <- ifelse(
    inherits(model_object, "bartmodel"),
    "bart",
    ifelse(inherits(model_object, "bcfmodel"), "bcf", "forest_samples")
  )
  if (model_type == "bart") {
    stopifnot(forest_type %in% c("mean", "variance"))
    if (forest_type == "mean") {
      if (!model_object$model_params$include_mean_forest) {
        stop("Mean forest was not sampled in the bart model provided")
      }
      forest_container <- model_object$mean_forests
    } else if (forest_type == "variance") {
      if (!model_object$model_params$include_variance_forest) {
        stop(
          "Variance forest was not sampled in the bart model provided"
        )
      }
      forest_container <- model_object$variance_forests
    }
  } else if (model_type == "bcf") {
    stopifnot(forest_type %in% c("prognostic", "treatment", "variance"))
    if (forest_type == "prognostic") {
      forest_container <- model_object$forests_mu
    } else if (forest_type == "treatment") {
      forest_container <- model_object$forests_tau
    } else if (forest_type == "variance") {
      if (!model_object$model_params$include_variance_forest) {
        stop(
          "Variance forest was not sampled in the bcf model provided"
        )
      }
      forest_container <- model_object$variance_forests
    }
  } else {
    forest_container <- model_object
  }

  # Preprocess forest indices
  num_forests <- forest_container$num_samples()
  if (is.null(forest_inds)) {
    forest_inds <- as.integer(1:num_forests - 1)
  } else {
    stopifnot(all(forest_inds <= num_forests - 1))
    stopifnot(all(forest_inds >= 0))
    forest_inds <- as.integer(forest_inds)
  }

  # Compute leaf indices
  output <- rep(NA, length(forest_inds))
  for (i in 1:length(forest_inds)) {
    output[i] <- forest_container_get_max_leaf_index_cpp(
      forest_container$forest_container_ptr,
      forest_inds[i]
    )
  }

  return(output)
}

## usethis namespace: start
#' @importFrom stats coef
#' @importFrom stats dnorm
#' @importFrom stats lm
#' @importFrom stats model.matrix
#' @importFrom stats predict
#' @importFrom stats qgamma
#' @importFrom stats qnorm
#' @importFrom stats quantile
#' @importFrom stats pnorm
#' @importFrom stats rbinom
#' @importFrom stats resid
#' @importFrom stats rnorm
#' @importFrom stats runif
#' @importFrom stats sd
#' @importFrom stats sigma
#' @importFrom stats var
#' @importFrom R6 R6Class
## usethis namespace: end
NULL

#' @useDynLib stochtree, .registration = TRUE
"_PACKAGE"

#' @title Calibrate Inverse Gamma Prior
#' @description
#' Calibrate the scale parameter on an inverse gamma prior for the global error variance as in Chipman et al (2022)
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' Chipman, H., George, E., Hahn, R., McCulloch, R., Pratola, M. and Sparapani, R. (2022). Bayesian Additive Regression Trees, Computational Approaches. In Wiley StatsRef: Statistics Reference Online (eds N. Balakrishnan, T. Colton, B. Everitt, W. Piegorsch, F. Ruggeri and J.L. Teugels). \doi{10.1002/9781118445112.stat08288}
#'
#' @param y Outcome to be modeled using BART, BCF or another nonparametric ensemble method.
#' @param X Covariates to be used to partition trees in an ensemble or series of ensemble.
#' @param W (Optional) Basis used to define a "leaf regression" model for each decision tree. The "classic" BART model assumes a constant leaf parameter, which is equivalent to a "leaf regression" on a basis of all ones, though it is not necessary to pass a vector of ones, here or to the BART function. Default: `NULL`.
#' @param nu The shape parameter for the global error variance's IG prior. The scale parameter in the Sparapani et al (2021) parameterization is defined as `nu*lambda` where `lambda` is the output of this function. Default: `3`.
#' @param quant (Optional) Quantile of the inverse gamma prior distribution represented by a linear-regression-based overestimate of `sigma^2`. Default: `0.9`.
#' @param standardize (Optional) Whether or not outcome should be standardized (`(y-mean(y))/sd(y)`) before calibration of `lambda`. Default: `TRUE`.
#'
#' @return Value of `lambda` which determines the scale parameter of the global error variance prior (`sigma^2 ~ IG(nu,nu*lambda)`)
#' @export
#'
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' y <- 10*X[,1] - 20*X[,2] + rnorm(n)
#' nu <- 3
#' lambda <- calibrateInverseGammaErrorVariance(y, X, nu = nu)
#' sigma2hat <- mean(resid(lm(y~X))^2)
#' mean(var(y)/rgamma(100000, nu, rate = nu*lambda) < sigma2hat)
calibrateInverseGammaErrorVariance <- function(
  y,
  X,
  W = NULL,
  nu = 3,
  quant = 0.9,
  standardize = TRUE
) {
  # Compute regression basis
  if (!is.null(W)) {
    basis <- cbind(X, W)
  } else {
    basis <- X
  }
  # Standardize outcome if requested
  if (standardize) {
    y <- (y - mean(y)) / sd(y)
  }
  # Compute the "regression-based" overestimate of sigma^2
  sigma2hat <- mean(resid(lm(y ~ basis))^2)
  # Calibrate lambda based on the implied quantile of sigma2hat
  return((sigma2hat * qgamma(1 - quant, nu)) / nu)
}

#' @title Forest Model Configuration Object
#'
#' @description
#' Object used to get / set parameters and other model configuration options
#' for a forest model in the "low-level" stochtree interface.
#' The "low-level" stochtree interface enables a high degreee of sampler
#' customization, in which users employ R wrappers around C++ objects
#' like ForestDataset, Outcome, CppRng, and ForestModel to run the
#' Gibbs sampler of a BART model with custom modifications.
#' ForestModelConfig allows users to specify / query the parameters of a
#' forest model they wish to run.
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

ForestModelConfig <- R6::R6Class(
  classname = "ForestModelConfig",
  cloneable = FALSE,
  public = list(
    #' @field feature_types Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)
    feature_types = NULL,

    #' @field sweep_update_indices Vector of trees to update in a sweep
    sweep_update_indices = NULL,

    #' @field num_trees Number of trees in the forest being sampled
    num_trees = NULL,

    #' @field num_features Number of features in training dataset
    num_features = NULL,

    #' @field num_observations Number of observations in training dataset
    num_observations = NULL,

    #' @field leaf_dimension Dimension of the leaf model
    leaf_dimension = NULL,

    #' @field alpha Root node split probability in tree prior
    alpha = NULL,

    #' @field beta Depth prior penalty in tree prior
    beta = NULL,

    #' @field min_samples_leaf Minimum number of samples in a tree leaf
    min_samples_leaf = NULL,

    #' @field max_depth Maximum depth of any tree in the ensemble in the model. Setting to `-1` does not enforce any depth limits on trees.
    max_depth = NULL,

    #' @field leaf_model_type Integer specifying the leaf model type (0 = constant leaf, 1 = univariate leaf regression, 2 = multivariate leaf regression)
    leaf_model_type = NULL,

    #' @field leaf_model_scale Scale parameter used in Gaussian leaf models
    leaf_model_scale = NULL,

    #' @field variable_weights Vector specifying sampling probability for all p covariates in ForestDataset
    variable_weights = NULL,

    #' @field variance_forest_shape Shape parameter for IG leaf models (applicable when `leaf_model_type = 3`)
    variance_forest_shape = NULL,

    #' @field variance_forest_scale Scale parameter for IG leaf models (applicable when `leaf_model_type = 3`)
    variance_forest_scale = NULL,

    #' @field cloglog_forest_shape Shape parameter for conditional gamma component of cloglog leaf models (applicable when `leaf_model_type = 4`)
    cloglog_forest_shape = NULL,

    #' @field cloglog_forest_rate Rate parameter for conditional gamma component of cloglog leaf models (applicable when `leaf_model_type = 4`)
    cloglog_forest_rate = NULL,

    #' @field cutpoint_grid_size Number of unique cutpoints to consider
    cutpoint_grid_size = NULL,

    #' @field num_features_subsample Number of features to subsample for the GFR algorithm
    num_features_subsample = NULL,

    #' Create a new ForestModelConfig object.
    #'
    #' @param feature_types Vector of integer-coded feature types (where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)
    #' @param sweep_update_indices Vector of (0-indexed) indices of trees to update in a sweep
    #' @param num_trees Number of trees in the forest being sampled
    #' @param num_features Number of features in training dataset
    #' @param num_observations Number of observations in training dataset
    #' @param variable_weights Vector specifying sampling probability for all p covariates in ForestDataset
    #' @param leaf_dimension Dimension of the leaf model (default: `1`)
    #' @param alpha Root node split probability in tree prior (default: `0.95`)
    #' @param beta Depth prior penalty in tree prior (default: `2.0`)
    #' @param min_samples_leaf Minimum number of samples in a tree leaf (default: `5`)
    #' @param max_depth Maximum depth of any tree in the ensemble in the model. Setting to `-1` does not enforce any depth limits on trees. Default: `-1`.
    #' @param leaf_model_type Integer specifying the leaf model type (0 = constant leaf, 1 = univariate leaf regression, 2 = multivariate leaf regression). Default: `0`.
    #' @param leaf_model_scale Scale parameter used in Gaussian leaf models (can either be a scalar or a q x q matrix, where q is the dimensionality of the basis and is only >1 when `leaf_model_int = 2`). Calibrated internally as `1/num_trees`, propagated along diagonal if needed for multivariate leaf models.
    #' @param variance_forest_shape Shape parameter for IG leaf models (applicable when `leaf_model_type = 3`). Default: `1`.
    #' @param variance_forest_scale Scale parameter for IG leaf models (applicable when `leaf_model_type = 3`). Default: `1`.
    #' @param cloglog_forest_shape Shape parameter for conditional gamma component of cloglog leaf models (applicable when `leaf_model_type = 4`). Default: `1`.
    #' @param cloglog_forest_rate Rate parameter for conditional gamma component of cloglog leaf models (applicable when `leaf_model_type = 4`). Default: `1`.
    #' @param cutpoint_grid_size Number of unique cutpoints to consider (default: `100`)
    #' @param num_features_subsample Number of features to subsample for the GFR algorithm
    #'
    #' @return A new ForestModelConfig object.
    initialize = function(
      feature_types = NULL,
      sweep_update_indices = NULL,
      num_trees = NULL,
      num_features = NULL,
      num_observations = NULL,
      variable_weights = NULL,
      leaf_dimension = 1,
      alpha = 0.95,
      beta = 2.0,
      min_samples_leaf = 5,
      max_depth = -1,
      leaf_model_type = 1,
      leaf_model_scale = NULL,
      variance_forest_shape = 1.0,
      variance_forest_scale = 1.0,
      cloglog_forest_shape = 2.0,
      cloglog_forest_rate = 2.0,
      cutpoint_grid_size = 100,
      num_features_subsample = NULL
    ) {
      if (is.null(feature_types)) {
        if (is.null(num_features)) {
          stop(
            "Neither of `num_features` nor `feature_types` (a vector from which `num_features` can be inferred) was provided. Please provide at least one of these inputs when creating a ForestModelConfig object."
          )
        }
        warning(
          "`feature_types` not provided, will be assumed to be numeric"
        )
        feature_types <- rep(0, num_features)
      } else {
        if (is.null(num_features)) {
          num_features <- length(feature_types)
        }
      }
      if (is.null(variable_weights)) {
        warning(
          "`variable_weights` not provided, will be assumed to be equal-weighted"
        )
        variable_weights <- rep(1 / num_features, num_features)
      }
      if (is.null(num_trees)) {
        stop("num_trees must be provided")
      }
      if (!is.null(sweep_update_indices)) {
        stopifnot(min(sweep_update_indices) >= 0)
        stopifnot(max(sweep_update_indices) < num_trees)
      }
      if (is.null(num_observations)) {
        stop("num_observations must be provided")
      }
      if (num_features != length(feature_types)) {
        stop("`feature_types` must have `num_features` total elements")
      }
      if (num_features != length(variable_weights)) {
        stop(
          "`variable_weights` must have `num_features` total elements"
        )
      }
      self$feature_types <- feature_types
      self$sweep_update_indices <- sweep_update_indices
      self$variable_weights <- variable_weights
      self$num_trees <- num_trees
      self$num_features <- num_features
      self$num_observations <- num_observations
      self$leaf_dimension <- leaf_dimension
      self$alpha <- alpha
      self$beta <- beta
      self$min_samples_leaf <- min_samples_leaf
      self$max_depth <- max_depth
      self$variance_forest_shape <- variance_forest_shape
      self$variance_forest_scale <- variance_forest_scale
      self$cloglog_forest_shape <- cloglog_forest_shape
      self$cloglog_forest_rate <- cloglog_forest_rate
      self$cutpoint_grid_size <- cutpoint_grid_size
      if (is.null(num_features_subsample)) {
        num_features_subsample <- num_features
      }
      if (num_features_subsample > num_features) {
        stop(
          "`num_features_subsample` cannot be larger than `num_features`"
        )
      }
      if (num_features_subsample <= 0) {
        stop("`num_features_subsample` must be at least 1")
      }
      self$num_features_subsample <- num_features_subsample

      if (!(as.integer(leaf_model_type) == leaf_model_type)) {
        stop("`leaf_model_type` must be an integer between 0 and 3")
        if ((leaf_model_type < 0) | (leaf_model_type > 3)) {
          stop("`leaf_model_type` must be an integer between 0 and 3")
        }
      }
      self$leaf_model_type <- leaf_model_type

      if (is.null(leaf_model_scale)) {
        self$leaf_model_scale <- diag(1 / num_trees, leaf_dimension)
      } else if (is.matrix(leaf_model_scale)) {
        if (ncol(leaf_model_scale) != nrow(leaf_model_scale)) {
          stop("`leaf_model_scale` must be a square matrix")
        }
        if (ncol(leaf_model_scale) != leaf_dimension) {
          stop(
            "`leaf_model_scale` must have `leaf_dimension` rows and columns"
          )
        }
        self$leaf_model_scale <- leaf_model_scale
      } else {
        if (leaf_model_scale <= 0) {
          stop(
            "`leaf_model_scale` must be positive, if provided as scalar"
          )
        }
        self$leaf_model_scale <- diag(leaf_model_scale, leaf_dimension)
      }
    },

    #' @description
    #' Update feature types
    #' @param feature_types Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)
    update_feature_types = function(feature_types) {
      stopifnot(length(feature_types) == self$num_features)
      self$feature_types <- feature_types
    },

    #' @description
    #' Update sweep update indices
    #' @param sweep_update_indices Vector of (0-indexed) indices of trees to update in a sweep
    update_sweep_indices = function(sweep_update_indices) {
      if (!is.null(sweep_update_indices)) {
        stopifnot(min(sweep_update_indices) >= 0)
        stopifnot(max(sweep_update_indices) < self$num_trees)
      }
      self$sweep_update_indices <- sweep_update_indices
    },

    #' @description
    #' Update variable weights
    #' @param variable_weights Vector specifying sampling probability for all p covariates in ForestDataset
    update_variable_weights = function(variable_weights) {
      stopifnot(length(variable_weights) == self$num_features)
      self$variable_weights <- variable_weights
    },

    #' @description
    #' Update root node split probability in tree prior
    #' @param alpha Root node split probability in tree prior
    update_alpha = function(alpha) {
      self$alpha <- alpha
    },

    #' @description
    #' Update depth prior penalty in tree prior
    #' @param beta Depth prior penalty in tree prior
    update_beta = function(beta) {
      self$beta <- beta
    },

    #' @description
    #' Update minimum number of samples per leaf node in the tree prior
    #' @param min_samples_leaf Minimum number of samples in a tree leaf
    update_min_samples_leaf = function(min_samples_leaf) {
      self$min_samples_leaf <- min_samples_leaf
    },

    #' @description
    #' Update max depth in the tree prior
    #' @param max_depth Maximum depth of any tree in the ensemble in the model
    update_max_depth = function(max_depth) {
      self$max_depth <- max_depth
    },

    #' @description
    #' Update scale parameter used in Gaussian leaf models
    #' @param leaf_model_scale Scale parameter used in Gaussian leaf models
    update_leaf_model_scale = function(leaf_model_scale) {
      if (is.matrix(leaf_model_scale)) {
        if (ncol(leaf_model_scale) != nrow(leaf_model_scale)) {
          stop("`leaf_model_scale` must be a square matrix")
        }
        if (ncol(leaf_model_scale) != self$leaf_dimension) {
          stop(
            "`leaf_model_scale` must have `leaf_dimension` rows and columns"
          )
        }
        self$leaf_model_scale <- leaf_model_scale
      } else {
        if (leaf_model_scale <= 0) {
          stop(
            "`leaf_model_scale` must be positive, if provided as scalar"
          )
        }
        self$leaf_model_scale <- diag(leaf_model_scale, self$leaf_dimension)
      }
    },

    #' @description
    #' Update shape parameter for IG leaf models
    #' @param variance_forest_shape Shape parameter for IG leaf models
    update_variance_forest_shape = function(variance_forest_shape) {
      self$variance_forest_shape <- variance_forest_shape
    },

    #' @description
    #' Update scale parameter for IG leaf models
    #' @param variance_forest_scale Scale parameter for IG leaf models
    update_variance_forest_scale = function(variance_forest_scale) {
      self$variance_forest_scale <- variance_forest_scale
    },

    #' @description
    #' Update shape parameter for conditional gamma component of cloglog leaf models
    #' @param cloglog_forest_shape Shape parameter for conditional gamma component of cloglog leaf models
    update_cloglog_forest_shape = function(cloglog_forest_shape) {
      self$cloglog_forest_shape <- cloglog_forest_shape
    },

    #' @description
    #' Update rate parameter for conditional gamma component of cloglog leaf models
    #' @param cloglog_forest_rate Rate parameter for conditional gamma component of cloglog leaf models
    update_cloglog_forest_rate = function(cloglog_forest_rate) {
      self$cloglog_forest_rate <- cloglog_forest_rate
    },

    #' @description
    #' Update number of unique cutpoints to consider
    #' @param cutpoint_grid_size Number of unique cutpoints to consider
    update_cutpoint_grid_size = function(cutpoint_grid_size) {
      self$cutpoint_grid_size <- cutpoint_grid_size
    },

    #' @description
    #' Update number of features to subsample for the GFR algorithm
    #' @param num_features_subsample Number of features to subsample for the GFR algorithm
    update_num_features_subsample = function(num_features_subsample) {
      if (num_features_subsample > self$num_features) {
        stop(
          "`num_features_subsample` cannot be larger than `num_features`"
        )
      }
      if (num_features_subsample <= 0) {
        stop("`num_features_subsample` must at least 1")
      }
      self$num_features_subsample <- num_features_subsample
    },

    #' @description
    #' Query feature types for this ForestModelConfig object
    #' @returns Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)
    get_feature_types = function() {
      return(self$feature_types)
    },

    #' @description
    #' Query sweep update indices for this ForestModelConfig object
    #' @returns Vector of (0-indexed) indices of trees to update in a sweep
    get_sweep_indices = function() {
      return(self$sweep_update_indices)
    },

    #' @description
    #' Query variable weights for this ForestModelConfig object
    #' @returns Vector specifying sampling probability for all p covariates in ForestDataset
    get_variable_weights = function() {
      return(self$variable_weights)
    },

    #' @description
    #' Query number of trees
    #' @returns Number of trees in a forest
    get_num_trees = function() {
      return(self$num_trees)
    },

    #' @description
    #' Query number of features
    #' @returns Number of features in a forest model training set
    get_num_features = function() {
      return(self$num_features)
    },

    #' @description
    #' Query number of observations
    #' @returns Number of observations in a forest model training set
    get_num_observations = function() {
      return(self$num_observations)
    },

    #' @description
    #' Query root node split probability in tree prior for this ForestModelConfig object
    #' @returns Root node split probability in tree prior
    get_alpha = function() {
      return(self$alpha)
    },

    #' @description
    #' Query depth prior penalty in tree prior for this ForestModelConfig object
    #' @returns Depth prior penalty in tree prior
    get_beta = function() {
      return(self$beta)
    },

    #' @description
    #' Query root node split probability in tree prior for this ForestModelConfig object
    #' @returns Minimum number of samples in a tree leaf
    get_min_samples_leaf = function() {
      return(self$min_samples_leaf)
    },

    #' @description
    #' Query root node split probability in tree prior for this ForestModelConfig object
    #' @returns Maximum depth of any tree in the ensemble in the model
    get_max_depth = function() {
      return(self$max_depth)
    },

    #' @description
    #' Query (integer-coded) type of leaf model
    #' @returns Integer coded leaf model type
    get_leaf_model_type = function() {
      return(self$leaf_model_type)
    },

    #' @description
    #' Query scale parameter used in Gaussian leaf models for this ForestModelConfig object
    #' @returns Scale parameter used in Gaussian leaf models
    get_leaf_model_scale = function() {
      return(self$leaf_model_scale)
    },

    #' @description
    #' Query shape parameter for IG leaf models for this ForestModelConfig object
    #' @returns Shape parameter for IG leaf models
    get_variance_forest_shape = function() {
      return(self$variance_forest_shape)
    },

    #' @description
    #' Query scale parameter for IG leaf models for this ForestModelConfig object
    #' @returns Scale parameter for IG leaf models
    get_variance_forest_scale = function() {
      return(self$variance_forest_scale)
    },

    #' @description
    #' Query shape parameter for conditional gamma component of cloglog leaf models for this ForestModelConfig object
    #' @returns Shape parameter for conditional gamma component of cloglog leaf models
    get_cloglog_forest_shape = function() {
      return(self$cloglog_forest_shape)
    },

    #' @description
    #' Query rate parameter for conditional gamma component of cloglog leaf models for this ForestModelConfig object
    #' @returns Rate parameter for conditional gamma component of cloglog leaf models
    get_cloglog_forest_rate = function() {
      return(self$cloglog_forest_rate)
    },

    #' @description
    #' Query number of unique cutpoints to consider for this ForestModelConfig object
    #' @returns Number of unique cutpoints to consider
    get_cutpoint_grid_size = function() {
      return(self$cutpoint_grid_size)
    },

    #' @description
    #' Query number of features to subsample for the GFR algorithm
    #' @returns Number of features to subsample for the GFR algorithm
    get_num_features_subsample = function() {
      return(self$num_features_subsample)
    }
  )
)

#' @title Global Model Configuration Object
#'
#' @description
#' Object used to get / set global parameters and other global model
#' configuration options in the "low-level" stochtree interface.
#' The "low-level" stochtree interface enables a high degreee of sampler
#' customization, in which users employ R wrappers around C++ objects
#' like ForestDataset, Outcome, CppRng, and ForestModel to run the
#' Gibbs sampler of a BART model with custom modifications.
#' GlobalModelConfig allows users to specify / query the global parameters
#' of a model they wish to run.
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

GlobalModelConfig <- R6::R6Class(
  classname = "GlobalModelConfig",
  cloneable = FALSE,
  public = list(
    #' @field global_error_variance Global error variance parameter
    global_error_variance = NULL,

    #' Create a new GlobalModelConfig object.
    #'
    #' @param global_error_variance Global error variance parameter (default: `1.0`)
    #'
    #' @return A new GlobalModelConfig object.
    initialize = function(global_error_variance = 1.0) {
      self$global_error_variance <- global_error_variance
    },

    #' @description
    #' Update global error variance parameter
    #' @param global_error_variance Global error variance parameter
    update_global_error_variance = function(global_error_variance) {
      self$global_error_variance <- global_error_variance
    },

    #' @description
    #' Query global error variance parameter for this GlobalModelConfig object
    #' @returns Global error variance parameter
    get_global_error_variance = function() {
      return(self$global_error_variance)
    }
  )
)

#' @title Create ForestModelConfig Object
#' @description
#' Create a forest model config object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param feature_types Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)
#' @param sweep_update_indices Vector of (0-indexed) indices of trees to update in a sweep
#' @param num_trees Number of trees in the forest being sampled
#' @param num_features Number of features in training dataset
#' @param num_observations Number of observations in training dataset
#' @param variable_weights Vector specifying sampling probability for all p covariates in ForestDataset
#' @param leaf_dimension Dimension of the leaf model (default: `1`)
#' @param alpha Root node split probability in tree prior (default: `0.95`)
#' @param beta Depth prior penalty in tree prior (default: `2.0`)
#' @param min_samples_leaf Minimum number of samples in a tree leaf (default: `5`)
#' @param max_depth Maximum depth of any tree in the ensemble in the model. Setting to `-1` does not enforce any depth limits on trees. Default: `-1`.
#' @param leaf_model_type Integer specifying the leaf model type (0 = constant leaf, 1 = univariate leaf regression, 2 = multivariate leaf regression). Default: `0`.
#' @param leaf_model_scale Scale parameter used in Gaussian leaf models (can either be a scalar or a q x q matrix, where q is the dimensionality of the basis and is only >1 when `leaf_model_int = 2`). Calibrated internally as `1/num_trees`, propagated along diagonal if needed for multivariate leaf models.
#' @param variance_forest_shape Shape parameter for IG leaf models (applicable when `leaf_model_type = 3`). Default: `1`.
#' @param variance_forest_scale Scale parameter for IG leaf models (applicable when `leaf_model_type = 3`). Default: `1`.
#' @param cloglog_forest_shape Shape parameter for conditional gamma component of cloglog leaf models (applicable when `leaf_model_type = 4`). Default: `1`.
#' @param cloglog_forest_rate Rate parameter for conditional gamma component of cloglog leaf models (applicable when `leaf_model_type = 4`). Default: `1`.
#' @param cutpoint_grid_size Number of unique cutpoints to consider (default: `100`)
#' @param num_features_subsample Number of features to subsample for the GFR algorithm
#' @return ForestModelConfig object
#' @export
#'
#' @examples
#' config <- createForestModelConfig(num_trees = 10, num_features = 5, num_observations = 100)
createForestModelConfig <- function(
  feature_types = NULL,
  sweep_update_indices = NULL,
  num_trees = NULL,
  num_features = NULL,
  num_observations = NULL,
  variable_weights = NULL,
  leaf_dimension = 1,
  alpha = 0.95,
  beta = 2.0,
  min_samples_leaf = 5,
  max_depth = -1,
  leaf_model_type = 1,
  leaf_model_scale = NULL,
  variance_forest_shape = 1.0,
  variance_forest_scale = 1.0,
  cloglog_forest_shape = 2.0,
  cloglog_forest_rate = 2.0,
  cutpoint_grid_size = 100,
  num_features_subsample = NULL
) {
  return(invisible(
    (ForestModelConfig$new(
      feature_types,
      sweep_update_indices,
      num_trees,
      num_features,
      num_observations,
      variable_weights,
      leaf_dimension,
      alpha,
      beta,
      min_samples_leaf,
      max_depth,
      leaf_model_type,
      leaf_model_scale,
      variance_forest_shape,
      variance_forest_scale,
      cutpoint_grid_size,
      num_features_subsample
    ))
  ))
}

#' @title Create GlobalModelConfig Object
#' @description
#' Create a global model config object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param global_error_variance Global error variance parameter (default: `1.0`)
#' @return GlobalModelConfig object
#' @export
#'
#' @examples
#' config <- createGlobalModelConfig(global_error_variance = 100)
createGlobalModelConfig <- function(global_error_variance = 1.0) {
  return(invisible((GlobalModelConfig$new(global_error_variance))))
}

#' Data Preprocessing Routines
#' @name DataPreprocessing
#' @description
#' The functions in this group are designed to handle data preprocessing for stochastic forest models.
#' For example, factor-valued columns in data frames are either one-hot encoded or converted to integer indices
#' before the dataframe is converted to a standard matrix format for sampling.
#' This preprocessing routine defines a set of "steps" that must be repeated on out-of-sample datasets before
#' predictions can be obtained from a sampling model.
#'
#' `preprocessTrainData` preprocesses covariates for the forest sampler routines, depending on the input type.
#' DataFrames will be preprocessed based on their column types (numeric columns are not modified, ordered factors are
#' integer coded, and unordered factors are one-hot encoded). Matrices are unmodified (assuming all columns are numeric).
#' This function also records and returns a "metadata" list with preprocessing details to ensure that other datasets can be preprocessing identically.
#'
#' `preprocessPredictionData` preprocesses covariates for the forest sampler routines, based on the steps outlined in a metadata list produced by `preprocessTrainData`.
#'
#' These procedures are handled internally in the [bart()] and [bcf()] functions, but they are provided in `stochtree` as convenience functions for users writing custom samplers.
#' Furthermore, while R lists can be serialized to `RDS` format, we offer a number of JSON serialization routines for the metadata list produced by `preprocessTrainData` for
#' consistency with the broader serialization approach of stochtree (see [BARTSerialization] and [BCFSerialization]).
#'
#' Following the API for serializing `bartmodel` and `bcfmodel` objects, we can convert metadata to JSON or JSON strings via
#' `savePreprocessorToJson` and `savePreprocessorToJsonString`. Similarly, we can reload a metadata list from JSON or JSON strings
#' via `createPreprocessorFromJson` and `createPreprocessorFromJsonString`.
#' @returns
#' `preprocessTrainData` returns a list with transformed matrix data and a "metadata" list with details on the preprocessing procedures applied.
#' `preprocessPredictionData` returns a matrix reflecting the data transformations specified in the provided metadata list.
#'
#' `savePreprocessorToJson` return an object of type `CppJson`.
#' `savePreprocessorToJsonString` returns a string dump of the preprocessor's JSON representation.
#'
#' `createPreprocessorFromJson` and `createPreprocessorFromJsonString` both return metadata lists.
#' @examples
#' # Check that running the same data through `preprocessTrainData`
#' # and `preprocessPredictionData` yields the same result
#' n <- 100
#' x1 <- rnorm(n)
#' x2 <- factor(sample(1:3, n, replace = TRUE), ordered = TRUE)
#' x3 <- factor(sample(1:3, n, replace = TRUE), ordered = FALSE)
#' df1 <- data.frame(x1 = x1, x2 = x2, x3 = x3)
#' df2 <- data.frame(x1 = x1, x2 = x2, x3 = x3)
#' preprocess_train_list <- preprocessTrainData(df1)
#' df1_process <- preprocess_train_list$data
#' df1_metadata <- preprocess_train_list$metadata
#' df2_process <- preprocessPredictionData(df2, df1_metadata)
#' all.equal(df1_process, df2_process)
#'
#' # Save to in-memory JSON
#' metadata_json <- savePreprocessorToJson(df1_metadata)
#' # Save to JSON string
#' metadata_json_string <- savePreprocessorToJsonString(df1_metadata)
#'
#' # Reload metadata list from in-memory JSON object
#' metadata_roundtrip <- createPreprocessorFromJson(metadata_json)
#' # Reload metadata list from JSON string
#' metadata_roundtrip <- createPreprocessorFromJsonString(metadata_json_string)
#'
NULL
#> NULL

#' Internal function that creates a new outcome model object without validation
#'
#' @param outcome Character string specifying the outcome type.
#' @param link Character string specifying the link function.
#'
#' @return An object of class `outcome_model`.
#' @noRd
new_outcome_model <- function(outcome = "continuous", link = NULL) {
  stopifnot((is.character(outcome)))
  stopifnot(is.character(link) || is.null(link))
  if (is.null(link)) {
    if (outcome == "continuous") {
      link <- "identity"
    } else if (outcome == "binary") {
      link <- "probit"
    } else if (outcome == "ordinal") {
      link <- "cloglog"
    }
  }
  structure(
    list(
      outcome = outcome,
      link = link
    ),
    class = "outcome_model"
  )
}

#' Internal helper function validates whether an outcome model object is correctly specified
#'
#' @param object Object of type `outcome_model`
#'
#' @return An object of class `outcome_model`.
#' @noRd
validate_outcome_model <- function(object) {
  if (!inherits(object, "outcome_model")) {
    stop("Invalid outcome model")
  }
  if (!(object$outcome %in% c("continuous", "binary", "ordinal"))) {
    stop("Outcome type must be one of 'continuous', 'binary', or 'ordinal'")
  }
  if (!(object$link %in% c("identity", "probit", "cloglog"))) {
    stop("Link function must be one of 'identity', 'probit', or 'cloglog'")
  }
  if (object$outcome == "continuous" && object$link != "identity") {
    stop("Link function must be 'identity' for continuous models")
  }
  if (
    object$outcome == "binary" && !(object$link %in% c("probit", "cloglog"))
  ) {
    stop("Link function must be 'probit' or 'cloglog' for binary models")
  }
  if (object$outcome == "ordinal" && object$link != "cloglog") {
    stop("Link function must be 'cloglog' for ordinal models")
  }
  object
}

#' Create a New Outcome Model Object
#'
#' @param outcome Character string specifying the outcome type.
#' @param link Character string specifying the link function.
#'
#' @return An object of class `outcome_model`.
#' @export
#'
#' @examples
#' my_model <- OutcomeModel(outcome = "continuous", link = "identity")
OutcomeModel <- function(outcome = "continuous", link = NULL) {
  validate_outcome_model(new_outcome_model(outcome, link))
}

#' Preprocess a parameter list, overriding defaults with any provided parameters.
#'
#' @param default_params List of parameters with default values set.
#' @param user_params (Optional) User-supplied overrides to `default_params`.
#' @noRd
#'
#' @return Parameter list with defaults overriden by values supplied in `user_params`
preprocessParams <- function(default_params, user_params = NULL) {
  # Override defaults from general_params
  if (!is.null(user_params)) {
    for (key in names(user_params)) {
      if (key %in% names(default_params)) {
        val <- user_params[[key]]
        if (!is.null(val)) default_params[[key]] <- val
      }
    }
  }

  # Return result
  return(default_params)
}

#' @title Preprocess Covariates for Model Training
#' @param input_data Covariates, provided as either a dataframe or a matrix
#' @export
#' @rdname DataPreprocessing
preprocessTrainData <- function(input_data) {
  # Input checks
  if ((!is.matrix(input_data)) && (!is.data.frame(input_data))) {
    stop("Covariates provided must be a dataframe or matrix")
  }

  # Routing the correct preprocessing function
  if (is.matrix(input_data)) {
    output <- preprocessTrainMatrix(input_data)
  } else {
    output <- preprocessTrainDataFrame(input_data)
  }

  return(output)
}

#' @title Preprocess Covariates for Model Prediction
#' @param input_data Covariates, provided as either a dataframe or a matrix
#' @param metadata List containing information on variables, including train set
#' categories for categorical variables
#' @export
#' @rdname DataPreprocessing
#'
#' @examples
#' cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#' metadata <- list(num_ordered_cat_vars = 0, num_unordered_cat_vars = 0,
#'                  num_numeric_vars = 3, numeric_vars = c("x1", "x2", "x3"))
#' X_preprocessed <- preprocessPredictionData(cov_df, metadata)
preprocessPredictionData <- function(input_data, metadata) {
  # Input checks
  if ((!is.matrix(input_data)) && (!is.data.frame(input_data))) {
    stop("Covariates provided must be a dataframe or matrix")
  }

  # Routing the correct preprocessing function
  if (is.matrix(input_data)) {
    X <- preprocessPredictionMatrix(input_data, metadata)
  } else {
    X <- preprocessPredictionDataFrame(input_data, metadata)
  }

  return(X)
}

#' Preprocess a matrix of covariate values, assuming all columns are numeric.
#' Returns a list including a matrix of preprocessed covariate values and associated tracking.
#'
#' @param input_matrix Covariate matrix.
#'
#' @return List with preprocessed (unmodified) data and details on the number of each type
#' of variable, unique categories associated with categorical variables, and the
#' vector of feature types needed for calls to BART and BCF.
#' @noRd
#'
#' @examples
#' cov_mat <- matrix(1:12, ncol = 3)
#' preprocess_list <- preprocessTrainMatrix(cov_mat)
#' X <- preprocess_list$X
preprocessTrainMatrix <- function(input_matrix) {
  # Input checks
  if (!is.matrix(input_matrix)) {
    stop("covariates provided must be a matrix")
  }

  # Unpack metadata (assuming all variables are numeric)
  names(input_matrix) <- paste0("x", 1:ncol(input_matrix))
  df_vars <- names(input_matrix)
  num_ordered_cat_vars <- 0
  num_unordered_cat_vars <- 0
  num_numeric_vars <- ncol(input_matrix)
  numeric_vars <- names(input_matrix)
  feature_types <- rep(0, ncol(input_matrix))

  # Unpack data
  X <- input_matrix

  # Aggregate results into a list
  metadata <- list(
    feature_types = feature_types,
    num_ordered_cat_vars = num_ordered_cat_vars,
    num_unordered_cat_vars = num_unordered_cat_vars,
    num_numeric_vars = num_numeric_vars,
    numeric_vars = numeric_vars,
    original_var_indices = 1:num_numeric_vars
  )
  output <- list(
    data = X,
    metadata = metadata
  )

  return(output)
}

#' Preprocess a matrix of covariate values, assuming all columns are numeric.
#'
#' @param input_matrix Covariate matrix.
#' @param metadata List containing information on variables, including train set
#' categories for categorical variables
#'
#' @return Preprocessed data with categorical variables appropriately preprocessed
#' @noRd
#'
#' @examples
#' cov_mat <- matrix(c(1:5, 5:1, 6:10), ncol = 3)
#' metadata <- list(num_ordered_cat_vars = 0, num_unordered_cat_vars = 0,
#'                  num_numeric_vars = 3, numeric_vars = c("x1", "x2", "x3"))
#' X_preprocessed <- preprocessPredictionMatrix(cov_mat, metadata)
preprocessPredictionMatrix <- function(input_matrix, metadata) {
  # Input checks
  if (!is.matrix(input_matrix)) {
    stop("covariates provided must be a matrix")
  }
  if (!(ncol(input_matrix) == metadata$num_numeric_vars)) {
    stop(
      "Prediction set covariates have inconsistent dimension from train set covariates"
    )
  }

  return(input_matrix)
}

#' Preprocess a dataframe of covariate values, converting categorical variables
#' to integers and one-hot encoding if need be. Returns a list including a
#' matrix of preprocessed covariate values and associated tracking.
#'
#' @param input_df Dataframe of covariates. Users must pre-process any
#' categorical variables as factors (ordered for ordered categorical).
#' @noRd
#'
#' @return List with preprocessed data and details on the number of each type
#' of variable, unique categories associated with categorical variables, and the
#' vector of feature types needed for calls to BART and BCF.
preprocessTrainDataFrame <- function(input_df) {
  # Input checks / details
  if (!is.data.frame(input_df)) {
    stop("covariates provided must be a data frame")
  }
  df_vars <- names(input_df)

  # Detect ordered and unordered categorical variables

  # First, ordered categorical: users must have explicitly
  # converted this to a factor with ordered = TRUE
  factor_mask <- sapply(input_df, is.factor)
  ordered_mask <- sapply(input_df, is.ordered)
  ordered_cat_matches <- factor_mask & ordered_mask
  ordered_cat_vars <- df_vars[ordered_cat_matches]
  ordered_cat_var_inds <- unname(which(ordered_cat_matches))
  num_ordered_cat_vars <- length(ordered_cat_vars)
  if (num_ordered_cat_vars > 0) {
    ordered_cat_df <- input_df[, ordered_cat_vars, drop = FALSE]
  }

  # Next, unordered categorical: we will convert character
  # columns but not integer columns (users must explicitly
  # convert these to factor)
  character_mask <- sapply(input_df, is.character)
  unordered_cat_matches <- (factor_mask & (!ordered_mask)) | character_mask
  unordered_cat_vars <- df_vars[unordered_cat_matches]
  unordered_cat_var_inds <- unname(which(unordered_cat_matches))
  num_unordered_cat_vars <- length(unordered_cat_vars)
  if (num_unordered_cat_vars > 0) {
    unordered_cat_df <- input_df[, unordered_cat_vars, drop = FALSE]
  }

  # Numeric variables
  numeric_matches <- (!ordered_cat_matches) & (!unordered_cat_matches)
  numeric_vars <- df_vars[numeric_matches]
  numeric_var_inds <- unname(which(numeric_matches))
  num_numeric_vars <- length(numeric_vars)
  if (num_numeric_vars > 0) {
    numeric_df <- input_df[, numeric_vars, drop = FALSE]
  }

  # Empty outputs
  X <- double(0)
  unordered_unique_levels <- list()
  ordered_unique_levels <- list()
  feature_types <- integer(0)
  original_var_indices <- integer(0)

  # First, extract the numeric covariates
  if (num_numeric_vars > 0) {
    Xnum <- double(0)
    for (i in 1:ncol(numeric_df)) {
      stopifnot(is.numeric(numeric_df[, i]))
      Xnum <- cbind(Xnum, numeric_df[, i])
    }
    X <- cbind(X, unname(Xnum))
    feature_types <- c(feature_types, rep(0, ncol(Xnum)))
    original_var_indices <- c(original_var_indices, numeric_var_inds)
  }

  # Next, run some simple preprocessing on the ordered categorical covariates
  if (num_ordered_cat_vars > 0) {
    Xordcat <- double(0)
    for (i in 1:ncol(ordered_cat_df)) {
      var_name <- names(ordered_cat_df)[i]
      preprocess_list <- orderedCatInitializeAndPreprocess(ordered_cat_df[,
        i
      ])
      ordered_unique_levels[[var_name]] <- preprocess_list$unique_levels
      Xordcat <- cbind(Xordcat, preprocess_list$x_preprocessed)
    }
    X <- cbind(X, unname(Xordcat))
    feature_types <- c(feature_types, rep(1, ncol(Xordcat)))
    original_var_indices <- c(original_var_indices, ordered_cat_var_inds)
  }

  # Finally, one-hot encode the unordered categorical covariates
  if (num_unordered_cat_vars > 0) {
    one_hot_mats <- list()
    for (i in 1:ncol(unordered_cat_df)) {
      var_name <- names(unordered_cat_df)[i]
      encode_list <- oneHotInitializeAndEncode(unordered_cat_df[, i])
      unordered_unique_levels[[var_name]] <- encode_list$unique_levels
      one_hot_mats[[var_name]] <- encode_list$Xtilde
      one_hot_var <- rep(
        unordered_cat_var_inds[i],
        ncol(encode_list$Xtilde)
      )
      original_var_indices <- c(original_var_indices, one_hot_var)
    }
    Xcat <- do.call(cbind, one_hot_mats)
    X <- cbind(X, unname(Xcat))
    feature_types <- c(feature_types, rep(1, ncol(Xcat)))
  }

  # Aggregate results into a list
  metadata <- list(
    feature_types = feature_types,
    num_ordered_cat_vars = num_ordered_cat_vars,
    num_unordered_cat_vars = num_unordered_cat_vars,
    num_numeric_vars = num_numeric_vars,
    original_var_indices = original_var_indices
  )
  if (num_ordered_cat_vars > 0) {
    metadata[["ordered_cat_vars"]] = ordered_cat_vars
    metadata[["ordered_unique_levels"]] = ordered_unique_levels
  }
  if (num_unordered_cat_vars > 0) {
    metadata[["unordered_cat_vars"]] = unordered_cat_vars
    metadata[["unordered_unique_levels"]] = unordered_unique_levels
  }
  if (num_numeric_vars > 0) {
    metadata[["numeric_vars"]] = numeric_vars
  }
  output <- list(
    data = X,
    metadata = metadata
  )

  return(output)
}

#' Preprocess a dataframe of covariate values, converting categorical variables
#' to integers and one-hot encoding if need be.
#'
#' @param input_df Dataframe of covariates. Users must pre-process any
#' categorical variables as factors (ordered for ordered categorical).
#' @param metadata List containing information on variables, including train set
#' categories for categorical variables
#'
#' @return Preprocessed data with categorical variables appropriately preprocessed
#' @noRd
#'
#' @examples
#' cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#' metadata <- list(num_ordered_cat_vars = 0, num_unordered_cat_vars = 0,
#'                  num_numeric_vars = 3, numeric_vars = c("x1", "x2", "x3"))
#' X_preprocessed <- preprocessPredictionDataFrame(cov_df, metadata)
preprocessPredictionDataFrame <- function(input_df, metadata) {
  if (!is.data.frame(input_df)) {
    stop("covariates provided must be a data frame")
  }
  df_vars <- names(input_df)
  num_ordered_cat_vars <- metadata$num_ordered_cat_vars
  num_unordered_cat_vars <- metadata$num_unordered_cat_vars
  num_numeric_vars <- metadata$num_numeric_vars

  if (num_ordered_cat_vars > 0) {
    ordered_cat_vars <- metadata$ordered_cat_vars
    ordered_cat_df <- input_df[, ordered_cat_vars, drop = FALSE]
  }
  if (num_unordered_cat_vars > 0) {
    unordered_cat_vars <- metadata$unordered_cat_vars
    unordered_cat_df <- input_df[, unordered_cat_vars, drop = FALSE]
  }
  if (num_numeric_vars > 0) {
    numeric_vars <- metadata$numeric_vars
    numeric_df <- input_df[, numeric_vars, drop = FALSE]
  }

  # Empty outputs
  X <- double(0)

  # First, extract the numeric covariates
  if (num_numeric_vars > 0) {
    Xnum <- double(0)
    for (i in 1:ncol(numeric_df)) {
      stopifnot(is.numeric(numeric_df[, i]))
      Xnum <- cbind(Xnum, numeric_df[, i])
    }
    X <- cbind(X, unname(Xnum))
  }

  # Next, run some simple preprocessing on the ordered categorical covariates
  if (num_ordered_cat_vars > 0) {
    Xordcat <- double(0)
    for (i in 1:ncol(ordered_cat_df)) {
      var_name <- names(ordered_cat_df)[i]
      x_preprocessed <- orderedCatPreprocess(
        ordered_cat_df[, i],
        metadata$ordered_unique_levels[[var_name]]
      )
      Xordcat <- cbind(Xordcat, x_preprocessed)
    }
    X <- cbind(X, unname(Xordcat))
  }

  # Finally, one-hot encode the unordered categorical covariates
  if (num_unordered_cat_vars > 0) {
    one_hot_mats <- list()
    for (i in 1:ncol(unordered_cat_df)) {
      var_name <- names(unordered_cat_df)[i]
      Xtilde <- oneHotEncode(
        unordered_cat_df[, i],
        metadata$unordered_unique_levels[[var_name]]
      )
      one_hot_mats[[var_name]] <- Xtilde
    }
    Xcat <- do.call(cbind, one_hot_mats)
    X <- cbind(X, unname(Xcat))
  }

  return(X)
}

#' @title Convert Covariate Preprocessor to [CppJson]
#' @param object List containing information on variables, including train set
#' categories for categorical variables
#' @export
#' @rdname DataPreprocessing
savePreprocessorToJson <- function(object) {
  jsonobj <- createCppJson()
  if (is.null(object$feature_types)) {
    stop("This covariate preprocessor has not yet been fit")
  }

  # Add internal scalars
  jsonobj$add_integer("num_numeric_vars", object$num_numeric_vars)
  jsonobj$add_integer("num_ordered_cat_vars", object$num_ordered_cat_vars)
  jsonobj$add_integer("num_unordered_cat_vars", object$num_unordered_cat_vars)

  # Add internal vectors
  jsonobj$add_vector("feature_types", object$feature_types)
  jsonobj$add_vector("original_var_indices", object$original_var_indices)
  if (object$num_numeric_vars > 0) {
    jsonobj$add_string_vector("numeric_vars", object$numeric_vars)
  }
  if (object$num_ordered_cat_vars > 0) {
    jsonobj$add_string_vector("ordered_cat_vars", object$ordered_cat_vars)
    for (i in 1:object$num_ordered_cat_vars) {
      var_key <- names(object$ordered_unique_levels)[i]
      jsonobj$add_string(
        paste0("key_", i),
        var_key,
        "ordered_unique_level_keys"
      )
      jsonobj$add_string_vector(
        var_key,
        object$ordered_unique_levels[[i]],
        "ordered_unique_levels"
      )
    }
  }
  if (object$num_unordered_cat_vars > 0) {
    jsonobj$add_string_vector(
      "unordered_cat_vars",
      object$unordered_cat_vars
    )
    for (i in 1:object$num_unordered_cat_vars) {
      var_key <- names(object$unordered_unique_levels)[i]
      jsonobj$add_string(
        paste0("key_", i),
        var_key,
        "unordered_unique_level_keys"
      )
      jsonobj$add_string_vector(
        var_key,
        object$unordered_unique_levels[[i]],
        "unordered_unique_levels"
      )
    }
  }

  return(jsonobj)
}

#' @title Convert Covariate Preprocessor to JSON String
#' @param object List containing information on variables, including train set
#' categories for categorical variables
#' @export
#' @rdname DataPreprocessing
savePreprocessorToJsonString <- function(object) {
  # Convert to Json
  jsonobj <- savePreprocessorToJson(object)

  # Dump to string
  return(jsonobj$return_json_string())
}

#' @title Reload Covariate Preprocessor from JSON String
#' @param json_object in-memory wrapper around JSON C++ object containing covariate preprocessor metadata
#' @export
#' @rdname DataPreprocessing
createPreprocessorFromJson <- function(json_object) {
  # Initialize the metadata list
  metadata <- list()

  # Unpack internal scalars
  metadata[["num_numeric_vars"]] <- json_object$get_integer(
    "num_numeric_vars"
  )
  metadata[["num_ordered_cat_vars"]] <- json_object$get_integer(
    "num_ordered_cat_vars"
  )
  metadata[["num_unordered_cat_vars"]] <- json_object$get_integer(
    "num_unordered_cat_vars"
  )

  # Unpack internal vectors
  metadata[["feature_types"]] <- json_object$get_vector("feature_types")
  metadata[["original_var_indices"]] <- json_object$get_vector(
    "original_var_indices"
  )
  if (metadata$num_numeric_vars > 0) {
    metadata[["numeric_vars"]] <- json_object$get_string_vector(
      "numeric_vars"
    )
  }
  if (metadata$num_ordered_cat_vars > 0) {
    metadata[["ordered_cat_vars"]] <- json_object$get_string_vector(
      "ordered_cat_vars"
    )
    ordered_unique_levels <- list()
    for (i in 1:metadata$num_ordered_cat_vars) {
      var_key <- json_object$get_string(
        paste0("key_", i),
        "ordered_unique_level_keys"
      )
      ordered_unique_levels[[var_key]] <- json_object$get_string_vector(
        var_key,
        "ordered_unique_levels"
      )
    }
    metadata[["ordered_unique_levels"]] <- ordered_unique_levels
  }
  if (metadata$num_unordered_cat_vars > 0) {
    metadata[["unordered_cat_vars"]] <- json_object$get_string_vector(
      "unordered_cat_vars"
    )
    unordered_unique_levels <- list()
    for (i in 1:metadata$num_unordered_cat_vars) {
      var_key <- json_object$get_string(
        paste0("key_", i),
        "unordered_unique_level_keys"
      )
      unordered_unique_levels[[var_key]] <- json_object$get_string_vector(
        var_key,
        "unordered_unique_levels"
      )
    }
    metadata[["unordered_unique_levels"]] <- unordered_unique_levels
  }

  return(metadata)
}

#' @title Reload Covariate Preprocessor from JSON String
#' @param json_string in-memory JSON string containing covariate preprocessor metadata
#' @export
#' @rdname DataPreprocessing
createPreprocessorFromJsonString <- function(json_string) {
  # Load a `CppJson` object from string
  preprocessor_json <- createCppJsonString(json_string)

  # Create and return the BCF object
  preprocessor_object <- createPreprocessorFromJson(preprocessor_json)

  return(preprocessor_object)
}

#' Preprocess a dataframe of covariate values, converting categorical variables
#' to integers and one-hot encoding if need be. Returns a list including a
#' matrix of preprocessed covariate values and associated tracking.
#'
#' @param input_data Dataframe or matrix of covariates. Users may pre-process any
#' categorical variables as factors but it is not necessary.
#' @param ordered_cat_vars (Optional) Vector of names of ordered categorical variables, or vector of column indices if `input_data` is a matrix.
#' @param unordered_cat_vars (Optional) Vector of names of unordered categorical variables, or vector of column indices if `input_data` is a matrix.
#'
#' @return List with preprocessed data and details on the number of each type
#' of variable, unique categories associated with categorical variables, and the
#' vector of feature types needed for calls to BART and BCF.
#' @noRd
#'
#' @examples
#' cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#' preprocess_list <- createForestCovariates(cov_df)
#' X <- preprocess_list$X
createForestCovariates <- function(
  input_data,
  ordered_cat_vars = NULL,
  unordered_cat_vars = NULL
) {
  if (is.matrix(input_data)) {
    input_df <- as.data.frame(input_data)
    names(input_df) <- paste0("x", 1:ncol(input_data))
    if (!is.null(ordered_cat_vars)) {
      if (is.numeric(ordered_cat_vars)) {
        ordered_cat_vars <- paste0("x", as.integer(ordered_cat_vars))
      }
    }
    if (!is.null(unordered_cat_vars)) {
      if (is.numeric(unordered_cat_vars)) {
        unordered_cat_vars <- paste0(
          "x",
          as.integer(unordered_cat_vars)
        )
      }
    }
  } else if (is.data.frame(input_data)) {
    input_df <- input_data
  } else {
    stop("input_data must be either a matrix or a data frame")
  }
  df_vars <- names(input_df)
  if (is.null(ordered_cat_vars)) {
    ordered_cat_matches <- rep(FALSE, length(df_vars))
  } else {
    ordered_cat_matches <- df_vars %in% ordered_cat_vars
  }
  if (is.null(unordered_cat_vars)) {
    unordered_cat_matches <- rep(FALSE, length(df_vars))
  } else {
    unordered_cat_matches <- df_vars %in% unordered_cat_vars
  }
  numeric_matches <- ((!ordered_cat_matches) & (!unordered_cat_matches))
  ordered_cat_vars <- df_vars[ordered_cat_matches]
  unordered_cat_vars <- df_vars[unordered_cat_matches]
  numeric_vars <- df_vars[numeric_matches]
  num_ordered_cat_vars <- length(ordered_cat_vars)
  num_unordered_cat_vars <- length(unordered_cat_vars)
  num_numeric_vars <- length(numeric_vars)
  if (num_ordered_cat_vars > 0) {
    ordered_cat_df <- input_df[, ordered_cat_vars, drop = FALSE]
  }
  if (num_unordered_cat_vars > 0) {
    unordered_cat_df <- input_df[, unordered_cat_vars, drop = FALSE]
  }
  if (num_numeric_vars > 0) {
    numeric_df <- input_df[, numeric_vars, drop = FALSE]
  }

  # Empty outputs
  X <- double(0)
  unordered_unique_levels <- list()
  ordered_unique_levels <- list()
  feature_types <- integer(0)

  # First, extract the numeric covariates
  if (num_numeric_vars > 0) {
    Xnum <- double(0)
    for (i in 1:ncol(numeric_df)) {
      stopifnot(is.numeric(numeric_df[, i]))
      Xnum <- cbind(Xnum, numeric_df[, i])
    }
    X <- cbind(X, unname(Xnum))
    feature_types <- c(feature_types, rep(0, ncol(Xnum)))
  }

  # Next, run some simple preprocessing on the ordered categorical covariates
  if (num_ordered_cat_vars > 0) {
    Xordcat <- double(0)
    for (i in 1:ncol(ordered_cat_df)) {
      var_name <- names(ordered_cat_df)[i]
      preprocess_list <- orderedCatInitializeAndPreprocess(ordered_cat_df[,
        i
      ])
      ordered_unique_levels[[var_name]] <- preprocess_list$unique_levels
      Xordcat <- cbind(Xordcat, preprocess_list$x_preprocessed)
    }
    X <- cbind(X, unname(Xordcat))
    feature_types <- c(feature_types, rep(1, ncol(Xordcat)))
  }

  # Finally, one-hot encode the unordered categorical covariates
  if (num_unordered_cat_vars > 0) {
    one_hot_mats <- list()
    for (i in 1:ncol(unordered_cat_df)) {
      var_name <- names(unordered_cat_df)[i]
      encode_list <- oneHotInitializeAndEncode(unordered_cat_df[, i])
      unordered_unique_levels[[var_name]] <- encode_list$unique_levels
      one_hot_mats[[var_name]] <- encode_list$Xtilde
    }
    Xcat <- do.call(cbind, one_hot_mats)
    X <- cbind(X, unname(Xcat))
    feature_types <- c(feature_types, rep(1, ncol(Xcat)))
  }

  # Aggregate results into a list
  metadata <- list(
    feature_types = feature_types,
    num_ordered_cat_vars = num_ordered_cat_vars,
    num_unordered_cat_vars = num_unordered_cat_vars,
    num_numeric_vars = num_numeric_vars
  )
  if (num_ordered_cat_vars > 0) {
    metadata[["ordered_cat_vars"]] = ordered_cat_vars
    metadata[["ordered_unique_levels"]] = ordered_unique_levels
  }
  if (num_unordered_cat_vars > 0) {
    metadata[["unordered_cat_vars"]] = unordered_cat_vars
    metadata[["unordered_unique_levels"]] = unordered_unique_levels
  }
  if (num_numeric_vars > 0) {
    metadata[["numeric_vars"]] = numeric_vars
  }
  output <- list(
    data = X,
    metadata = metadata
  )

  return(output)
}

#' Preprocess a dataframe of covariate values, converting categorical variables
#' to integers and one-hot encoding if need be. Returns a list including a
#' matrix of preprocessed covariate values and associated tracking.
#'
#' @param input_data Dataframe or matrix of covariates. Users may pre-process any
#' categorical variables as factors but it is not necessary.
#' @param metadata List containing information on variables, including train set
#' categories for categorical variables
#'
#' @return Preprocessed data with categorical variables appropriately preprocessed
#' @noRd
#'
#' @examples
#' cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#' metadata <- list(num_ordered_cat_vars = 0, num_unordered_cat_vars = 0,
#'                  num_numeric_vars = 3, numeric_vars = c("x1", "x2", "x3"))
#' X_preprocessed <- createForestCovariatesFromMetadata(cov_df, metadata)
createForestCovariatesFromMetadata <- function(input_data, metadata) {
  if (is.matrix(input_data)) {
    input_df <- as.data.frame(input_data)
    names(input_df) <- paste0("x", 1:ncol(input_data))
  } else if (is.data.frame(input_data)) {
    input_df <- input_data
  } else {
    stop("input_data must be either a matrix or a data frame")
  }
  df_vars <- names(input_df)
  num_ordered_cat_vars <- metadata$num_ordered_cat_vars
  num_unordered_cat_vars <- metadata$num_unordered_cat_vars
  num_numeric_vars <- metadata$num_numeric_vars

  if (num_ordered_cat_vars > 0) {
    ordered_cat_vars <- metadata$ordered_cat_vars
    ordered_cat_df <- input_df[, ordered_cat_vars, drop = FALSE]
  }
  if (num_unordered_cat_vars > 0) {
    unordered_cat_vars <- metadata$unordered_cat_vars
    unordered_cat_df <- input_df[, unordered_cat_vars, drop = FALSE]
  }
  if (num_numeric_vars > 0) {
    numeric_vars <- metadata$numeric_vars
    numeric_df <- input_df[, numeric_vars, drop = FALSE]
  }

  # Empty outputs
  X <- double(0)

  # First, extract the numeric covariates
  if (num_numeric_vars > 0) {
    Xnum <- double(0)
    for (i in 1:ncol(numeric_df)) {
      stopifnot(is.numeric(numeric_df[, i]))
      Xnum <- cbind(Xnum, numeric_df[, i])
    }
    X <- cbind(X, unname(Xnum))
  }

  # Next, run some simple preprocessing on the ordered categorical covariates
  if (num_ordered_cat_vars > 0) {
    Xordcat <- double(0)
    for (i in 1:ncol(ordered_cat_df)) {
      var_name <- names(ordered_cat_df)[i]
      x_preprocessed <- orderedCatPreprocess(
        ordered_cat_df[, i],
        metadata$ordered_unique_levels[[var_name]]
      )
      Xordcat <- cbind(Xordcat, x_preprocessed)
    }
    X <- cbind(X, unname(Xordcat))
  }

  # Finally, one-hot encode the unordered categorical covariates
  if (num_unordered_cat_vars > 0) {
    one_hot_mats <- list()
    for (i in 1:ncol(unordered_cat_df)) {
      var_name <- names(unordered_cat_df)[i]
      Xtilde <- oneHotEncode(
        unordered_cat_df[, i],
        metadata$unordered_unique_levels[[var_name]]
      )
      one_hot_mats[[var_name]] <- Xtilde
    }
    Xcat <- do.call(cbind, one_hot_mats)
    X <- cbind(X, unname(Xcat))
  }

  return(X)
}

#' Convert a vector of unordered categorical data (either numeric or character
#' labels) to a "one-hot" encoded matrix in which a 1 in a column indicates
#' the presence of the relevant category.
#'
#' To allow for prediction on "unseen" categories in a test dataset, this
#' procedure pads the one-hot matrix with a blank "other" column.
#' Test set observations that contain categories not in `levels(factor(x_input))`
#' will all be mapped to this column.
#'
#' @param x_input Vector of unordered categorical data (typically either strings
#' integers, but this function also accepts floating point data).
#'
#' @return List containing a binary one-hot matrix and the unique levels of the
#' input variable. These unique levels are used in the BCF and BART functions.
#' @noRd
#'
#' @examples
#' x <- c("a","c","b","c","d","a","c","a","b","d")
#' x_onehot <- oneHotInitializeAndEncode(x)
oneHotInitializeAndEncode <- function(x_input) {
  stopifnot((is.null(dim(x_input)) && length(x_input) > 0))
  if (is.factor(x_input) && is.ordered(x_input)) {
    warning("One-hot encoding an ordered categorical variable")
  }
  x_factor <- factor(x_input)
  unique_levels <- levels(x_factor)
  Xtilde <- cbind(unname(model.matrix(~ 0 + x_factor)), 0)
  output <- list(Xtilde = Xtilde, unique_levels = unique_levels)
  return(output)
}

#' Convert a vector of unordered categorical data (either numeric or character
#' labels) to a "one-hot" encoded matrix in which a 1 in a column indicates
#' the presence of the relevant category.
#'
#' This procedure assumes that a reference set of observations for this variable
#' (typically a training set that was used to sample a forest) has already been
#' one-hot encoded and that the unique levels of the training set variable are
#' available (and passed as `unique_levels`). Test set observations that contain
#' categories not in `unique_levels` will all be mapped to the last column of
#' this matrix
#'
#' @param x_input Vector of unordered categorical data (typically either strings
#' integers, but this function also accepts floating point data).
#' @param unique_levels Unique values of the categorical variable used to create
#' the initial one-hot matrix (typically a training set)
#'
#' @return Binary one-hot matrix
#' @noRd
#'
#' @examples
#' x <- sample(1:8, 100, TRUE)
#' x_test <- sample(1:9, 10, TRUE)
#' x_onehot <- oneHotEncode(x_test, levels(factor(x)))
oneHotEncode <- function(x_input, unique_levels) {
  stopifnot((is.null(dim(x_input)) && length(x_input) > 0))
  stopifnot((is.null(dim(unique_levels)) && length(unique_levels) > 0))
  num_unique_levels <- length(unique_levels)
  in_sample <- x_input %in% unique_levels
  out_of_sample <- !(x_input %in% unique_levels)
  has_out_of_sample <- sum(out_of_sample) > 0
  if (has_out_of_sample) {
    x_factor_insample <- factor(x_input[in_sample], levels = unique_levels)
    Xtilde <- matrix(
      0,
      nrow = length(x_input),
      ncol = num_unique_levels + 1
    )
    Xtilde_insample <- cbind(
      unname(model.matrix(~ 0 + x_factor_insample)),
      0
    )
    Xtilde_out_of_sample <- cbind(
      matrix(0, nrow = sum(out_of_sample), ncol = num_unique_levels),
      1
    )
    Xtilde[in_sample, ] <- Xtilde_insample
    Xtilde[out_of_sample, ] <- Xtilde_out_of_sample
  } else {
    x_factor <- factor(x_input, levels = unique_levels)
    Xtilde <- cbind(unname(model.matrix(~ 0 + x_factor)), 0)
  }
  return(Xtilde)
}

#' Run some simple preprocessing of ordered categorical variables, converting
#' ordered levels to integers if necessary, and storing the unique levels of a
#' variable.
#'
#' @param x_input Vector of ordered categorical data. If the data is not already
#' stored as an ordered factor, it will be converted to one using the default
#' sort order.
#'
#' @return List containing a preprocessed vector of integer-converted ordered
#' categorical observations and the unique level of the original ordered
#' categorical feature.
#' @noRd
#'
#' @examples
#' x <- c("1. Strongly disagree", "3. Neither agree nor disagree", "2. Disagree",
#'        "4. Agree", "3. Neither agree nor disagree", "5. Strongly agree", "4. Agree")
#' preprocess_list <- orderedCatInitializeAndPreprocess(x)
#' x_preprocessed <- preprocess_list$x_preprocessed
orderedCatInitializeAndPreprocess <- function(x_input) {
  stopifnot((is.null(dim(x_input)) && length(x_input) > 0))
  already_ordered_factor <- (is.factor(x_input)) && (is.ordered(x_input))
  if (already_ordered_factor) {
    x_preprocessed <- as.integer(x_input)
    unique_levels <- levels(x_input)
  } else {
    x_factor <- factor(x_input, ordered = TRUE)
    x_preprocessed <- as.integer(x_factor)
    unique_levels <- levels(x_factor)
  }
  return(list(x_preprocessed = x_preprocessed, unique_levels = unique_levels))
}

#' Run some simple preprocessing of ordered categorical variables, converting
#' ordered levels to integers if necessary, and storing the unique levels of a
#' variable.
#'
#' @param x_input Vector of ordered categorical data. If the data is not already
#' stored as an ordered factor, it will be converted to one using the default
#' sort order.
#' @param unique_levels Vector of unique levels for a categorical feature.
#' @param var_name (Optional) Name of variable.
#'
#' @return List containing a preprocessed vector of integer-converted ordered
#' categorical observations and the unique level of the original ordered
#' categorical feature.
#' @noRd
#'
#' @examples
#' x_levels <- c("1. Strongly disagree", "2. Disagree",
#'               "3. Neither agree nor disagree",
#'               "4. Agree", "5. Strongly agree")
#' x <- c("1. Strongly disagree", "3. Neither agree nor disagree", "2. Disagree",
#'        "4. Agree", "3. Neither agree nor disagree", "5. Strongly agree", "4. Agree")
#' x_processed <- orderedCatPreprocess(x, x_levels)
orderedCatPreprocess <- function(x_input, unique_levels, var_name = NULL) {
  stopifnot((is.null(dim(x_input)) && length(x_input) > 0))
  stopifnot((is.null(dim(unique_levels)) && length(unique_levels) > 0))
  already_ordered_factor <- (is.factor(x_input)) && (is.ordered(x_input))
  if (already_ordered_factor) {
    # Run time checks
    levels_not_in_reflist <- !(levels(x_input) %in% unique_levels)
    if (sum(levels_not_in_reflist) > 0) {
      if (!is.null(var_name)) {
        warning_message <- paste0(
          "Variable ",
          var_name,
          " includes ordered categorical levels not included in the original training set"
        )
      } else {
        warning_message <- paste0(
          "Variable includes ordered categorical levels not included in the original training set"
        )
      }
      warning(warning_message)
    }
    # Preprocessing
    x_string <- as.character(x_input)
    x_factor <- factor(x_string, unique_levels, ordered = TRUE)
    x_preprocessed <- as.integer(x_factor)
    x_preprocessed[is.na(x_preprocessed)] <- length(unique_levels) + 1
  } else {
    x_factor <- factor(x_input, ordered = TRUE)
    # Run time checks
    levels_not_in_reflist <- !(levels(x_factor) %in% unique_levels)
    if (sum(levels_not_in_reflist) > 0) {
      if (!is.null(var_name)) {
        warning_message <- paste0(
          "Variable ",
          var_name,
          " includes ordered categorical levels not included in the original training set"
        )
      } else {
        warning_message <- paste0(
          "Variable includes ordered categorical levels not included in the original training set"
        )
      }
      warning(warning_message)
    }
    # Preprocessing
    x_string <- as.character(x_input)
    x_factor <- factor(x_string, unique_levels, ordered = TRUE)
    x_preprocessed <- as.integer(x_factor)
    x_preprocessed[is.na(x_preprocessed)] <- length(unique_levels) + 1
  }
  return(x_preprocessed)
}

#' Convert scalar input to vector of dimension `output_size`,
#' or check that input array is equivalent to a vector of dimension `output_size`.
#'
#' @param input Input to be converted to a vector (or passed through as-is)
#' @param output_size Intended size of the output vector
#' @return A vector of length `output_size`
#' @noRd
expand_dims_1d <- function(input, output_size) {
  if (length(input) == 1) {
    output <- rep(input, output_size)
  } else if (is.numeric(input)) {
    if (length(input) != output_size) {
      stop("`input` must be a 1D numpy array with `output_size` elements")
    }
    output <- input
  } else {
    stop(
      "`input` must be either a 1D numpy array or a scalar that can be repeated `output_size` times"
    )
  }
  return(output)
}

#' Ensures that input is propagated appropriately to a matrix of dimension `output_rows` x `output_cols`.
#'
#' Ensures that input is propagated appropriately to a matrix of dimension `output_rows` x `output_cols`.
#' Handles the following cases:
#'  1. `input` is a scalar: output is simply a (`output_rows`, `output_cols`) matrix with `input` repeated for each element
#'  2. `input` is a vector of length `output_rows`: output is a (`output_rows`, `output_cols`) array with `input` broadcast across each of `output_cols` columns
#'  3. `input` is a vector of length `output_cols`: output is a (`output_rows`, `output_cols`) array with `input` broadcast across each of `output_rows` rows
#'  4. `input` is a matrix of dimension (`output_rows`, `output_cols`): input is passed through as-is
#' All other cases throw an error.
#'
#' @param input Input to be converted to a matrix (or passed through as-is)
#' @param output_rows Intended number of rows in the output array
#' @param output_cols Intended number of columns in the output array
#' @return A matrix of dimension `output_rows` x `output_cols`
#' @noRd
expand_dims_2d <- function(input, output_rows, output_cols) {
  if (length(input) == 1) {
    output <- matrix(
      rep(input, output_rows * output_cols),
      ncol = output_cols
    )
  } else if (is.numeric(input)) {
    if (length(input) == output_cols) {
      output <- matrix(
        rep(input, output_rows),
        nrow = output_rows,
        byrow = T
      )
    } else if (length(input) == output_rows) {
      output <- matrix(
        rep(input, output_cols),
        ncol = output_cols,
        byrow = F
      )
    } else {
      stop(
        "If `input` is a vector, it must either contain `output_rows` or `output_cols` elements"
      )
    }
  } else if (is.matrix(input)) {
    if (nrow(input) != output_rows) {
      stop("`input` must be a matrix with `output_rows` rows")
    }
    if (ncol(input) != output_cols) {
      stop("`input` must be a matrix with `output_cols` columns")
    }
    output <- input
  } else {
    stop("`input` must be either a matrix, vector or a scalar")
  }
  return(output)
}

#' Convert scalar input to square matrix of dimension `output_size` x `output_size` with `input` along the diagonal,
#' or check that input array is equivalent to a square matrix of dimension `output_size` x `output_size`.
#'
#' @param input Input to be converted to a square matrix (or passed through as-is)
#' @param output_size Intended row and column dimension of the square output matrix
#' @return A square matrix of dimension `output_size` x `output_size`
#' @noRd
expand_dims_2d_diag <- function(input, output_size) {
  if (length(input) == 1) {
    output <- as.matrix(diag(input, output_size))
  } else if (is.matrix(input)) {
    if (nrow(input) != ncol(input)) {
      stop("`input` must be a square matrix")
    }
    if (nrow(input) != output_size) {
      stop(
        "`input` must be a square matrix with `output_size` rows and columns"
      )
    }
    output <- input
  } else {
    stop("`input` must be either a square matrix or a scalar")
  }
  return(output)
}

#' Return the current stochtree package version string.
#'
#' Falls back to "dev" if the package metadata is unavailable (e.g. during
#' development installs that have not been properly registered).
#'
#' @return A character string such as "0.4.1" or "dev".
#' @noRd
getStochtreeVersion <- function() {
  tryCatch(
    as.character(utils::packageVersion("stochtree")),
    error = function(e) "dev"
  )
}

#' Infer the stochtree version bracket from the fields present in a JSON object.
#'
#' When a JSON was serialized before version stamping was introduced, the version
#' can be approximated by checking which fields are present. The returned string
#' is intended for use in warning messages only, not to gate deserialization
#' behavior.
#'
#' @param json_object A \code{CppJson} object as returned by \code{createCppJson()}
#'   or the various \code{saveBARTModelToJson} / \code{saveBCFModelToJson} functions.
#' @return A character string: the stamp value if \code{stochtree_version} is
#'   present, otherwise a bracket string such as \code{"<0.4.1"}.
#' @noRd
inferStochtreeJsonVersion <- function(json_object) {
  has_field <- function(name) {
    json_contains_field_cpp(json_object$json_ptr, name)
  }
  has_subfolder_field <- function(subfolder, name) {
    json_contains_field_subfolder_cpp(json_object$json_ptr, subfolder, name)
  }

  if (has_field("stochtree_version")) {
    return(json_object$get_string("stochtree_version"))
  }

  # outcome/link in outcome_model were added in ~0.4.1
  if (
    !has_subfolder_field("outcome_model", "outcome") ||
      !has_subfolder_field("outcome_model", "link")
  ) {
    return("<0.4.1")
  }

  # has_rfx_basis / num_rfx_basis were added in ~0.4.0
  if (!has_field("has_rfx_basis") || !has_field("num_rfx_basis")) {
    return("<0.4.0")
  }

  # internal_propensity_model was added in ~0.3.2 (BCF only)
  if (
    has_field("propensity_covariate") && !has_field("internal_propensity_model")
  ) {
    return("<0.3.2")
  }

  # rfx_model_spec and preprocessor_metadata were added in ~0.3.0
  if (!has_field("rfx_model_spec") || !has_field("preprocessor_metadata")) {
    return("<0.3.0")
  }

  return("unknown")
}

test_that("OutcomeModel initialization", {
  # Valid initializations with both terms specified
  model <- OutcomeModel(outcome = "continuous", link = "identity")
  expect_equal(model$outcome, "continuous")
  expect_equal(model$link, "identity")
  model <- OutcomeModel(outcome = "binary", link = "probit")
  expect_equal(model$outcome, "binary")
  expect_equal(model$link, "probit")
  model <- OutcomeModel(outcome = "ordinal", link = "cloglog")
  expect_equal(model$outcome, "ordinal")
  expect_equal(model$link, "cloglog")

  # Valid initializations with only outcome specified
  model <- OutcomeModel(outcome = "continuous")
  expect_equal(model$outcome, "continuous")
  expect_equal(model$link, "identity")
  model <- OutcomeModel(outcome = "binary")
  expect_equal(model$outcome, "binary")
  expect_equal(model$link, "probit")
  model <- OutcomeModel(outcome = "ordinal")
  expect_equal(model$outcome, "ordinal")
  expect_equal(model$link, "cloglog")

  # Invalid initializations
  expect_error(OutcomeModel(outcome = "continuous", link = "other"))
  expect_error(OutcomeModel(outcome = "binary", link = "other"))
  expect_error(OutcomeModel(outcome = "ordinal", link = "other"))
  expect_error(OutcomeModel(outcome = "other", link = "identity"))
  expect_error(OutcomeModel(outcome = "other", link = "probit"))
  expect_error(OutcomeModel(outcome = "other", link = "cloglog"))
})

test_that("Array conversion", {
  skip_on_cran()

  # Test data
  scalar_1 <- 1.5
  scalar_2 <- -2.5
  scalar_3 <- 4
  array_1d_1 <- c(1.6, 3.4, 7.6, 8.7)
  array_1d_2 <- c(2.5, 3.1, 5.6)
  array_1d_3 <- c(2.5)
  array_2d_1 <- matrix(
    c(2.5, 1.2, 4.3, 7.4, 1.7, 2.9, 3.6, 9.1, 7.2, 4.5, 6.7, 1.4),
    nrow = 3,
    ncol = 4,
    byrow = T
  )
  array_2d_2 <- matrix(
    c(2.5, 1.2, 4.3, 7.4, 1.7, 2.9, 3.6, 9.1),
    nrow = 2,
    ncol = 4,
    byrow = T
  )
  array_square_1 <- matrix(
    c(2.5, 1.2, 1.7, 2.9),
    nrow = 2,
    ncol = 2,
    byrow = T
  )
  array_square_2 <- matrix(
    c(2.5, 0.0, 0.0, 2.9),
    nrow = 2,
    ncol = 2,
    byrow = T
  )
  array_square_3 <- matrix(
    c(2.5, 0.0, 0.0, 0.0, 2.9, 0.0, 0.0, 0.0, 5.6),
    nrow = 3,
    ncol = 3,
    byrow = T
  )

  # Error cases
  expect_error(stochtree:::expand_dims_1d(array_1d_1, 5))
  expect_error(stochtree:::expand_dims_1d(array_1d_2, 4))
  expect_error(stochtree:::expand_dims_2d(array_2d_1, 2, 4))
  expect_error(stochtree:::expand_dims_2d(array_2d_2, 3, 4))
  expect_error(stochtree:::expand_dims_2d_diag(array_square_1, 4))
  expect_error(stochtree:::expand_dims_2d_diag(array_square_2, 3))
  expect_error(stochtree:::expand_dims_2d_diag(array_square_3, 2))

  # Working cases
  expect_equal(
    c(scalar_1, scalar_1, scalar_1),
    stochtree:::expand_dims_1d(scalar_1, 3)
  )
  expect_equal(
    c(scalar_2, scalar_2, scalar_2, scalar_2),
    stochtree:::expand_dims_1d(scalar_2, 4)
  )
  expect_equal(c(scalar_3, scalar_3), stochtree:::expand_dims_1d(scalar_3, 2))
  expect_equal(
    c(array_1d_3, array_1d_3, array_1d_3),
    stochtree:::expand_dims_1d(array_1d_3, 3)
  )

  output_exp <- matrix(rep(scalar_1, 6), nrow = 2, byrow = T)
  expect_equal(output_exp, stochtree:::expand_dims_2d(scalar_1, 2, 3))
  output_exp <- matrix(rep(scalar_2, 8), nrow = 2, byrow = T)
  expect_equal(output_exp, stochtree:::expand_dims_2d(scalar_2, 2, 4))
  output_exp <- matrix(rep(scalar_3, 6), nrow = 3, byrow = T)
  expect_equal(output_exp, stochtree:::expand_dims_2d(scalar_3, 3, 2))
  output_exp <- matrix(rep(array_1d_3, 6), nrow = 3, byrow = T)
  expect_equal(output_exp, stochtree:::expand_dims_2d(array_1d_3, 3, 2))
  output_exp <- unname(rbind(array_1d_1, array_1d_1))
  expect_equal(output_exp, stochtree:::expand_dims_2d(array_1d_1, 2, 4))
  output_exp <- unname(rbind(array_1d_2, array_1d_2, array_1d_2))
  expect_equal(output_exp, stochtree:::expand_dims_2d(array_1d_2, 3, 3))
  output_exp <- unname(cbind(array_1d_2, array_1d_2, array_1d_2, array_1d_2))
  expect_equal(output_exp, stochtree:::expand_dims_2d(array_1d_2, 3, 4))
  output_exp <- unname(cbind(array_1d_3, array_1d_3, array_1d_3, array_1d_3))
  expect_equal(output_exp, stochtree:::expand_dims_2d(array_1d_3, 1, 4))
  output_exp <- unname(rbind(array_1d_3, array_1d_3, array_1d_3, array_1d_3))
  expect_equal(output_exp, stochtree:::expand_dims_2d(array_1d_3, 4, 1))

  expect_equal(diag(scalar_1, 3), stochtree:::expand_dims_2d_diag(scalar_1, 3))
  expect_equal(diag(scalar_2, 2), stochtree:::expand_dims_2d_diag(scalar_2, 2))
  expect_equal(diag(scalar_3, 4), stochtree:::expand_dims_2d_diag(scalar_3, 4))
  expect_equal(
    diag(array_1d_3, 2),
    stochtree:::expand_dims_2d_diag(array_1d_3, 2)
  )
})

test_that("In-sample one-hot encoding works for unordered categorical variables", {
  x1 <- c(3, 2, 1, 4, 3, 2, 3, 2, 4, 2)
  x1_onehot <- oneHotInitializeAndEncode(x1)
  # fmt: skip
  x1_expected <- matrix(
        c(0,0,1,0,0,
          0,1,0,0,0,
          1,0,0,0,0,
          0,0,0,1,0,
          0,0,1,0,0,
          0,1,0,0,0,
          0,0,1,0,0,
          0,1,0,0,0,
          0,0,0,1,0,
          0,1,0,0,0),
    byrow = TRUE, ncol = 5)
  x1_levels_expected <- c("1", "2", "3", "4")

  x2 <- c("a", "c", "b", "c", "d", "a", "c", "a", "b", "d")
  x2_onehot <- oneHotInitializeAndEncode(x2)
  # fmt: skip
  x2_expected <- matrix(
        c(1,0,0,0,0,
          0,0,1,0,0,
          0,1,0,0,0,
          0,0,1,0,0,
          0,0,0,1,0,
          1,0,0,0,0,
          0,0,1,0,0,
          1,0,0,0,0,
          0,1,0,0,0,
          0,0,0,1,0),
        byrow = TRUE, ncol = 5)
  x2_levels_expected <- c("a", "b", "c", "d")

  x3 <- c(3.2, 2.4, 1.5, 4.6, 3.2, 2.4, 3.2, 2.4, 4.6, 2.4)
  x3_onehot <- oneHotInitializeAndEncode(x3)
  # fmt: skip
  x3_expected <- matrix(
        c(0,0,1,0,0,
          0,1,0,0,0,
          1,0,0,0,0,
          0,0,0,1,0,
          0,0,1,0,0,
          0,1,0,0,0,
          0,0,1,0,0,
          0,1,0,0,0,
          0,0,0,1,0,
          0,1,0,0,0),
        byrow = TRUE, ncol = 5)
  x3_levels_expected <- c("1.5", "2.4", "3.2", "4.6")

  expect_equal(x1_onehot$Xtilde, x1_expected)
  expect_equal(x2_onehot$Xtilde, x2_expected)
  expect_equal(x3_onehot$Xtilde, x3_expected)
  expect_equal(x1_onehot$unique_levels, x1_levels_expected)
  expect_equal(x2_onehot$unique_levels, x2_levels_expected)
  expect_equal(x3_onehot$unique_levels, x3_levels_expected)
})

test_that("Out-of-sample one-hot encoding works for unordered categorical variables", {
  x1 <- c(3, 2, 1, 4, 3, 2, 3, 2, 4, 2)
  x1_test <- c(1, 2, 4, 3, 5)
  x1_test_onehot <- oneHotEncode(x1_test, levels(factor(x1)))
  # fmt: skip
  x1_test_expected <- matrix(
        c(1,0,0,0,0,
          0,1,0,0,0,
          0,0,0,1,0,
          0,0,1,0,0,
          0,0,0,0,1),
        byrow = TRUE, ncol = 5)

  x2 <- c("a", "c", "b", "c", "d", "a", "c", "a", "b", "d")
  x2_test <- c("a", "c", "g", "b", "f")
  x2_test_onehot <- oneHotEncode(x2_test, levels(factor(x2)))
  # fmt: skip
  x2_test_expected <- matrix(
        c(1,0,0,0,0,
          0,0,1,0,0,
          0,0,0,0,1,
          0,1,0,0,0,
          0,0,0,0,1),
        byrow = TRUE, ncol = 5)

  x3 <- c(3.2, 2.4, 1.5, 4.6, 3.2, 2.4, 3.2, 2.4, 4.6, 2.4)
  x3_test <- c(10.3, -0.5, 4.6, 3.2, 1.8)
  x3_test_onehot <- oneHotEncode(x3_test, levels(factor(x3)))
  # fmt: skip
  x3_test_expected <- matrix(
        c(0,0,0,0,1,
          0,0,0,0,1,
          0,0,0,1,0,
          0,0,1,0,0,
          0,0,0,0,1),
        byrow = TRUE, ncol = 5)

  expect_equal(x1_test_onehot, x1_test_expected)
  expect_equal(x2_test_onehot, x2_test_expected)
  expect_equal(x3_test_onehot, x3_test_expected)
})

test_that("In-sample preprocessing for ordered categorical variables", {
  string_var_response_levels <- c(
    "1. Strongly disagree",
    "2. Disagree",
    "3. Neither agree nor disagree",
    "4. Agree",
    "5. Strongly agree"
  )

  x1 <- c(
    "1. Strongly disagree",
    "3. Neither agree nor disagree",
    "2. Disagree",
    "4. Agree",
    "3. Neither agree nor disagree",
    "5. Strongly agree",
    "4. Agree"
  )
  x1_preprocessing <- orderedCatInitializeAndPreprocess(x1)
  x1_vector_expected <- c(1, 3, 2, 4, 3, 5, 4)
  x1_levels_expected <- string_var_response_levels

  x2 <- factor(x1, levels = string_var_response_levels, ordered = TRUE)
  x2_preprocessing <- orderedCatInitializeAndPreprocess(x2)
  x2_vector_expected <- c(1, 3, 2, 4, 3, 5, 4)
  x2_levels_expected <- string_var_response_levels

  string_var_levels_reordered <- c(
    "5. Strongly agree",
    "4. Agree",
    "3. Neither agree nor disagree",
    "2. Disagree",
    "1. Strongly disagree"
  )
  x3 <- factor(x1, levels = string_var_levels_reordered, ordered = TRUE)
  x3_preprocessing <- orderedCatInitializeAndPreprocess(x3)
  x3_vector_expected <- c(5, 3, 4, 2, 3, 1, 2)
  x3_levels_expected <- string_var_levels_reordered

  x4 <- c(3, 2, 4, 6, 5, 2, 3, 1, 3, 4, 6)
  x4_preprocessing <- orderedCatInitializeAndPreprocess(x4)
  x4_vector_expected <- c(3, 2, 4, 6, 5, 2, 3, 1, 3, 4, 6)
  x4_levels_expected <- c("1", "2", "3", "4", "5", "6")

  expect_equal(x1_preprocessing$x_preprocessed, x1_vector_expected)
  expect_equal(x2_preprocessing$x_preprocessed, x2_vector_expected)
  expect_equal(x3_preprocessing$x_preprocessed, x3_vector_expected)
  expect_equal(x4_preprocessing$x_preprocessed, x4_vector_expected)
  expect_equal(x1_preprocessing$unique_levels, x1_levels_expected)
  expect_equal(x2_preprocessing$unique_levels, x2_levels_expected)
  expect_equal(x3_preprocessing$unique_levels, x3_levels_expected)
  expect_equal(x4_preprocessing$unique_levels, x4_levels_expected)
})

test_that("Out-of-sample preprocessing for ordered categorical variables", {
  string_var_response_levels <- c(
    "1. Strongly disagree",
    "2. Disagree",
    "3. Neither agree nor disagree",
    "4. Agree",
    "5. Strongly agree"
  )

  x1 <- c(
    "1. Strongly disagree",
    "3. Neither agree nor disagree",
    "2. Disagree",
    "4. Agree",
    "3. Neither agree nor disagree",
    "5. Strongly agree",
    "4. Agree"
  )
  x1_preprocessing <- orderedCatPreprocess(x1, string_var_response_levels)
  x1_vector_expected <- c(1, 3, 2, 4, 3, 5, 4)

  x2 <- factor(x1, levels = string_var_response_levels, ordered = TRUE)
  x2_preprocessing <- orderedCatPreprocess(x2, string_var_response_levels)
  x2_vector_expected <- c(1, 3, 2, 4, 3, 5, 4)

  x3 <- c(
    "1. Strongly disagree",
    "6. Other",
    "7. Also other",
    "4. Agree",
    "3. Neither agree nor disagree",
    "5. Strongly agree",
    "4. Agree"
  )
  expected_warning_message <- "Variable includes ordered categorical levels not included in the original training set"
  expect_warning(
    x3_preprocessing <- orderedCatPreprocess(x3, string_var_response_levels),
    expected_warning_message
  )
  x3_vector_expected <- c(1, 6, 6, 4, 3, 5, 4)

  x4 <- c(3, 2, 4, 6, 5, 2, 3, 1, 3, 4, 6)
  x4_preprocessing <- orderedCatPreprocess(x4, c("1", "2", "3", "4", "5", "6"))
  x4_vector_expected <- c(3, 2, 4, 6, 5, 2, 3, 1, 3, 4, 6)

  expect_equal(x1_preprocessing, x1_vector_expected)
  expect_equal(x2_preprocessing, x2_vector_expected)
  expect_equal(x3_preprocessing, x3_vector_expected)
  expect_equal(x4_preprocessing, x4_vector_expected)
})

# test_that("Preprocessing of all-numeric covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_mat <- matrix(c(
#         1,2,3,4,5,
#         5,4,3,2,1,
#         6,7,8,9,10
#     ), ncol = 3, byrow = FALSE)
#     preprocess_list <- createForestCovariates(cov_df)
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, rep(0,3))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 3)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$numeric_vars, c("x1","x2","x3"))
# })
#
# test_that("Preprocessing of all-unordered-categorical covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_mat <- matrix(c(
#         1,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,
#         0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,
#         0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,
#         0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,0,
#         0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0
#     ), nrow = 5, byrow = TRUE)
#     preprocess_list <- createForestCovariates(cov_df, unordered_cat_vars = c("x1","x2","x3"))
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, rep(1,18))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 0)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 3)
#     expect_equal(preprocess_list$metadata$unordered_cat_vars, c("x1","x2","x3"))
#     expect_equal(preprocess_list$metadata$unordered_unique_levels,
#                  list(x1=c("1","2","3","4","5"),
#                       x2=c("1","2","3","4","5"),
#                       x3=c("6","7","8","9","10"))
#     )
# })
#
# test_that("Preprocessing of all-ordered-categorical covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_mat <- matrix(c(
#         1,2,3,4,5,
#         5,4,3,2,1,
#         1,2,3,4,5
#     ), ncol = 3, byrow = FALSE)
#     preprocess_list <- createForestCovariates(cov_df, ordered_cat_vars = c("x1","x2","x3"))
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, rep(1,3))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 0)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 3)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$ordered_cat_vars, c("x1","x2","x3"))
#     expect_equal(preprocess_list$metadata$ordered_unique_levels,
#                  list(x1=c("1","2","3","4","5"),
#                       x2=c("1","2","3","4","5"),
#                       x3=c("6","7","8","9","10"))
#     )
# })
#
# test_that("Preprocessing of mixed-covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_mat <- matrix(c(
#         1,5,1,0,0,0,0,0,
#         2,4,0,1,0,0,0,0,
#         3,3,0,0,1,0,0,0,
#         4,2,0,0,0,1,0,0,
#         5,1,0,0,0,0,1,0
#     ), nrow = 5, byrow = TRUE)
#     preprocess_list <- createForestCovariates(cov_df, ordered_cat_vars = c("x2"), unordered_cat_vars = "x3")
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, c(0, rep(1,7)))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 1)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 1)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 1)
#     expect_equal(preprocess_list$metadata$ordered_cat_vars, c("x2"))
#     expect_equal(preprocess_list$metadata$unordered_cat_vars, c("x3"))
#     expect_equal(preprocess_list$metadata$ordered_unique_levels, list(x2=c("1","2","3","4","5")))
#     expect_equal(preprocess_list$metadata$unordered_unique_levels, list(x3=c("6","7","8","9","10")))
# })
#
# test_that("Preprocessing of mixed-covariate matrix works", {
#     cov_input <- matrix(c(1:5,5:1,6:10),ncol=3,byrow=FALSE)
#     cov_mat <- matrix(c(
#         1,5,1,0,0,0,0,0,
#         2,4,0,1,0,0,0,0,
#         3,3,0,0,1,0,0,0,
#         4,2,0,0,0,1,0,0,
#         5,1,0,0,0,0,1,0
#     ), nrow = 5, byrow = TRUE)
#     preprocess_list <- createForestCovariates(cov_input, ordered_cat_vars = 2, unordered_cat_vars = 3)
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, c(0, rep(1,7)))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 1)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 1)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 1)
#     expect_equal(preprocess_list$metadata$ordered_cat_vars, c("x2"))
#     expect_equal(preprocess_list$metadata$unordered_cat_vars, c("x3"))
#     expect_equal(preprocess_list$metadata$ordered_unique_levels, list(x2=c("1","2","3","4","5")))
#     expect_equal(preprocess_list$metadata$unordered_unique_levels, list(x3=c("6","7","8","9","10")))
#
#     alt_preprocess_list <- createForestCovariates(cov_input, ordered_cat_vars = "x2", unordered_cat_vars = "x3")
#     expect_equal(alt_preprocess_list$data, cov_mat)
#     expect_equal(alt_preprocess_list$metadata$feature_types, c(0, rep(1,7)))
#     expect_equal(alt_preprocess_list$metadata$num_numeric_vars, 1)
#     expect_equal(alt_preprocess_list$metadata$num_ordered_cat_vars, 1)
#     expect_equal(alt_preprocess_list$metadata$num_unordered_cat_vars, 1)
#     expect_equal(alt_preprocess_list$metadata$ordered_cat_vars, c("x2"))
#     expect_equal(alt_preprocess_list$metadata$unordered_cat_vars, c("x3"))
#     expect_equal(alt_preprocess_list$metadata$ordered_unique_levels, list(x2=c("1","2","3","4","5")))
#     expect_equal(alt_preprocess_list$metadata$unordered_unique_levels, list(x3=c("6","7","8","9","10")))
# })
#
# test_that("Preprocessing of out-of-sample mixed-covariate dataset works", {
#     metadata <- list(
#         num_numeric_vars = 1,
#         num_ordered_cat_vars = 1,
#         num_unordered_cat_vars = 1,
#         numeric_vars = c("x1"),
#         ordered_cat_vars = c("x2"),
#         unordered_cat_vars = c("x3"),
#         ordered_unique_levels = list(x2=c("1","2","3","4","5")),
#         unordered_unique_levels = list(x3=c("6","7","8","9","10"))
#     )
#     cov_df <- data.frame(x1 = c(1:5,1), x2 = c(5:1,5), x3 = 6:11)
#     cov_mat <- matrix(c(
#         1,5,1,0,0,0,0,0,
#         2,4,0,1,0,0,0,0,
#         3,3,0,0,1,0,0,0,
#         4,2,0,0,0,1,0,0,
#         5,1,0,0,0,0,1,0,
#         1,5,0,0,0,0,0,1
#     ), nrow = 6, byrow = TRUE)
#     X_preprocessed <- createForestCovariatesFromMetadata(cov_df, metadata)
#     expect_equal(X_preprocessed, cov_mat)
# })
#
# test_that("Preprocessing of all-numeric covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_mat <- matrix(c(
#         1,2,3,4,5,
#         5,4,3,2,1,
#         6,7,8,9,10
#     ), ncol = 3, byrow = FALSE)
#     preprocess_list <- preprocessTrainDataFrame(cov_df)
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, rep(0,3))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 3)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$original_var_indices, 1:3)
#     expect_equal(preprocess_list$metadata$numeric_vars, c("x1","x2","x3"))
# })
#
# test_that("Preprocessing of all-unordered-categorical covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_df$x1 <- factor(cov_df$x1)
#     cov_df$x2 <- factor(cov_df$x2)
#     cov_df$x3 <- factor(cov_df$x3)
#     cov_mat <- matrix(c(
#         1,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,
#         0,1,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,
#         0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,
#         0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,1,0,0,
#         0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0
#     ), nrow = 5, byrow = TRUE)
#     preprocess_list <- preprocessTrainDataFrame(cov_df)
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, rep(1,18))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 0)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 3)
#     expect_equal(preprocess_list$metadata$unordered_cat_vars, c("x1","x2","x3"))
#     expected_var_indices <- c(rep(1,6),rep(2,6),rep(3,6))
#     expect_equal(preprocess_list$metadata$original_var_indices, expected_var_indices)
#     expect_equal(preprocess_list$metadata$unordered_unique_levels,
#                  list(x1=c("1","2","3","4","5"),
#                       x2=c("1","2","3","4","5"),
#                       x3=c("6","7","8","9","10"))
#     )
# })
#
# test_that("Preprocessing of all-ordered-categorical covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_df$x1 <- factor(cov_df$x1, ordered = TRUE)
#     cov_df$x2 <- factor(cov_df$x2, ordered = TRUE)
#     cov_df$x3 <- factor(cov_df$x3, ordered = TRUE)
#     cov_mat <- matrix(c(
#         1,2,3,4,5,
#         5,4,3,2,1,
#         1,2,3,4,5
#     ), ncol = 3, byrow = FALSE)
#     preprocess_list <- preprocessTrainDataFrame(cov_df)
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, rep(1,3))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 0)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 3)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 0)
#     expect_equal(preprocess_list$metadata$ordered_cat_vars, c("x1","x2","x3"))
#     expect_equal(preprocess_list$metadata$original_var_indices, 1:3)
#     expect_equal(preprocess_list$metadata$ordered_unique_levels,
#                  list(x1=c("1","2","3","4","5"),
#                       x2=c("1","2","3","4","5"),
#                       x3=c("6","7","8","9","10"))
#     )
# })
#
# test_that("Preprocessing of mixed-covariate dataset works", {
#     cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
#     cov_df$x2 <- factor(cov_df$x2, ordered = TRUE)
#     cov_df$x3 <- factor(cov_df$x3)
#     cov_mat <- matrix(c(
#         1,5,1,0,0,0,0,0,
#         2,4,0,1,0,0,0,0,
#         3,3,0,0,1,0,0,0,
#         4,2,0,0,0,1,0,0,
#         5,1,0,0,0,0,1,0
#     ), nrow = 5, byrow = TRUE)
#     preprocess_list <- preprocessTrainDataFrame(cov_df)
#     expect_equal(preprocess_list$data, cov_mat)
#     expect_equal(preprocess_list$metadata$feature_types, c(0, rep(1,7)))
#     expect_equal(preprocess_list$metadata$num_numeric_vars, 1)
#     expect_equal(preprocess_list$metadata$num_ordered_cat_vars, 1)
#     expect_equal(preprocess_list$metadata$num_unordered_cat_vars, 1)
#     expect_equal(preprocess_list$metadata$ordered_cat_vars, c("x2"))
#     expect_equal(preprocess_list$metadata$unordered_cat_vars, c("x3"))
#     expected_var_indices <- c(1,2,rep(3,6))
#     expect_equal(preprocess_list$metadata$original_var_indices, expected_var_indices)
#     expect_equal(preprocess_list$metadata$ordered_unique_levels, list(x2=c("1","2","3","4","5")))
#     expect_equal(preprocess_list$metadata$unordered_unique_levels, list(x3=c("6","7","8","9","10")))
# })
#
# test_that("Preprocessing of out-of-sample mixed-covariate dataset works", {
#     metadata <- list(
#         num_numeric_vars = 1,
#         num_ordered_cat_vars = 1,
#         num_unordered_cat_vars = 1,
#         original_var_indices = c(1, 2, 3, 3, 3, 3, 3, 3),
#         numeric_vars = c("x1"),
#         ordered_cat_vars = c("x2"),
#         unordered_cat_vars = c("x3"),
#         ordered_unique_levels = list(x2=c("1","2","3","4","5")),
#         unordered_unique_levels = list(x3=c("6","7","8","9","10"))
#     )
#     cov_df <- data.frame(x1 = c(1:5,1), x2 = c(5:1,5), x3 = 6:11)
#     var_weights <- rep(1./3., 3)
#     cov_mat <- matrix(c(
#         1,5,1,0,0,0,0,0,
#         2,4,0,1,0,0,0,0,
#         3,3,0,0,1,0,0,0,
#         4,2,0,0,0,1,0,0,
#         5,1,0,0,0,0,1,0,
#         1,5,0,0,0,0,0,1
#     ), nrow = 6, byrow = TRUE)
#     X_preprocessed <- preprocessPredictionDataFrame(cov_df, metadata)
#     expect_equal(X_preprocessed, cov_mat)
# })

make_bart_data <- function(n = 100, p = 5, seed = 42) {
  set.seed(seed)
  X <- matrix(runif(n * p), ncol = p)
  y <- sin(X[, 1] * pi) + rnorm(n, 0, 0.1)
  n_train <- as.integer(0.8 * n)
  list(
    X_train = X[1:n_train, ], y_train = y[1:n_train],
    X_test = X[(n_train + 1):n, ], n_train = n_train, n_test = n - n_train
  )
}

make_bcf_data <- function(n = 100, p = 5, seed = 42) {
  set.seed(seed)
  X <- matrix(runif(n * p), ncol = p)
  pi_X <- 0.25 + 0.5 * X[, 1]
  Z <- rbinom(n, 1, pi_X)
  y <- pi_X * 5 + X[, 2] * 2 * Z + rnorm(n, 0, 1)
  n_train <- as.integer(0.8 * n)
  list(
    X_train = X[1:n_train, ], Z_train = Z[1:n_train], y_train = y[1:n_train],
    pi_train = pi_X[1:n_train], X_test = X[(n_train + 1):n, ],
    Z_test = Z[(n_train + 1):n], pi_test = pi_X[(n_train + 1):n],
    n_train = n_train, n_test = n - n_train
  )
}

test_that("BART: uniform weights produce identical predictions to no weights", {
  skip_on_cran()
  d <- make_bart_data()
  num_mcmc <- 10

  set.seed(1)
  m1 <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = num_mcmc,
    general_params = list(random_seed = 1L)
  )

  set.seed(1)
  m2 <- bart(
    X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
    observation_weights = rep(1.0, d$n_train),
    num_gfr = 0, num_burnin = 0, num_mcmc = num_mcmc,
    general_params = list(random_seed = 1L)
  )

  expect_equal(m1$y_hat_train, m2$y_hat_train)
  expect_equal(m1$y_hat_test, m2$y_hat_test)
})

test_that("BART: non-uniform weights run and produce correct output shape", {
  skip_on_cran()
  d <- make_bart_data()
  num_mcmc <- 10
  weights <- runif(d$n_train, 0.5, 2.0)

  expect_no_error(
    m <- bart(
      X_train = d$X_train, y_train = d$y_train, X_test = d$X_test,
      observation_weights = weights,
      num_gfr = 0, num_burnin = 0, num_mcmc = num_mcmc
    )
  )
  expect_equal(dim(m$y_hat_train), c(d$n_train, num_mcmc))
  expect_equal(dim(m$y_hat_test), c(d$n_test, num_mcmc))
})

test_that("BART: all-zero weights (prior mode) run with num_gfr = 0", {
  skip_on_cran()
  d <- make_bart_data()
  num_mcmc <- 10

  expect_no_error(
    m <- bart(
      X_train = d$X_train, y_train = d$y_train,
      observation_weights = rep(0.0, d$n_train),
      num_gfr = 0, num_burnin = 0, num_mcmc = num_mcmc
    )
  )
  expect_equal(dim(m$y_hat_train), c(d$n_train, num_mcmc))
})

test_that("BART: non-numeric observation_weights raises error", {
  skip_on_cran()
  d <- make_bart_data()
  expect_error(
    bart(
      X_train = d$X_train, y_train = d$y_train,
      observation_weights = as.character(rep(1, d$n_train)),
      num_gfr = 0, num_burnin = 0, num_mcmc = 5
    ),
    "numeric"
  )
})

test_that("BART: wrong-length observation_weights raises error", {
  skip_on_cran()
  d <- make_bart_data()
  expect_error(
    bart(
      X_train = d$X_train, y_train = d$y_train,
      observation_weights = rep(1.0, d$n_train + 1),
      num_gfr = 0, num_burnin = 0, num_mcmc = 5
    ),
    "nrow"
  )
})

test_that("BART: negative observation_weights raises error", {
  skip_on_cran()
  d <- make_bart_data()
  weights <- rep(1.0, d$n_train)
  weights[1] <- -1.0
  expect_error(
    bart(
      X_train = d$X_train, y_train = d$y_train,
      observation_weights = weights,
      num_gfr = 0, num_burnin = 0, num_mcmc = 5
    ),
    "negative"
  )
})

test_that("BART: all-zero weights with num_gfr > 0 raises error", {
  skip_on_cran()
  d <- make_bart_data()
  expect_error(
    bart(
      X_train = d$X_train, y_train = d$y_train,
      observation_weights = rep(0.0, d$n_train),
      num_gfr = 5, num_burnin = 0, num_mcmc = 10
    ),
    "num_gfr"
  )
})

test_that("BART: observation_weights with cloglog outcome raises error", {
  skip_on_cran()
  d <- make_bart_data()
  y_ord <- sample(1:3, d$n_train, replace = TRUE)
  expect_error(
    bart(
      X_train = d$X_train, y_train = y_ord,
      observation_weights = rep(1.0, d$n_train),
      num_gfr = 0, num_burnin = 0, num_mcmc = 5,
      general_params = list(outcome_model = OutcomeModel(outcome = "ordinal", link = "cloglog"))
    ),
    "cloglog"
  )
})

test_that("BART: observation_weights with variance forest raises warning", {
  skip_on_cran()
  d <- make_bart_data()
  expect_warning(
    bart(
      X_train = d$X_train, y_train = d$y_train,
      observation_weights = rep(1.0, d$n_train),
      num_gfr = 0, num_burnin = 0, num_mcmc = 5,
      variance_forest_params = list(num_trees = 5)
    ),
    "variance forest"
  )
})

test_that("BCF: uniform weights produce identical predictions to no weights", {
  skip_on_cran()
  d <- make_bcf_data()
  num_mcmc <- 10

  set.seed(1)
  m1 <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train, X_test = d$X_test,
    Z_test = d$Z_test, propensity_test = d$pi_test,
    num_gfr = 0, num_burnin = 0, num_mcmc = num_mcmc,
    general_params = list(random_seed = 1L)
  )

  set.seed(1)
  m2 <- bcf(
    X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
    propensity_train = d$pi_train, X_test = d$X_test,
    Z_test = d$Z_test, propensity_test = d$pi_test,
    observation_weights = rep(1.0, d$n_train),
    num_gfr = 0, num_burnin = 0, num_mcmc = num_mcmc,
    general_params = list(random_seed = 1L)
  )

  expect_equal(m1$y_hat_train, m2$y_hat_train)
  expect_equal(m1$tau_hat_train, m2$tau_hat_train)
})

test_that("BCF: non-uniform weights run and produce correct output shape", {
  skip_on_cran()
  d <- make_bcf_data()
  num_mcmc <- 10
  weights <- runif(d$n_train, 0.5, 2.0)

  expect_no_error(
    m <- bcf(
      X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
      propensity_train = d$pi_train, X_test = d$X_test,
      Z_test = d$Z_test, propensity_test = d$pi_test,
      observation_weights = weights,
      num_gfr = 0, num_burnin = 0, num_mcmc = num_mcmc
    )
  )
  expect_equal(dim(m$y_hat_train), c(d$n_train, num_mcmc))
  expect_equal(dim(m$tau_hat_train), c(d$n_train, num_mcmc))
  expect_equal(dim(m$y_hat_test), c(d$n_test, num_mcmc))
  expect_equal(dim(m$tau_hat_test), c(d$n_test, num_mcmc))
})

test_that("BCF: negative observation_weights raises error", {
  skip_on_cran()
  d <- make_bcf_data()
  weights <- rep(1.0, d$n_train)
  weights[1] <- -1.0
  expect_error(
    bcf(
      X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
      propensity_train = d$pi_train,
      observation_weights = weights,
      num_gfr = 0, num_burnin = 0, num_mcmc = 5
    ),
    "negative"
  )
})

test_that("BCF: all-zero weights with num_gfr > 0 raises error", {
  skip_on_cran()
  d <- make_bcf_data()
  expect_error(
    bcf(
      X_train = d$X_train, Z_train = d$Z_train, y_train = d$y_train,
      propensity_train = d$pi_train,
      observation_weights = rep(0.0, d$n_train),
      num_gfr = 5, num_burnin = 0, num_mcmc = 10
    ),
    "num_gfr"
  )
})

test_that("BCF: observation_weights with cloglog outcome raises error", {
  skip_on_cran()
  d <- make_bcf_data()
  y_bin <- rbinom(d$n_train, 1, 0.5)
  expect_error(
    bcf(
      X_train = d$X_train, Z_train = d$Z_train, y_train = y_bin,
      propensity_train = d$pi_train,
      observation_weights = rep(1.0, d$n_train),
      num_gfr = 0, num_burnin = 0, num_mcmc = 5,
      general_params = list(outcome_model = OutcomeModel(outcome = "binary", link = "cloglog"))
    ),
    "cloglog"
  )
})

#' Forest State Management Routines
#' @name ForestStateManagement
#' @description
#' A forest sampler features two types of state: ephemeral and persistent.
#' Persistent state includes objects like [ForestSamples] and [RandomEffectSamples]
#' which constitute part of the final sampled model.
#' Ephemeral state supports the sampling computations, but is not retained after the sampler finishes.
#'
#' The two primary forest-based bits of ephemeral state are the [Forest] and [ForestModel] classes,
#' which represent the current state of a forest and its corresponding tracking data structures.
#'
#' In a linear sampling loop, this ephemeral state is updated with each iteration of the sampler
#' and any retained forests are copied to a [ForestSamples] object. However, in multi-chain settings,
#' the state of a forest must typically be "reset" at the beginning of a new chain. These function enable
#' this process by synchronizing the state of a [Forest] and [ForestModel] with a corresponding element
#' of a [ForestSamples] object, or by resetting both to their default (root) state.
#'
#' `resetActiveForest` resets a [Forest] object, either from a specific forest in a [ForestSamples]
#' object or to an ensemble of single-node (i.e. root) trees.
#' `resetForestModel` re-initializes a forest model (tracking data structures) from a specific forest in a [ForestSamples] object.
#'
#' These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#' @returns
#' Both functions have no return type and operate in-place on the relevant [Forest] or [ForestModel] objects
#' @examples
#' n <- 100
#' p <- 10
#' num_trees <- 100
#' leaf_dimension <- 1
#' is_leaf_constant <- TRUE
#' is_exponentiated <- FALSE
#' alpha <- 0.95
#' beta <- 2.0
#' min_samples_leaf <- 2
#' max_depth <- 10
#' feature_types <- as.integer(rep(0, p))
#' leaf_model <- 0
#' sigma2 <- 1.0
#' leaf_scale <- as.matrix(1.0)
#' variable_weights <- rep(1/p, p)
#' a_forest <- 1
#' b_forest <- 1
#' cutpoint_grid_size <- 100
#' X <- matrix(runif(n*p), ncol = p)
#' forest_dataset <- createForestDataset(X)
#' y <- -5 + 10*(X[,1] > 0.5) + rnorm(n)
#' outcome <- createOutcome(y)
#' rng <- createCppRNG(1234)
#' global_model_config <- createGlobalModelConfig(global_error_variance=sigma2)
#' forest_model_config <- createForestModelConfig(feature_types=feature_types,
#'                                                num_trees=num_trees, num_observations=n,
#'                                                num_features=p, alpha=alpha, beta=beta,
#'                                                min_samples_leaf=min_samples_leaf,
#'                                                max_depth=max_depth,
#'                                                variable_weights=variable_weights,
#'                                                cutpoint_grid_size=cutpoint_grid_size,
#'                                                leaf_model_type=leaf_model,
#'                                                leaf_model_scale=leaf_scale)
#' forest_model <- createForestModel(forest_dataset, forest_model_config, global_model_config)
#' active_forest <- createForest(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
#' forest_samples <- createForestSamples(num_trees, leaf_dimension,
#'                                       is_leaf_constant, is_exponentiated)
#' active_forest$prepare_for_sampler(forest_dataset, outcome, forest_model, 0, 0.)
#' forest_model$sample_one_iteration(
#'     forest_dataset, outcome, forest_samples, active_forest,
#'     rng, forest_model_config, global_model_config,
#'     keep_forest = TRUE, gfr = FALSE
#' )
#' resetActiveForest(active_forest, forest_samples, 0)
#' resetForestModel(forest_model, active_forest, forest_dataset, outcome, TRUE)
#'
NULL
#> NULL

#' @title Forest Container C++ Wrapper
#' @description
#' Wrapper around a C++ class that stores draws from an random ensemble of decision trees.
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

ForestSamples <- R6::R6Class(
  classname = "ForestSamples",
  cloneable = FALSE,
  public = list(
    #' @field forest_container_ptr External pointer to a C++ ForestContainer class
    forest_container_ptr = NULL,

    #' @description
    #' Create a new ForestContainer object.
    #' @param num_trees Number of trees
    #' @param leaf_dimension Dimensionality of the outcome model
    #' @param is_leaf_constant Whether leaf is constant
    #' @param is_exponentiated Whether forest predictions should be exponentiated before being returned
    #' @return A new `ForestContainer` object.
    initialize = function(
      num_trees,
      leaf_dimension = 1,
      is_leaf_constant = FALSE,
      is_exponentiated = FALSE
    ) {
      self$forest_container_ptr <- forest_container_cpp(
        num_trees,
        leaf_dimension,
        is_leaf_constant,
        is_exponentiated
      )
    },

    #' @description
    #' Collapse forests in this container by a pre-specified batch size.
    #' For example, if we have a container of twenty 10-tree forests, and we
    #' specify a `batch_size` of 5, then this method will yield four 50-tree
    #' forests. "Excess" forests remaining after the size of a forest container
    #' is divided by `batch_size` will be pruned from the beginning of the
    #' container (i.e. earlier sampled forests will be deleted). This method
    #' has no effect if `batch_size` is larger than the number of forests
    #' in a container.
    #' @param batch_size Number of forests to be collapsed into a single forest
    collapse = function(batch_size) {
      container_size <- self$num_samples()
      if ((batch_size <= container_size) && (batch_size > 1)) {
        reverse_container_inds <- seq(container_size, 1, -1)
        num_clean_batches <- container_size %/% batch_size
        batch_inds <- (reverse_container_inds -
          (container_size -
            (container_size %/% num_clean_batches) *
              num_clean_batches) -
          1) %/%
          batch_size
        for (batch_ind in unique(batch_inds[batch_inds >= 0])) {
          merge_forest_inds <- sort(
            reverse_container_inds[batch_inds == batch_ind] - 1
          )
          num_merge_forests <- length(merge_forest_inds)
          self$combine_forests(merge_forest_inds)
          for (i in num_merge_forests:2) {
            self$delete_sample(merge_forest_inds[i])
          }
          forest_scale_factor <- 1.0 / num_merge_forests
          self$multiply_forest(
            merge_forest_inds[1],
            forest_scale_factor
          )
        }
        if (min(batch_inds) < 0) {
          delete_forest_inds <- sort(
            reverse_container_inds[batch_inds < 0] - 1
          )
          for (i in length(delete_forest_inds):1) {
            self$delete_sample(delete_forest_inds[i])
          }
        }
      }
    },

    #' @description
    #' Merge specified forests into a single forest
    #' @param forest_inds Indices of forests to be combined (0-indexed)
    combine_forests = function(forest_inds) {
      stopifnot(max(forest_inds) < self$num_samples())
      stopifnot(min(forest_inds) >= 0)
      stopifnot(length(forest_inds) > 1)
      stopifnot(all(as.integer(forest_inds) == forest_inds))
      forest_inds_sorted <- as.integer(sort(forest_inds))
      combine_forests_forest_container_cpp(
        self$forest_container_ptr,
        forest_inds_sorted
      )
    },

    #' @description
    #' Add a constant value to every leaf of every tree of a given forest
    #' @param forest_index Index of forest whose leaves will be modified (0-indexed)
    #' @param constant_value Value to add to every leaf of every tree of the forest at `forest_index`
    add_to_forest = function(forest_index, constant_value) {
      stopifnot(forest_index < self$num_samples())
      stopifnot(forest_index >= 0)
      add_to_forest_forest_container_cpp(
        self$forest_container_ptr,
        forest_index,
        constant_value
      )
    },

    #' @description
    #' Multiply every leaf of every tree of a given forest by constant value
    #' @param forest_index Index of forest whose leaves will be modified (0-indexed)
    #' @param constant_multiple Value to multiply through by every leaf of every tree of the forest at `forest_index`
    multiply_forest = function(forest_index, constant_multiple) {
      stopifnot(forest_index < self$num_samples())
      stopifnot(forest_index >= 0)
      multiply_forest_forest_container_cpp(
        self$forest_container_ptr,
        forest_index,
        constant_multiple
      )
    },

    #' @description
    #' Create a new `ForestContainer` object from a json object
    #' @param json_object Object of class `CppJson`
    #' @param json_forest_label Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy
    #' @return A new `ForestContainer` object.
    load_from_json = function(json_object, json_forest_label) {
      self$forest_container_ptr <- forest_container_from_json_cpp(
        json_object$json_ptr,
        json_forest_label
      )
    },

    #' @description
    #' Append to a `ForestContainer` object from a json object
    #' @param json_object Object of class `CppJson`
    #' @param json_forest_label Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy
    #' @return None
    append_from_json = function(json_object, json_forest_label) {
      forest_container_append_from_json_cpp(
        self$forest_container_ptr,
        json_object$json_ptr,
        json_forest_label
      )
    },

    #' @description
    #' Create a new `ForestContainer` object from a json object
    #' @param json_string JSON string which parses into object of class `CppJson`
    #' @param json_forest_label Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy
    #' @return A new `ForestContainer` object.
    load_from_json_string = function(json_string, json_forest_label) {
      self$forest_container_ptr <- forest_container_from_json_string_cpp(
        json_string,
        json_forest_label
      )
    },

    #' @description
    #' Append to a `ForestContainer` object from a json object
    #' @param json_string JSON string which parses into object of class `CppJson`
    #' @param json_forest_label Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy
    #' @return None
    append_from_json_string = function(json_string, json_forest_label) {
      forest_container_append_from_json_string_cpp(
        self$forest_container_ptr,
        json_string,
        json_forest_label
      )
    },

    #' @description
    #' Predict every tree ensemble on every sample in `forest_dataset`
    #' @param forest_dataset `ForestDataset` R class
    #' @return matrix of predictions with as many rows as in forest_dataset
    #' and as many columns as samples in the `ForestContainer`
    predict = function(forest_dataset) {
      stopifnot(!is.null(forest_dataset$data_ptr))
      return(predict_forest_cpp(
        self$forest_container_ptr,
        forest_dataset$data_ptr
      ))
    },

    #' @description
    #' Predict "raw" leaf values (without being multiplied by basis) for every tree ensemble on every sample in `forest_dataset`
    #' @param forest_dataset `ForestDataset` R class
    #' @return Array of predictions for each observation in `forest_dataset` and
    #' each sample in the `ForestSamples` class with each prediction having the
    #' dimensionality of the forests' leaf model. In the case of a constant leaf model
    #' or univariate leaf regression, this array is two-dimensional (number of observations,
    #' number of forest samples). In the case of a multivariate leaf regression,
    #' this array is three-dimension (number of observations, leaf model dimension,
    #' number of samples).
    predict_raw = function(forest_dataset) {
      stopifnot(!is.null(forest_dataset$data_ptr))
      # Unpack dimensions
      output_dim <- leaf_dimension_forest_container_cpp(
        self$forest_container_ptr
      )
      num_samples <- num_samples_forest_container_cpp(
        self$forest_container_ptr
      )
      n <- dataset_num_rows_cpp(forest_dataset$data_ptr)

      # Predict leaf values from forest
      predictions <- predict_forest_raw_cpp(
        self$forest_container_ptr,
        forest_dataset$data_ptr
      )
      if (output_dim > 1) {
        dim(predictions) <- c(n, output_dim, num_samples)
      } else {
        dim(predictions) <- c(n, num_samples)
      }

      return(predictions)
    },

    #' @description
    #' Predict "raw" leaf values (without being multiplied by basis) for a specific forest on every sample in `forest_dataset`
    #' @param forest_dataset `ForestDataset` R class
    #' @param forest_num Index of the forest sample within the container
    #' @return matrix of predictions with as many rows as in forest_dataset
    #' and as many columns as dimensions in the leaves of trees in `ForestContainer`
    predict_raw_single_forest = function(forest_dataset, forest_num) {
      stopifnot(!is.null(forest_dataset$data_ptr))
      # Unpack dimensions
      output_dim <- leaf_dimension_forest_container_cpp(
        self$forest_container_ptr
      )
      n <- dataset_num_rows_cpp(forest_dataset$data_ptr)

      # Predict leaf values from forest
      output <- predict_forest_raw_single_forest_cpp(
        self$forest_container_ptr,
        forest_dataset$data_ptr,
        forest_num
      )
      return(output)
    },

    #' @description
    #' Predict "raw" leaf values (without being multiplied by basis) for a specific tree in a specific forest on every observation in `forest_dataset`
    #' @param forest_dataset `ForestDataset` R class
    #' @param forest_num Index of the forest sample within the container
    #' @param tree_num Index of the tree to be queried
    #' @return matrix of predictions with as many rows as in `forest_dataset`
    #' and as many columns as dimensions in the leaves of trees in `ForestContainer`
    predict_raw_single_tree = function(
      forest_dataset,
      forest_num,
      tree_num
    ) {
      stopifnot(!is.null(forest_dataset$data_ptr))

      # Predict leaf values from forest
      output <- predict_forest_raw_single_tree_cpp(
        self$forest_container_ptr,
        forest_dataset$data_ptr,
        forest_num,
        tree_num
      )
      return(output)
    },

    #' @description
    #' Set a constant predicted value for every tree in the ensemble.
    #' Stops program if any tree is more than a root node.
    #' @param forest_num Index of the forest sample within the container.
    #' @param leaf_value Constant leaf value(s) to be fixed for each tree in the ensemble indexed by `forest_num`. Can be either a single number or a vector, depending on the forest's leaf dimension.
    set_root_leaves = function(forest_num, leaf_value) {
      stopifnot(!is.null(self$forest_container_ptr))
      stopifnot(
        num_samples_forest_container_cpp(self$forest_container_ptr) == 0
      )

      # Set leaf values
      if (length(leaf_value) == 1) {
        stopifnot(
          leaf_dimension_forest_container_cpp(
            self$forest_container_ptr
          ) ==
            1
        )
        set_leaf_value_forest_container_cpp(
          self$forest_container_ptr,
          leaf_value
        )
      } else if (length(leaf_value) > 1) {
        stopifnot(
          leaf_dimension_forest_container_cpp(
            self$forest_container_ptr
          ) ==
            length(leaf_value)
        )
        set_leaf_vector_forest_container_cpp(
          self$forest_container_ptr,
          leaf_value
        )
      } else {
        stop(
          "leaf_value must be a numeric value or vector of length >= 1"
        )
      }
    },

    #' @description
    #' Set a constant predicted value for every tree in the ensemble.
    #' Stops program if any tree is more than a root node.
    #' @param dataset `ForestDataset` Dataset class (covariates, basis, etc...)
    #' @param outcome `Outcome` Outcome class (residual / partial residual)
    #' @param forest_model `ForestModel` object storing tracking structures used in training / sampling
    #' @param leaf_model_int Integer value encoding the leaf model type (0 = constant gaussian, 1 = univariate gaussian, 2 = multivariate gaussian, 3 = log linear variance).
    #' @param leaf_value Constant leaf value(s) to be fixed for each tree in the ensemble indexed by `forest_num`. Can be either a single number or a vector, depending on the forest's leaf dimension.
    prepare_for_sampler = function(
      dataset,
      outcome,
      forest_model,
      leaf_model_int,
      leaf_value
    ) {
      stopifnot(!is.null(dataset$data_ptr))
      stopifnot(!is.null(outcome$data_ptr))
      stopifnot(!is.null(forest_model$tracker_ptr))
      stopifnot(!is.null(self$forest_container_ptr))
      stopifnot(
        num_samples_forest_container_cpp(self$forest_container_ptr) == 0
      )

      # Initialize the model
      initialize_forest_model_cpp(
        dataset$data_ptr,
        outcome$data_ptr,
        self$forest_container_ptr,
        forest_model$tracker_ptr,
        leaf_value,
        leaf_model_int
      )
    },

    #' @description
    #' Adjusts residual based on the predictions of a forest
    #'
    #' This is typically run just once at the beginning of a forest sampling algorithm.
    #' After trees are initialized with constant root node predictions, their root predictions are subtracted out of the residual.
    #' @param dataset `ForestDataset` object storing the covariates and bases for a given forest
    #' @param outcome `Outcome` object storing the residuals to be updated based on forest predictions
    #' @param forest_model `ForestModel` object storing tracking structures used in training / sampling
    #' @param requires_basis Whether or not a forest requires a basis for prediction
    #' @param forest_num Index of forest used to update residuals
    #' @param add Whether forest predictions should be added to or subtracted from residuals
    adjust_residual = function(
      dataset,
      outcome,
      forest_model,
      requires_basis,
      forest_num,
      add
    ) {
      stopifnot(!is.null(dataset$data_ptr))
      stopifnot(!is.null(outcome$data_ptr))
      stopifnot(!is.null(forest_model$tracker_ptr))
      stopifnot(!is.null(self$forest_container_ptr))

      adjust_residual_forest_container_cpp(
        dataset$data_ptr,
        outcome$data_ptr,
        self$forest_container_ptr,
        forest_model$tracker_ptr,
        requires_basis,
        forest_num,
        add
      )
    },

    #' @description
    #' Store the trees and metadata of `ForestDataset` class in a json file
    #' @param json_filename Name of output json file (must end in ".json")
    save_json = function(json_filename) {
      invisible(json_save_forest_container_cpp(
        self$forest_container_ptr,
        json_filename
      ))
    },

    #' @description
    #' Load trees and metadata for an ensemble from a json file. Note that
    #' any trees and metadata already present in `ForestDataset` class will
    #' be overwritten.
    #' @param json_filename Name of model input json file (must end in ".json")
    load_json = function(json_filename) {
      invisible(json_load_forest_container_cpp(
        self$forest_container_ptr,
        json_filename
      ))
    },

    #' @description
    #' Return number of samples in a `ForestContainer` object
    #' @return Sample count
    num_samples = function() {
      return(num_samples_forest_container_cpp(self$forest_container_ptr))
    },

    #' @description
    #' Return number of trees in each ensemble of a `ForestContainer` object
    #' @return Tree count
    num_trees = function() {
      return(num_trees_forest_container_cpp(self$forest_container_ptr))
    },

    #' @description
    #' Return output dimension of trees in a `ForestContainer` object
    #' @return Leaf node parameter size
    leaf_dimension = function() {
      return(leaf_dimension_forest_container_cpp(
        self$forest_container_ptr
      ))
    },

    #' @description
    #' Return constant leaf status of trees in a `ForestContainer` object
    #' @return `TRUE` if leaves are constant, `FALSE` otherwise
    is_leaf_constant = function() {
      return(is_leaf_constant_forest_container_cpp(
        self$forest_container_ptr
      ))
    },

    #' @description
    #' Return exponentiation status of trees in a `ForestContainer` object
    #' @return `TRUE` if leaf predictions must be exponentiated, `FALSE` otherwise
    is_exponentiated = function() {
      return(is_exponentiated_forest_container_cpp(
        self$forest_container_ptr
      ))
    },

    #' @description
    #' Add a new all-root ensemble to the container, with all of the leaves
    #' set to the value / vector provided
    #' @param leaf_value Value (or vector of values) to initialize root nodes in tree
    add_forest_with_constant_leaves = function(leaf_value) {
      if (length(leaf_value) > 1) {
        add_sample_vector_forest_container_cpp(
          self$forest_container_ptr,
          leaf_value
        )
      } else {
        add_sample_value_forest_container_cpp(
          self$forest_container_ptr,
          leaf_value
        )
      }
    },

    #' @description
    #' Add a numeric (i.e. `X[,i] <= c`) split to a given tree in the ensemble
    #' @param forest_num Index of the forest which contains the tree to be split
    #' @param tree_num Index of the tree to be split
    #' @param leaf_num Leaf to be split
    #' @param feature_num Feature that defines the new split
    #' @param split_threshold Value that defines the cutoff of the new split
    #' @param left_leaf_value Value (or vector of values) to assign to the newly created left node
    #' @param right_leaf_value Value (or vector of values) to assign to the newly created right node
    add_numeric_split_tree = function(
      forest_num,
      tree_num,
      leaf_num,
      feature_num,
      split_threshold,
      left_leaf_value,
      right_leaf_value
    ) {
      if (length(left_leaf_value) > 1) {
        add_numeric_split_tree_vector_forest_container_cpp(
          self$forest_container_ptr,
          forest_num,
          tree_num,
          leaf_num,
          feature_num,
          split_threshold,
          left_leaf_value,
          right_leaf_value
        )
      } else {
        add_numeric_split_tree_value_forest_container_cpp(
          self$forest_container_ptr,
          forest_num,
          tree_num,
          leaf_num,
          feature_num,
          split_threshold,
          left_leaf_value,
          right_leaf_value
        )
      }
    },

    #' @description
    #' Retrieve a vector of indices of leaf nodes for a given tree in a given forest
    #' @param forest_num Index of the forest which contains tree `tree_num`
    #' @param tree_num Index of the tree for which leaf indices will be retrieved
    get_tree_leaves = function(forest_num, tree_num) {
      return(get_tree_leaves_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num
      ))
    },

    #' @description
    #' Retrieve a vector of split counts for every training set variable in a given tree in a given forest
    #' @param forest_num Index of the forest which contains tree `tree_num`
    #' @param tree_num Index of the tree for which split counts will be retrieved
    #' @param num_features Total number of features in the training set
    get_tree_split_counts = function(forest_num, tree_num, num_features) {
      return(get_tree_split_counts_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        num_features
      ))
    },

    #' @description
    #' Retrieve a vector of split counts for every training set variable in a given forest
    #' @param forest_num Index of the forest for which split counts will be retrieved
    #' @param num_features Total number of features in the training set
    get_forest_split_counts = function(forest_num, num_features) {
      return(get_forest_split_counts_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        num_features
      ))
    },

    #' @description
    #' Retrieve a vector of split counts for every training set variable in a given forest, aggregated across ensembles and trees
    #' @param num_features Total number of features in the training set
    get_aggregate_split_counts = function(num_features) {
      return(get_overall_split_counts_forest_container_cpp(
        self$forest_container_ptr,
        num_features
      ))
    },

    #' @description
    #' Retrieve a vector of split counts for every training set variable in a given forest, reported separately for each ensemble and tree
    #' @param num_features Total number of features in the training set
    get_granular_split_counts = function(num_features) {
      n_samples <- self$num_samples()
      n_trees <- self$num_trees()
      output <- get_granular_split_count_array_forest_container_cpp(
        self$forest_container_ptr,
        num_features
      )
      dim(output) <- c(n_samples, n_trees, num_features)
      return(output)
    },

    #' @description
    #' Maximum depth of a specific tree in a specific ensemble in a `ForestSamples` object
    #' @param ensemble_num Ensemble number
    #' @param tree_num Tree index within ensemble `ensemble_num`
    #' @return Maximum leaf depth
    ensemble_tree_max_depth = function(ensemble_num, tree_num) {
      return(ensemble_tree_max_depth_forest_container_cpp(
        self$forest_container_ptr,
        ensemble_num,
        tree_num
      ))
    },

    #' @description
    #' Average the maximum depth of each tree in a given ensemble in a `ForestSamples` object
    #' @param ensemble_num Ensemble number
    #' @return Average maximum depth
    average_ensemble_max_depth = function(ensemble_num) {
      return(ensemble_average_max_depth_forest_container_cpp(
        self$forest_container_ptr,
        ensemble_num
      ))
    },

    #' @description
    #' Average the maximum depth of each tree in each ensemble in a `ForestContainer` object
    #' @return Average maximum depth
    average_max_depth = function() {
      return(average_max_depth_forest_container_cpp(
        self$forest_container_ptr
      ))
    },

    #' @description
    #' Number of leaves in a given ensemble in a `ForestSamples` object
    #' @param forest_num Index of the ensemble to be queried
    #' @return Count of leaves in the ensemble stored at `forest_num`
    num_forest_leaves = function(forest_num) {
      return(num_leaves_ensemble_forest_container_cpp(
        self$forest_container_ptr,
        forest_num
      ))
    },

    #' @description
    #' Sum of squared (raw) leaf values in a given ensemble in a `ForestSamples` object
    #' @param forest_num Index of the ensemble to be queried
    #' @return Average maximum depth
    sum_leaves_squared = function(forest_num) {
      return(sum_leaves_squared_ensemble_forest_container_cpp(
        self$forest_container_ptr,
        forest_num
      ))
    },

    #' @description
    #' Whether or not a given node of a given tree in a given forest in the `ForestSamples` is a leaf
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return `TRUE` if node is a leaf, `FALSE` otherwise
    is_leaf_node = function(forest_num, tree_num, node_id) {
      return(is_leaf_node_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Whether or not a given node of a given tree in a given forest in the `ForestSamples` is a numeric split node
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return `TRUE` if node is a numeric split node, `FALSE` otherwise
    is_numeric_split_node = function(forest_num, tree_num, node_id) {
      return(is_numeric_split_node_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Whether or not a given node of a given tree in a given forest in the `ForestSamples` is a categorical split node
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return `TRUE` if node is a categorical split node, `FALSE` otherwise
    is_categorical_split_node = function(forest_num, tree_num, node_id) {
      return(is_categorical_split_node_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Parent node of given node of a given tree in a given forest in a `ForestSamples` object
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Integer ID of the parent node
    parent_node = function(forest_num, tree_num, node_id) {
      return(parent_node_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Left child node of given node of a given tree in a given forest in a `ForestSamples` object
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Integer ID of the left child node
    left_child_node = function(forest_num, tree_num, node_id) {
      return(left_child_node_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Right child node of given node of a given tree in a given forest in a `ForestSamples` object
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Integer ID of the right child node
    right_child_node = function(forest_num, tree_num, node_id) {
      return(right_child_node_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Depth of given node of a given tree in a given forest in a `ForestSamples` object, with 0 depth for the root node.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Integer valued depth of the node
    node_depth = function(forest_num, tree_num, node_id) {
      return(node_depth_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Split index of given node of a given tree in a given forest in a `ForestSamples` object. Returns `-1` is node is a leaf.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Integer valued depth of the node
    node_split_index = function(forest_num, tree_num, node_id) {
      if (self$is_leaf_node(forest_num, tree_num, node_id)) {
        return(-1)
      } else {
        return(split_index_forest_container_cpp(
          self$forest_container_ptr,
          forest_num,
          tree_num,
          node_id
        ))
      }
    },

    #' @description
    #' Threshold that defines a numeric split for a given node of a given tree in a given forest in a `ForestSamples` object.
    #' Returns `Inf` if the node is a leaf or a categorical split node.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Threshold defining a split for the node
    node_split_threshold = function(forest_num, tree_num, node_id) {
      if (
        self$is_leaf_node(forest_num, tree_num, node_id) ||
          self$is_categorical_split_node(
            forest_num,
            tree_num,
            node_id
          )
      ) {
        return(Inf)
      } else {
        return(split_theshold_forest_container_cpp(
          self$forest_container_ptr,
          forest_num,
          tree_num,
          node_id
        ))
      }
    },

    #' @description
    #' Array of category indices that define a categorical split for a given node of a given tree in a given forest in a `ForestSamples` object.
    #' Returns `c(Inf)` if the node is a leaf or a numeric split node.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Categories defining a split for the node
    node_split_categories = function(forest_num, tree_num, node_id) {
      if (
        self$is_leaf_node(forest_num, tree_num, node_id) ||
          self$is_numeric_split_node(forest_num, tree_num, node_id)
      ) {
        return(c(Inf))
      } else {
        return(split_categories_forest_container_cpp(
          self$forest_container_ptr,
          forest_num,
          tree_num,
          node_id
        ))
      }
    },

    #' @description
    #' Leaf node value(s) for a given node of a given tree in a given forest in a `ForestSamples` object.
    #' Values are stale if the node is a split node.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @param node_id Index of the node to be queried
    #' @return Vector (often univariate) of leaf values
    node_leaf_values = function(forest_num, tree_num, node_id) {
      return(leaf_values_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num,
        node_id
      ))
    },

    #' @description
    #' Number of nodes in a given tree in a given forest in a `ForestSamples` object.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @return Count of total tree nodes
    num_nodes = function(forest_num, tree_num) {
      return(num_nodes_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num
      ))
    },

    #' @description
    #' Number of leaves in a given tree in a given forest in a `ForestSamples` object.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @return Count of total tree leaves
    num_leaves = function(forest_num, tree_num) {
      return(num_leaves_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num
      ))
    },

    #' @description
    #' Number of leaf parents (split nodes with two leaves as children) in a given tree in a given forest in a `ForestSamples` object.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @return Count of total tree leaf parents
    num_leaf_parents = function(forest_num, tree_num) {
      return(num_leaf_parents_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num
      ))
    },

    #' @description
    #' Number of split nodes in a given tree in a given forest in a `ForestSamples` object.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @return Count of total tree split nodes
    num_split_nodes = function(forest_num, tree_num) {
      return(num_split_nodes_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num
      ))
    },

    #' @description
    #' Array of node indices in a given tree in a given forest in a `ForestSamples` object.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @return Indices of tree nodes
    nodes = function(forest_num, tree_num) {
      return(nodes_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num
      ))
    },

    #' @description
    #' Array of leaf indices in a given tree in a given forest in a `ForestSamples` object.
    #' @param forest_num Index of the forest to be queried
    #' @param tree_num Index of the tree to be queried
    #' @return Indices of leaf nodes
    leaves = function(forest_num, tree_num) {
      return(leaves_forest_container_cpp(
        self$forest_container_ptr,
        forest_num,
        tree_num
      ))
    },

    #' @description
    #' Modify the ``ForestSamples`` object by removing the forest sample indexed by `forest_num
    #' @param forest_num Index of the forest to be removed
    delete_sample = function(forest_num) {
      return(remove_sample_forest_container_cpp(
        self$forest_container_ptr,
        forest_num
      ))
    }
  )
)

#' @title Forest C++ Wrapper
#' @description
#' Wrapper around a C++ class that stores a single ensemble of decision trees (often treated as the "active forest" / current state of a forest term in a sampling loop in R)
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

Forest <- R6::R6Class(
  classname = "Forest",
  cloneable = FALSE,
  public = list(
    #' @field forest_ptr External pointer to a C++ TreeEnsemble class
    forest_ptr = NULL,

    #' @field internal_forest_is_empty Whether the forest has not yet been "initialized" such that its `predict` function can be called.
    internal_forest_is_empty = TRUE,

    #' @description
    #' Create a new Forest object.
    #' @param num_trees Number of trees in the forest
    #' @param leaf_dimension Dimensionality of the outcome model
    #' @param is_leaf_constant Whether leaf is constant
    #' @param is_exponentiated Whether forest predictions should be exponentiated before being returned
    #' @return A new `Forest` object.
    initialize = function(
      num_trees,
      leaf_dimension = 1,
      is_leaf_constant = FALSE,
      is_exponentiated = FALSE
    ) {
      self$forest_ptr <- active_forest_cpp(
        num_trees,
        leaf_dimension,
        is_leaf_constant,
        is_exponentiated
      )
      self$internal_forest_is_empty <- TRUE
    },

    #' @description
    #' Create a larger forest by merging the trees of this forest with those of another forest
    #' @param forest Forest to be merged into this forest
    merge_forest = function(forest) {
      stopifnot(self$leaf_dimension() == forest$leaf_dimension())
      stopifnot(self$is_leaf_constant() == forest$is_leaf_constant())
      stopifnot(self$is_exponentiated() == forest$is_exponentiated())
      forest_merge_cpp(self$forest_ptr, forest$forest_ptr)
    },

    #' @description
    #' Add a constant value to every leaf of every tree in an ensemble. If leaves are multi-dimensional, `constant_value` will be added to every dimension of the leaves.
    #' @param constant_value Value that will be added to every leaf of every tree
    add_constant = function(constant_value) {
      forest_add_constant_cpp(self$forest_ptr, constant_value)
    },

    #' @description
    #' Multiply every leaf of every tree by a constant value. If leaves are multi-dimensional, `constant_multiple` will be multiplied through every dimension of the leaves.
    #' @param constant_multiple Value that will be multiplied by every leaf of every tree
    multiply_constant = function(constant_multiple) {
      forest_multiply_constant_cpp(self$forest_ptr, constant_multiple)
    },

    #' @description
    #' Predict forest on every sample in `forest_dataset`
    #' @param forest_dataset `ForestDataset` R class
    #' @return vector of predictions with as many rows as in `forest_dataset`
    predict = function(forest_dataset) {
      stopifnot(!is.null(forest_dataset$data_ptr))
      stopifnot(!is.null(self$forest_ptr))
      return(predict_active_forest_cpp(
        self$forest_ptr,
        forest_dataset$data_ptr
      ))
    },

    #' @description
    #' Predict "raw" leaf values (without being multiplied by basis) for every sample in `forest_dataset`
    #' @param forest_dataset `ForestDataset` R class
    #' @return Array of predictions for each observation in `forest_dataset` and
    #' each sample in the `ForestSamples` class with each prediction having the
    #' dimensionality of the forests' leaf model. In the case of a constant leaf model
    #' or univariate leaf regression, this array is a vector (length is the number of
    #' observations). In the case of a multivariate leaf regression,
    #' this array is a matrix (number of observations by leaf model dimension,
    #' number of samples).
    predict_raw = function(forest_dataset) {
      stopifnot(!is.null(forest_dataset$data_ptr))
      # Unpack dimensions
      output_dim <- leaf_dimension_active_forest_cpp(self$forest_ptr)
      n <- dataset_num_rows_cpp(forest_dataset$data_ptr)

      # Predict leaf values from forest
      predictions <- predict_raw_active_forest_cpp(
        self$forest_ptr,
        forest_dataset$data_ptr
      )
      if (output_dim > 1) {
        dim(predictions) <- c(n, output_dim)
      }

      return(predictions)
    },

    #' @description
    #' Set a constant predicted value for every tree in the ensemble.
    #' Stops program if any tree is more than a root node.
    #' @param leaf_value Constant leaf value(s) to be fixed for each tree in the ensemble indexed by `forest_num`. Can be either a single number or a vector, depending on the forest's leaf dimension.
    set_root_leaves = function(leaf_value) {
      stopifnot(!is.null(self$forest_ptr))
      stopifnot(self$internal_forest_is_empty)

      # Set leaf values
      if (length(leaf_value) == 1) {
        stopifnot(
          leaf_dimension_active_forest_cpp(self$forest_ptr) == 1
        )
        set_leaf_value_active_forest_cpp(self$forest_ptr, leaf_value)
      } else if (length(leaf_value) > 1) {
        stopifnot(
          leaf_dimension_active_forest_cpp(self$forest_ptr) ==
            length(leaf_value)
        )
        set_leaf_vector_active_forest_cpp(self$forest_ptr, leaf_value)
      } else {
        stop(
          "leaf_value must be a numeric value or vector of length >= 1"
        )
      }

      self$internal_forest_is_empty = FALSE
    },

    #' @description
    #' Set a constant predicted value for every tree in the ensemble.
    #' Stops program if any tree is more than a root node.
    #' @param dataset `ForestDataset` Dataset class (covariates, basis, etc...)
    #' @param outcome `Outcome` Outcome class (residual / partial residual)
    #' @param forest_model `ForestModel` object storing tracking structures used in training / sampling
    #' @param leaf_model_int Integer value encoding the leaf model type (0 = constant gaussian, 1 = univariate gaussian, 2 = multivariate gaussian, 3 = log linear variance).
    #' @param leaf_value Constant leaf value(s) to be fixed for each tree in the ensemble indexed by `forest_num`. Can be either a single number or a vector, depending on the forest's leaf dimension.
    prepare_for_sampler = function(
      dataset,
      outcome,
      forest_model,
      leaf_model_int,
      leaf_value
    ) {
      stopifnot(!is.null(dataset$data_ptr))
      stopifnot(!is.null(outcome$data_ptr))
      stopifnot(!is.null(forest_model$tracker_ptr))
      stopifnot(!is.null(self$forest_ptr))
      stopifnot(self$internal_forest_is_empty)

      # Initialize the model
      initialize_forest_model_active_forest_cpp(
        dataset$data_ptr,
        outcome$data_ptr,
        self$forest_ptr,
        forest_model$tracker_ptr,
        leaf_value,
        leaf_model_int
      )

      self$internal_forest_is_empty = FALSE
    },

    #' @description
    #' Adjusts residual based on the predictions of a forest
    #'
    #' This is typically run just once at the beginning of a forest sampling algorithm.
    #' After trees are initialized with constant root node predictions, their root predictions are subtracted out of the residual.
    #' @param dataset `ForestDataset` object storing the covariates and bases for a given forest
    #' @param outcome `Outcome` object storing the residuals to be updated based on forest predictions
    #' @param forest_model `ForestModel` object storing tracking structures used in training / sampling
    #' @param requires_basis Whether or not a forest requires a basis for prediction
    #' @param add Whether forest predictions should be added to or subtracted from residuals
    adjust_residual = function(
      dataset,
      outcome,
      forest_model,
      requires_basis,
      add
    ) {
      stopifnot(!is.null(dataset$data_ptr))
      stopifnot(!is.null(outcome$data_ptr))
      stopifnot(!is.null(forest_model$tracker_ptr))
      stopifnot(!is.null(self$forest_ptr))

      adjust_residual_active_forest_cpp(
        dataset$data_ptr,
        outcome$data_ptr,
        self$forest_ptr,
        forest_model$tracker_ptr,
        requires_basis,
        add
      )
    },

    #' @description
    #' Return number of trees in each ensemble of a `Forest` object
    #' @return Tree count
    num_trees = function() {
      return(num_trees_active_forest_cpp(self$forest_ptr))
    },

    #' @description
    #' Return output dimension of trees in a `Forest` object
    #' @return Leaf node parameter size
    leaf_dimension = function() {
      return(leaf_dimension_active_forest_cpp(self$forest_ptr))
    },

    #' @description
    #' Return constant leaf status of trees in a `Forest` object
    #' @return `TRUE` if leaves are constant, `FALSE` otherwise
    is_leaf_constant = function() {
      return(is_leaf_constant_active_forest_cpp(self$forest_ptr))
    },

    #' @description
    #' Return exponentiation status of trees in a `Forest` object
    #' @return `TRUE` if leaf predictions must be exponentiated, `FALSE` otherwise
    is_exponentiated = function() {
      return(is_exponentiated_active_forest_cpp(self$forest_ptr))
    },

    #' @description
    #' Add a numeric (i.e. `X[,i] <= c`) split to a given tree in the ensemble
    #' @param tree_num Index of the tree to be split
    #' @param leaf_num Leaf to be split
    #' @param feature_num Feature that defines the new split
    #' @param split_threshold Value that defines the cutoff of the new split
    #' @param left_leaf_value Value (or vector of values) to assign to the newly created left node
    #' @param right_leaf_value Value (or vector of values) to assign to the newly created right node
    add_numeric_split_tree = function(
      tree_num,
      leaf_num,
      feature_num,
      split_threshold,
      left_leaf_value,
      right_leaf_value
    ) {
      if (length(left_leaf_value) > 1) {
        add_numeric_split_tree_vector_active_forest_cpp(
          self$forest_ptr,
          tree_num,
          leaf_num,
          feature_num,
          split_threshold,
          left_leaf_value,
          right_leaf_value
        )
      } else {
        add_numeric_split_tree_value_active_forest_cpp(
          self$forest_ptr,
          tree_num,
          leaf_num,
          feature_num,
          split_threshold,
          left_leaf_value,
          right_leaf_value
        )
      }
    },

    #' @description
    #' Retrieve a vector of indices of leaf nodes for a given tree in a given forest
    #' @param tree_num Index of the tree for which leaf indices will be retrieved
    get_tree_leaves = function(tree_num) {
      return(get_tree_leaves_active_forest_cpp(self$forest_ptr, tree_num))
    },

    #' @description
    #' Retrieve a vector of split counts for every training set variable in a given tree in the forest
    #' @param tree_num Index of the tree for which split counts will be retrieved
    #' @param num_features Total number of features in the training set
    get_tree_split_counts = function(tree_num, num_features) {
      return(get_tree_split_counts_active_forest_cpp(
        self$forest_ptr,
        tree_num,
        num_features
      ))
    },

    #' @description
    #' Retrieve a vector of split counts for every training set variable in the forest
    #' @param num_features Total number of features in the training set
    get_forest_split_counts = function(num_features) {
      return(get_overall_split_counts_active_forest_cpp(
        self$forest_ptr,
        num_features
      ))
    },

    #' @description
    #' Maximum depth of a specific tree in the forest
    #' @param tree_num Tree index within forest
    #' @return Maximum leaf depth
    tree_max_depth = function(tree_num) {
      return(ensemble_tree_max_depth_active_forest_cpp(
        self$forest_ptr,
        tree_num
      ))
    },

    #' @description
    #' Average the maximum depth of each tree in the forest
    #' @return Average maximum depth
    average_max_depth = function() {
      return(ensemble_average_max_depth_active_forest_cpp(
        self$forest_ptr
      ))
    },

    #' @description
    #' When a forest object is created, it is "empty" in the sense that none
    #' of its component trees have leaves with values. There are two ways to
    #' "initialize" a Forest object. First, the `set_root_leaves()` method
    #' simply initializes every tree in the forest to a single node carrying
    #' the same (user-specified) leaf value. Second, the `prepare_for_sampler()`
    #' method initializes every tree in the forest to a single node with the
    #' same value and also propagates this information through to a ForestModel
    #' object, which must be synchronized with a Forest during a forest
    #' sampler loop.
    #' @return `TRUE` if a Forest has not yet been initialized with a constant
    #' root value, `FALSE` otherwise if the forest has already been
    #' initialized / grown.
    is_empty = function() {
      return(self$internal_forest_is_empty)
    }
  )
)

#' @title Create ForestSamples Object
#' @description Create a container of forest samples
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param num_trees Number of trees
#' @param leaf_dimension Dimensionality of the outcome model
#' @param is_leaf_constant Whether leaf is constant
#' @param is_exponentiated Whether forest predictions should be exponentiated before being returned
#'
#' @return `ForestSamples` object
#' @export
#'
#' @examples
#' num_trees <- 100
#' leaf_dimension <- 2
#' is_leaf_constant <- FALSE
#' is_exponentiated <- FALSE
#' forest_samples <- createForestSamples(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
createForestSamples <- function(
  num_trees,
  leaf_dimension = 1,
  is_leaf_constant = FALSE,
  is_exponentiated = FALSE
) {
  return(invisible(
    (ForestSamples$new(
      num_trees,
      leaf_dimension,
      is_leaf_constant,
      is_exponentiated
    ))
  ))
}

#' @title Create Forest Object
#' @description Create a forest
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param num_trees Number of trees in the forest
#' @param leaf_dimension Dimensionality of the outcome model
#' @param is_leaf_constant Whether leaf is constant
#' @param is_exponentiated Whether forest predictions should be exponentiated before being returned
#'
#' @return `Forest` object
#' @export
#'
#' @examples
#' num_trees <- 100
#' leaf_dimension <- 2
#' is_leaf_constant <- FALSE
#' is_exponentiated <- FALSE
#' forest <- createForest(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
createForest <- function(
  num_trees,
  leaf_dimension = 1,
  is_leaf_constant = FALSE,
  is_exponentiated = FALSE
) {
  return(invisible(
    (Forest$new(
      num_trees,
      leaf_dimension,
      is_leaf_constant,
      is_exponentiated
    ))
  ))
}

#' @title Reset Active Forest
#' @rdname ForestStateManagement
#' @param active_forest Current active forest
#' @param forest_samples (Optional) Container of forest samples from which to re-initialize active forest. If not provided, active forest will be reset to an ensemble of single-node (i.e. root) trees.
#' @param forest_num (Optional) Index of forest samples from which to initialize active forest. If not provided, active forest will be reset to an ensemble of single-node (i.e. root) trees.
#' @export
resetActiveForest <- function(
  active_forest,
  forest_samples = NULL,
  forest_num = NULL
) {
  if (is.null(forest_samples)) {
    root_reset_active_forest_cpp(active_forest$forest_ptr)
    active_forest$internal_forest_is_empty = TRUE
  } else {
    if (is.null(forest_num)) {
      stop(
        "`forest_num` must be specified if `forest_samples` is provided"
      )
    }
    reset_active_forest_cpp(
      active_forest$forest_ptr,
      forest_samples$forest_container_ptr,
      forest_num
    )
  }
}

#' @title Reset Forest Model
#' @rdname ForestStateManagement
#' @param forest_model Forest model with tracking data structures
#' @param forest Forest from which to re-initialize forest model
#' @param dataset Training dataset object
#' @param residual Residual which will also be updated
#' @param is_mean_model Whether the model being updated is a conditional mean model
#' @export
resetForestModel <- function(
  forest_model,
  forest,
  dataset,
  residual,
  is_mean_model
) {
  reset_forest_model_cpp(
    forest_model$tracker_ptr,
    forest$forest_ptr,
    dataset$data_ptr,
    residual$data_ptr,
    is_mean_model
  )
}

#' @title Summarize a ForestSamples Object
#' @description Prints a summary of the ForestSamples object, including number of forests and the underlying model of each forest.
#' @param x ForestSamples object
#' @param ... Additional arguments
#' @export
#' @return ForestSamples object unchanged after printing summary
print.ForestSamples <- function(x, ...) {
  # Construct summary message
  num_samples <- x$num_samples()
  summary_message <- paste0(
    "Forest container with ",
    num_samples,
    " samples. Each forest contains "
  )
  num_trees <- x$num_trees()
  is_leaf_constant <- x$is_leaf_constant()
  leaf_dimension <- x$leaf_dimension()
  summary_message <- paste0(summary_message, num_trees, " trees with ")
  if (is_leaf_constant) {
    summary_message <- paste0(summary_message, "a constant leaf model.")
  } else {
    summary_message <- paste0(
      summary_message,
      "a ",
      leaf_dimension,
      "-dimensional leaf regression model."
    )
  }

  # Print the forest container details
  cat(summary_message, "\n")

  # Return forest container invisibly
  invisible(x)
}

test_that("Univariate forest construction", {
  # Create dataset and forest container
  num_trees <- 10
  # fmt: skip
  X = matrix(c(1.5, 8.7, 1.2, 
                 2.7, 3.4, 5.4, 
                 3.6, 1.2, 9.3, 
                 4.4, 5.4, 10.4, 
                 5.3, 9.3, 3.6, 
                 6.1, 10.4, 4.4), 
               byrow = TRUE, nrow = 6)
  n <- nrow(X)
  p <- ncol(X)
  forest_dataset = createForestDataset(X)
  forest <- createForest(num_trees, 1, TRUE)

  # Initialize forest with 0.0 root predictions
  forest$set_root_leaves(0.)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest$predict(forest_dataset)
  pred_raw <- forest$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Split the root of the first tree in the ensemble at X[,1] > 4.0
  forest$add_numeric_split_tree(0, 0, 0, 4.0, -5., 5.)

  # Check that predictions are the same (since the leaf is constant)
  pred <- forest$predict(forest_dataset)
  pred_raw <- forest$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Split the left leaf of the first tree in the ensemble at X[,2] > 4.0
  forest$add_numeric_split_tree(0, 1, 1, 4.0, -7.5, -2.5)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest$predict(forest_dataset)
  pred_raw <- forest$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Check the split count for the first tree in the ensemble
  split_counts <- forest$get_tree_split_counts(0, p)
  split_counts_expected <- c(1, 1, 0)

  # Assertion
  expect_equal(split_counts, split_counts_expected)
})

test_that("Univariate forest construction and low-level merge / arithmetic ops", {
  # Create dataset and forest container
  num_trees <- 10
  # fmt: skip
  X = matrix(c(1.5, 8.7, 1.2, 
                 2.7, 3.4, 5.4, 
                 3.6, 1.2, 9.3, 
                 4.4, 5.4, 10.4, 
                 5.3, 9.3, 3.6, 
                 6.1, 10.4, 4.4), 
               byrow = TRUE, nrow = 6)
  n <- nrow(X)
  p <- ncol(X)
  forest_dataset = createForestDataset(X)
  forest1 <- createForest(num_trees, 1, TRUE)
  forest2 <- createForest(num_trees, 1, TRUE)

  # Initialize forests with 0.0 root predictions
  forest1$set_root_leaves(0.)
  forest2$set_root_leaves(0.)

  # Check that predictions are as expected
  pred1 <- forest1$predict(forest_dataset)
  pred2 <- forest2$predict(forest_dataset)
  pred_exp1 <- c(0, 0, 0, 0, 0, 0)
  pred_exp2 <- c(0, 0, 0, 0, 0, 0)

  # Assertion
  expect_equal(pred1, pred_exp1)
  expect_equal(pred2, pred_exp2)

  # Split the root of the first tree of the first forest in the ensemble at X[,1] > 4.0
  forest1$add_numeric_split_tree(0, 0, 0, 4.0, -5., 5.)

  # Split the root of the first tree of the second forest in the ensemble at X[,1] > 3.0
  forest2$add_numeric_split_tree(0, 0, 0, 3.0, -1., 1.)

  # Check that predictions are as expected
  pred1 <- forest1$predict(forest_dataset)
  pred2 <- forest2$predict(forest_dataset)
  pred_exp1 <- c(-5, -5, -5, 5, 5, 5)
  pred_exp2 <- c(-1, -1, 1, 1, 1, 1)

  # Assertion
  expect_equal(pred1, pred_exp1)
  expect_equal(pred2, pred_exp2)

  # Split the left leaf of the first tree of the first forest in the ensemble at X[,2] > 4.0
  forest1$add_numeric_split_tree(0, 1, 1, 4.0, -7.5, -2.5)

  # Split the left leaf of the first tree of the first forest in the ensemble at X[,2] > 4.0
  forest2$add_numeric_split_tree(0, 1, 1, 4.0, -1.5, -0.5)

  # Check that predictions are as expected
  pred1 <- forest1$predict(forest_dataset)
  pred2 <- forest2$predict(forest_dataset)
  pred_exp1 <- c(-2.5, -7.5, -7.5, 5, 5, 5)
  pred_exp2 <- c(-0.5, -1.5, 1, 1, 1, 1)

  # Assertion
  expect_equal(pred1, pred_exp1)
  expect_equal(pred2, pred_exp2)

  # Merge forests
  forest1$merge_forest(forest2)

  # Check that predictions are as expected
  pred <- forest1$predict(forest_dataset)
  pred_exp <- c(-3.0, -9.0, -6.5, 6.0, 6.0, 6.0)

  # Assertion
  expect_equal(pred, pred_exp)

  # Add constant to every value of the combined forest
  forest1$add_constant(0.5)

  # Check that predictions are as expected
  pred <- forest1$predict(forest_dataset)
  pred_exp <- c(7.0, 1.0, 3.5, 16.0, 16.0, 16.0)

  # Assertion
  expect_equal(pred, pred_exp)

  # Check that "old" forest is still intact
  pred <- forest2$predict(forest_dataset)
  pred_exp <- c(-0.5, -1.5, 1, 1, 1, 1)

  # Assertion
  expect_equal(pred, pred_exp)

  # Subtract constant back off of every value of the combined forest
  forest1$add_constant(-0.5)

  # Check that predictions are as expected
  pred <- forest1$predict(forest_dataset)
  pred_exp <- c(-3.0, -9.0, -6.5, 6.0, 6.0, 6.0)

  # Assertion
  expect_equal(pred, pred_exp)

  # Multiply every value of the combined forest by a constant
  forest1$multiply_constant(2.0)

  # Check that predictions are as expected
  pred <- forest1$predict(forest_dataset)
  pred_exp <- c(-6.0, -18.0, -13.0, 12.0, 12.0, 12.0)

  # Assertion
  expect_equal(pred, pred_exp)
})

test_that("Univariate constant forest container", {
  # Create dataset and forest container
  num_trees <- 10
  # fmt: skip
  X = matrix(c(1.5, 8.7, 1.2, 
                 2.7, 3.4, 5.4, 
                 3.6, 1.2, 9.3, 
                 4.4, 5.4, 10.4, 
                 5.3, 9.3, 3.6, 
                 6.1, 10.4, 4.4), 
               byrow = TRUE, nrow = 6)
  n <- nrow(X)
  p <- ncol(X)
  forest_dataset = createForestDataset(X)
  forest_samples <- createForestSamples(num_trees, 1, TRUE)

  # Initialize a forest with constant root predictions
  forest_samples$add_forest_with_constant_leaves(0.)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Split the root of the first tree in the ensemble at X[,1] > 4.0
  forest_samples$add_numeric_split_tree(0, 0, 0, 0, 4.0, -5., 5.)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Split the left leaf of the first tree in the ensemble at X[,2] > 4.0
  forest_samples$add_numeric_split_tree(0, 0, 1, 1, 4.0, -7.5, -2.5)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Multiply first forest by 2.0
  forest_samples$multiply_forest(0, 2.0)

  # Check that predictions are all double
  pred_orig <- pred
  pred_expected <- pred * 2.0
  pred <- forest_samples$predict(forest_dataset)

  # Assertion
  expect_equal(pred, pred_expected)

  # Add 1.0 to every tree in first forest
  forest_samples$add_to_forest(0, 1.0)

  # Check that predictions are += num_trees
  pred_expected <- pred + num_trees
  pred <- forest_samples$predict(forest_dataset)

  # Assertion
  expect_equal(pred, pred_expected)

  # Initialize a new forest with constant root predictions
  forest_samples$add_forest_with_constant_leaves(0.)

  # Split the second forest as the first forest was split
  forest_samples$add_numeric_split_tree(1, 0, 0, 0, 4.0, -5., 5.)
  forest_samples$add_numeric_split_tree(1, 0, 1, 1, 4.0, -7.5, -2.5)

  # Check that predictions are as expected
  pred <- forest_samples$predict(forest_dataset)
  pred_expected_new <- cbind(pred_expected, pred_orig)

  # Assertion
  expect_equal(pred, pred_expected_new)

  # Combine second forest with the first forest
  forest_samples$combine_forests(c(0, 1))

  # Check that predictions are as expected
  pred <- forest_samples$predict(forest_dataset)
  pred_expected_new <- cbind(pred_expected + pred_orig, pred_orig)

  # Assertion
  expect_equal(pred, pred_expected_new)

  # Divide first forest predictions by 2
  forest_samples$multiply_forest(0, 0.5)

  # Check that predictions are as expected
  pred <- forest_samples$predict(forest_dataset)
  pred_expected_new <- cbind((pred_expected + pred_orig) / 2.0, pred_orig)

  # Assertion
  expect_equal(pred, pred_expected_new)

  # Delete second forest
  forest_samples$delete_sample(1)

  # Check that predictions are as expected
  pred <- forest_samples$predict(forest_dataset)
  pred_expected_new <- (pred_expected + pred_orig) / 2.0

  # Assertion
  expect_equal(pred, pred_expected_new)

  # Test print method
  forest_output <- capture.output(print(forest_samples))
  expect_true(any(grepl("constant", forest_output, fixed = TRUE)))
  expect_true(any(grepl(num_trees, forest_output, fixed = TRUE)))
  expect_true(any(grepl(
    forest_samples$num_samples(),
    forest_output,
    fixed = TRUE
  )))
})

test_that("Collapse forests", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (
        ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) + 
        ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) + 
        ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) + 
        ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
    )
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Create forest dataset
  forest_dataset_test <- createForestDataset(covariates = X_test)

  # Run BART for 50 iterations
  num_mcmc <- 50
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = general_param_list
  )

  # Extract the mean forest container
  mean_forest_container <- bart_model$mean_forests

  # Predict from the original container
  pred_orig <- mean_forest_container$predict(forest_dataset_test)

  # Collapse the container in batches of 5
  batch_size <- 5
  mean_forest_container$collapse(batch_size)

  # Predict from the modified container
  pred_new <- mean_forest_container$predict(forest_dataset_test)

  # Check that corresponding (sums of) predictions match
  batch_inds <- (seq(1, num_mcmc, 1) -
    (num_mcmc -
      (num_mcmc %/% (num_mcmc %/% batch_size)) * (num_mcmc %/% batch_size)) -
    1) %/%
    batch_size +
    1
  pred_orig_collapsed <- matrix(
    NA,
    nrow = nrow(pred_orig),
    ncol = max(batch_inds)
  )
  for (i in 1:max(batch_inds)) {
    pred_orig_collapsed[, i] <- rowSums(pred_orig[, batch_inds == i]) /
      sum(batch_inds == i)
  }

  # Assertion
  expect_equal(pred_orig_collapsed, pred_new)

  # Now run BART for 52 iterations
  num_mcmc <- 52
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = general_param_list
  )

  # Extract the mean forest container
  mean_forest_container <- bart_model$mean_forests

  # Predict from the original container
  pred_orig <- mean_forest_container$predict(forest_dataset_test)

  # Collapse the container in batches of 5
  batch_size <- 5
  mean_forest_container$collapse(batch_size)

  # Predict from the modified container
  pred_new <- mean_forest_container$predict(forest_dataset_test)

  # Check that corresponding (sums of) predictions match
  batch_inds <- (seq(1, num_mcmc, 1) -
    (num_mcmc -
      (num_mcmc %/% (num_mcmc %/% batch_size)) * (num_mcmc %/% batch_size)) -
    1) %/%
    batch_size +
    1
  pred_orig_collapsed <- matrix(
    NA,
    nrow = nrow(pred_orig),
    ncol = max(batch_inds)
  )
  for (i in 1:max(batch_inds)) {
    pred_orig_collapsed[, i] <- rowSums(pred_orig[, batch_inds == i]) /
      sum(batch_inds == i)
  }

  # Assertion
  expect_equal(pred_orig_collapsed, pred_new)

  # Now run BART for 5 iterations
  num_mcmc <- 5
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = general_param_list
  )

  # Extract the mean forest container
  mean_forest_container <- bart_model$mean_forests

  # Predict from the original container
  pred_orig <- mean_forest_container$predict(forest_dataset_test)

  # Collapse the container in batches of 5
  batch_size <- 5
  mean_forest_container$collapse(batch_size)

  # Predict from the modified container
  pred_new <- mean_forest_container$predict(forest_dataset_test)

  # Check that corresponding (sums of) predictions match
  pred_orig_collapsed <- as.matrix(rowSums(pred_orig) / batch_size)

  # Assertion
  expect_equal(pred_orig_collapsed, pred_new)

  # Now run BART for 4 iterations
  num_mcmc <- 4
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = general_param_list
  )

  # Extract the mean forest container
  mean_forest_container <- bart_model$mean_forests

  # Predict from the original container
  pred_orig <- mean_forest_container$predict(forest_dataset_test)

  # Collapse the container in batches of 5
  batch_size <- 5
  mean_forest_container$collapse(batch_size)

  # Predict from the modified container
  pred_new <- mean_forest_container$predict(forest_dataset_test)

  # Check that corresponding (sums of) predictions match
  pred_orig_collapsed <- pred_orig

  # Assertion
  expect_equal(pred_orig_collapsed, pred_new)

  # Test print method
  forest_output <- capture.output(print(mean_forest_container))
  expect_true(any(grepl("constant", forest_output, fixed = TRUE)))
  expect_true(any(grepl(
    mean_forest_container$num_trees(),
    forest_output,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    mean_forest_container$num_samples(),
    forest_output,
    fixed = TRUE
  )))
})

test_that("Prediction from trees with constant leaf", {
  # Create dataset and forest container
  num_trees <- 10
  # fmt: skip
  X = matrix(c(1.5, 8.7, 1.2, 
                 2.7, 3.4, 5.4, 
                 3.6, 1.2, 9.3, 
                 4.4, 5.4, 10.4, 
                 5.3, 9.3, 3.6, 
                 6.1, 10.4, 4.4), 
               byrow = TRUE, nrow = 6)
  n <- nrow(X)
  p <- ncol(X)
  forest_dataset = createForestDataset(X)
  forest_samples <- createForestSamples(num_trees, 1, TRUE)

  # Initialize a forest with constant root predictions
  forest_samples$add_forest_with_constant_leaves(0.)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Split the root of the first tree in the ensemble at X[,1] > 4.0
  forest_samples$add_numeric_split_tree(0, 0, 0, 0, 4.0, -5., 5.)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Split the left leaf of the first tree in the ensemble at X[,2] > 4.0
  forest_samples$add_numeric_split_tree(0, 0, 1, 1, 4.0, -7.5, -2.5)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Check the split count for the first tree in the ensemble
  split_counts <- forest_samples$get_tree_split_counts(0, 0, p)
  split_counts_expected <- c(1, 1, 0)

  # Assertion
  expect_equal(split_counts, split_counts_expected)
})

test_that("Prediction from trees with univariate leaf basis", {
  # Create dataset and forest container
  num_trees <- 10
  # fmt: skip
  X = matrix(c(1.5, 8.7, 1.2, 
                 2.7, 3.4, 5.4, 
                 3.6, 1.2, 9.3, 
                 4.4, 5.4, 10.4, 
                 5.3, 9.3, 3.6, 
                 6.1, 10.4, 4.4), 
               byrow = TRUE, nrow = 6)
  W = as.matrix(c(-1, -1, -1, 1, 1, 1))
  n <- nrow(X)
  p <- ncol(X)
  forest_dataset = createForestDataset(X, W)
  forest_samples <- createForestSamples(num_trees, 1, FALSE)

  # Initialize a forest with constant root predictions
  forest_samples$add_forest_with_constant_leaves(0.)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)

  # Assertion
  expect_equal(pred, pred_raw)

  # Split the root of the first tree in the ensemble at X[,1] > 4.0
  forest_samples$add_numeric_split_tree(0, 0, 0, 0, 4.0, -5., 5.)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)
  pred_manual <- pred_raw * W

  # Assertion
  expect_equal(pred, pred_manual)

  # Split the left leaf of the first tree in the ensemble at X[,2] > 4.0
  forest_samples$add_numeric_split_tree(0, 0, 1, 1, 4.0, -7.5, -2.5)

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)
  pred_manual <- pred_raw * W

  # Assertion
  expect_equal(pred, pred_manual)

  # Check the split count for the first tree in the ensemble
  split_counts <- forest_samples$get_tree_split_counts(0, 0, p)
  split_counts_expected <- c(1, 1, 0)

  # Assertion
  expect_equal(split_counts, split_counts_expected)
})

test_that("Prediction from trees with multivariate leaf basis", {
  # Create dataset and forest container
  num_trees <- 10
  output_dim <- 2
  num_samples <- 0
  # fmt: skip
  X = matrix(c(1.5, 8.7, 1.2, 
                 2.7, 3.4, 5.4, 
                 3.6, 1.2, 9.3, 
                 4.4, 5.4, 10.4, 
                 5.3, 9.3, 3.6, 
                 6.1, 10.4, 4.4), 
               byrow = TRUE, nrow = 6)
  n <- nrow(X)
  p <- ncol(X)
  W = matrix(c(1, 1, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1), byrow = FALSE, nrow = 6)
  forest_dataset = createForestDataset(X, W)
  forest_samples <- createForestSamples(num_trees, output_dim, FALSE)

  # Initialize a forest with constant root predictions
  forest_samples$add_forest_with_constant_leaves(c(1., 1.))
  num_samples <- num_samples + 1

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)
  pred_intermediate <- as.numeric(pred_raw) * as.numeric(W)
  dim(pred_intermediate) <- c(n, output_dim, num_samples)
  pred_manual <- apply(pred_intermediate, 3, function(x) rowSums(x))

  # Assertion
  expect_equal(pred, pred_manual)

  # Split the root of the first tree in the ensemble at X[,1] > 4.0
  forest_samples$add_numeric_split_tree(0, 0, 0, 0, 4.0, c(-5., -1.), c(5., 1.))

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)
  pred_intermediate <- as.numeric(pred_raw) * as.numeric(W)
  dim(pred_intermediate) <- c(n, output_dim, num_samples)
  pred_manual <- apply(pred_intermediate, 3, function(x) rowSums(x))

  # Assertion
  expect_equal(pred, pred_manual)

  # Split the left leaf of the first tree in the ensemble at X[,2] > 4.0
  forest_samples$add_numeric_split_tree(
    0,
    0,
    1,
    1,
    4.0,
    c(-7.5, 2.5),
    c(-2.5, 7.5)
  )

  # Check that regular and "raw" predictions are the same (since the leaf is constant)
  pred <- forest_samples$predict(forest_dataset)
  pred_raw <- forest_samples$predict_raw(forest_dataset)
  pred_intermediate <- as.numeric(pred_raw) * as.numeric(W)
  dim(pred_intermediate) <- c(n, output_dim, num_samples)
  pred_manual <- apply(pred_intermediate, 3, function(x) rowSums(x))

  # Assertion
  expect_equal(pred, pred_manual)

  # Check the split count for the first tree in the ensemble
  split_counts <- forest_samples$get_tree_split_counts(0, 0, 3)
  split_counts_expected <- c(1, 1, 0)

  # Assertion
  expect_equal(split_counts, split_counts_expected)
})

test_that("BART predictions with pre-summarization", {
  # Generate data and test-train split
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
                 ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
                 ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
                 ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit a "classic" BART model
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10
  )

  # Check that the default predict method returns a list
  pred <- predict(bart_model, X = X_test)
  y_hat_posterior_test <- pred$y_hat
  expect_equal(dim(y_hat_posterior_test), c(20, 10))

  # Check that the pre-aggregated predictions match with those computed by rowMeans
  pred_mean <- predict(bart_model, X = X_test, type = "mean")
  y_hat_mean_test <- pred_mean$y_hat
  expect_equal(y_hat_mean_test, rowMeans(y_hat_posterior_test))

  # Check that we warn and return a NULL when requesting terms that weren't fit
  expect_warning({
    pred_mean <- predict(
      bart_model,
      X = X_test,
      type = "mean",
      terms = c("rfx", "variance_forest")
    )
  })
  expect_equal(NULL, pred_mean)

  # Fit a heteroskedastic BART model
  var_params <- list(num_trees = 20)
  het_bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    variance_forest_params = var_params
  )

  # Check that the default predict method returns a list
  pred <- predict(het_bart_model, X = X_test)
  y_hat_posterior_test <- pred$y_hat
  sigma2_hat_posterior_test <- pred$variance_forest_predictions

  # Assertion
  expect_equal(dim(y_hat_posterior_test), c(20, 10))
  expect_equal(dim(sigma2_hat_posterior_test), c(20, 10))

  # Check that the pre-aggregated predictions match with those computed by rowMeans
  pred_mean <- predict(het_bart_model, X = X_test, type = "mean")
  y_hat_mean_test <- pred_mean$y_hat
  sigma2_hat_mean_test <- pred_mean$variance_forest_predictions

  # Assertion
  expect_equal(y_hat_mean_test, rowMeans(y_hat_posterior_test))
  expect_equal(sigma2_hat_mean_test, rowMeans(sigma2_hat_posterior_test))

  # Check that the "single-term" pre-aggregated predictions
  # match those computed by pre-aggregated predictions returned in a list
  y_hat_mean_test_single_term <- predict(
    het_bart_model,
    X = X_test,
    type = "mean",
    terms = "y_hat"
  )
  sigma2_hat_mean_test_single_term <- predict(
    het_bart_model,
    X = X_test,
    type = "mean",
    terms = "variance_forest"
  )

  # Assertion
  expect_equal(y_hat_mean_test, y_hat_mean_test_single_term)
  expect_equal(sigma2_hat_mean_test, sigma2_hat_mean_test_single_term)
})

test_that("BART predictions with random effects", {
  # Generate data and test-train split
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
                 ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
                 ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
                 ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  noise_sd <- 1
  rfx_group_ids <- sample(1:3, n, replace = TRUE)
  rfx_coefs <- c(-2, 0, 2)
  rfx_term <- rfx_coefs[rfx_group_ids]
  rfx_basis <- matrix(1, nrow = n, ncol = 1)
  y <- f_XW + rfx_term + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit a "classic" BART model
  rfx_params <- list(model_spec = "intercept_only")
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    rfx_group_ids_train = rfx_group_ids_train,
    random_effects_params = rfx_params,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10
  )

  # Check that the default predict method returns a list
  pred <- predict(bart_model, X = X_test, rfx_group_ids = rfx_group_ids_test)
  y_hat_posterior_test <- pred$y_hat
  expect_equal(dim(y_hat_posterior_test), c(20, 10))

  # Check that the pre-aggregated predictions match with those computed by rowMeans
  pred_mean <- predict(
    bart_model,
    X = X_test,
    rfx_group_ids = rfx_group_ids_test,
    type = "mean"
  )
  y_hat_mean_test <- pred_mean$y_hat
  expect_equal(y_hat_mean_test, rowMeans(y_hat_posterior_test))

  # Check that we warn and return a NULL when requesting terms that weren't fit
  expect_warning({
    pred_mean <- predict(
      bart_model,
      X = X_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "mean",
      terms = c("variance_forest")
    )
  })
  expect_equal(NULL, pred_mean)
})

test_that("BCF predictions with pre-summarization", {
  # Generate data and test-train split
  n <- 100
  g <- function(x) {
    ifelse(x[, 5] == 1, 2, ifelse(x[, 5] == 2, -1, -4))
  }
  x1 <- rnorm(n)
  x2 <- rnorm(n)
  x3 <- rnorm(n)
  x4 <- as.numeric(rbinom(n, 1, 0.5))
  x5 <- as.numeric(sample(1:3, n, replace = TRUE))
  X <- cbind(x1, x2, x3, x4, x5)
  p <- ncol(X)
  mu_x <- 1 + g(X) + X[, 1] * X[, 3]
  tau_x <- 1 + 2 * X[, 2] * X[, 4]
  pi_x <- 0.8 *
    pnorm((3 * mu_x / sd(mu_x)) - 0.5 * X[, 1]) +
    0.05 +
    runif(n) / 10
  Z <- rbinom(n, 1, pi_x)
  E_XZ <- mu_x + Z * tau_x
  snr <- 2
  y <- E_XZ + rnorm(n, 0, 1) * (sd(E_XZ) / snr)
  X <- as.data.frame(X)
  X$x4 <- factor(X$x4, ordered = TRUE)
  X$x5 <- factor(X$x5, ordered = TRUE)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  pi_test <- pi_x[test_inds]
  pi_train <- pi_x[train_inds]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit a "classic" BCF model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10
  )

  # Check that the default predict method returns a list
  pred <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test
  )
  y_hat_posterior_test <- pred$y_hat
  expect_equal(dim(y_hat_posterior_test), c(20, 10))

  # Check that the pre-aggregated predictions match with those computed by rowMeans
  pred_mean <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    type = "mean"
  )
  y_hat_mean_test <- pred_mean$y_hat
  expect_equal(y_hat_mean_test, rowMeans(y_hat_posterior_test))

  # Check that we warn and return a NULL when requesting terms that weren't fit
  expect_warning({
    pred_mean <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      type = "mean",
      terms = c("rfx", "variance_forest")
    )
  })
  expect_equal(NULL, pred_mean)

  # Fit a heteroskedastic BCF model
  var_params <- list(num_trees = 20)
  expect_warning(
    het_bcf_model <- bcf(
      X_train = X_train,
      Z_train = Z_train,
      y_train = y_train,
      propensity_train = pi_train,
      X_test = X_test,
      Z_test = Z_test,
      propensity_test = pi_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      variance_forest_params = var_params
    )
  )

  # Check that the default predict method returns a list
  pred <- predict(
    het_bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test
  )
  y_hat_posterior_test <- pred$y_hat
  sigma2_hat_posterior_test <- pred$variance_forest_predictions

  # Assertion
  expect_equal(dim(y_hat_posterior_test), c(20, 10))
  expect_equal(dim(sigma2_hat_posterior_test), c(20, 10))

  # Check that the pre-aggregated predictions match with those computed by rowMeans
  pred_mean <- predict(
    het_bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    type = "mean"
  )
  y_hat_mean_test <- pred_mean$y_hat
  sigma2_hat_mean_test <- pred_mean$variance_forest_predictions

  # Assertion
  expect_equal(y_hat_mean_test, rowMeans(y_hat_posterior_test))
  expect_equal(sigma2_hat_mean_test, rowMeans(sigma2_hat_posterior_test))

  # Check that the "single-term" pre-aggregated predictions
  # match those computed by pre-aggregated predictions returned in a list
  y_hat_mean_test_single_term <- predict(
    het_bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    type = "mean",
    terms = "y_hat"
  )
  sigma2_hat_mean_test_single_term <- predict(
    het_bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    type = "mean",
    terms = "variance_forest"
  )

  # Assertion
  expect_equal(y_hat_mean_test, y_hat_mean_test_single_term)
  expect_equal(sigma2_hat_mean_test, sigma2_hat_mean_test_single_term)
})

test_that("BCF predictions with random effects", {
  # Generate data and test-train split
  n <- 100
  g <- function(x) {
    ifelse(x[, 5] == 1, 2, ifelse(x[, 5] == 2, -1, -4))
  }
  x1 <- rnorm(n)
  x2 <- rnorm(n)
  x3 <- rnorm(n)
  x4 <- as.numeric(rbinom(n, 1, 0.5))
  x5 <- as.numeric(sample(1:3, n, replace = TRUE))
  X <- cbind(x1, x2, x3, x4, x5)
  p <- ncol(X)
  mu_x <- 1 + g(X) + X[, 1] * X[, 3]
  tau_x <- 1 + 2 * X[, 2] * X[, 4]
  pi_x <- 0.8 *
    pnorm((3 * mu_x / sd(mu_x)) - 0.5 * X[, 1]) +
    0.05 +
    runif(n) / 10
  Z <- rbinom(n, 1, pi_x)
  E_XZ <- mu_x + Z * tau_x
  rfx_group_ids <- sample(1:3, n, replace = TRUE)
  rfx_basis <- cbind(1, Z)
  rfx_coefs <- matrix(
    c(
      -2,
      -0.5,
      0,
      0.0,
      2,
      0.5
    ),
    byrow = T,
    ncol = 2
  )
  rfx_term <- rowSums(rfx_basis * rfx_coefs[rfx_group_ids, ])
  snr <- 2
  y <- E_XZ + rfx_term + rnorm(n, 0, 1) * (sd(E_XZ + rfx_term) / snr)
  X <- as.data.frame(X)
  X$x4 <- factor(X$x4, ordered = TRUE)
  X$x5 <- factor(X$x5, ordered = TRUE)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  pi_test <- pi_x[test_inds]
  pi_train <- pi_x[train_inds]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit a BCF model with random intercept and random slope on Z
  rfx_params = list(model_spec = "intercept_plus_treatment")
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    rfx_group_ids_train = rfx_group_ids_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    rfx_group_ids_test = rfx_group_ids_test,
    random_effects_params = rfx_params,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10
  )

  # Check that the default predict method returns a list
  pred <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    rfx_group_ids = rfx_group_ids_test
  )
  y_hat_posterior_test <- pred$y_hat
  expect_equal(dim(y_hat_posterior_test), c(20, 10))

  # Check that the pre-aggregated predictions match with those computed by rowMeans
  pred_mean <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    rfx_group_ids = rfx_group_ids_test,
    type = "mean"
  )
  y_hat_mean_test <- pred_mean$y_hat
  expect_equal(y_hat_mean_test, rowMeans(y_hat_posterior_test))

  # Check that we warn and return a NULL when requesting terms that weren't fit
  expect_warning({
    pred_mean <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      type = "mean",
      terms = c("variance_forest")
    )
  })
  expect_equal(NULL, pred_mean)

  # Fit a BCF model with random intercept only
  # Fit a BCF model with random intercept and random slope on Z
  rfx_params = list(model_spec = "intercept_only")
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    rfx_group_ids_train = rfx_group_ids_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    rfx_group_ids_test = rfx_group_ids_test,
    random_effects_params = rfx_params,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10
  )

  # Check that the default predict method returns a list
  pred <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    rfx_group_ids = rfx_group_ids_test
  )
  y_hat_posterior_test <- pred$y_hat
  expect_equal(dim(y_hat_posterior_test), c(20, 10))

  # Check that the pre-aggregated predictions match with those computed by rowMeans
  pred_mean <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    rfx_group_ids = rfx_group_ids_test,
    type = "mean"
  )
  y_hat_mean_test <- pred_mean$y_hat
  expect_equal(y_hat_mean_test, rowMeans(y_hat_posterior_test))

  # Check that we warn and return a NULL when requesting terms that weren't fit
  expect_warning({
    pred_mean <- predict(
      bcf_model,
      X = X_test,
      Z = Z_test,
      propensity = pi_test,
      type = "mean",
      terms = c("variance_forest")
    )
  })
})

test_that("BART cloglog binary: posterior interval and contrast", {
  # Generate binary cloglog data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  true_lambda <- X %*% beta
  prob_y1 <- 1 - exp(-exp(true_lambda))
  y <- rbinom(n, 1, prob_y1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]

  # Fit binary cloglog BART model
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "binary", link = "cloglog")
    )
  )

  # Test posterior interval on linear scale
  interval_linear <- computeBARTPosteriorInterval(
    bart_model,
    terms = "mean_forest",
    level = 0.95,
    scale = "linear",
    X = X_test
  )
  expect_true(is.list(interval_linear))
  expect_true("lower" %in% names(interval_linear))
  expect_true("upper" %in% names(interval_linear))
  expect_equal(length(interval_linear$lower), n_test)
  expect_equal(length(interval_linear$upper), n_test)
  expect_true(all(interval_linear$lower <= interval_linear$upper))

  # Test posterior interval on probability scale
  interval_prob <- computeBARTPosteriorInterval(
    bart_model,
    terms = "mean_forest",
    level = 0.95,
    scale = "probability",
    X = X_test
  )
  expect_true(is.list(interval_prob))
  expect_equal(length(interval_prob$lower), n_test)
  expect_true(all(interval_prob$lower >= 0 & interval_prob$lower <= 1))
  expect_true(all(interval_prob$upper >= 0 & interval_prob$upper <= 1))
  expect_true(all(interval_prob$lower <= interval_prob$upper))

  # Test contrast on linear scale
  X0 <- X_test
  X1 <- X_test + 0.5
  contrast_linear <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "posterior",
    scale = "linear"
  )
  expect_true(is.matrix(contrast_linear))
  expect_equal(nrow(contrast_linear), n_test)

  # Test contrast on probability scale
  contrast_prob <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "posterior",
    scale = "probability"
  )
  expect_true(is.matrix(contrast_prob))
  expect_equal(nrow(contrast_prob), n_test)

  # Test contrast with type = "mean"
  contrast_mean_linear <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "mean",
    scale = "linear"
  )
  expect_true(is.numeric(contrast_mean_linear))
  expect_equal(length(contrast_mean_linear), n_test)

  contrast_mean_prob <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "mean",
    scale = "probability"
  )
  expect_true(is.numeric(contrast_mean_prob))
  expect_equal(length(contrast_mean_prob), n_test)
})

test_that("BART cloglog ordinal: posterior interval and contrast", {
  # Generate ordinal cloglog data (3 categories)
  set.seed(42)
  n <- 500
  p <- 3
  n_categories <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  true_lambda <- X %*% beta
  gamma_true <- c(-1.5, -0.5)

  # Compute class probabilities
  true_probs <- matrix(0, nrow = n, ncol = n_categories)
  for (j in 1:n_categories) {
    if (j == 1) {
      true_probs[, j] <- 1 - exp(-exp(gamma_true[j] + true_lambda))
    } else if (j == n_categories) {
      true_probs[, j] <- 1 - rowSums(true_probs[, 1:(j - 1), drop = FALSE])
    } else {
      true_probs[, j] <- exp(-exp(gamma_true[j - 1] + true_lambda)) *
        (1 - exp(-exp(gamma_true[j] + true_lambda)))
    }
  }

  # Generate ordinal outcomes
  y <- sapply(1:n, function(i) {
    sample(1:n_categories, 1, prob = true_probs[i, ])
  })

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]

  # Fit ordinal cloglog BART model
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "ordinal", link = "cloglog")
    )
  )

  # Test posterior interval on linear scale
  interval_linear <- computeBARTPosteriorInterval(
    bart_model,
    terms = "mean_forest",
    level = 0.95,
    scale = "linear",
    X = X_test
  )
  expect_true(is.list(interval_linear))
  expect_equal(length(interval_linear$lower), n_test)
  expect_true(all(interval_linear$lower <= interval_linear$upper))

  # Test posterior interval on probability scale
  # For ordinal models, probability scale returns survival probabilities P(Y > k)
  # which are (n_test, n_categories - 1) matrices
  interval_prob <- computeBARTPosteriorInterval(
    bart_model,
    terms = "mean_forest",
    level = 0.95,
    scale = "probability",
    X = X_test
  )
  expect_true(is.list(interval_prob))
  expect_equal(nrow(interval_prob$lower), n_test)
  expect_equal(ncol(interval_prob$lower), n_categories - 1)
  expect_equal(nrow(interval_prob$upper), n_test)
  expect_equal(ncol(interval_prob$upper), n_categories - 1)
  expect_true(all(interval_prob$lower >= 0 & interval_prob$lower <= 1))
  expect_true(all(interval_prob$upper >= 0 & interval_prob$upper <= 1))
  expect_true(all(interval_prob$lower <= interval_prob$upper))

  # Test contrast on linear scale
  X0 <- X_test
  X1 <- X_test + 0.5
  contrast_linear <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "posterior",
    scale = "linear"
  )
  expect_true(is.matrix(contrast_linear))
  expect_equal(nrow(contrast_linear), n_test)

  # Test contrast on probability scale (ordinal returns 3D array)
  contrast_prob <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "posterior",
    scale = "probability"
  )
  expect_true(is.array(contrast_prob))
  expect_equal(dim(contrast_prob)[1], n_test)
  expect_equal(dim(contrast_prob)[2], n_categories - 1)

  # Test contrast with type = "mean" on linear scale
  contrast_mean_linear <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "mean",
    scale = "linear"
  )
  expect_true(is.numeric(contrast_mean_linear))
  expect_equal(length(contrast_mean_linear), n_test)

  # Test contrast with type = "mean" on probability scale
  # For ordinal, mean contrast should be (n_test, n_categories - 1) matrix
  contrast_mean_prob <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "mean",
    scale = "probability"
  )
  expect_true(is.matrix(contrast_mean_prob))
  expect_equal(nrow(contrast_mean_prob), n_test)
  expect_equal(ncol(contrast_mean_prob), n_categories - 1)
})

test_that("BART cloglog binary: sample posterior predictive", {
  # Generate binary cloglog data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  true_lambda <- X %*% beta
  prob_y1 <- 1 - exp(-exp(true_lambda))
  y <- rbinom(n, 1, prob_y1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit binary cloglog BART model
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "binary", link = "cloglog")
    )
  )

  # Test with multiple draws per sample
  ppd <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 3
  )
  expect_equal(dim(ppd), c(n_test, num_mcmc, 3))
  expect_true(all(ppd %in% c(0, 1)))

  # Test with single draw per sample
  ppd1 <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 1
  )
  expect_equal(dim(ppd1), c(n_test, num_mcmc))
  expect_true(all(ppd1 %in% c(0, 1)))
})

test_that("BART cloglog ordinal: sample posterior predictive", {
  # Generate ordinal cloglog data (3 categories)
  set.seed(42)
  n <- 500
  p <- 3
  n_categories <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  true_lambda <- X %*% beta
  gamma_true <- c(-1.5, -0.5)

  # Compute class probabilities
  true_probs <- matrix(0, nrow = n, ncol = n_categories)
  for (j in 1:n_categories) {
    if (j == 1) {
      true_probs[, j] <- 1 - exp(-exp(gamma_true[j] + true_lambda))
    } else if (j == n_categories) {
      true_probs[, j] <- 1 - rowSums(true_probs[, 1:(j - 1), drop = FALSE])
    } else {
      true_probs[, j] <- exp(-exp(gamma_true[j - 1] + true_lambda)) *
        (1 - exp(-exp(gamma_true[j] + true_lambda)))
    }
  }

  # Generate ordinal outcomes
  y <- sapply(1:n, function(i) {
    sample(1:n_categories, 1, prob = true_probs[i, ])
  })

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit ordinal cloglog BART model
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "ordinal", link = "cloglog")
    )
  )

  # Test with multiple draws per sample
  ppd <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 3
  )
  expect_equal(dim(ppd), c(n_test, num_mcmc, 3))
  expect_true(all(ppd %in% 1:n_categories))

  # Test with single draw per sample
  ppd1 <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 1
  )
  expect_equal(dim(ppd1), c(n_test, num_mcmc))
  expect_true(all(ppd1 %in% 1:n_categories))
})

test_that("BART gaussian: posterior interval and contrast", {
  # Generate gaussian data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  mu <- X %*% beta
  y <- rnorm(n, mu, 1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit gaussian BART model (default link)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc
  )

  # Test posterior interval on linear scale (mean_forest term)
  interval_linear <- computeBARTPosteriorInterval(
    bart_model,
    terms = "mean_forest",
    level = 0.95,
    scale = "linear",
    X = X_test
  )
  expect_true(is.list(interval_linear))
  expect_true("lower" %in% names(interval_linear))
  expect_true("upper" %in% names(interval_linear))
  expect_equal(length(interval_linear$lower), n_test)
  expect_equal(length(interval_linear$upper), n_test)
  expect_true(all(interval_linear$lower <= interval_linear$upper))

  # Test posterior interval for y_hat term
  interval_yhat <- computeBARTPosteriorInterval(
    bart_model,
    terms = "y_hat",
    level = 0.95,
    scale = "linear",
    X = X_test
  )
  expect_true(is.list(interval_yhat))
  expect_equal(length(interval_yhat$lower), n_test)
  expect_true(all(interval_yhat$lower <= interval_yhat$upper))

  # Test contrast on linear scale, type = "posterior"
  X0 <- X_test
  X1 <- X_test + 0.5
  contrast_posterior <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "posterior",
    scale = "linear"
  )
  expect_true(is.matrix(contrast_posterior))
  expect_equal(nrow(contrast_posterior), n_test)
  expect_equal(ncol(contrast_posterior), num_mcmc)

  # Test contrast with type = "mean"
  contrast_mean <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "mean",
    scale = "linear"
  )
  expect_true(is.numeric(contrast_mean))
  expect_equal(length(contrast_mean), n_test)
})

test_that("BART binary probit: posterior interval and contrast", {
  # Generate binary probit data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  mu <- X %*% beta
  prob_y1 <- pnorm(mu)
  y <- rbinom(n, 1, prob_y1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit binary probit BART model
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "binary", link = "probit")
    )
  )

  # Test posterior interval on linear scale
  interval_linear <- computeBARTPosteriorInterval(
    bart_model,
    terms = "mean_forest",
    level = 0.95,
    scale = "linear",
    X = X_test
  )
  expect_true(is.list(interval_linear))
  expect_true("lower" %in% names(interval_linear))
  expect_true("upper" %in% names(interval_linear))
  expect_equal(length(interval_linear$lower), n_test)
  expect_equal(length(interval_linear$upper), n_test)
  expect_true(all(interval_linear$lower <= interval_linear$upper))

  # Test posterior interval on probability scale
  interval_prob <- computeBARTPosteriorInterval(
    bart_model,
    terms = "mean_forest",
    level = 0.95,
    scale = "probability",
    X = X_test
  )
  expect_true(is.list(interval_prob))
  expect_equal(length(interval_prob$lower), n_test)
  expect_equal(length(interval_prob$upper), n_test)
  expect_true(all(interval_prob$lower >= 0 & interval_prob$lower <= 1))
  expect_true(all(interval_prob$upper >= 0 & interval_prob$upper <= 1))
  expect_true(all(interval_prob$lower <= interval_prob$upper))

  # Test contrast on linear scale
  X0 <- X_test
  X1 <- X_test + 0.5
  contrast_linear <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "posterior",
    scale = "linear"
  )
  expect_true(is.matrix(contrast_linear))
  expect_equal(nrow(contrast_linear), n_test)
  expect_equal(ncol(contrast_linear), num_mcmc)

  # Test contrast on probability scale
  contrast_prob <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "posterior",
    scale = "probability"
  )
  expect_true(is.matrix(contrast_prob))
  expect_equal(nrow(contrast_prob), n_test)
  expect_equal(ncol(contrast_prob), num_mcmc)

  # Test contrast with type = "mean" on linear scale
  contrast_mean_linear <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "mean",
    scale = "linear"
  )
  expect_true(is.numeric(contrast_mean_linear))
  expect_equal(length(contrast_mean_linear), n_test)

  # Test contrast with type = "mean" on probability scale
  contrast_mean_prob <- computeContrastBARTModel(
    bart_model,
    X_0 = X0,
    X_1 = X1,
    type = "mean",
    scale = "probability"
  )
  expect_true(is.numeric(contrast_mean_prob))
  expect_equal(length(contrast_mean_prob), n_test)
})

test_that("BART gaussian: sample posterior predictive", {
  # Generate gaussian data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  mu <- X %*% beta
  y <- rnorm(n, mu, 1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit gaussian BART model (default link)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc
  )

  # Test with multiple draws per sample
  ppd <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 3
  )
  expect_equal(dim(ppd), c(n_test, num_mcmc, 3))
  expect_true(is.numeric(ppd))

  # Test with single draw per sample
  ppd1 <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 1
  )
  expect_equal(dim(ppd1), c(n_test, num_mcmc))
  expect_true(is.numeric(ppd1))
})

test_that("BART binary probit: sample posterior predictive", {
  # Generate binary probit data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  beta <- rep(1 / sqrt(p), p)
  mu <- X %*% beta
  prob_y1 <- pnorm(mu)
  y <- rbinom(n, 1, prob_y1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit binary probit BART model
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "binary", link = "probit")
    )
  )

  # Test with multiple draws per sample
  ppd <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 3
  )
  expect_equal(dim(ppd), c(n_test, num_mcmc, 3))
  expect_true(all(ppd %in% c(0, 1)))

  # Test with single draw per sample
  ppd1 <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    num_draws_per_sample = 1
  )
  expect_equal(dim(ppd1), c(n_test, num_mcmc))
  expect_true(all(ppd1 %in% c(0, 1)))
})

test_that("BCF gaussian: posterior interval and contrast", {
  # Generate gaussian causal data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  pi_x <- pnorm(X[, 1])
  Z <- rbinom(n, 1, pi_x)
  mu_x <- X[, 1] + X[, 2]
  tau_x <- rep(1, n)
  y <- mu_x + tau_x * Z + rnorm(n)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]
  Z_test <- Z[test_inds]
  pi_train <- pi_x[train_inds]
  pi_test <- pi_x[test_inds]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit gaussian BCF model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc
  )

  # Test posterior interval on linear scale (prognostic_function term)
  interval_prog <- computeBCFPosteriorInterval(
    bcf_model,
    terms = "prognostic_function",
    level = 0.95,
    scale = "linear",
    X = X_test,
    Z = Z_test,
    propensity = pi_test
  )
  expect_true(is.list(interval_prog))
  expect_true("lower" %in% names(interval_prog))
  expect_true("upper" %in% names(interval_prog))
  expect_equal(length(interval_prog$lower), n_test)
  expect_equal(length(interval_prog$upper), n_test)
  expect_true(all(interval_prog$lower <= interval_prog$upper))

  # Test posterior interval for cate term
  interval_cate <- computeBCFPosteriorInterval(
    bcf_model,
    terms = "cate",
    level = 0.95,
    scale = "linear",
    X = X_test,
    Z = Z_test,
    propensity = pi_test
  )
  expect_true(is.list(interval_cate))
  expect_equal(length(interval_cate$lower), n_test)
  expect_true(all(interval_cate$lower <= interval_cate$upper))

  # Test posterior interval for y_hat term
  interval_yhat <- computeBCFPosteriorInterval(
    bcf_model,
    terms = "y_hat",
    level = 0.95,
    scale = "linear",
    X = X_test,
    Z = Z_test,
    propensity = pi_test
  )
  expect_true(is.list(interval_yhat))
  expect_equal(length(interval_yhat$lower), n_test)
  expect_true(all(interval_yhat$lower <= interval_yhat$upper))

  # Test contrast on linear scale (CATE: Z=1 vs Z=0)
  contrast_posterior <- computeContrastBCFModel(
    bcf_model,
    X_0 = X_test,
    X_1 = X_test,
    Z_0 = rep(0, n_test),
    Z_1 = rep(1, n_test),
    propensity_0 = pi_test,
    propensity_1 = pi_test,
    type = "posterior",
    scale = "linear"
  )
  expect_true(is.matrix(contrast_posterior))
  expect_equal(nrow(contrast_posterior), n_test)
  expect_equal(ncol(contrast_posterior), num_mcmc)

  # Test contrast with type = "mean"
  contrast_mean <- computeContrastBCFModel(
    bcf_model,
    X_0 = X_test,
    X_1 = X_test,
    Z_0 = rep(0, n_test),
    Z_1 = rep(1, n_test),
    propensity_0 = pi_test,
    propensity_1 = pi_test,
    type = "mean",
    scale = "linear"
  )
  expect_true(is.numeric(contrast_mean))
  expect_equal(length(contrast_mean), n_test)
})

test_that("BCF binary probit: posterior interval and contrast", {
  # Generate binary probit causal data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  pi_x <- pnorm(X[, 1])
  Z <- rbinom(n, 1, pi_x)
  mu_x <- X[, 1] + X[, 2]
  tau_x <- rep(0.5, n)
  prob_y1 <- pnorm(mu_x + tau_x * Z)
  y <- rbinom(n, 1, prob_y1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]
  Z_test <- Z[test_inds]
  pi_train <- pi_x[train_inds]
  pi_test <- pi_x[test_inds]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit binary probit BCF model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "binary", link = "probit")
    )
  )

  # Test posterior interval on linear scale
  interval_linear <- computeBCFPosteriorInterval(
    bcf_model,
    terms = "prognostic_function",
    level = 0.95,
    scale = "linear",
    X = X_test,
    Z = Z_test,
    propensity = pi_test
  )
  expect_true(is.list(interval_linear))
  expect_true("lower" %in% names(interval_linear))
  expect_true("upper" %in% names(interval_linear))
  expect_equal(length(interval_linear$lower), n_test)
  expect_equal(length(interval_linear$upper), n_test)
  expect_true(all(interval_linear$lower <= interval_linear$upper))

  # Test posterior interval on probability scale
  interval_prob <- computeBCFPosteriorInterval(
    bcf_model,
    terms = "prognostic_function",
    level = 0.95,
    scale = "probability",
    X = X_test,
    Z = Z_test,
    propensity = pi_test
  )
  expect_true(is.list(interval_prob))
  expect_equal(length(interval_prob$lower), n_test)
  expect_equal(length(interval_prob$upper), n_test)
  expect_true(all(interval_prob$lower >= 0 & interval_prob$lower <= 1))
  expect_true(all(interval_prob$upper >= 0 & interval_prob$upper <= 1))
  expect_true(all(interval_prob$lower <= interval_prob$upper))

  # Test contrast on linear scale (CATE: Z=1 vs Z=0)
  contrast_linear <- computeContrastBCFModel(
    bcf_model,
    X_0 = X_test,
    X_1 = X_test,
    Z_0 = rep(0, n_test),
    Z_1 = rep(1, n_test),
    propensity_0 = pi_test,
    propensity_1 = pi_test,
    type = "posterior",
    scale = "linear"
  )
  expect_true(is.matrix(contrast_linear))
  expect_equal(nrow(contrast_linear), n_test)
  expect_equal(ncol(contrast_linear), num_mcmc)

  # Test contrast on probability scale
  contrast_prob <- computeContrastBCFModel(
    bcf_model,
    X_0 = X_test,
    X_1 = X_test,
    Z_0 = rep(0, n_test),
    Z_1 = rep(1, n_test),
    propensity_0 = pi_test,
    propensity_1 = pi_test,
    type = "posterior",
    scale = "probability"
  )
  expect_true(is.matrix(contrast_prob))
  expect_equal(nrow(contrast_prob), n_test)
  expect_equal(ncol(contrast_prob), num_mcmc)

  # Test contrast with type = "mean" on linear scale
  contrast_mean_linear <- computeContrastBCFModel(
    bcf_model,
    X_0 = X_test,
    X_1 = X_test,
    Z_0 = rep(0, n_test),
    Z_1 = rep(1, n_test),
    propensity_0 = pi_test,
    propensity_1 = pi_test,
    type = "mean",
    scale = "linear"
  )
  expect_true(is.numeric(contrast_mean_linear))
  expect_equal(length(contrast_mean_linear), n_test)

  # Test contrast with type = "mean" on probability scale
  contrast_mean_prob <- computeContrastBCFModel(
    bcf_model,
    X_0 = X_test,
    X_1 = X_test,
    Z_0 = rep(0, n_test),
    Z_1 = rep(1, n_test),
    propensity_0 = pi_test,
    propensity_1 = pi_test,
    type = "mean",
    scale = "probability"
  )
  expect_true(is.numeric(contrast_mean_prob))
  expect_equal(length(contrast_mean_prob), n_test)
})

test_that("BCF gaussian: sample posterior predictive", {
  # Generate gaussian causal data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  pi_x <- pnorm(X[, 1])
  Z <- rbinom(n, 1, pi_x)
  mu_x <- X[, 1] + X[, 2]
  tau_x <- rep(1, n)
  y <- mu_x + tau_x * Z + rnorm(n)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]
  Z_test <- Z[test_inds]
  pi_train <- pi_x[train_inds]
  pi_test <- pi_x[test_inds]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit gaussian BCF model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc
  )

  # Test with multiple draws per sample
  ppd <- sampleBCFPosteriorPredictive(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    num_draws_per_sample = 3
  )
  expect_equal(dim(ppd), c(n_test, num_mcmc, 3))
  expect_true(is.numeric(ppd))

  # Test with single draw per sample
  ppd1 <- sampleBCFPosteriorPredictive(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    num_draws_per_sample = 1
  )
  expect_equal(dim(ppd1), c(n_test, num_mcmc))
  expect_true(is.numeric(ppd1))
})

test_that("BCF binary probit: sample posterior predictive", {
  # Generate binary probit causal data
  set.seed(42)
  n <- 100
  p <- 3
  X <- matrix(runif(n * p), ncol = p)
  pi_x <- pnorm(X[, 1])
  Z <- rbinom(n, 1, pi_x)
  mu_x <- X[, 1] + X[, 2]
  tau_x <- rep(0.5, n)
  prob_y1 <- pnorm(mu_x + tau_x * Z)
  y <- rbinom(n, 1, prob_y1)

  # Train/test split
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]
  Z_test <- Z[test_inds]
  pi_train <- pi_x[train_inds]
  pi_test <- pi_x[test_inds]
  y_train <- y[train_inds]
  num_mcmc <- 10

  # Fit binary probit BCF model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(
      sample_sigma2_global = FALSE,
      outcome_model = OutcomeModel(outcome = "binary", link = "probit")
    )
  )

  # Test with multiple draws per sample
  ppd <- sampleBCFPosteriorPredictive(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    num_draws_per_sample = 3
  )
  expect_equal(dim(ppd), c(n_test, num_mcmc, 3))
  expect_true(all(ppd %in% c(0, 1)))

  # Test with single draw per sample
  ppd1 <- sampleBCFPosteriorPredictive(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = pi_test,
    num_draws_per_sample = 1
  )
  expect_equal(dim(ppd1), c(n_test, num_mcmc))
  expect_true(all(ppd1 %in% c(0, 1)))
})

test_that("Residual updates correctly propagated after forest sampling step", {
  # Setup
  # Create dataset
  # fmt: skip
  X = matrix(c(1.5, 8.7, 1.2, 
                 2.7, 3.4, 5.4, 
                 3.6, 1.2, 9.3, 
                 4.4, 5.4, 10.4, 
                 5.3, 9.3, 3.6, 
                 6.1, 10.4, 4.4), 
               byrow = TRUE, nrow = 6)
  W = matrix(c(1, 1, 1, 1, 1, 1), nrow = 6)
  n = nrow(X)
  p = ncol(X)
  y = as.matrix(ifelse(X[, 1] > 4, -5, 5) + rnorm(n, 0, 1))
  y_bar = mean(y)
  y_std = sd(y)
  resid = (y - y_bar) / y_std
  forest_dataset = createForestDataset(X, W)
  residual = createOutcome(resid)
  variable_weights = rep(1.0 / p, p)
  feature_types = as.integer(rep(0, p))

  # Forest parameters
  num_trees = 50
  alpha = 0.95
  beta = 2.0
  min_samples_leaf = 1
  current_sigma2 = 1.
  current_leaf_scale = as.matrix(1. / num_trees, nrow = 1, ncol = 1)
  cutpoint_grid_size = 100
  max_depth = 10
  a_forest = 0
  b_forest = 0

  # RNG
  cpp_rng = createCppRNG(-1)

  # Create forest sampler and forest container
  global_model_config = createGlobalModelConfig(
    global_error_variance = current_sigma2
  )
  forest_model_config = createForestModelConfig(
    feature_types = feature_types,
    num_trees = num_trees,
    num_observations = n,
    alpha = alpha,
    beta = beta,
    min_samples_leaf = min_samples_leaf,
    max_depth = max_depth,
    leaf_model_type = 0,
    leaf_model_scale = current_leaf_scale,
    variable_weights = variable_weights,
    variance_forest_shape = a_forest,
    variance_forest_scale = b_forest,
    cutpoint_grid_size = cutpoint_grid_size
  )
  forest_model = createForestModel(
    forest_dataset,
    forest_model_config,
    global_model_config
  )
  forest_samples = createForestSamples(num_trees, 1, FALSE)
  active_forest = createForest(num_trees, 1, FALSE)

  # Initialize the leaves of each tree in the prognostic forest
  active_forest$prepare_for_sampler(
    forest_dataset,
    residual,
    forest_model,
    0,
    mean(resid)
  )
  active_forest$adjust_residual(
    forest_dataset,
    residual,
    forest_model,
    FALSE,
    FALSE
  )

  # Run the forest sampling algorithm for a single iteration
  forest_model$sample_one_iteration(
    forest_dataset,
    residual,
    forest_samples,
    active_forest,
    cpp_rng,
    forest_model_config,
    global_model_config,
    keep_forest = TRUE,
    gfr = TRUE
  )

  # Get the current residual after running the sampler
  initial_resid = residual$get_data()

  # Get initial prediction from the tree ensemble
  initial_yhat = as.numeric(forest_samples$predict(forest_dataset))

  # Update the basis vector
  scalar = 2.0
  W_update = W * scalar
  forest_dataset$update_basis(W_update)

  # Update residual to reflect adjusted basis
  forest_model$propagate_basis_update(forest_dataset, residual, active_forest)

  # Get updated prediction from the tree ensemble
  updated_yhat = as.numeric(forest_samples$predict(forest_dataset))

  # Compute the expected residual
  expected_resid = initial_resid + initial_yhat - updated_yhat

  # Get the current residual after running the sampler
  updated_resid = residual$get_data()

  # Assertion
  expect_equal(expected_resid, updated_resid)
})

#' @title Compute Contrast for BCF Model
#' @description
#' Compute a contrast using a BCF model by making two sets of outcome predictions and taking their difference.
#' For simple BCF models with binary treatment, this will yield the same prediction as requesting `terms = "cate"`
#' in the `predict.bcfmodel` function. For more general models, such as models with continuous / multivariate treatments or
#' an additive random effects term with a coefficient on the treatment, this function provides the flexibility to compute a
#' any contrast of interest by specifying covariates, treatment, and random effects bases and IDs for both sides of a two term
#' contrast. For simplicity, we refer to the subtrahend of the contrast as the "control" or `Y0` term and the minuend of the
#' contrast as the `Y1` term, though the requested contrast need not match the "control vs treatment" terminology of a classic
#' two-arm experiment. We mirror the function calls and terminology of the `predict.bcfmodel` function, labeling each prediction
#' data term with a `1` to denote its contribution to the treatment prediction of a contrast and `0` to denote inclusion in the
#' control prediction.
#'
#' @param object Object of type `bcfmodel` containing draws of a Bayesian causal forest model and associated sampling outputs.
#' @param X_0 Covariates used for prediction in the "control" case. Must be a matrix or dataframe.
#' @param X_1 Covariates used for prediction in the "treatment" case. Must be a matrix or dataframe.
#' @param Z_0 Treatments used for prediction in the "control" case. Must be a matrix or vector.
#' @param Z_1 Treatments used for prediction in the "treatment" case. Must be a matrix or vector.
#' @param propensity_0 (Optional) Propensities used for prediction in the "control" case. Must be a matrix or vector.
#' @param propensity_1 (Optional) Propensities used for prediction in the "treatment" case. Must be a matrix or vector.
#' @param rfx_group_ids_0 (Optional) Test set group labels used for prediction from an additive random effects
#' model in the "control" case. We do not currently support (but plan to in the near future), test set evaluation
#' for group labels that were not in the training set. Must be a vector.
#' @param rfx_group_ids_1 (Optional) Test set group labels used for prediction from an additive random effects
#' model in the "treatment" case. We do not currently support (but plan to in the near future), test set evaluation
#' for group labels that were not in the training set. Must be a vector.
#' @param rfx_basis_0 (Optional) Test set basis for used for prediction from an additive random effects model in the "control" case.  Must be a matrix or vector.
#' @param rfx_basis_1 (Optional) Test set basis for used for prediction from an additive random effects model in the "treatment" case. Must be a matrix or vector.
#' @param type (Optional) Aggregation level of the contrast. Options are "mean", which averages the contrast evaluations over every draw of a BCF model, and "posterior", which returns the entire matrix of posterior contrast estimates. Default: "posterior".
#' @param scale (Optional) Scale of the contrast. Options are "linear", which returns a contrast on the original scale of the mean forest / RFX terms, and "probability", which transforms each contrast term into a probability of observing `y == 1` before taking their difference. "probability" is only valid for models fit with a probit outcome model. Default: "linear".
#'
#' @return List of prediction matrices or single prediction matrix / vector, depending on the terms requested.
#' @export
#'
#' @examples
#' n <- 500
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' mu_x <- (
#'     ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
#'     ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
#'     ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
#'     ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
#' )
#' pi_x <- (
#'     ((0 <= X[,1]) & (0.25 > X[,1])) * (0.2) +
#'     ((0.25 <= X[,1]) & (0.5 > X[,1])) * (0.4) +
#'     ((0.5 <= X[,1]) & (0.75 > X[,1])) * (0.6) +
#'     ((0.75 <= X[,1]) & (1 > X[,1])) * (0.8)
#' )
#' tau_x <- (
#'     ((0 <= X[,2]) & (0.25 > X[,2])) * (0.5) +
#'     ((0.25 <= X[,2]) & (0.5 > X[,2])) * (1.0) +
#'     ((0.5 <= X[,2]) & (0.75 > X[,2])) * (1.5) +
#'     ((0.75 <= X[,2]) & (1 > X[,2])) * (2.0)
#' )
#' Z <- rbinom(n, 1, pi_x)
#' noise_sd <- 1
#' y <- mu_x + tau_x*Z + rnorm(n, 0, noise_sd)
#' test_set_pct <- 0.2
#' n_test <- round(test_set_pct*n)
#' n_train <- n - n_test
#' test_inds <- sort(sample(1:n, n_test, replace = FALSE))
#' train_inds <- (1:n)[!((1:n) %in% test_inds)]
#' X_test <- X[test_inds,]
#' X_train <- X[train_inds,]
#' pi_test <- pi_x[test_inds]
#' pi_train <- pi_x[train_inds]
#' Z_test <- Z[test_inds]
#' Z_train <- Z[train_inds]
#' y_test <- y[test_inds]
#' y_train <- y[train_inds]
#' mu_test <- mu_x[test_inds]
#' mu_train <- mu_x[train_inds]
#' tau_test <- tau_x[test_inds]
#' tau_train <- tau_x[train_inds]
#' bcf_model <- bcf(X_train = X_train, Z_train = Z_train, y_train = y_train,
#'                  propensity_train = pi_train, num_gfr = 10,
#'                  num_burnin = 0, num_mcmc = 10)
#' tau_hat_test <- computeContrastBCFModel(
#'     bcf_model, X_0=X_test, X_1=X_test, Z_0=rep(0, n_test), Z_1=rep(1, n_test),
#'     propensity_0 = pi_test, propensity_1 = pi_test
#' )
computeContrastBCFModel <- function(
  object,
  X_0,
  X_1,
  Z_0,
  Z_1,
  propensity_0 = NULL,
  propensity_1 = NULL,
  rfx_group_ids_0 = NULL,
  rfx_group_ids_1 = NULL,
  rfx_basis_0 = NULL,
  rfx_basis_1 = NULL,
  type = "posterior",
  scale = "linear"
) {
  # Handle mean function scale
  if (!is.character(scale)) {
    stop("scale must be a string or character vector")
  }
  if (!(scale %in% c("linear", "probability"))) {
    stop("scale must either be 'linear' or 'probability'")
  }
  is_probit <- object$model_params$outcome_model$link == "probit"
  if ((scale == "probability") && (!is_probit)) {
    stop(
      "scale cannot be 'probability' for models not fit with a probit outcome model"
    )
  }
  probability_scale <- scale == "probability"

  # Handle prediction type
  if (!is.character(type)) {
    stop("type must be a string or character vector")
  }
  if (!(type %in% c("mean", "posterior"))) {
    stop("type must either be 'mean' or 'posterior")
  }
  predict_mean <- type == "mean"

  # Make sure covariates are matrix or data frame
  if ((!is.data.frame(X_0)) && (!is.matrix(X_0))) {
    stop("X_0 must be a matrix or dataframe")
  }
  if ((!is.data.frame(X_1)) && (!is.matrix(X_1))) {
    stop("X_1 must be a matrix or dataframe")
  }

  # Convert all input data to matrices if not already converted
  if ((is.null(dim(Z_0))) && (!is.null(Z_0))) {
    Z_0 <- as.matrix(as.numeric(Z_0))
  }
  if ((is.null(dim(Z_1))) && (!is.null(Z_1))) {
    Z_1 <- as.matrix(as.numeric(Z_1))
  }
  if ((is.null(dim(propensity_0))) && (!is.null(propensity_0))) {
    propensity_0 <- as.matrix(propensity_0)
  }
  if ((is.null(dim(propensity_1))) && (!is.null(propensity_1))) {
    propensity_1 <- as.matrix(propensity_1)
  }
  if ((is.null(dim(rfx_basis_0))) && (!is.null(rfx_basis_0))) {
    rfx_basis_0 <- as.matrix(rfx_basis_0)
  }
  if ((is.null(dim(rfx_basis_1))) && (!is.null(rfx_basis_1))) {
    rfx_basis_1 <- as.matrix(rfx_basis_1)
  }

  # Data checks
  if (
    (object$model_params$propensity_covariate != "none") &&
      ((is.null(propensity_0)) ||
        (is.null(propensity_1)))
  ) {
    if (!object$model_params$internal_propensity_model) {
      stop("propensity_0 and propensity_1 must be provided for this model")
    }
  }
  if (nrow(X_0) != nrow(Z_0)) {
    stop("X_0 and Z_0 must have the same number of rows")
  }
  if (nrow(X_1) != nrow(Z_1)) {
    stop("X_1 and Z_1 must have the same number of rows")
  }
  if (object$model_params$num_covariates != ncol(X_0)) {
    stop(
      "X_0 and must have the same number of columns as the covariates used to train the model"
    )
  }
  if (object$model_params$num_covariates != ncol(X_1)) {
    stop(
      "X_1 and must have the same number of columns as the covariates used to train the model"
    )
  }
  if ((object$model_params$has_rfx) && (is.null(rfx_group_ids_0))) {
    stop(
      "Random effect group labels (rfx_group_ids_0) must be provided for this model"
    )
  }
  if ((object$model_params$has_rfx) && (is.null(rfx_group_ids_1))) {
    stop(
      "Random effect group labels (rfx_group_ids_1) must be provided for this model"
    )
  }
  if ((object$model_params$has_rfx_basis) && (is.null(rfx_basis_0))) {
    stop("Random effects basis (rfx_basis_0) must be provided for this model")
  }
  if ((object$model_params$has_rfx_basis) && (is.null(rfx_basis_1))) {
    stop("Random effects basis (rfx_basis_1) must be provided for this model")
  }
  if (
    (object$model_params$num_rfx_basis > 0) &&
      (ncol(rfx_basis_0) != object$model_params$num_rfx_basis)
  ) {
    stop(
      "Random effects basis has a different dimension than the basis used to train this model"
    )
  }
  if (
    (object$model_params$num_rfx_basis > 0) &&
      (ncol(rfx_basis_1) != object$model_params$num_rfx_basis)
  ) {
    stop(
      "Random effects basis has a different dimension than the basis used to train this model"
    )
  }

  # Predict for the control arm
  control_preds <- predict(
    object = object,
    X = X_0,
    Z = Z_0,
    propensity = propensity_0,
    rfx_group_ids = rfx_group_ids_0,
    rfx_basis = rfx_basis_0,
    type = "posterior",
    term = "y_hat",
    scale = "linear"
  )

  # Predict for the treatment arm
  treatment_preds <- predict(
    object = object,
    X = X_1,
    Z = Z_1,
    propensity = propensity_1,
    rfx_group_ids = rfx_group_ids_1,
    rfx_basis = rfx_basis_1,
    type = "posterior",
    term = "y_hat",
    scale = "linear"
  )

  # Transform to probability scale if requested
  if (probability_scale) {
    treatment_preds <- pnorm(treatment_preds)
    control_preds <- pnorm(control_preds)
  }

  # Compute and return contrast
  if (predict_mean) {
    return(rowMeans(treatment_preds - control_preds))
  } else {
    return(treatment_preds - control_preds)
  }
}

#' @title Compute Contrast for BART Model
#' @description
#' Compute a contrast using a BART model by making two sets of outcome predictions and taking their difference.
#' This function provides the flexibility to compute any contrast of interest by specifying covariates, leaf basis, and random effects
#' bases / IDs for both sides of a two term contrast. For simplicity, we refer to the subtrahend of the contrast as the "control" or
#' `Y0` term and the minuend of the contrast as the `Y1` term, though the requested contrast need not match the "control vs treatment"
#' terminology of a classic two-treatment causal inference problem. We mirror the function calls and terminology of the `predict.bartmodel`
#' function, labeling each prediction data term with a `1` to denote its contribution to the treatment prediction of a contrast and
#' `0` to denote inclusion in the control prediction.
#'
#' Only valid when there is either a mean forest or a random effects term in the BART model.
#'
#' @param object Object of type `bart` containing draws of a regression forest and associated sampling outputs.
#' @param X_0 Covariates used for prediction in the "control" case. Must be a matrix or dataframe.
#' @param X_1 Covariates used for prediction in the "treatment" case. Must be a matrix or dataframe.
#' @param leaf_basis_0 (Optional) Bases used for prediction in the "control" case (by e.g. dot product with leaf values). Default: `NULL`.
#' @param leaf_basis_1 (Optional) Bases used for prediction in the "treatment" case (by e.g. dot product with leaf values). Default: `NULL`.
#' @param rfx_group_ids_0 (Optional) Test set group labels used for prediction from an additive random effects
#' model in the "control" case. We do not currently support (but plan to in the near future), test set evaluation
#' for group labels that were not in the training set. Must be a vector.
#' @param rfx_group_ids_1 (Optional) Test set group labels used for prediction from an additive random effects
#' model in the "treatment" case. We do not currently support (but plan to in the near future), test set evaluation
#' for group labels that were not in the training set. Must be a vector.
#' @param rfx_basis_0 (Optional) Test set basis for used for prediction from an additive random effects model in the "control" case.  Must be a matrix or vector.
#' @param rfx_basis_1 (Optional) Test set basis for used for prediction from an additive random effects model in the "treatment" case. Must be a matrix or vector.
#' @param type (Optional) Aggregation level of the contrast. Options are "mean", which averages the contrast evaluations over every draw of a BART model, and "posterior", which returns the entire matrix of posterior contrast estimates. Default: "posterior".
#' @param scale (Optional) Scale of the contrast. Options are "linear", which returns contrast of predictions on the original scale of the mean forest / RFX terms, and "probability". `scale = "probability"` is only valid for models fit with a probit / cloglog links on binary or ordinal outcomes. For binary outcome models, `scale = "probability"` will return contrasts of the probability that `y == 1`. For ordinal outcome models, `scale = "probability"` will return contrasts over the "survival function" `P(y > k)` for `k = 1, 2, ..., K-1` where `K` is the total number of categories. Default: "linear".
#'
#' @return Contrast matrix or vector, depending on whether type = "mean" or "posterior".
#' @export
#'
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' W <- matrix(runif(n*1), ncol = 1)
#' f_XW <- (
#'     ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5*W[,1]) +
#'     ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5*W[,1]) +
#'     ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5*W[,1]) +
#'     ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5*W[,1])
#' )
#' noise_sd <- 1
#' y <- f_XW + rnorm(n, 0, noise_sd)
#' test_set_pct <- 0.2
#' n_test <- round(test_set_pct*n)
#' n_train <- n - n_test
#' test_inds <- sort(sample(1:n, n_test, replace = FALSE))
#' train_inds <- (1:n)[!((1:n) %in% test_inds)]
#' X_test <- X[test_inds,]
#' X_train <- X[train_inds,]
#' W_test <- W[test_inds,]
#' W_train <- W[train_inds,]
#' y_test <- y[test_inds]
#' y_train <- y[train_inds]
#' bart_model <- bart(X_train = X_train, leaf_basis_train = W_train, y_train = y_train,
#'                    num_gfr = 10, num_burnin = 0, num_mcmc = 10)
#' contrast_test <- computeContrastBARTModel(
#'     bart_model,
#'     X_0 = X_test,
#'     X_1 = X_test,
#'     leaf_basis_0 = matrix(0, nrow = n_test, ncol = 1),
#'     leaf_basis_1 = matrix(1, nrow = n_test, ncol = 1),
#'     type = "posterior",
#'     scale = "linear"
#' )
computeContrastBARTModel <- function(
  object,
  X_0,
  X_1,
  leaf_basis_0 = NULL,
  leaf_basis_1 = NULL,
  rfx_group_ids_0 = NULL,
  rfx_group_ids_1 = NULL,
  rfx_basis_0 = NULL,
  rfx_basis_1 = NULL,
  type = "posterior",
  scale = "linear"
) {
  # Handle mean function scale
  if (!is.character(scale)) {
    stop("scale must be a string or character vector")
  }
  if (!(scale %in% c("linear", "probability"))) {
    stop("scale must either be 'linear' or 'probability'")
  }
  is_probit <- object$model_params$outcome_model$link == "probit"
  is_cloglog <- object$model_params$outcome_model$link == "cloglog"
  is_binary_cloglog <- is_cloglog &&
    (object$model_params$outcome_model$outcome == "binary")
  is_ordinal_cloglog <- is_cloglog &&
    (object$model_params$outcome_model$outcome == "ordinal")
  if ((scale == "probability") && (!(is_probit || is_cloglog))) {
    stop(
      "scale cannot be 'probability' for models not fit with a probit or cloglog outcome model"
    )
  }
  probability_scale <- scale == "probability"

  # Handle prediction type
  if (!is.character(type)) {
    stop("type must be a string or character vector")
  }
  if (!(type %in% c("mean", "posterior"))) {
    stop("type must either be 'mean' or 'posterior'")
  }
  predict_mean <- type == "mean"

  # Handle prediction terms
  has_mean_forest <- object$model_params$include_mean_forest
  has_rfx <- object$model_params$has_rfx
  if ((!has_mean_forest) && (!has_rfx)) {
    stop(
      "Model must have either or both of mean forest or random effects terms to compute the requested contrast."
    )
  }

  # Check that covariates are matrix or data frame
  if ((!is.data.frame(X_0)) && (!is.matrix(X_0))) {
    stop("X_0 must be a matrix or dataframe")
  }
  if ((!is.data.frame(X_1)) && (!is.matrix(X_1))) {
    stop("X_1 must be a matrix or dataframe")
  }

  # Convert all input data to matrices if not already converted
  if ((is.null(dim(leaf_basis_0))) && (!is.null(leaf_basis_0))) {
    leaf_basis_0 <- as.matrix(leaf_basis_0)
  }
  if ((is.null(dim(leaf_basis_1))) && (!is.null(leaf_basis_1))) {
    leaf_basis_1 <- as.matrix(leaf_basis_1)
  }
  if ((is.null(dim(rfx_basis_0))) && (!is.null(rfx_basis_0))) {
    rfx_basis_0 <- as.matrix(rfx_basis_0)
  }
  if ((is.null(dim(rfx_basis_1))) && (!is.null(rfx_basis_1))) {
    rfx_basis_1 <- as.matrix(rfx_basis_1)
  }

  # Data checks
  if (
    (object$model_params$requires_basis) &&
      (is.null(leaf_basis_0) || is.null(leaf_basis_1))
  ) {
    stop("leaf_basis_0 and leaf_basis_1 must be provided for this model")
  }
  if ((!is.null(leaf_basis_0)) && (nrow(X_0) != nrow(leaf_basis_0))) {
    stop("X_0 and leaf_basis_0 must have the same number of rows")
  }
  if ((!is.null(leaf_basis_1)) && (nrow(X_1) != nrow(leaf_basis_1))) {
    stop("X_1 and leaf_basis_1 must have the same number of rows")
  }
  if (object$model_params$num_covariates != ncol(X_0)) {
    stop(
      "X_0 must contain the same number of columns as the BART model's training dataset"
    )
  }
  if (object$model_params$num_covariates != ncol(X_1)) {
    stop(
      "X_1 must contain the same number of columns as the BART model's training dataset"
    )
  }
  if ((has_rfx) && (is.null(rfx_group_ids_0) || is.null(rfx_group_ids_1))) {
    stop(
      "rfx_group_ids_0 and rfx_group_ids_1 must be provided for this model"
    )
  }
  if (has_rfx) {
    if (object$model_params$rfx_model_spec == "custom") {
      if ((is.null(rfx_basis_0) || is.null(rfx_basis_1))) {
        stop(
          "A user-provided basis (`rfx_basis_0` and `rfx_basis_1`) must be provided when the model was sampled with a random effects model spec set to 'custom'"
        )
      }
      if (!is.matrix(rfx_basis_0) || !is.matrix(rfx_basis_1)) {
        stop("'rfx_basis_0' and 'rfx_basis_1' must be matrices")
      }
      if (
        (nrow(rfx_basis_0) != nrow(X_0)) || (nrow(rfx_basis_1) != nrow(X_1))
      ) {
        stop(
          "'rfx_basis_0' and 'rfx_basis_1' must have the same number of rows as 'X_0' and 'X_1'"
        )
      }
      if (
        (object$model_params$num_rfx_basis > 0) &&
          ((ncol(rfx_basis_0) != object$model_params$num_rfx_basis) ||
            (ncol(rfx_basis_1) != object$model_params$num_rfx_basis))
      ) {
        stop(
          "rfx_basis_0 and / or rfx_basis_1 have a different dimension than the basis used to train this model"
        )
      }
    }
  }

  # Predict for the control arm
  control_preds <- predict(
    object = object,
    X = X_0,
    leaf_basis = leaf_basis_0,
    rfx_group_ids = rfx_group_ids_0,
    rfx_basis = rfx_basis_0,
    type = "posterior",
    term = "y_hat",
    scale = scale
  )

  # Predict for the treatment arm
  treatment_preds <- predict(
    object = object,
    X = X_1,
    leaf_basis = leaf_basis_1,
    rfx_group_ids = rfx_group_ids_1,
    rfx_basis = rfx_basis_1,
    type = "posterior",
    term = "y_hat",
    scale = scale
  )

  # Convert ordinal class probabilities to "survival" probabilities
  if (is_ordinal_cloglog) {
    if (probability_scale) {
      num_categories <- object$model_params$cloglog_num_categories
      control_preds <- class_probs_to_survival_probs(
        control_preds,
        num_categories
      )
      treatment_preds <- class_probs_to_survival_probs(
        treatment_preds,
        num_categories
      )
    }
  }

  # Compute and return contrast
  if (predict_mean) {
    output_dim <- dim(treatment_preds)
    if (length(output_dim) == 3) {
      treatment_collapsed <- apply(treatment_preds, c(1, 2), mean)
      control_collapsed <- apply(control_preds, c(1, 2), mean)
      return(treatment_collapsed - control_collapsed)
    } else if (length(output_dim) == 2) {
      return(rowMeans(treatment_preds - control_preds))
    } else {
      stop("Unexpected output dimension")
    }
  } else {
    return(treatment_preds - control_preds)
  }
}

#' @title Sample BCF Posterior Predictive
#' @description
#' Sample from the posterior predictive distribution for outcomes modeled by BCF
#'
#' @param model_object A fitted BCF model object of class `bcfmodel`.
#' @param X A matrix or data frame of covariates.
#' @param Z A vector or matrix of treatment assignments.
#' @param propensity (Optional) A vector or matrix of propensity scores. Required if the underlying model depends on user-provided propensities.
#' @param rfx_group_ids (Optional) A vector of group IDs for random effects model. Required if the BCF model includes random effects.
#' @param rfx_basis (Optional) A matrix of bases for random effects model. Required if the BCF model includes random effects.
#' @param num_draws_per_sample (Optional) The number of samples to draw from the likelihood for each draw of the posterior. Defaults to a heuristic based on the number of samples in a BCF model (i.e. if the BCF model has >1000 draws, we use 1 draw from the likelihood per sample, otherwise we upsample to ensure at least 1000 posterior predictive draws).
#'
#' @returns Array of posterior predictive samples with dimensions (num_observations, num_posterior_samples, num_draws_per_sample) if num_draws_per_sample > 1, otherwise (num_observations, num_posterior_samples).
#'
#' @export
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(rnorm(n * p), nrow = n, ncol = p)
#' pi_X <- pnorm(X[,1] / 2)
#' Z <- rbinom(n, 1, pi_X)
#' y <- 2 * X[,2] + 0.5 * X[,2] * Z + rnorm(n)
#' bcf_model <- bcf(X_train = X, Z_train = Z, y_train = y, propensity_train = pi_X)
#' ppd_samples <- sampleBCFPosteriorPredictive(
#'   model_object = bcf_model, X = X,
#'   Z = Z, propensity = pi_X
#' )
sampleBCFPosteriorPredictive <- function(
  model_object,
  X = NULL,
  Z = NULL,
  propensity = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL,
  num_draws_per_sample = NULL
) {
  # Check the provided model object
  check_model_is_valid(model_object)

  # Determine whether the outcome is continuous (Gaussian) or binary (probit-link)
  is_probit <- model_object$model_params$outcome_model$link == "probit"

  # Check that all the necessary inputs were provided for interval computation
  needs_covariates <- TRUE
  if (needs_covariates) {
    if (is.null(X)) {
      stop(
        "'X' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(X) && !is.data.frame(X)) {
      stop("'X' must be a matrix or data frame")
    }
  }
  needs_treatment <- needs_covariates
  if (needs_treatment) {
    if (is.null(Z)) {
      stop(
        "'Z' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(Z) && !is.numeric(Z)) {
      stop("'Z' must be a numeric vector or matrix")
    }
    if (is.matrix(Z)) {
      if (nrow(Z) != nrow(X)) {
        stop("'Z' must have the same number of rows as 'X'")
      }
    } else {
      if (length(Z) != nrow(X)) {
        stop(
          "'Z' must have the same number of elements as 'X'"
        )
      }
    }
  }
  uses_propensity <- model_object$model_params$propensity_covariate != "none"
  internal_propensity_model <- model_object$model_params$internal_propensity_model
  needs_propensity <- (needs_covariates &&
    uses_propensity &&
    (!internal_propensity_model))
  if (needs_propensity) {
    if (is.null(propensity)) {
      stop(
        "'propensity' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(propensity) && !is.numeric(propensity)) {
      stop("'propensity' must be a numeric vector or matrix")
    }
    if (is.matrix(propensity)) {
      if (nrow(propensity) != nrow(X)) {
        stop("'propensity' must have the same number of rows as 'X'")
      }
    } else {
      if (length(propensity) != nrow(X)) {
        stop(
          "'propensity' must have the same number of elements as 'X'"
        )
      }
    }
  }
  needs_rfx_data <- model_object$model_params$has_rfx
  if (needs_rfx_data) {
    if (is.null(rfx_group_ids)) {
      stop(
        "'rfx_group_ids' must be provided in order to compute the requested intervals"
      )
    }
    if (length(rfx_group_ids) != nrow(X)) {
      stop(
        "'rfx_group_ids' must have the same length as the number of rows in 'X'"
      )
    }

    if (model_object$model_params$rfx_model_spec == "custom") {
      if (is.null(rfx_basis)) {
        stop(
          "A user-provided basis (`rfx_basis`) must be provided when the model was sampled with a random effects model spec set to 'custom'"
        )
      }
    }

    if (!is.null(rfx_basis)) {
      if (!is.matrix(rfx_basis)) {
        stop("'rfx_basis' must be a matrix")
      }
      if (nrow(rfx_basis) != nrow(X)) {
        stop("'rfx_basis' must have the same number of rows as 'X'")
      }
    }
  }

  # Compute posterior samples
  bcf_preds <- predict(
    model_object,
    X = X,
    Z = Z,
    propensity = propensity,
    rfx_group_ids = rfx_group_ids,
    rfx_basis = rfx_basis,
    type = "posterior",
    terms = c("all"),
    scale = "linear"
  )

  # Compute outcome mean and variance for every posterior draw
  has_rfx <- model_object$model_params$has_rfx
  has_variance_forest <- model_object$model_params$include_variance_forest
  samples_global_variance <- model_object$model_params$sample_sigma2_global
  num_posterior_draws <- model_object$model_params$num_samples
  num_observations <- nrow(X)
  ppd_mean <- bcf_preds$y_hat
  if (has_variance_forest) {
    ppd_variance <- bcf_preds$variance_forest_predictions
  } else {
    if (samples_global_variance) {
      ppd_variance <- matrix(
        rep(
          model_object$sigma2_global_samples,
          each = num_observations
        ),
        nrow = num_observations
      )
    } else {
      ppd_variance <- model_object$model_params$initial_sigma2
    }
  }

  # Sample from the posterior predictive distribution
  if (is.null(num_draws_per_sample)) {
    ppd_draw_multiplier <- posterior_predictive_heuristic_multiplier(
      num_posterior_draws,
      num_observations
    )
  } else {
    ppd_draw_multiplier <- num_draws_per_sample
  }
  num_ppd_draws <- ppd_draw_multiplier * num_posterior_draws * num_observations
  ppd_vector <- rnorm(num_ppd_draws, ppd_mean, sqrt(ppd_variance))

  # Reshape data
  if (ppd_draw_multiplier > 1) {
    ppd_array <- array(
      ppd_vector,
      dim = c(num_observations, num_posterior_draws, ppd_draw_multiplier)
    )
  } else {
    ppd_array <- array(
      ppd_vector,
      dim = c(num_observations, num_posterior_draws)
    )
  }

  # Binarize outcomes for probit models
  if (is_probit) {
    ppd_array <- (ppd_array > 0.0) * 1
  }

  return(ppd_array)
}

#' @title Sample BART Posterior Predictive
#' @description
#' Sample from the posterior predictive distribution for outcomes modeled by BART
#'
#' @param model_object A fitted BART model object of class `bartmodel`.
#' @param X A matrix or data frame of covariates. Required if the BART model depends on covariates (e.g., contains a mean or variance forest).
#' @param leaf_basis A matrix of bases for mean forest models with regression defined in the leaves. Required for "leaf regression" models.
#' @param rfx_group_ids A vector of group IDs for random effects model. Required if the BART model includes random effects.
#' @param rfx_basis A matrix of bases for random effects model. Required if the BART model includes random effects.
#' @param num_draws_per_sample The number of posterior predictive samples to draw for each posterior sample. Defaults to a heuristic based on the number of samples in a BART model (i.e. if the BART model has >1000 draws, we use 1 draw from the likelihood per sample, otherwise we upsample to ensure intervals are based on at least 1000 posterior predictive draws).
#'
#' @returns Array of posterior predictive samples with dimensions (num_observations, num_posterior_samples, num_draws_per_sample) if num_draws_per_sample > 1, otherwise (num_observations, num_posterior_samples).
#'
#' @export
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(rnorm(n * p), nrow = n, ncol = p)
#' y <- 2 * X[,1] + rnorm(n)
#' bart_model <- bart(y_train = y, X_train = X)
#' ppd_samples <- sampleBARTPosteriorPredictive(
#'   model_object = bart_model, X = X
#' )
sampleBARTPosteriorPredictive <- function(
  model_object,
  X = NULL,
  leaf_basis = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL,
  num_draws_per_sample = NULL
) {
  # Check the provided model object
  check_model_is_valid(model_object)

  # Determine whether the outcome is continuous (Gaussian) or binary (probit-link)
  is_gaussian <- model_object$model_params$outcome_model$link == "identity"
  is_probit <- model_object$model_params$outcome_model$link == "probit"
  is_cloglog <- model_object$model_params$outcome_model$link == "cloglog"
  is_binary_cloglog <- is_cloglog &&
    model_object$model_params$outcome_model$outcome == "binary"
  is_ordinal_cloglog <- is_cloglog &&
    model_object$model_params$outcome_model$outcome == "ordinal"

  # Check that all the necessary inputs were provided for interval computation
  needs_covariates <- model_object$model_params$include_mean_forest
  if (needs_covariates) {
    if (is.null(X)) {
      stop(
        "'X' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(X) && !is.data.frame(X)) {
      stop("'X' must be a matrix or data frame")
    }
  }
  needs_basis <- needs_covariates && model_object$model_params$has_basis
  if (needs_basis) {
    if (is.null(leaf_basis)) {
      stop(
        "'leaf_basis' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(leaf_basis)) {
      stop("'leaf_basis' must be a matrix")
    }
    if (is.matrix(leaf_basis)) {
      if (nrow(leaf_basis) != nrow(X)) {
        stop("'leaf_basis' must have the same number of rows as 'X'")
      }
    } else {
      if (length(leaf_basis) != nrow(X)) {
        stop("'leaf_basis' must have the same number of elements as 'X'")
      }
    }
  }
  needs_rfx_data <- model_object$model_params$has_rfx
  if (needs_rfx_data) {
    if (is.null(rfx_group_ids)) {
      stop(
        "'rfx_group_ids' must be provided in order to compute the requested intervals"
      )
    }
    if (length(rfx_group_ids) != nrow(X)) {
      stop(
        "'rfx_group_ids' must have the same length as the number of rows in 'X'"
      )
    }
    if (model_object$model_params$rfx_model_spec == "custom") {
      if (is.null(rfx_basis)) {
        stop(
          "A user-provided basis (`rfx_basis`) must be provided when the model was sampled with a random effects model spec set to 'custom'"
        )
      }
      if (!is.matrix(rfx_basis)) {
        stop("'rfx_basis' must be a matrix")
      }
      if (nrow(rfx_basis) != nrow(X)) {
        stop("'rfx_basis' must have the same number of rows as 'X'")
      }
    }
  }

  if (is_gaussian) {
    # Compute posterior samples
    bart_preds <- predict(
      model_object,
      X = X,
      leaf_basis = leaf_basis,
      rfx_group_ids = rfx_group_ids,
      rfx_basis = rfx_basis,
      type = "posterior",
      terms = c("all"),
      scale = "linear"
    )

    # Compute outcome mean and variance for every posterior draw
    has_mean_term <- (model_object$model_params$include_mean_forest ||
      model_object$model_params$has_rfx)
    has_variance_forest <- model_object$model_params$include_variance_forest
    samples_global_variance <- model_object$model_params$sample_sigma2_global
    num_posterior_draws <- model_object$model_params$num_samples
    num_observations <- nrow(X)
    if (has_mean_term) {
      ppd_mean <- bart_preds$y_hat
    } else {
      ppd_mean <- 0
    }
    if (has_variance_forest) {
      ppd_variance <- bart_preds$variance_forest_predictions
    } else {
      if (samples_global_variance) {
        ppd_variance <- matrix(
          rep(
            model_object$sigma2_global_samples,
            each = num_observations
          ),
          nrow = num_observations
        )
      } else {
        ppd_variance <- model_object$model_params$sigma2_init
      }
    }

    # Sample from the posterior predictive distribution
    if (is.null(num_draws_per_sample)) {
      ppd_draw_multiplier <- posterior_predictive_heuristic_multiplier(
        num_posterior_draws,
        num_observations
      )
    } else {
      ppd_draw_multiplier <- num_draws_per_sample
    }
    num_ppd_draws <- ppd_draw_multiplier *
      num_posterior_draws *
      num_observations
    ppd_vector <- rnorm(num_ppd_draws, ppd_mean, sqrt(ppd_variance))

    # Reshape data
    if (ppd_draw_multiplier > 1) {
      ppd_array <- array(
        ppd_vector,
        dim = c(num_observations, num_posterior_draws, ppd_draw_multiplier)
      )
    } else {
      ppd_array <- array(
        ppd_vector,
        dim = c(num_observations, num_posterior_draws)
      )
    }
  } else if (is_probit || is_binary_cloglog) {
    # Compute posterior probability samples
    bart_preds <- predict(
      model_object,
      X = X,
      leaf_basis = leaf_basis,
      rfx_group_ids = rfx_group_ids,
      rfx_basis = rfx_basis,
      type = "posterior",
      terms = "y_hat",
      scale = "probability"
    )

    # Sample from the posterior predictive distribution
    num_posterior_draws <- model_object$model_params$num_samples
    num_observations <- nrow(X)
    if (is.null(num_draws_per_sample)) {
      ppd_draw_multiplier <- posterior_predictive_heuristic_multiplier(
        num_posterior_draws,
        num_observations
      )
    } else {
      ppd_draw_multiplier <- num_draws_per_sample
    }
    num_ppd_draws <- ppd_draw_multiplier *
      num_posterior_draws *
      num_observations
    ppd_vector <- rbinom(num_ppd_draws, size = 1, prob = bart_preds)

    # Reshape data
    if (ppd_draw_multiplier > 1) {
      ppd_array <- array(
        ppd_vector,
        dim = c(num_observations, num_posterior_draws, ppd_draw_multiplier)
      )
    } else {
      ppd_array <- array(
        ppd_vector,
        dim = c(num_observations, num_posterior_draws)
      )
    }
  } else if (is_ordinal_cloglog) {
    # Compute posterior probability samples
    bart_preds <- predict(
      model_object,
      X = X,
      leaf_basis = leaf_basis,
      rfx_group_ids = rfx_group_ids,
      rfx_basis = rfx_basis,
      type = "posterior",
      terms = "y_hat",
      scale = "probability"
    )

    # Sample from the posterior predictive distribution
    num_categories <- model_object$model_params$cloglog_num_categories
    num_posterior_draws <- model_object$model_params$num_samples
    num_observations <- nrow(X)
    if (is.null(num_draws_per_sample)) {
      ppd_draw_multiplier <- posterior_predictive_heuristic_multiplier(
        num_posterior_draws,
        num_observations
      )
    } else {
      ppd_draw_multiplier <- num_draws_per_sample
    }
    ppd_vector <- apply(bart_preds, c(1, 3), function(x) {
      sample(1:num_categories, ppd_draw_multiplier, replace = T, prob = x)
    })

    # Reshape data
    if (ppd_draw_multiplier > 1) {
      # apply returns (ppd_draw_multiplier x num_observations x num_posterior_draws)
      ppd_array <- aperm(ppd_vector, c(2, 3, 1))
    } else {
      # apply returns (num_observations x num_posterior_draws) matrix
      ppd_array <- ppd_vector
    }
  }

  return(ppd_array)
}

posterior_predictive_heuristic_multiplier <- function(
  num_samples,
  num_observations
) {
  if (num_samples >= 1000) {
    return(1)
  } else {
    return(ceiling(1000 / num_samples))
  }
}

#' @title Compute BCF Posterior Credible Intervals
#' @description
#' Compute posterior credible intervals for specified terms from a fitted BCF model. Supports intervals for prognostic forests, CATE forests, variance forests, random effects, and overall mean outcome predictions.
#'
#' @param model_object A fitted BCF model object of class `bcfmodel`.
#' @param terms A character string specifying the model term(s) for which to compute intervals. Options for BCF models are `"prognostic_function"`, `"mu"`, `"cate"`, `"tau"`, `"variance_forest"`, `"rfx"`, or `"y_hat"`. Note that `"mu"` is only different from `"prognostic_function"` if random effects are included with a model spec of `"intercept_only"` or `"intercept_plus_treatment"` and `"tau"` is only different from `"cate"` if random effects are included with a model spec of `"intercept_plus_treatment"`.
#' @param level A numeric value between 0 and 1 specifying the credible interval level (default is 0.95 for a 95% credible interval).
#' @param scale (Optional) Scale of mean function predictions. Options are "linear", which returns predictions on the original scale of the mean forest / RFX terms, and "probability", which transforms predictions into a probability of observing `y == 1`. "probability" is only valid for models fit with a probit outcome model. Default: "linear".
#' @param X (Optional) A matrix or data frame of covariates at which to compute the intervals. Required if the requested term depends on covariates (e.g., prognostic forest, CATE forest, variance forest, or overall predictions).
#' @param Z (Optional) A vector or matrix of treatment assignments. Required if the requested term is `"y_hat"` (overall predictions).
#' @param propensity (Optional) A vector or matrix of propensity scores. Required if the underlying model depends on user-provided propensities.
#' @param rfx_group_ids An optional vector of group IDs for random effects. Required if the requested term includes random effects.
#' @param rfx_basis An optional matrix of basis function evaluations for random effects. Required if the requested term includes random effects.
#'
#' @returns A list containing the lower and upper bounds of the credible interval for the specified term. If multiple terms are requested, a named list with intervals for each term is returned.
#'
#' @export
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(rnorm(n * p), nrow = n, ncol = p)
#' pi_X <- pnorm(0.5 * X[,1])
#' Z <- rbinom(n, 1, pi_X)
#' mu_X <- X[,1]
#' tau_X <- 0.25 * X[,2]
#' y <- mu_X + tau_X * Z + rnorm(n)
#' bcf_model <- bcf(X_train = X, Z_train = Z, y_train = y,
#'                  propensity_train = pi_X)
#' intervals <- computeBCFPosteriorInterval(
#'  model_object = bcf_model,
#'  terms = c("prognostic_function", "cate"),
#'  X = X,
#'  Z = Z,
#'  propensity = pi_X,
#'  level = 0.90
#' )
computeBCFPosteriorInterval <- function(
  model_object,
  terms,
  level = 0.95,
  scale = "linear",
  X = NULL,
  Z = NULL,
  propensity = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL
) {
  # Check the provided model object and requested term
  check_model_is_valid(model_object)
  for (term in terms) {
    check_model_has_term(model_object, term)
  }

  # Handle mean function scale
  if (!is.character(scale)) {
    stop("scale must be a string or character vector")
  }
  if (!(scale %in% c("linear", "probability"))) {
    stop("scale must either be 'linear' or 'probability'")
  }
  is_probit <- model_object$model_params$outcome_model$link == "probit"
  if ((scale == "probability") && (!is_probit)) {
    stop(
      "scale cannot be 'probability' for models not fit with a probit outcome model"
    )
  }

  # Check that all the necessary inputs were provided for interval computation
  for (term in terms) {
    if (
      !(term %in%
        c(
          "prognostic_function",
          "mu",
          "cate",
          "tau",
          "variance_forest",
          "rfx",
          "y_hat",
          "all"
        ))
    ) {
      stop(
        paste0(
          "Term '",
          term,
          "' was requested. Valid terms are 'prognostic_function', 'mu', 'cate', 'tau', 'variance_forest', 'rfx', 'y_hat', and 'all'."
        )
      )
    }
  }
  predict_terms <- terms
  needs_covariates_intermediate <- ((("y_hat" %in% predict_terms) ||
    ("all" %in% predict_terms)))
  needs_covariates <- (("prognostic_function" %in% predict_terms) ||
    ("cate" %in% predict_terms) ||
    ("variance_forest" %in% predict_terms) ||
    (needs_covariates_intermediate))
  if (needs_covariates) {
    if (is.null(X)) {
      stop(
        "'X' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(X) && !is.data.frame(X)) {
      stop("'X' must be a matrix or data frame")
    }
  }
  needs_treatment <- needs_covariates
  if (needs_treatment) {
    if (is.null(Z)) {
      stop(
        "'Z' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(Z) && !is.numeric(Z)) {
      stop("'Z' must be a numeric vector or matrix")
    }
    if (is.matrix(Z)) {
      if (nrow(Z) != nrow(X)) {
        stop("'Z' must have the same number of rows as 'X'")
      }
    } else {
      if (length(Z) != nrow(X)) {
        stop(
          "'Z' must have the same number of elements as 'X'"
        )
      }
    }
  }
  uses_propensity <- model_object$model_params$propensity_covariate != "none"
  internal_propensity_model <- model_object$model_params$internal_propensity_model
  needs_propensity <- (needs_covariates &&
    uses_propensity &&
    (!internal_propensity_model))
  if (needs_propensity) {
    if (is.null(propensity)) {
      stop(
        "'propensity' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(propensity) && !is.numeric(propensity)) {
      stop("'propensity' must be a numeric vector or matrix")
    }
    if (is.matrix(propensity)) {
      if (nrow(propensity) != nrow(X)) {
        stop("'propensity' must have the same number of rows as 'X'")
      }
    } else {
      if (length(propensity) != nrow(X)) {
        stop(
          "'propensity' must have the same number of elements as 'X'"
        )
      }
    }
  }
  needs_rfx_data_intermediate <- ((("y_hat" %in% predict_terms) ||
    ("all" %in% predict_terms)) &&
    model_object$model_params$has_rfx)
  needs_rfx_data <- (("rfx" %in% predict_terms) ||
    (needs_rfx_data_intermediate))
  if (needs_rfx_data) {
    if (is.null(rfx_group_ids)) {
      stop(
        "'rfx_group_ids' must be provided in order to compute the requested intervals"
      )
    }
    if (length(rfx_group_ids) != nrow(X)) {
      stop(
        "'rfx_group_ids' must have the same length as the number of rows in 'X'"
      )
    }

    if (model_object$model_params$rfx_model_spec == "custom") {
      if (is.null(rfx_basis)) {
        stop(
          "A user-provided basis (`rfx_basis`) must be provided when the model was sampled with a random effects model spec set to 'custom'"
        )
      }
    }

    if (!is.null(rfx_basis)) {
      if (!is.matrix(rfx_basis)) {
        stop("'rfx_basis' must be a matrix")
      }
      if (nrow(rfx_basis) != nrow(X)) {
        stop("'rfx_basis' must have the same number of rows as 'X'")
      }
    }
  }

  result <- list()

  # Compute posterior matrices for predict-able terms (if any)
  if (length(predict_terms) > 0) {
    predictions <- predict(
      model_object,
      X = X,
      Z = Z,
      propensity = propensity,
      rfx_group_ids = rfx_group_ids,
      rfx_basis = rfx_basis,
      type = "posterior",
      terms = predict_terms,
      scale = scale
    )
    if (is.list(predictions)) {
      for (term_name in names(predictions)) {
        if (!is.null(predictions[[term_name]])) {
          result[[term_name]] <- summarize_interval(
            predictions[[term_name]],
            sample_dim = 2,
            level = level
          )
        } else {
          result[[term_name]] <- NULL
        }
      }
    } else {
      result[[predict_terms]] <- summarize_interval(
        predictions,
        sample_dim = 2,
        level = level
      )
    }
  }

  # Return single interval directly if only one term was requested
  if (length(terms) == 1) {
    return(result[[terms]])
  }
  return(result)
}

#' @title Compute BART Posterior Credible Intervals
#' @description
#' Compute posterior credible intervals for specified terms from a fitted BART model. Supports intervals for mean functions, variance functions, random effects, and overall outcome predictions.
#'
#' @param model_object A fitted BART or BCF model object of class `bartmodel`.
#' @param terms A character string specifying the model term(s) for which to compute intervals. Options for BART models are `"mean_forest"`, `"variance_forest"`, `"rfx"`, or `"y_hat"`.
#' @param level A numeric value between 0 and 1 specifying the credible interval level (default is 0.95 for a 95% credible interval).
#' @param scale (Optional) Scale of mean function predictions. Options are "linear", which returns predictions on the original scale of the mean forest / RFX terms, and "probability". `scale = "probability"` is only valid for models fit with a probit / cloglog links on binary or ordinal outcomes. For binary outcome models, `scale = "probability"` will return an interval over the probability that `y == 1`. For ordinal outcome models, `scale = "probability"` will return intervals over the "survival function" `P(y > k)` for `k = 1, 2, ..., K-1` where `K` is the total number of categories. Default: "linear".
#' @param X A matrix or data frame of covariates at which to compute the intervals. Required if the requested term depends on covariates (e.g., mean forest, variance forest, or overall predictions).
#' @param leaf_basis An optional matrix of basis function evaluations for mean forest models with regression defined in the leaves. Required for "leaf regression" models.
#' @param rfx_group_ids An optional vector of group IDs for random effects. Required if the requested term includes random effects.
#' @param rfx_basis An optional matrix of basis function evaluations for random effects. Required if the requested term includes random effects.
#'
#' @returns A list containing the lower and upper bounds of the credible interval for the specified term. If multiple terms are requested, a named list with intervals for each term is returned.
#'
#' @export
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(rnorm(n * p), nrow = n, ncol = p)
#' y <- 2 * X[,1] + rnorm(n)
#' bart_model <- bart(y_train = y, X_train = X)
#' intervals <- computeBARTPosteriorInterval(
#'  model_object = bart_model,
#'  terms = c("mean_forest", "y_hat"),
#'  X = X,
#'  level = 0.90
#' )
#' @export
computeBARTPosteriorInterval <- function(
  model_object,
  terms,
  level = 0.95,
  scale = "linear",
  X = NULL,
  leaf_basis = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL
) {
  # Check the provided model object and requested term
  check_model_is_valid(model_object)
  for (term in terms) {
    check_model_has_term(model_object, term)
  }

  # Handle mean function scale
  if (!is.character(scale)) {
    stop("scale must be a string or character vector")
  }
  if (!(scale %in% c("linear", "probability"))) {
    stop("scale must either be 'linear' or 'probability'")
  }
  is_probit <- model_object$model_params$outcome_model$link == "probit"
  is_cloglog <- model_object$model_params$outcome_model$link == "cloglog"
  is_binary_cloglog <- is_cloglog &&
    (model_object$model_params$outcome_model$outcome == "binary")
  is_ordinal_cloglog <- is_cloglog &&
    (model_object$model_params$outcome_model$outcome == "ordinal")
  if ((scale == "probability") && (!(is_probit || is_cloglog))) {
    stop(
      "scale cannot be 'probability' for models not fit with a probit or cloglog link for binary / ordinal outcome data"
    )
  }

  # Check that all the necessary inputs were provided for interval computation
  needs_covariates_intermediate <- ((("y_hat" %in% terms) ||
    ("all" %in% terms)) &&
    model_object$model_params$include_mean_forest)
  needs_covariates <- (("mean_forest" %in% terms) ||
    ("variance_forest" %in% terms) ||
    (needs_covariates_intermediate))
  if (needs_covariates) {
    if (is.null(X)) {
      stop(
        "'X' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(X) && !is.data.frame(X)) {
      stop("'X' must be a matrix or data frame")
    }
  }
  needs_basis <- needs_covariates && model_object$model_params$has_basis
  if (needs_basis) {
    if (is.null(leaf_basis)) {
      stop(
        "'leaf_basis' must be provided in order to compute the requested intervals"
      )
    }
    if (!is.matrix(leaf_basis)) {
      stop("'leaf_basis' must be a matrix")
    }
    if (is.matrix(leaf_basis)) {
      if (nrow(leaf_basis) != nrow(X)) {
        stop("'leaf_basis' must have the same number of rows as 'X'")
      }
    } else {
      if (length(leaf_basis) != nrow(X)) {
        stop("'leaf_basis' must have the same number of elements as 'X'")
      }
    }
  }
  needs_rfx_data_intermediate <- ((("y_hat" %in% terms) ||
    ("all" %in% terms)) &&
    model_object$model_params$has_rfx)
  needs_rfx_data <- (("rfx" %in% terms) ||
    (needs_rfx_data_intermediate))
  if (needs_rfx_data) {
    if (is.null(rfx_group_ids)) {
      stop(
        "'rfx_group_ids' must be provided in order to compute the requested intervals"
      )
    }
    if (length(rfx_group_ids) != nrow(X)) {
      stop(
        "'rfx_group_ids' must have the same length as the number of rows in 'X'"
      )
    }
    if (model_object$model_params$rfx_model_spec == "custom") {
      if (is.null(rfx_basis)) {
        stop(
          "A user-provided basis (`rfx_basis`) must be provided when the model was sampled with a random effects model spec set to 'custom'"
        )
      }
      if (!is.matrix(rfx_basis)) {
        stop("'rfx_basis' must be a matrix")
      }
      if (nrow(rfx_basis) != nrow(X)) {
        stop("'rfx_basis' must have the same number of rows as 'X'")
      }
    }
  }

  # Extract number of samples in the model
  num_samples <- model_object$model_params$num_samples

  # Compute posterior matrices for the requested model terms
  predictions <- predict(
    model_object,
    X = X,
    leaf_basis = leaf_basis,
    rfx_group_ids = rfx_group_ids,
    rfx_basis = rfx_basis,
    type = "posterior",
    terms = terms,
    scale = scale
  )
  has_multiple_terms <- ifelse(is.list(predictions), TRUE, FALSE)

  # Convert ordinal class probabilities to "survival" probabilities
  if (is_ordinal_cloglog) {
    if (scale == "probability") {
      num_categories <- model_object$model_params$cloglog_num_categories
      if (has_multiple_terms) {
        for (term_name in names(predictions)) {
          if (!is.null(predictions[[term_name]])) {
            predictions[[term_name]] <- class_probs_to_survival_probs(
              predictions[[term_name]],
              num_categories
            )
          } else {
            result[[term_name]] <- NULL
          }
        }
      } else {
        predictions <- class_probs_to_survival_probs(
          predictions,
          num_categories
        )
      }
    }
  }

  # Compute the interval
  if (has_multiple_terms) {
    result <- list()
    for (term_name in names(predictions)) {
      if (!is.null(predictions[[term_name]])) {
        sample_dim <- compute_sample_dim(predictions[[term_name]], num_samples)
        result[[term_name]] <- summarize_interval(
          predictions[[term_name]],
          sample_dim = sample_dim,
          level = level
        )
      } else {
        result[[term_name]] <- NULL
      }
    }
    return(result)
  } else {
    sample_dim <- compute_sample_dim(predictions, num_samples)
    return(summarize_interval(
      predictions,
      sample_dim = sample_dim,
      level = level
    ))
  }
}

class_probs_to_survival_probs <- function(probs, num_categories) {
  pred_dims <- dim(probs)
  pred_dims[2] <- pred_dims[2] - 1
  output <- array(NA_real_, dim = pred_dims)
  for (i in 2:num_categories) {
    output[, i - 1, ] <- apply(
      probs[,
        i:num_categories,
        ,
        drop = F
      ],
      c(1, 3),
      sum
    )
  }
  return(output)
}

compute_sample_dim <- function(predictions, num_samples) {
  term_dims <- dim(predictions)
  if (length(term_dims) <= 1) {
    sample_dim <- 1
  } else {
    sample_dim <- which(term_dims == num_samples)
    if (length(sample_dim) > 1) {
      # Use a heuristic and raise a warning
      warning(
        "Multiple posterior dimensions matching the number of posterior draws found in the array, using the last one as the MCMC index"
      )
      sample_dim <- sample_dim[length(sample_dim)]
    } else if (length(sample_dim) == 0) {
      # "No hit" case raises an error
      stop(
        "No posterior dimension was found that matches the number of posterior draws"
      )
    }
  }
  return(sample_dim)
}

summarize_interval <- function(array, sample_dim = 2, level = 0.95) {
  # Check that the array is numeric and at least 2 dimensional
  stopifnot(is.numeric(array) && length(dim(array)) >= 2)

  # Compute lower and upper quantiles based on the requested interval
  quantile_lb <- (1 - level) / 2
  quantile_ub <- 1 - quantile_lb

  # Determine the dimensions over which interval is computed
  apply_dim <- setdiff(1:length(dim(array)), sample_dim)

  # Calculate the interval
  result_lb <- apply(array, apply_dim, function(x) {
    quantile(x, probs = quantile_lb, names = FALSE)
  })
  result_ub <- apply(array, apply_dim, function(x) {
    quantile(x, probs = quantile_ub, names = FALSE)
  })

  return(list(lower = result_lb, upper = result_ub))
}

check_model_is_valid <- function(model_object) {
  if (
    (!inherits(model_object, "bartmodel")) &&
      (!inherits(model_object, "bcfmodel"))
  ) {
    stop("'model_object' must be a bartmodel or bcfmodel")
  }
}

check_model_has_term <- function(model_object, term) {
  # Parse inputs
  if (!is.character(term) || length(term) != 1) {
    stop("'term' must be a single character string")
  }
  if (
    (!inherits(model_object, "bartmodel")) &&
      (!inherits(model_object, "bcfmodel"))
  ) {
    stop("'model_object' must be a bartmodel or bcfmodel")
  }
  model_type <- ifelse(inherits(model_object, "bartmodel"), "bart", "bcf")

  # Check if the term was fitted as part of the provided model
  if (model_type == "bart") {
    validate_bart_term(term)
    return(bart_model_has_term(model_object, term))
  } else {
    validate_bcf_term(term)
    return(bcf_model_has_term(model_object, term))
  }
}

bart_model_has_term <- function(model_object, term) {
  if (term == "mean_forest") {
    return(model_object$model_params$include_mean_forest)
  } else if (term == "variance_forest") {
    return(model_object$model_params$include_variance_forest)
  } else if (term == "rfx") {
    return(model_object$model_params$has_rfx)
  } else if (term == "y_hat") {
    return(
      model_object$model_params$include_mean_forest ||
        model_object$model_params$has_rfx
    )
  } else if (term == "all") {
    return(TRUE)
  } else {
    return(FALSE)
  }
}

bcf_model_has_term <- function(model_object, term) {
  if (term == "prognostic_function") {
    return(TRUE)
  } else if (term == "mu") {
    return(TRUE)
  } else if (term == "cate") {
    return(TRUE)
  } else if (term == "tau") {
    return(TRUE)
  } else if (term == "variance_forest") {
    return(model_object$model_params$include_variance_forest)
  } else if (term == "rfx") {
    return(model_object$model_params$has_rfx)
  } else if (term == "y_hat") {
    return(TRUE)
  } else if (term == "all") {
    return(TRUE)
  } else {
    return(FALSE)
  }
}

validate_bart_term <- function(term) {
  model_terms <- c("mean_forest", "variance_forest", "rfx", "y_hat", "all")
  if (!(term %in% model_terms)) {
    stop(
      "'term' must be one of 'mean_forest', 'variance_forest', 'rfx', 'y_hat', or 'all' for bartmodel objects"
    )
  }
}

validate_bcf_term <- function(term) {
  model_terms <- c(
    "prognostic_function",
    "mu",
    "cate",
    "tau",
    "variance_forest",
    "rfx",
    "y_hat",
    "all"
  )
  if (!(term %in% model_terms)) {
    stop(
      "'term' must be one of 'prognostic_function', 'mu', 'cate', 'tau', 'variance_forest', 'rfx', 'y_hat', or 'all' for bcfmodel objects"
    )
  }
}

run_bart_factorial <- function(
  bart_data_train,
  bart_data_test,
  leaf_reg = "none",
  variance_forest = FALSE,
  random_effects = "none",
  sampling_global_error_scale = FALSE,
  sampling_leaf_scale = FALSE,
  outcome_type = "continuous",
  num_chains = 1
) {
  # Unpack BART training data
  y <- bart_data_train[["y"]]
  X <- bart_data_train[["X"]]
  if (leaf_reg != "none") {
    leaf_basis <- bart_data_train[["leaf_basis"]]
  } else {
    leaf_basis <- NULL
  }
  if (random_effects != "none") {
    rfx_group_ids <- bart_data_train[["rfx_group_ids"]]
  } else {
    rfx_group_ids <- NULL
  }
  if (random_effects == "custom") {
    rfx_basis <- bart_data_train[["rfx_basis"]]
  } else {
    rfx_basis <- NULL
  }

  # Set BART model parameters
  general_params <- list(
    num_chains = num_chains,
    sample_sigma2_global = sampling_global_error_scale,
    outcome_model = OutcomeModel(outcome = outcome_type)
  )
  mean_forest_params <- list(
    sample_sigma2_leaf = sampling_leaf_scale
  )
  variance_forest_params <- list(
    num_trees = ifelse(variance_forest, 20, 0)
  )
  rfx_params <- list(
    model_spec = ifelse(random_effects == "none", "custom", random_effects)
  )

  # Sample BART model
  bart_model <- stochtree::bart(
    X_train = X,
    y_train = y,
    leaf_basis_train = leaf_basis,
    rfx_group_ids_train = rfx_group_ids,
    rfx_basis_train = rfx_basis,
    general_params = general_params,
    mean_forest_params = mean_forest_params,
    variance_forest_params = variance_forest_params,
    random_effects_params = rfx_params
  )

  # Unpack test set data
  y_test <- bart_data_test[["y"]]
  X_test <- bart_data_test[["X"]]
  if (leaf_reg != "none") {
    leaf_basis_test <- bart_data_test[["leaf_basis"]]
  } else {
    leaf_basis_test <- NULL
  }
  if (random_effects != "none") {
    rfx_group_ids_test <- bart_data_test[["rfx_group_ids"]]
  } else {
    rfx_group_ids_test <- NULL
  }
  if (random_effects == "custom") {
    rfx_basis_test <- bart_data_test[["rfx_basis"]]
  } else {
    rfx_basis_test <- NULL
  }

  # Predict on test set
  mean_preds <- predict(
    bart_model,
    X = X_test,
    leaf_basis = leaf_basis_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test,
    type = "mean",
    terms = "all",
    scale = ifelse(outcome_type == "binary", "probability", "linear")
  )
  posterior_preds <- predict(
    bart_model,
    X = X_test,
    leaf_basis = leaf_basis_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test,
    type = "posterior",
    terms = "all",
    scale = ifelse(outcome_type == "binary", "probability", "linear")
  )

  # Compute intervals
  posterior_interval <- computeBARTPosteriorInterval(
    bart_model,
    terms = "all",
    level = 0.95,
    scale = ifelse(outcome_type == "binary", "probability", "linear"),
    X = X_test,
    leaf_basis = leaf_basis_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test
  )

  # Sample posterior predictive
  posterior_predictive_draws <- sampleBARTPosteriorPredictive(
    bart_model,
    X = X_test,
    leaf_basis = leaf_basis_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test,
    num_draws_per_sample = 5
  )
}

run_bcf_factorial <- function(
  bcf_data_train,
  bcf_data_test,
  treatment_type = "binary",
  variance_forest = FALSE,
  random_effects = "none",
  sampling_global_error_scale = FALSE,
  sampling_mu_leaf_scale = FALSE,
  sampling_tau_leaf_scale = FALSE,
  outcome_type = "continuous",
  num_chains = 1,
  adaptive_coding = TRUE,
  include_propensity = TRUE
) {
  # Unpack BART training data
  y <- bcf_data_train[["y"]]
  X <- bcf_data_train[["X"]]
  Z <- bcf_data_train[["Z"]]
  if (include_propensity) {
    propensity_train <- bcf_data_train[["propensity"]]
  } else {
    propensity_train <- NULL
  }
  if (random_effects != "none") {
    rfx_group_ids <- bcf_data_train[["rfx_group_ids"]]
  } else {
    rfx_group_ids <- NULL
  }
  if (random_effects == "custom") {
    rfx_basis <- bcf_data_train[["rfx_basis"]]
  } else {
    rfx_basis <- NULL
  }

  # Set BART model parameters
  general_params <- list(
    num_chains = num_chains,
    sample_sigma2_global = sampling_global_error_scale,
    outcome_model = OutcomeModel(outcome = outcome_type),
    adaptive_coding = adaptive_coding
  )
  mu_forest_params <- list(
    sample_sigma2_leaf = sampling_mu_leaf_scale
  )
  tau_forest_params <- list(
    sample_sigma2_leaf = sampling_tau_leaf_scale
  )
  variance_forest_params <- list(
    num_trees = ifelse(variance_forest, 20, 0)
  )
  rfx_params <- list(
    model_spec = ifelse(random_effects == "none", "custom", random_effects)
  )

  # Sample BART model
  bcf_model <- stochtree::bcf(
    X_train = X,
    y_train = y,
    Z_train = Z,
    propensity_train = propensity_train,
    rfx_group_ids_train = rfx_group_ids,
    rfx_basis_train = rfx_basis,
    general_params = general_params,
    prognostic_forest_params = mu_forest_params,
    treatment_effect_forest_params = tau_forest_params,
    variance_forest_params = variance_forest_params,
    random_effects_params = rfx_params
  )

  # Unpack test set data
  y_test <- bcf_data_test[["y"]]
  X_test <- bcf_data_test[["X"]]
  Z_test <- bcf_data_test[["Z"]]
  if (include_propensity) {
    propensity_test <- bcf_data_test[["propensity"]]
  } else {
    propensity_test <- NULL
  }
  if (random_effects != "none") {
    rfx_group_ids_test <- bcf_data_test[["rfx_group_ids"]]
  } else {
    rfx_group_ids_test <- NULL
  }
  if (random_effects == "custom") {
    rfx_basis_test <- bcf_data_test[["rfx_basis"]]
  } else {
    rfx_basis_test <- NULL
  }

  # Predict on test set
  mean_preds <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = propensity_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test,
    type = "mean",
    terms = "all",
    scale = ifelse(outcome_type == "binary", "probability", "linear")
  )
  posterior_preds <- predict(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = propensity_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test,
    type = "posterior",
    terms = "all",
    scale = ifelse(outcome_type == "binary", "probability", "linear")
  )

  # Compute intervals
  posterior_interval <- computeBCFPosteriorInterval(
    bcf_model,
    terms = "all",
    level = 0.95,
    scale = ifelse(outcome_type == "binary", "probability", "linear"),
    X = X_test,
    Z = Z_test,
    propensity = propensity_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test
  )

  # Sample posterior predictive
  posterior_predictive_draws <- sampleBCFPosteriorPredictive(
    bcf_model,
    X = X_test,
    Z = Z_test,
    propensity = propensity_test,
    rfx_group_ids = rfx_group_ids_test,
    rfx_basis = rfx_basis_test,
    num_draws_per_sample = 5
  )
}

# Construct chained expectations without writing out every combination of function calls
construct_chained_expectation_bart <- function(
  error_cond,
  warning_cond_1,
  warning_cond_2,
  warning_cond_3
) {
  # Build the chain from innermost to outermost
  function_text <- "x"
  if (warning_cond_1) {
    function_text <- paste0(
      "warning_fun_1(",
      function_text,
      ")"
    )
  }
  if (warning_cond_2) {
    function_text <- paste0(
      "warning_fun_2(",
      function_text,
      ")"
    )
  }
  if (warning_cond_3) {
    function_text <- paste0(
      "warning_fun_3(",
      function_text,
      ")"
    )
  }
  if (error_cond) {
    function_text <- paste0(
      "expect_error(",
      function_text,
      ")"
    )
  }
  return(as.function(
    c(alist(x = ), parse(text = function_text)[[1]]),
    envir = parent.frame()
  ))
}

construct_chained_expectation_bcf <- function(
  error_cond,
  warning_cond_1,
  warning_cond_2,
  warning_cond_3,
  warning_cond_4,
  warning_cond_5,
  warning_cond_6
) {
  # Build the chain from innermost to outermost
  function_text <- "x"
  if (warning_cond_1) {
    function_text <- paste0(
      "warning_fun_1(",
      function_text,
      ")"
    )
  }
  if (warning_cond_2) {
    function_text <- paste0(
      "warning_fun_2(",
      function_text,
      ")"
    )
  }
  if (warning_cond_3) {
    function_text <- paste0(
      "warning_fun_3(",
      function_text,
      ")"
    )
  }
  if (warning_cond_4) {
    function_text <- paste0(
      "warning_fun_4(",
      function_text,
      ")"
    )
  }
  if (warning_cond_5) {
    function_text <- paste0(
      "warning_fun_5(",
      function_text,
      ")"
    )
  }
  if (warning_cond_6) {
    function_text <- paste0(
      "warning_fun_6(",
      function_text,
      ")"
    )
  }
  if (error_cond) {
    function_text <- paste0(
      "expect_error(",
      function_text,
      ")"
    )
  }
  return(as.function(
    c(alist(x = ), parse(text = function_text)[[1]]),
    envir = parent.frame()
  ))
}

test_that("Quick check of interactions between components of BART functionality", {
  skip_on_cran()
  # Code from: https://github.com/r-lib/testthat/blob/main/R/skip.R
  skip_if(
    isFALSE(as.logical(Sys.getenv("RUN_SLOW_TESTS", "false"))),
    "skipping slow tests"
  )

  # Overall, we have seven components of a BART sampler which can be on / off or set to different levels:
  # 1. Leaf regression: none, univariate, multivariate
  # 2. Variance forest: no, yes
  # 3. Random effects: no, custom basis, `intercept_only`
  # 4. Sampling global error scale: no, yes
  # 5. Sampling leaf scale on mean forest: no, yes (only available for constant leaf or univariate leaf regression)
  # 6. Outcome type: continuous (identity link), binary (probit link)
  # 7. Number of chains: 1, >1
  #
  # For each of the possible models this implies,
  # we'd like to be sure that stochtree functions that operate on BART models
  # will run without error. Since there are so many possible models implied by the
  # options above, this test is designed to be quick (small sample size, low dimensional data)
  # and we are only interested in ensuring no errors are triggered.

  # Generate data with random effects
  n <- 50
  p <- 3
  num_basis <- 2
  num_rfx_groups <- 3
  num_rfx_basis <- 2
  X <- matrix(runif(n * p), ncol = p)
  leaf_basis <- matrix(runif(n * num_basis), ncol = num_basis)
  leaf_coefs <- runif(num_basis)
  group_ids <- sample(1:num_rfx_groups, n, replace = T)
  rfx_basis <- matrix(runif(n * num_rfx_basis), ncol = num_rfx_basis)
  rfx_coefs <- matrix(
    runif(num_rfx_groups * num_rfx_basis),
    ncol = num_rfx_basis
  )
  mean_term <- sin(X[, 1]) * rowSums(leaf_basis * leaf_coefs)
  rfx_term <- rowSums(rfx_coefs[group_ids, ] * rfx_basis)
  E_y <- mean_term + rfx_term
  E_y <- E_y - mean(E_y)
  epsilon <- rnorm(n, 0, 1)
  y_continuous <- E_y + epsilon
  y_binary <- 1 * (y_continuous > 0)

  # Split into test and train sets
  test_set_pct <- 0.5
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  leaf_basis_test <- leaf_basis[test_inds, ]
  leaf_basis_train <- leaf_basis[train_inds, ]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  group_ids_test <- group_ids[test_inds]
  group_ids_train <- group_ids[train_inds]
  y_continuous_test <- y_continuous[test_inds]
  y_continuous_train <- y_continuous[train_inds]
  y_binary_test <- y_binary[test_inds]
  y_binary_train <- y_binary[train_inds]

  # Run the power set of models
  leaf_reg_options <- c("none", "univariate", "multivariate")
  variance_forest_options <- c(FALSE, TRUE)
  random_effects_options <- c("none", "custom", "intercept_only")
  sampling_global_error_scale_options <- c(FALSE, TRUE)
  sampling_leaf_scale_options <- c(FALSE, TRUE)
  outcome_type_options <- c("continuous", "binary")
  num_chains_options <- c(1, 3)
  model_options_df <- expand.grid(
    leaf_reg = leaf_reg_options,
    variance_forest = variance_forest_options,
    random_effects = random_effects_options,
    sampling_global_error_scale = sampling_global_error_scale_options,
    sampling_leaf_scale = sampling_leaf_scale_options,
    outcome_type = outcome_type_options,
    num_chains = num_chains_options,
    stringsAsFactors = FALSE
  )
  for (i in 1:nrow(model_options_df)) {
    # Determine which errors and warnings should be triggered
    error_cond <- (model_options_df$variance_forest[i]) &&
      (model_options_df$outcome_type[i] == "binary")
    warning_cond_1 <- (model_options_df$sampling_leaf_scale[i]) &&
      (model_options_df$leaf_reg[i] == "multivariate")
    warning_fun_1 <- function(x) {
      expect_warning(
        x,
        "Sampling leaf scale not yet supported for multivariate leaf models, so the leaf scale parameter will not be sampled in this model."
      )
    }
    warning_cond_2 <- (model_options_df$sampling_global_error_scale[i]) &&
      (model_options_df$outcome_type[i] == "binary")
    warning_fun_2 <- function(x) {
      expect_warning(
        x,
        "Global error variance will not be sampled with a probit link as it is fixed at 1"
      )
    }
    warning_cond_3 <- (model_options_df$sampling_global_error_scale[i]) &&
      (model_options_df$variance_forest[i])
    warning_fun_3 <- function(x) {
      expect_warning(
        x,
        "Global error variance will not be sampled with a heteroskedasticity"
      )
    }
    warning_cond <- warning_cond_1 || warning_cond_2 || warning_cond_3

    if (error_cond || warning_cond) {
      test_fun <- construct_chained_expectation_bart(
        error_cond = error_cond,
        warning_cond_1 = warning_cond_1,
        warning_cond_2 = warning_cond_2,
        warning_cond_3 = warning_cond_3
      )
    } else {
      test_fun <- expect_no_error
    }

    # Prepare test function arguments
    bart_data_train <- list(X = X_train)
    bart_data_test <- list(X = X_test)
    if (model_options_df$outcome_type[i] == "continuous") {
      bart_data_train[["y"]] <- y_continuous_train
      bart_data_test[["y"]] <- y_continuous_test
    } else {
      bart_data_train[["y"]] <- y_binary_train
      bart_data_test[["y"]] <- y_binary_test
    }
    if (model_options_df$leaf_reg[i] != "none") {
      if (model_options_df$leaf_reg[i] == "univariate") {
        bart_data_train[["leaf_basis"]] <- leaf_basis_train[, 1, drop = FALSE]
        bart_data_test[["leaf_basis"]] <- leaf_basis_test[, 1, drop = FALSE]
      } else {
        bart_data_train[["leaf_basis"]] <- leaf_basis_train
        bart_data_test[["leaf_basis"]] <- leaf_basis_test
      }
    } else {
      bart_data_train[["leaf_basis"]] <- NULL
      bart_data_test[["leaf_basis"]] <- NULL
    }
    if (model_options_df$random_effects[i] != "none") {
      bart_data_train[["rfx_group_ids"]] <- group_ids_train
      bart_data_test[["rfx_group_ids"]] <- group_ids_test
    } else {
      bart_data_train[["rfx_group_ids"]] <- NULL
      bart_data_test[["rfx_group_ids"]] <- NULL
    }
    if (model_options_df$random_effects[i] == "custom") {
      bart_data_train[["rfx_basis"]] <- rfx_basis_train
      bart_data_test[["rfx_basis"]] <- rfx_basis_test
    } else {
      bart_data_train[["rfx_basis"]] <- NULL
      bart_data_test[["rfx_basis"]] <- NULL
    }

    # Apply testthat expectation(s)
    test_fun({
      run_bart_factorial(
        bart_data_train = bart_data_train,
        bart_data_test = bart_data_test,
        leaf_reg = model_options_df$leaf_reg[i],
        variance_forest = model_options_df$variance_forest[i],
        random_effects = model_options_df$random_effects[i],
        sampling_global_error_scale = model_options_df$sampling_global_error_scale[
          i
        ],
        sampling_leaf_scale = model_options_df$sampling_leaf_scale[
          i
        ],
        outcome_type = model_options_df$outcome_type[i],
        num_chains = model_options_df$num_chains[i]
      )
    })
  }
})

test_that("Quick check of interactions between components of BCF functionality", {
  skip_on_cran()
  # Code from: https://github.com/r-lib/testthat/blob/main/R/skip.R
  skip_if(
    isFALSE(as.logical(Sys.getenv("RUN_SLOW_TESTS", "false"))),
    "skipping slow tests"
  )

  # Overall, we have nine components of a BCF sampler which can be on / off or set to different levels:
  # 1. treatment: binary, univariate continuous, multivariate
  # 2. Variance forest: no, yes
  # 3. Random effects: no, custom basis, `intercept_only`, `intercept_plus_treatment`
  # 4. Sampling global error scale: no, yes
  # 5. Sampling leaf scale on prognostic forest: no, yes
  # 6. Sampling leaf scale on treatment forest: no, yes (only available for univariate treatment)
  # 7. Outcome type: continuous (identity link), binary (probit link)
  # 8. Number of chains: 1, >1
  # 9. Adaptive coding: no, yes
  #
  # For each of the possible models this implies,
  # we'd like to be sure that stochtree functions that operate on BCF models
  # will run without error. Since there are so many possible models implied by the
  # options above, this test is designed to be quick (small sample size, low dimensional data)
  # and we are only interested in ensuring no errors are triggered.

  # Generate data with random effects
  n <- 50
  p <- 3
  num_rfx_groups <- 3
  num_rfx_basis <- 2
  X <- matrix(runif(n * p), ncol = p)
  binary_treatment <- rbinom(n, 1, 0.5)
  continuous_treatment <- runif(n, 0, 1)
  multivariate_treatment <- cbind(
    binary_treatment,
    continuous_treatment
  )
  group_ids <- sample(1:num_rfx_groups, n, replace = T)
  rfx_basis <- matrix(runif(n * num_rfx_basis), ncol = num_rfx_basis)
  rfx_coefs <- matrix(
    runif(num_rfx_groups * num_rfx_basis),
    ncol = num_rfx_basis
  )
  propensity <- runif(n)
  prognostic_term <- sin(X[, 1])
  binary_treatment_effect <- X[, 2]
  continuous_treatment_effect <- X[, 3]
  rfx_term <- rowSums(rfx_coefs[group_ids, ] * rfx_basis)
  E_y <- prognostic_term +
    binary_treatment_effect * binary_treatment +
    continuous_treatment_effect * continuous_treatment +
    rfx_term
  E_y <- E_y - mean(E_y)
  epsilon <- rnorm(n, 0, 1)
  y_continuous <- E_y + epsilon
  y_binary <- 1 * (y_continuous > 0)

  # Split into test and train sets
  test_set_pct <- 0.5
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  binary_treatment_test <- binary_treatment[test_inds]
  binary_treatment_train <- binary_treatment[train_inds]
  propensity_test <- propensity[test_inds]
  propensity_train <- propensity[train_inds]
  continuous_treatment_test <- continuous_treatment[test_inds]
  continuous_treatment_train <- continuous_treatment[train_inds]
  multivariate_treatment_test <- multivariate_treatment[test_inds, ]
  multivariate_treatment_train <- multivariate_treatment[train_inds, ]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  group_ids_test <- group_ids[test_inds]
  group_ids_train <- group_ids[train_inds]
  y_continuous_test <- y_continuous[test_inds]
  y_continuous_train <- y_continuous[train_inds]
  y_binary_test <- y_binary[test_inds]
  y_binary_train <- y_binary[train_inds]

  # Run the power set of models
  treatment_options <- c("binary", "univariate_continuous", "multivariate")
  variance_forest_options <- c(FALSE, TRUE)
  random_effects_options <- c(
    "none",
    "custom",
    "intercept_only",
    "intercept_plus_treatment"
  )
  sampling_global_error_scale_options <- c(FALSE, TRUE)
  sampling_mu_leaf_scale_options <- c(FALSE, TRUE)
  sampling_tau_leaf_scale_options <- c(FALSE, TRUE)
  outcome_type_options <- c("continuous", "binary")
  num_chains_options <- c(1, 3)
  adaptive_coding_options <- c(FALSE, TRUE)
  include_propensity_options <- c(FALSE, TRUE)
  model_options_df <- expand.grid(
    treatment_type = treatment_options,
    variance_forest = variance_forest_options,
    random_effects = random_effects_options,
    sampling_global_error_scale = sampling_global_error_scale_options,
    sampling_mu_leaf_scale = sampling_mu_leaf_scale_options,
    sampling_tau_leaf_scale = sampling_tau_leaf_scale_options,
    outcome_type = outcome_type_options,
    num_chains = num_chains_options,
    adaptive_coding = adaptive_coding_options,
    include_propensity = include_propensity_options,
    stringsAsFactors = FALSE
  )
  for (i in 1:nrow(model_options_df)) {
    # Determine which errors and warnings should be triggered
    error_cond <- (model_options_df$variance_forest[i]) &&
      (model_options_df$outcome_type[i] == "binary")
    warning_cond_1 <- (model_options_df$sampling_tau_leaf_scale[i]) &&
      (model_options_df$treatment_type[i] == "multivariate")
    warning_fun_1 <- function(x) {
      expect_warning(
        x,
        "Sampling leaf scale not yet supported for multivariate leaf models, so the leaf scale parameter will not be sampled for the treatment forest in this model.",
        fixed = TRUE
      )
    }
    warning_cond_2 <- (!model_options_df$include_propensity[i]) &&
      (model_options_df$treatment_type[i] == "multivariate")
    warning_fun_2 <- function(x) {
      expect_warning(
        x,
        "No propensities were provided for the multivariate treatment; an internal propensity model will not be fitted to the multivariate treatment and propensity_covariate will be set to 'none'",
        fixed = TRUE
      )
    }
    warning_cond_3 <- (model_options_df$adaptive_coding[i]) &&
      (model_options_df$treatment_type[i] != "binary")
    warning_fun_3 <- function(x) {
      expect_warning(
        x,
        "Adaptive coding is only compatible with binary (univariate) treatment and, as a result, will be ignored in sampling this model",
        fixed = TRUE
      )
    }
    warning_cond_4 <- (model_options_df$sampling_global_error_scale[i]) &&
      (model_options_df$outcome_type[i] == "binary")
    warning_fun_4 <- function(x) {
      expect_warning(
        x,
        "Global error variance will not be sampled with a probit link as it is fixed at 1",
        fixed = TRUE
      )
    }
    warning_cond_5 <- (model_options_df$sampling_global_error_scale[i]) &&
      (model_options_df$variance_forest[i])
    warning_fun_5 <- function(x) {
      expect_warning(
        x,
        "Global error variance will not be sampled with a heteroskedasticity",
        fixed = TRUE
      )
    }
    warning_cond_6 <- (model_options_df$treatment_type[i] == "multivariate") &&
      (model_options_df$random_effects[i] == "intercept_plus_treatment")
    warning_fun_6 <- function(x) {
      expect_warning(
        x,
        "Random effects `intercept_plus_treatment` specification is not currently implemented for multivariate treatments. This model will be fit under the `intercept_only` specification instead. Please provide a custom `rfx_basis_train` if you wish to have random slopes on multivariate treatment variables.",
        fixed = TRUE
      )
    }
    warning_cond <- (warning_cond_1 ||
      warning_cond_2 ||
      warning_cond_3 ||
      warning_cond_4 ||
      warning_cond_5 ||
      warning_cond_6)

    # Generate something like the below code but for all five warnings
    if (error_cond || warning_cond) {
      test_fun <- construct_chained_expectation_bcf(
        error_cond = error_cond,
        warning_cond_1 = warning_cond_1,
        warning_cond_2 = warning_cond_2,
        warning_cond_3 = warning_cond_3,
        warning_cond_4 = warning_cond_4,
        warning_cond_5 = warning_cond_5,
        warning_cond_6 = warning_cond_6
      )
    } else {
      test_fun <- expect_no_error
    }

    # Prepare test function arguments
    bcf_data_train <- list(X = X_train)
    bcf_data_test <- list(X = X_test)
    if (model_options_df$outcome_type[i] == "continuous") {
      bcf_data_train[["y"]] <- y_continuous_train
      bcf_data_test[["y"]] <- y_continuous_test
    } else {
      bcf_data_train[["y"]] <- y_binary_train
      bcf_data_test[["y"]] <- y_binary_test
    }
    if (model_options_df$include_propensity[i]) {
      bcf_data_train[["propensity"]] <- propensity_train
      bcf_data_test[["propensity"]] <- propensity_test
    } else {
      bcf_data_train[["propensity"]] <- NULL
      bcf_data_test[["propensity"]] <- NULL
    }
    if (model_options_df$treatment_type[i] == "binary") {
      bcf_data_train[["Z"]] <- binary_treatment_train
      bcf_data_test[["Z"]] <- binary_treatment_test
    } else if (model_options_df$treatment_type[i] == "univariate_continuous") {
      bcf_data_train[["Z"]] <- continuous_treatment_train
      bcf_data_test[["Z"]] <- continuous_treatment_test
    } else {
      bcf_data_train[["Z"]] <- multivariate_treatment_train
      bcf_data_test[["Z"]] <- multivariate_treatment_test
    }
    if (model_options_df$random_effects[i] != "none") {
      bcf_data_train[["rfx_group_ids"]] <- group_ids_train
      bcf_data_test[["rfx_group_ids"]] <- group_ids_test
    } else {
      bcf_data_train[["rfx_group_ids"]] <- NULL
      bcf_data_test[["rfx_group_ids"]] <- NULL
    }
    if (model_options_df$random_effects[i] == "custom") {
      bcf_data_train[["rfx_basis"]] <- rfx_basis_train
      bcf_data_test[["rfx_basis"]] <- rfx_basis_test
    } else {
      bcf_data_train[["rfx_basis"]] <- NULL
      bcf_data_test[["rfx_basis"]] <- NULL
    }

    # Apply testthat expectation(s)
    test_fun({
      run_bcf_factorial(
        bcf_data_train = bcf_data_train,
        bcf_data_test = bcf_data_test,
        treatment_type = model_options_df$treatment_type[i],
        variance_forest = model_options_df$variance_forest[i],
        random_effects = model_options_df$random_effects[i],
        sampling_global_error_scale = model_options_df$sampling_global_error_scale[
          i
        ],
        sampling_mu_leaf_scale = model_options_df$sampling_mu_leaf_scale[
          i
        ],
        sampling_tau_leaf_scale = model_options_df$sampling_tau_leaf_scale[
          i
        ],
        outcome_type = model_options_df$outcome_type[i],
        num_chains = model_options_df$num_chains[i],
        adaptive_coding = model_options_df$adaptive_coding[i],
        include_propensity = model_options_df$include_propensity[i]
      )
    })
  }
})

#' @title Random Number Generator C++ Wrapper
#' @description
#' Wrapper around a C++ random number generator object (for reproducibility).
#' The class persists a C++ random number generator throughout an R session to
#' ensure a given seed generates the same outputs (on the same OS). If no seed is provided,
#' the C++ random number generator is initialized using `std::random_device`.

CppRNG <- R6::R6Class(
  classname = "CppRNG",
  cloneable = FALSE,
  public = list(
    #' @field rng_ptr External pointer to a C++ std::mt19937 class
    rng_ptr = NULL,

    #' @description
    #' Create a new CppRNG object.
    #' @param random_seed (Optional) random seed for sampling
    #' @return A new `CppRNG` object.
    initialize = function(random_seed = -1) {
      self$rng_ptr <- rng_cpp(random_seed)
    }
  )
)

#' @title Forest Model C++ Wrapper
#' @description
#' Wraps the C++ data structures needed to sample an ensemble of decision
#' trees and exposes functionality to run a forest sampler
#' (using either MCMC or the grow-from-root algorithm).
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

ForestModel <- R6::R6Class(
  classname = "ForestModel",
  cloneable = FALSE,
  public = list(
    #' @field tracker_ptr External pointer to a C++ ForestTracker class
    tracker_ptr = NULL,

    #' @field tree_prior_ptr External pointer to a C++ TreePrior class
    tree_prior_ptr = NULL,

    #' @description
    #' Create a new ForestModel object.
    #' @param forest_dataset `ForestDataset` object, used to initialize forest sampling data structures
    #' @param feature_types Feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)
    #' @param num_trees Number of trees in the forest being sampled
    #' @param n Number of observations in `forest_dataset`
    #' @param alpha Root node split probability in tree prior
    #' @param beta Depth prior penalty in tree prior
    #' @param min_samples_leaf Minimum number of samples in a tree leaf
    #' @param max_depth Maximum depth that any tree can reach
    #' @return A new `ForestModel` object.
    initialize = function(
      forest_dataset,
      feature_types,
      num_trees,
      n,
      alpha,
      beta,
      min_samples_leaf,
      max_depth = -1
    ) {
      stopifnot(!is.null(forest_dataset$data_ptr))
      self$tracker_ptr <- forest_tracker_cpp(
        forest_dataset$data_ptr,
        feature_types,
        num_trees,
        n
      )
      self$tree_prior_ptr <- tree_prior_cpp(
        alpha,
        beta,
        min_samples_leaf,
        max_depth
      )
    },

    #' @description
    #' Run a single iteration of the forest sampling algorithm (MCMC or GFR)
    #' @param forest_dataset Dataset used to sample the forest
    #' @param residual Outcome used to sample the forest
    #' @param forest_samples Container of forest samples
    #' @param active_forest "Active" forest updated by the sampler in each iteration
    #' @param rng Wrapper around C++ random number generator
    #' @param forest_model_config ForestModelConfig object containing forest model parameters and settings
    #' @param global_model_config GlobalModelConfig object containing global model parameters and settings
    #' @param num_threads Number of threads to use in the GFR and MCMC algorithms, as well as prediction. If OpenMP is not available on a user's system, this will default to `1`, otherwise to the maximum number of available threads.
    #' @param keep_forest (Optional) Whether the updated forest sample should be saved to `forest_samples`. Default: `TRUE`.
    #' @param gfr (Optional) Whether or not the forest should be sampled using the "grow-from-root" (GFR) algorithm. Default: `TRUE`.
    sample_one_iteration = function(
      forest_dataset,
      residual,
      forest_samples,
      active_forest,
      rng,
      forest_model_config,
      global_model_config,
      num_threads = -1,
      keep_forest = TRUE,
      gfr = TRUE
    ) {
      if (active_forest$is_empty()) {
        stop(
          "`active_forest` has not yet been initialized, which is necessary to run the sampler. Please set constant values for `active_forest`'s leaves using either the `set_root_leaves` or `prepare_for_sampler` methods."
        )
      }

      # Unpack parameters from model config object
      feature_types <- forest_model_config$feature_types
      sweep_update_indices <- forest_model_config$sweep_update_indices
      leaf_model_int <- forest_model_config$leaf_model_type
      leaf_model_scale <- forest_model_config$leaf_model_scale
      variable_weights <- forest_model_config$variable_weights
      if (leaf_model_int == 4) {
        a_forest <- forest_model_config$cloglog_forest_shape
        b_forest <- forest_model_config$cloglog_forest_rate
      } else {
        a_forest <- forest_model_config$variance_forest_shape
        b_forest <- forest_model_config$variance_forest_scale
      }
      global_scale <- global_model_config$global_error_variance
      cutpoint_grid_size <- forest_model_config$cutpoint_grid_size
      num_features_subsample <- forest_model_config$num_features_subsample

      # Default to empty integer vector if sweep_update_indices is NULL
      if (is.null(sweep_update_indices)) {
        # sweep_update_indices <- integer(0)
        sweep_update_indices <- 0:(forest_model_config$num_trees - 1)
      }

      # Detect changes to tree prior
      if (
        forest_model_config$alpha !=
          get_alpha_tree_prior_cpp(self$tree_prior_ptr)
      ) {
        update_alpha_tree_prior_cpp(
          self$tree_prior_ptr,
          forest_model_config$alpha
        )
      }
      if (
        forest_model_config$beta != get_beta_tree_prior_cpp(self$tree_prior_ptr)
      ) {
        update_beta_tree_prior_cpp(
          self$tree_prior_ptr,
          forest_model_config$beta
        )
      }
      if (
        forest_model_config$min_samples_leaf !=
          get_min_samples_leaf_tree_prior_cpp(self$tree_prior_ptr)
      ) {
        update_min_samples_leaf_tree_prior_cpp(
          self$tree_prior_ptr,
          forest_model_config$min_samples_leaf
        )
      }
      if (
        forest_model_config$max_depth !=
          get_max_depth_tree_prior_cpp(self$tree_prior_ptr)
      ) {
        update_max_depth_tree_prior_cpp(
          self$tree_prior_ptr,
          forest_model_config$max_depth
        )
      }

      # Run the sampler
      if (gfr) {
        sample_gfr_one_iteration_cpp(
          forest_dataset$data_ptr,
          residual$data_ptr,
          forest_samples$forest_container_ptr,
          active_forest$forest_ptr,
          self$tracker_ptr,
          self$tree_prior_ptr,
          rng$rng_ptr,
          sweep_update_indices,
          feature_types,
          cutpoint_grid_size,
          leaf_model_scale,
          variable_weights,
          a_forest,
          b_forest,
          global_scale,
          leaf_model_int,
          keep_forest,
          num_features_subsample,
          num_threads
        )
      } else {
        sample_mcmc_one_iteration_cpp(
          forest_dataset$data_ptr,
          residual$data_ptr,
          forest_samples$forest_container_ptr,
          active_forest$forest_ptr,
          self$tracker_ptr,
          self$tree_prior_ptr,
          rng$rng_ptr,
          sweep_update_indices,
          feature_types,
          cutpoint_grid_size,
          leaf_model_scale,
          variable_weights,
          a_forest,
          b_forest,
          global_scale,
          leaf_model_int,
          keep_forest,
          num_threads
        )
      }
    },

    #' @description
    #' Extract an internally-cached prediction of a forest on the training dataset in a sampler.
    #' @return Vector with as many elements as observations in the training dataset
    get_cached_forest_predictions = function() {
      get_cached_forest_predictions_cpp(self$tracker_ptr)
    },

    #' @description
    #' Propagates basis update through to the (full/partial) residual by iteratively
    #' (a) adding back in the previous prediction of each tree, (b) recomputing predictions
    #' for each tree (caching on the C++ side), (c) subtracting the new predictions from the residual.
    #'
    #' This is useful in cases where a basis (for e.g. leaf regression) is updated outside
    #' of a tree sampler (as with e.g. adaptive coding for binary treatment BCF).
    #' Once a basis has been updated, the overall "function" represented by a tree model has
    #' changed and this should be reflected through to the residual before the next sampling loop is run.
    #' @param dataset `ForestDataset` object storing the covariates and bases for a given forest
    #' @param outcome `Outcome` object storing the residuals to be updated based on forest predictions
    #' @param active_forest "Active" forest updated by the sampler in each iteration
    propagate_basis_update = function(dataset, outcome, active_forest) {
      stopifnot(!is.null(dataset$data_ptr))
      stopifnot(!is.null(outcome$data_ptr))
      stopifnot(!is.null(self$tracker_ptr))
      stopifnot(!is.null(active_forest$forest_ptr))

      propagate_basis_update_active_forest_cpp(
        dataset$data_ptr,
        outcome$data_ptr,
        active_forest$forest_ptr,
        self$tracker_ptr
      )
    },

    #' @description
    #' Update the current state of the outcome (i.e. partial residual) data by subtracting the current predictions of each tree.
    #' This function is run after the `Outcome` class's `update_data` method, which overwrites the partial residual with an entirely new stream of outcome data.
    #' @param residual Outcome used to sample the forest
    #' @return None
    propagate_residual_update = function(residual) {
      propagate_trees_column_vector_cpp(
        self$tracker_ptr,
        residual$data_ptr
      )
    },

    #' @description
    #' Update alpha in the tree prior
    #' @param alpha New value of alpha to be used
    #' @return None
    update_alpha = function(alpha) {
      update_alpha_tree_prior_cpp(self$tree_prior_ptr, alpha)
    },

    #' @description
    #' Update beta in the tree prior
    #' @param beta New value of beta to be used
    #' @return None
    update_beta = function(beta) {
      update_beta_tree_prior_cpp(self$tree_prior_ptr, beta)
    },

    #' @description
    #' Update min_samples_leaf in the tree prior
    #' @param min_samples_leaf New value of min_samples_leaf to be used
    #' @return None
    update_min_samples_leaf = function(min_samples_leaf) {
      update_min_samples_leaf_tree_prior_cpp(
        self$tree_prior_ptr,
        min_samples_leaf
      )
    },

    #' @description
    #' Update max_depth in the tree prior
    #' @param max_depth New value of max_depth to be used
    #' @return None
    update_max_depth = function(max_depth) {
      update_max_depth_tree_prior_cpp(self$tree_prior_ptr, max_depth)
    },

    #' @description
    #' Update alpha in the tree prior
    #' @return Value of alpha in the tree prior
    get_alpha = function() {
      get_alpha_tree_prior_cpp(self$tree_prior_ptr)
    },

    #' @description
    #' Update beta in the tree prior
    #' @return Value of beta in the tree prior
    get_beta = function() {
      get_beta_tree_prior_cpp(self$tree_prior_ptr)
    },

    #' @description
    #' Query min_samples_leaf in the tree prior
    #' @return Value of min_samples_leaf in the tree prior
    get_min_samples_leaf = function() {
      get_min_samples_leaf_tree_prior_cpp(self$tree_prior_ptr)
    },

    #' @description
    #' Query max_depth in the tree prior
    #' @return Value of max_depth in the tree prior
    get_max_depth = function() {
      get_max_depth_tree_prior_cpp(self$tree_prior_ptr)
    }
  )
)

#' @title Create CppRNG Object
#' @description
#' Create an R class that wraps a C++ random number generator
#'
#' @param random_seed (Optional) random seed for sampling
#'
#' @return `CppRng` object
#' @export
#'
#' @examples
#' rng <- createCppRNG(1234)
#' rng <- createCppRNG()
createCppRNG <- function(random_seed = -1) {
  return(invisible((CppRNG$new(random_seed))))
}

#' @title Create ForestModel Object
#' @description
#' Create a forest model object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param forest_dataset ForestDataset object, used to initialize forest sampling data structures
#' @param forest_model_config ForestModelConfig object containing forest model parameters and settings
#' @param global_model_config GlobalModelConfig object containing global model parameters and settings
#'
#' @return `ForestModel` object
#' @export
#'
#' @examples
#' num_trees <- 100
#' n <- 100
#' p <- 10
#' alpha <- 0.95
#' beta <- 2.0
#' min_samples_leaf <- 2
#' max_depth <- 10
#' feature_types <- as.integer(rep(0, p))
#' X <- matrix(runif(n*p), ncol = p)
#' forest_dataset <- createForestDataset(X)
#' forest_model_config <- createForestModelConfig(feature_types=feature_types,
#'                                                num_trees=num_trees, num_features=p,
#'                                                num_observations=n, alpha=alpha, beta=beta,
#'                                                min_samples_leaf=min_samples_leaf,
#'                                                max_depth=max_depth, leaf_model_type=1)
#' global_model_config <- createGlobalModelConfig(global_error_variance=1.0)
#' forest_model <- createForestModel(forest_dataset, forest_model_config, global_model_config)
createForestModel <- function(
  forest_dataset,
  forest_model_config,
  global_model_config
) {
  return(invisible(
    (ForestModel$new(
      forest_dataset,
      forest_model_config$feature_types,
      forest_model_config$num_trees,
      forest_model_config$num_observations,
      forest_model_config$alpha,
      forest_model_config$beta,
      forest_model_config$min_samples_leaf,
      forest_model_config$max_depth
    ))
  ))
}


#' @title Sample Without Replacement
#' @description
#' Draw `sample_size` samples from `population_vector` without replacement, weighted by `sampling_probabilities`
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param population_vector Vector from which to draw samples.
#' @param sampling_probabilities Vector of probabilities of drawing each element of `population_vector`.
#' @param sample_size Number of samples to draw from `population_vector`. Must be less than or equal to `length(population_vector)`
#'
#' @returns Vector of size `sample_size`
#' @export
#'
#' @examples
#' a <- as.integer(c(4,3,2,5,1,9,7))
#' p <- c(0.7,0.2,0.05,0.02,0.01,0.01,0.01)
#' num_samples <- 5
#' sample_without_replacement(a, p, num_samples)
sample_without_replacement <- function(
  population_vector,
  sampling_probabilities,
  sample_size
) {
  return(sample_without_replacement_integer_cpp(
    population_vector,
    sampling_probabilities,
    sample_size
  ))
}

#' BART Serialization Routines
#' @name BARTSerialization
#' @description
#' BART models contains external pointers to C++ objects, which means they cannot
#' be correctly serialized to `.Rds` from an R session in their default state.
#' This group of serialization functions allow us to convert between C++ data structures and a persistent JSON
#' representation. The `CppJson` class wraps a performant C++ JSON API, and the functions
#' `saveBARTModelToJson` and `createBARTModelFromJson` save to and load from this format.
#' This representation, of course, also relies on external C++ pointers, so in order to
#' save and reload BART models across sessions, we provide two other interfaces.
#'
#' `saveBARTModelToJsonString` converts a BART model to an in-memory string containing the model's
#' JSON representation and `createBARTModelFromJsonString` converts this representation back to a BART model object.
#'
#' `saveBARTModelToJsonFile` and `createBARTModelFromJsonFile` save or reload a BART model
#' directly to / from a `.json` file.
#'
#' Finally, for cases in which multiple BART models have been sampled (for instance, multiple processes
#' run via `doParallel`), we offer `createBARTModelFromCombinedJson` and `createBARTModelFromCombinedJsonString` for
#' loading a new combined BART model from a list of BART JSON objects / strings.
#' @returns
#' `saveBARTModelToJson` return an object of type `CppJson`.
#' `saveBARTModelToJsonString` returns a string dump of the BART model's JSON representation.
#' `saveBARTModelToJsonFile` returns nothing, but writes to the provided filename.
#'
#' `createBARTModelFromJson`, `createBARTModelFromJsonFile`, `createBARTModelFromJsonString`,
#' `createBARTModelFromCombinedJson`, and `createBARTModelFromCombinedJsonString` all return
#' objects of type `bartmodel`.
#' @examples
#' # Generate data
#' n <- 100
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' y <- X[,1] + rnorm(n, 0, 1)
#'
#' # Sample BART model
#' bart_model <- bart(X_train = X, y_train = y, num_gfr = 0,
#'                    num_burnin = 0, num_mcmc = 10)
#'
#' # Save to in-memory JSON
#' bart_json <- saveBARTModelToJson(bart_model)
#' # Save to JSON string
#' bart_json_string <- saveBARTModelToJsonString(bart_model)
#' # Save to JSON file
#' tmpjson <- tempfile(fileext = ".json")
#' saveBARTModelToJsonFile(bart_model, file.path(tmpjson))
#'
#' # Reload BART model from in-memory JSON object
#' bart_model_roundtrip <- createBARTModelFromJson(bart_json)
#' # Reload BART model from JSON string
#' bart_model_roundtrip <- createBARTModelFromJsonString(bart_json_string)
#' # Reload BART model from JSON file
#' bart_model_roundtrip <- createBARTModelFromJsonFile(file.path(tmpjson))
#' unlink(tmpjson)
#' # Reload BART model from list of JSON objects
#' bart_model_roundtrip <- createBARTModelFromCombinedJson(list(bart_json))
#' # Reload BART model from list of JSON strings
#' bart_model_roundtrip <- createBARTModelFromCombinedJsonString(list(bart_json_string))
#'
NULL
#> NULL

#' @title Run BART for Supervised Learning
#' @description
#' Run the BART algorithm for supervised learning.
#'
#' @param X_train Covariates used to split trees in the ensemble. May be provided either as a dataframe or a matrix.
#' Matrix covariates will be assumed to be all numeric. Covariates passed as a dataframe will be
#' preprocessed based on the variable types (e.g. categorical columns stored as unordered factors will be one-hot encoded,
#' categorical columns stored as ordered factors will passed as integers to the core algorithm, along with the metadata
#' that the column is ordered categorical).
#' @param y_train Outcome to be modeled by the ensemble.
#' @param leaf_basis_train (Optional) Bases used to define a regression model `y ~ W` in
#' each leaf of each regression tree. By default, BART assumes constant leaf node
#' parameters, implicitly regressing on a constant basis of ones (i.e. `y ~ 1`).
#' @param rfx_group_ids_train (Optional) Group labels used for an additive random effects model.
#' @param rfx_basis_train (Optional) Basis for "random-slope" regression in an additive random effects model.
#' If `rfx_group_ids_train` is provided with a regression basis, an intercept-only random effects model
#' will be estimated.
#' @param X_test (Optional) Test set of covariates used to define "out of sample" evaluation data.
#' May be provided either as a dataframe or a matrix, but the format of `X_test` must be consistent with
#' that of `X_train`.
#' @param leaf_basis_test (Optional) Test set of bases used to define "out of sample" evaluation data.
#' While a test set is optional, the structure of any provided test set must match that
#' of the training set (i.e. if both `X_train` and `leaf_basis_train` are provided, then a test set must
#' consist of `X_test` and `leaf_basis_test` with the same number of columns).
#' @param rfx_group_ids_test (Optional) Test set group labels used for an additive random effects model.
#' We do not currently support (but plan to in the near future), test set evaluation for group labels
#' that were not in the training set.
#' @param rfx_basis_test (Optional) Test set basis for "random-slope" regression in additive random effects model.
#' @param observation_weights (Optional) Numeric vector of observation weights of length `nrow(X_train)`. Weights are
#'   applied as `y_i | - ~ N(mu(X_i), sigma^2 / w_i)`, so larger weights increase an observation's influence on the fit.
#'   All weights must be non-negative. Default: `NULL` (all observations equally weighted). Compatible with Gaussian
#'   (continuous/identity) and probit outcome models; not compatible with cloglog link functions. Note: these are
#'   referred to internally in the C++ layer as "variance weights" (`var_weights`), since they scale the residual variance.
#' @param num_gfr Number of "warm-start" iterations run using the grow-from-root algorithm (He and Hahn, 2021). Default: 5.
#' @param num_burnin Number of "burn-in" iterations of the MCMC sampler. Default: 0.
#' @param num_mcmc Number of "retained" iterations of the MCMC sampler. Default: 100.
#' @param previous_model_json (Optional) JSON string containing a previous BART model. This can be used to "continue" a sampler interactively after inspecting the samples or to run parallel chains "warm-started" from existing forest samples. Default: `NULL`.
#' @param previous_model_warmstart_sample_num (Optional) Sample number from `previous_model_json` that will be used to warmstart this BART sampler. One-indexed (so that the first sample is used for warm-start by setting `previous_model_warmstart_sample_num = 1`). Default: `NULL`.  If `num_chains` in the `general_params` list is > 1, then each successive chain will be initialized from a different sample, counting backwards from `previous_model_warmstart_sample_num`. That is, if `previous_model_warmstart_sample_num = 10` and `num_chains = 4`, then chain 1 will be initialized from sample 10, chain 2 from sample 9, chain 3 from sample 8, and chain 4 from sample 7. If `previous_model_json` is provided but `previous_model_warmstart_sample_num` is NULL, the last sample in the previous model will be used to initialize the first chain, counting backwards as noted before. If more chains are requested than there are samples in `previous_model_json`, a warning will be raised and only the last sample will be used.
#' @param general_params (Optional) A list of general (non-forest-specific) model parameters, each of which has a default value processed internally, so this argument list is optional.
#'
#'   - `cutpoint_grid_size` Maximum size of the "grid" of potential cutpoints to consider in the GFR algorithm. Default: `100`.
#'   - `standardize` Whether or not to standardize the outcome (and store the offset / scale in the model object). Default: `TRUE`.
#'   - `sample_sigma2_global` Whether or not to update the `sigma^2` global error variance parameter based on `IG(sigma2_global_shape, sigma2_global_scale)`. Default: `TRUE`.
#'   - `sigma2_global_init` Starting value of global error variance parameter. Calibrated internally as `1.0*var(y_train)`, where `y_train` is the possibly standardized outcome, if not set.
#'   - `sigma2_global_shape` Shape parameter in the `IG(sigma2_global_shape, sigma2_global_scale)` global error variance model. Default: `0`.
#'   - `sigma2_global_scale` Scale parameter in the `IG(sigma2_global_shape, sigma2_global_scale)` global error variance model. Default: `0`.
#'   - `variable_weights` Numeric weights reflecting the relative probability of splitting on each variable. Does not need to sum to 1 but cannot be negative. Defaults to `rep(1/ncol(X_train), ncol(X_train))` if not set here. Note that if the propensity score is included as a covariate in either forest, its weight will default to `1/ncol(X_train)`.
#'   - `random_seed` Integer parameterizing the C++ random number generator. If not specified, the C++ random number generator is seeded according to `std::random_device`.
#'   - `keep_burnin` Whether or not "burnin" samples should be included in the stored samples of forests and other parameters. Default `FALSE`. Ignored if `num_mcmc = 0`.
#'   - `keep_gfr` Whether or not "grow-from-root" samples should be included in the stored samples of forests and other parameters. Default `FALSE`. Ignored if `num_mcmc = 0`.
#'   - `keep_every` How many iterations of the burned-in MCMC sampler should be run before forests and parameters are retained. Default `1`. Setting `keep_every <- k` for some `k > 1` will "thin" the MCMC samples by retaining every `k`-th sample, rather than simply every sample. This can reduce the autocorrelation of the MCMC samples.
#'   - `num_chains` How many independent MCMC chains should be sampled. If `num_mcmc = 0`, this is ignored. If `num_gfr = 0`, then each chain is run from root for `num_mcmc * keep_every + num_burnin` iterations, with `num_mcmc` samples retained. If `num_gfr > 0`, each MCMC chain will be initialized from a separate GFR ensemble, with the requirement that `num_gfr >= num_chains`. Default: `1`. Note that if `num_chains > 1`, the returned model object will contain samples from all chains, stored consecutively. That is, if there are 4 chains with 100 samples each, the first 100 samples will be from chain 1, the next 100 samples will be from chain 2, etc... For more detail on working with multi-chain BART models, see [the multi chain vignette](https://stochtree.ai/vignettes/multi-chain.html).
#'   - `verbose` Whether or not to print progress during the sampling loops. Default: `FALSE`.
#'   - `outcome_model` A structured `OutcomeModel` object that specifies the outcome type and desired link function. This argument pre-empts the legacy (deprecated) `probit_outcome_model` option. Default: `OutcomeModel(outcome='continuous', link='identity')`.
#'   - `probit_outcome_model` Deprecated in favor of `outcome_model`. Whether or not the outcome should be modeled as explicitly binary via a probit link. If `TRUE`, `y` must only contain the values `0` and `1`. Default: `FALSE`.
#'   - `num_threads` Number of threads to use in the GFR and MCMC algorithms, as well as prediction. If OpenMP is not available on a user's setup, this will default to `1`, otherwise to the maximum number of available threads.
#'
#' @param mean_forest_params (Optional) A list of mean forest model parameters, each of which has a default value processed internally, so this argument list is optional.
#'
#'   - `num_trees` Number of trees in the ensemble for the conditional mean model. Default: `200`. If `num_trees = 0`, the conditional mean will not be modeled using a forest, and the function will only proceed if `num_trees > 0` for the variance forest.
#'   - `alpha` Prior probability of splitting for a tree of depth 0 in the mean model. Tree split prior combines `alpha` and `beta` via `alpha*(1+node_depth)^-beta`. Default: `0.95`.
#'   - `beta` Exponent that decreases split probabilities for nodes of depth > 0 in the mean model. Tree split prior combines `alpha` and `beta` via `alpha*(1+node_depth)^-beta`. Default: `2`.
#'   - `min_samples_leaf` Minimum allowable size of a leaf, in terms of training samples, in the mean model. Default: `5`.
#'   - `max_depth` Maximum depth of any tree in the ensemble in the mean model. Default: `10`. Can be overridden with ``-1`` which does not enforce any depth limits on trees.
#'   - `sample_sigma2_leaf` Whether or not to update the leaf scale variance parameter based on `IG(sigma2_leaf_shape, sigma2_leaf_scale)`. Cannot (currently) be set to true if `ncol(leaf_basis_train)>1`. Default: `FALSE`.
#'   - `sigma2_leaf_init` Starting value of leaf node scale parameter. Calibrated internally as `1/num_trees` if not set here.
#'   - `sigma2_leaf_shape` Shape parameter in the `IG(sigma2_leaf_shape, sigma2_leaf_scale)` leaf node parameter variance model. Default: `3`.
#'   - `sigma2_leaf_scale` Scale parameter in the `IG(sigma2_leaf_shape, sigma2_leaf_scale)` leaf node parameter variance model. Calibrated internally as `0.5/num_trees` if not set here.
#'   - `keep_vars` Vector of variable names or column indices denoting variables that should be included in the forest. Default: `NULL`.
#'   - `drop_vars` Vector of variable names or column indices denoting variables that should be excluded from the forest. Default: `NULL`. If both `drop_vars` and `keep_vars` are set, `drop_vars` will be ignored.
#'   - `num_features_subsample` How many features to subsample when growing each tree for the GFR algorithm. Defaults to the number of features in the training dataset.
#'
#' @param variance_forest_params (Optional) A list of variance forest model parameters, each of which has a default value processed internally, so this argument list is optional.
#'
#'   - `num_trees` Number of trees in the ensemble for the conditional variance model. Default: `0`. Variance is only modeled using a tree / forest if `num_trees > 0`.
#'   - `alpha` Prior probability of splitting for a tree of depth 0 in the variance model. Tree split prior combines `alpha` and `beta` via `alpha*(1+node_depth)^-beta`. Default: `0.95`.
#'   - `beta` Exponent that decreases split probabilities for nodes of depth > 0 in the variance model. Tree split prior combines `alpha` and `beta` via `alpha*(1+node_depth)^-beta`. Default: `2`.
#'   - `min_samples_leaf` Minimum allowable size of a leaf, in terms of training samples, in the variance model. Default: `5`.
#'   - `max_depth` Maximum depth of any tree in the ensemble in the variance model. Default: `10`. Can be overridden with ``-1`` which does not enforce any depth limits on trees.
#'   - `leaf_prior_calibration_param` Hyperparameter used to calibrate the `IG(var_forest_prior_shape, var_forest_prior_scale)` conditional error variance model. If `var_forest_prior_shape` and `var_forest_prior_scale` are not set below, this calibration parameter is used to set these values to `num_trees / leaf_prior_calibration_param^2 + 0.5` and `num_trees / leaf_prior_calibration_param^2`, respectively. Default: `1.5`.
#'   - `var_forest_leaf_init` Starting value of root forest prediction in conditional (heteroskedastic) error variance model. Calibrated internally as `log(0.6*var(y_train))/num_trees`, where `y_train` is the possibly standardized outcome, if not set.
#'   - `var_forest_prior_shape` Shape parameter in the `IG(var_forest_prior_shape, var_forest_prior_scale)` conditional error variance model (which is only sampled if `num_trees > 0`). Calibrated internally as `num_trees / leaf_prior_calibration_param^2 + 0.5` if not set.
#'   - `var_forest_prior_scale` Scale parameter in the `IG(var_forest_prior_shape, var_forest_prior_scale)` conditional error variance model (which is only sampled if `num_trees > 0`). Calibrated internally as `num_trees / leaf_prior_calibration_param^2` if not set.
#'   - `keep_vars` Vector of variable names or column indices denoting variables that should be included in the forest. Default: `NULL`.
#'   - `drop_vars` Vector of variable names or column indices denoting variables that should be excluded from the forest. Default: `NULL`. If both `drop_vars` and `keep_vars` are set, `drop_vars` will be ignored.
#'   - `num_features_subsample` How many features to subsample when growing each tree for the GFR algorithm. Defaults to the number of features in the training dataset.
#'
#' @param random_effects_params (Optional) A list of random effects model parameters, each of which has a default value processed internally, so this argument list is optional.
#'
#'   - `model_spec` Specification of the random effects model. Options are "custom" and "intercept_only". If "custom" is specified, then a user-provided basis must be passed through `rfx_basis_train`. If "intercept_only" is specified, a random effects basis of all ones will be dispatched internally at sampling and prediction time. If "intercept_plus_treatment" is specified, a random effects basis that combines an "intercept" basis of all ones with the treatment variable (`Z_train`) will be dispatched internally at sampling and prediction time. Default: "custom". If "intercept_only" is specified, `rfx_basis_train` and `rfx_basis_test` (if provided) will be ignored.
#'   - `working_parameter_prior_mean` Prior mean for the random effects "working parameter". Default: `NULL`. Must be a vector whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a vector.
#'   - `group_parameters_prior_mean` Prior mean for the random effects "group parameters." Default: `NULL`. Must be a vector whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a vector.
#'   - `working_parameter_prior_cov` Prior covariance matrix for the random effects "working parameter." Default: `NULL`. Must be a square matrix whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a diagonal matrix.
#'   - `group_parameter_prior_cov` Prior covariance matrix for the random effects "group parameters." Default: `NULL`. Must be a square matrix whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a diagonal matrix.
#'   - `variance_prior_shape` Shape parameter for the inverse gamma prior on the variance of the random effects "group parameter." Default: `1`.
#'   - `variance_prior_scale` Scale parameter for the inverse gamma prior on the variance of the random effects "group parameter." Default: `1`.
#'
#' @return List of sampling outputs and a wrapper around the sampled forests (which can be used for in-memory prediction on new data, or serialized to JSON on disk).
#' @export
#'
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' f_XW <- (
#'     ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
#'     ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
#'     ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
#'     ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
#' )
#' noise_sd <- 1
#' y <- f_XW + rnorm(n, 0, noise_sd)
#' test_set_pct <- 0.2
#' n_test <- round(test_set_pct*n)
#' n_train <- n - n_test
#' test_inds <- sort(sample(1:n, n_test, replace = FALSE))
#' train_inds <- (1:n)[!((1:n) %in% test_inds)]
#' X_test <- X[test_inds,]
#' X_train <- X[train_inds,]
#' y_test <- y[test_inds]
#' y_train <- y[train_inds]
#'
#' bart_model <- bart(X_train = X_train, y_train = y_train, X_test = X_test,
#'                    num_gfr = 10, num_burnin = 0, num_mcmc = 10)
bart <- function(
  X_train,
  y_train,
  leaf_basis_train = NULL,
  rfx_group_ids_train = NULL,
  rfx_basis_train = NULL,
  X_test = NULL,
  leaf_basis_test = NULL,
  rfx_group_ids_test = NULL,
  rfx_basis_test = NULL,
  observation_weights = NULL,
  num_gfr = 5,
  num_burnin = 0,
  num_mcmc = 100,
  previous_model_json = NULL,
  previous_model_warmstart_sample_num = NULL,
  general_params = list(),
  mean_forest_params = list(),
  variance_forest_params = list(),
  random_effects_params = list()
) {
  # Update general BART parameters
  general_params_default <- list(
    cutpoint_grid_size = 100,
    standardize = TRUE,
    sample_sigma2_global = TRUE,
    sigma2_global_init = NULL,
    sigma2_global_shape = 0,
    sigma2_global_scale = 0,
    variable_weights = NULL,
    random_seed = -1,
    keep_burnin = FALSE,
    keep_gfr = FALSE,
    keep_every = 1,
    num_chains = 1,
    verbose = FALSE,
    outcome_model = OutcomeModel(outcome = "continuous", link = "identity"),
    probit_outcome_model = FALSE,
    num_threads = -1
  )
  general_params_updated <- preprocessParams(
    general_params_default,
    general_params
  )
  # TODO: think about validation and deprecation flow for probit_outcome_model

  # Update mean forest BART parameters
  mean_forest_params_default <- list(
    num_trees = 200,
    alpha = 0.95,
    beta = 2.0,
    min_samples_leaf = 5,
    max_depth = 10,
    sample_sigma2_leaf = TRUE,
    sigma2_leaf_init = NULL,
    sigma2_leaf_shape = 3,
    sigma2_leaf_scale = NULL,
    keep_vars = NULL,
    drop_vars = NULL,
    num_features_subsample = NULL,
    cloglog_leaf_prior_shape = 2.0,
    cloglog_leaf_prior_scale = 2.0
  )
  mean_forest_params_updated <- preprocessParams(
    mean_forest_params_default,
    mean_forest_params
  )

  # Update variance forest BART parameters
  variance_forest_params_default <- list(
    num_trees = 0,
    alpha = 0.95,
    beta = 2.0,
    min_samples_leaf = 5,
    max_depth = 10,
    leaf_prior_calibration_param = 1.5,
    var_forest_leaf_init = NULL,
    var_forest_prior_shape = NULL,
    var_forest_prior_scale = NULL,
    keep_vars = NULL,
    drop_vars = NULL,
    num_features_subsample = NULL
  )
  variance_forest_params_updated <- preprocessParams(
    variance_forest_params_default,
    variance_forest_params
  )

  # Update rfx parameters
  rfx_params_default <- list(
    model_spec = "custom",
    working_parameter_prior_mean = NULL,
    group_parameter_prior_mean = NULL,
    working_parameter_prior_cov = NULL,
    group_parameter_prior_cov = NULL,
    variance_prior_shape = 1,
    variance_prior_scale = 1
  )
  rfx_params_updated <- preprocessParams(
    rfx_params_default,
    random_effects_params
  )

  ### Unpack all parameter values
  # 1. General parameters
  cutpoint_grid_size <- general_params_updated$cutpoint_grid_size
  standardize <- general_params_updated$standardize
  sample_sigma2_global <- general_params_updated$sample_sigma2_global
  sigma2_init <- general_params_updated$sigma2_global_init
  a_global <- general_params_updated$sigma2_global_shape
  b_global <- general_params_updated$sigma2_global_scale
  variable_weights <- general_params_updated$variable_weights
  random_seed <- general_params_updated$random_seed
  keep_burnin <- general_params_updated$keep_burnin
  keep_gfr <- general_params_updated$keep_gfr
  keep_every <- general_params_updated$keep_every
  num_chains <- general_params_updated$num_chains
  verbose <- general_params_updated$verbose
  outcome_model <- general_params_updated$outcome_model
  probit_outcome_model <- general_params_updated$probit_outcome_model
  num_threads <- general_params_updated$num_threads

  # 2. Mean forest parameters
  num_trees_mean <- mean_forest_params_updated$num_trees
  alpha_mean <- mean_forest_params_updated$alpha
  beta_mean <- mean_forest_params_updated$beta
  min_samples_leaf_mean <- mean_forest_params_updated$min_samples_leaf
  max_depth_mean <- mean_forest_params_updated$max_depth
  sample_sigma2_leaf <- mean_forest_params_updated$sample_sigma2_leaf
  sigma2_leaf_init <- mean_forest_params_updated$sigma2_leaf_init
  a_leaf <- mean_forest_params_updated$sigma2_leaf_shape
  b_leaf <- mean_forest_params_updated$sigma2_leaf_scale
  keep_vars_mean <- mean_forest_params_updated$keep_vars
  drop_vars_mean <- mean_forest_params_updated$drop_vars
  num_features_subsample_mean <- mean_forest_params_updated$num_features_subsample
  cloglog_leaf_prior_shape <- mean_forest_params_updated$cloglog_leaf_prior_shape
  cloglog_leaf_prior_scale <- mean_forest_params_updated$cloglog_leaf_prior_scale

  # 3. Variance forest parameters
  num_trees_variance <- variance_forest_params_updated$num_trees
  alpha_variance <- variance_forest_params_updated$alpha
  beta_variance <- variance_forest_params_updated$beta
  min_samples_leaf_variance <- variance_forest_params_updated$min_samples_leaf
  max_depth_variance <- variance_forest_params_updated$max_depth
  a_0 <- variance_forest_params_updated$leaf_prior_calibration_param
  variance_forest_init <- variance_forest_params_updated$var_forest_leaf_init
  a_forest <- variance_forest_params_updated$var_forest_prior_shape
  b_forest <- variance_forest_params_updated$var_forest_prior_scale
  keep_vars_variance <- variance_forest_params_updated$keep_vars
  drop_vars_variance <- variance_forest_params_updated$drop_vars
  num_features_subsample_variance <- variance_forest_params_updated$num_features_subsample

  # 4. RFX parameters
  rfx_model_spec <- rfx_params_updated$model_spec
  rfx_working_parameter_prior_mean <- rfx_params_updated$working_parameter_prior_mean
  rfx_group_parameter_prior_mean <- rfx_params_updated$group_parameter_prior_mean
  rfx_working_parameter_prior_cov <- rfx_params_updated$working_parameter_prior_cov
  rfx_group_parameter_prior_cov <- rfx_params_updated$group_parameter_prior_cov
  rfx_variance_prior_shape <- rfx_params_updated$variance_prior_shape
  rfx_variance_prior_scale <- rfx_params_updated$variance_prior_scale

  # Raise a deprecation warning to use `outcome_model` if `probit_outcome_model = TRUE` is specified
  if (probit_outcome_model) {
    warning(
      "Specifying a probit link through `general_params = list(probit_outcome_model = TRUE)` is deprecated and will be removed in a future version. Please use `general_params = list(outcome_model = OutcomeModel(outcome = 'binary', link = 'probit'))` instead."
    )
  }

  # Unpack outcome model details
  link_is_linear <- FALSE
  link_is_probit <- FALSE
  link_is_cloglog <- FALSE
  outcome_is_continuous <- FALSE
  outcome_is_binary <- FALSE
  outcome_is_ordinal <- FALSE
  if (
    outcome_model$outcome == "continuous" && outcome_model$link == "identity"
  ) {
    link_is_linear <- TRUE
    outcome_is_continuous <- TRUE
  } else if (
    outcome_model$outcome == "binary" && outcome_model$link == "probit"
  ) {
    link_is_probit <- TRUE
    outcome_is_binary <- TRUE
  } else if (
    outcome_model$outcome == "binary" && outcome_model$link == "cloglog"
  ) {
    link_is_cloglog <- TRUE
    outcome_is_binary <- TRUE
  } else if (
    outcome_model$outcome == "ordinal" && outcome_model$link == "cloglog"
  ) {
    link_is_cloglog <- TRUE
    outcome_is_ordinal <- TRUE
  } else {
    stop(paste0(
      "Invalid outcome model specification, outcome = ",
      outcome_model$outcome,
      ", link = ",
      outcome_model$link
    ))
  }

  # Set a function-scoped RNG if user provided a random seed
  custom_rng <- random_seed >= 0
  has_existing_random_seed <- F
  if (custom_rng) {
    # Cache original global environment RNG state (if it exists)
    if (exists(".Random.seed", envir = .GlobalEnv)) {
      original_global_seed <- .Random.seed
      has_existing_random_seed <- T
    }
    # Set new seed and store associated RNG state
    set.seed(random_seed)
  }

  # Check if there are enough GFR samples to seed num_chains samplers
  if (num_gfr > 0) {
    if (num_chains > num_gfr) {
      stop(
        "num_chains > num_gfr, meaning we do not have enough GFR samples to seed num_chains distinct MCMC chains"
      )
    }
  }

  # Override keep_gfr if there are no MCMC samples
  if (num_mcmc == 0) {
    keep_gfr <- TRUE
  }

  # Check if previous model JSON is provided and parse it if so
  has_prev_model <- !is.null(previous_model_json)
  has_prev_model_index <- !is.null(previous_model_warmstart_sample_num)
  if (has_prev_model) {
    previous_bart_model <- createBARTModelFromJsonString(
      previous_model_json
    )
    prev_num_samples <- previous_bart_model$model_params$num_samples
    if (!has_prev_model_index) {
      previous_model_warmstart_sample_num <- prev_num_samples
      warning(
        "`previous_model_warmstart_sample_num` was not provided alongside `previous_model_json`, so it will be set to the number of samples available in `previous_model_json`"
      )
    } else {
      if (previous_model_warmstart_sample_num < 1) {
        stop(
          "`previous_model_warmstart_sample_num` must be a positive integer"
        )
      }
      if (previous_model_warmstart_sample_num > prev_num_samples) {
        stop(
          "`previous_model_warmstart_sample_num` exceeds the number of samples in `previous_model_json`"
        )
      }
    }
    previous_model_decrement <- T
    if (num_chains > previous_model_warmstart_sample_num) {
      warning(
        "The number of chains being sampled exceeds the number of previous model samples available from the requested position in `previous_model_json`. All chains will be initialized from the same sample."
      )
      previous_model_decrement <- F
    }
    previous_y_bar <- previous_bart_model$model_params$outcome_mean
    previous_y_scale <- previous_bart_model$model_params$outcome_scale
    if (previous_bart_model$model_params$include_mean_forest) {
      previous_forest_samples_mean <- previous_bart_model$mean_forests
    } else {
      previous_forest_samples_mean <- NULL
    }
    if (previous_bart_model$model_params$include_variance_forest) {
      previous_forest_samples_variance <- previous_bart_model$variance_forests
    } else {
      previous_forest_samples_variance <- NULL
    }
    if (previous_bart_model$model_params$sample_sigma2_global) {
      previous_global_var_samples <- previous_bart_model$sigma2_global_samples /
        (previous_y_scale * previous_y_scale)
    } else {
      previous_global_var_samples <- NULL
    }
    if (previous_bart_model$model_params$sample_sigma2_leaf) {
      previous_leaf_var_samples <- previous_bart_model$sigma2_leaf_samples
    } else {
      previous_leaf_var_samples <- NULL
    }
    if (previous_bart_model$model_params$has_rfx) {
      previous_rfx_samples <- previous_bart_model$rfx_samples
    } else {
      previous_rfx_samples <- NULL
    }
    if (previous_bart_model$model_params$outcome_model$link == "cloglog") {
      previous_cloglog_cutpoint_samples <- previous_bart_model$cloglog_cutpoint_samples
      previous_cloglog_num_categories <- previous_bart_model$cloglog_num_categories
    } else {
      previous_cloglog_cutpoint_samples <- NULL
      previous_cloglog_num_categories <- 0
    }
    previous_model_num_samples <- previous_bart_model$model_params$num_samples
    if (previous_model_warmstart_sample_num > previous_model_num_samples) {
      stop(
        "`previous_model_warmstart_sample_num` exceeds the number of samples in `previous_model_json`"
      )
    }
  } else {
    previous_y_bar <- NULL
    previous_y_scale <- NULL
    previous_global_var_samples <- NULL
    previous_leaf_var_samples <- NULL
    previous_rfx_samples <- NULL
    previous_forest_samples_mean <- NULL
    previous_forest_samples_variance <- NULL
    previous_cloglog_cutpoint_samples <- NULL
    previous_cloglog_num_categories <- 0
    previous_model_num_samples <- 0
  }

  # Determine whether conditional mean, variance, or both will be modeled
  if (num_trees_variance > 0) {
    include_variance_forest = TRUE
  } else {
    include_variance_forest = FALSE
  }
  if (num_trees_mean > 0) {
    include_mean_forest = TRUE
  } else {
    include_mean_forest = FALSE
  }

  # observation_weights compatibility checks
  if (!is.null(observation_weights)) {
    if (link_is_cloglog) {
      stop(
        "observation_weights are not compatible with cloglog link functions."
      )
    }
    if (include_variance_forest) {
      warning(
        "Results may be unreliable when observation_weights are deployed alongside a variance forest model."
      )
    }
  }

  # Set the variance forest priors if not set
  if (include_variance_forest) {
    if (is.null(a_forest)) {
      a_forest <- num_trees_variance / (a_0^2) + 0.5
    }
    if (is.null(b_forest)) b_forest <- num_trees_variance / (a_0^2)
  } else {
    a_forest <- 1.
    b_forest <- 1.
  }

  # Override tau sampling if there is no mean forest
  if (!include_mean_forest) {
    sample_sigma2_leaf <- FALSE
  }

  # Variable weight preprocessing (and initialization if necessary)
  if (is.null(variable_weights)) {
    variable_weights = rep(1 / ncol(X_train), ncol(X_train))
  }
  if (any(variable_weights < 0)) {
    stop("variable_weights cannot have any negative weights")
  }

  # Observation weight validation
  if (!is.null(observation_weights)) {
    if (!is.numeric(observation_weights)) {
      stop("observation_weights must be a numeric vector")
    }
    if (length(observation_weights) != nrow(X_train)) {
      stop("length(observation_weights) must equal nrow(X_train)")
    }
    if (any(observation_weights < 0)) {
      stop("observation_weights cannot have any negative values")
    }
    if (all(observation_weights == 0) && num_gfr > 0) {
      stop(
        "observation_weights are all zero (prior sampling mode) but num_gfr > 0. ",
        "GFR warm-start is data-dependent and ill-defined with zero weights. ",
        "Set num_gfr = 0 when using all-zero observation_weights."
      )
    }
  }

  # Check covariates are matrix or dataframe
  if ((!is.data.frame(X_train)) && (!is.matrix(X_train))) {
    stop("X_train must be a matrix or dataframe")
  }
  if (!is.null(X_test)) {
    if ((!is.data.frame(X_test)) && (!is.matrix(X_test))) {
      stop("X_test must be a matrix or dataframe")
    }
  }
  num_cov_orig <- ncol(X_train)

  # Raise a warning if the data have ties and only GFR is being run
  if ((num_gfr > 0) && (num_mcmc == 0) && (num_burnin == 0)) {
    num_values <- nrow(X_train)
    max_grid_size <- ifelse(
      num_values > cutpoint_grid_size,
      floor(num_values / cutpoint_grid_size),
      1
    )
    x_is_df <- is.data.frame(X_train)
    covs_warning_1 <- NULL
    covs_warning_2 <- NULL
    covs_warning_3 <- NULL
    covs_warning_4 <- NULL
    for (i in 1:num_cov_orig) {
      # Skip check for variables that are treated as categorical
      x_numeric <- T
      if (x_is_df) {
        if (is.factor(X_train[, i])) {
          x_numeric <- F
        }
      }
      if (x_numeric) {
        # Determine the number of unique values
        num_unique_values <- length(unique(X_train[, i]))

        # Determine a "name" for the covariate
        cov_name <- ifelse(
          is.null(colnames(X_train)),
          paste0("X", i),
          colnames(X_train)[i]
        )

        # Check for a small relative number of unique values
        unique_full_ratio <- num_unique_values / num_values
        if (unique_full_ratio < 0.2) {
          covs_warning_1 <- c(covs_warning_1, cov_name)
        }

        # Check for a small absolute number of unique values
        if (num_values > 100) {
          if (num_unique_values < 20) {
            covs_warning_2 <- c(covs_warning_2, cov_name)
          }
        }

        # Check for a large number of duplicates of any individual value
        x_j_hist <- table(X_train[, i])
        if (any(x_j_hist > 2 * max_grid_size)) {
          covs_warning_3 <- c(covs_warning_3, cov_name)
        }

        # Check for binary variables
        if (num_unique_values == 2) {
          covs_warning_4 <- c(covs_warning_4, cov_name)
        }
      }
    }

    if (!is.null(covs_warning_1)) {
      warning(
        paste0(
          "Covariate(s) ",
          paste(covs_warning_1, collapse = ", "),
          " have a ratio of unique to overall observations of less than 0.2. ",
          "This might present some issues with the grow-from-root (GFR) algorithm. ",
          "Consider running with `num_mcmc > 0` and `num_burnin > 0` to improve your model's performance."
        )
      )
    }

    if (!is.null(covs_warning_2)) {
      warning(
        paste0(
          "Covariate(s) ",
          paste(covs_warning_2, collapse = ", "),
          " have fewer than 20 unique values. ",
          "This might present some issues with the grow-from-root (GFR) algorithm. ",
          "Consider running with `num_mcmc > 0` and `num_burnin > 0` to improve your model's performance."
        )
      )
    }

    if (!is.null(covs_warning_3)) {
      warning(
        paste0(
          "Covariates ",
          paste(covs_warning_3, collapse = ", "),
          " have some observed values with more than ",
          2 * max_grid_size,
          " repeated observations. ",
          "This might present some issues with the grow-from-root (GFR) algorithm. ",
          "Consider running with `num_mcmc > 0` and `num_burnin > 0` to improve your model's performance."
        )
      )
    }

    if (!is.null(covs_warning_4)) {
      warning(
        paste0(
          "Covariates ",
          paste(covs_warning_4, collapse = ", "),
          " appear to be binary but are currently treated by stochtree as continuous. ",
          "This might present some issues with the grow-from-root (GFR) algorithm. ",
          "Consider converting binary variables to ordered factor (i.e. `factor(..., ordered = T)`."
        )
      )
    }
  }

  # Standardize the keep variable lists to numeric indices
  if (!is.null(keep_vars_mean)) {
    if (is.character(keep_vars_mean)) {
      if (!all(keep_vars_mean %in% names(X_train))) {
        stop(
          "keep_vars_mean includes some variable names that are not in X_train"
        )
      }
      variable_subset_mean <- unname(which(
        names(X_train) %in% keep_vars_mean
      ))
    } else {
      if (any(keep_vars_mean > ncol(X_train))) {
        stop(
          "keep_vars_mean includes some variable indices that exceed the number of columns in X_train"
        )
      }
      if (any(keep_vars_mean < 0)) {
        stop("keep_vars_mean includes some negative variable indices")
      }
      variable_subset_mean <- keep_vars_mean
    }
  } else if ((is.null(keep_vars_mean)) && (!is.null(drop_vars_mean))) {
    if (is.character(drop_vars_mean)) {
      if (!all(drop_vars_mean %in% names(X_train))) {
        stop(
          "drop_vars_mean includes some variable names that are not in X_train"
        )
      }
      variable_subset_mean <- unname(which(
        !(names(X_train) %in% drop_vars_mean)
      ))
    } else {
      if (any(drop_vars_mean > ncol(X_train))) {
        stop(
          "drop_vars_mean includes some variable indices that exceed the number of columns in X_train"
        )
      }
      if (any(drop_vars_mean < 0)) {
        stop("drop_vars_mean includes some negative variable indices")
      }
      variable_subset_mean <- (1:ncol(X_train))[
        !(1:ncol(X_train) %in% drop_vars_mean)
      ]
    }
  } else {
    variable_subset_mean <- 1:ncol(X_train)
  }
  if (!is.null(keep_vars_variance)) {
    if (is.character(keep_vars_variance)) {
      if (!all(keep_vars_variance %in% names(X_train))) {
        stop(
          "keep_vars_variance includes some variable names that are not in X_train"
        )
      }
      variable_subset_variance <- unname(which(
        names(X_train) %in% keep_vars_variance
      ))
    } else {
      if (any(keep_vars_variance > ncol(X_train))) {
        stop(
          "keep_vars_variance includes some variable indices that exceed the number of columns in X_train"
        )
      }
      if (any(keep_vars_variance < 0)) {
        stop(
          "keep_vars_variance includes some negative variable indices"
        )
      }
      variable_subset_variance <- keep_vars_variance
    }
  } else if ((is.null(keep_vars_variance)) && (!is.null(drop_vars_variance))) {
    if (is.character(drop_vars_variance)) {
      if (!all(drop_vars_variance %in% names(X_train))) {
        stop(
          "drop_vars_variance includes some variable names that are not in X_train"
        )
      }
      variable_subset_variance <- unname(which(
        !(names(X_train) %in% drop_vars_variance)
      ))
    } else {
      if (any(drop_vars_variance > ncol(X_train))) {
        stop(
          "drop_vars_variance includes some variable indices that exceed the number of columns in X_train"
        )
      }
      if (any(drop_vars_variance < 0)) {
        stop(
          "drop_vars_variance includes some negative variable indices"
        )
      }
      variable_subset_variance <- (1:ncol(X_train))[
        !(1:ncol(X_train) %in% drop_vars_variance)
      ]
    }
  } else {
    variable_subset_variance <- 1:ncol(X_train)
  }

  # Preprocess covariates
  if ((!is.data.frame(X_train)) && (!is.matrix(X_train))) {
    stop("X_train must be a matrix or dataframe")
  }
  if (!is.null(X_test)) {
    if ((!is.data.frame(X_test)) && (!is.matrix(X_test))) {
      stop("X_test must be a matrix or dataframe")
    }
  }
  if (ncol(X_train) != length(variable_weights)) {
    stop("length(variable_weights) must equal ncol(X_train)")
  }
  train_cov_preprocess_list <- preprocessTrainData(X_train)
  X_train_metadata <- train_cov_preprocess_list$metadata
  X_train <- train_cov_preprocess_list$data
  original_var_indices <- X_train_metadata$original_var_indices
  feature_types <- X_train_metadata$feature_types
  if (!is.null(X_test)) {
    X_test <- preprocessPredictionData(X_test, X_train_metadata)
  }

  # Update variable weights
  variable_weights_mean <- variable_weights_variance <- variable_weights
  variable_weights_adj <- 1 /
    sapply(original_var_indices, function(x) sum(original_var_indices == x))
  if (include_mean_forest) {
    variable_weights_mean <- variable_weights_mean[original_var_indices] *
      variable_weights_adj
    variable_weights_mean[
      !(original_var_indices %in% variable_subset_mean)
    ] <- 0
  }
  if (include_variance_forest) {
    variable_weights_variance <- variable_weights_variance[
      original_var_indices
    ] *
      variable_weights_adj
    variable_weights_variance[
      !(original_var_indices %in% variable_subset_variance)
    ] <- 0
  }

  # Set num_features_subsample to default, ncol(X_train), if not already set
  if (is.null(num_features_subsample_mean)) {
    num_features_subsample_mean <- ncol(X_train)
  }
  if (is.null(num_features_subsample_variance)) {
    num_features_subsample_variance <- ncol(X_train)
  }

  # Convert all input data to matrices if not already converted
  if ((is.null(dim(leaf_basis_train))) && (!is.null(leaf_basis_train))) {
    leaf_basis_train <- as.matrix(leaf_basis_train)
  }
  if ((is.null(dim(leaf_basis_test))) && (!is.null(leaf_basis_test))) {
    leaf_basis_test <- as.matrix(leaf_basis_test)
  }
  if ((is.null(dim(rfx_basis_train))) && (!is.null(rfx_basis_train))) {
    rfx_basis_train <- as.matrix(rfx_basis_train)
  }
  if ((is.null(dim(rfx_basis_test))) && (!is.null(rfx_basis_test))) {
    rfx_basis_test <- as.matrix(rfx_basis_test)
  }

  # Recode group IDs to integer vector (if passed as, for example, a vector of county names, etc...)
  has_rfx <- FALSE
  has_rfx_test <- FALSE
  if (!is.null(rfx_group_ids_train)) {
    group_ids_factor <- factor(rfx_group_ids_train)
    rfx_group_ids_train <- as.integer(group_ids_factor)
    has_rfx <- TRUE
    if (!is.null(rfx_group_ids_test)) {
      group_ids_factor_test <- factor(
        rfx_group_ids_test,
        levels = levels(group_ids_factor)
      )
      if (sum(is.na(group_ids_factor_test)) > 0) {
        stop(
          "All random effect group labels provided in rfx_group_ids_test must be present in rfx_group_ids_train"
        )
      }
      rfx_group_ids_test <- as.integer(group_ids_factor_test)
      has_rfx_test <- TRUE
    }
  }

  # Data consistency checks
  if ((!is.null(X_test)) && (ncol(X_test) != ncol(X_train))) {
    stop("X_train and X_test must have the same number of columns")
  }
  if (
    (!is.null(leaf_basis_test)) &&
      (ncol(leaf_basis_test) != ncol(leaf_basis_train))
  ) {
    stop(
      "leaf_basis_train and leaf_basis_test must have the same number of columns"
    )
  }
  if (
    (!is.null(leaf_basis_train)) &&
      (nrow(leaf_basis_train) != nrow(X_train))
  ) {
    stop("leaf_basis_train and X_train must have the same number of rows")
  }
  if ((!is.null(leaf_basis_test)) && (nrow(leaf_basis_test) != nrow(X_test))) {
    stop("leaf_basis_test and X_test must have the same number of rows")
  }
  if (nrow(X_train) != length(y_train)) {
    stop("X_train and y_train must have the same number of observations")
  }
  if (
    (!is.null(rfx_basis_test)) &&
      (ncol(rfx_basis_test) != ncol(rfx_basis_train))
  ) {
    stop(
      "rfx_basis_train and rfx_basis_test must have the same number of columns"
    )
  }
  if (!is.null(rfx_group_ids_train)) {
    if (!is.null(rfx_group_ids_test)) {
      if ((!is.null(rfx_basis_train)) && (is.null(rfx_basis_test))) {
        stop(
          "rfx_basis_train is provided but rfx_basis_test is not provided"
        )
      }
    }
  }

  # Handle the rfx basis matrices
  has_basis_rfx <- FALSE
  num_basis_rfx <- 0
  if (has_rfx) {
    if (rfx_model_spec == "custom") {
      if (is.null(rfx_basis_train)) {
        stop(
          "A user-provided basis (`rfx_basis_train`) must be provided when the random effects model spec is 'custom'"
        )
      }
      has_basis_rfx <- TRUE
      num_basis_rfx <- ncol(rfx_basis_train)
    } else if (rfx_model_spec == "intercept_only") {
      rfx_basis_train <- matrix(
        rep(1, nrow(X_train)),
        nrow = nrow(X_train),
        ncol = 1
      )
      has_basis_rfx <- TRUE
      num_basis_rfx <- 1
    }
    num_rfx_groups <- length(unique(rfx_group_ids_train))
    num_rfx_components <- ncol(rfx_basis_train)
    if (num_rfx_groups == 1) {
      warning(
        "Only one group was provided for random effect sampling, so the random effects model is likely overkill"
      )
    }
  }
  if (has_rfx_test) {
    if (rfx_model_spec == "custom") {
      if (is.null(rfx_basis_test)) {
        stop(
          "A user-provided basis (`rfx_basis_test`) must be provided when the random effects model spec is 'custom'"
        )
      }
    } else if (rfx_model_spec == "intercept_only") {
      rfx_basis_test <- matrix(
        rep(1, nrow(X_test)),
        nrow = nrow(X_test),
        ncol = 1
      )
    }
  }

  # Convert y_train to numeric vector if not already converted
  if (!is.null(dim(y_train))) {
    y_train <- as.matrix(y_train)
  }

  # Determine whether a basis vector is provided
  has_basis = !is.null(leaf_basis_train)

  # Determine whether a test set is provided
  has_test = !is.null(X_test)

  # Preliminary runtime checks for probit link
  if (!include_mean_forest) {
    link_is_probit <- FALSE
    # TODO: think about allowing binary models with probit link for homoskedastic RFX-only models?
  }
  if (link_is_probit) {
    if (!(length(unique(y_train)) == 2)) {
      stop(
        "You specified a probit link, but supplied an outcome with more than 2 unique values. Probit is only currently supported for binary outcomes."
      )
    }
    unique_outcomes <- sort(unique(y_train))
    if (!(all(unique_outcomes == c(0, 1)))) {
      stop(
        "You specified a probit link, but supplied an outcome with 2 unique values other than 0 and 1"
      )
    }
    if (include_variance_forest) {
      stop("We do not support heteroskedasticity with a probit link")
    }
    if (sample_sigma2_global) {
      warning(
        "Global error variance will not be sampled with a probit link as it is fixed at 1"
      )
      sample_sigma2_global <- F
    }
  }

  # Preliminary runtime checks for cloglog link
  if (!include_mean_forest) {
    link_is_cloglog <- FALSE
    # TODO: think about allowing binary models with cloglog link for homoskedastic RFX-only models?
  }
  if (link_is_cloglog) {
    if (!all(as.integer(y_train) == y_train)) {
      stop(
        "You specified a cloglog link, but supplied an outcome with non-integer values. Cloglog is only currently supported for integer outcomes."
      )
    }
    unique_outcomes <- sort(unique(y_train))
    if (!(min(unique_outcomes) %in% c(0, 1))) {
      stop(
        "You specified a cloglog link, but supplied an integer outcome that does not start with 0 or 1. Please remap / shift the outcomes so that the smallest category label is either 0 or 1."
      )
    }
    if (!all(diff(unique_outcomes) == 1)) {
      stop(
        "You specified a cloglog link, but supplied an integer outcome that is not a sequence of consecutive integers"
      )
    }
    if (include_variance_forest) {
      stop("We do not support heteroskedasticity with a cloglog link")
    }
    if (has_basis) {
      stop("We do not support leaf basis regression with a cloglog link")
    }
    if (sample_sigma2_global) {
      warning(
        "Global error variance will not be sampled with a cloglog link"
      )
      sample_sigma2_global <- F
    }
    if (sample_sigma2_leaf) {
      warning(
        "Leaf scale parameter will not be sampled with a cloglog link"
      )
      sample_sigma2_leaf <- F
    }
  }

  # Runtime checks for variance forest
  if (include_variance_forest) {
    if (sample_sigma2_global) {
      warning(
        "Global error variance will not be sampled with a heteroskedasticity forest"
      )
      sample_sigma2_global <- F
    }
  }

  # Handle standardization, prior calibration, and initialization of forest
  # differently for binary and continuous outcomes
  if (link_is_probit) {
    # Probit-scale intercept: center the forest on the population-average latent mean.
    # The forest predicts mu(X) and y_bar_train is added back at prediction time.
    # The latent z sampling uses y_bar_train to set the correct truncated normal mean and to center z before the residual update.
    y_bar_train <- qnorm(mean_cpp(as.numeric(y_train)))
    y_std_train <- 1
    standardize <- FALSE

    # Set a pseudo outcome by subtracting mean_cpp(y_train) from y_train
    resid_train <- y_train - mean_cpp(as.numeric(y_train))

    # Set initial values of root nodes to 0.0 (in probit scale)
    init_val_mean <- 0.0

    # Calibrate priors for sigma^2 and tau
    # Set sigma2_init to 1, ignoring default provided
    sigma2_init <- 1.0
    # Skip variance_forest_init, since variance forests are not supported with probit link
    if (is.null(b_leaf)) {
      b_leaf <- 1 / (num_trees_mean)
    }
    if (has_basis) {
      if (ncol(leaf_basis_train) > 1) {
        if (is.null(sigma2_leaf_init)) {
          sigma2_leaf_init <- diag(
            2 / (num_trees_mean),
            ncol(leaf_basis_train)
          )
        }
        if (!is.matrix(sigma2_leaf_init)) {
          current_leaf_scale <- as.matrix(diag(
            sigma2_leaf_init,
            ncol(leaf_basis_train)
          ))
        } else {
          current_leaf_scale <- sigma2_leaf_init
        }
      } else {
        if (is.null(sigma2_leaf_init)) {
          sigma2_leaf_init <- as.matrix(2 / (num_trees_mean))
        }
        if (!is.matrix(sigma2_leaf_init)) {
          current_leaf_scale <- as.matrix(diag(sigma2_leaf_init, 1))
        } else {
          current_leaf_scale <- sigma2_leaf_init
        }
      }
    } else {
      if (is.null(sigma2_leaf_init)) {
        sigma2_leaf_init <- as.matrix(2 / (num_trees_mean))
      }
      if (!is.matrix(sigma2_leaf_init)) {
        current_leaf_scale <- as.matrix(diag(sigma2_leaf_init, 1))
      } else {
        current_leaf_scale <- sigma2_leaf_init
      }
    }
    current_sigma2 <- sigma2_init
  } else if (link_is_cloglog) {
    # Fix offset to 0 and scale to 1
    y_bar_train <- 0
    y_std_train <- 1
    standardize <- FALSE

    # Remap outcomes to start from 0
    resid_train <- as.numeric(y_train - min(unique_outcomes))
    cloglog_num_categories <- max(resid_train) + 1

    # Set initial values of root nodes to 0.0 (in linear scale)
    init_val_mean <- 0.0

    # Calibrate priors for sigma^2 and tau
    # Set sigma2_init to 1, ignoring default provided
    sigma2_init <- 1.0
    if (is.null(sigma2_leaf_init)) {
      sigma2_leaf_init <- as.matrix(2 / (num_trees_mean))
    }
    current_sigma2 <- sigma2_init
    current_leaf_scale <- sigma2_leaf_init

    # Set first cutpoint to 0 for identifiability
    cloglog_cutpoint_0 <- 0

    # Set shape and rate parameters for conditional gamma model
    cloglog_forest_shape <- 2.0
    cloglog_forest_rate <- 2.0
  } else {
    # Only standardize if user requested
    if (standardize) {
      y_bar_train <- mean_cpp(as.numeric(y_train))
      y_std_train <- sd_cpp(as.numeric(y_train))
    } else {
      y_bar_train <- 0
      y_std_train <- 1
    }

    # Compute standardized outcome
    resid_train <- (y_train - y_bar_train) / y_std_train

    # Compute initial value of root nodes in mean forest
    init_val_mean <- mean_cpp(as.numeric(resid_train))

    # Calibrate priors for sigma^2 and tau
    if (is.null(sigma2_init)) {
      sigma2_init <- 1.0 * var_cpp(as.numeric(resid_train))
    }
    if (is.null(variance_forest_init)) {
      variance_forest_init <- 1.0 * var_cpp(as.numeric(resid_train))
    }
    if (is.null(b_leaf)) {
      b_leaf <- var_cpp(as.numeric(resid_train)) / (2 * num_trees_mean)
    }
    if (has_basis) {
      if (ncol(leaf_basis_train) > 1) {
        if (is.null(sigma2_leaf_init)) {
          sigma2_leaf_init <- diag(
            2 * var_cpp(as.numeric(resid_train)) / (num_trees_mean),
            ncol(leaf_basis_train)
          )
        }
        if (!is.matrix(sigma2_leaf_init)) {
          current_leaf_scale <- as.matrix(diag(
            sigma2_leaf_init,
            ncol(leaf_basis_train)
          ))
        } else {
          current_leaf_scale <- sigma2_leaf_init
        }
      } else {
        if (is.null(sigma2_leaf_init)) {
          sigma2_leaf_init <- as.matrix(
            2 * var_cpp(as.numeric(resid_train)) / (num_trees_mean)
          )
        }
        if (!is.matrix(sigma2_leaf_init)) {
          current_leaf_scale <- as.matrix(diag(sigma2_leaf_init, 1))
        } else {
          current_leaf_scale <- sigma2_leaf_init
        }
      }
    } else {
      if (is.null(sigma2_leaf_init)) {
        sigma2_leaf_init <- as.matrix(
          2 * var_cpp(as.numeric(resid_train)) / (num_trees_mean)
        )
      }
      if (!is.matrix(sigma2_leaf_init)) {
        current_leaf_scale <- as.matrix(diag(sigma2_leaf_init, 1))
      } else {
        current_leaf_scale <- sigma2_leaf_init
      }
    }
    current_sigma2 <- sigma2_init
  }

  # Determine leaf model type
  if ((!has_basis) && (!link_is_cloglog)) {
    leaf_model_mean_forest <- 0
  } else if ((!has_basis) && (link_is_cloglog)) {
    leaf_model_mean_forest <- 4
  } else if (ncol(leaf_basis_train) == 1) {
    leaf_model_mean_forest <- 1
  } else if (ncol(leaf_basis_train) > 1) {
    leaf_model_mean_forest <- 2
  } else {
    stop("leaf_basis_train passed must be a matrix with at least 1 column")
  }

  # Set variance leaf model type (currently only one option)
  leaf_model_variance_forest <- 3

  # Unpack model type info
  if (leaf_model_mean_forest == 0) {
    leaf_dimension = 1
    is_leaf_constant = TRUE
    leaf_regression = FALSE
  } else if (leaf_model_mean_forest == 1) {
    stopifnot(has_basis)
    stopifnot(ncol(leaf_basis_train) == 1)
    leaf_dimension = 1
    is_leaf_constant = FALSE
    leaf_regression = TRUE
  } else if (leaf_model_mean_forest == 2) {
    stopifnot(has_basis)
    stopifnot(ncol(leaf_basis_train) > 1)
    leaf_dimension = ncol(leaf_basis_train)
    is_leaf_constant = FALSE
    leaf_regression = TRUE
    if (sample_sigma2_leaf) {
      warning(
        "Sampling leaf scale not yet supported for multivariate leaf models, so the leaf scale parameter will not be sampled in this model."
      )
      sample_sigma2_leaf <- FALSE
    }
  } else if (leaf_model_mean_forest == 4) {
    leaf_dimension = 1
    is_leaf_constant = TRUE
    leaf_regression = FALSE
  }

  # Data
  if (leaf_regression) {
    forest_dataset_train <- createForestDataset(
      X_train,
      leaf_basis_train,
      observation_weights
    )
    if (has_test) {
      forest_dataset_test <- createForestDataset(X_test, leaf_basis_test)
    }
    requires_basis <- TRUE
  } else {
    forest_dataset_train <- createForestDataset(
      X_train,
      variance_weights = observation_weights
    )
    if (has_test) {
      forest_dataset_test <- createForestDataset(X_test)
    }
    requires_basis <- FALSE
  }
  outcome_train <- createOutcome(resid_train)

  # Random number generator (std::mt19937)
  if (is.null(random_seed)) {
    random_seed = sample(1:10000, 1, FALSE)
  }
  rng <- createCppRNG(random_seed)

  # Separate ordinal sampler object for cloglog
  if (link_is_cloglog) {
    ordinal_sampler <- ordinal_sampler_cpp()
  }

  # Sampling data structures
  feature_types <- as.integer(feature_types)
  global_model_config <- createGlobalModelConfig(
    global_error_variance = current_sigma2
  )
  if (include_mean_forest) {
    forest_model_config_mean <- createForestModelConfig(
      feature_types = feature_types,
      num_trees = num_trees_mean,
      num_features = ncol(X_train),
      num_observations = nrow(X_train),
      variable_weights = variable_weights_mean,
      leaf_dimension = leaf_dimension,
      alpha = alpha_mean,
      beta = beta_mean,
      min_samples_leaf = min_samples_leaf_mean,
      max_depth = max_depth_mean,
      leaf_model_type = leaf_model_mean_forest,
      leaf_model_scale = current_leaf_scale,
      cutpoint_grid_size = cutpoint_grid_size,
      num_features_subsample = num_features_subsample_mean
    )
    if (link_is_cloglog) {
      forest_model_config_mean$update_cloglog_forest_shape(cloglog_forest_shape)
      forest_model_config_mean$update_cloglog_forest_rate(cloglog_forest_rate)
    }
    forest_model_mean <- createForestModel(
      forest_dataset_train,
      forest_model_config_mean,
      global_model_config
    )
  }
  if (include_variance_forest) {
    forest_model_config_variance <- createForestModelConfig(
      feature_types = feature_types,
      num_trees = num_trees_variance,
      num_features = ncol(X_train),
      num_observations = nrow(X_train),
      variable_weights = variable_weights_variance,
      leaf_dimension = 1,
      alpha = alpha_variance,
      beta = beta_variance,
      min_samples_leaf = min_samples_leaf_variance,
      max_depth = max_depth_variance,
      leaf_model_type = leaf_model_variance_forest,
      variance_forest_shape = a_forest,
      variance_forest_scale = b_forest,
      cutpoint_grid_size = cutpoint_grid_size,
      num_features_subsample = num_features_subsample_variance
    )
    forest_model_variance <- createForestModel(
      forest_dataset_train,
      forest_model_config_variance,
      global_model_config
    )
  }

  # Container of forest samples
  if (include_mean_forest) {
    forest_samples_mean <- createForestSamples(
      num_trees_mean,
      leaf_dimension,
      is_leaf_constant,
      FALSE
    )
    active_forest_mean <- createForest(
      num_trees_mean,
      leaf_dimension,
      is_leaf_constant,
      FALSE
    )
  }
  if (include_variance_forest) {
    forest_samples_variance <- createForestSamples(
      num_trees_variance,
      1,
      TRUE,
      TRUE
    )
    active_forest_variance <- createForest(
      num_trees_variance,
      1,
      TRUE,
      TRUE
    )
  }

  # Random effects initialization
  if (has_rfx) {
    # Prior parameters
    if (is.null(rfx_working_parameter_prior_mean)) {
      if (num_rfx_components == 1) {
        alpha_init <- c(0)
      } else if (num_rfx_components > 1) {
        alpha_init <- rep(0, num_rfx_components)
      } else {
        stop("There must be at least 1 random effect component")
      }
    } else {
      alpha_init <- expand_dims_1d(
        rfx_working_parameter_prior_mean,
        num_rfx_components
      )
    }

    if (is.null(rfx_group_parameter_prior_mean)) {
      xi_init <- matrix(
        rep(alpha_init, num_rfx_groups),
        num_rfx_components,
        num_rfx_groups
      )
    } else {
      xi_init <- expand_dims_2d(
        rfx_group_parameter_prior_mean,
        num_rfx_components,
        num_rfx_groups
      )
    }

    if (is.null(rfx_working_parameter_prior_cov)) {
      sigma_alpha_init <- diag(1, num_rfx_components, num_rfx_components)
    } else {
      sigma_alpha_init <- expand_dims_2d_diag(
        rfx_working_parameter_prior_cov,
        num_rfx_components
      )
    }

    if (is.null(rfx_group_parameter_prior_cov)) {
      sigma_xi_init <- diag(1, num_rfx_components, num_rfx_components)
    } else {
      sigma_xi_init <- expand_dims_2d_diag(
        rfx_group_parameter_prior_cov,
        num_rfx_components
      )
    }

    sigma_xi_shape <- rfx_variance_prior_shape
    sigma_xi_scale <- rfx_variance_prior_scale

    # Random effects data structure and storage container
    rfx_dataset_train <- createRandomEffectsDataset(
      rfx_group_ids_train,
      rfx_basis_train
    )
    rfx_tracker_train <- createRandomEffectsTracker(rfx_group_ids_train)
    rfx_model <- createRandomEffectsModel(
      num_rfx_components,
      num_rfx_groups
    )
    rfx_model$set_working_parameter(alpha_init)
    rfx_model$set_group_parameters(xi_init)
    rfx_model$set_working_parameter_cov(sigma_alpha_init)
    rfx_model$set_group_parameter_cov(sigma_xi_init)
    rfx_model$set_variance_prior_shape(sigma_xi_shape)
    rfx_model$set_variance_prior_scale(sigma_xi_scale)
    rfx_samples <- createRandomEffectSamples(
      num_rfx_components,
      num_rfx_groups,
      rfx_tracker_train
    )
  }

  # Container of parameter samples
  num_actual_mcmc_iter <- num_mcmc * keep_every
  num_samples <- num_gfr + num_burnin + num_actual_mcmc_iter
  # Delete GFR samples from these containers after the fact if desired
  # num_retained_samples <- ifelse(keep_gfr, num_gfr, 0) + ifelse(keep_burnin, num_burnin, 0) + num_mcmc
  num_retained_samples <- num_gfr +
    ifelse(keep_burnin, num_burnin, 0) +
    num_mcmc * num_chains
  if (sample_sigma2_global) {
    global_var_samples <- rep(NA, num_retained_samples)
  }
  if (sample_sigma2_leaf) {
    leaf_scale_samples <- rep(NA, num_retained_samples)
  }
  if (link_is_cloglog) {
    cloglog_cutpoint_samples <- matrix(
      NA_real_,
      cloglog_num_categories - 1,
      num_retained_samples
    )
  }
  if (include_mean_forest) {
    mean_forest_pred_train <- matrix(
      NA_real_,
      nrow(X_train),
      num_retained_samples
    )
  }
  if (include_variance_forest) {
    variance_forest_pred_train <- matrix(
      NA_real_,
      nrow(X_train),
      num_retained_samples
    )
  }
  sample_counter <- 0

  # Initialize the leaves of each tree in the mean forest
  if (include_mean_forest) {
    if (requires_basis) {
      # Handle the case in which we must initialize root values in a leaf basis regression
      # when init_val_mean != 0. To do this, we regress rep(init_val_mean, nrow(y_train))
      # on leaf_basis_train and use (coefs / num_trees_mean) as initial values
      if (abs(init_val_mean) > 0.00001) {
        init_val_y <- rep(init_val_mean, nrow(y_train))
        init_val_model <- lm(init_val_y ~ 0 + leaf_basis_train)
        init_values_mean_forest <- coef(init_val_model)
        if (any(is.na(init_values_mean_forest))) {
          init_values_mean_forest[which(is.na(init_values_mean_forest))] <- 0.
        }
      } else {
        init_values_mean_forest <- rep(init_val_mean, ncol(leaf_basis_train))
      }
    } else {
      init_values_mean_forest <- init_val_mean
    }
    active_forest_mean$prepare_for_sampler(
      forest_dataset_train,
      outcome_train,
      forest_model_mean,
      leaf_model_mean_forest,
      init_values_mean_forest
    )
  }

  # Initialize the leaves of each tree in the variance forest
  if (include_variance_forest) {
    active_forest_variance$prepare_for_sampler(
      forest_dataset_train,
      outcome_train,
      forest_model_variance,
      leaf_model_variance_forest,
      variance_forest_init
    )
  }

  # Initialize auxiliary data for cloglog
  if (link_is_cloglog) {
    ## Allocate auxiliary data
    train_size <- nrow(X_train)
    # Latent variable (Z in Alam et al (2025) notation)
    forest_dataset_train$add_auxiliary_dimension(train_size)
    # Forest predictions (eta in Alam et al (2025) notation)
    forest_dataset_train$add_auxiliary_dimension(train_size)
    # Log-scale non-cumulative cutpoint (gamma in Alam et al (2025) notation)
    forest_dataset_train$add_auxiliary_dimension(cloglog_num_categories - 1)
    # Exponentiated cumulative cutpoints (exp(c_k) in Alam et al (2025) notation)
    # This auxiliary series is designed so that the element stored at position `i`
    # corresponds to the sum of all exponentiated gamma_j values for j < i.
    # It has cloglog_num_categories elements instead of cloglog_num_categories - 1 because
    # even the largest categorical index has a valid value of sum_{j < i} exp(gamma_j)
    forest_dataset_train$add_auxiliary_dimension(cloglog_num_categories)

    ## Set initial values for auxiliary data
    # Initialize latent variables to zero (slot 0)
    for (i in 1:train_size) {
      forest_dataset_train$set_auxiliary_data_value(0, i - 1, 0.0)
    }
    # Initialize forest predictions to zero (slot 1)
    for (i in 1:train_size) {
      forest_dataset_train$set_auxiliary_data_value(1, i - 1, 0.0)
    }
    # Initialize log-scale cutpoints to 0
    initial_gamma <- rep(0.0, cloglog_num_categories - 1)
    for (i in seq_along(initial_gamma)) {
      forest_dataset_train$set_auxiliary_data_value(2, i - 1, initial_gamma[i])
    }
    # Convert to cumulative exponentiated cutpoints directly in C++
    ordinal_sampler_update_cumsum_exp_cpp(
      ordinal_sampler,
      forest_dataset_train$data_ptr
    )
  }

  # Run GFR (warm start) if specified
  if (num_gfr > 0) {
    for (i in 1:num_gfr) {
      # Keep all GFR samples at this stage -- remove from ForestSamples after MCMC
      # keep_sample <- ifelse(keep_gfr, TRUE, FALSE)
      keep_sample <- TRUE
      if (keep_sample) {
        sample_counter <- sample_counter + 1
      }
      # Print progress
      if (verbose) {
        if ((i %% 10 == 0) || (i == num_gfr)) {
          cat(
            "Sampling",
            i,
            "out of",
            num_gfr,
            "XBART (grow-from-root) draws\n"
          )
        }
      }

      if (include_mean_forest) {
        if (link_is_probit) {
          # Sample latent probit variable, z | -
          # outcome_pred is the centered forest prediction (not including y_bar_train).
          # The truncated normal mean is outcome_pred + y_bar_train (the full eta on the probit scale).
          # The residual stored is z - y_bar_train - outcome_pred so the forest sees a
          # zero-centered signal and the prior shrinkage toward 0 is well-calibrated.
          outcome_pred <- active_forest_mean$predict(
            forest_dataset_train
          )
          if (has_rfx) {
            rfx_pred <- rfx_model$predict(
              rfx_dataset_train,
              rfx_tracker_train
            )
            outcome_pred <- outcome_pred + rfx_pred
          }
          eta_pred <- outcome_pred + y_bar_train
          mu0 <- eta_pred[y_train == 0]
          mu1 <- eta_pred[y_train == 1]
          u0 <- runif(sum(y_train == 0), 0, pnorm(0 - mu0))
          u1 <- runif(sum(y_train == 1), pnorm(0 - mu1), 1)
          resid_train[y_train == 0] <- mu0 + qnorm(u0)
          resid_train[y_train == 1] <- mu1 + qnorm(u1)

          # Update outcome: center z by y_bar_train before passing to forest
          outcome_train$update_data(resid_train - y_bar_train - outcome_pred)
        }

        # Sample mean forest
        forest_model_mean$sample_one_iteration(
          forest_dataset = forest_dataset_train,
          residual = outcome_train,
          forest_samples = forest_samples_mean,
          active_forest = active_forest_mean,
          rng = rng,
          forest_model_config = forest_model_config_mean,
          global_model_config = global_model_config,
          num_threads = num_threads,
          keep_forest = keep_sample,
          gfr = TRUE
        )

        # Cache train set predictions since they are already computed during sampling
        if (keep_sample) {
          mean_forest_pred_train[,
            sample_counter
          ] <- forest_model_mean$get_cached_forest_predictions()
        }

        # Additional Gibbs updates needed for the cloglog model
        if (link_is_cloglog) {
          # Update auxiliary data to current forest predictions
          forest_pred_current <- forest_model_mean$get_cached_forest_predictions()
          for (i in 1:train_size) {
            forest_dataset_train$set_auxiliary_data_value(
              1,
              i - 1,
              forest_pred_current[i]
            )
          }

          # Sample latent z_i's using truncated exponential
          ordinal_sampler_update_latent_variables_cpp(
            ordinal_sampler,
            forest_dataset_train$data_ptr,
            outcome_train$data_ptr,
            rng$rng_ptr
          )

          # Sample gamma parameters (cutpoints)
          ordinal_sampler_update_gamma_params_cpp(
            ordinal_sampler,
            forest_dataset_train$data_ptr,
            outcome_train$data_ptr,
            cloglog_forest_shape,
            cloglog_forest_rate,
            cloglog_cutpoint_0,
            rng$rng_ptr
          )

          # Update cumulative sum of exp(gamma) values
          ordinal_sampler_update_cumsum_exp_cpp(
            ordinal_sampler,
            forest_dataset_train$data_ptr
          )

          # Retain cutpoint draw
          if (keep_sample) {
            cloglog_cutpoints <- forest_dataset_train$get_auxiliary_data_vector(
              2
            )
            cloglog_cutpoint_samples[, sample_counter] <- cloglog_cutpoints
          }
        }
      }
      if (include_variance_forest) {
        forest_model_variance$sample_one_iteration(
          forest_dataset = forest_dataset_train,
          residual = outcome_train,
          forest_samples = forest_samples_variance,
          active_forest = active_forest_variance,
          rng = rng,
          forest_model_config = forest_model_config_variance,
          global_model_config = global_model_config,
          num_threads = num_threads,
          keep_forest = keep_sample,
          gfr = TRUE
        )

        # Cache train set predictions since they are already computed during sampling
        if (keep_sample) {
          variance_forest_pred_train[,
            sample_counter
          ] <- forest_model_variance$get_cached_forest_predictions()
        }
      }
      if (sample_sigma2_global) {
        current_sigma2 <- sampleGlobalErrorVarianceOneIteration(
          outcome_train,
          forest_dataset_train,
          rng,
          a_global,
          b_global
        )
        if (keep_sample) {
          global_var_samples[sample_counter] <- current_sigma2
        }
        global_model_config$update_global_error_variance(current_sigma2)
      }
      if (sample_sigma2_leaf) {
        leaf_scale_double <- sampleLeafVarianceOneIteration(
          active_forest_mean,
          rng,
          a_leaf,
          b_leaf
        )
        current_leaf_scale <- as.matrix(leaf_scale_double)
        if (keep_sample) {
          leaf_scale_samples[sample_counter] <- leaf_scale_double
        }
        forest_model_config_mean$update_leaf_model_scale(
          current_leaf_scale
        )
      }
      if (has_rfx) {
        rfx_model$sample_random_effect(
          rfx_dataset_train,
          outcome_train,
          rfx_tracker_train,
          rfx_samples,
          keep_sample,
          current_sigma2,
          rng
        )
      }
    }
  }

  # Run MCMC
  if (num_burnin + num_mcmc > 0) {
    for (chain_num in 1:num_chains) {
      if (verbose) {
        cat("Sampling chain", chain_num, "of", num_chains, "\n")
      }
      if (num_gfr > 0) {
        # Reset state of active_forest and forest_model based on a previous GFR sample
        forest_ind <- num_gfr - chain_num
        if (include_mean_forest) {
          resetActiveForest(
            active_forest_mean,
            forest_samples_mean,
            forest_ind
          )
          resetForestModel(
            forest_model_mean,
            active_forest_mean,
            forest_dataset_train,
            outcome_train,
            TRUE
          )
          if (sample_sigma2_leaf) {
            leaf_scale_double <- leaf_scale_samples[forest_ind + 1]
            current_leaf_scale <- as.matrix(leaf_scale_double)
            forest_model_config_mean$update_leaf_model_scale(
              current_leaf_scale
            )
          }
          if (link_is_cloglog) {
            # Restore ordinal labels corrupted by resetForestModel's
            # residual adjustment (outcome stores category labels, not residuals)
            outcome_train$update_data(resid_train)
            # We can reset cutpoints from warm-start since cutpoints are retained
            current_cutpoints <- cloglog_cutpoint_samples[, forest_ind + 1]
            for (i in seq_along(current_cutpoints)) {
              forest_dataset_train$set_auxiliary_data_value(
                2,
                i - 1,
                current_cutpoints[i]
              )
            }
            ordinal_sampler_update_cumsum_exp_cpp(
              ordinal_sampler,
              forest_dataset_train$data_ptr
            )
            # Re-predict from the reconstituted active forest
            active_forest_preds <- active_forest_mean$predict(
              forest_dataset_train
            )
            for (i in 1:train_size) {
              forest_dataset_train$set_auxiliary_data_value(
                1,
                i - 1,
                active_forest_preds[i]
              )
              # Latent variables must be reset to 0 and burnt in
              forest_dataset_train$set_auxiliary_data_value(0, i - 1, 0.0)
            }
          }
        }
        if (include_variance_forest) {
          resetActiveForest(
            active_forest_variance,
            forest_samples_variance,
            forest_ind
          )
          resetForestModel(
            forest_model_variance,
            active_forest_variance,
            forest_dataset_train,
            outcome_train,
            FALSE
          )
        }
        if (has_rfx) {
          resetRandomEffectsModel(
            rfx_model,
            rfx_samples,
            forest_ind,
            sigma_alpha_init
          )
          resetRandomEffectsTracker(
            rfx_tracker_train,
            rfx_model,
            rfx_dataset_train,
            outcome_train,
            rfx_samples
          )
        }
        if (sample_sigma2_global) {
          current_sigma2 <- global_var_samples[forest_ind + 1]
          global_model_config$update_global_error_variance(
            current_sigma2
          )
        }
      } else if (has_prev_model) {
        warmstart_index <- ifelse(
          previous_model_decrement,
          previous_model_warmstart_sample_num - chain_num + 1,
          previous_model_warmstart_sample_num
        )
        if (include_mean_forest) {
          resetActiveForest(
            active_forest_mean,
            previous_forest_samples_mean,
            warmstart_index - 1
          )
          resetForestModel(
            forest_model_mean,
            active_forest_mean,
            forest_dataset_train,
            outcome_train,
            TRUE
          )
          if (
            sample_sigma2_leaf &&
              (!is.null(previous_leaf_var_samples))
          ) {
            leaf_scale_double <- previous_leaf_var_samples[
              warmstart_index
            ]
            current_leaf_scale <- as.matrix(leaf_scale_double)
            forest_model_config_mean$update_leaf_model_scale(
              current_leaf_scale
            )
          }
          if (link_is_cloglog) {
            # Restore ordinal labels corrupted by resetForestModel's
            # residual adjustment (outcome stores category labels, not residuals)
            outcome_train$update_data(resid_train)
            # We can reset cutpoints from warm-start since cutpoints are retained
            current_cutpoints <- previous_cloglog_cutpoint_samples[,
              warmstart_index
            ]
            for (i in seq_along(current_cutpoints)) {
              forest_dataset_train$set_auxiliary_data_value(
                2,
                i - 1,
                current_cutpoints[i]
              )
            }
            ordinal_sampler_update_cumsum_exp_cpp(
              ordinal_sampler,
              forest_dataset_train$data_ptr
            )
            # Re-predict from the reconstituted active forest
            active_forest_preds <- active_forest_mean$predict(
              forest_dataset_train
            )
            for (i in 1:train_size) {
              forest_dataset_train$set_auxiliary_data_value(
                1,
                i - 1,
                active_forest_preds[i]
              )
              # Latent variables must be reset to 0 and burnt in
              forest_dataset_train$set_auxiliary_data_value(0, i - 1, 0.0)
            }
          }
        }
        if (include_variance_forest) {
          resetActiveForest(
            active_forest_variance,
            previous_forest_samples_variance,
            warmstart_index - 1
          )
          resetForestModel(
            forest_model_variance,
            active_forest_variance,
            forest_dataset_train,
            outcome_train,
            FALSE
          )
        }
        if (has_rfx) {
          if (is.null(previous_rfx_samples)) {
            warning(
              "`previous_model_json` did not have any random effects samples, so the RFX sampler will be run from scratch while the forests and any other parameters are warm started"
            )
            rootResetRandomEffectsModel(
              rfx_model,
              alpha_init,
              xi_init,
              sigma_alpha_init,
              sigma_xi_init,
              sigma_xi_shape,
              sigma_xi_scale
            )
            rootResetRandomEffectsTracker(
              rfx_tracker_train,
              rfx_model,
              rfx_dataset_train,
              outcome_train
            )
          } else {
            resetRandomEffectsModel(
              rfx_model,
              previous_rfx_samples,
              warmstart_index - 1,
              sigma_alpha_init
            )
            resetRandomEffectsTracker(
              rfx_tracker_train,
              rfx_model,
              rfx_dataset_train,
              outcome_train,
              rfx_samples
            )
          }
        }
        if (sample_sigma2_global) {
          if (!is.null(previous_global_var_samples)) {
            current_sigma2 <- previous_global_var_samples[
              warmstart_index
            ]
            global_model_config$update_global_error_variance(
              current_sigma2
            )
          }
        }
      } else {
        if (include_mean_forest) {
          resetActiveForest(active_forest_mean)
          active_forest_mean$set_root_leaves(
            init_values_mean_forest / num_trees_mean
          )
          resetForestModel(
            forest_model_mean,
            active_forest_mean,
            forest_dataset_train,
            outcome_train,
            TRUE
          )
          if (sample_sigma2_leaf) {
            current_leaf_scale <- as.matrix(sigma2_leaf_init)
            forest_model_config_mean$update_leaf_model_scale(
              current_leaf_scale
            )
          }
          if (link_is_cloglog) {
            # Restore ordinal labels corrupted by resetForestModel's
            # residual adjustment (outcome stores category labels, not residuals)
            outcome_train$update_data(resid_train)
            # Reset all cloglog parameters to default values
            for (i in 1:train_size) {
              forest_dataset_train$set_auxiliary_data_value(0, i - 1, 0.0)
              forest_dataset_train$set_auxiliary_data_value(1, i - 1, 0.0)
            }
            # Initialize log-scale cutpoints to 0
            initial_gamma <- rep(0.0, cloglog_num_categories - 1)
            for (i in seq_along(initial_gamma)) {
              forest_dataset_train$set_auxiliary_data_value(
                2,
                i - 1,
                initial_gamma[i]
              )
            }
            # Convert to cumulative exponentiated cutpoints directly in C++
            ordinal_sampler_update_cumsum_exp_cpp(
              ordinal_sampler,
              forest_dataset_train$data_ptr
            )
          }
        }
        if (include_variance_forest) {
          resetActiveForest(active_forest_variance)
          active_forest_variance$set_root_leaves(
            log(variance_forest_init) / num_trees_variance
          )
          resetForestModel(
            forest_model_variance,
            active_forest_variance,
            forest_dataset_train,
            outcome_train,
            FALSE
          )
        }
        if (has_rfx) {
          rootResetRandomEffectsModel(
            rfx_model,
            alpha_init,
            xi_init,
            sigma_alpha_init,
            sigma_xi_init,
            sigma_xi_shape,
            sigma_xi_scale
          )
          rootResetRandomEffectsTracker(
            rfx_tracker_train,
            rfx_model,
            rfx_dataset_train,
            outcome_train
          )
        }
        if (sample_sigma2_global) {
          current_sigma2 <- sigma2_init
          global_model_config$update_global_error_variance(
            current_sigma2
          )
        }
      }
      for (i in (num_gfr + 1):num_samples) {
        is_mcmc <- i > (num_gfr + num_burnin)
        if (is_mcmc) {
          mcmc_counter <- i - (num_gfr + num_burnin)
          if (mcmc_counter %% keep_every == 0) {
            keep_sample <- TRUE
          } else {
            keep_sample <- FALSE
          }
        } else {
          if (keep_burnin) {
            keep_sample <- TRUE
          } else {
            keep_sample <- FALSE
          }
        }
        if (keep_sample) {
          sample_counter <- sample_counter + 1
        }
        # Print progress
        if (verbose) {
          if (num_burnin > 0 && !is_mcmc) {
            if (
              ((i - num_gfr) %% 100 == 0) ||
                ((i - num_gfr) == num_burnin)
            ) {
              cat(
                "Sampling",
                i - num_gfr,
                "out of",
                num_burnin,
                "BART burn-in draws; Chain number ",
                chain_num,
                "\n"
              )
            }
          }
          if (num_mcmc > 0 && is_mcmc) {
            raw_iter <- i - num_gfr - num_burnin
            if ((raw_iter %% 100 == 0) || (i == num_samples)) {
              if (keep_every == 1) {
                cat(
                  "Sampling",
                  raw_iter,
                  "out of",
                  num_mcmc,
                  "BART MCMC draws; Chain number ",
                  chain_num,
                  "\n"
                )
              } else {
                cat(
                  "Sampling raw draw",
                  raw_iter,
                  "of",
                  num_actual_mcmc_iter,
                  "BART MCMC draws (thinning by",
                  keep_every,
                  ":",
                  raw_iter %/% keep_every,
                  "of",
                  num_mcmc,
                  "retained); Chain number ",
                  chain_num,
                  "\n"
                )
              }
            }
          }
        }

        if (include_mean_forest) {
          if (link_is_probit) {
            # Sample latent probit variable, z | -
            outcome_pred <- active_forest_mean$predict(
              forest_dataset_train
            )
            if (has_rfx) {
              rfx_pred <- rfx_model$predict(
                rfx_dataset_train,
                rfx_tracker_train
              )
              outcome_pred <- outcome_pred + rfx_pred
            }
            eta_pred <- outcome_pred + y_bar_train
            mu0 <- eta_pred[y_train == 0]
            mu1 <- eta_pred[y_train == 1]
            u0 <- runif(sum(y_train == 0), 0, pnorm(0 - mu0))
            u1 <- runif(sum(y_train == 1), pnorm(0 - mu1), 1)
            resid_train[y_train == 0] <- mu0 + qnorm(u0)
            resid_train[y_train == 1] <- mu1 + qnorm(u1)

            # Update outcome: center z by y_bar_train before passing to forest
            outcome_train$update_data(
              resid_train - y_bar_train - outcome_pred
            )
          }

          forest_model_mean$sample_one_iteration(
            forest_dataset = forest_dataset_train,
            residual = outcome_train,
            forest_samples = forest_samples_mean,
            active_forest = active_forest_mean,
            rng = rng,
            forest_model_config = forest_model_config_mean,
            global_model_config = global_model_config,
            num_threads = num_threads,
            keep_forest = keep_sample,
            gfr = FALSE
          )

          # Cache train set predictions since they are already computed during sampling
          if (keep_sample) {
            mean_forest_pred_train[,
              sample_counter
            ] <- forest_model_mean$get_cached_forest_predictions()
          }

          # Additional Gibbs updates needed for the cloglog model
          if (link_is_cloglog) {
            # Update auxiliary data to current forest predictions
            forest_pred_current <- forest_model_mean$get_cached_forest_predictions()
            for (i in 1:train_size) {
              forest_dataset_train$set_auxiliary_data_value(
                1,
                i - 1,
                forest_pred_current[i]
              )
            }

            # Sample latent z_i's using truncated exponential
            ordinal_sampler_update_latent_variables_cpp(
              ordinal_sampler,
              forest_dataset_train$data_ptr,
              outcome_train$data_ptr,
              rng$rng_ptr
            )

            # Sample gamma parameters (cutpoints)
            ordinal_sampler_update_gamma_params_cpp(
              ordinal_sampler,
              forest_dataset_train$data_ptr,
              outcome_train$data_ptr,
              cloglog_forest_shape,
              cloglog_forest_rate,
              cloglog_cutpoint_0,
              rng$rng_ptr
            )

            # Update cumulative sum of exp(gamma) values
            ordinal_sampler_update_cumsum_exp_cpp(
              ordinal_sampler,
              forest_dataset_train$data_ptr
            )

            # Retain cutpoint draw
            if (keep_sample) {
              cloglog_cutpoints <- forest_dataset_train$get_auxiliary_data_vector(
                2
              )
              cloglog_cutpoint_samples[, sample_counter] <- cloglog_cutpoints
            }
          }
        }
        if (include_variance_forest) {
          forest_model_variance$sample_one_iteration(
            forest_dataset = forest_dataset_train,
            residual = outcome_train,
            forest_samples = forest_samples_variance,
            active_forest = active_forest_variance,
            rng = rng,
            forest_model_config = forest_model_config_variance,
            global_model_config = global_model_config,
            num_threads = num_threads,
            keep_forest = keep_sample,
            gfr = FALSE
          )

          # Cache train set predictions since they are already computed during sampling
          if (keep_sample) {
            variance_forest_pred_train[,
              sample_counter
            ] <- forest_model_variance$get_cached_forest_predictions()
          }
        }
        if (sample_sigma2_global) {
          current_sigma2 <- sampleGlobalErrorVarianceOneIteration(
            outcome_train,
            forest_dataset_train,
            rng,
            a_global,
            b_global
          )
          if (keep_sample) {
            global_var_samples[sample_counter] <- current_sigma2
          }
          global_model_config$update_global_error_variance(
            current_sigma2
          )
        }
        if (sample_sigma2_leaf) {
          leaf_scale_double <- sampleLeafVarianceOneIteration(
            active_forest_mean,
            rng,
            a_leaf,
            b_leaf
          )
          current_leaf_scale <- as.matrix(leaf_scale_double)
          if (keep_sample) {
            leaf_scale_samples[sample_counter] <- leaf_scale_double
          }
          forest_model_config_mean$update_leaf_model_scale(
            current_leaf_scale
          )
        }
        if (has_rfx) {
          rfx_model$sample_random_effect(
            rfx_dataset_train,
            outcome_train,
            rfx_tracker_train,
            rfx_samples,
            keep_sample,
            current_sigma2,
            rng
          )
        }
      }
    }
  }

  # Remove GFR samples if they are not to be retained
  if ((!keep_gfr) && (num_gfr > 0)) {
    for (i in 1:num_gfr) {
      if (include_mean_forest) {
        forest_samples_mean$delete_sample(0)
      }
      if (include_variance_forest) {
        forest_samples_variance$delete_sample(0)
      }
      if (has_rfx) {
        rfx_samples$delete_sample(0)
      }
    }
    if (include_mean_forest) {
      mean_forest_pred_train <- mean_forest_pred_train[,
        (num_gfr + 1):ncol(mean_forest_pred_train)
      ]
      if (link_is_cloglog) {
        cloglog_cutpoint_samples <- cloglog_cutpoint_samples[,
          (num_gfr + 1):ncol(cloglog_cutpoint_samples),
          drop = FALSE
        ]
      }
    }
    if (include_variance_forest) {
      variance_forest_pred_train <- variance_forest_pred_train[,
        (num_gfr + 1):ncol(variance_forest_pred_train)
      ]
    }
    if (sample_sigma2_global) {
      global_var_samples <- global_var_samples[
        (num_gfr + 1):length(global_var_samples)
      ]
    }
    if (sample_sigma2_leaf) {
      leaf_scale_samples <- leaf_scale_samples[
        (num_gfr + 1):length(leaf_scale_samples)
      ]
    }
    num_retained_samples <- num_retained_samples - num_gfr
  }

  # Mean forest predictions
  if (include_mean_forest) {
    # y_hat_train <- forest_samples_mean$predict(forest_dataset_train)*y_std_train + y_bar_train
    y_hat_train <- mean_forest_pred_train * y_std_train + y_bar_train
    if (has_test) {
      y_hat_test <- forest_samples_mean$predict(forest_dataset_test) *
        y_std_train +
        y_bar_train
    }
  }

  # Variance forest predictions
  if (include_variance_forest) {
    # sigma2_x_hat_train <- forest_samples_variance$predict(forest_dataset_train)
    sigma2_x_hat_train <- exp(variance_forest_pred_train)
    if (has_test) {
      sigma2_x_hat_test <- forest_samples_variance$predict(
        forest_dataset_test
      )
    }
  }

  # Random effects predictions
  if (has_rfx) {
    rfx_preds_train <- rfx_samples$predict(
      rfx_group_ids_train,
      rfx_basis_train
    ) *
      y_std_train
    y_hat_train <- y_hat_train + rfx_preds_train
  }
  if ((has_rfx_test) && (has_test)) {
    rfx_preds_test <- rfx_samples$predict(
      rfx_group_ids_test,
      rfx_basis_test
    ) *
      y_std_train
    y_hat_test <- y_hat_test + rfx_preds_test
  }

  # Global error variance
  if (sample_sigma2_global) {
    sigma2_global_samples <- global_var_samples * (y_std_train^2)
  }

  # Leaf parameter variance
  if (sample_sigma2_leaf) {
    tau_samples <- leaf_scale_samples
  }

  # Rescale variance forest prediction by global sigma2 (sampled or constant)
  if (include_variance_forest) {
    if (sample_sigma2_global) {
      sigma2_x_hat_train <- sapply(1:num_retained_samples, function(i) {
        sigma2_x_hat_train[, i] * sigma2_global_samples[i]
      })
      if (has_test) {
        sigma2_x_hat_test <- sapply(
          1:num_retained_samples,
          function(i) {
            sigma2_x_hat_test[, i] * sigma2_global_samples[i]
          }
        )
      }
    } else {
      sigma2_x_hat_train <- sigma2_x_hat_train *
        sigma2_init *
        y_std_train *
        y_std_train
      if (has_test) {
        sigma2_x_hat_test <- sigma2_x_hat_test *
          sigma2_init *
          y_std_train *
          y_std_train
      }
    }
  }

  # Return results as a list
  model_params <- list(
    "sigma2_init" = sigma2_init,
    "sigma2_leaf_init" = sigma2_leaf_init,
    "a_global" = a_global,
    "b_global" = b_global,
    "a_leaf" = a_leaf,
    "b_leaf" = b_leaf,
    "a_forest" = a_forest,
    "b_forest" = b_forest,
    "outcome_mean" = y_bar_train,
    "outcome_scale" = y_std_train,
    "standardize" = standardize,
    "leaf_dimension" = leaf_dimension,
    "is_leaf_constant" = is_leaf_constant,
    "leaf_regression" = leaf_regression,
    "requires_basis" = requires_basis,
    "num_covariates" = num_cov_orig,
    "num_basis" = ifelse(
      is.null(leaf_basis_train),
      0,
      ncol(leaf_basis_train)
    ),
    "num_samples" = num_retained_samples,
    "num_gfr" = num_gfr,
    "num_burnin" = num_burnin,
    "num_mcmc" = num_mcmc,
    "keep_every" = keep_every,
    "num_chains" = num_chains,
    "has_basis" = !is.null(leaf_basis_train),
    "has_rfx" = has_rfx,
    "has_rfx_basis" = has_basis_rfx,
    "num_rfx_basis" = num_basis_rfx,
    "sample_sigma2_global" = sample_sigma2_global,
    "sample_sigma2_leaf" = sample_sigma2_leaf,
    "include_mean_forest" = include_mean_forest,
    "include_variance_forest" = include_variance_forest,
    "outcome_model" = outcome_model,
    "probit_outcome_model" = probit_outcome_model,
    "cloglog_num_categories" = ifelse(
      link_is_cloglog,
      cloglog_num_categories,
      0
    ),
    "rfx_model_spec" = rfx_model_spec
  )
  result <- list(
    "model_params" = model_params,
    "train_set_metadata" = X_train_metadata
  )
  if (include_mean_forest) {
    result[["mean_forests"]] = forest_samples_mean
    result[["y_hat_train"]] = y_hat_train
    if (has_test) {
      result[["y_hat_test"]] = y_hat_test
    }
    if (link_is_cloglog && !outcome_is_binary) {
      result[["cloglog_cutpoint_samples"]] = cloglog_cutpoint_samples
    }
  }
  if (include_variance_forest) {
    result[["variance_forests"]] = forest_samples_variance
    result[["sigma2_x_hat_train"]] = sigma2_x_hat_train
    if (has_test) result[["sigma2_x_hat_test"]] = sigma2_x_hat_test
  }
  if (sample_sigma2_global) {
    result[["sigma2_global_samples"]] = sigma2_global_samples
  }
  if (sample_sigma2_leaf) {
    result[["sigma2_leaf_samples"]] = tau_samples
  }
  if (has_rfx) {
    result[["rfx_samples"]] = rfx_samples
    result[["rfx_preds_train"]] = rfx_preds_train
    result[["rfx_unique_group_ids"]] = levels(group_ids_factor)
  }
  if ((has_rfx_test) && (has_test)) {
    result[["rfx_preds_test"]] = rfx_preds_test
  }
  class(result) <- "bartmodel"

  # Clean up classes with external pointers to C++ data structures
  if (include_mean_forest) {
    rm(forest_model_mean)
  }
  if (include_variance_forest) {
    rm(forest_model_variance)
  }
  rm(forest_dataset_train)
  if (has_test) {
    rm(forest_dataset_test)
  }
  if (has_rfx) {
    rm(rfx_dataset_train, rfx_tracker_train, rfx_model)
  }
  rm(outcome_train)
  rm(rng)

  # Restore global RNG state if user provided a random seed
  if (custom_rng) {
    if (has_existing_random_seed) {
      .Random.seed <- original_global_seed
    } else {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  }

  return(result)
}

#' @title Predict from a BART Model
#' @description
#' Predict from a sampled BART model on new data
#'
#' @param object Object of type `bart` containing draws of a regression forest and associated sampling outputs.
#' @param X Covariates used to determine tree leaf predictions for each observation. Must be passed as a matrix or dataframe.
#' @param leaf_basis (Optional) Bases used for prediction (by e.g. dot product with leaf values). Default: `NULL`.
#' @param rfx_group_ids (Optional) Test set group labels used for an additive random effects model.
#' We do not currently support (but plan to in the near future), test set evaluation for group labels
#' that were not in the training set.
#' @param rfx_basis (Optional) Test set basis for "random-slope" regression in additive random effects model.
#' @param type (Optional) Type of prediction to return. Options are "mean", which averages the predictions from every draw of a BART model, and "posterior", which returns the entire matrix of posterior predictions. Default: "posterior".
#' @param terms (Optional) Which model terms to include in the prediction. This can be a single term or a list of model terms. Options include "y_hat", "mean_forest", "rfx", "variance_forest", or "all". If a model doesn't have mean forest, random effects, or variance forest predictions, but one of those terms is request, the request will simply be ignored. If none of the requested terms are present in a model, this function will return `NULL` along with a warning. Default: "all".
#' @param scale (Optional) Scale of mean function predictions. Options are "linear", which returns predictions on the original scale of the mean forest / RFX terms, "probability", which transforms predictions into class probabilities for models with discrete outcomes, and "class", which returns predicted outcome categories for discrete outcome models. "probability" is only valid for outcome models with `outcome == 'binary'` or `outcome == 'ordinal'`. For binary outcomes, this will return the probability that `y == 1`, and for ordinal outcomes, this will return probabilities for each outcome label. Default: "linear".
#' @param ... (Optional) Other prediction parameters.
#'
#' @return List of prediction matrices or single prediction matrix / vector, depending on the terms requested.
#' @export
#'
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' f_XW <- (
#'     ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
#'     ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
#'     ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
#'     ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
#' )
#' noise_sd <- 1
#' y <- f_XW + rnorm(n, 0, noise_sd)
#' test_set_pct <- 0.2
#' n_test <- round(test_set_pct*n)
#' n_train <- n - n_test
#' test_inds <- sort(sample(1:n, n_test, replace = FALSE))
#' train_inds <- (1:n)[!((1:n) %in% test_inds)]
#' X_test <- X[test_inds,]
#' X_train <- X[train_inds,]
#' y_test <- y[test_inds]
#' y_train <- y[train_inds]
#' bart_model <- bart(X_train = X_train, y_train = y_train,
#'                    num_gfr = 10, num_burnin = 0, num_mcmc = 10)
#' y_hat_test <- predict(bart_model, X=X_test)$y_hat
predict.bartmodel <- function(
  object,
  X,
  leaf_basis = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL,
  type = "posterior",
  terms = "all",
  scale = "linear",
  ...
) {
  # Handle mean function scale
  if (!is.character(scale)) {
    stop("scale must be a string or character vector")
  }
  if (!(scale %in% c("linear", "probability", "class"))) {
    stop("scale must either be 'linear', 'probability', or 'class'")
  }
  outcome_model <- object$model_params$outcome_model
  is_probit <- (outcome_model$link == "probit" &&
    outcome_model$outcome == "binary")
  is_binary_cloglog <- (outcome_model$link == "cloglog" &&
    outcome_model$outcome == "binary")
  is_ordinal_cloglog <- (outcome_model$link == "cloglog" &&
    outcome_model$outcome == "ordinal")
  is_cloglog <- is_binary_cloglog || is_ordinal_cloglog
  if ((scale == "probability") && (!(is_probit || is_cloglog))) {
    stop(
      "scale cannot be 'probability' for models not fit with a probit or cloglog outcome model"
    )
  }
  if ((scale == "class") && (!(is_probit || is_cloglog))) {
    stop(
      "scale cannot be 'class' for models not fit with a probit or cloglog outcome model"
    )
  }
  probability_scale <- scale == "probability"
  class_scale <- scale == "class"

  # Handle prediction type
  if (!is.character(type)) {
    stop("type must be a string or character vector")
  }
  if (!(type %in% c("mean", "posterior"))) {
    stop("type must either be 'mean' or 'posterior'")
  }
  predict_mean <- type == "mean"
  if (predict_mean && class_scale) {
    stop("Posterior mean predictions are not supported for scale = 'class'")
  }

  # Handle prediction terms
  rfx_model_spec <- object$model_params$rfx_model_spec
  rfx_intercept <- rfx_model_spec == "intercept_only"
  if (!is.character(terms)) {
    stop("type must be a string or character vector")
  }
  num_terms <- length(terms)
  has_mean_forest <- object$model_params$include_mean_forest
  has_variance_forest <- object$model_params$include_variance_forest
  has_rfx <- object$model_params$has_rfx
  has_y_hat <- has_mean_forest || has_rfx
  predict_y_hat <- (((has_y_hat) && ("y_hat" %in% terms)) ||
    ((has_y_hat) && ("all" %in% terms)))
  predict_mean_forest <- (((has_mean_forest) && ("mean_forest" %in% terms)) ||
    ((has_mean_forest) && ("all" %in% terms)))
  predict_rfx <- (((has_rfx) && ("rfx" %in% terms)) ||
    ((has_rfx) && ("all" %in% terms)))
  predict_variance_forest <- (((has_variance_forest) &&
    ("variance_forest" %in% terms)) ||
    ((has_variance_forest) && ("all" %in% terms)))
  predict_count <- sum(c(
    predict_y_hat,
    predict_mean_forest,
    predict_rfx,
    predict_variance_forest
  ))
  if (predict_count == 0) {
    warning(paste0(
      "None of the requested model terms, ",
      paste(terms, collapse = ", "),
      ", were fit in this model"
    ))
    return(NULL)
  }
  predict_rfx_intermediate <- (predict_y_hat && has_rfx)
  predict_mean_forest_intermediate <- (predict_y_hat && has_mean_forest)
  if (class_scale) {
    if (!((predict_count == 1) && (predict_y_hat))) {
      stop("Class scale can only be used with y_hat predictions")
    }
  }

  # Check that we have at least one term to predict on probability scale
  if (
    probability_scale &&
      !predict_y_hat &&
      !predict_mean_forest &&
      !predict_rfx
  ) {
    stop(
      "scale can only be 'probability' if at least one mean term is requested"
    )
  }

  # Check that covariates are matrix or data frame
  if ((!is.data.frame(X)) && (!is.matrix(X))) {
    stop("X must be a matrix or dataframe")
  }

  # Convert all input data to matrices if not already converted
  if ((is.null(dim(leaf_basis))) && (!is.null(leaf_basis))) {
    leaf_basis <- as.matrix(leaf_basis)
  }
  if ((is.null(dim(rfx_basis))) && (!is.null(rfx_basis))) {
    if (predict_rfx) rfx_basis <- as.matrix(rfx_basis)
  }

  # Data checks
  if ((object$model_params$requires_basis) && (is.null(leaf_basis))) {
    stop("Basis (leaf_basis) must be provided for this model")
  }
  if ((!is.null(leaf_basis)) && (nrow(X) != nrow(leaf_basis))) {
    stop("X and leaf_basis must have the same number of rows")
  }
  if (object$model_params$num_covariates != ncol(X)) {
    stop(
      "X must contain the same number of columns as the BART model's training dataset"
    )
  }
  if ((predict_rfx) && (is.null(rfx_group_ids))) {
    stop(
      "Random effect group labels (rfx_group_ids) must be provided for this model"
    )
  }
  if ((predict_rfx) && (is.null(rfx_basis)) && (!rfx_intercept)) {
    stop("Random effects basis (rfx_basis) must be provided for this model")
  }
  if ((object$model_params$num_rfx_basis > 0) && (!rfx_intercept)) {
    if (ncol(rfx_basis) != object$model_params$num_rfx_basis) {
      stop(
        "Random effects basis has a different dimension than the basis used to train this model"
      )
    }
  }

  # Preprocess covariates
  train_set_metadata <- object$train_set_metadata
  X <- preprocessPredictionData(X, train_set_metadata)

  # Recode group IDs to integer vector (if passed as, for example, a vector of county names, etc...)
  if (!is.null(rfx_group_ids)) {
    rfx_unique_group_ids <- object$rfx_unique_group_ids
    group_ids_factor <- factor(rfx_group_ids, levels = rfx_unique_group_ids)
    if (sum(is.na(group_ids_factor)) > 0) {
      stop(
        "All random effect group labels provided in rfx_group_ids must have been present in rfx_group_ids_train"
      )
    }
    rfx_group_ids <- as.integer(group_ids_factor)
  }

  # Handle RFX model specification
  if (has_rfx) {
    if (object$model_params$rfx_model_spec == "custom") {
      if (is.null(rfx_basis)) {
        stop(
          "A user-provided basis (`rfx_basis`) must be provided when the model was sampled with a random effects model spec set to 'custom'"
        )
      }
    } else if (object$model_params$rfx_model_spec == "intercept_only") {
      # Only construct a basis if user-provided basis missing
      if (is.null(rfx_basis)) {
        rfx_basis <- matrix(
          rep(1, nrow(X)),
          nrow = nrow(X),
          ncol = 1
        )
      }
    }
  }

  # Create prediction dataset
  if (!is.null(leaf_basis)) {
    prediction_dataset <- createForestDataset(X, leaf_basis)
  } else {
    prediction_dataset <- createForestDataset(X)
  }

  # Compute variance forest predictions
  if (predict_variance_forest) {
    s_x_raw <- object$variance_forests$predict(prediction_dataset)
  }

  # Scale variance forest predictions
  num_samples <- object$model_params$num_samples
  y_std <- object$model_params$outcome_scale
  y_bar <- object$model_params$outcome_mean
  sigma2_init <- object$model_params$sigma2_init
  if (predict_variance_forest) {
    if (object$model_params$sample_sigma2_global) {
      sigma2_global_samples <- object$sigma2_global_samples
      variance_forest_predictions <- sapply(1:num_samples, function(i) {
        s_x_raw[, i] * sigma2_global_samples[i]
      })
    } else {
      variance_forest_predictions <- s_x_raw * sigma2_init * y_std * y_std
    }
    if (predict_mean) {
      variance_forest_predictions <- rowMeans(variance_forest_predictions)
    }
  }

  # Compute mean forest predictions
  if (predict_mean_forest || predict_mean_forest_intermediate) {
    mean_forest_predictions <- object$mean_forests$predict(
      prediction_dataset
    ) *
      y_std +
      y_bar
  }

  # Compute rfx predictions (if needed)
  if (predict_rfx || predict_rfx_intermediate) {
    if (!is.null(rfx_basis)) {
      rfx_predictions <- object$rfx_samples$predict(
        rfx_group_ids,
        rfx_basis
      ) *
        y_std
    } else {
      # Sanity check -- this branch should only occur if rfx_model_spec == "intercept_only"
      if (!rfx_intercept) {
        stop(
          "rfx_basis must be provided for random effects models with random slopes"
        )
      }

      # Extract the raw RFX samples and scale by train set outcome standard deviation
      rfx_param_list <- object$rfx_samples$extract_parameter_samples()
      rfx_beta_draws <- rfx_param_list$beta_samples * y_std

      # Promote to an array with consistent dimensions when there's one rfx term
      if (length(dim(rfx_beta_draws)) == 2) {
        dim(rfx_beta_draws) <- c(1, dim(rfx_beta_draws))
      }

      # Construct a matrix with the appropriate group random effects arranged for each observation
      rfx_predictions_raw <- array(
        NA,
        dim = c(
          nrow(X),
          ncol(rfx_basis),
          object$model_params$num_samples
        )
      )
      for (i in 1:nrow(X)) {
        rfx_predictions_raw[i, , ] <-
          rfx_beta_draws[, rfx_group_ids[i], ]
      }

      # Intercept-only model, so the random effect prediction is simply the
      # value of the respective group's intercept coefficient for each observation
      rfx_predictions = rfx_predictions_raw[, 1, ]
    }
  }

  # Combine into y hat predictions
  if (probability_scale) {
    if (is_probit) {
      if (predict_y_hat) {
        if (has_mean_forest && has_rfx) {
          y_hat <- pnorm(mean_forest_predictions + rfx_predictions)
          mean_forest_predictions <- pnorm(mean_forest_predictions)
          rfx_predictions <- pnorm(rfx_predictions)
        } else if (has_mean_forest) {
          y_hat <- pnorm(mean_forest_predictions)
          mean_forest_predictions <- pnorm(mean_forest_predictions)
        } else if (has_rfx) {
          y_hat <- pnorm(rfx_predictions)
          rfx_predictions <- pnorm(rfx_predictions)
        }
      } else {
        if (has_mean_forest && has_rfx) {
          mean_forest_predictions <- pnorm(mean_forest_predictions)
          rfx_predictions <- pnorm(rfx_predictions)
        } else if (has_mean_forest) {
          mean_forest_predictions <- pnorm(mean_forest_predictions)
        } else if (has_rfx) {
          rfx_predictions <- pnorm(rfx_predictions)
        }
      }
    } else if (is_binary_cloglog) {
      mean_forest_predictions <- exp(-exp(mean_forest_predictions))
      if (predict_y_hat) {
        y_hat <- mean_forest_predictions
      }
    } else if (is_ordinal_cloglog) {
      cloglog_num_categories <- object$model_params$cloglog_num_categories
      cloglog_cutpoint_samples <- object$cloglog_cutpoint_samples
      mean_forest_probabilities <- array(
        NA_real_,
        dim = c(
          nrow(X),
          cloglog_num_categories,
          object$model_params$num_samples
        )
      )
      for (j in 1:cloglog_num_categories) {
        if (j == 1) {
          mean_forest_probabilities[, j, ] <- (1 -
            exp(
              -exp(
                sweep(
                  mean_forest_predictions,
                  2,
                  cloglog_cutpoint_samples[j, ],
                  "+"
                )
              )
            ))
        } else if (j == cloglog_num_categories) {
          mean_forest_probabilities[, j, ] <- 1 -
            apply(
              mean_forest_probabilities[, 1:(j - 1), , drop = FALSE],
              c(1, 3),
              sum
            )
        } else {
          mean_forest_probabilities[, j, ] <- (exp(
            -exp(
              sweep(
                mean_forest_predictions,
                2,
                cloglog_cutpoint_samples[j - 1, ],
                "+"
              )
            )
          ) *
            (1 -
              exp(
                -exp(
                  sweep(
                    mean_forest_predictions,
                    2,
                    cloglog_cutpoint_samples[j, ],
                    "+"
                  )
                )
              )))
        }
      }
      if (predict_y_hat) {
        y_hat <- mean_forest_probabilities
      }
      mean_forest_predictions <- mean_forest_probabilities
    }
  } else {
    if (predict_y_hat && has_mean_forest && has_rfx) {
      y_hat <- mean_forest_predictions + rfx_predictions
    } else if (predict_y_hat && has_mean_forest) {
      y_hat <- mean_forest_predictions
    } else if (predict_y_hat && has_rfx) {
      y_hat <- rfx_predictions
    }
  }

  # Collapse to posterior mean predictions if requested
  if (predict_mean) {
    if (predict_mean_forest) {
      if (is_ordinal_cloglog && probability_scale) {
        mean_forest_predictions <- apply(mean_forest_predictions, c(1, 2), mean)
      } else {
        mean_forest_predictions <- rowMeans(mean_forest_predictions)
      }
    }
    if (predict_rfx) {
      # Note: random effects not supported for cloglog, so we don't have to handle the ordinal / cloglog case here
      rfx_predictions <- rowMeans(rfx_predictions)
    }
    if (predict_y_hat) {
      if (is_ordinal_cloglog && probability_scale) {
        y_hat <- apply(y_hat, c(1, 2), mean)
      } else {
        y_hat <- rowMeans(y_hat)
      }
    }
  }

  # Convert probabilities to classes if requested
  if (class_scale) {
    if (is_ordinal_cloglog) {
      y_hat <- apply(y_hat, c(1, 3), which.max)
    } else {
      y_hat <- ifelse(y_hat < 0.5, 0, 1)
    }
  }

  if (predict_count == 1) {
    if (predict_y_hat) {
      return(y_hat)
    } else if (predict_mean_forest) {
      return(mean_forest_predictions)
    } else if (predict_rfx) {
      return(rfx_predictions)
    } else if (predict_variance_forest) {
      return(variance_forest_predictions)
    }
  } else {
    result <- list()
    if (predict_y_hat) {
      result[["y_hat"]] = y_hat
    } else {
      result[["y_hat"]] <- NULL
    }
    if (predict_mean_forest) {
      result[["mean_forest_predictions"]] = mean_forest_predictions
    } else {
      result[["mean_forest_predictions"]] <- NULL
    }
    if (predict_rfx) {
      result[["rfx_predictions"]] = rfx_predictions
    } else {
      result[["rfx_predictions"]] <- NULL
    }
    if (predict_variance_forest) {
      result[["variance_forest_predictions"]] = variance_forest_predictions
    } else {
      result[["variance_forest_predictions"]] <- NULL
    }
    return(result)
  }
}

#' @title Print Summary of BART Model
#' @description Prints a summary of the BART model, including the model terms and their specifications.
#' @param x The BART model object
#' @param ... Additional arguments
#' @export
#' @return BART model object unchanged after printing summary
print.bartmodel <- function(x, ...) {
  # What type of model was run
  model_terms <- c()
  if (x$model_params$include_mean_forest) {
    model_terms <- c(model_terms, "mean forest")
  }
  if (x$model_params$include_variance_forest) {
    model_terms <- c(model_terms, "variance forest")
  }
  if (x$model_params$has_rfx) {
    model_terms <- c(model_terms, "additive random effects")
  }
  if (x$model_params$sample_sigma2_global) {
    model_terms <- c(model_terms, "global error variance model")
  }
  if (x$model_params$sample_sigma2_leaf) {
    model_terms <- c(model_terms, "mean forest leaf scale model")
  }
  if (length(model_terms) > 2) {
    summary_message <- paste0(
      "stochtree::bart() run with ",
      paste0(
        paste0(model_terms[1:(length(model_terms) - 1)], collapse = ", "),
        ", and ",
        model_terms[length(model_terms)]
      )
    )
  } else if (length(model_terms) == 2) {
    summary_message <- paste0(
      "stochtree::bart() run with ",
      paste0(model_terms, collapse = " and ")
    )
  } else {
    summary_message <- paste0("stochtree::bart() run with ", model_terms)
  }

  # Outcome and leaf model details
  outcome_model <- x$model_params$outcome_model
  is_probit <- (outcome_model$link == "probit" &&
    outcome_model$outcome == "binary")
  is_binary_cloglog <- (outcome_model$link == "cloglog" &&
    outcome_model$outcome == "binary")
  is_ordinal_cloglog <- (outcome_model$link == "cloglog" &&
    outcome_model$outcome == "ordinal")
  if (is_ordinal_cloglog) {
    num_categories <- x$model_params$cloglog_num_categories
    outcome_model_summary <- paste0(
      "Ordinal outcome with ",
      num_categories,
      " categories was modeled with a complementary log-log (cloglog) link function"
    )
  } else if (is_binary_cloglog) {
    outcome_model_summary <- paste0(
      "Binary outcome was modeled with a complementary log-log (cloglog) link function"
    )
  } else if (is_probit) {
    outcome_model_summary <- paste0(
      "Binary outcome was modeled with a probit link function"
    )
  } else {
    outcome_model_summary <- paste0(
      "Continuous outcome was modeled as Gaussian"
    )
  }
  if (x$model_params$leaf_regression) {
    summary_message <- paste0(
      summary_message,
      "\n",
      outcome_model_summary,
      " with a leaf regression prior with ",
      x$model_params$leaf_dimension,
      " bases for the mean forest"
    )
  } else if (x$model_params$include_mean_forest) {
    summary_message <- paste0(
      summary_message,
      "\n",
      outcome_model_summary,
      " with a constant leaf prior for the mean forest"
    )
  } else {
    summary_message <- paste0(
      summary_message,
      "\n",
      outcome_model_summary,
    )
  }

  # Standardization
  if (x$model_params$standardize) {
    summary_message <- paste0(
      summary_message,
      "\n",
      "Outcome was standardized"
    )
  }

  # Random effects details
  if (x$model_params$has_rfx) {
    if (x$model_params$rfx_model_spec == "custom") {
      summary_message <- paste0(
        summary_message,
        "\n",
        "Random effects were fit with a user-supplied basis"
      )
    } else if (x$model_params$rfx_model_spec == "intercept_only") {
      summary_message <- paste0(
        summary_message,
        "\n",
        "Random effects were fit with an 'intercept-only' parameterization"
      )
    }
  }

  # Sampler details
  summary_message <- paste0(
    summary_message,
    "\n",
    "The sampler was run for ",
    x$model_params$num_gfr,
    " GFR iterations, with ",
    x$model_params$num_chains,
    ifelse(
      x$model_params$num_chains == 1,
      " chain of ",
      " chains of "
    ),
    x$model_params$num_burnin,
    " burn-in iterations and ",
    x$model_params$num_mcmc,
    " MCMC iterations, ",
    ifelse(
      x$model_params$keep_every == 1,
      "retaining every iteration (i.e. no thinning)",
      paste0(
        "retaining every ",
        x$model_params$keep_every,
        "th iteration (i.e. thinning)"
      )
    )
  )

  # Print the model details
  cat(summary_message, "\n")

  # Return bart_model invisibly
  invisible(x)
}

#' @title Summarize BART Model Fit and Parameters
#' @description Summarize a BART fit with a description of the model that was fit and numeric summaries of any sampled quantities.
#' @param object The BART model object
#' @param ... Additional arguments
#' @export
#' @return BART model object unchanged after summarizing
summary.bartmodel <- function(object, ...) {
  # First, print the BART model
  tmp <- print(object)

  # Summarize any sampled quantities

  # Global error scale
  if (object$model_params$sample_sigma2_global) {
    sigma2_samples <- object$sigma2_global_samples
    n_samples <- length(sigma2_samples)
    mean_sigma2 <- mean(sigma2_samples)
    sd_sigma2 <- sd(sigma2_samples)
    quantiles_sigma2 <- quantile(
      sigma2_samples,
      probs = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
    )
    cat(sprintf(
      "Summary of sigma^2 posterior: \n%d samples, mean = %.3f, standard deviation = %.3f, quantiles:\n",
      n_samples,
      mean_sigma2,
      sd_sigma2
    ))
    print(quantiles_sigma2)
  }

  # Leaf scale
  if (object$model_params$sample_sigma2_leaf) {
    sigma2_leaf_samples <- object$sigma2_leaf_samples
    n_samples <- length(sigma2_leaf_samples)
    mean_sigma2 <- mean(sigma2_leaf_samples)
    sd_sigma2 <- sd(sigma2_leaf_samples)
    quantiles_sigma2 <- quantile(
      sigma2_leaf_samples,
      probs = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
    )
    cat(sprintf(
      "Summary of leaf scale posterior: \n%d samples, mean = %.3f, standard deviation = %.3f, quantiles:\n",
      n_samples,
      mean_sigma2,
      sd_sigma2
    ))
    print(quantiles_sigma2)
  }

  # Determine whether outcome model is binary / ordinal
  outcome_model <- object$model_params$outcome_model
  is_probit <- (outcome_model$link == "probit" &&
    outcome_model$outcome == "binary")
  is_binary_cloglog <- (outcome_model$link == "cloglog" &&
    outcome_model$outcome == "binary")
  is_ordinal_cloglog <- (outcome_model$link == "cloglog" &&
    outcome_model$outcome == "ordinal")
  non_continuous_outcome <- (is_probit ||
    is_binary_cloglog ||
    is_ordinal_cloglog)

  # In-sample predictions
  if (!is.null(object$y_hat_train)) {
    y_hat_train_mean <- rowMeans(object$y_hat_train)
    n_y_hat_train <- length(y_hat_train_mean)
    mean_y_hat_train <- mean(y_hat_train_mean)
    sd_y_hat_train <- sd(y_hat_train_mean)
    quantiles_y_hat_train <- quantile(
      y_hat_train_mean,
      probs = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
    )
    if (non_continuous_outcome) {
      summary_text <- "Summary of in-sample inverse-link-scale posterior predictions: \n%d observations, mean = %.3f, standard deviation = %.3f, quantiles:\n"
    } else {
      summary_text <- "Summary of in-sample posterior mean predictions: \n%d observations, mean = %.3f, standard deviation = %.3f, quantiles:\n"
    }
    cat(sprintf(
      summary_text,
      n_y_hat_train,
      mean_y_hat_train,
      sd_y_hat_train
    ))
    print(quantiles_y_hat_train)
  }

  # Test-set predictions
  if (!is.null(object$y_hat_test)) {
    y_hat_test_mean <- rowMeans(object$y_hat_test)
    n_y_hat_test <- length(y_hat_test_mean)
    mean_y_hat_test <- mean(y_hat_test_mean)
    sd_y_hat_test <- sd(y_hat_test_mean)
    quantiles_y_hat_test <- quantile(
      y_hat_test_mean,
      probs = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
    )
    if (non_continuous_outcome) {
      summary_text <- "Summary of test-set inverse-link-scale posterior predictions: \n%d observations, mean = %.3f, standard deviation = %.3f, quantiles:\n"
    } else {
      summary_text <- "Summary of test-set posterior mean predictions: \n%d observations, mean = %.3f, standard deviation = %.3f, quantiles:\n"
    }
    cat(sprintf(
      summary_text,
      n_y_hat_test,
      mean_y_hat_test,
      sd_y_hat_test
    ))
    print(quantiles_y_hat_test)
  }

  # Random effects
  if (object$model_params$has_rfx) {
    rfx_samples <- getRandomEffectSamples(object)
    rfx_beta_samples <- rfx_samples$beta_samples
    if (length(dim(rfx_beta_samples)) > 2) {
      cat(
        "Random effects summary of variance components across groups and posterior draws:\n"
      )
      rfx_component_means <- apply(rfx_beta_samples, 1, mean)
      rfx_component_sds <- apply(rfx_beta_samples, 1, sd)
      cat("Variance component means: ", rfx_component_means, "\n")
      cat("Variance component standard deviations: ", rfx_component_sds, "\n")
      quantile_summary <- t(apply(
        rfx_beta_samples,
        1,
        quantile,
        probs = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
      ))
      cat("Variance component quantiles:\n")
      print(quantile_summary)
    } else {
      cat(
        "Random effects summary of variance components across groups and posterior draws:\n"
      )
      rfx_component_means <- mean(rfx_beta_samples)
      rfx_component_sds <- sd(rfx_beta_samples)
      cat("Random effects overall mean: ", rfx_component_means, "\n")
      cat(
        "Random effects overall standard deviation: ",
        rfx_component_sds,
        "\n"
      )
      cat("Random effects overall quantiles:\n")
      quantile_summary <- quantile(
        rfx_beta_samples,
        probs = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
      )
      cat("Random effects overall quantiles:\n")
      print(quantile_summary)
    }
  }

  # Return bart_model invisibly
  invisible(object)
}

#' @title Plot BART Model Fit
#' @description Plot the BART model fit and any relevant sampled quantities. This will default to a traceplot of the global error scale and the in-sample mean forest predictions for the first train set observation. Since `stochtree::bart()` is flexible and it's possible to sample a model with a fixed global error scale and no mean forest, this procedure is adaptive and will attempt to plot a trace of whichever model terms are included if these two default terms are omitted.
#' @param x The BART model object
#' @param ... Additional arguments
#' @export
#' @return BART model object unchanged after summarizing
plot.bartmodel <- function(x, ...) {
  # Check if model has global error scale samples
  has_sigma2_samples <- x$model_params$sample_sigma2_global
  has_mean_forest_preds <- !is.null(x$y_hat_train)

  # Check if model is ordinal / binary
  is_probit <- (x$model_params$outcome_model$link == "probit" &&
    x$model_params$outcome_model$outcome == "binary")
  is_binary_cloglog <- (x$model_params$outcome_model$link == "cloglog" &&
    x$model_params$outcome_model$outcome == "binary")
  is_ordinal_cloglog <- (x$model_params$outcome_model$link == "cloglog" &&
    x$model_params$outcome_model$outcome == "ordinal")
  non_continuous_outcome <- (is_probit ||
    is_binary_cloglog ||
    is_ordinal_cloglog)

  # First try combinations of sigma2 and mean forest predictions
  if (has_sigma2_samples || has_mean_forest_preds) {
    if (has_sigma2_samples) {
      plot(
        x$sigma2_global_samples,
        type = "l",
        ylab = "Sigma^2",
        main = "Global error scale traceplot"
      )
    } else if (has_mean_forest_preds) {
      if (non_continuous_outcome) {
        plot_text <- "In-sample inverse-link-scale prediction trace for the first train set observation"
      } else {
        plot_text <- "In-sample mean function trace for the first train set observation"
      }
      plot(
        x$y_hat_train[1, ],
        type = "l",
        ylab = "Predictions",
        main = plot_text
      )
    }
  } else {
    stop(
      "This model does not have enough model terms / parameter traces to produce stochtree's default plots. See `predict.bartmodel()` for examples of how to further investigate your model."
    )
  }

  # Return x invisibly
  invisible(x)
}

#' @title Extract Random Effects Samples from BART Model
#' @description
#' Extract raw sample values for each of the random effect parameter terms.
#'
#' @param object Object of type `bartmodel` containing draws of a BART model and associated sampling outputs.
#' @param ... Other parameters to be used in random effects extraction
#' @return List of arrays. The alpha array has dimension (`num_components`, `num_samples`) and is simply a vector if `num_components = 1`.
#' The xi and beta arrays have dimension (`num_components`, `num_groups`, `num_samples`) and is simply a matrix if `num_components = 1`.
#' The sigma array has dimension (`num_components`, `num_samples`) and is simply a vector if `num_components = 1`.
#' @export
#'
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' f_XW <- (
#'     ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
#'     ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
#'     ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
#'     ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
#' )
#' snr <- 3
#' group_ids <- rep(c(1,2), n %/% 2)
#' rfx_coefs <- matrix(c(-1, -1, 1, 1), nrow=2, byrow=TRUE)
#' rfx_basis <- cbind(1, runif(n, -1, 1))
#' rfx_term <- rowSums(rfx_coefs[group_ids,] * rfx_basis)
#' E_y <- f_XW + rfx_term
#' y <- E_y + rnorm(n, 0, 1)*(sd(E_y)/snr)
#' test_set_pct <- 0.2
#' n_test <- round(test_set_pct*n)
#' n_train <- n - n_test
#' test_inds <- sort(sample(1:n, n_test, replace = FALSE))
#' train_inds <- (1:n)[!((1:n) %in% test_inds)]
#' X_test <- X[test_inds,]
#' X_train <- X[train_inds,]
#' y_test <- y[test_inds]
#' y_train <- y[train_inds]
#' rfx_group_ids_test <- group_ids[test_inds]
#' rfx_group_ids_train <- group_ids[train_inds]
#' rfx_basis_test <- rfx_basis[test_inds,]
#' rfx_basis_train <- rfx_basis[train_inds,]
#' rfx_term_test <- rfx_term[test_inds]
#' rfx_term_train <- rfx_term[train_inds]
#' bart_model <- bart(X_train = X_train, y_train = y_train, X_test = X_test,
#'                    rfx_group_ids_train = rfx_group_ids_train,
#'                    rfx_group_ids_test = rfx_group_ids_test,
#'                    rfx_basis_train = rfx_basis_train,
#'                    rfx_basis_test = rfx_basis_test,
#'                    num_gfr = 10, num_burnin = 0, num_mcmc = 10)
#' rfx_samples <- getRandomEffectSamples(bart_model)
getRandomEffectSamples.bartmodel <- function(object, ...) {
  result = list()

  if (!object$model_params$has_rfx) {
    warning("This model has no RFX terms, returning an empty list")
    return(result)
  }

  # Extract the samples
  result <- object$rfx_samples$extract_parameter_samples()

  # Scale by sd(y_train)
  result$beta_samples <- result$beta_samples *
    object$model_params$outcome_scale
  result$xi_samples <- result$xi_samples * object$model_params$outcome_scale
  result$alpha_samples <- result$alpha_samples *
    object$model_params$outcome_scale
  result$sigma_samples <- result$sigma_samples *
    (object$model_params$outcome_scale^2)

  return(result)
}

#' @title Extract BART Parameter Samples
#' @description Extract a vector, matrix or array of parameter samples from a BART model by name.
#' Random effects are handled by a separate `getRandomEffectSamples` function due to the complexity of the random effects parameters.
#' If the requested model term is not found, an error is thrown.
#' The following conventions are used for parameter names:
#' - Global error variance: `"sigma2"`, `"global_error_scale"`, `"sigma2_global"`
#' - Leaf scale: `"sigma2_leaf"`, `"leaf_scale"`
#' - In-sample mean function predictions: `"y_hat_train"`
#' - Test set mean function predictions: `"y_hat_test"`
#' - In-sample variance forest predictions: `"sigma2_x_train"`, `"var_x_train"`
#' - Test set variance forest predictions: `"sigma2_x_test"`, `"var_x_test"`
#' - Ordinal model cutpoints (valid only for ordinal cloglog models): `"cloglog_cutpoints"`, `"cutpoints"`
#'
#' @param object Object of type `bartmodel` containing draws of a BART model and associated sampling outputs.
#' @param term Name of the parameter to extract (e.g., `"sigma2"`, `"y_hat_train"`, etc.)
#' @return Array of parameter samples. If the underlying parameter is a scalar, this will be a vector of length `num_samples`.
#' If the underlying parameter is vector-valued, this will be (`parameter_dimension` x `num_samples`) matrix, and if the underlying
#' parameter is multidimensional, this will be an array of dimension (`parameter_dimension_1` x `parameter_dimension_2` x ... x `num_samples`).
#' @export
#'
#' @examples
#' n <- 100
#' p <- 5
#' X <- matrix(runif(n*p), ncol = p)
#' f_XW <- (
#'     ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
#'     ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
#'     ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
#'     ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
#' )
#' snr <- 3
#' group_ids <- rep(c(1,2), n %/% 2)
#' rfx_coefs <- matrix(c(-1, -1, 1, 1), nrow=2, byrow=TRUE)
#' rfx_basis <- cbind(1, runif(n, -1, 1))
#' rfx_term <- rowSums(rfx_coefs[group_ids,] * rfx_basis)
#' E_y <- f_XW + rfx_term
#' y <- E_y + rnorm(n, 0, 1)*(sd(E_y)/snr)
#' test_set_pct <- 0.2
#' n_test <- round(test_set_pct*n)
#' n_train <- n - n_test
#' test_inds <- sort(sample(1:n, n_test, replace = FALSE))
#' train_inds <- (1:n)[!((1:n) %in% test_inds)]
#' X_test <- X[test_inds,]
#' X_train <- X[train_inds,]
#' y_test <- y[test_inds]
#' y_train <- y[train_inds]
#' rfx_group_ids_test <- group_ids[test_inds]
#' rfx_group_ids_train <- group_ids[train_inds]
#' rfx_basis_test <- rfx_basis[test_inds,]
#' rfx_basis_train <- rfx_basis[train_inds,]
#' rfx_term_test <- rfx_term[test_inds]
#' rfx_term_train <- rfx_term[train_inds]
#' bart_model <- bart(X_train = X_train, y_train = y_train, X_test = X_test,
#'                    rfx_group_ids_train = rfx_group_ids_train,
#'                    rfx_group_ids_test = rfx_group_ids_test,
#'                    rfx_basis_train = rfx_basis_train,
#'                    rfx_basis_test = rfx_basis_test,
#'                    num_gfr = 10, num_burnin = 0, num_mcmc = 10)
#' sigma2_samples <- extractParameter(bart_model, "sigma2")
extractParameter.bartmodel <- function(object, term) {
  if (term %in% c("sigma2", "global_error_scale", "sigma2_global")) {
    if (!is.null(object$sigma2_global_samples)) {
      return(object$sigma2_global_samples)
    } else {
      stop("This model does not have global variance parameter samples")
    }
  }

  if (term %in% c("sigma2_leaf", "leaf_scale")) {
    if (!is.null(object$sigma2_leaf_samples)) {
      return(object$sigma2_leaf_samples)
    } else {
      stop("This model does not have leaf variance parameter samples")
    }
  }

  if (term %in% c("y_hat_train")) {
    if (!is.null(object$y_hat_train)) {
      return(object$y_hat_train)
    } else {
      stop(
        "This model does not have in-sample mean function prediction samples"
      )
    }
  }

  if (term %in% c("y_hat_test")) {
    if (!is.null(object$y_hat_test)) {
      return(object$y_hat_test)
    } else {
      stop("This model does not have test set mean function prediction samples")
    }
  }

  if (term %in% c("sigma2_x_train", "var_x_train")) {
    if (!is.null(object$sigma2_x_hat_train)) {
      return(object$sigma2_x_hat_train)
    } else {
      stop("This model does not have in-sample variance forest predictions")
    }
  }

  if (term %in% c("sigma2_x_test", "var_x_test")) {
    if (!is.null(object$sigma2_x_hat_test)) {
      return(object$sigma2_x_hat_test)
    } else {
      stop("This model does not have test set variance forest predictions")
    }
  }

  if (term %in% c("cloglog_cutpoints", "cutpoints")) {
    if (!is.null(object$cloglog_cutpoint_samples)) {
      return(object$cloglog_cutpoint_samples)
    } else {
      stop("This model does not have ordinal cutpoint samples")
    }
  }

  stop(paste0("term ", term, " is not a valid BART model term"))
}

#' @title Convert BART Model to JSON
#' @rdname BARTSerialization
#' @param object Object of type `bartmodel` containing draws of a BART model and associated sampling outputs.
#' @export
saveBARTModelToJson <- function(object) {
  jsonobj <- createCppJson()

  if (!inherits(object, "bartmodel")) {
    stop("`object` must be a BART model")
  }

  if (is.null(object$model_params)) {
    stop("This BCF model has not yet been sampled")
  }

  # Add the forests
  if (object$model_params$include_mean_forest) {
    jsonobj$add_forest(object$mean_forests)
  }
  if (object$model_params$include_variance_forest) {
    jsonobj$add_forest(object$variance_forests)
  }

  # Add metadata
  jsonobj$add_scalar(
    "num_numeric_vars",
    object$train_set_metadata$num_numeric_vars
  )
  jsonobj$add_scalar(
    "num_ordered_cat_vars",
    object$train_set_metadata$num_ordered_cat_vars
  )
  jsonobj$add_scalar(
    "num_unordered_cat_vars",
    object$train_set_metadata$num_unordered_cat_vars
  )
  if (object$train_set_metadata$num_numeric_vars > 0) {
    jsonobj$add_string_vector(
      "numeric_vars",
      object$train_set_metadata$numeric_vars
    )
  }
  if (object$train_set_metadata$num_ordered_cat_vars > 0) {
    jsonobj$add_string_vector(
      "ordered_cat_vars",
      object$train_set_metadata$ordered_cat_vars
    )
    jsonobj$add_string_list(
      "ordered_unique_levels",
      object$train_set_metadata$ordered_unique_levels
    )
  }
  if (object$train_set_metadata$num_unordered_cat_vars > 0) {
    jsonobj$add_string_vector(
      "unordered_cat_vars",
      object$train_set_metadata$unordered_cat_vars
    )
    jsonobj$add_string_list(
      "unordered_unique_levels",
      object$train_set_metadata$unordered_unique_levels
    )
  }

  # Add version stamp and global parameters
  jsonobj$add_string("stochtree_version", getStochtreeVersion())
  jsonobj$add_scalar("outcome_scale", object$model_params$outcome_scale)
  jsonobj$add_scalar("outcome_mean", object$model_params$outcome_mean)
  jsonobj$add_boolean("standardize", object$model_params$standardize)
  jsonobj$add_scalar("sigma2_init", object$model_params$sigma2_init)
  jsonobj$add_boolean(
    "sample_sigma2_global",
    object$model_params$sample_sigma2_global
  )
  jsonobj$add_boolean(
    "sample_sigma2_leaf",
    object$model_params$sample_sigma2_leaf
  )
  jsonobj$add_boolean(
    "include_mean_forest",
    object$model_params$include_mean_forest
  )
  jsonobj$add_boolean(
    "include_variance_forest",
    object$model_params$include_variance_forest
  )
  jsonobj$add_boolean("has_rfx", object$model_params$has_rfx)
  jsonobj$add_boolean("has_rfx_basis", object$model_params$has_rfx_basis)
  jsonobj$add_scalar("num_rfx_basis", object$model_params$num_rfx_basis)
  jsonobj$add_scalar("num_gfr", object$model_params$num_gfr)
  jsonobj$add_scalar("num_burnin", object$model_params$num_burnin)
  jsonobj$add_scalar("num_mcmc", object$model_params$num_mcmc)
  jsonobj$add_scalar("num_samples", object$model_params$num_samples)
  jsonobj$add_scalar("num_covariates", object$model_params$num_covariates)
  jsonobj$add_scalar("num_basis", object$model_params$num_basis)
  jsonobj$add_scalar("num_chains", object$model_params$num_chains)
  jsonobj$add_scalar("keep_every", object$model_params$keep_every)
  jsonobj$add_boolean("requires_basis", object$model_params$requires_basis)
  jsonobj$add_string(
    "outcome",
    object$model_params$outcome_model$outcome,
    "outcome_model"
  )
  jsonobj$add_string(
    "link",
    object$model_params$outcome_model$link,
    "outcome_model"
  )
  jsonobj$add_boolean(
    "probit_outcome_model",
    object$model_params$probit_outcome_model
  )
  jsonobj$add_string(
    "rfx_model_spec",
    object$model_params$rfx_model_spec
  )
  if (object$model_params$outcome_model$link == "cloglog") {
    jsonobj$add_scalar(
      "cloglog_num_categories",
      object$model_params$cloglog_num_categories
    )
    if (object$model_params$outcome_model$outcome == "ordinal") {
      for (i in 1:(object$model_params$cloglog_num_categories - 1)) {
        jsonobj$add_vector(
          paste0("cloglog_cutpoint_samples_", i),
          object$cloglog_cutpoint_samples[i, ],
          "parameters"
        )
      }
    }
  }
  if (object$model_params$sample_sigma2_global) {
    jsonobj$add_vector(
      "sigma2_global_samples",
      object$sigma2_global_samples,
      "parameters"
    )
  }
  if (object$model_params$sample_sigma2_leaf) {
    jsonobj$add_vector(
      "sigma2_leaf_samples",
      object$sigma2_leaf_samples,
      "parameters"
    )
  }

  # Add random effects (if present)
  if (object$model_params$has_rfx) {
    jsonobj$add_random_effects(object$rfx_samples)
    jsonobj$add_string_vector(
      "rfx_unique_group_ids",
      object$rfx_unique_group_ids
    )
  }

  # Add covariate preprocessor metadata
  preprocessor_metadata_string <- savePreprocessorToJsonString(
    object$train_set_metadata
  )
  jsonobj$add_string("preprocessor_metadata", preprocessor_metadata_string)

  return(jsonobj)
}

#' @title Save BART Model to JSON File
#' @rdname BARTSerialization
#' @param object Object of type `bartmodel` containing draws of a BART model and associated sampling outputs.
#' @param filename String of filepath, must end in ".json"
#'
#' @export
saveBARTModelToJsonFile <- function(object, filename) {
  # Convert to Json
  jsonobj <- saveBARTModelToJson(object)

  # Save to file
  jsonobj$save_file(filename)
}

#' @title Convert BART Model to JSON String
#' @rdname BARTSerialization
#' @param object Object of type `bartmodel` containing draws of a BART model and associated sampling outputs.
#' @export
saveBARTModelToJsonString <- function(object) {
  # Convert to Json
  jsonobj <- saveBARTModelToJson(object)

  # Dump to string
  return(jsonobj$return_json_string())
}

#' @title Convert JSON to BART Model
#' @rdname BARTSerialization
#' @param json_object Object of type `CppJson` containing Json representation of a BART model
#' @export
createBARTModelFromJson <- function(json_object) {
  # Initialize the BCF model
  output <- list()

  # Helpers for optional-field presence checks
  .ver <- inferStochtreeJsonVersion(json_object)
  has_field <- function(name) {
    json_contains_field_cpp(json_object$json_ptr, name)
  }
  has_subfolder_field <- function(subfolder, name) {
    json_contains_field_subfolder_cpp(json_object$json_ptr, subfolder, name)
  }

  # Unpack the forests
  include_mean_forest <- json_object$get_boolean("include_mean_forest")
  include_variance_forest <- json_object$get_boolean(
    "include_variance_forest"
  )
  if (include_mean_forest) {
    output[["mean_forests"]] <- loadForestContainerJson(
      json_object,
      "forest_0"
    )
    if (include_variance_forest) {
      output[["variance_forests"]] <- loadForestContainerJson(
        json_object,
        "forest_1"
      )
    }
  } else {
    output[["variance_forests"]] <- loadForestContainerJson(
      json_object,
      "forest_0"
    )
  }

  # Unpack metadata
  train_set_metadata = list()
  train_set_metadata[["num_numeric_vars"]] <- json_object$get_scalar(
    "num_numeric_vars"
  )
  train_set_metadata[["num_ordered_cat_vars"]] <- json_object$get_scalar(
    "num_ordered_cat_vars"
  )
  train_set_metadata[["num_unordered_cat_vars"]] <- json_object$get_scalar(
    "num_unordered_cat_vars"
  )
  if (train_set_metadata[["num_numeric_vars"]] > 0) {
    train_set_metadata[["numeric_vars"]] <- json_object$get_string_vector(
      "numeric_vars"
    )
  }
  if (train_set_metadata[["num_ordered_cat_vars"]] > 0) {
    train_set_metadata[[
      "ordered_cat_vars"
    ]] <- json_object$get_string_vector("ordered_cat_vars")
    train_set_metadata[[
      "ordered_unique_levels"
    ]] <- json_object$get_string_list(
      "ordered_unique_levels",
      train_set_metadata[["ordered_cat_vars"]]
    )
  }
  if (train_set_metadata[["num_unordered_cat_vars"]] > 0) {
    train_set_metadata[[
      "unordered_cat_vars"
    ]] <- json_object$get_string_vector("unordered_cat_vars")
    train_set_metadata[[
      "unordered_unique_levels"
    ]] <- json_object$get_string_list(
      "unordered_unique_levels",
      train_set_metadata[["unordered_cat_vars"]]
    )
  }
  output[["train_set_metadata"]] <- train_set_metadata

  # Unpack model params
  model_params = list()
  model_params[["outcome_scale"]] <- json_object$get_scalar("outcome_scale")
  model_params[["outcome_mean"]] <- json_object$get_scalar("outcome_mean")
  model_params[["standardize"]] <- json_object$get_boolean("standardize")
  model_params[["sigma2_init"]] <- json_object$get_scalar("sigma2_init")
  model_params[["sample_sigma2_global"]] <- json_object$get_boolean(
    "sample_sigma2_global"
  )
  model_params[["sample_sigma2_leaf"]] <- json_object$get_boolean(
    "sample_sigma2_leaf"
  )
  model_params[["include_mean_forest"]] <- include_mean_forest
  model_params[["include_variance_forest"]] <- include_variance_forest
  model_params[["has_rfx"]] <- json_object$get_boolean("has_rfx")

  if (has_field("has_rfx_basis")) {
    model_params[["has_rfx_basis"]] <- json_object$get_boolean("has_rfx_basis")
    model_params[["num_rfx_basis"]] <- json_object$get_scalar("num_rfx_basis")
  } else {
    model_params[["has_rfx_basis"]] <- FALSE
    model_params[["num_rfx_basis"]] <- 1
    warning(paste0(
      "Fields 'has_rfx_basis' and 'num_rfx_basis' not found in JSON (model appears to have been ",
      "serialized under stochtree ",
      .ver,
      "). Defaulting to FALSE / 1. ",
      "Re-save your model to suppress this warning."
    ))
  }

  model_params[["num_gfr"]] <- json_object$get_scalar("num_gfr")
  model_params[["num_burnin"]] <- json_object$get_scalar("num_burnin")
  model_params[["num_mcmc"]] <- json_object$get_scalar("num_mcmc")
  model_params[["num_samples"]] <- json_object$get_scalar("num_samples")
  model_params[["num_covariates"]] <- if (has_field("num_covariates")) {
    json_object$get_scalar("num_covariates")
  } else {
    NA_real_
  }
  model_params[["num_basis"]] <- json_object$get_scalar("num_basis")
  model_params[["requires_basis"]] <- json_object$get_boolean("requires_basis")

  if (has_field("num_chains")) {
    model_params[["num_chains"]] <- json_object$get_scalar("num_chains")
  } else {
    model_params[["num_chains"]] <- 1
    warning(paste0(
      "Field 'num_chains' not found in JSON (model appears to have been serialized under stochtree ",
      .ver,
      "). Defaulting to 1. Re-save your model to suppress this warning."
    ))
  }

  if (has_field("keep_every")) {
    model_params[["keep_every"]] <- json_object$get_scalar("keep_every")
  } else {
    model_params[["keep_every"]] <- 1
    warning(paste0(
      "Field 'keep_every' not found in JSON (model appears to have been serialized under stochtree ",
      .ver,
      "). Defaulting to 1. Re-save your model to suppress this warning."
    ))
  }

  model_params[["probit_outcome_model"]] <- if (
    has_field("probit_outcome_model")
  ) {
    json_object$get_boolean("probit_outcome_model")
  } else {
    FALSE
  }

  if (
    has_subfolder_field("outcome_model", "outcome") &&
      has_subfolder_field("outcome_model", "link")
  ) {
    outcome_model_outcome <- json_object$get_string("outcome", "outcome_model")
    outcome_model_link <- json_object$get_string("link", "outcome_model")
  } else {
    outcome_model_outcome <- "continuous"
    outcome_model_link <- "identity"
    warning(paste0(
      "Fields 'outcome' and 'link' not found under 'outcome_model' in JSON (model appears to have ",
      "been serialized under stochtree ",
      .ver,
      "). Defaulting to outcome='continuous', ",
      "link='identity'. Re-save your model to suppress this warning."
    ))
  }
  model_params[["outcome_model"]] <- OutcomeModel(
    outcome = outcome_model_outcome,
    link = outcome_model_link
  )

  if (has_field("rfx_model_spec")) {
    model_params[["rfx_model_spec"]] <- json_object$get_string("rfx_model_spec")
  } else {
    model_params[["rfx_model_spec"]] <- ""
    if (model_params[["has_rfx"]]) {
      warning(paste0(
        "Field 'rfx_model_spec' not found in JSON (model appears to have been serialized under ",
        "stochtree ",
        .ver,
        "). Defaulting to ''. Re-save your model to suppress this warning."
      ))
    }
  }
  if (model_params[["outcome_model"]]$link == "cloglog") {
    cloglog_num_categories <- json_object$get_scalar("cloglog_num_categories")
    model_params[["cloglog_num_categories"]] <- cloglog_num_categories
  } else {
    model_params[["cloglog_num_categories"]] <- 0
  }

  output[["model_params"]] <- model_params

  # Unpack sampled parameters
  if (model_params[["sample_sigma2_global"]]) {
    output[["sigma2_global_samples"]] <- json_object$get_vector(
      "sigma2_global_samples",
      "parameters"
    )
  }
  if (model_params[["sample_sigma2_leaf"]]) {
    output[["sigma2_leaf_samples"]] <- json_object$get_vector(
      "sigma2_leaf_samples",
      "parameters"
    )
  }
  if (
    model_params[["outcome_model"]]$link == "cloglog" &&
      model_params[["outcome_model"]]$outcome == "ordinal"
  ) {
    cloglog_cutpoint_samples <- matrix(
      NA_real_,
      model_params[["cloglog_num_categories"]] - 1,
      model_params[["num_samples"]]
    )
    for (i in 1:(model_params[["cloglog_num_categories"]] - 1)) {
      cloglog_cutpoint_samples[i, ] <- json_object$get_vector(
        paste0("cloglog_cutpoint_samples_", i),
        "parameters"
      )
    }
    output[["cloglog_cutpoint_samples"]] <- cloglog_cutpoint_samples
  }

  # Unpack random effects
  if (model_params[["has_rfx"]]) {
    output[["rfx_unique_group_ids"]] <- json_object$get_string_vector(
      "rfx_unique_group_ids"
    )
    output[["rfx_samples"]] <- loadRandomEffectSamplesJson(json_object, 0)
  }

  # Unpack covariate preprocessor
  if (has_field("preprocessor_metadata")) {
    preprocessor_metadata_string <- json_object$get_string(
      "preprocessor_metadata"
    )
    output[["train_set_metadata"]] <- createPreprocessorFromJsonString(
      preprocessor_metadata_string
    )
  } else {
    output[["train_set_metadata"]] <- NULL
    warning(paste0(
      "Field 'preprocessor_metadata' not found in JSON (model appears to have been serialized ",
      "under stochtree ",
      .ver,
      "). DataFrame covariates will not be supported for prediction. ",
      "Re-save your model to suppress this warning."
    ))
  }

  class(output) <- "bartmodel"
  return(output)
}

#' @title Convert JSON File to BART Model
#' @rdname BARTSerialization
#' @param json_filename String of filepath, must end in ".json"
#' @export
createBARTModelFromJsonFile <- function(json_filename) {
  # Load a `CppJson` object from file
  bart_json <- createCppJsonFile(json_filename)

  # Create and return the BART object
  bart_object <- createBARTModelFromJson(bart_json)

  return(bart_object)
}

#' @title Convert JSON String to BART Model
#' @rdname BARTSerialization
#' @param json_string JSON string dump
#' @export
createBARTModelFromJsonString <- function(json_string) {
  # Load a `CppJson` object from string
  bart_json <- createCppJsonString(json_string)

  # Create and return the BART object
  bart_object <- createBARTModelFromJson(bart_json)

  return(bart_object)
}

#' @title Convert JSON List to Single BART Model
#' @rdname BARTSerialization
#' @param json_object_list List of objects of type `CppJson` containing Json representation of a BART model
#' @export
createBARTModelFromCombinedJson <- function(json_object_list) {
  # Initialize the BCF model
  output <- list()

  # For scalar / preprocessing details which aren't sample-dependent,
  # defer to the first json
  json_object_default <- json_object_list[[1]]

  # Helpers for optional-field presence checks
  .ver <- inferStochtreeJsonVersion(json_object_default)
  has_field <- function(name) {
    json_contains_field_cpp(json_object_default$json_ptr, name)
  }
  has_subfolder_field <- function(subfolder, name) {
    json_contains_field_subfolder_cpp(
      json_object_default$json_ptr,
      subfolder,
      name
    )
  }

  # Unpack the forests
  include_mean_forest <- json_object_default$get_boolean(
    "include_mean_forest"
  )
  include_variance_forest <- json_object_default$get_boolean(
    "include_variance_forest"
  )
  if (include_mean_forest) {
    output[["mean_forests"]] <- loadForestContainerCombinedJson(
      json_object_list,
      "forest_0"
    )
    if (include_variance_forest) {
      output[["variance_forests"]] <- loadForestContainerCombinedJson(
        json_object_list,
        "forest_1"
      )
    }
  } else {
    output[["variance_forests"]] <- loadForestContainerCombinedJson(
      json_object_list,
      "forest_0"
    )
  }

  # Unpack metadata
  train_set_metadata = list()
  train_set_metadata[["num_numeric_vars"]] <- json_object_default$get_scalar(
    "num_numeric_vars"
  )
  train_set_metadata[[
    "num_ordered_cat_vars"
  ]] <- json_object_default$get_scalar("num_ordered_cat_vars")
  train_set_metadata[[
    "num_unordered_cat_vars"
  ]] <- json_object_default$get_scalar("num_unordered_cat_vars")
  if (train_set_metadata[["num_numeric_vars"]] > 0) {
    train_set_metadata[[
      "numeric_vars"
    ]] <- json_object_default$get_string_vector("numeric_vars")
  }
  if (train_set_metadata[["num_ordered_cat_vars"]] > 0) {
    train_set_metadata[[
      "ordered_cat_vars"
    ]] <- json_object_default$get_string_vector("ordered_cat_vars")
    train_set_metadata[[
      "ordered_unique_levels"
    ]] <- json_object_default$get_string_list(
      "ordered_unique_levels",
      train_set_metadata[["ordered_cat_vars"]]
    )
  }
  if (train_set_metadata[["num_unordered_cat_vars"]] > 0) {
    train_set_metadata[[
      "unordered_cat_vars"
    ]] <- json_object_default$get_string_vector("unordered_cat_vars")
    train_set_metadata[[
      "unordered_unique_levels"
    ]] <- json_object_default$get_string_list(
      "unordered_unique_levels",
      train_set_metadata[["unordered_cat_vars"]]
    )
  }
  output[["train_set_metadata"]] <- train_set_metadata

  # Unpack model params
  model_params = list()
  model_params[["outcome_scale"]] <- json_object_default$get_scalar(
    "outcome_scale"
  )
  model_params[["outcome_mean"]] <- json_object_default$get_scalar(
    "outcome_mean"
  )
  model_params[["standardize"]] <- json_object_default$get_boolean(
    "standardize"
  )
  model_params[["sigma2_init"]] <- json_object_default$get_scalar(
    "sigma2_init"
  )
  model_params[["sample_sigma2_global"]] <- json_object_default$get_boolean(
    "sample_sigma2_global"
  )
  model_params[["sample_sigma2_leaf"]] <- json_object_default$get_boolean(
    "sample_sigma2_leaf"
  )
  model_params[["include_mean_forest"]] <- include_mean_forest
  model_params[["include_variance_forest"]] <- include_variance_forest
  model_params[["has_rfx"]] <- json_object_default$get_boolean("has_rfx")

  if (has_field("has_rfx_basis")) {
    model_params[["has_rfx_basis"]] <- json_object_default$get_boolean(
      "has_rfx_basis"
    )
    model_params[["num_rfx_basis"]] <- json_object_default$get_scalar(
      "num_rfx_basis"
    )
  } else {
    model_params[["has_rfx_basis"]] <- FALSE
    model_params[["num_rfx_basis"]] <- 1
    warning(paste0(
      "Fields 'has_rfx_basis' and 'num_rfx_basis' not found in JSON (model appears to have been ",
      "serialized under stochtree ",
      .ver,
      "). Defaulting to FALSE / 1. ",
      "Re-save your model to suppress this warning."
    ))
  }

  model_params[["num_covariates"]] <- if (has_field("num_covariates")) {
    json_object_default$get_scalar("num_covariates")
  } else {
    NA_real_
  }
  model_params[["num_basis"]] <- json_object_default$get_scalar("num_basis")
  model_params[["requires_basis"]] <- json_object_default$get_boolean(
    "requires_basis"
  )

  model_params[["probit_outcome_model"]] <- if (
    has_field("probit_outcome_model")
  ) {
    json_object_default$get_boolean("probit_outcome_model")
  } else {
    FALSE
  }

  if (
    has_subfolder_field("outcome_model", "outcome") &&
      has_subfolder_field("outcome_model", "link")
  ) {
    outcome_model_outcome <- json_object_default$get_string(
      "outcome",
      "outcome_model"
    )
    outcome_model_link <- json_object_default$get_string(
      "link",
      "outcome_model"
    )
  } else {
    outcome_model_outcome <- "continuous"
    outcome_model_link <- "identity"
    warning(paste0(
      "Fields 'outcome' and 'link' not found under 'outcome_model' in JSON (model appears to have ",
      "been serialized under stochtree ",
      .ver,
      "). Defaulting to outcome='continuous', ",
      "link='identity'. Re-save your model to suppress this warning."
    ))
  }
  model_params[["outcome_model"]] <- OutcomeModel(
    outcome = outcome_model_outcome,
    link = outcome_model_link
  )

  if (has_field("rfx_model_spec")) {
    model_params[["rfx_model_spec"]] <- json_object_default$get_string(
      "rfx_model_spec"
    )
  } else {
    model_params[["rfx_model_spec"]] <- ""
    if (model_params[["has_rfx"]]) {
      warning(paste0(
        "Field 'rfx_model_spec' not found in JSON (model appears to have been serialized under ",
        "stochtree ",
        .ver,
        "). Defaulting to ''. Re-save your model to suppress this warning."
      ))
    }
  }

  if (has_field("num_chains")) {
    model_params[["num_chains"]] <- json_object_default$get_scalar("num_chains")
  } else {
    model_params[["num_chains"]] <- 1
    warning(paste0(
      "Field 'num_chains' not found in JSON (model appears to have been serialized under stochtree ",
      .ver,
      "). Defaulting to 1. Re-save your model to suppress this warning."
    ))
  }

  if (has_field("keep_every")) {
    model_params[["keep_every"]] <- json_object_default$get_scalar("keep_every")
  } else {
    model_params[["keep_every"]] <- 1
    warning(paste0(
      "Field 'keep_every' not found in JSON (model appears to have been serialized under stochtree ",
      .ver,
      "). Defaulting to 1. Re-save your model to suppress this warning."
    ))
  }
  if (model_params[["outcome_model"]]$link == "cloglog") {
    cloglog_num_categories <- json_object_default$get_scalar(
      "cloglog_num_categories"
    )
    model_params[["cloglog_num_categories"]] <- cloglog_num_categories
  } else {
    model_params[["cloglog_num_categories"]] <- 0
  }

  # Combine values that are sample-specific
  for (i in 1:length(json_object_list)) {
    json_object <- json_object_list[[i]]
    if (i == 1) {
      model_params[["num_gfr"]] <- json_object$get_scalar("num_gfr")
      model_params[["num_burnin"]] <- json_object$get_scalar("num_burnin")
      model_params[["num_mcmc"]] <- json_object$get_scalar("num_mcmc")
      model_params[["num_samples"]] <- json_object$get_scalar(
        "num_samples"
      )
    } else {
      prev_json <- json_object_list[[i - 1]]
      model_params[["num_gfr"]] <- model_params[["num_gfr"]] +
        json_object$get_scalar("num_gfr")
      model_params[["num_burnin"]] <- model_params[["num_burnin"]] +
        json_object$get_scalar("num_burnin")
      model_params[["num_mcmc"]] <- model_params[["num_mcmc"]] +
        json_object$get_scalar("num_mcmc")
      model_params[["num_samples"]] <- model_params[["num_samples"]] +
        json_object$get_scalar("num_samples")
    }
  }
  output[["model_params"]] <- model_params

  # Unpack sampled parameters
  if (model_params[["sample_sigma2_global"]]) {
    for (i in 1:length(json_object_list)) {
      json_object <- json_object_list[[i]]
      if (i == 1) {
        output[["sigma2_global_samples"]] <- json_object$get_vector(
          "sigma2_global_samples",
          "parameters"
        )
      } else {
        output[["sigma2_global_samples"]] <- c(
          output[["sigma2_global_samples"]],
          json_object$get_vector(
            "sigma2_global_samples",
            "parameters"
          )
        )
      }
    }
  }
  if (model_params[["sample_sigma2_leaf"]]) {
    for (i in 1:length(json_object_list)) {
      json_object <- json_object_list[[i]]
      if (i == 1) {
        output[["sigma2_leaf_samples"]] <- json_object$get_vector(
          "sigma2_leaf_samples",
          "parameters"
        )
      } else {
        output[["sigma2_leaf_samples"]] <- c(
          output[["sigma2_leaf_samples"]],
          json_object$get_vector("sigma2_leaf_samples", "parameters")
        )
      }
    }
  }
  if (
    model_params[["outcome_model"]]$link == "cloglog" &&
      model_params[["outcome_model"]]$outcome == "ordinal"
  ) {
    cloglog_cutpoint_samples <- matrix(
      NA_real_,
      model_params[["cloglog_num_categories"]] - 1,
      model_params[["num_samples"]]
    )
    index_start <- 1
    for (i in 1:length(json_object_list)) {
      json_object <- json_object_list[[i]]
      num_samples <- json_object$get_scalar("num_samples")
      subset_inds <- index_start:(index_start + num_samples - 1)
      for (j in 1:(model_params[["cloglog_num_categories"]] - 1)) {
        cloglog_cutpoint_samples[j, subset_inds] <- json_object$get_vector(
          paste0("cloglog_cutpoint_samples_", j),
          "parameters"
        )
      }
    }
    output[["cloglog_cutpoint_samples"]] <- cloglog_cutpoint_samples
  }

  # Unpack random effects
  if (model_params[["has_rfx"]]) {
    output[[
      "rfx_unique_group_ids"
    ]] <- json_object_default$get_string_vector("rfx_unique_group_ids")
    output[["rfx_samples"]] <- loadRandomEffectSamplesCombinedJson(
      json_object_list,
      0
    )
  }

  # Unpack covariate preprocessor
  if (has_field("preprocessor_metadata")) {
    preprocessor_metadata_string <- json_object_default$get_string(
      "preprocessor_metadata"
    )
    output[["train_set_metadata"]] <- createPreprocessorFromJsonString(
      preprocessor_metadata_string
    )
  } else {
    output[["train_set_metadata"]] <- NULL
    warning(paste0(
      "Field 'preprocessor_metadata' not found in JSON (model appears to have been serialized ",
      "under stochtree ",
      .ver,
      "). DataFrame covariates will not be supported for prediction. ",
      "Re-save your model to suppress this warning."
    ))
  }

  class(output) <- "bartmodel"
  return(output)
}

#' @title Convert JSON String List to Single BART Model
#' @rdname BARTSerialization
#' @param json_string_list List of JSON strings which can be parsed to objects of type `CppJson` containing Json representation of a BART model
#' @export
createBARTModelFromCombinedJsonString <- function(json_string_list) {
  # Initialize the BCF model
  output <- list()

  # Convert JSON strings
  json_object_list <- list()
  for (i in 1:length(json_string_list)) {
    json_string <- json_string_list[[i]]
    json_object_list[[i]] <- createCppJsonString(json_string)
  }

  # For scalar / preprocessing details which aren't sample-dependent,
  # defer to the first json
  json_object_default <- json_object_list[[1]]

  # Helpers for optional-field presence checks
  .ver <- inferStochtreeJsonVersion(json_object_default)
  has_field <- function(name) {
    json_contains_field_cpp(json_object_default$json_ptr, name)
  }
  has_subfolder_field <- function(subfolder, name) {
    json_contains_field_subfolder_cpp(
      json_object_default$json_ptr,
      subfolder,
      name
    )
  }

  # Unpack the forests
  include_mean_forest <- json_object_default$get_boolean(
    "include_mean_forest"
  )
  include_variance_forest <- json_object_default$get_boolean(
    "include_variance_forest"
  )
  if (include_mean_forest) {
    output[["mean_forests"]] <- loadForestContainerCombinedJson(
      json_object_list,
      "forest_0"
    )
    if (include_variance_forest) {
      output[["variance_forests"]] <- loadForestContainerCombinedJson(
        json_object_list,
        "forest_1"
      )
    }
  } else {
    output[["variance_forests"]] <- loadForestContainerCombinedJson(
      json_object_list,
      "forest_0"
    )
  }

  # Unpack metadata
  train_set_metadata = list()
  train_set_metadata[["num_numeric_vars"]] <- json_object_default$get_scalar(
    "num_numeric_vars"
  )
  train_set_metadata[[
    "num_ordered_cat_vars"
  ]] <- json_object_default$get_scalar("num_ordered_cat_vars")
  train_set_metadata[[
    "num_unordered_cat_vars"
  ]] <- json_object_default$get_scalar("num_unordered_cat_vars")
  if (train_set_metadata[["num_numeric_vars"]] > 0) {
    train_set_metadata[[
      "numeric_vars"
    ]] <- json_object_default$get_string_vector("numeric_vars")
  }
  if (train_set_metadata[["num_ordered_cat_vars"]] > 0) {
    train_set_metadata[[
      "ordered_cat_vars"
    ]] <- json_object_default$get_string_vector("ordered_cat_vars")
    train_set_metadata[[
      "ordered_unique_levels"
    ]] <- json_object_default$get_string_list(
      "ordered_unique_levels",
      train_set_metadata[["ordered_cat_vars"]]
    )
  }
  if (train_set_metadata[["num_unordered_cat_vars"]] > 0) {
    train_set_metadata[[
      "unordered_cat_vars"
    ]] <- json_object_default$get_string_vector("unordered_cat_vars")
    train_set_metadata[[
      "unordered_unique_levels"
    ]] <- json_object_default$get_string_list(
      "unordered_unique_levels",
      train_set_metadata[["unordered_cat_vars"]]
    )
  }
  output[["train_set_metadata"]] <- train_set_metadata

  # Unpack model params
  model_params = list()
  model_params[["outcome_scale"]] <- json_object_default$get_scalar(
    "outcome_scale"
  )
  model_params[["outcome_mean"]] <- json_object_default$get_scalar(
    "outcome_mean"
  )
  model_params[["standardize"]] <- json_object_default$get_boolean(
    "standardize"
  )
  model_params[["sigma2_init"]] <- json_object_default$get_scalar(
    "sigma2_init"
  )
  model_params[["sample_sigma2_global"]] <- json_object_default$get_boolean(
    "sample_sigma2_global"
  )
  model_params[["sample_sigma2_leaf"]] <- json_object_default$get_boolean(
    "sample_sigma2_leaf"
  )
  model_params[["include_mean_forest"]] <- include_mean_forest
  model_params[["include_variance_forest"]] <- include_variance_forest
  model_params[["has_rfx"]] <- json_object_default$get_boolean("has_rfx")

  if (has_field("has_rfx_basis")) {
    model_params[["has_rfx_basis"]] <- json_object_default$get_boolean(
      "has_rfx_basis"
    )
    model_params[["num_rfx_basis"]] <- json_object_default$get_scalar(
      "num_rfx_basis"
    )
  } else {
    model_params[["has_rfx_basis"]] <- FALSE
    model_params[["num_rfx_basis"]] <- 1
    warning(paste0(
      "Fields 'has_rfx_basis' and 'num_rfx_basis' not found in JSON (model appears to have been ",
      "serialized under stochtree ",
      .ver,
      "). Defaulting to FALSE / 1. ",
      "Re-save your model to suppress this warning."
    ))
  }

  model_params[["num_covariates"]] <- if (has_field("num_covariates")) {
    json_object_default$get_scalar("num_covariates")
  } else {
    NA_real_
  }
  model_params[["num_basis"]] <- json_object_default$get_scalar("num_basis")
  model_params[["requires_basis"]] <- json_object_default$get_boolean(
    "requires_basis"
  )

  model_params[["probit_outcome_model"]] <- if (
    has_field("probit_outcome_model")
  ) {
    json_object_default$get_boolean("probit_outcome_model")
  } else {
    FALSE
  }

  if (
    has_subfolder_field("outcome_model", "outcome") &&
      has_subfolder_field("outcome_model", "link")
  ) {
    outcome_model_outcome <- json_object_default$get_string(
      "outcome",
      "outcome_model"
    )
    outcome_model_link <- json_object_default$get_string(
      "link",
      "outcome_model"
    )
  } else {
    outcome_model_outcome <- "continuous"
    outcome_model_link <- "identity"
    warning(paste0(
      "Fields 'outcome' and 'link' not found under 'outcome_model' in JSON (model appears to have ",
      "been serialized under stochtree ",
      .ver,
      "). Defaulting to outcome='continuous', ",
      "link='identity'. Re-save your model to suppress this warning."
    ))
  }
  model_params[["outcome_model"]] <- OutcomeModel(
    outcome = outcome_model_outcome,
    link = outcome_model_link
  )

  if (has_field("rfx_model_spec")) {
    model_params[["rfx_model_spec"]] <- json_object_default$get_string(
      "rfx_model_spec"
    )
  } else {
    model_params[["rfx_model_spec"]] <- ""
    if (model_params[["has_rfx"]]) {
      warning(paste0(
        "Field 'rfx_model_spec' not found in JSON (model appears to have been serialized under ",
        "stochtree ",
        .ver,
        "). Defaulting to ''. Re-save your model to suppress this warning."
      ))
    }
  }

  if (has_field("num_chains")) {
    model_params[["num_chains"]] <- json_object_default$get_scalar("num_chains")
  } else {
    model_params[["num_chains"]] <- 1
    warning(paste0(
      "Field 'num_chains' not found in JSON (model appears to have been serialized under stochtree ",
      .ver,
      "). Defaulting to 1. Re-save your model to suppress this warning."
    ))
  }

  if (has_field("keep_every")) {
    model_params[["keep_every"]] <- json_object_default$get_scalar("keep_every")
  } else {
    model_params[["keep_every"]] <- 1
    warning(paste0(
      "Field 'keep_every' not found in JSON (model appears to have been serialized under stochtree ",
      .ver,
      "). Defaulting to 1. Re-save your model to suppress this warning."
    ))
  }

  if (model_params[["outcome_model"]]$link == "cloglog") {
    cloglog_num_categories <- json_object_default$get_scalar(
      "cloglog_num_categories"
    )
    model_params[["cloglog_num_categories"]] <- cloglog_num_categories
  } else {
    model_params[["cloglog_num_categories"]] <- 0
  }

  # Combine values that are sample-specific
  for (i in 1:length(json_object_list)) {
    json_object <- json_object_list[[i]]
    if (i == 1) {
      model_params[["num_gfr"]] <- json_object$get_scalar("num_gfr")
      model_params[["num_burnin"]] <- json_object$get_scalar("num_burnin")
      model_params[["num_mcmc"]] <- json_object$get_scalar("num_mcmc")
      model_params[["num_samples"]] <- json_object$get_scalar(
        "num_samples"
      )
    } else {
      prev_json <- json_object_list[[i - 1]]
      model_params[["num_gfr"]] <- model_params[["num_gfr"]] +
        json_object$get_scalar("num_gfr")
      model_params[["num_burnin"]] <- model_params[["num_burnin"]] +
        json_object$get_scalar("num_burnin")
      model_params[["num_mcmc"]] <- model_params[["num_mcmc"]] +
        json_object$get_scalar("num_mcmc")
      model_params[["num_samples"]] <- model_params[["num_samples"]] +
        json_object$get_scalar("num_samples")
    }
  }
  output[["model_params"]] <- model_params

  # Unpack sampled parameters
  if (model_params[["sample_sigma2_global"]]) {
    for (i in 1:length(json_object_list)) {
      json_object <- json_object_list[[i]]
      if (i == 1) {
        output[["sigma2_global_samples"]] <- json_object$get_vector(
          "sigma2_global_samples",
          "parameters"
        )
      } else {
        output[["sigma2_global_samples"]] <- c(
          output[["sigma2_global_samples"]],
          json_object$get_vector(
            "sigma2_global_samples",
            "parameters"
          )
        )
      }
    }
  }
  if (model_params[["sample_sigma2_leaf"]]) {
    for (i in 1:length(json_object_list)) {
      json_object <- json_object_list[[i]]
      if (i == 1) {
        output[["sigma2_leaf_samples"]] <- json_object$get_vector(
          "sigma2_leaf_samples",
          "parameters"
        )
      } else {
        output[["sigma2_leaf_samples"]] <- c(
          output[["sigma2_leaf_samples"]],
          json_object$get_vector("sigma2_leaf_samples", "parameters")
        )
      }
    }
  }
  if (
    model_params[["outcome_model"]]$link == "cloglog" &&
      model_params[["outcome_model"]]$outcome == "ordinal"
  ) {
    cloglog_cutpoint_samples <- matrix(
      NA_real_,
      model_params[["cloglog_num_categories"]] - 1,
      model_params[["num_samples"]]
    )
    index_start <- 1
    for (i in 1:length(json_object_list)) {
      json_object <- json_object_list[[i]]
      num_samples <- json_object$get_scalar("num_samples")
      subset_inds <- index_start:(index_start + num_samples - 1)
      for (j in 1:(model_params[["cloglog_num_categories"]] - 1)) {
        cloglog_cutpoint_samples[j, subset_inds] <- json_object$get_vector(
          paste0("cloglog_cutpoint_samples_", j),
          "parameters"
        )
      }
    }
    output[["cloglog_cutpoint_samples"]] <- cloglog_cutpoint_samples
  }

  # Unpack random effects
  if (model_params[["has_rfx"]]) {
    output[[
      "rfx_unique_group_ids"
    ]] <- json_object_default$get_string_vector("rfx_unique_group_ids")
    output[["rfx_samples"]] <- loadRandomEffectSamplesCombinedJson(
      json_object_list,
      0
    )
  }

  # Unpack covariate preprocessor
  if (has_field("preprocessor_metadata")) {
    preprocessor_metadata_string <- json_object_default$get_string(
      "preprocessor_metadata"
    )
    output[["train_set_metadata"]] <- createPreprocessorFromJsonString(
      preprocessor_metadata_string
    )
  } else {
    output[["train_set_metadata"]] <- NULL
    warning(paste0(
      "Field 'preprocessor_metadata' not found in JSON (model appears to have been serialized ",
      "under stochtree ",
      .ver,
      "). DataFrame covariates will not be supported for prediction. ",
      "Re-save your model to suppress this warning."
    ))
  }

  class(output) <- "bartmodel"
  return(output)
}

test_that("BART print method", {
  skip_on_cran()

  # Generate simulated data
  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  f_X <- 5 * X[, 1]
  rfx_group_ids <- sample(1:4, size = n, replace = TRUE)
  rfx_basis <- cbind(rep(1, n), runif(n))
  y <- f_X + rnorm(n)
  test_inds <- sort(sample(1:n, 20))
  train_inds <- setdiff(1:n, test_inds)
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  y_test <- y[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_basis_train <- rfx_basis[train_inds, ]
  rfx_basis_test <- rfx_basis[test_inds, ]

  # --- 1 model term: mean forest only (no global variance, no leaf scale) ---
  bart_model_1 <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  out_1 <- capture.output(result_1 <- print(bart_model_1))
  # Return value is the model object
  expect_identical(result_1, bart_model_1)
  # Key strings in output
  expect_true(any(grepl("stochtree::bart()", out_1, fixed = TRUE)))
  expect_true(any(grepl("mean forest", out_1, fixed = TRUE)))
  expect_true(any(grepl("constant leaf prior", out_1, fixed = TRUE)))
  expect_true(any(grepl("Outcome was standardized", out_1, fixed = TRUE)))
  expect_true(any(grepl("1 chain of", out_1, fixed = TRUE)))
  expect_true(any(grepl("retaining every iteration", out_1, fixed = TRUE)))

  # --- 2 model terms: mean forest + global error variance ---
  bart_model_2 <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = TRUE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  out_2 <- capture.output(print(bart_model_2))
  expect_true(any(grepl("mean forest and global error variance model", out_2, fixed = TRUE)))

  # --- >2 model terms: mean forest + global error variance + leaf scale (Oxford comma) ---
  bart_model_3 <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = TRUE),
    mean_forest_params = list(sample_sigma2_leaf = TRUE)
  )
  out_3 <- capture.output(print(bart_model_3))
  expect_true(any(grepl(", and mean forest leaf scale model", out_3, fixed = TRUE)))

  # --- Leaf regression: basis supplied, exercises leaf_regression branch ---
  basis_train <- matrix(runif(length(y_train)), ncol = 1)
  basis_test <- matrix(runif(length(y_test)), ncol = 1)
  bart_model_lr <- bart(
    X_train = X_train, y_train = y_train, leaf_basis_train = basis_train,
    X_test = X_test, leaf_basis_test = basis_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  out_lr <- capture.output(print(bart_model_lr))
  expect_true(any(grepl("leaf regression prior", out_lr, fixed = TRUE)))

  # --- Intercept-only random effects ---
  bart_model_rfx_1 <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    rfx_group_ids_train = rfx_group_ids_train,
    rfx_group_ids_test = rfx_group_ids_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE),
    random_effects_params = list(model_spec = "intercept_only")
  )
  out_rfx_1 <- capture.output(print(bart_model_rfx_1))
  expect_true(any(grepl("additive random effects", out_rfx_1, fixed = TRUE)))
  expect_true(any(grepl("intercept-only", out_rfx_1, fixed = TRUE)))

  # --- Custom (multi-component) random effects ---
  bart_model_rfx_2 <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    rfx_group_ids_train = rfx_group_ids_train,
    rfx_group_ids_test = rfx_group_ids_test,
    rfx_basis_train = rfx_basis_train,
    rfx_basis_test = rfx_basis_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE),
    random_effects_params = list(model_spec = "custom")
  )
  out_rfx_2 <- capture.output(print(bart_model_rfx_2))
  expect_true(any(grepl("additive random effects", out_rfx_2, fixed = TRUE)))
  expect_true(any(grepl("user-supplied basis", out_rfx_2, fixed = TRUE)))

  # --- GFR count is printed correctly ---
  bart_model_gfr <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    num_gfr = 10, num_burnin = 0, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  out_gfr <- capture.output(print(bart_model_gfr))
  expect_true(any(grepl("10 GFR iterations", out_gfr, fixed = TRUE)))

  # --- standardize = FALSE: no "standardized" line ---
  bart_model_nostd <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(standardize = FALSE, sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  out_nostd <- capture.output(print(bart_model_nostd))
  expect_false(any(grepl("standardized", out_nostd, fixed = TRUE)))

  # --- Multiple chains + thinning ---
  bart_model_mc <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(num_chains = 2, keep_every = 2, sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  out_mc <- capture.output(print(bart_model_mc))
  expect_true(any(grepl("2 chains of", out_mc, fixed = TRUE)))
  expect_true(any(grepl("thinning", out_mc, fixed = TRUE)))
})

test_that("BART summary method", {
  skip_on_cran()

  # Generate simulated data
  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  y <- 5 * X[, 1] + rnorm(n)
  rfx_group_ids <- sample(1:4, size = n, replace = TRUE)
  rfx_basis <- cbind(rep(1, n), runif(n))
  test_inds <- sort(sample(1:n, 20))
  train_inds <- setdiff(1:n, test_inds)
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  y_test <- y[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_basis_train <- rfx_basis[train_inds, ]
  rfx_basis_test <- rfx_basis[test_inds, ]

  # With sigma2_global, sigma2_leaf, and a test set
  bart_model <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = TRUE),
    mean_forest_params = list(sample_sigma2_leaf = TRUE)
  )
  out <- capture.output(result <- summary(bart_model))
  # Return value is the model object
  expect_identical(result, bart_model)
  # Sampled quantities are summarized
  expect_true(any(grepl("sigma^2", out, fixed = TRUE)))
  expect_true(any(grepl("leaf scale", out, fixed = TRUE)))
  # Both in-sample and test-set prediction summaries appear
  expect_true(any(grepl("in-sample", out, fixed = TRUE)))
  expect_true(any(grepl("test-set", out, fixed = TRUE)))

  # Without a test set: no test-set summary line
  bart_model_notestset <- bart(
    X_train = X_train, y_train = y_train,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  out_notestset <- capture.output(summary(bart_model_notestset))
  expect_false(any(grepl("test-set", out_notestset, fixed = TRUE)))

  # Intercept-only RFX: single-component random effects summary
  bart_model_rfx_1 <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    rfx_group_ids_train = rfx_group_ids_train,
    rfx_group_ids_test = rfx_group_ids_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE),
    random_effects_params = list(model_spec = "intercept_only")
  )
  out_rfx_1 <- capture.output(summary(bart_model_rfx_1))
  expect_true(any(grepl("Random effects", out_rfx_1, fixed = TRUE)))
  expect_true(any(grepl("Random effects overall mean", out_rfx_1, fixed = TRUE)))

  # Custom (multi-component) RFX: multi-component random effects summary
  bart_model_rfx_2 <- bart(
    X_train = X_train, y_train = y_train, X_test = X_test,
    rfx_group_ids_train = rfx_group_ids_train,
    rfx_group_ids_test = rfx_group_ids_test,
    rfx_basis_train = rfx_basis_train,
    rfx_basis_test = rfx_basis_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE),
    random_effects_params = list(model_spec = "custom")
  )
  out_rfx_2 <- capture.output(summary(bart_model_rfx_2))
  expect_true(any(grepl("Random effects", out_rfx_2, fixed = TRUE)))
  expect_true(any(grepl("Variance component means", out_rfx_2, fixed = TRUE)))
})

test_that("BCF print method", {
  skip_on_cran()

  # Generate simulated data
  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  pi_X <- 0.2 + 0.6 * X[, 1]
  Z <- rbinom(n, 1, pi_X)
  mu_X <- 5 * X[, 1]
  tau_X <- 2 * X[, 2]
  y <- mu_X + tau_X * Z + rnorm(n)
  test_inds <- sort(sample(1:n, 20))
  train_inds <- setdiff(1:n, test_inds)
  X_train <- X[train_inds, ]; X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]; Z_test <- Z[test_inds]
  pi_train <- pi_X[train_inds]; pi_test <- pi_X[test_inds]
  y_train <- y[train_inds]; y_test <- y[test_inds]

  # --- User-provided propensity, binary treatment, default coding (defaults) ---
  bcf_model <- bcf(
    X_train = X_train, y_train = y_train, Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test, Z_test = Z_test, propensity_test = pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10
  )
  out <- capture.output(result <- print(bcf_model))
  # Return value is the model object
  expect_identical(result, bcf_model)
  # Key strings in output
  expect_true(any(grepl("stochtree::bcf()", out, fixed = TRUE)))
  expect_true(any(grepl("prognostic forest", out, fixed = TRUE)))
  expect_true(any(grepl("treatment effect forest", out, fixed = TRUE)))
  expect_true(any(grepl("User-provided propensity scores", out, fixed = TRUE)))
  expect_true(any(grepl("default coding", out, fixed = TRUE)))
  expect_true(any(grepl("1 chain of", out, fixed = TRUE)))
  expect_true(any(grepl("retaining every iteration", out, fixed = TRUE)))

  # --- Binary treatment, adaptive coding disabled ---
  bcf_model_noac <- bcf(
    X_train = X_train, y_train = y_train, Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test, Z_test = Z_test, propensity_test = pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(adaptive_coding = FALSE)
  )
  out_noac <- capture.output(print(bcf_model_noac))
  expect_true(any(grepl("default coding", out_noac, fixed = TRUE)))

  # --- Propensity excluded from both forests ---
  bcf_model_noprop <- bcf(
    X_train = X_train, y_train = y_train, Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test, Z_test = Z_test, propensity_test = pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(propensity_covariate = "none")
  )
  out_noprop <- capture.output(print(bcf_model_noprop))
  expect_true(any(grepl("not used in either forest", out_noprop, fixed = TRUE)))

  # --- Continuous (non-binary) treatment ---
  Z_cont_train <- rnorm(length(train_inds))
  Z_cont_test <- rnorm(length(test_inds))
  y_cont_train <- mu_X[train_inds] + 2 * Z_cont_train + rnorm(length(train_inds))
  bcf_model_cont <- bcf(
    X_train = X_train, y_train = y_cont_train, Z_train = Z_cont_train,
    X_test = X_test, Z_test = Z_cont_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(propensity_covariate = "none")
  )
  out_cont <- capture.output(print(bcf_model_cont))
  expect_true(any(grepl("univariate but not binary", out_cont, fixed = TRUE)))

  # --- Multiple chains + thinning ---
  bcf_model_mc <- bcf(
    X_train = X_train, y_train = y_train, Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test, Z_test = Z_test, propensity_test = pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(num_chains = 2, keep_every = 2)
  )
  out_mc <- capture.output(print(bcf_model_mc))
  expect_true(any(grepl("2 chains of", out_mc, fixed = TRUE)))
  expect_true(any(grepl("thinning", out_mc, fixed = TRUE)))
})

test_that("BCF summary method", {
  skip_on_cran()

  # Generate simulated data
  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  pi_X <- 0.2 + 0.6 * X[, 1]
  Z <- rbinom(n, 1, pi_X)
  mu_X <- 5 * X[, 1]
  tau_X <- 2 * X[, 2]
  y <- mu_X + tau_X * Z + rnorm(n)
  test_inds <- sort(sample(1:n, 20))
  train_inds <- setdiff(1:n, test_inds)
  X_train <- X[train_inds, ]; X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]; Z_test <- Z[test_inds]
  pi_train <- pi_X[train_inds]; pi_test <- pi_X[test_inds]
  y_train <- y[train_inds]; y_test <- y[test_inds]

  # With sigma2_global, both leaf scales, adaptive coding, and a test set
  bcf_model <- bcf(
    X_train = X_train, y_train = y_train, Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test, Z_test = Z_test, propensity_test = pi_test,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10,
    general_params = list(sample_sigma2_global = TRUE, adaptive_coding = TRUE),
    prognostic_forest_params = list(sample_sigma2_leaf = TRUE),
    treatment_effect_forest_params = list(sample_sigma2_leaf = TRUE)
  )
  out <- capture.output(result <- summary(bcf_model))
  # Return value is the model object
  expect_identical(result, bcf_model)
  # Sampled quantities are summarized
  expect_true(any(grepl("sigma^2", out, fixed = TRUE)))
  expect_true(any(grepl("prognostic forest leaf scale", out, fixed = TRUE)))
  expect_true(any(grepl("treatment effect forest leaf scale", out, fixed = TRUE)))
  expect_true(any(grepl("adaptive coding parameters", out, fixed = TRUE)))
  # Both in-sample and test-set prediction summaries appear
  expect_true(any(grepl("in-sample", out, fixed = TRUE)))
  expect_true(any(grepl("test-set", out, fixed = TRUE)))
  # CATE summaries appear
  expect_true(any(grepl("CATEs", out, fixed = TRUE)))

  # Without a test set: no test-set summary lines
  bcf_model_notestset <- bcf(
    X_train = X_train, y_train = y_train, Z_train = Z_train,
    propensity_train = pi_train,
    num_gfr = 0, num_burnin = 10, num_mcmc = 10
  )
  out_notestset <- capture.output(summary(bcf_model_notestset))
  expect_false(any(grepl("test-set", out_notestset, fixed = TRUE)))
})

test_that("MCMC BART", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # 1 chain, no thinning
  general_param_list <- list(num_chains = 1, keep_every = 1)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 3 chains, no thinning
  general_param_list <- list(num_chains = 3, keep_every = 1)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 1 chain, thinning
  general_param_list <- list(num_chains = 1, keep_every = 5)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 3 chains, thinning
  general_param_list <- list(num_chains = 3, keep_every = 5)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Generate simulated data with a leaf basis
  n <- 100
  p <- 5
  p_w <- 2
  X <- matrix(runif(n * p), ncol = p)
  W <- matrix(runif(n * p_w), ncol = p_w)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5 * W[, 1]) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5 * W[, 1]) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5 * W[, 1]) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5 * W[, 1]))
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  W_test <- W[test_inds, ]
  W_train <- W[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # 3 chains, thinning, leaf regression
  general_param_list <- list(num_chains = 3, keep_every = 5)
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      mean_forest_params = mean_forest_param_list
    )
  )

  # 3 chains, thinning, leaf regression with a scalar leaf scale
  general_param_list <- list(num_chains = 3, keep_every = 5)
  mean_forest_param_list <- list(
    sample_sigma2_leaf = FALSE,
    sigma2_leaf_init = 0.5
  )
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      mean_forest_params = mean_forest_param_list
    )
  )

  # 3 chains, thinning, leaf regression with a scalar leaf scale, random leaf scale
  general_param_list <- list(num_chains = 3, keep_every = 5)
  mean_forest_param_list <- list(sample_sigma2_leaf = T, sigma2_leaf_init = 0.5)
  expect_warning(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      mean_forest_params = mean_forest_param_list
    )
  )
})

test_that("GFR BART", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # 1 chain, no thinning
  general_param_list <- list(num_chains = 1, keep_every = 1)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 3 chains, no thinning
  general_param_list <- list(num_chains = 3, keep_every = 1)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 1 chain, thinning
  general_param_list <- list(num_chains = 1, keep_every = 5)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # 3 chains, thinning
  general_param_list <- list(num_chains = 3, keep_every = 5)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check for error when more chains than GFR forests
  general_param_list <- list(num_chains = 11, keep_every = 1)
  expect_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check for error when more chains than GFR forests
  general_param_list <- list(num_chains = 11, keep_every = 5)
  expect_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )
})

test_that("Warmstart BART", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Run a BART model with only GFR
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 0,
    general_params = general_param_list
  )

  # Save to JSON string
  bart_model_json_string <- saveBARTModelToJsonString(bart_model)

  # Run a new BART chain from the existing (X)BART model
  general_param_list <- list(num_chains = 3, keep_every = 5)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      previous_model_json = bart_model_json_string,
      previous_model_warmstart_sample_num = 10,
      general_params = general_param_list
    )
  )
  expect_warning(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      previous_model_json = bart_model_json_string,
      previous_model_warmstart_sample_num = 1,
      general_params = general_param_list
    )
  )

  # Generate simulated data with random effects
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
  rfx_basis <- rep(1, n)
  rfx_coefs <- c(-5, 5)
  rfx_term <- rfx_coefs[rfx_group_ids] * rfx_basis
  noise_sd <- 1
  y <- f_XW + rfx_term + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds]
  rfx_basis_train <- rfx_basis[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Run a BART model with only GFR
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    rfx_group_ids_train = rfx_group_ids_train,
    rfx_group_ids_test = rfx_group_ids_test,
    rfx_basis_train = rfx_basis_train,
    rfx_basis_test = rfx_basis_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 0,
    general_params = general_param_list
  )

  # Save to JSON string
  bart_model_json_string <- saveBARTModelToJsonString(bart_model)

  # Run a new BART chain from the existing (X)BART model
  general_param_list <- list(num_chains = 4, keep_every = 5)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      previous_model_json = bart_model_json_string,
      previous_model_warmstart_sample_num = 10,
      general_params = general_param_list
    )
  )
  expect_warning(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      previous_model_json = bart_model_json_string,
      previous_model_warmstart_sample_num = 1,
      general_params = general_param_list
    )
  )
})

test_that("BART Predictions", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5))
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Run a BART model with only GFR
  general_params <- list(num_chains = 1, sample_sigma2_global = FALSE)
  variance_forest_params <- list(num_trees = 50)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 10,
    num_burnin = 0,
    num_mcmc = 10,
    general_params = general_params,
    variance_forest_params = variance_forest_params
  )

  # Check that cached predictions agree with results of predict() function
  train_preds <- predict(bart_model, X = X_train)
  train_preds_mean_cached <- bart_model$y_hat_train
  train_preds_mean_recomputed <- train_preds$mean_forest_predictions
  train_preds_variance_cached <- bart_model$sigma2_x_hat_train
  train_preds_variance_recomputed <- train_preds$variance_forest_predictions

  # Assertion
  expect_equal(train_preds_mean_cached, train_preds_mean_recomputed)
  expect_equal(train_preds_variance_cached, train_preds_variance_recomputed)
})

test_that("Random Effects BART", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  p_w <- 2
  X <- matrix(runif(n * p), ncol = p)
  W <- matrix(runif(n * p_w), ncol = p_w)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5 * W[, 1]) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5 * W[, 1]) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5 * W[, 1]) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5 * W[, 1]))
  rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
  rfx_basis <- cbind(rep(1, n), runif(n))
  num_rfx_components <- ncol(rfx_basis)
  num_rfx_groups <- length(unique(rfx_group_ids))
  rfx_coefs <- matrix(c(-5, 5, 1, -1), ncol = 2, byrow = T)
  rfx_term <- rowSums(rfx_coefs[rfx_group_ids, ] * rfx_basis)
  noise_sd <- 1
  y <- f_XW + rfx_term + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  W_test <- W[test_inds, ]
  W_train <- W[train_inds, ]
  rfx_group_ids_test <- rfx_group_ids[test_inds]
  rfx_group_ids_train <- rfx_group_ids[train_inds]
  rfx_basis_test <- rfx_basis[test_inds, ]
  rfx_basis_train <- rfx_basis[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Specify no rfx parameters
  general_param_list <- list()
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      mean_forest_params = mean_forest_param_list
    )
  )

  # Specify all rfx parameters as scalars
  rfx_param_list <- list(
    working_parameter_prior_mean = 1.,
    group_parameter_prior_mean = 1.,
    working_parameter_prior_cov = 1.,
    group_parameter_prior_cov = 1.,
    variance_prior_shape = 1,
    variance_prior_scale = 1
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
  )

  # Specify all relevant rfx parameters as vectors
  rfx_param_list <- list(
    working_parameter_prior_mean = c(1., 1.),
    group_parameter_prior_mean = c(1., 1.),
    working_parameter_prior_cov = diag(1., 2),
    group_parameter_prior_cov = diag(1., 2),
    variance_prior_shape = 1,
    variance_prior_scale = 1
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      rfx_basis_train = rfx_basis_train,
      rfx_basis_test = rfx_basis_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
  )

  # Specify simpler intercept-only RFX model
  rfx_param_list <- list(
    model_spec = "intercept_only"
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error({
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      rfx_group_ids_train = rfx_group_ids_train,
      rfx_group_ids_test = rfx_group_ids_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      mean_forest_params = mean_forest_param_list,
      random_effects_params = rfx_param_list
    )
    preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      rfx_group_ids = rfx_group_ids_test,
      type = "posterior",
      terms = "rfx"
    )
  })
})

test_that("Probit BART", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  p_w <- 2
  X <- matrix(runif(n * p), ncol = p)
  W <- matrix(runif(n * p_w), ncol = p_w)
  # fmt: skip
  f_XW <- (((0 <= X[, 1]) & (0.25 > X[, 1])) * (-7.5 * W[, 1]) +
    ((0.25 <= X[, 1]) & (0.5 > X[, 1])) * (-2.5 * W[, 1]) +
    ((0.5 <= X[, 1]) & (0.75 > X[, 1])) * (2.5 * W[, 1]) +
    ((0.75 <= X[, 1]) & (1 > X[, 1])) * (7.5 * W[, 1]))
  z <- f_XW + rnorm(n, 0, 1)
  y <- as.numeric(z > 0)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  W_test <- W[test_inds, ]
  W_train <- W[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit probit model
  general_param_list <- list(
    outcome_model = OutcomeModel(outcome = "binary", link = "probit"),
    sample_sigma2_global = F
  )
  mean_forest_param_list <- list(sample_sigma2_leaf = FALSE)
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      leaf_basis_train = W_train,
      leaf_basis_test = W_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list,
      mean_forest_params = mean_forest_param_list
    )
  )

  # Predict from model on linear scale
  expect_no_error({
    preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      type = "posterior",
      scale = "linear"
    )
  })

  # Predict from model on probability scale
  expect_no_error({
    preds <- predict(
      bart_model,
      X = X_test,
      leaf_basis = W_test,
      type = "posterior",
      scale = "probability"
    )
  })
})

test_that("Cloglog Binary BART", {
  skip_on_cran()

  # Generate simulated data
  n <- 200
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  f_X <- 2 * (X[, 1] > 0.5) - 1
  prob <- 1 - exp(-exp(f_X))
  y <- rbinom(n, 1, prob)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit cloglog binary model (MCMC only)
  general_param_list <- list(
    outcome_model = OutcomeModel(outcome = "binary", link = "cloglog"),
    sample_sigma2_global = FALSE,
    num_chains = 1
  )
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check model outputs
  expect_true(!is.null(bart_model$y_hat_train))
  expect_true(!is.null(bart_model$y_hat_test))
  expect_true(is.null(bart_model$cloglog_cutpoint_samples))
  expect_equal(nrow(bart_model$y_hat_train), n_train)
  expect_equal(ncol(bart_model$y_hat_train), 10)
  expect_equal(nrow(bart_model$y_hat_test), n_test)
  expect_equal(ncol(bart_model$y_hat_test), 10)

  # Predict from model on linear scale (terms = "y_hat" for single matrix return)
  expect_no_error({
    preds_linear <- predict(
      bart_model,
      X = X_test,
      type = "posterior",
      scale = "linear",
      terms = "y_hat"
    )
  })
  expect_equal(nrow(preds_linear), n_test)
  expect_equal(ncol(preds_linear), 10)

  # Predict from model on probability scale
  expect_no_error({
    preds_prob <- predict(
      bart_model,
      X = X_test,
      type = "posterior",
      scale = "probability",
      terms = "y_hat"
    )
  })
  expect_equal(nrow(preds_prob), n_test)
  expect_equal(ncol(preds_prob), 10)

  # Predict posterior mean on linear scale
  expect_no_error({
    preds_mean <- predict(
      bart_model,
      X = X_test,
      type = "mean",
      scale = "linear",
      terms = "y_hat"
    )
  })
  expect_equal(length(preds_mean), n_test)

  # Predict posterior mean on probability scale
  expect_no_error({
    preds_mean_prob <- predict(
      bart_model,
      X = X_test,
      type = "mean",
      scale = "probability",
      terms = "y_hat"
    )
  })
  expect_equal(length(preds_mean_prob), n_test)

  # Predict class labels
  expect_no_error({
    preds_class <- predict(
      bart_model,
      X = X_test,
      type = "posterior",
      scale = "class",
      terms = "y_hat"
    )
  })
  expect_true(all(preds_class >= 0 & preds_class <= 1))
})

test_that("Cloglog Binary BART with GFR", {
  skip_on_cran()

  # Generate simulated data
  n <- 200
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  f_X <- 2 * (X[, 1] > 0.5) - 1
  prob <- 1 - exp(-exp(f_X))
  y <- rbinom(n, 1, prob)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit cloglog binary model with GFR warmstart
  general_param_list <- list(
    outcome_model = OutcomeModel(outcome = "binary", link = "cloglog"),
    sample_sigma2_global = FALSE,
    num_chains = 1
  )
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check model outputs
  expect_true(!is.null(bart_model$y_hat_train))
  expect_true(!is.null(bart_model$y_hat_test))
  expect_true(is.null(bart_model$cloglog_cutpoint_samples))
  expect_equal(nrow(bart_model$y_hat_train), n_train)
  expect_equal(ncol(bart_model$y_hat_train), 10)
})

test_that("Cloglog Ordinal BART", {
  skip_on_cran()

  # Generate simulated ordinal data (3 categories)
  n <- 300
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  f_X <- 2 * (X[, 1] > 0.5) - 1
  gamma_true <- c(-1.0, 0.5)
  n_categories <- 3

  # Compute ordinal class probabilities
  true_probs <- matrix(0, n, n_categories)
  true_probs[, 1] <- 1 - exp(-exp(gamma_true[1] + f_X))
  true_probs[, 2] <- exp(-exp(gamma_true[1] + f_X)) *
    (1 - exp(-exp(gamma_true[2] + f_X)))
  true_probs[, 3] <- 1 - true_probs[, 1] - true_probs[, 2]

  # Generate ordinal outcomes (1-indexed)
  y <- sapply(1:n, function(i) {
    sample(1:n_categories, 1, prob = true_probs[i, ])
  })

  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit cloglog ordinal model (MCMC only)
  general_param_list <- list(
    outcome_model = OutcomeModel(outcome = "ordinal", link = "cloglog"),
    sample_sigma2_global = FALSE,
    num_chains = 1
  )
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check model outputs
  expect_true(!is.null(bart_model$y_hat_train))
  expect_true(!is.null(bart_model$y_hat_test))
  expect_true(!is.null(bart_model$cloglog_cutpoint_samples))
  expect_equal(nrow(bart_model$y_hat_train), n_train)
  expect_equal(ncol(bart_model$y_hat_train), 10)
  expect_equal(nrow(bart_model$y_hat_test), n_test)
  expect_equal(ncol(bart_model$y_hat_test), 10)
  # 3 categories means 2 cutpoint rows
  expect_equal(nrow(bart_model$cloglog_cutpoint_samples), 2)
  expect_equal(ncol(bart_model$cloglog_cutpoint_samples), 10)
  expect_equal(bart_model$model_params$cloglog_num_categories, 3)

  # Predict from model on linear scale
  expect_no_error({
    preds_linear <- predict(
      bart_model,
      X = X_test,
      type = "posterior",
      scale = "linear",
      terms = "y_hat"
    )
  })
  expect_equal(nrow(preds_linear), n_test)
  expect_equal(ncol(preds_linear), 10)

  # Predict posterior mean on linear scale
  expect_no_error({
    preds_mean <- predict(
      bart_model,
      X = X_test,
      type = "mean",
      scale = "linear",
      terms = "y_hat"
    )
  })
  expect_equal(length(preds_mean), n_test)
})

test_that("Cloglog Ordinal BART with GFR", {
  skip_on_cran()

  # Generate simulated ordinal data (3 categories)
  n <- 300
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  f_X <- 0.5 * (X[, 1] > 0.5) - 0.25
  gamma_true <- c(-0.5, 0.5)
  n_categories <- 3

  # Compute ordinal class probabilities
  true_probs <- matrix(0, n, n_categories)
  true_probs[, 1] <- 1 - exp(-exp(gamma_true[1] + f_X))
  true_probs[, 2] <- exp(-exp(gamma_true[1] + f_X)) *
    (1 - exp(-exp(gamma_true[2] + f_X)))
  true_probs[, 3] <- 1 - true_probs[, 1] - true_probs[, 2]

  # Generate ordinal outcomes (1-indexed)
  y <- sapply(1:n, function(i) {
    sample(1:n_categories, 1, prob = true_probs[i, ])
  })

  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit cloglog ordinal model with GFR warmstart
  general_param_list <- list(
    outcome_model = OutcomeModel(outcome = "ordinal", link = "cloglog"),
    sample_sigma2_global = FALSE,
    num_chains = 1
  )
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check model outputs
  expect_true(!is.null(bart_model$y_hat_train))
  expect_true(!is.null(bart_model$y_hat_test))
  expect_true(!is.null(bart_model$cloglog_cutpoint_samples))
  expect_equal(nrow(bart_model$y_hat_train), n_train)
  expect_equal(ncol(bart_model$y_hat_train), 10)
  # 3 categories means 2 cutpoint rows
  expect_equal(nrow(bart_model$cloglog_cutpoint_samples), 2)
  expect_equal(ncol(bart_model$cloglog_cutpoint_samples), 10)
  expect_equal(bart_model$model_params$cloglog_num_categories, 3)
})

test_that("Cloglog BART multi-chain", {
  skip_on_cran()

  # Generate simulated data
  n <- 200
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  f_X <- 2 * (X[, 1] > 0.5) - 1
  prob <- 1 - exp(-exp(f_X))
  y <- rbinom(n, 1, prob)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit cloglog binary model with 2 chains
  general_param_list <- list(
    outcome_model = OutcomeModel(outcome = "binary", link = "cloglog"),
    sample_sigma2_global = FALSE,
    num_chains = 2
  )
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 10,
      num_burnin = 0,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Check model outputs: 2 chains x 10 MCMC = 20 total samples
  expect_true(!is.null(bart_model$y_hat_train))
  expect_true(is.null(bart_model$cloglog_cutpoint_samples))
  expect_equal(nrow(bart_model$y_hat_train), n_train)
  expect_equal(ncol(bart_model$y_hat_train), 20)
})

test_that("Cloglog BART JSON round-trip", {
  skip_on_cran()

  # Generate simulated ordinal data (3 categories)
  n <- 200
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  f_X <- 2 * (X[, 1] > 0.5) - 1
  gamma_true <- c(-1.0, 0.5)
  n_categories <- 3
  true_probs <- matrix(0, n, n_categories)
  true_probs[, 1] <- 1 - exp(-exp(gamma_true[1] + f_X))
  true_probs[, 2] <- exp(-exp(gamma_true[1] + f_X)) *
    (1 - exp(-exp(gamma_true[2] + f_X)))
  true_probs[, 3] <- 1 - true_probs[, 1] - true_probs[, 2]
  y <- sapply(1:n, function(i) {
    sample(1:n_categories, 1, prob = true_probs[i, ])
  })

  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Fit cloglog ordinal model
  general_param_list <- list(
    outcome_model = OutcomeModel(outcome = "ordinal", link = "cloglog"),
    sample_sigma2_global = FALSE,
    num_chains = 1
  )
  expect_no_error(
    bart_model <- bart(
      X_train = X_train,
      y_train = y_train,
      X_test = X_test,
      num_gfr = 0,
      num_burnin = 10,
      num_mcmc = 10,
      general_params = general_param_list
    )
  )

  # Save to JSON
  json_string <- saveBARTModelToJsonString(bart_model)
  expect_true(nchar(json_string) > 0)

  # Load from JSON
  expect_no_error(
    bart_model_loaded <- createBARTModelFromJsonString(json_string)
  )

  # Check that cloglog parameters survive round-trip
  expect_equal(
    bart_model_loaded$model_params$cloglog_num_categories,
    bart_model$model_params$cloglog_num_categories
  )
  expect_equal(
    bart_model_loaded$cloglog_cutpoint_samples,
    bart_model$cloglog_cutpoint_samples
  )

  # Check that forest predictions survive round-trip
  pred_orig <- predict(
    bart_model,
    X = X_test,
    type = "posterior",
    scale = "linear",
    terms = "y_hat"
  )
  pred_loaded <- predict(
    bart_model_loaded,
    X = X_test,
    type = "posterior",
    scale = "linear",
    terms = "y_hat"
  )
  expect_equal(pred_orig, pred_loaded)
})

test_that("extractParameter.bartmodel", {
  skip_on_cran()

  # Generate simulated data
  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  y <- 5 * X[, 1] + rnorm(n)
  test_inds <- sort(sample(1:n, 20))
  train_inds <- setdiff(1:n, test_inds)
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  y_train <- y[train_inds]
  n_train <- length(train_inds)
  n_test <- length(test_inds)
  num_mcmc <- 10

  # sigma2 / global_error_scale / sigma2_global all return same samples
  bart_sigma2 <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = TRUE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  s2 <- extractParameter(bart_sigma2, "sigma2")
  expect_length(s2, num_mcmc)
  expect_equal(s2, extractParameter(bart_sigma2, "global_error_scale"))
  expect_equal(s2, extractParameter(bart_sigma2, "sigma2_global"))

  # sigma2_leaf / leaf_scale both return leaf-scale samples
  bart_leaf <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = TRUE)
  )
  sl <- extractParameter(bart_leaf, "sigma2_leaf")
  expect_length(sl, num_mcmc)
  expect_equal(sl, extractParameter(bart_leaf, "leaf_scale"))

  # y_hat_train returns matrix of shape (n_train, num_mcmc)
  bart_base <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  yht <- extractParameter(bart_base, "y_hat_train")
  expect_equal(dim(yht), c(n_train, num_mcmc))

  # y_hat_test returns matrix of shape (n_test, num_mcmc)
  yhtest <- extractParameter(bart_base, "y_hat_test")
  expect_equal(dim(yhtest), c(n_test, num_mcmc))

  # sigma2_x_train / var_x_train return variance forest in-sample predictions
  bart_var <- bart(
    X_train = X_train,
    y_train = y_train,
    X_test = X_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE),
    variance_forest_params = list(num_trees = 10)
  )
  s2x_train <- extractParameter(bart_var, "sigma2_x_train")
  expect_equal(dim(s2x_train), c(n_train, num_mcmc))
  expect_equal(s2x_train, extractParameter(bart_var, "var_x_train"))

  # sigma2_x_test / var_x_test return variance forest test-set predictions
  s2x_test <- extractParameter(bart_var, "sigma2_x_test")
  expect_equal(dim(s2x_test), c(n_test, num_mcmc))
  expect_equal(s2x_test, extractParameter(bart_var, "var_x_test"))

  # Error: sigma2 when not sampled
  bart_nosigma2 <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    mean_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  expect_error(
    extractParameter(bart_nosigma2, "sigma2"),
    "global variance"
  )

  # Error: sigma2_leaf when not sampled
  expect_error(
    extractParameter(bart_nosigma2, "sigma2_leaf"),
    "leaf variance"
  )

  # Error: y_hat_test without a test set
  expect_error(
    extractParameter(bart_nosigma2, "y_hat_test"),
    "test set"
  )

  # Error: sigma2_x_train without variance forest
  expect_error(
    extractParameter(bart_nosigma2, "sigma2_x_train"),
    "variance forest"
  )

  # Error: invalid term
  expect_error(
    extractParameter(bart_nosigma2, "not_a_real_term"),
    "not a valid BART model term"
  )
})

test_that("extractParameter.bcfmodel", {
  skip_on_cran()

  # Generate simulated data
  set.seed(42)
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  pi_X <- 0.2 + 0.6 * X[, 1]
  Z <- rbinom(n, 1, pi_X)
  mu_X <- 5 * X[, 1]
  tau_X <- 2 * X[, 2]
  y <- mu_X + tau_X * Z + rnorm(n)
  test_inds <- sort(sample(1:n, 20))
  train_inds <- setdiff(1:n, test_inds)
  X_train <- X[train_inds, ]
  X_test <- X[test_inds, ]
  Z_train <- Z[train_inds]
  Z_test <- Z[test_inds]
  pi_train <- pi_X[train_inds]
  pi_test <- pi_X[test_inds]
  y_train <- y[train_inds]
  y_test <- y[test_inds]
  n_train <- length(train_inds)
  n_test <- length(test_inds)
  num_mcmc <- 10

  # sigma2 / global_error_scale / sigma2_global all return same samples
  bcf_sigma2 <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = TRUE)
  )
  s2 <- extractParameter(bcf_sigma2, "sigma2")
  expect_length(s2, num_mcmc)
  expect_equal(s2, extractParameter(bcf_sigma2, "global_error_scale"))
  expect_equal(s2, extractParameter(bcf_sigma2, "sigma2_global"))

  # sigma2_leaf_mu / leaf_scale_mu / mu_leaf_scale for prognostic forest
  bcf_leaf_mu <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    prognostic_forest_params = list(sample_sigma2_leaf = TRUE),
    treatment_effect_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  sl_mu <- extractParameter(bcf_leaf_mu, "sigma2_leaf_mu")
  expect_length(sl_mu, num_mcmc)
  expect_equal(sl_mu, extractParameter(bcf_leaf_mu, "leaf_scale_mu"))
  expect_equal(sl_mu, extractParameter(bcf_leaf_mu, "mu_leaf_scale"))

  # sigma2_leaf_tau / leaf_scale_tau / tau_leaf_scale for treatment effect forest
  bcf_leaf_tau <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    prognostic_forest_params = list(sample_sigma2_leaf = FALSE),
    treatment_effect_forest_params = list(sample_sigma2_leaf = TRUE)
  )
  sl_tau <- extractParameter(bcf_leaf_tau, "sigma2_leaf_tau")
  expect_length(sl_tau, num_mcmc)
  expect_equal(sl_tau, extractParameter(bcf_leaf_tau, "leaf_scale_tau"))
  expect_equal(sl_tau, extractParameter(bcf_leaf_tau, "tau_leaf_scale"))

  # adaptive_coding returns matrix with 2 rows (control, treated)
  bcf_ac <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE, adaptive_coding = TRUE)
  )
  ac <- extractParameter(bcf_ac, "adaptive_coding")
  expect_equal(dim(ac), c(2L, num_mcmc))

  # y_hat_train and y_hat_test
  bcf_base <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE)
  )
  yht <- extractParameter(bcf_base, "y_hat_train")
  expect_equal(dim(yht), c(n_train, num_mcmc))
  yhtest <- extractParameter(bcf_base, "y_hat_test")
  expect_equal(dim(yhtest), c(n_test, num_mcmc))

  # mu_hat_train / prognostic_function_train
  mht <- extractParameter(bcf_base, "mu_hat_train")
  expect_equal(dim(mht), c(n_train, num_mcmc))
  expect_equal(mht, extractParameter(bcf_base, "prognostic_function_train"))

  # mu_hat_test / prognostic_function_test
  mhtest <- extractParameter(bcf_base, "mu_hat_test")
  expect_equal(dim(mhtest), c(n_test, num_mcmc))
  expect_equal(mhtest, extractParameter(bcf_base, "prognostic_function_test"))

  # tau_hat_train / cate_train
  tht <- extractParameter(bcf_base, "tau_hat_train")
  expect_equal(dim(tht), c(n_train, num_mcmc))
  expect_equal(tht, extractParameter(bcf_base, "cate_train"))

  # tau_hat_test / cate_test
  thtest <- extractParameter(bcf_base, "tau_hat_test")
  expect_equal(dim(thtest), c(n_test, num_mcmc))
  expect_equal(thtest, extractParameter(bcf_base, "cate_test"))

  # sigma2_x_train / var_x_train and sigma2_x_test / var_x_test (variance forest)
  bcf_var <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    X_test = X_test,
    Z_test = Z_test,
    propensity_test = pi_test,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    variance_forest_params = list(num_trees = 10)
  )
  s2x_train <- extractParameter(bcf_var, "sigma2_x_train")
  expect_equal(dim(s2x_train), c(n_train, num_mcmc))
  expect_equal(s2x_train, extractParameter(bcf_var, "var_x_train"))

  s2x_test <- extractParameter(bcf_var, "sigma2_x_test")
  expect_equal(dim(s2x_test), c(n_test, num_mcmc))
  expect_equal(s2x_test, extractParameter(bcf_var, "var_x_test"))

  # Error: sigma2 when not sampled
  bcf_noextras <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE),
    prognostic_forest_params = list(sample_sigma2_leaf = FALSE),
    treatment_effect_forest_params = list(sample_sigma2_leaf = FALSE)
  )
  expect_error(
    extractParameter(bcf_noextras, "sigma2"),
    "global variance"
  )

  # Error: sigma2_leaf_mu when not sampled
  expect_error(
    extractParameter(bcf_noextras, "sigma2_leaf_mu"),
    "prognostic forest leaf variance"
  )

  # Error: sigma2_leaf_tau when not sampled
  expect_error(
    extractParameter(bcf_noextras, "sigma2_leaf_tau"),
    "treatment effect forest leaf variance"
  )

  # Error: adaptive_coding when disabled
  bcf_noac <- bcf(
    X_train = X_train,
    y_train = y_train,
    Z_train = Z_train,
    propensity_train = pi_train,
    num_gfr = 0,
    num_burnin = 0,
    num_mcmc = num_mcmc,
    general_params = list(sample_sigma2_global = FALSE, adaptive_coding = FALSE)
  )
  expect_error(
    extractParameter(bcf_noac, "adaptive_coding"),
    "adaptive coding"
  )

  # Error: y_hat_test without a test set
  expect_error(
    extractParameter(bcf_noextras, "y_hat_test"),
    "test set"
  )

  # Error: tau_hat_test without a test set
  expect_error(
    extractParameter(bcf_noextras, "tau_hat_test"),
    "test set"
  )

  # Error: sigma2_x_train without variance forest
  expect_error(
    extractParameter(bcf_noextras, "sigma2_x_train"),
    "variance forest"
  )

  # Error: invalid term
  expect_error(
    extractParameter(bcf_noextras, "not_a_real_term"),
    "not a valid BCF model term"
  )
})

test_that("BART Serialization", {
  skip_on_cran()

  # Generate simulated data
  n <- 100
  p <- 5
  X <- matrix(runif(n * p), ncol = p)
  # fmt: skip
  f_XW <- (
        ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) + 
        ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) + 
        ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) + 
        ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
    )
  noise_sd <- 1
  y <- f_XW + rnorm(n, 0, noise_sd)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  y_test <- y[test_inds]
  y_train <- y[train_inds]

  # Sample a BART model
  general_param_list <- list(num_chains = 1, keep_every = 1)
  bart_model <- bart(
    X_train = X_train,
    y_train = y_train,
    num_gfr = 0,
    num_burnin = 10,
    num_mcmc = 10,
    general_params = general_param_list
  )
  y_hat_orig <- rowMeans(predict(bart_model, X = X_test)$y_hat)

  # Save to JSON
  bart_json_string <- saveBARTModelToJsonString(bart_model)

  # Reload as a BART model
  bart_model_roundtrip <- createBARTModelFromJsonString(bart_json_string)

  # Predict from the roundtrip BART model
  y_hat_reloaded <- rowMeans(predict(bart_model_roundtrip, X = X_test)$y_hat)

  # Assertion
  expect_equal(y_hat_orig, y_hat_reloaded)
})

test_that("BCF Serialization", {
  skip_on_cran()

  n <- 500
  x1 <- runif(n)
  x2 <- runif(n)
  x3 <- runif(n)
  x4 <- runif(n)
  x5 <- runif(n)
  X <- cbind(x1, x2, x3, x4, x5)
  p <- ncol(X)
  pi_x <- 0.25 + 0.5 * X[, 1]
  mu_x <- pi_x * 5
  tau_x <- X[, 2] * 2
  Z <- rbinom(n, 1, pi_x)
  E_XZ <- mu_x + Z * tau_x
  y <- E_XZ + rnorm(n, 0, 1)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  pi_test <- pi_x[test_inds]
  pi_train <- pi_x[train_inds]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]
  mu_test <- mu_x[test_inds]
  mu_train <- mu_x[train_inds]
  tau_test <- tau_x[test_inds]
  tau_train <- tau_x[train_inds]

  # Sample a BCF model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    propensity_train = pi_train,
    num_gfr = 100,
    num_burnin = 0,
    num_mcmc = 100
  )
  bcf_preds_orig <- predict(bcf_model, X_test, Z_test, pi_test)
  mu_hat_orig <- rowMeans(bcf_preds_orig[["mu_hat"]])
  tau_hat_orig <- rowMeans(bcf_preds_orig[["tau_hat"]])
  y_hat_orig <- rowMeans(bcf_preds_orig[["y_hat"]])

  # Save to JSON
  bcf_json_string <- saveBCFModelToJsonString(bcf_model)

  # Reload as a BCF model
  bcf_model_roundtrip <- createBCFModelFromJsonString(bcf_json_string)

  # Predict from the roundtrip BCF model
  bcf_preds_reloaded <- predict(bcf_model_roundtrip, X_test, Z_test, pi_test)
  mu_hat_reloaded <- rowMeans(bcf_preds_reloaded[["mu_hat"]])
  tau_hat_reloaded <- rowMeans(bcf_preds_reloaded[["tau_hat"]])
  y_hat_reloaded <- rowMeans(bcf_preds_reloaded[["y_hat"]])

  # Assertion
  expect_equal(y_hat_orig, y_hat_reloaded)
})

test_that("BCF Serialization (no propensity)", {
  skip_on_cran()

  n <- 500
  x1 <- runif(n)
  x2 <- runif(n)
  x3 <- runif(n)
  x4 <- runif(n)
  x5 <- runif(n)
  X <- cbind(x1, x2, x3, x4, x5)
  p <- ncol(X)
  pi_x <- 0.25 + 0.5 * X[, 1]
  mu_x <- pi_x * 5
  tau_x <- X[, 2] * 2
  Z <- rbinom(n, 1, pi_x)
  E_XZ <- mu_x + Z * tau_x
  y <- E_XZ + rnorm(n, 0, 1)
  test_set_pct <- 0.2
  n_test <- round(test_set_pct * n)
  n_train <- n - n_test
  test_inds <- sort(sample(1:n, n_test, replace = FALSE))
  train_inds <- (1:n)[!((1:n) %in% test_inds)]
  X_test <- X[test_inds, ]
  X_train <- X[train_inds, ]
  pi_test <- pi_x[test_inds]
  pi_train <- pi_x[train_inds]
  Z_test <- Z[test_inds]
  Z_train <- Z[train_inds]
  y_test <- y[test_inds]
  y_train <- y[train_inds]
  mu_test <- mu_x[test_inds]
  mu_train <- mu_x[train_inds]
  tau_test <- tau_x[test_inds]
  tau_train <- tau_x[train_inds]

  # Sample a BCF model
  bcf_model <- bcf(
    X_train = X_train,
    Z_train = Z_train,
    y_train = y_train,
    num_gfr = 100,
    num_burnin = 0,
    num_mcmc = 100
  )
  bcf_preds_orig <- predict(bcf_model, X_test, Z_test)
  mu_hat_orig <- rowMeans(bcf_preds_orig[["mu_hat"]])
  tau_hat_orig <- rowMeans(bcf_preds_orig[["tau_hat"]])
  y_hat_orig <- rowMeans(bcf_preds_orig[["y_hat"]])

  # Save to JSON
  bcf_json_string <- saveBCFModelToJsonString(bcf_model)

  # Reload as a BCF model
  bcf_model_roundtrip <- createBCFModelFromJsonString(bcf_json_string)

  # Predict from the roundtrip BCF model
  bcf_preds_reloaded <- predict(bcf_model_roundtrip, X_test, Z_test)
  mu_hat_reloaded <- rowMeans(bcf_preds_reloaded[["mu_hat"]])
  tau_hat_reloaded <- rowMeans(bcf_preds_reloaded[["tau_hat"]])
  y_hat_reloaded <- rowMeans(bcf_preds_reloaded[["y_hat"]])

  # Assertion
  expect_equal(y_hat_orig, y_hat_reloaded)
})

test_that("BCF JSON uses canonical field names (sigma2_init, b1_samples, b0_samples)", {
  skip_on_cran()

  set.seed(1)
  n <- 100
  X <- matrix(runif(n * 5), ncol = 5)
  pi_x <- 0.25 + 0.5 * X[, 1]
  Z <- rbinom(n, 1, pi_x)
  y <- pi_x * 5 + Z * X[, 2] * 2 + rnorm(n)

  bcf_model <- bcf(
    X_train = X, Z_train = Z, y_train = y,
    propensity_train = pi_x,
    num_gfr = 0, num_burnin = 0, num_mcmc = 10,
    general_params = list(adaptive_coding = TRUE)
  )
  json_string <- saveBCFModelToJsonString(bcf_model)

  # New canonical names must be present
  expect_true(grepl('"sigma2_init"', json_string))
  expect_true(grepl('"b1_samples"', json_string))
  expect_true(grepl('"b0_samples"', json_string))

  # Legacy names must not be present
  expect_false(grepl('"initial_sigma2"', json_string))
  expect_false(grepl('"b_1_samples"', json_string))
  expect_false(grepl('"b_0_samples"', json_string))
})

test_that("BCF JSON deserialization handles legacy field names with warnings", {
  skip_on_cran()

  set.seed(2)
  n <- 100
  X <- matrix(runif(n * 5), ncol = 5)
  pi_x <- 0.25 + 0.5 * X[, 1]
  Z <- rbinom(n, 1, pi_x)
  y <- pi_x * 5 + Z * X[, 2] * 2 + rnorm(n)
  X_test <- matrix(runif(20 * 5), ncol = 5)
  pi_test <- rep(0.5, 20)
  Z_test <- rbinom(20, 1, 0.5)

  bcf_model <- bcf(
    X_train = X, Z_train = Z, y_train = y,
    propensity_train = pi_x,
    num_gfr = 0, num_burnin = 0, num_mcmc = 10,
    general_params = list(adaptive_coding = TRUE)
  )
  preds_orig <- predict(bcf_model, X_test, Z_test, pi_test)

  # Simulate a legacy JSON by replacing canonical names with old names
  json_new <- saveBCFModelToJsonString(bcf_model)
  json_legacy <- gsub('"sigma2_init"', '"initial_sigma2"', json_new, fixed = TRUE)
  json_legacy <- gsub('"b1_samples"', '"b_1_samples"', json_legacy, fixed = TRUE)
  json_legacy <- gsub('"b0_samples"', '"b_0_samples"', json_legacy, fixed = TRUE)

  # Loading a legacy JSON should emit deprecation warnings for all renamed fields
  all_warnings <- character(0)
  withCallingHandlers(
    bcf_legacy <- createBCFModelFromJsonString(json_legacy),
    warning = function(w) {
      all_warnings <<- c(all_warnings, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )
  expect_true(any(grepl("initial_sigma2.*deprecated|deprecated.*initial_sigma2", all_warnings)))
  expect_true(any(grepl("b_1_samples.*deprecated|b_0_samples.*deprecated|deprecated.*b_[01]_samples", all_warnings)))

  # Predictions must still match
  preds_legacy <- predict(bcf_legacy, X_test, Z_test, pi_test)
  expect_equal(rowMeans(preds_legacy[["y_hat"]]), rowMeans(preds_orig[["y_hat"]]))
  expect_equal(rowMeans(preds_legacy[["tau_hat"]]), rowMeans(preds_orig[["tau_hat"]]))
})

#' @title Forest Dataset C++ Wrapper
#'
#' @description
#' Wrapper around a C++ dataset class used to sample a forest.
#' A dataset consists of three matrices / vectors: covariates,
#' bases, and variance weights. Both the basis vector and variance
#' weights are optional.
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

ForestDataset <- R6::R6Class(
  classname = "ForestDataset",
  cloneable = FALSE,
  public = list(
    #' @field data_ptr External pointer to a C++ ForestDataset class
    data_ptr = NULL,

    #' @description
    #' Create a new ForestDataset object.
    #' @param covariates Matrix of covariates
    #' @param basis (Optional) Matrix of bases used to define a leaf regression
    #' @param variance_weights (Optional) Vector of observation-specific variance weights
    #' @return A new `ForestDataset` object.
    initialize = function(
      covariates,
      basis = NULL,
      variance_weights = NULL
    ) {
      self$data_ptr <- create_forest_dataset_cpp()
      forest_dataset_add_covariates_cpp(self$data_ptr, covariates)
      if (!is.null(basis)) {
        forest_dataset_add_basis_cpp(self$data_ptr, basis)
      }
      if (!is.null(variance_weights)) {
        forest_dataset_add_weights_cpp(self$data_ptr, variance_weights)
      }
    },

    #' @description
    #' Update basis matrix in a dataset
    #' @param basis Updated matrix of bases used to define a leaf regression
    update_basis = function(basis) {
      stopifnot(self$has_basis())
      forest_dataset_update_basis_cpp(self$data_ptr, basis)
    },

    #' @description
    #' Update variance_weights in a dataset
    #' @param variance_weights Updated vector of variance weights used to define individual variance / case weights
    #' @param exponentiate Whether or not input vector should be exponentiated before being written to the Dataset's variance weights. Default: F.
    update_variance_weights = function(variance_weights, exponentiate = F) {
      stopifnot(self$has_variance_weights())
      forest_dataset_update_var_weights_cpp(
        self$data_ptr,
        variance_weights,
        exponentiate
      )
    },

    #' @description
    #' Return number of observations in a `ForestDataset` object
    #' @return Observation count
    num_observations = function() {
      return(dataset_num_rows_cpp(self$data_ptr))
    },

    #' @description
    #' Return number of covariates in a `ForestDataset` object
    #' @return Covariate count
    num_covariates = function() {
      return(dataset_num_covariates_cpp(self$data_ptr))
    },

    #' @description
    #' Return number of bases in a `ForestDataset` object
    #' @return Basis count
    num_basis = function() {
      return(dataset_num_basis_cpp(self$data_ptr))
    },

    #' @description
    #' Return covariates as an R matrix
    #' @return Covariate data
    get_covariates = function() {
      return(forest_dataset_get_covariates_cpp(self$data_ptr))
    },

    #' @description
    #' Return bases as an R matrix
    #' @return Basis data
    get_basis = function() {
      return(forest_dataset_get_basis_cpp(self$data_ptr))
    },

    #' @description
    #' Return variance weights as an R vector
    #' @return Variance weight data
    get_variance_weights = function() {
      return(forest_dataset_get_variance_weights_cpp(self$data_ptr))
    },

    #' @description
    #' Whether or not a dataset has a basis matrix
    #' @return True if basis matrix is loaded, false otherwise
    has_basis = function() {
      return(dataset_has_basis_cpp(self$data_ptr))
    },

    #' @description
    #' Whether or not a dataset has variance weights
    #' @return True if variance weights are loaded, false otherwise
    has_variance_weights = function() {
      return(dataset_has_variance_weights_cpp(self$data_ptr))
    },

    #' @description
    #' Whether or not a dataset has auxiliary data stored at the dimension indicated
    #' @param dim_idx Dimension of auxiliary data
    #' @return True if auxiliary data has been allocated for `dim_idx` False otherwise
    has_auxiliary_dimension = function(dim_idx) {
      return(forest_dataset_has_auxiliary_dimension_cpp(self$data_ptr, dim_idx))
    },

    #' @description
    #' Initialize a new dimension / lane of auxiliary data and allocate data in its place
    #' @param dim_size Size of the new vector of data to allocate
    #' @return None
    add_auxiliary_dimension = function(dim_size) {
      return(forest_dataset_add_auxiliary_dimension_cpp(
        self$data_ptr,
        dim_size
      ))
    },

    #' @description
    #' Retrieve auxiliary data value
    #' @param dim_idx Dimension from which data value to be retrieved
    #' @param element_idx Element to retrieve from dimension `dim_idx`
    #' @return Floating point value stored in the requested auxiliary data space
    get_auxiliary_data_value = function(dim_idx, element_idx) {
      return(forest_dataset_get_auxiliary_data_value_cpp(
        self$data_ptr,
        dim_idx,
        element_idx
      ))
    },

    #' @description
    #' Set auxiliary data value
    #' @param dim_idx Dimension in which data value to be set
    #' @param element_idx Element to set within dimension `dim_idx`
    #' @param value Data value to set at auxiliary data dimension `dim_idx` and element `element_idx`
    #' @return None
    set_auxiliary_data_value = function(dim_idx, element_idx, value) {
      return(forest_dataset_set_auxiliary_data_value_cpp(
        self$data_ptr,
        dim_idx,
        element_idx,
        value
      ))
    },

    #' @description
    #' Retrieve entire auxiliary data vector
    #' @param dim_idx Dimension to retrieve
    #' @return Vector of all of the auxiliary data stored at dimension `dim_idx`
    get_auxiliary_data_vector = function(dim_idx) {
      return(forest_dataset_get_auxiliary_data_vector_cpp(
        self$data_ptr,
        dim_idx
      ))
    }
  )
)

#' @title Outcome Data C++ Wrapper
#'
#' @description
#' Outcome / partial residual used to sample an additive model.
#' The outcome class is a wrapper around a vector of (mutable)
#' outcomes for ML tasks (supervised learning, causal inference).
#' When an additive tree ensemble is sampled, the outcome used to
#' sample a specific model term is the "partial residual" consisting
#' of the outcome minus the predictions of every other model term
#' (trees, group random effects, etc...).
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

Outcome <- R6::R6Class(
  classname = "Outcome",
  cloneable = FALSE,
  public = list(
    #' @field data_ptr External pointer to a C++ `Outcome` class
    data_ptr = NULL,

    #' @description
    #' Create a new Outcome object.
    #' @param outcome Vector of outcome values
    #' @return A new `Outcome` object.
    initialize = function(outcome) {
      self$data_ptr <- create_column_vector_cpp(outcome)
    },

    #' @description
    #' Extract raw data in R from the underlying C++ object
    #' @return R vector containing (copy of) the values in `Outcome` object
    get_data = function() {
      return(get_residual_cpp(self$data_ptr))
    },

    #' @description
    #' Update the current state of the outcome (i.e. partial residual) data by adding the values of `update_vector`
    #' @param update_vector Vector to be added to outcome
    #' @return None
    add_vector = function(update_vector) {
      if (!is.numeric(update_vector)) {
        stop("update_vector must be a numeric vector or 2d matrix")
      } else {
        dim_vec <- dim(update_vector)
        if (!is.null(dim_vec)) {
          if (length(dim_vec) > 2) {
            stop(
              "if update_vector is provided as a matrix, it must be 2d"
            )
          }
          update_vector <- as.numeric(update_vector)
        }
      }
      add_to_column_vector_cpp(self$data_ptr, update_vector)
    },

    #' @description
    #' Update the current state of the outcome (i.e. partial residual) data by subtracting the values of `update_vector`
    #' @param update_vector Vector to be subtracted from outcome
    #' @return None
    subtract_vector = function(update_vector) {
      if (!is.numeric(update_vector)) {
        stop("update_vector must be a numeric vector or 2d matrix")
      } else {
        dim_vec <- dim(update_vector)
        if (!is.null(dim_vec)) {
          if (length(dim_vec) > 2) {
            stop(
              "if update_vector is provided as a matrix, it must be 2d"
            )
          }
          update_vector <- as.numeric(update_vector)
        }
      }
      subtract_from_column_vector_cpp(self$data_ptr, update_vector)
    },

    #' @description
    #' Update the current state of the outcome (i.e. partial residual) data by replacing each element with the elements of `new_vector`
    #' @param new_vector Vector from which to overwrite the current data
    #' @return None
    update_data = function(new_vector) {
      if (!is.numeric(new_vector)) {
        stop("update_vector must be a numeric vector or 2d matrix")
      } else {
        dim_vec <- dim(new_vector)
        if (!is.null(dim_vec)) {
          if (length(dim_vec) > 2) {
            stop(
              "if update_vector is provided as a matrix, it must be 2d"
            )
          }
          new_vector <- as.numeric(new_vector)
        }
      }
      overwrite_column_vector_cpp(self$data_ptr, new_vector)
    }
  )
)

#' @title Random Effects Dataset C++ Wrapper
#'
#' @description
#' Dataset used to sample a random effects model.
#' A random effects dataset consists of three matrices / vectors: group labels,
#' bases, and variance weights. Variance weights are optional.
#'
#' This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>

RandomEffectsDataset <- R6::R6Class(
  classname = "RandomEffectsDataset",
  cloneable = FALSE,
  public = list(
    #' @field data_ptr External pointer to a C++ RandomEffectsDataset class
    data_ptr = NULL,

    #' @description
    #' Create a new RandomEffectsDataset object.
    #' @param group_labels Vector of group labels
    #' @param basis Matrix of bases used to define the random effects regression (for an intercept-only model, pass an array of ones)
    #' @param variance_weights (Optional) Vector of observation-specific variance weights
    #' @return A new `RandomEffectsDataset` object.
    initialize = function(group_labels, basis, variance_weights = NULL) {
      self$data_ptr <- create_rfx_dataset_cpp()
      rfx_dataset_add_group_labels_cpp(self$data_ptr, group_labels)
      rfx_dataset_add_basis_cpp(self$data_ptr, basis)
      if (!is.null(variance_weights)) {
        rfx_dataset_add_weights_cpp(self$data_ptr, variance_weights)
      }
    },

    #' @description
    #' Update basis matrix in a dataset
    #' @param basis Updated matrix of bases used to define random slopes / intercepts
    update_basis = function(basis) {
      stopifnot(self$has_basis())
      rfx_dataset_update_basis_cpp(self$data_ptr, basis)
    },

    #' @description
    #' Update variance_weights in a dataset
    #' @param variance_weights Updated vector of variance weights used to define individual variance / case weights
    #' @param exponentiate Whether or not input vector should be exponentiated before being written to the RandomEffectsDataset's variance weights. Default: F.
    update_variance_weights = function(variance_weights, exponentiate = F) {
      stopifnot(self$has_variance_weights())
      rfx_dataset_update_var_weights_cpp(
        self$data_ptr,
        variance_weights,
        exponentiate
      )
    },

    #' @description
    #' Return number of observations in a `RandomEffectsDataset` object
    #' @return Observation count
    num_observations = function() {
      return(rfx_dataset_num_rows_cpp(self$data_ptr))
    },

    #' @description
    #' Return dimension of the basis matrix in a `RandomEffectsDataset` object
    #' @return Basis vector count
    num_basis = function() {
      return(rfx_dataset_num_basis_cpp(self$data_ptr))
    },

    #' @description
    #' Return group labels as an R vector
    #' @return Group label data
    get_group_labels = function() {
      return(rfx_dataset_get_group_labels_cpp(self$data_ptr))
    },

    #' @description
    #' Return bases as an R matrix
    #' @return Basis data
    get_basis = function() {
      return(rfx_dataset_get_basis_cpp(self$data_ptr))
    },

    #' @description
    #' Return variance weights as an R vector
    #' @return Variance weight data
    get_variance_weights = function() {
      return(rfx_dataset_get_variance_weights_cpp(self$data_ptr))
    },

    #' @description
    #' Whether or not a dataset has group label indices
    #' @return True if group label vector is loaded, false otherwise
    has_group_labels = function() {
      return(rfx_dataset_has_group_labels_cpp(self$data_ptr))
    },

    #' @description
    #' Whether or not a dataset has a basis matrix
    #' @return True if basis matrix is loaded, false otherwise
    has_basis = function() {
      return(rfx_dataset_has_basis_cpp(self$data_ptr))
    },

    #' @description
    #' Whether or not a dataset has variance weights
    #' @return True if variance weights are loaded, false otherwise
    has_variance_weights = function() {
      return(rfx_dataset_has_variance_weights_cpp(self$data_ptr))
    }
  )
)

#' @title Create ForestDataset Object
#' @description
#' Create a forest dataset object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param covariates Matrix of covariates
#' @param basis (Optional) Matrix of bases used to define a leaf regression
#' @param variance_weights (Optional) Vector of observation-specific variance weights
#'
#' @return `ForestDataset` object
#' @export
#'
#' @examples
#' covariate_matrix <- matrix(runif(10*100), ncol = 10)
#' basis_matrix <- matrix(rnorm(3*100), ncol = 3)
#' weight_vector <- rnorm(100)
#' forest_dataset <- createForestDataset(covariate_matrix)
#' forest_dataset <- createForestDataset(covariate_matrix, basis_matrix)
#' forest_dataset <- createForestDataset(covariate_matrix, basis_matrix, weight_vector)
createForestDataset <- function(
  covariates,
  basis = NULL,
  variance_weights = NULL
) {
  return(invisible((ForestDataset$new(covariates, basis, variance_weights))))
}

#' @title Create Outcome Object
#' @description
#' Create an outcome object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param outcome Vector of outcome values
#'
#' @return `Outcome` object
#' @export
#'
#' @examples
#' X <- matrix(runif(10*100), ncol = 10)
#' y <- -5 + 10*(X[,1] > 0.5) + rnorm(100)
#' outcome <- createOutcome(y)
createOutcome <- function(outcome) {
  return(invisible((Outcome$new(outcome))))
}

#' @title Create RandomEffectsDataset Object
#' @description
#' Create a random effects dataset object
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param group_labels Vector of group labels
#' @param basis Matrix of bases used to define the random effects regression (for an intercept-only model, pass an array of ones)
#' @param variance_weights (Optional) Vector of observation-specific variance weights
#'
#' @return `RandomEffectsDataset` object
#' @export
#'
#' @examples
#' rfx_group_ids <- sample(1:2, size = 100, replace = TRUE)
#' rfx_basis <- matrix(rnorm(3*100), ncol = 3)
#' weight_vector <- rnorm(100)
#' rfx_dataset <- createRandomEffectsDataset(rfx_group_ids, rfx_basis)
#' rfx_dataset <- createRandomEffectsDataset(rfx_group_ids, rfx_basis, weight_vector)
createRandomEffectsDataset <- function(
  group_labels,
  basis,
  variance_weights = NULL
) {
  return(invisible(
    (RandomEffectsDataset$new(group_labels, basis, variance_weights))
  ))
}

test_that("ForestDataset can be constructed and updated", {
  # Generate data
  n <- 20
  num_covariates <- 10
  num_basis <- 5
  covariates <- matrix(runif(n * num_covariates), ncol = num_covariates)
  basis <- matrix(runif(n * num_basis), ncol = num_basis)
  variance_weights <- runif(n)

  # Copy data to a ForestDataset object
  forest_dataset <- createForestDataset(covariates, basis, variance_weights)

  # Run first round of expectations
  expect_equal(forest_dataset$num_observations(), n)
  expect_equal(forest_dataset$num_covariates(), num_covariates)
  expect_equal(forest_dataset$num_basis(), num_basis)
  expect_equal(forest_dataset$has_variance_weights(), T)

  # Update data
  new_basis <- matrix(runif(n * num_basis), ncol = num_basis)
  new_variance_weights <- runif(n)
  expect_no_error(
    forest_dataset$update_basis(new_basis)
  )
  expect_no_error(
    forest_dataset$update_variance_weights(new_variance_weights)
  )

  # Check that we recover the correct data through get_covariates, get_basis, and get_variance_weights
  expect_equal(covariates, forest_dataset$get_covariates())
  expect_equal(new_basis, forest_dataset$get_basis())
  expect_equal(new_variance_weights, forest_dataset$get_variance_weights())
})

test_that("RandomEffectsDataset can be constructed and updated", {
  # Generate data
  n <- 20
  num_groups <- 4
  num_basis <- 5
  group_ids <- sample(as.integer(1:num_groups), size = n, replace = T)
  rfx_basis <- cbind(
    1,
    matrix(runif(n * (num_basis - 1)), ncol = (num_basis - 1))
  )
  variance_weights <- runif(n)

  # Copy data to a RandomEffectsDataset object
  rfx_dataset <- createRandomEffectsDataset(
    group_ids,
    rfx_basis,
    variance_weights
  )

  # Run first round of expectations
  expect_equal(rfx_dataset$num_observations(), n)
  expect_equal(rfx_dataset$num_basis(), num_basis)
  expect_equal(rfx_dataset$has_variance_weights(), T)

  # Update data
  new_rfx_basis <- matrix(runif(n * num_basis), ncol = num_basis)
  new_variance_weights <- runif(n)
  expect_no_error(
    rfx_dataset$update_basis(new_rfx_basis)
  )
  expect_no_error(
    rfx_dataset$update_variance_weights(new_variance_weights)
  )

  # Check that we recover the correct data through get_group_labels, get_basis, and get_variance_weights
  expect_equal(group_ids, rfx_dataset$get_group_labels())
  expect_equal(new_rfx_basis, rfx_dataset$get_basis())
  expect_equal(new_variance_weights, rfx_dataset$get_variance_weights())
})

#' @title Sample Global Error Variance
#' @description
#' Sample one iteration of the (inverse gamma) global variance model
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param residual Outcome class
#' @param dataset ForestDataset class
#' @param rng C++ random number generator
#' @param a Global variance shape parameter
#' @param b Global variance scale parameter
#' @return None
#' @export
#'
#' @examples
#' X <- matrix(runif(10*100), ncol = 10)
#' y <- -5 + 10*(X[,1] > 0.5) + rnorm(100)
#' y_std <- (y-mean(y))/sd(y)
#' forest_dataset <- createForestDataset(X)
#' outcome <- createOutcome(y_std)
#' rng <- createCppRNG(1234)
#' a <- 1.0
#' b <- 1.0
#' sigma2 <- sampleGlobalErrorVarianceOneIteration(outcome, forest_dataset, rng, a, b)
sampleGlobalErrorVarianceOneIteration <- function(
  residual,
  dataset,
  rng,
  a,
  b
) {
  return(sample_sigma2_one_iteration_cpp(
    residual$data_ptr,
    dataset$data_ptr,
    rng$rng_ptr,
    a,
    b
  ))
}

#' @title Sample Leaf Scale
#' @description
#' Sample one iteration of the leaf parameter variance model (only for univariate basis and constant leaf!)
#'
#' This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
#' Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
#' For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at <https://stochtree.ai/>
#'
#' @param forest C++ forest
#' @param rng C++ random number generator
#' @param a Leaf variance shape parameter
#' @param b Leaf variance scale parameter
#' @return None
#' @export
#'
#' @examples
#' num_trees <- 100
#' leaf_dimension <- 1
#' is_leaf_constant <- TRUE
#' is_exponentiated <- FALSE
#' active_forest <- createForest(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
#' rng <- createCppRNG(1234)
#' a <- 1.0
#' b <- 1.0
#' tau <- sampleLeafVarianceOneIteration(active_forest, rng, a, b)
sampleLeafVarianceOneIteration <- function(forest, rng, a, b) {
  return(sample_tau_one_iteration_cpp(forest$forest_ptr, rng$rng_ptr, a, b))
}

# Generated by cpp11: do not edit by hand

create_forest_dataset_cpp <- function() {
  .Call(`_stochtree_create_forest_dataset_cpp`)
}

dataset_num_rows_cpp <- function(dataset) {
  .Call(`_stochtree_dataset_num_rows_cpp`, dataset)
}

dataset_num_covariates_cpp <- function(dataset) {
  .Call(`_stochtree_dataset_num_covariates_cpp`, dataset)
}

dataset_num_basis_cpp <- function(dataset) {
  .Call(`_stochtree_dataset_num_basis_cpp`, dataset)
}

dataset_has_basis_cpp <- function(dataset) {
  .Call(`_stochtree_dataset_has_basis_cpp`, dataset)
}

dataset_has_variance_weights_cpp <- function(dataset) {
  .Call(`_stochtree_dataset_has_variance_weights_cpp`, dataset)
}

forest_dataset_add_covariates_cpp <- function(dataset_ptr, covariates) {
  invisible(.Call(`_stochtree_forest_dataset_add_covariates_cpp`, dataset_ptr, covariates))
}

forest_dataset_add_basis_cpp <- function(dataset_ptr, basis) {
  invisible(.Call(`_stochtree_forest_dataset_add_basis_cpp`, dataset_ptr, basis))
}

forest_dataset_update_basis_cpp <- function(dataset_ptr, basis) {
  invisible(.Call(`_stochtree_forest_dataset_update_basis_cpp`, dataset_ptr, basis))
}

forest_dataset_update_var_weights_cpp <- function(dataset_ptr, weights, exponentiate) {
  invisible(.Call(`_stochtree_forest_dataset_update_var_weights_cpp`, dataset_ptr, weights, exponentiate))
}

forest_dataset_add_weights_cpp <- function(dataset_ptr, weights) {
  invisible(.Call(`_stochtree_forest_dataset_add_weights_cpp`, dataset_ptr, weights))
}

forest_dataset_get_covariates_cpp <- function(dataset_ptr) {
  .Call(`_stochtree_forest_dataset_get_covariates_cpp`, dataset_ptr)
}

forest_dataset_get_basis_cpp <- function(dataset_ptr) {
  .Call(`_stochtree_forest_dataset_get_basis_cpp`, dataset_ptr)
}

forest_dataset_get_variance_weights_cpp <- function(dataset_ptr) {
  .Call(`_stochtree_forest_dataset_get_variance_weights_cpp`, dataset_ptr)
}

forest_dataset_has_auxiliary_dimension_cpp <- function(dataset_ptr, dim_idx) {
  .Call(`_stochtree_forest_dataset_has_auxiliary_dimension_cpp`, dataset_ptr, dim_idx)
}

forest_dataset_add_auxiliary_dimension_cpp <- function(dataset_ptr, dim_size) {
  invisible(.Call(`_stochtree_forest_dataset_add_auxiliary_dimension_cpp`, dataset_ptr, dim_size))
}

forest_dataset_get_auxiliary_data_value_cpp <- function(dataset_ptr, dim_idx, element_idx) {
  .Call(`_stochtree_forest_dataset_get_auxiliary_data_value_cpp`, dataset_ptr, dim_idx, element_idx)
}

forest_dataset_set_auxiliary_data_value_cpp <- function(dataset_ptr, dim_idx, element_idx, value) {
  invisible(.Call(`_stochtree_forest_dataset_set_auxiliary_data_value_cpp`, dataset_ptr, dim_idx, element_idx, value))
}

forest_dataset_get_auxiliary_data_vector_cpp <- function(dataset_ptr, dim_idx) {
  .Call(`_stochtree_forest_dataset_get_auxiliary_data_vector_cpp`, dataset_ptr, dim_idx)
}

create_column_vector_cpp <- function(outcome) {
  .Call(`_stochtree_create_column_vector_cpp`, outcome)
}

add_to_column_vector_cpp <- function(outcome, update_vector) {
  invisible(.Call(`_stochtree_add_to_column_vector_cpp`, outcome, update_vector))
}

subtract_from_column_vector_cpp <- function(outcome, update_vector) {
  invisible(.Call(`_stochtree_subtract_from_column_vector_cpp`, outcome, update_vector))
}

overwrite_column_vector_cpp <- function(outcome, new_vector) {
  invisible(.Call(`_stochtree_overwrite_column_vector_cpp`, outcome, new_vector))
}

propagate_trees_column_vector_cpp <- function(tracker, residual) {
  invisible(.Call(`_stochtree_propagate_trees_column_vector_cpp`, tracker, residual))
}

get_residual_cpp <- function(vector_ptr) {
  .Call(`_stochtree_get_residual_cpp`, vector_ptr)
}

create_rfx_dataset_cpp <- function() {
  .Call(`_stochtree_create_rfx_dataset_cpp`)
}

rfx_dataset_update_basis_cpp <- function(dataset_ptr, basis) {
  invisible(.Call(`_stochtree_rfx_dataset_update_basis_cpp`, dataset_ptr, basis))
}

rfx_dataset_update_var_weights_cpp <- function(dataset_ptr, weights, exponentiate) {
  invisible(.Call(`_stochtree_rfx_dataset_update_var_weights_cpp`, dataset_ptr, weights, exponentiate))
}

rfx_dataset_update_group_labels_cpp <- function(dataset_ptr, group_labels) {
  invisible(.Call(`_stochtree_rfx_dataset_update_group_labels_cpp`, dataset_ptr, group_labels))
}

rfx_dataset_num_basis_cpp <- function(dataset) {
  .Call(`_stochtree_rfx_dataset_num_basis_cpp`, dataset)
}

rfx_dataset_num_rows_cpp <- function(dataset) {
  .Call(`_stochtree_rfx_dataset_num_rows_cpp`, dataset)
}

rfx_dataset_has_group_labels_cpp <- function(dataset) {
  .Call(`_stochtree_rfx_dataset_has_group_labels_cpp`, dataset)
}

rfx_dataset_has_basis_cpp <- function(dataset) {
  .Call(`_stochtree_rfx_dataset_has_basis_cpp`, dataset)
}

rfx_dataset_has_variance_weights_cpp <- function(dataset) {
  .Call(`_stochtree_rfx_dataset_has_variance_weights_cpp`, dataset)
}

rfx_dataset_add_group_labels_cpp <- function(dataset_ptr, group_labels) {
  invisible(.Call(`_stochtree_rfx_dataset_add_group_labels_cpp`, dataset_ptr, group_labels))
}

rfx_dataset_add_basis_cpp <- function(dataset_ptr, basis) {
  invisible(.Call(`_stochtree_rfx_dataset_add_basis_cpp`, dataset_ptr, basis))
}

rfx_dataset_add_weights_cpp <- function(dataset_ptr, weights) {
  invisible(.Call(`_stochtree_rfx_dataset_add_weights_cpp`, dataset_ptr, weights))
}

rfx_dataset_get_group_labels_cpp <- function(dataset_ptr) {
  .Call(`_stochtree_rfx_dataset_get_group_labels_cpp`, dataset_ptr)
}

rfx_dataset_get_basis_cpp <- function(dataset_ptr) {
  .Call(`_stochtree_rfx_dataset_get_basis_cpp`, dataset_ptr)
}

rfx_dataset_get_variance_weights_cpp <- function(dataset_ptr) {
  .Call(`_stochtree_rfx_dataset_get_variance_weights_cpp`, dataset_ptr)
}

rfx_container_cpp <- function(num_components, num_groups) {
  .Call(`_stochtree_rfx_container_cpp`, num_components, num_groups)
}

rfx_container_from_json_cpp <- function(json_ptr, rfx_label) {
  .Call(`_stochtree_rfx_container_from_json_cpp`, json_ptr, rfx_label)
}

rfx_label_mapper_from_json_cpp <- function(json_ptr, rfx_label) {
  .Call(`_stochtree_rfx_label_mapper_from_json_cpp`, json_ptr, rfx_label)
}

rfx_group_ids_from_json_cpp <- function(json_ptr, rfx_label) {
  .Call(`_stochtree_rfx_group_ids_from_json_cpp`, json_ptr, rfx_label)
}

rfx_container_append_from_json_cpp <- function(rfx_container_ptr, json_ptr, rfx_label) {
  invisible(.Call(`_stochtree_rfx_container_append_from_json_cpp`, rfx_container_ptr, json_ptr, rfx_label))
}

rfx_container_from_json_string_cpp <- function(json_string, rfx_label) {
  .Call(`_stochtree_rfx_container_from_json_string_cpp`, json_string, rfx_label)
}

rfx_label_mapper_from_json_string_cpp <- function(json_string, rfx_label) {
  .Call(`_stochtree_rfx_label_mapper_from_json_string_cpp`, json_string, rfx_label)
}

rfx_group_ids_from_json_string_cpp <- function(json_string, rfx_label) {
  .Call(`_stochtree_rfx_group_ids_from_json_string_cpp`, json_string, rfx_label)
}

rfx_container_append_from_json_string_cpp <- function(rfx_container_ptr, json_string, rfx_label) {
  invisible(.Call(`_stochtree_rfx_container_append_from_json_string_cpp`, rfx_container_ptr, json_string, rfx_label))
}

rfx_model_cpp <- function(num_components, num_groups) {
  .Call(`_stochtree_rfx_model_cpp`, num_components, num_groups)
}

rfx_tracker_cpp <- function(group_labels) {
  .Call(`_stochtree_rfx_tracker_cpp`, group_labels)
}

rfx_label_mapper_cpp <- function(rfx_tracker) {
  .Call(`_stochtree_rfx_label_mapper_cpp`, rfx_tracker)
}

rfx_model_sample_random_effects_cpp <- function(rfx_model, rfx_dataset, residual, rfx_tracker, rfx_container, keep_sample, global_variance, rng) {
  invisible(.Call(`_stochtree_rfx_model_sample_random_effects_cpp`, rfx_model, rfx_dataset, residual, rfx_tracker, rfx_container, keep_sample, global_variance, rng))
}

rfx_model_predict_cpp <- function(rfx_model, rfx_dataset, rfx_tracker) {
  .Call(`_stochtree_rfx_model_predict_cpp`, rfx_model, rfx_dataset, rfx_tracker)
}

rfx_container_predict_cpp <- function(rfx_container, rfx_dataset, label_mapper) {
  .Call(`_stochtree_rfx_container_predict_cpp`, rfx_container, rfx_dataset, label_mapper)
}

rfx_container_num_samples_cpp <- function(rfx_container) {
  .Call(`_stochtree_rfx_container_num_samples_cpp`, rfx_container)
}

rfx_container_num_components_cpp <- function(rfx_container) {
  .Call(`_stochtree_rfx_container_num_components_cpp`, rfx_container)
}

rfx_container_num_groups_cpp <- function(rfx_container) {
  .Call(`_stochtree_rfx_container_num_groups_cpp`, rfx_container)
}

rfx_container_delete_sample_cpp <- function(rfx_container, sample_num) {
  invisible(.Call(`_stochtree_rfx_container_delete_sample_cpp`, rfx_container, sample_num))
}

rfx_model_set_working_parameter_cpp <- function(rfx_model, working_param_init) {
  invisible(.Call(`_stochtree_rfx_model_set_working_parameter_cpp`, rfx_model, working_param_init))
}

rfx_model_set_group_parameters_cpp <- function(rfx_model, group_params_init) {
  invisible(.Call(`_stochtree_rfx_model_set_group_parameters_cpp`, rfx_model, group_params_init))
}

rfx_model_set_working_parameter_covariance_cpp <- function(rfx_model, working_param_cov_init) {
  invisible(.Call(`_stochtree_rfx_model_set_working_parameter_covariance_cpp`, rfx_model, working_param_cov_init))
}

rfx_model_set_group_parameter_covariance_cpp <- function(rfx_model, group_param_cov_init) {
  invisible(.Call(`_stochtree_rfx_model_set_group_parameter_covariance_cpp`, rfx_model, group_param_cov_init))
}

rfx_model_set_variance_prior_shape_cpp <- function(rfx_model, shape) {
  invisible(.Call(`_stochtree_rfx_model_set_variance_prior_shape_cpp`, rfx_model, shape))
}

rfx_model_set_variance_prior_scale_cpp <- function(rfx_model, scale) {
  invisible(.Call(`_stochtree_rfx_model_set_variance_prior_scale_cpp`, rfx_model, scale))
}

rfx_tracker_get_unique_group_ids_cpp <- function(rfx_tracker) {
  .Call(`_stochtree_rfx_tracker_get_unique_group_ids_cpp`, rfx_tracker)
}

rfx_container_get_beta_cpp <- function(rfx_container_ptr) {
  .Call(`_stochtree_rfx_container_get_beta_cpp`, rfx_container_ptr)
}

rfx_container_get_alpha_cpp <- function(rfx_container_ptr) {
  .Call(`_stochtree_rfx_container_get_alpha_cpp`, rfx_container_ptr)
}

rfx_container_get_xi_cpp <- function(rfx_container_ptr) {
  .Call(`_stochtree_rfx_container_get_xi_cpp`, rfx_container_ptr)
}

rfx_container_get_sigma_cpp <- function(rfx_container_ptr) {
  .Call(`_stochtree_rfx_container_get_sigma_cpp`, rfx_container_ptr)
}

rfx_label_mapper_to_list_cpp <- function(label_mapper_ptr) {
  .Call(`_stochtree_rfx_label_mapper_to_list_cpp`, label_mapper_ptr)
}

reset_rfx_model_cpp <- function(rfx_model, rfx_container, sample_num) {
  invisible(.Call(`_stochtree_reset_rfx_model_cpp`, rfx_model, rfx_container, sample_num))
}

reset_rfx_tracker_cpp <- function(tracker, dataset, residual, rfx_model) {
  invisible(.Call(`_stochtree_reset_rfx_tracker_cpp`, tracker, dataset, residual, rfx_model))
}

root_reset_rfx_tracker_cpp <- function(tracker, dataset, residual, rfx_model) {
  invisible(.Call(`_stochtree_root_reset_rfx_tracker_cpp`, tracker, dataset, residual, rfx_model))
}

sum_cpp <- function(x) {
  .Call(`_stochtree_sum_cpp`, x)
}

mean_cpp <- function(x) {
  .Call(`_stochtree_mean_cpp`, x)
}

var_cpp <- function(x) {
  .Call(`_stochtree_var_cpp`, x)
}

sd_cpp <- function(x) {
  .Call(`_stochtree_sd_cpp`, x)
}

active_forest_cpp <- function(num_trees, output_dimension, is_leaf_constant, is_exponentiated) {
  .Call(`_stochtree_active_forest_cpp`, num_trees, output_dimension, is_leaf_constant, is_exponentiated)
}

forest_container_cpp <- function(num_trees, output_dimension, is_leaf_constant, is_exponentiated) {
  .Call(`_stochtree_forest_container_cpp`, num_trees, output_dimension, is_leaf_constant, is_exponentiated)
}

forest_container_from_json_cpp <- function(json_ptr, forest_label) {
  .Call(`_stochtree_forest_container_from_json_cpp`, json_ptr, forest_label)
}

forest_container_append_from_json_cpp <- function(forest_sample_ptr, json_ptr, forest_label) {
  invisible(.Call(`_stochtree_forest_container_append_from_json_cpp`, forest_sample_ptr, json_ptr, forest_label))
}

forest_container_from_json_string_cpp <- function(json_string, forest_label) {
  .Call(`_stochtree_forest_container_from_json_string_cpp`, json_string, forest_label)
}

forest_merge_cpp <- function(inbound_forest_ptr, outbound_forest_ptr) {
  invisible(.Call(`_stochtree_forest_merge_cpp`, inbound_forest_ptr, outbound_forest_ptr))
}

forest_add_constant_cpp <- function(forest_ptr, constant_value) {
  invisible(.Call(`_stochtree_forest_add_constant_cpp`, forest_ptr, constant_value))
}

forest_multiply_constant_cpp <- function(forest_ptr, constant_multiple) {
  invisible(.Call(`_stochtree_forest_multiply_constant_cpp`, forest_ptr, constant_multiple))
}

forest_container_append_from_json_string_cpp <- function(forest_sample_ptr, json_string, forest_label) {
  invisible(.Call(`_stochtree_forest_container_append_from_json_string_cpp`, forest_sample_ptr, json_string, forest_label))
}

combine_forests_forest_container_cpp <- function(forest_samples, forest_inds) {
  invisible(.Call(`_stochtree_combine_forests_forest_container_cpp`, forest_samples, forest_inds))
}

add_to_forest_forest_container_cpp <- function(forest_samples, forest_index, constant_value) {
  invisible(.Call(`_stochtree_add_to_forest_forest_container_cpp`, forest_samples, forest_index, constant_value))
}

multiply_forest_forest_container_cpp <- function(forest_samples, forest_index, constant_multiple) {
  invisible(.Call(`_stochtree_multiply_forest_forest_container_cpp`, forest_samples, forest_index, constant_multiple))
}

num_samples_forest_container_cpp <- function(forest_samples) {
  .Call(`_stochtree_num_samples_forest_container_cpp`, forest_samples)
}

ensemble_tree_max_depth_forest_container_cpp <- function(forest_samples, ensemble_num, tree_num) {
  .Call(`_stochtree_ensemble_tree_max_depth_forest_container_cpp`, forest_samples, ensemble_num, tree_num)
}

ensemble_average_max_depth_forest_container_cpp <- function(forest_samples, ensemble_num) {
  .Call(`_stochtree_ensemble_average_max_depth_forest_container_cpp`, forest_samples, ensemble_num)
}

average_max_depth_forest_container_cpp <- function(forest_samples) {
  .Call(`_stochtree_average_max_depth_forest_container_cpp`, forest_samples)
}

num_leaves_ensemble_forest_container_cpp <- function(forest_samples, forest_num) {
  .Call(`_stochtree_num_leaves_ensemble_forest_container_cpp`, forest_samples, forest_num)
}

sum_leaves_squared_ensemble_forest_container_cpp <- function(forest_samples, forest_num) {
  .Call(`_stochtree_sum_leaves_squared_ensemble_forest_container_cpp`, forest_samples, forest_num)
}

num_trees_forest_container_cpp <- function(forest_samples) {
  .Call(`_stochtree_num_trees_forest_container_cpp`, forest_samples)
}

json_save_forest_container_cpp <- function(forest_samples, json_filename) {
  invisible(.Call(`_stochtree_json_save_forest_container_cpp`, forest_samples, json_filename))
}

json_load_forest_container_cpp <- function(forest_samples, json_filename) {
  invisible(.Call(`_stochtree_json_load_forest_container_cpp`, forest_samples, json_filename))
}

leaf_dimension_forest_container_cpp <- function(forest_samples) {
  .Call(`_stochtree_leaf_dimension_forest_container_cpp`, forest_samples)
}

is_leaf_constant_forest_container_cpp <- function(forest_samples) {
  .Call(`_stochtree_is_leaf_constant_forest_container_cpp`, forest_samples)
}

is_exponentiated_forest_container_cpp <- function(forest_samples) {
  .Call(`_stochtree_is_exponentiated_forest_container_cpp`, forest_samples)
}

all_roots_forest_container_cpp <- function(forest_samples, forest_num) {
  .Call(`_stochtree_all_roots_forest_container_cpp`, forest_samples, forest_num)
}

add_sample_forest_container_cpp <- function(forest_samples) {
  invisible(.Call(`_stochtree_add_sample_forest_container_cpp`, forest_samples))
}

set_leaf_value_forest_container_cpp <- function(forest_samples, leaf_value) {
  invisible(.Call(`_stochtree_set_leaf_value_forest_container_cpp`, forest_samples, leaf_value))
}

add_sample_value_forest_container_cpp <- function(forest_samples, leaf_value) {
  invisible(.Call(`_stochtree_add_sample_value_forest_container_cpp`, forest_samples, leaf_value))
}

add_sample_vector_forest_container_cpp <- function(forest_samples, leaf_vector) {
  invisible(.Call(`_stochtree_add_sample_vector_forest_container_cpp`, forest_samples, leaf_vector))
}

add_numeric_split_tree_value_forest_container_cpp <- function(forest_samples, forest_num, tree_num, leaf_num, feature_num, split_threshold, left_leaf_value, right_leaf_value) {
  invisible(.Call(`_stochtree_add_numeric_split_tree_value_forest_container_cpp`, forest_samples, forest_num, tree_num, leaf_num, feature_num, split_threshold, left_leaf_value, right_leaf_value))
}

add_numeric_split_tree_vector_forest_container_cpp <- function(forest_samples, forest_num, tree_num, leaf_num, feature_num, split_threshold, left_leaf_vector, right_leaf_vector) {
  invisible(.Call(`_stochtree_add_numeric_split_tree_vector_forest_container_cpp`, forest_samples, forest_num, tree_num, leaf_num, feature_num, split_threshold, left_leaf_vector, right_leaf_vector))
}

get_tree_leaves_forest_container_cpp <- function(forest_samples, forest_num, tree_num) {
  .Call(`_stochtree_get_tree_leaves_forest_container_cpp`, forest_samples, forest_num, tree_num)
}

get_tree_split_counts_forest_container_cpp <- function(forest_samples, forest_num, tree_num, num_features) {
  .Call(`_stochtree_get_tree_split_counts_forest_container_cpp`, forest_samples, forest_num, tree_num, num_features)
}

get_forest_split_counts_forest_container_cpp <- function(forest_samples, forest_num, num_features) {
  .Call(`_stochtree_get_forest_split_counts_forest_container_cpp`, forest_samples, forest_num, num_features)
}

get_overall_split_counts_forest_container_cpp <- function(forest_samples, num_features) {
  .Call(`_stochtree_get_overall_split_counts_forest_container_cpp`, forest_samples, num_features)
}

get_granular_split_count_array_forest_container_cpp <- function(forest_samples, num_features) {
  .Call(`_stochtree_get_granular_split_count_array_forest_container_cpp`, forest_samples, num_features)
}

set_leaf_vector_forest_container_cpp <- function(forest_samples, leaf_vector) {
  invisible(.Call(`_stochtree_set_leaf_vector_forest_container_cpp`, forest_samples, leaf_vector))
}

is_leaf_node_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_is_leaf_node_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

is_numeric_split_node_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_is_numeric_split_node_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

is_categorical_split_node_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_is_categorical_split_node_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

parent_node_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_parent_node_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

left_child_node_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_left_child_node_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

right_child_node_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_right_child_node_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

node_depth_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_node_depth_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

split_index_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_split_index_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

split_theshold_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_split_theshold_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

split_categories_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_split_categories_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

leaf_values_forest_container_cpp <- function(forest_samples, forest_num, tree_num, node_id) {
  .Call(`_stochtree_leaf_values_forest_container_cpp`, forest_samples, forest_num, tree_num, node_id)
}

num_nodes_forest_container_cpp <- function(forest_samples, forest_num, tree_num) {
  .Call(`_stochtree_num_nodes_forest_container_cpp`, forest_samples, forest_num, tree_num)
}

num_leaves_forest_container_cpp <- function(forest_samples, forest_num, tree_num) {
  .Call(`_stochtree_num_leaves_forest_container_cpp`, forest_samples, forest_num, tree_num)
}

num_leaf_parents_forest_container_cpp <- function(forest_samples, forest_num, tree_num) {
  .Call(`_stochtree_num_leaf_parents_forest_container_cpp`, forest_samples, forest_num, tree_num)
}

num_split_nodes_forest_container_cpp <- function(forest_samples, forest_num, tree_num) {
  .Call(`_stochtree_num_split_nodes_forest_container_cpp`, forest_samples, forest_num, tree_num)
}

nodes_forest_container_cpp <- function(forest_samples, forest_num, tree_num) {
  .Call(`_stochtree_nodes_forest_container_cpp`, forest_samples, forest_num, tree_num)
}

leaves_forest_container_cpp <- function(forest_samples, forest_num, tree_num) {
  .Call(`_stochtree_leaves_forest_container_cpp`, forest_samples, forest_num, tree_num)
}

initialize_forest_model_cpp <- function(data, residual, forest_samples, tracker, init_values, leaf_model_int) {
  invisible(.Call(`_stochtree_initialize_forest_model_cpp`, data, residual, forest_samples, tracker, init_values, leaf_model_int))
}

adjust_residual_forest_container_cpp <- function(data, residual, forest_samples, tracker, requires_basis, forest_num, add) {
  invisible(.Call(`_stochtree_adjust_residual_forest_container_cpp`, data, residual, forest_samples, tracker, requires_basis, forest_num, add))
}

propagate_basis_update_forest_container_cpp <- function(data, residual, forest_samples, tracker, forest_num) {
  invisible(.Call(`_stochtree_propagate_basis_update_forest_container_cpp`, data, residual, forest_samples, tracker, forest_num))
}

remove_sample_forest_container_cpp <- function(forest_samples, forest_num) {
  invisible(.Call(`_stochtree_remove_sample_forest_container_cpp`, forest_samples, forest_num))
}

predict_forest_cpp <- function(forest_samples, dataset) {
  .Call(`_stochtree_predict_forest_cpp`, forest_samples, dataset)
}

predict_forest_raw_cpp <- function(forest_samples, dataset) {
  .Call(`_stochtree_predict_forest_raw_cpp`, forest_samples, dataset)
}

predict_forest_raw_single_forest_cpp <- function(forest_samples, dataset, forest_num) {
  .Call(`_stochtree_predict_forest_raw_single_forest_cpp`, forest_samples, dataset, forest_num)
}

predict_forest_raw_single_tree_cpp <- function(forest_samples, dataset, forest_num, tree_num) {
  .Call(`_stochtree_predict_forest_raw_single_tree_cpp`, forest_samples, dataset, forest_num, tree_num)
}

predict_active_forest_cpp <- function(active_forest, dataset) {
  .Call(`_stochtree_predict_active_forest_cpp`, active_forest, dataset)
}

predict_raw_active_forest_cpp <- function(active_forest, dataset) {
  .Call(`_stochtree_predict_raw_active_forest_cpp`, active_forest, dataset)
}

leaf_dimension_active_forest_cpp <- function(active_forest) {
  .Call(`_stochtree_leaf_dimension_active_forest_cpp`, active_forest)
}

average_max_depth_active_forest_cpp <- function(active_forest) {
  .Call(`_stochtree_average_max_depth_active_forest_cpp`, active_forest)
}

num_trees_active_forest_cpp <- function(active_forest) {
  .Call(`_stochtree_num_trees_active_forest_cpp`, active_forest)
}

ensemble_tree_max_depth_active_forest_cpp <- function(active_forest, tree_num) {
  .Call(`_stochtree_ensemble_tree_max_depth_active_forest_cpp`, active_forest, tree_num)
}

is_leaf_constant_active_forest_cpp <- function(active_forest) {
  .Call(`_stochtree_is_leaf_constant_active_forest_cpp`, active_forest)
}

is_exponentiated_active_forest_cpp <- function(active_forest) {
  .Call(`_stochtree_is_exponentiated_active_forest_cpp`, active_forest)
}

all_roots_active_forest_cpp <- function(active_forest) {
  .Call(`_stochtree_all_roots_active_forest_cpp`, active_forest)
}

set_leaf_value_active_forest_cpp <- function(active_forest, leaf_value) {
  invisible(.Call(`_stochtree_set_leaf_value_active_forest_cpp`, active_forest, leaf_value))
}

set_leaf_vector_active_forest_cpp <- function(active_forest, leaf_vector) {
  invisible(.Call(`_stochtree_set_leaf_vector_active_forest_cpp`, active_forest, leaf_vector))
}

add_numeric_split_tree_value_active_forest_cpp <- function(active_forest, tree_num, leaf_num, feature_num, split_threshold, left_leaf_value, right_leaf_value) {
  invisible(.Call(`_stochtree_add_numeric_split_tree_value_active_forest_cpp`, active_forest, tree_num, leaf_num, feature_num, split_threshold, left_leaf_value, right_leaf_value))
}

add_numeric_split_tree_vector_active_forest_cpp <- function(active_forest, tree_num, leaf_num, feature_num, split_threshold, left_leaf_vector, right_leaf_vector) {
  invisible(.Call(`_stochtree_add_numeric_split_tree_vector_active_forest_cpp`, active_forest, tree_num, leaf_num, feature_num, split_threshold, left_leaf_vector, right_leaf_vector))
}

get_tree_leaves_active_forest_cpp <- function(active_forest, tree_num) {
  .Call(`_stochtree_get_tree_leaves_active_forest_cpp`, active_forest, tree_num)
}

get_tree_split_counts_active_forest_cpp <- function(active_forest, tree_num, num_features) {
  .Call(`_stochtree_get_tree_split_counts_active_forest_cpp`, active_forest, tree_num, num_features)
}

get_overall_split_counts_active_forest_cpp <- function(active_forest, num_features) {
  .Call(`_stochtree_get_overall_split_counts_active_forest_cpp`, active_forest, num_features)
}

get_granular_split_count_array_active_forest_cpp <- function(active_forest, num_features) {
  .Call(`_stochtree_get_granular_split_count_array_active_forest_cpp`, active_forest, num_features)
}

initialize_forest_model_active_forest_cpp <- function(data, residual, active_forest, tracker, init_values, leaf_model_int) {
  invisible(.Call(`_stochtree_initialize_forest_model_active_forest_cpp`, data, residual, active_forest, tracker, init_values, leaf_model_int))
}

adjust_residual_active_forest_cpp <- function(data, residual, active_forest, tracker, requires_basis, add) {
  invisible(.Call(`_stochtree_adjust_residual_active_forest_cpp`, data, residual, active_forest, tracker, requires_basis, add))
}

propagate_basis_update_active_forest_cpp <- function(data, residual, active_forest, tracker) {
  invisible(.Call(`_stochtree_propagate_basis_update_active_forest_cpp`, data, residual, active_forest, tracker))
}

reset_active_forest_cpp <- function(active_forest, forest_samples, forest_num) {
  invisible(.Call(`_stochtree_reset_active_forest_cpp`, active_forest, forest_samples, forest_num))
}

reset_forest_model_cpp <- function(forest_tracker, forest, data, residual, is_mean_model) {
  invisible(.Call(`_stochtree_reset_forest_model_cpp`, forest_tracker, forest, data, residual, is_mean_model))
}

root_reset_active_forest_cpp <- function(active_forest) {
  invisible(.Call(`_stochtree_root_reset_active_forest_cpp`, active_forest))
}

forest_container_get_max_leaf_index_cpp <- function(forest_container, forest_num) {
  .Call(`_stochtree_forest_container_get_max_leaf_index_cpp`, forest_container, forest_num)
}

compute_leaf_indices_cpp <- function(forest_container, covariates, forest_nums) {
  .Call(`_stochtree_compute_leaf_indices_cpp`, forest_container, covariates, forest_nums)
}

sample_gfr_one_iteration_cpp <- function(data, residual, forest_samples, active_forest, tracker, split_prior, rng, sweep_indices, feature_types, cutpoint_grid_size, leaf_model_scale_input, variable_weights, a_forest, b_forest, global_variance, leaf_model_int, keep_forest, num_features_subsample, num_threads) {
  invisible(.Call(`_stochtree_sample_gfr_one_iteration_cpp`, data, residual, forest_samples, active_forest, tracker, split_prior, rng, sweep_indices, feature_types, cutpoint_grid_size, leaf_model_scale_input, variable_weights, a_forest, b_forest, global_variance, leaf_model_int, keep_forest, num_features_subsample, num_threads))
}

sample_mcmc_one_iteration_cpp <- function(data, residual, forest_samples, active_forest, tracker, split_prior, rng, sweep_indices, feature_types, cutpoint_grid_size, leaf_model_scale_input, variable_weights, a_forest, b_forest, global_variance, leaf_model_int, keep_forest, num_threads) {
  invisible(.Call(`_stochtree_sample_mcmc_one_iteration_cpp`, data, residual, forest_samples, active_forest, tracker, split_prior, rng, sweep_indices, feature_types, cutpoint_grid_size, leaf_model_scale_input, variable_weights, a_forest, b_forest, global_variance, leaf_model_int, keep_forest, num_threads))
}

sample_sigma2_one_iteration_cpp <- function(residual, dataset, rng, a, b) {
  .Call(`_stochtree_sample_sigma2_one_iteration_cpp`, residual, dataset, rng, a, b)
}

sample_tau_one_iteration_cpp <- function(active_forest, rng, a, b) {
  .Call(`_stochtree_sample_tau_one_iteration_cpp`, active_forest, rng, a, b)
}

rng_cpp <- function(random_seed) {
  .Call(`_stochtree_rng_cpp`, random_seed)
}

tree_prior_cpp <- function(alpha, beta, min_samples_leaf, max_depth) {
  .Call(`_stochtree_tree_prior_cpp`, alpha, beta, min_samples_leaf, max_depth)
}

update_alpha_tree_prior_cpp <- function(tree_prior_ptr, alpha) {
  invisible(.Call(`_stochtree_update_alpha_tree_prior_cpp`, tree_prior_ptr, alpha))
}

update_beta_tree_prior_cpp <- function(tree_prior_ptr, beta) {
  invisible(.Call(`_stochtree_update_beta_tree_prior_cpp`, tree_prior_ptr, beta))
}

update_min_samples_leaf_tree_prior_cpp <- function(tree_prior_ptr, min_samples_leaf) {
  invisible(.Call(`_stochtree_update_min_samples_leaf_tree_prior_cpp`, tree_prior_ptr, min_samples_leaf))
}

update_max_depth_tree_prior_cpp <- function(tree_prior_ptr, max_depth) {
  invisible(.Call(`_stochtree_update_max_depth_tree_prior_cpp`, tree_prior_ptr, max_depth))
}

get_alpha_tree_prior_cpp <- function(tree_prior_ptr) {
  .Call(`_stochtree_get_alpha_tree_prior_cpp`, tree_prior_ptr)
}

get_beta_tree_prior_cpp <- function(tree_prior_ptr) {
  .Call(`_stochtree_get_beta_tree_prior_cpp`, tree_prior_ptr)
}

get_min_samples_leaf_tree_prior_cpp <- function(tree_prior_ptr) {
  .Call(`_stochtree_get_min_samples_leaf_tree_prior_cpp`, tree_prior_ptr)
}

get_max_depth_tree_prior_cpp <- function(tree_prior_ptr) {
  .Call(`_stochtree_get_max_depth_tree_prior_cpp`, tree_prior_ptr)
}

forest_tracker_cpp <- function(data, feature_types, num_trees, n) {
  .Call(`_stochtree_forest_tracker_cpp`, data, feature_types, num_trees, n)
}

get_cached_forest_predictions_cpp <- function(tracker_ptr) {
  .Call(`_stochtree_get_cached_forest_predictions_cpp`, tracker_ptr)
}

sample_without_replacement_integer_cpp <- function(population_vector, sampling_probs, sample_size) {
  .Call(`_stochtree_sample_without_replacement_integer_cpp`, population_vector, sampling_probs, sample_size)
}

ordinal_sampler_cpp <- function() {
  .Call(`_stochtree_ordinal_sampler_cpp`)
}

ordinal_sampler_update_latent_variables_cpp <- function(sampler_ptr, data_ptr, outcome_ptr, rng_ptr) {
  invisible(.Call(`_stochtree_ordinal_sampler_update_latent_variables_cpp`, sampler_ptr, data_ptr, outcome_ptr, rng_ptr))
}

ordinal_sampler_update_gamma_params_cpp <- function(sampler_ptr, data_ptr, outcome_ptr, alpha_gamma, beta_gamma, gamma_0, rng_ptr) {
  invisible(.Call(`_stochtree_ordinal_sampler_update_gamma_params_cpp`, sampler_ptr, data_ptr, outcome_ptr, alpha_gamma, beta_gamma, gamma_0, rng_ptr))
}

ordinal_sampler_update_cumsum_exp_cpp <- function(sampler_ptr, data_ptr) {
  invisible(.Call(`_stochtree_ordinal_sampler_update_cumsum_exp_cpp`, sampler_ptr, data_ptr))
}

init_json_cpp <- function() {
  .Call(`_stochtree_init_json_cpp`)
}

json_add_double_subfolder_cpp <- function(json_ptr, subfolder_name, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_double_subfolder_cpp`, json_ptr, subfolder_name, field_name, field_value))
}

json_add_double_cpp <- function(json_ptr, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_double_cpp`, json_ptr, field_name, field_value))
}

json_add_integer_subfolder_cpp <- function(json_ptr, subfolder_name, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_integer_subfolder_cpp`, json_ptr, subfolder_name, field_name, field_value))
}

json_add_integer_cpp <- function(json_ptr, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_integer_cpp`, json_ptr, field_name, field_value))
}

json_add_bool_subfolder_cpp <- function(json_ptr, subfolder_name, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_bool_subfolder_cpp`, json_ptr, subfolder_name, field_name, field_value))
}

json_add_bool_cpp <- function(json_ptr, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_bool_cpp`, json_ptr, field_name, field_value))
}

json_add_vector_subfolder_cpp <- function(json_ptr, subfolder_name, field_name, field_vector) {
  invisible(.Call(`_stochtree_json_add_vector_subfolder_cpp`, json_ptr, subfolder_name, field_name, field_vector))
}

json_add_vector_cpp <- function(json_ptr, field_name, field_vector) {
  invisible(.Call(`_stochtree_json_add_vector_cpp`, json_ptr, field_name, field_vector))
}

json_add_integer_vector_subfolder_cpp <- function(json_ptr, subfolder_name, field_name, field_vector) {
  invisible(.Call(`_stochtree_json_add_integer_vector_subfolder_cpp`, json_ptr, subfolder_name, field_name, field_vector))
}

json_add_integer_vector_cpp <- function(json_ptr, field_name, field_vector) {
  invisible(.Call(`_stochtree_json_add_integer_vector_cpp`, json_ptr, field_name, field_vector))
}

json_add_string_vector_subfolder_cpp <- function(json_ptr, subfolder_name, field_name, field_vector) {
  invisible(.Call(`_stochtree_json_add_string_vector_subfolder_cpp`, json_ptr, subfolder_name, field_name, field_vector))
}

json_add_string_vector_cpp <- function(json_ptr, field_name, field_vector) {
  invisible(.Call(`_stochtree_json_add_string_vector_cpp`, json_ptr, field_name, field_vector))
}

json_add_string_subfolder_cpp <- function(json_ptr, subfolder_name, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_string_subfolder_cpp`, json_ptr, subfolder_name, field_name, field_value))
}

json_add_string_cpp <- function(json_ptr, field_name, field_value) {
  invisible(.Call(`_stochtree_json_add_string_cpp`, json_ptr, field_name, field_value))
}

json_contains_field_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_contains_field_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_contains_field_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_contains_field_cpp`, json_ptr, field_name)
}

json_extract_double_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_extract_double_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_extract_double_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_extract_double_cpp`, json_ptr, field_name)
}

json_extract_integer_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_extract_integer_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_extract_integer_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_extract_integer_cpp`, json_ptr, field_name)
}

json_extract_bool_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_extract_bool_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_extract_bool_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_extract_bool_cpp`, json_ptr, field_name)
}

json_extract_string_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_extract_string_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_extract_string_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_extract_string_cpp`, json_ptr, field_name)
}

json_extract_vector_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_extract_vector_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_extract_vector_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_extract_vector_cpp`, json_ptr, field_name)
}

json_extract_integer_vector_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_extract_integer_vector_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_extract_integer_vector_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_extract_integer_vector_cpp`, json_ptr, field_name)
}

json_extract_string_vector_subfolder_cpp <- function(json_ptr, subfolder_name, field_name) {
  .Call(`_stochtree_json_extract_string_vector_subfolder_cpp`, json_ptr, subfolder_name, field_name)
}

json_extract_string_vector_cpp <- function(json_ptr, field_name) {
  .Call(`_stochtree_json_extract_string_vector_cpp`, json_ptr, field_name)
}

json_add_forest_cpp <- function(json_ptr, forest_samples) {
  .Call(`_stochtree_json_add_forest_cpp`, json_ptr, forest_samples)
}

json_increment_rfx_count_cpp <- function(json_ptr) {
  invisible(.Call(`_stochtree_json_increment_rfx_count_cpp`, json_ptr))
}

json_add_rfx_container_cpp <- function(json_ptr, rfx_samples) {
  .Call(`_stochtree_json_add_rfx_container_cpp`, json_ptr, rfx_samples)
}

json_add_rfx_label_mapper_cpp <- function(json_ptr, label_mapper) {
  .Call(`_stochtree_json_add_rfx_label_mapper_cpp`, json_ptr, label_mapper)
}

json_add_rfx_groupids_cpp <- function(json_ptr, groupids) {
  .Call(`_stochtree_json_add_rfx_groupids_cpp`, json_ptr, groupids)
}

get_json_string_cpp <- function(json_ptr) {
  .Call(`_stochtree_get_json_string_cpp`, json_ptr)
}

json_save_file_cpp <- function(json_ptr, filename) {
  invisible(.Call(`_stochtree_json_save_file_cpp`, json_ptr, filename))
}

json_load_file_cpp <- function(json_ptr, filename) {
  invisible(.Call(`_stochtree_json_load_file_cpp`, json_ptr, filename))
}

json_load_string_cpp <- function(json_ptr, json_string) {
  invisible(.Call(`_stochtree_json_load_string_cpp`, json_ptr, json_string))
}

#' @title Extract Random Effect Samples Generic Function
#' @description
#' Generic function for extracting random effect samples from a model object (BCF, BART, etc...)
#'
#' @param object Fitted model object from which to extract random effects
#' @param ... Other parameters to be used in random effects extraction
#' @return List of random effect samples
#' @export
#'
#' @examples
#' n <- 100
#' p <- 10
#' X <- matrix(runif(n*p), ncol = p)
#' rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
#' rfx_basis <- rep(1.0, n)
#' y <- (-5 + 10*(X[,1] > 0.5)) + (-2*(rfx_group_ids==1)+2*(rfx_group_ids==2)) + rnorm(n)
#' bart_model <- bart(X_train=X, y_train=y, rfx_group_ids_train=rfx_group_ids,
#'                    rfx_basis_train = rfx_basis, num_gfr=0, num_mcmc=10)
#' rfx_samples <- getRandomEffectSamples(bart_model)
getRandomEffectSamples <- function(object, ...) {
  UseMethod("getRandomEffectSamples")
}

#' @title Extract Parameter Samples Generic Function
#' @description
#' Generic function for extracting parameter samples from a model object (BCF, BART, etc...)
#'
#' @param object Fitted model object from which to extract parameter samples
#' @param term Name of the parameter to extract (e.g., `"sigma2"`, `"y_hat_train"`, etc.)
#' @return Parameter sample array
#' @export
#'
#' @examples
#' n <- 100
#' p <- 10
#' X <- matrix(runif(n*p), ncol = p)
#' rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
#' rfx_basis <- rep(1.0, n)
#' y <- (-5 + 10*(X[,1] > 0.5)) + (-2*(rfx_group_ids==1)+2*(rfx_group_ids==2)) + rnorm(n)
#' bart_model <- bart(X_train=X, y_train=y, rfx_group_ids_train=rfx_group_ids,
#'                    rfx_basis_train = rfx_basis, num_gfr=0, num_mcmc=10)
#' sigma2_samples <- extractParameter(bart_model, "sigma2")
extractParameter <- function(object, term) {
  UseMethod("extractParameter")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialization.R
\name{CppJson}
\alias{CppJson}
\title{JSON C++ Object Wrapper}
\description{
Wrapper around a C++ \code{nlohmann::json} object

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{json_ptr}}{External pointer to a C++ nlohmann::json object}

\item{\code{num_forests}}{Number of forests in the nlohmann::json object}

\item{\code{forest_labels}}{Names of forest objects in the overall nlohmann::json object}

\item{\code{num_rfx}}{Number of random effects terms in the nlohman::json object}

\item{\code{rfx_container_labels}}{Names of rfx container objects in the overall nlohmann::json object}

\item{\code{rfx_mapper_labels}}{Names of rfx label mapper objects in the overall nlohmann::json object}

\item{\code{rfx_groupid_labels}}{Names of rfx group id objects in the overall nlohmann::json object}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CppJson-new}{\code{CppJson$new()}}
\item \href{#method-CppJson-add_forest}{\code{CppJson$add_forest()}}
\item \href{#method-CppJson-add_random_effects}{\code{CppJson$add_random_effects()}}
\item \href{#method-CppJson-add_scalar}{\code{CppJson$add_scalar()}}
\item \href{#method-CppJson-add_integer}{\code{CppJson$add_integer()}}
\item \href{#method-CppJson-add_boolean}{\code{CppJson$add_boolean()}}
\item \href{#method-CppJson-add_string}{\code{CppJson$add_string()}}
\item \href{#method-CppJson-add_vector}{\code{CppJson$add_vector()}}
\item \href{#method-CppJson-add_integer_vector}{\code{CppJson$add_integer_vector()}}
\item \href{#method-CppJson-add_string_vector}{\code{CppJson$add_string_vector()}}
\item \href{#method-CppJson-add_list}{\code{CppJson$add_list()}}
\item \href{#method-CppJson-add_string_list}{\code{CppJson$add_string_list()}}
\item \href{#method-CppJson-get_scalar}{\code{CppJson$get_scalar()}}
\item \href{#method-CppJson-get_integer}{\code{CppJson$get_integer()}}
\item \href{#method-CppJson-get_boolean}{\code{CppJson$get_boolean()}}
\item \href{#method-CppJson-get_string}{\code{CppJson$get_string()}}
\item \href{#method-CppJson-get_vector}{\code{CppJson$get_vector()}}
\item \href{#method-CppJson-get_integer_vector}{\code{CppJson$get_integer_vector()}}
\item \href{#method-CppJson-get_string_vector}{\code{CppJson$get_string_vector()}}
\item \href{#method-CppJson-get_numeric_list}{\code{CppJson$get_numeric_list()}}
\item \href{#method-CppJson-get_string_list}{\code{CppJson$get_string_list()}}
\item \href{#method-CppJson-return_json_string}{\code{CppJson$return_json_string()}}
\item \href{#method-CppJson-save_file}{\code{CppJson$save_file()}}
\item \href{#method-CppJson-load_from_file}{\code{CppJson$load_from_file()}}
\item \href{#method-CppJson-load_from_string}{\code{CppJson$load_from_string()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-new"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-new}{}}}
\subsection{Method \code{new()}}{
Create a new CppJson object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$new()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A new \code{CppJson} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_forest"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_forest}{}}}
\subsection{Method \code{add_forest()}}{
Convert a forest container to json and add to the current \code{CppJson} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_forest(forest_samples)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_samples}}{\code{ForestSamples} R class}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_random_effects"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_random_effects}{}}}
\subsection{Method \code{add_random_effects()}}{
Convert a random effects container to json and add to the current \code{CppJson} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_random_effects(rfx_samples)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{rfx_samples}}{\code{RandomEffectSamples} R class}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_scalar"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_scalar}{}}}
\subsection{Method \code{add_scalar()}}{
Add a scalar to the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_scalar(field_name, field_value, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_value}}{Numeric value of the field to be added to json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which to place the value}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_integer"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_integer}{}}}
\subsection{Method \code{add_integer()}}{
Add a scalar to the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_integer(field_name, field_value, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_value}}{Integer value of the field to be added to json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which to place the value}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_boolean"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_boolean}{}}}
\subsection{Method \code{add_boolean()}}{
Add a boolean value to the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_boolean(field_name, field_value, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_value}}{Numeric value of the field to be added to json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which to place the value}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_string"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_string}{}}}
\subsection{Method \code{add_string()}}{
Add a string value to the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_string(field_name, field_value, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_value}}{Numeric value of the field to be added to json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which to place the value}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_vector"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_vector}{}}}
\subsection{Method \code{add_vector()}}{
Add a vector to the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_vector(field_name, field_vector, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_vector}}{Vector to be stored in json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which to place the value}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_integer_vector"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_integer_vector}{}}}
\subsection{Method \code{add_integer_vector()}}{
Add an integer vector to the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_integer_vector(field_name, field_vector, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_vector}}{Vector to be stored in json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which to place the value}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_string_vector"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_string_vector}{}}}
\subsection{Method \code{add_string_vector()}}{
Add an array to the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_string_vector(field_name, field_vector, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_vector}}{Character vector to be stored in json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which to place the value}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_list"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_list}{}}}
\subsection{Method \code{add_list()}}{
Add a list of vectors (as an object map of arrays) to the json object under the name "field_name"
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_list(field_name, field_list)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_list}}{List to be stored in json}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-add_string_list"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-add_string_list}{}}}
\subsection{Method \code{add_string_list()}}{
Add a list of vectors (as an object map of arrays) to the json object under the name "field_name"
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$add_string_list(field_name, field_list)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{field_list}}{List to be stored in json}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_scalar"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_scalar}{}}}
\subsection{Method \code{get_scalar()}}{
Retrieve a scalar value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_scalar(field_name, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be accessed from json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which the field is stored}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_integer"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_integer}{}}}
\subsection{Method \code{get_integer()}}{
Retrieve a integer value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_integer(field_name, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be accessed from json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which the field is stored}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_boolean"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_boolean}{}}}
\subsection{Method \code{get_boolean()}}{
Retrieve a boolean value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_boolean(field_name, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be accessed from json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which the field is stored}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_string"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_string}{}}}
\subsection{Method \code{get_string()}}{
Retrieve a string value from the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_string(field_name, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be accessed from json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which the field is stored}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_vector"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_vector}{}}}
\subsection{Method \code{get_vector()}}{
Retrieve a vector from the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_vector(field_name, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be accessed from json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which the field is stored}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_integer_vector"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_integer_vector}{}}}
\subsection{Method \code{get_integer_vector()}}{
Retrieve an integer vector from the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_integer_vector(field_name, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be accessed from json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which the field is stored}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_string_vector"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_string_vector}{}}}
\subsection{Method \code{get_string_vector()}}{
Retrieve a character vector from the json object under the name "field_name" (with optional subfolder "subfolder_name")
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_string_vector(field_name, subfolder_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be accessed from json}

\item{\code{subfolder_name}}{(Optional) Name of the subfolder / hierarchy under which the field is stored}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_numeric_list"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_numeric_list}{}}}
\subsection{Method \code{get_numeric_list()}}{
Reconstruct a list of numeric vectors from the json object stored under "field_name"
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_numeric_list(field_name, key_names)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{key_names}}{Vector of names of list elements (each of which is a vector)}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-get_string_list"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-get_string_list}{}}}
\subsection{Method \code{get_string_list()}}{
Reconstruct a list of string vectors from the json object stored under "field_name"
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$get_string_list(field_name, key_names)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field_name}}{The name of the field to be added to json}

\item{\code{key_names}}{Vector of names of list elements (each of which is a vector)}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-return_json_string"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-return_json_string}{}}}
\subsection{Method \code{return_json_string()}}{
Convert a JSON object to in-memory string
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$return_json_string()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
JSON string
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-save_file"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-save_file}{}}}
\subsection{Method \code{save_file()}}{
Save a json object to file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$save_file(filename)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{filename}}{String of filepath, must end in ".json"}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-load_from_file"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-load_from_file}{}}}
\subsection{Method \code{load_from_file()}}{
Load a json object from file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$load_from_file(filename)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{filename}}{String of filepath, must end in ".json"}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppJson-load_from_string"></a>}}
\if{latex}{\out{\hypertarget{method-CppJson-load_from_string}{}}}
\subsection{Method \code{load_from_string()}}{
Load a json object from string
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppJson$load_from_string(json_string)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_string}}{JSON string dump}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialization.R
\name{loadVectorJson}
\alias{loadVectorJson}
\title{Load Vector from JSON}
\usage{
loadVectorJson(json_object, json_vector_label, subfolder_name = NULL)
}
\arguments{
\item{json_object}{Object of class \code{CppJson}}

\item{json_vector_label}{Label referring to a particular vector (i.e. "sigma2_global_samples") in the overall json hierarchy}

\item{subfolder_name}{(Optional) Name of the subfolder / hierarchy under which vector sits}
}
\value{
R vector
}
\description{
Load a vector from json

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
example_vec <- runif(10)
example_json <- createCppJson()
example_json$add_vector("myvec", example_vec)
roundtrip_vec <- loadVectorJson(example_json, "myvec")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_effects.R
\name{RandomEffectStateManagement}
\alias{RandomEffectStateManagement}
\alias{resetRandomEffectsModel}
\alias{resetRandomEffectsTracker}
\alias{rootResetRandomEffectsModel}
\alias{rootResetRandomEffectsTracker}
\title{Random Effects State Management Routines}
\usage{
resetRandomEffectsModel(rfx_model, rfx_samples, sample_num, sigma_alpha_init)

resetRandomEffectsTracker(
  rfx_tracker,
  rfx_model,
  rfx_dataset,
  residual,
  rfx_samples
)

rootResetRandomEffectsModel(
  rfx_model,
  alpha_init,
  xi_init,
  sigma_alpha_init,
  sigma_xi_init,
  sigma_xi_shape,
  sigma_xi_scale
)

rootResetRandomEffectsTracker(rfx_tracker, rfx_model, rfx_dataset, residual)
}
\arguments{
\item{rfx_model}{Object of type \code{RandomEffectsModel}.}

\item{rfx_samples}{Object of type \code{RandomEffectSamples}.}

\item{sample_num}{Index of sample stored in \code{rfx_samples} from which to reset the state of a random effects model. Zero-indexed, so resetting based on the first sample would require setting \code{sample_num = 0}.}

\item{sigma_alpha_init}{Initial value of the "working parameter" scale parameter.}

\item{rfx_tracker}{Object of type \code{RandomEffectsTracker}.}

\item{rfx_dataset}{Object of type \code{RandomEffectsDataset}.}

\item{residual}{Object of type \code{Outcome}.}

\item{alpha_init}{Initial value of the "working parameter".}

\item{xi_init}{Initial value of the "group parameters".}

\item{sigma_xi_init}{Initial value of the "group parameters" scale parameter.}

\item{sigma_xi_shape}{Shape parameter for the inverse gamma variance model on the group parameters.}

\item{sigma_xi_scale}{Scale parameter for the inverse gamma variance model on the group parameters.}
}
\value{
All four functions have no return type and operate in-place on the relevant \link{RandomEffectsModel} or \link{RandomEffectsTracker} objects
}
\description{
A forest sampler features two types of state: ephemeral and persistent.
Persistent state includes objects like \link{ForestSamples} and \link{RandomEffectSamples}
which constitute part of the final sampled model.
Ephemeral state supports the sampling computations, but is not retained after the sampler finishes.

The two primary random-effects-based bits of ephemeral state are the \link{RandomEffectsModel} and \link{RandomEffectsTracker} classes,
which represent the current state of a random effects model and its corresponding tracking data structures.

In a linear sampling loop, this ephemeral state is updated with each iteration of the sampler
and any retained forests are copied to a \link{RandomEffectSamples} object. However, in multi-chain settings,
the state of a random effects model must typically be "reset" at the beginning of a new chain. These function enable
this process by synchronizing the state of a \link{RandomEffectsModel} and \link{RandomEffectsTracker} with a corresponding element
of a \link{RandomEffectSamples} object, or by resetting both to their default (root) state.

\code{resetRandomEffectsModel} resets a \code{RandomEffectsModel} object based on the parameters indexed by \code{sample_num} in a \code{RandomEffectsSamples} object.
\code{resetRandomEffectsTracker} resets a \code{RandomEffectsTracker} object based on the parameters indexed by \code{sample_num} in a \code{RandomEffectsSamples} object.

\code{rootResetRandomEffectsModel} resets a \code{RandomEffectsModel} object to its "default" state.
\code{rootResetRandomEffectsTracker} resets a \code{RandomEffectsTracker} object to its "default" state.

These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
n <- 100
p <- 10
rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
rfx_basis <- matrix(rep(1.0, n), ncol=1)
rfx_dataset <- createRandomEffectsDataset(rfx_group_ids, rfx_basis)
y <- (-2*(rfx_group_ids==1)+2*(rfx_group_ids==2)) + rnorm(n)
y_std <- (y-mean(y))/sd(y)
outcome <- createOutcome(y_std)
rng <- createCppRNG(1234)
num_groups <- length(unique(rfx_group_ids))
num_components <- ncol(rfx_basis)
rfx_model <- createRandomEffectsModel(num_components, num_groups)
rfx_tracker <- createRandomEffectsTracker(rfx_group_ids)
rfx_samples <- createRandomEffectSamples(num_components, num_groups, rfx_tracker)
alpha_init <- rep(1,num_components)
xi_init <- matrix(rep(alpha_init, num_groups),num_components,num_groups)
sigma_alpha_init <- diag(1,num_components,num_components)
sigma_xi_init <- diag(1,num_components,num_components)
sigma_xi_shape <- 1
sigma_xi_scale <- 1
rfx_model$set_working_parameter(alpha_init)
rfx_model$set_group_parameters(xi_init)
rfx_model$set_working_parameter_cov(sigma_alpha_init)
rfx_model$set_group_parameter_cov(sigma_xi_init)
rfx_model$set_variance_prior_shape(sigma_xi_shape)
rfx_model$set_variance_prior_scale(sigma_xi_scale)
for (i in 1:3) {
    rfx_model$sample_random_effect(rfx_dataset=rfx_dataset, residual=outcome,
                                   rfx_tracker=rfx_tracker, rfx_samples=rfx_samples,
                                   keep_sample=TRUE, global_variance=1.0, rng=rng)
}
resetRandomEffectsModel(rfx_model, rfx_samples, 0, 1.0)
resetRandomEffectsTracker(rfx_tracker, rfx_model, rfx_dataset, outcome, rfx_samples)
rootResetRandomEffectsModel(rfx_model, alpha_init, xi_init, sigma_alpha_init,
                            sigma_xi_init, sigma_xi_shape, sigma_xi_scale)
rootResetRandomEffectsTracker(rfx_tracker, rfx_model, rfx_dataset, outcome)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcf.R
\name{summary.bcfmodel}
\alias{summary.bcfmodel}
\title{Summarize BCF Model}
\usage{
\method{summary}{bcfmodel}(object, ...)
}
\arguments{
\item{object}{The BCF model object}

\item{...}{Additional arguments}
}
\value{
BCF model object unchanged after summarizing
}
\description{
Summarize a BCF fit with a description of the model that was fit and numeric summaries of any sampled quantities.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcf.R
\name{predict.bcfmodel}
\alias{predict.bcfmodel}
\title{Predict from BCF Model}
\usage{
\method{predict}{bcfmodel}(
  object,
  X,
  Z,
  propensity = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL,
  type = "posterior",
  terms = "all",
  scale = "linear",
  ...
)
}
\arguments{
\item{object}{Object of type \code{bcfmodel} containing draws of a Bayesian causal forest model and associated sampling outputs.}

\item{X}{Covariates used to determine tree leaf predictions for each observation. Must be passed as a matrix or dataframe.}

\item{Z}{Treatments used for prediction.}

\item{propensity}{(Optional) Propensities used for prediction.}

\item{rfx_group_ids}{(Optional) Test set group labels used for an additive random effects model.
We do not currently support (but plan to in the near future), test set evaluation for group labels
that were not in the training set.}

\item{rfx_basis}{(Optional) Test set basis for "random-slope" regression in additive random effects model. If the model was sampled with a random effects \code{model_spec} of "intercept_only" or "intercept_plus_treatment", this is optional, but if it is provided, it will be used.}

\item{type}{(Optional) Type of prediction to return. Options are "mean", which averages the predictions from every draw of a BCF model, and "posterior", which returns the entire matrix of posterior predictions. Default: "posterior".}

\item{terms}{(Optional) Which model terms to include in the prediction. This can be a single term or a list of model terms. Options include "y_hat", "prognostic_function", "mu", "cate", "tau", "rfx", "variance_forest", or "all". If a model doesn't have random effects or variance forest predictions, but one of those terms is request, the request will simply be ignored. If a model has random effects fit with either "intercept_only" or "intercept_plus_treatment" model_spec, then "prognostic_function" refers to the predictions of the prognostic forest plus the random intercept and "cate" refers to the predictions of the treatment effect forest plus the random slope on the treatment variable. For these models, the forest predictions alone can be requested via "mu" (prognostic forest) and "tau" (treatment effect forest). In all other cases, "mu" will return exactly the same result as "prognostic_function" and "tau" will return exactly the same result as "cate". If none of the requested terms are present in a model, this function will return \code{NULL} along with a warning. Default: "all".}

\item{scale}{(Optional) Scale of mean function predictions. Options are "linear", which returns predictions on the original scale of the mean forest / RFX terms, and "probability", which transforms predictions into a probability of observing \code{y == 1}. "probability" is only valid for models fit with a probit outcome model. Default: "linear".}

\item{...}{(Optional) Other prediction parameters.}
}
\value{
List of prediction matrices or single prediction matrix / vector, depending on the terms requested.
}
\description{
Predict from a sampled BCF model on new data
}
\examples{
n <- 500
p <- 5
X <- matrix(runif(n*p), ncol = p)
mu_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
pi_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (0.2) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (0.4) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (0.6) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (0.8)
)
tau_x <- (
    ((0 <= X[,2]) & (0.25 > X[,2])) * (0.5) +
    ((0.25 <= X[,2]) & (0.5 > X[,2])) * (1.0) +
    ((0.5 <= X[,2]) & (0.75 > X[,2])) * (1.5) +
    ((0.75 <= X[,2]) & (1 > X[,2])) * (2.0)
)
Z <- rbinom(n, 1, pi_x)
noise_sd <- 1
y <- mu_x + tau_x*Z + rnorm(n, 0, noise_sd)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
pi_test <- pi_x[test_inds]
pi_train <- pi_x[train_inds]
Z_test <- Z[test_inds]
Z_train <- Z[train_inds]
y_test <- y[test_inds]
y_train <- y[train_inds]
mu_test <- mu_x[test_inds]
mu_train <- mu_x[train_inds]
tau_test <- tau_x[test_inds]
tau_train <- tau_x[train_inds]
bcf_model <- bcf(X_train = X_train, Z_train = Z_train, y_train = y_train,
                 propensity_train = pi_train, num_gfr = 10,
                 num_burnin = 0, num_mcmc = 10)
preds <- predict(bcf_model, X_test, Z_test, pi_test)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior_transformation.R
\name{computeBCFPosteriorInterval}
\alias{computeBCFPosteriorInterval}
\title{Compute BCF Posterior Credible Intervals}
\usage{
computeBCFPosteriorInterval(
  model_object,
  terms,
  level = 0.95,
  scale = "linear",
  X = NULL,
  Z = NULL,
  propensity = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL
)
}
\arguments{
\item{model_object}{A fitted BCF model object of class \code{bcfmodel}.}

\item{terms}{A character string specifying the model term(s) for which to compute intervals. Options for BCF models are \code{"prognostic_function"}, \code{"mu"}, \code{"cate"}, \code{"tau"}, \code{"variance_forest"}, \code{"rfx"}, or \code{"y_hat"}. Note that \code{"mu"} is only different from \code{"prognostic_function"} if random effects are included with a model spec of \code{"intercept_only"} or \code{"intercept_plus_treatment"} and \code{"tau"} is only different from \code{"cate"} if random effects are included with a model spec of \code{"intercept_plus_treatment"}.}

\item{level}{A numeric value between 0 and 1 specifying the credible interval level (default is 0.95 for a 95\% credible interval).}

\item{scale}{(Optional) Scale of mean function predictions. Options are "linear", which returns predictions on the original scale of the mean forest / RFX terms, and "probability", which transforms predictions into a probability of observing \code{y == 1}. "probability" is only valid for models fit with a probit outcome model. Default: "linear".}

\item{X}{(Optional) A matrix or data frame of covariates at which to compute the intervals. Required if the requested term depends on covariates (e.g., prognostic forest, CATE forest, variance forest, or overall predictions).}

\item{Z}{(Optional) A vector or matrix of treatment assignments. Required if the requested term is \code{"y_hat"} (overall predictions).}

\item{propensity}{(Optional) A vector or matrix of propensity scores. Required if the underlying model depends on user-provided propensities.}

\item{rfx_group_ids}{An optional vector of group IDs for random effects. Required if the requested term includes random effects.}

\item{rfx_basis}{An optional matrix of basis function evaluations for random effects. Required if the requested term includes random effects.}
}
\value{
A list containing the lower and upper bounds of the credible interval for the specified term. If multiple terms are requested, a named list with intervals for each term is returned.
}
\description{
Compute posterior credible intervals for specified terms from a fitted BCF model. Supports intervals for prognostic forests, CATE forests, variance forests, random effects, and overall mean outcome predictions.
}
\examples{
n <- 100
p <- 5
X <- matrix(rnorm(n * p), nrow = n, ncol = p)
pi_X <- pnorm(0.5 * X[,1])
Z <- rbinom(n, 1, pi_X)
mu_X <- X[,1]
tau_X <- 0.25 * X[,2]
y <- mu_X + tau_X * Z + rnorm(n)
bcf_model <- bcf(X_train = X, Z_train = Z, y_train = y,
                 propensity_train = pi_X)
intervals <- computeBCFPosteriorInterval(
 model_object = bcf_model,
 terms = c("prognostic_function", "cate"),
 X = X,
 Z = Z,
 propensity = pi_X,
 level = 0.90
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{RandomEffectsDataset}
\alias{RandomEffectsDataset}
\title{Random Effects Dataset C++ Wrapper}
\description{
Dataset used to sample a random effects model.
A random effects dataset consists of three matrices / vectors: group labels,
bases, and variance weights. Variance weights are optional.

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{data_ptr}}{External pointer to a C++ RandomEffectsDataset class}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-RandomEffectsDataset-new}{\code{RandomEffectsDataset$new()}}
\item \href{#method-RandomEffectsDataset-update_basis}{\code{RandomEffectsDataset$update_basis()}}
\item \href{#method-RandomEffectsDataset-update_variance_weights}{\code{RandomEffectsDataset$update_variance_weights()}}
\item \href{#method-RandomEffectsDataset-num_observations}{\code{RandomEffectsDataset$num_observations()}}
\item \href{#method-RandomEffectsDataset-num_basis}{\code{RandomEffectsDataset$num_basis()}}
\item \href{#method-RandomEffectsDataset-get_group_labels}{\code{RandomEffectsDataset$get_group_labels()}}
\item \href{#method-RandomEffectsDataset-get_basis}{\code{RandomEffectsDataset$get_basis()}}
\item \href{#method-RandomEffectsDataset-get_variance_weights}{\code{RandomEffectsDataset$get_variance_weights()}}
\item \href{#method-RandomEffectsDataset-has_group_labels}{\code{RandomEffectsDataset$has_group_labels()}}
\item \href{#method-RandomEffectsDataset-has_basis}{\code{RandomEffectsDataset$has_basis()}}
\item \href{#method-RandomEffectsDataset-has_variance_weights}{\code{RandomEffectsDataset$has_variance_weights()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-new"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-new}{}}}
\subsection{Method \code{new()}}{
Create a new RandomEffectsDataset object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$new(group_labels, basis, variance_weights = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{group_labels}}{Vector of group labels}

\item{\code{basis}}{Matrix of bases used to define the random effects regression (for an intercept-only model, pass an array of ones)}

\item{\code{variance_weights}}{(Optional) Vector of observation-specific variance weights}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{RandomEffectsDataset} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-update_basis"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-update_basis}{}}}
\subsection{Method \code{update_basis()}}{
Update basis matrix in a dataset
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$update_basis(basis)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{basis}}{Updated matrix of bases used to define random slopes / intercepts}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-update_variance_weights"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-update_variance_weights}{}}}
\subsection{Method \code{update_variance_weights()}}{
Update variance_weights in a dataset
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$update_variance_weights(
  variance_weights,
  exponentiate = F
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{variance_weights}}{Updated vector of variance weights used to define individual variance / case weights}

\item{\code{exponentiate}}{Whether or not input vector should be exponentiated before being written to the RandomEffectsDataset's variance weights. Default: F.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-num_observations"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-num_observations}{}}}
\subsection{Method \code{num_observations()}}{
Return number of observations in a \code{RandomEffectsDataset} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$num_observations()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Observation count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-num_basis"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-num_basis}{}}}
\subsection{Method \code{num_basis()}}{
Return dimension of the basis matrix in a \code{RandomEffectsDataset} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$num_basis()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Basis vector count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-get_group_labels"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-get_group_labels}{}}}
\subsection{Method \code{get_group_labels()}}{
Return group labels as an R vector
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$get_group_labels()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Group label data
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-get_basis"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-get_basis}{}}}
\subsection{Method \code{get_basis()}}{
Return bases as an R matrix
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$get_basis()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Basis data
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-get_variance_weights"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-get_variance_weights}{}}}
\subsection{Method \code{get_variance_weights()}}{
Return variance weights as an R vector
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$get_variance_weights()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Variance weight data
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-has_group_labels"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-has_group_labels}{}}}
\subsection{Method \code{has_group_labels()}}{
Whether or not a dataset has group label indices
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$has_group_labels()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
True if group label vector is loaded, false otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-has_basis"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-has_basis}{}}}
\subsection{Method \code{has_basis()}}{
Whether or not a dataset has a basis matrix
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$has_basis()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
True if basis matrix is loaded, false otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsDataset-has_variance_weights"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsDataset-has_variance_weights}{}}}
\subsection{Method \code{has_variance_weights()}}{
Whether or not a dataset has variance weights
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsDataset$has_variance_weights()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
True if variance weights are loaded, false otherwise
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcf.R
\name{BCFSerialization}
\alias{BCFSerialization}
\alias{saveBCFModelToJson}
\alias{saveBCFModelToJsonFile}
\alias{saveBCFModelToJsonString}
\alias{createBCFModelFromJson}
\alias{createBCFModelFromJsonFile}
\alias{createBCFModelFromJsonString}
\alias{createBCFModelFromCombinedJson}
\alias{createBCFModelFromCombinedJsonString}
\title{BCF Serialization Routines}
\usage{
saveBCFModelToJson(object)

saveBCFModelToJsonFile(object, filename)

saveBCFModelToJsonString(object)

createBCFModelFromJson(json_object)

createBCFModelFromJsonFile(json_filename)

createBCFModelFromJsonString(json_string)

createBCFModelFromCombinedJson(json_object_list)

createBCFModelFromCombinedJsonString(json_string_list)
}
\arguments{
\item{object}{Object of type \code{bcfmodel} containing draws of a Bayesian causal forest model and associated sampling outputs.}

\item{filename}{String of filepath, must end in ".json"}

\item{json_object}{Object of type \code{CppJson} containing Json representation of a BCF model}

\item{json_filename}{String of filepath, must end in ".json"}

\item{json_string}{JSON string dump}

\item{json_object_list}{List of objects of type \code{CppJson} containing Json representation of a BCF model}

\item{json_string_list}{List of JSON strings which can be parsed to objects of type \code{CppJson} containing Json representation of a BCF model}
}
\value{
\code{saveBCFModelToJson} return an object of type \code{CppJson}.
\code{saveBCFModelToJsonFile} returns nothing, but writes to the provided filename.
\code{saveBCFModelToJsonString} returns a string dump of the BCF model's JSON representation.

\code{createBCFModelFromJson}, \code{createBCFModelFromJsonFile}, \code{createBCFModelFromJsonString},
\code{createBCFModelFromCombinedJson}, and \code{createBCFModelFromCombinedJsonString} all return
objects of type \code{bcfmodel}.
}
\description{
BCF models contains external pointers to C++ objects, which means they cannot
be correctly serialized to \code{.Rds} from an R session in their default state.
These functions allow us to convert between C++ data structures and a persistent JSON
representation. The \code{CppJson} class wraps a performant C++ JSON API, and the functions
\code{saveBCFModelToJson} and \code{createBCFModelFromJson} save to and load from this format.
This representation, of course, also relies on external C++ pointers, so in order to
save and reload BCF models across sessions, we provide two other interfaces.

\code{saveBCFModelToJsonFile} and \code{createBCFModelFromJsonFile} save or reload a BCF model's JSON
representation directly to / from a \code{.json} file.

\code{saveBCFModelToJsonString} and \code{createBCFModelFromJsonString} handle in-memory strings containing JSON data,
which can be written to disk or passed between processes.

Finally, for cases in which multiple BCF models have been sampled (for instance, sampled in multiple processes
via \code{doParallel}), we offer \code{createBCFModelFromCombinedJson} and \code{createBCFModelFromCombinedJsonString} for
loading a new combined BCF model from a list of BCF JSON objects or strings.
}
\examples{
# Generate data
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
pi_X <- runif(n, 0.3, 0.7)
Z <- rbinom(n, p = pi_X, size = 1)
y <- X[,1] + Z + rnorm(n, 0, 1)

# Sample BCF model
bcf_model <- bcf(X_train = X, Z_train = Z, propensity_train = pi_X, y_train = y,
                 num_gfr = 0, num_burnin = 0, num_mcmc = 10)

# Save to in-memory JSON
bcf_json <- saveBCFModelToJson(bcf_model)
# Save to JSON string
bcf_json_string <- saveBCFModelToJsonString(bcf_model)
# Save to JSON file
tmpjson <- tempfile(fileext = ".json")
saveBCFModelToJsonFile(bcf_model, file.path(tmpjson))

# Reload BCF model from in-memory JSON object
bcf_model_roundtrip <- createBCFModelFromJson(bcf_json)
# Reload BCF model from JSON string
bcf_model_roundtrip <- createBCFModelFromJsonString(bcf_json_string)
# Reload BCF model from JSON file
bcf_model_roundtrip <- createBCFModelFromJsonFile(file.path(tmpjson))
unlink(tmpjson)
# Reload BCF model from list of JSON objects
bcf_model_roundtrip <- createBCFModelFromCombinedJson(list(bcf_json))
# Reload BCF model from list of JSON strings
bcf_model_roundtrip <- createBCFModelFromCombinedJsonString(list(bcf_json_string))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forest.R
\name{Forest}
\alias{Forest}
\title{Forest C++ Wrapper}
\description{
Wrapper around a C++ class that stores a single ensemble of decision trees (often treated as the "active forest" / current state of a forest term in a sampling loop in R)

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{forest_ptr}}{External pointer to a C++ TreeEnsemble class}

\item{\code{internal_forest_is_empty}}{Whether the forest has not yet been "initialized" such that its \code{predict} function can be called.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Forest-new}{\code{Forest$new()}}
\item \href{#method-Forest-merge_forest}{\code{Forest$merge_forest()}}
\item \href{#method-Forest-add_constant}{\code{Forest$add_constant()}}
\item \href{#method-Forest-multiply_constant}{\code{Forest$multiply_constant()}}
\item \href{#method-Forest-predict}{\code{Forest$predict()}}
\item \href{#method-Forest-predict_raw}{\code{Forest$predict_raw()}}
\item \href{#method-Forest-set_root_leaves}{\code{Forest$set_root_leaves()}}
\item \href{#method-Forest-prepare_for_sampler}{\code{Forest$prepare_for_sampler()}}
\item \href{#method-Forest-adjust_residual}{\code{Forest$adjust_residual()}}
\item \href{#method-Forest-num_trees}{\code{Forest$num_trees()}}
\item \href{#method-Forest-leaf_dimension}{\code{Forest$leaf_dimension()}}
\item \href{#method-Forest-is_leaf_constant}{\code{Forest$is_leaf_constant()}}
\item \href{#method-Forest-is_exponentiated}{\code{Forest$is_exponentiated()}}
\item \href{#method-Forest-add_numeric_split_tree}{\code{Forest$add_numeric_split_tree()}}
\item \href{#method-Forest-get_tree_leaves}{\code{Forest$get_tree_leaves()}}
\item \href{#method-Forest-get_tree_split_counts}{\code{Forest$get_tree_split_counts()}}
\item \href{#method-Forest-get_forest_split_counts}{\code{Forest$get_forest_split_counts()}}
\item \href{#method-Forest-tree_max_depth}{\code{Forest$tree_max_depth()}}
\item \href{#method-Forest-average_max_depth}{\code{Forest$average_max_depth()}}
\item \href{#method-Forest-is_empty}{\code{Forest$is_empty()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-new"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-new}{}}}
\subsection{Method \code{new()}}{
Create a new Forest object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$new(
  num_trees,
  leaf_dimension = 1,
  is_leaf_constant = FALSE,
  is_exponentiated = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_trees}}{Number of trees in the forest}

\item{\code{leaf_dimension}}{Dimensionality of the outcome model}

\item{\code{is_leaf_constant}}{Whether leaf is constant}

\item{\code{is_exponentiated}}{Whether forest predictions should be exponentiated before being returned}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{Forest} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-merge_forest"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-merge_forest}{}}}
\subsection{Method \code{merge_forest()}}{
Create a larger forest by merging the trees of this forest with those of another forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$merge_forest(forest)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest}}{Forest to be merged into this forest}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-add_constant"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-add_constant}{}}}
\subsection{Method \code{add_constant()}}{
Add a constant value to every leaf of every tree in an ensemble. If leaves are multi-dimensional, \code{constant_value} will be added to every dimension of the leaves.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$add_constant(constant_value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{constant_value}}{Value that will be added to every leaf of every tree}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-multiply_constant"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-multiply_constant}{}}}
\subsection{Method \code{multiply_constant()}}{
Multiply every leaf of every tree by a constant value. If leaves are multi-dimensional, \code{constant_multiple} will be multiplied through every dimension of the leaves.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$multiply_constant(constant_multiple)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{constant_multiple}}{Value that will be multiplied by every leaf of every tree}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-predict"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-predict}{}}}
\subsection{Method \code{predict()}}{
Predict forest on every sample in \code{forest_dataset}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$predict(forest_dataset)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{\code{ForestDataset} R class}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
vector of predictions with as many rows as in \code{forest_dataset}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-predict_raw"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-predict_raw}{}}}
\subsection{Method \code{predict_raw()}}{
Predict "raw" leaf values (without being multiplied by basis) for every sample in \code{forest_dataset}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$predict_raw(forest_dataset)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{\code{ForestDataset} R class}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Array of predictions for each observation in \code{forest_dataset} and
each sample in the \code{ForestSamples} class with each prediction having the
dimensionality of the forests' leaf model. In the case of a constant leaf model
or univariate leaf regression, this array is a vector (length is the number of
observations). In the case of a multivariate leaf regression,
this array is a matrix (number of observations by leaf model dimension,
number of samples).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-set_root_leaves"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-set_root_leaves}{}}}
\subsection{Method \code{set_root_leaves()}}{
Set a constant predicted value for every tree in the ensemble.
Stops program if any tree is more than a root node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$set_root_leaves(leaf_value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{leaf_value}}{Constant leaf value(s) to be fixed for each tree in the ensemble indexed by \code{forest_num}. Can be either a single number or a vector, depending on the forest's leaf dimension.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-prepare_for_sampler"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-prepare_for_sampler}{}}}
\subsection{Method \code{prepare_for_sampler()}}{
Set a constant predicted value for every tree in the ensemble.
Stops program if any tree is more than a root node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$prepare_for_sampler(
  dataset,
  outcome,
  forest_model,
  leaf_model_int,
  leaf_value
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dataset}}{\code{ForestDataset} Dataset class (covariates, basis, etc...)}

\item{\code{outcome}}{\code{Outcome} Outcome class (residual / partial residual)}

\item{\code{forest_model}}{\code{ForestModel} object storing tracking structures used in training / sampling}

\item{\code{leaf_model_int}}{Integer value encoding the leaf model type (0 = constant gaussian, 1 = univariate gaussian, 2 = multivariate gaussian, 3 = log linear variance).}

\item{\code{leaf_value}}{Constant leaf value(s) to be fixed for each tree in the ensemble indexed by \code{forest_num}. Can be either a single number or a vector, depending on the forest's leaf dimension.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-adjust_residual"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-adjust_residual}{}}}
\subsection{Method \code{adjust_residual()}}{
Adjusts residual based on the predictions of a forest

This is typically run just once at the beginning of a forest sampling algorithm.
After trees are initialized with constant root node predictions, their root predictions are subtracted out of the residual.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$adjust_residual(dataset, outcome, forest_model, requires_basis, add)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dataset}}{\code{ForestDataset} object storing the covariates and bases for a given forest}

\item{\code{outcome}}{\code{Outcome} object storing the residuals to be updated based on forest predictions}

\item{\code{forest_model}}{\code{ForestModel} object storing tracking structures used in training / sampling}

\item{\code{requires_basis}}{Whether or not a forest requires a basis for prediction}

\item{\code{add}}{Whether forest predictions should be added to or subtracted from residuals}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-num_trees"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-num_trees}{}}}
\subsection{Method \code{num_trees()}}{
Return number of trees in each ensemble of a \code{Forest} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$num_trees()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Tree count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-leaf_dimension"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-leaf_dimension}{}}}
\subsection{Method \code{leaf_dimension()}}{
Return output dimension of trees in a \code{Forest} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$leaf_dimension()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Leaf node parameter size
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-is_leaf_constant"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-is_leaf_constant}{}}}
\subsection{Method \code{is_leaf_constant()}}{
Return constant leaf status of trees in a \code{Forest} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$is_leaf_constant()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if leaves are constant, \code{FALSE} otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-is_exponentiated"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-is_exponentiated}{}}}
\subsection{Method \code{is_exponentiated()}}{
Return exponentiation status of trees in a \code{Forest} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$is_exponentiated()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if leaf predictions must be exponentiated, \code{FALSE} otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-add_numeric_split_tree"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-add_numeric_split_tree}{}}}
\subsection{Method \code{add_numeric_split_tree()}}{
Add a numeric (i.e. \code{X[,i] <= c}) split to a given tree in the ensemble
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$add_numeric_split_tree(
  tree_num,
  leaf_num,
  feature_num,
  split_threshold,
  left_leaf_value,
  right_leaf_value
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tree_num}}{Index of the tree to be split}

\item{\code{leaf_num}}{Leaf to be split}

\item{\code{feature_num}}{Feature that defines the new split}

\item{\code{split_threshold}}{Value that defines the cutoff of the new split}

\item{\code{left_leaf_value}}{Value (or vector of values) to assign to the newly created left node}

\item{\code{right_leaf_value}}{Value (or vector of values) to assign to the newly created right node}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-get_tree_leaves"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-get_tree_leaves}{}}}
\subsection{Method \code{get_tree_leaves()}}{
Retrieve a vector of indices of leaf nodes for a given tree in a given forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$get_tree_leaves(tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tree_num}}{Index of the tree for which leaf indices will be retrieved}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-get_tree_split_counts"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-get_tree_split_counts}{}}}
\subsection{Method \code{get_tree_split_counts()}}{
Retrieve a vector of split counts for every training set variable in a given tree in the forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$get_tree_split_counts(tree_num, num_features)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tree_num}}{Index of the tree for which split counts will be retrieved}

\item{\code{num_features}}{Total number of features in the training set}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-get_forest_split_counts"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-get_forest_split_counts}{}}}
\subsection{Method \code{get_forest_split_counts()}}{
Retrieve a vector of split counts for every training set variable in the forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$get_forest_split_counts(num_features)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_features}}{Total number of features in the training set}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-tree_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-tree_max_depth}{}}}
\subsection{Method \code{tree_max_depth()}}{
Maximum depth of a specific tree in the forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$tree_max_depth(tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tree_num}}{Tree index within forest}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Maximum leaf depth
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-average_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-average_max_depth}{}}}
\subsection{Method \code{average_max_depth()}}{
Average the maximum depth of each tree in the forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$average_max_depth()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Average maximum depth
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Forest-is_empty"></a>}}
\if{latex}{\out{\hypertarget{method-Forest-is_empty}{}}}
\subsection{Method \code{is_empty()}}{
When a forest object is created, it is "empty" in the sense that none
of its component trees have leaves with values. There are two ways to
"initialize" a Forest object. First, the \code{set_root_leaves()} method
simply initializes every tree in the forest to a single node carrying
the same (user-specified) leaf value. Second, the \code{prepare_for_sampler()}
method initializes every tree in the forest to a single node with the
same value and also propagates this information through to a ForestModel
object, which must be synchronized with a Forest during a forest
sampler loop.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Forest$is_empty()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if a Forest has not yet been initialized with a constant
root value, \code{FALSE} otherwise if the forest has already been
initialized / grown.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variance.R
\name{sampleGlobalErrorVarianceOneIteration}
\alias{sampleGlobalErrorVarianceOneIteration}
\title{Sample Global Error Variance}
\usage{
sampleGlobalErrorVarianceOneIteration(residual, dataset, rng, a, b)
}
\arguments{
\item{residual}{Outcome class}

\item{dataset}{ForestDataset class}

\item{rng}{C++ random number generator}

\item{a}{Global variance shape parameter}

\item{b}{Global variance scale parameter}
}
\value{
None
}
\description{
Sample one iteration of the (inverse gamma) global variance model

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
X <- matrix(runif(10*100), ncol = 10)
y <- -5 + 10*(X[,1] > 0.5) + rnorm(100)
y_std <- (y-mean(y))/sd(y)
forest_dataset <- createForestDataset(X)
outcome <- createOutcome(y_std)
rng <- createCppRNG(1234)
a <- 1.0
b <- 1.0
sigma2 <- sampleGlobalErrorVarianceOneIteration(outcome, forest_dataset, rng, a, b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{createForestModelConfig}
\alias{createForestModelConfig}
\title{Create ForestModelConfig Object}
\usage{
createForestModelConfig(
  feature_types = NULL,
  sweep_update_indices = NULL,
  num_trees = NULL,
  num_features = NULL,
  num_observations = NULL,
  variable_weights = NULL,
  leaf_dimension = 1,
  alpha = 0.95,
  beta = 2,
  min_samples_leaf = 5,
  max_depth = -1,
  leaf_model_type = 1,
  leaf_model_scale = NULL,
  variance_forest_shape = 1,
  variance_forest_scale = 1,
  cloglog_forest_shape = 2,
  cloglog_forest_rate = 2,
  cutpoint_grid_size = 100,
  num_features_subsample = NULL
)
}
\arguments{
\item{feature_types}{Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)}

\item{sweep_update_indices}{Vector of (0-indexed) indices of trees to update in a sweep}

\item{num_trees}{Number of trees in the forest being sampled}

\item{num_features}{Number of features in training dataset}

\item{num_observations}{Number of observations in training dataset}

\item{variable_weights}{Vector specifying sampling probability for all p covariates in ForestDataset}

\item{leaf_dimension}{Dimension of the leaf model (default: \code{1})}

\item{alpha}{Root node split probability in tree prior (default: \code{0.95})}

\item{beta}{Depth prior penalty in tree prior (default: \code{2.0})}

\item{min_samples_leaf}{Minimum number of samples in a tree leaf (default: \code{5})}

\item{max_depth}{Maximum depth of any tree in the ensemble in the model. Setting to \code{-1} does not enforce any depth limits on trees. Default: \code{-1}.}

\item{leaf_model_type}{Integer specifying the leaf model type (0 = constant leaf, 1 = univariate leaf regression, 2 = multivariate leaf regression). Default: \code{0}.}

\item{leaf_model_scale}{Scale parameter used in Gaussian leaf models (can either be a scalar or a q x q matrix, where q is the dimensionality of the basis and is only >1 when \code{leaf_model_int = 2}). Calibrated internally as \code{1/num_trees}, propagated along diagonal if needed for multivariate leaf models.}

\item{variance_forest_shape}{Shape parameter for IG leaf models (applicable when \code{leaf_model_type = 3}). Default: \code{1}.}

\item{variance_forest_scale}{Scale parameter for IG leaf models (applicable when \code{leaf_model_type = 3}). Default: \code{1}.}

\item{cloglog_forest_shape}{Shape parameter for conditional gamma component of cloglog leaf models (applicable when \code{leaf_model_type = 4}). Default: \code{1}.}

\item{cloglog_forest_rate}{Rate parameter for conditional gamma component of cloglog leaf models (applicable when \code{leaf_model_type = 4}). Default: \code{1}.}

\item{cutpoint_grid_size}{Number of unique cutpoints to consider (default: \code{100})}

\item{num_features_subsample}{Number of features to subsample for the GFR algorithm}
}
\value{
ForestModelConfig object
}
\description{
Create a forest model config object

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
config <- createForestModelConfig(num_trees = 10, num_features = 5, num_observations = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_effects.R
\name{print.RandomEffectSamples}
\alias{print.RandomEffectSamples}
\title{Summarize a RandomEffectSamples Object}
\usage{
\method{print}{RandomEffectSamples}(x, ...)
}
\arguments{
\item{x}{RandomEffectSamples object}

\item{...}{Additional arguments}
}
\value{
RandomEffectSamples object unchanged after printing summary
}
\description{
Prints a summary of the RandomEffectSamples object, including number of forests and the underlying model of each forest.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcf.R
\name{plot.bcfmodel}
\alias{plot.bcfmodel}
\title{Plot BCF Model}
\usage{
\method{plot}{bcfmodel}(x, ...)
}
\arguments{
\item{x}{The BCF model object}

\item{...}{Additional arguments}
}
\value{
BCF model object unchanged after summarizing
}
\description{
Plot the BCF model fit and any relevant sampled quantities. This will default to a traceplot of the global error scale and the in-sample mean forest predictions for the first train set observation. Since \code{stochtree::bcf()} is flexible and it's possible to sample a model with a fixed global error scale and no mean forest, this procedure will throw an error if these two default terms are omitted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{createForestDataset}
\alias{createForestDataset}
\title{Create ForestDataset object}
\usage{
createForestDataset(covariates, basis = NULL, variance_weights = NULL)
}
\arguments{
\item{covariates}{Matrix of covariates}

\item{basis}{(Optional) Matrix of bases used to define a leaf regression}

\item{variance_weights}{(Optional) Vector of observation-specific variance weights}
}
\value{
\code{ForestDataset} object
}
\description{
Create a forest dataset object

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
covariate_matrix <- matrix(runif(10*100), ncol = 10)
basis_matrix <- matrix(rnorm(3*100), ncol = 3)
weight_vector <- rnorm(100)
forest_dataset <- createForestDataset(covariate_matrix)
forest_dataset <- createForestDataset(covariate_matrix, basis_matrix)
forest_dataset <- createForestDataset(covariate_matrix, basis_matrix, weight_vector)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{BARTSerialization}
\alias{BARTSerialization}
\alias{saveBARTModelToJson}
\alias{saveBARTModelToJsonFile}
\alias{saveBARTModelToJsonString}
\alias{createBARTModelFromJson}
\alias{createBARTModelFromJsonFile}
\alias{createBARTModelFromJsonString}
\alias{createBARTModelFromCombinedJson}
\alias{createBARTModelFromCombinedJsonString}
\title{BART Serialization Routines}
\usage{
saveBARTModelToJson(object)

saveBARTModelToJsonFile(object, filename)

saveBARTModelToJsonString(object)

createBARTModelFromJson(json_object)

createBARTModelFromJsonFile(json_filename)

createBARTModelFromJsonString(json_string)

createBARTModelFromCombinedJson(json_object_list)

createBARTModelFromCombinedJsonString(json_string_list)
}
\arguments{
\item{object}{Object of type \code{bartmodel} containing draws of a BART model and associated sampling outputs.}

\item{filename}{String of filepath, must end in ".json"}

\item{json_object}{Object of type \code{CppJson} containing Json representation of a BART model}

\item{json_filename}{String of filepath, must end in ".json"}

\item{json_string}{JSON string dump}

\item{json_object_list}{List of objects of type \code{CppJson} containing Json representation of a BART model}

\item{json_string_list}{List of JSON strings which can be parsed to objects of type \code{CppJson} containing Json representation of a BART model}
}
\value{
\code{saveBARTModelToJson} return an object of type \code{CppJson}.
\code{saveBARTModelToJsonString} returns a string dump of the BART model's JSON representation.
\code{saveBARTModelToJsonFile} returns nothing, but writes to the provided filename.

\code{createBARTModelFromJson}, \code{createBARTModelFromJsonFile}, \code{createBARTModelFromJsonString},
\code{createBARTModelFromCombinedJson}, and \code{createBARTModelFromCombinedJsonString} all return
objects of type \code{bartmodel}.
}
\description{
BART models contains external pointers to C++ objects, which means they cannot
be correctly serialized to \code{.Rds} from an R session in their default state.
This group of serialization functions allow us to convert between C++ data structures and a persistent JSON
representation. The \code{CppJson} class wraps a performant C++ JSON API, and the functions
\code{saveBARTModelToJson} and \code{createBARTModelFromJson} save to and load from this format.
This representation, of course, also relies on external C++ pointers, so in order to
save and reload BART models across sessions, we provide two other interfaces.

\code{saveBARTModelToJsonString} converts a BART model to an in-memory string containing the model's
JSON representation and \code{createBARTModelFromJsonString} converts this representation back to a BART model object.

\code{saveBARTModelToJsonFile} and \code{createBARTModelFromJsonFile} save or reload a BART model
directly to / from a \code{.json} file.

Finally, for cases in which multiple BART models have been sampled (for instance, multiple processes
run via \code{doParallel}), we offer \code{createBARTModelFromCombinedJson} and \code{createBARTModelFromCombinedJsonString} for
loading a new combined BART model from a list of BART JSON objects / strings.
}
\examples{
# Generate data
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
y <- X[,1] + rnorm(n, 0, 1)

# Sample BART model
bart_model <- bart(X_train = X, y_train = y, num_gfr = 0,
                   num_burnin = 0, num_mcmc = 10)

# Save to in-memory JSON
bart_json <- saveBARTModelToJson(bart_model)
# Save to JSON string
bart_json_string <- saveBARTModelToJsonString(bart_model)
# Save to JSON file
tmpjson <- tempfile(fileext = ".json")
saveBARTModelToJsonFile(bart_model, file.path(tmpjson))

# Reload BART model from in-memory JSON object
bart_model_roundtrip <- createBARTModelFromJson(bart_json)
# Reload BART model from JSON string
bart_model_roundtrip <- createBARTModelFromJsonString(bart_json_string)
# Reload BART model from JSON file
bart_model_roundtrip <- createBARTModelFromJsonFile(file.path(tmpjson))
unlink(tmpjson)
# Reload BART model from list of JSON objects
bart_model_roundtrip <- createBARTModelFromCombinedJson(list(bart_json))
# Reload BART model from list of JSON strings
bart_model_roundtrip <- createBARTModelFromCombinedJsonString(list(bart_json_string))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialization.R
\name{JSONCreation}
\alias{JSONCreation}
\alias{createCppJson}
\alias{createCppJsonFile}
\alias{createCppJsonString}
\title{JSON Creation Routines}
\usage{
createCppJson()

createCppJsonFile(json_filename)

createCppJsonString(json_string)
}
\arguments{
\item{json_filename}{Name of JSON file to read. Must end in \code{.json}.}

\item{json_string}{JSON string}
}
\value{
Each of the functions in this group returns a \code{CppJson} object.
}
\description{
The \link{CppJson} class wraps an external pointer to C++ JSON object for a seamless R serialization interface.
This function group provides several utilities for creating \link{CppJson} objects: from scratch (i.e. initializing an empty JSON object),
from a JSON file, or from a string that parses to valid JSON.

\code{createCppJson} creates a new (empty) \link{CppJson} object. \code{createCppJsonFile} creates and populates a \link{CppJson} object from data in a JSON file.
\code{createCppJsonString} creates and populates a \link{CppJson} object from a JSON string.

These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
example_vec <- runif(10)
example_json <- createCppJson()
example_json$add_vector("myvec", example_vec)
tmpjson <- tempfile(fileext = ".json")
example_json$save_file(file.path(tmpjson))
example_json_roundtrip <- createCppJsonFile(file.path(tmpjson))
unlink(tmpjson)
example_json_string <- example_json$return_json_string()
example_json_roundtrip <- createCppJsonString(example_json_string)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{DataPreprocessing}
\alias{DataPreprocessing}
\alias{preprocessTrainData}
\alias{preprocessPredictionData}
\alias{savePreprocessorToJson}
\alias{savePreprocessorToJsonString}
\alias{createPreprocessorFromJson}
\alias{createPreprocessorFromJsonString}
\title{Data Preprocessing Routines}
\usage{
preprocessTrainData(input_data)

preprocessPredictionData(input_data, metadata)

savePreprocessorToJson(object)

savePreprocessorToJsonString(object)

createPreprocessorFromJson(json_object)

createPreprocessorFromJsonString(json_string)
}
\arguments{
\item{input_data}{Covariates, provided as either a dataframe or a matrix}

\item{metadata}{List containing information on variables, including train set
categories for categorical variables}

\item{object}{List containing information on variables, including train set
categories for categorical variables}

\item{json_object}{in-memory wrapper around JSON C++ object containing covariate preprocessor metadata}

\item{json_string}{in-memory JSON string containing covariate preprocessor metadata}
}
\value{
\code{preprocessTrainData} returns a list with transformed matrix data and a "metadata" list with details on the preprocessing procedures applied.
\code{preprocessPredictionData} returns a matrix reflecting the data transformations specified in the provided metadata list.

\code{savePreprocessorToJson} return an object of type \code{CppJson}.
\code{savePreprocessorToJsonString} returns a string dump of the preprocessor's JSON representation.

\code{createPreprocessorFromJson} and \code{createPreprocessorFromJsonString} both return metadata lists.
}
\description{
The functions in this group are designed to handle data preprocessing for stochastic forest models.
For example, factor-valued columns in data frames are either one-hot encoded or converted to integer indices
before the dataframe is converted to a standard matrix format for sampling.
This preprocessing routine defines a set of "steps" that must be repeated on out-of-sample datasets before
predictions can be obtained from a sampling model.

\code{preprocessTrainData} preprocesses covariates for the forest sampler routines, depending on the input type.
DataFrames will be preprocessed based on their column types (numeric columns are not modified, ordered factors are
integer coded, and unordered factors are one-hot encoded). Matrices are unmodified (assuming all columns are numeric).
This function also records and returns a "metadata" list with preprocessing details to ensure that other datasets can be preprocessing identically.

\code{preprocessPredictionData} preprocesses covariates for the forest sampler routines, based on the steps outlined in a metadata list produced by \code{preprocessTrainData}.

These procedures are handled internally in the \code{\link[=bart]{bart()}} and \code{\link[=bcf]{bcf()}} functions, but they are provided in \code{stochtree} as convenience functions for users writing custom samplers.
Furthermore, while R lists can be serialized to \code{RDS} format, we offer a number of JSON serialization routines for the metadata list produced by \code{preprocessTrainData} for
consistency with the broader serialization approach of stochtree (see \link{BARTSerialization} and \link{BCFSerialization}).

Following the API for serializing \code{bartmodel} and \code{bcfmodel} objects, we can convert metadata to JSON or JSON strings via
\code{savePreprocessorToJson} and \code{savePreprocessorToJsonString}. Similarly, we can reload a metadata list from JSON or JSON strings
via \code{createPreprocessorFromJson} and \code{createPreprocessorFromJsonString}.
}
\examples{
# Check that running the same data through `preprocessTrainData`
# and `preprocessPredictionData` yields the same result
n <- 100
x1 <- rnorm(n)
x2 <- factor(sample(1:3, n, replace = TRUE), ordered = TRUE)
x3 <- factor(sample(1:3, n, replace = TRUE), ordered = FALSE)
df1 <- data.frame(x1 = x1, x2 = x2, x3 = x3)
df2 <- data.frame(x1 = x1, x2 = x2, x3 = x3)
preprocess_train_list <- preprocessTrainData(df1)
df1_process <- preprocess_train_list$data
df1_metadata <- preprocess_train_list$metadata
df2_process <- preprocessPredictionData(df2, df1_metadata)
all.equal(df1_process, df2_process)

# Save to in-memory JSON
metadata_json <- savePreprocessorToJson(df1_metadata)
# Save to JSON string
metadata_json_string <- savePreprocessorToJsonString(df1_metadata)

# Reload metadata list from in-memory JSON object
metadata_roundtrip <- createPreprocessorFromJson(metadata_json)
# Reload metadata list from JSON string
metadata_roundtrip <- createPreprocessorFromJsonString(metadata_json_string)

cov_df <- data.frame(x1 = 1:5, x2 = 5:1, x3 = 6:10)
metadata <- list(num_ordered_cat_vars = 0, num_unordered_cat_vars = 0,
                 num_numeric_vars = 3, numeric_vars = c("x1", "x2", "x3"))
X_preprocessed <- preprocessPredictionData(cov_df, metadata)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{print.bartmodel}
\alias{print.bartmodel}
\title{Print Summary of BART Model}
\usage{
\method{print}{bartmodel}(x, ...)
}
\arguments{
\item{x}{The BART model object}

\item{...}{Additional arguments}
}
\value{
BART model object unchanged after printing summary
}
\description{
Prints a summary of the BART model, including the model terms and their specifications.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{sample_without_replacement}
\alias{sample_without_replacement}
\title{Sample Without Replacement}
\usage{
sample_without_replacement(
  population_vector,
  sampling_probabilities,
  sample_size
)
}
\arguments{
\item{population_vector}{Vector from which to draw samples.}

\item{sampling_probabilities}{Vector of probabilities of drawing each element of \code{population_vector}.}

\item{sample_size}{Number of samples to draw from \code{population_vector}. Must be less than or equal to \code{length(population_vector)}}
}
\value{
Vector of size \code{sample_size}
}
\description{
Draw \code{sample_size} samples from \code{population_vector} without replacement, weighted by \code{sampling_probabilities}

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
a <- as.integer(c(4,3,2,5,1,9,7))
p <- c(0.7,0.2,0.05,0.02,0.01,0.01,0.01)
num_samples <- 5
sample_without_replacement(a, p, num_samples)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialization.R
\name{loadScalarJson}
\alias{loadScalarJson}
\title{Load Scalar from JSON}
\usage{
loadScalarJson(json_object, json_scalar_label, subfolder_name = NULL)
}
\arguments{
\item{json_object}{Object of class \code{CppJson}}

\item{json_scalar_label}{Label referring to a particular scalar / string value (i.e. "num_samples") in the overall json hierarchy}

\item{subfolder_name}{(Optional) Name of the subfolder / hierarchy under which vector sits}
}
\value{
R vector
}
\description{
Load a scalar from json

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
example_scalar <- 5.4
example_json <- createCppJson()
example_json$add_scalar("myscalar", example_scalar)
roundtrip_scalar <- loadScalarJson(example_json, "myscalar")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calibration.R
\name{calibrateInverseGammaErrorVariance}
\alias{calibrateInverseGammaErrorVariance}
\title{Calibrate Inverse Gamma Prior}
\usage{
calibrateInverseGammaErrorVariance(
  y,
  X,
  W = NULL,
  nu = 3,
  quant = 0.9,
  standardize = TRUE
)
}
\arguments{
\item{y}{Outcome to be modeled using BART, BCF or another nonparametric ensemble method.}

\item{X}{Covariates to be used to partition trees in an ensemble or series of ensemble.}

\item{W}{(Optional) Basis used to define a "leaf regression" model for each decision tree. The "classic" BART model assumes a constant leaf parameter, which is equivalent to a "leaf regression" on a basis of all ones, though it is not necessary to pass a vector of ones, here or to the BART function. Default: \code{NULL}.}

\item{nu}{The shape parameter for the global error variance's IG prior. The scale parameter in the Sparapani et al (2021) parameterization is defined as \code{nu*lambda} where \code{lambda} is the output of this function. Default: \code{3}.}

\item{quant}{(Optional) Quantile of the inverse gamma prior distribution represented by a linear-regression-based overestimate of \code{sigma^2}. Default: \code{0.9}.}

\item{standardize}{(Optional) Whether or not outcome should be standardized (\code{(y-mean(y))/sd(y)}) before calibration of \code{lambda}. Default: \code{TRUE}.}
}
\value{
Value of \code{lambda} which determines the scale parameter of the global error variance prior (\code{sigma^2 ~ IG(nu,nu*lambda)})
}
\description{
Calibrate the scale parameter on an inverse gamma prior for the global error variance as in Chipman et al (2022)

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}

Chipman, H., George, E., Hahn, R., McCulloch, R., Pratola, M. and Sparapani, R. (2022). Bayesian Additive Regression Trees, Computational Approaches. In Wiley StatsRef: Statistics Reference Online (eds N. Balakrishnan, T. Colton, B. Everitt, W. Piegorsch, F. Ruggeri and J.L. Teugels). \doi{10.1002/9781118445112.stat08288}
}
\examples{
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
y <- 10*X[,1] - 20*X[,2] + rnorm(n)
nu <- 3
lambda <- calibrateInverseGammaErrorVariance(y, X, nu = nu)
sigma2hat <- mean(resid(lm(y~X))^2)
mean(var(y)/rgamma(100000, nu, rate = nu*lambda) < sigma2hat)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{plot.bartmodel}
\alias{plot.bartmodel}
\title{Plot BART Model Fit}
\usage{
\method{plot}{bartmodel}(x, ...)
}
\arguments{
\item{x}{The BART model object}

\item{...}{Additional arguments}
}
\value{
BART model object unchanged after summarizing
}
\description{
Plot the BART model fit and any relevant sampled quantities. This will default to a traceplot of the global error scale and the in-sample mean forest predictions for the first train set observation. Since \code{stochtree::bart()} is flexible and it's possible to sample a model with a fixed global error scale and no mean forest, this procedure is adaptive and will attempt to plot a trace of whichever model terms are included if these two default terms are omitted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{createRandomEffectsDataset}
\alias{createRandomEffectsDataset}
\title{Create RandomEffectsDataset Object}
\usage{
createRandomEffectsDataset(group_labels, basis, variance_weights = NULL)
}
\arguments{
\item{group_labels}{Vector of group labels}

\item{basis}{Matrix of bases used to define the random effects regression (for an intercept-only model, pass an array of ones)}

\item{variance_weights}{(Optional) Vector of observation-specific variance weights}
}
\value{
\code{RandomEffectsDataset} object
}
\description{
Create a random effects dataset object

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
rfx_group_ids <- sample(1:2, size = 100, replace = TRUE)
rfx_basis <- matrix(rnorm(3*100), ncol = 3)
weight_vector <- rnorm(100)
rfx_dataset <- createRandomEffectsDataset(rfx_group_ids, rfx_basis)
rfx_dataset <- createRandomEffectsDataset(rfx_group_ids, rfx_basis, weight_vector)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_effects.R
\name{RandomEffectSamples}
\alias{RandomEffectSamples}
\title{Random Effect Container C++ Wrapper}
\description{
Class that wraps the "persistent" aspects of a C++ random effects model, including
draws of the parameters and a map from the original label indices to the
0-indexed label numbers used to place group samples in memory (i.e. the
first label is stored in column 0 of the sample matrix, the second label
is store in column 1 of the sample matrix, etc...)

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}

Coordinates various C++ random effects classes and persists those
needed for prediction / serialization
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{rfx_container_ptr}}{External pointer to a C++ StochTree::RandomEffectsContainer class}

\item{\code{label_mapper_ptr}}{External pointer to a C++ StochTree::LabelMapper class}

\item{\code{training_group_ids}}{Unique vector of group IDs that were in the training dataset}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-RandomEffectSamples-new}{\code{RandomEffectSamples$new()}}
\item \href{#method-RandomEffectSamples-load_in_session}{\code{RandomEffectSamples$load_in_session()}}
\item \href{#method-RandomEffectSamples-load_from_json}{\code{RandomEffectSamples$load_from_json()}}
\item \href{#method-RandomEffectSamples-append_from_json}{\code{RandomEffectSamples$append_from_json()}}
\item \href{#method-RandomEffectSamples-load_from_json_string}{\code{RandomEffectSamples$load_from_json_string()}}
\item \href{#method-RandomEffectSamples-append_from_json_string}{\code{RandomEffectSamples$append_from_json_string()}}
\item \href{#method-RandomEffectSamples-predict}{\code{RandomEffectSamples$predict()}}
\item \href{#method-RandomEffectSamples-extract_parameter_samples}{\code{RandomEffectSamples$extract_parameter_samples()}}
\item \href{#method-RandomEffectSamples-delete_sample}{\code{RandomEffectSamples$delete_sample()}}
\item \href{#method-RandomEffectSamples-extract_label_mapping}{\code{RandomEffectSamples$extract_label_mapping()}}
\item \href{#method-RandomEffectSamples-num_samples}{\code{RandomEffectSamples$num_samples()}}
\item \href{#method-RandomEffectSamples-num_components}{\code{RandomEffectSamples$num_components()}}
\item \href{#method-RandomEffectSamples-num_groups}{\code{RandomEffectSamples$num_groups()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-new"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-new}{}}}
\subsection{Method \code{new()}}{
Create a new RandomEffectSamples object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$new()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A new \code{RandomEffectSamples} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-load_in_session"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-load_in_session}{}}}
\subsection{Method \code{load_in_session()}}{
Construct RandomEffectSamples object from other "in-session" R objects
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$load_in_session(
  num_components,
  num_groups,
  random_effects_tracker
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_components}}{Number of "components" or bases defining the random effects regression}

\item{\code{num_groups}}{Number of random effects groups}

\item{\code{random_effects_tracker}}{Object of type \code{RandomEffectsTracker}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-load_from_json"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-load_from_json}{}}}
\subsection{Method \code{load_from_json()}}{
Construct RandomEffectSamples object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$load_from_json(
  json_object,
  json_rfx_container_label,
  json_rfx_mapper_label,
  json_rfx_groupids_label
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_object}}{Object of class \code{CppJson}}

\item{\code{json_rfx_container_label}}{Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy}

\item{\code{json_rfx_mapper_label}}{Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy}

\item{\code{json_rfx_groupids_label}}{Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{RandomEffectSamples} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-append_from_json"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-append_from_json}{}}}
\subsection{Method \code{append_from_json()}}{
Append random effect draws to \code{RandomEffectSamples} object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$append_from_json(
  json_object,
  json_rfx_container_label,
  json_rfx_mapper_label,
  json_rfx_groupids_label
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_object}}{Object of class \code{CppJson}}

\item{\code{json_rfx_container_label}}{Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy}

\item{\code{json_rfx_mapper_label}}{Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy}

\item{\code{json_rfx_groupids_label}}{Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-load_from_json_string"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-load_from_json_string}{}}}
\subsection{Method \code{load_from_json_string()}}{
Construct RandomEffectSamples object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$load_from_json_string(
  json_string,
  json_rfx_container_label,
  json_rfx_mapper_label,
  json_rfx_groupids_label
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_string}}{JSON string which parses into object of class \code{CppJson}}

\item{\code{json_rfx_container_label}}{Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy}

\item{\code{json_rfx_mapper_label}}{Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy}

\item{\code{json_rfx_groupids_label}}{Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{RandomEffectSamples} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-append_from_json_string"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-append_from_json_string}{}}}
\subsection{Method \code{append_from_json_string()}}{
Append random effect draws to \code{RandomEffectSamples} object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$append_from_json_string(
  json_string,
  json_rfx_container_label,
  json_rfx_mapper_label,
  json_rfx_groupids_label
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_string}}{JSON string which parses into object of class \code{CppJson}}

\item{\code{json_rfx_container_label}}{Label referring to a particular rfx sample container (i.e. "random_effect_container_0") in the overall json hierarchy}

\item{\code{json_rfx_mapper_label}}{Label referring to a particular rfx label mapper (i.e. "random_effect_label_mapper_0") in the overall json hierarchy}

\item{\code{json_rfx_groupids_label}}{Label referring to a particular set of rfx group IDs (i.e. "random_effect_groupids_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-predict"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-predict}{}}}
\subsection{Method \code{predict()}}{
Predict random effects for each observation implied by \code{rfx_group_ids} and \code{rfx_basis}.
If a random effects model is "intercept-only" the \code{rfx_basis} will be a vector of ones of size \code{length(rfx_group_ids)}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$predict(rfx_group_ids, rfx_basis = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{rfx_group_ids}}{Indices of random effects groups in a prediction set}

\item{\code{rfx_basis}}{(Optional) Basis used for random effects prediction}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Matrix with as many rows as observations provided and as many columns as samples drawn of the model.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-extract_parameter_samples"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-extract_parameter_samples}{}}}
\subsection{Method \code{extract_parameter_samples()}}{
Extract the random effects parameters sampled. With the "redundant parameterization"
of Gelman et al (2008), this includes four parameters: alpha (the "working parameter"
shared across every group), xi (the "group parameter" sampled separately for each group),
beta (the product of alpha and xi, which corresponds to the overall group-level random effects),
and sigma (group-independent prior variance for each component of xi).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$extract_parameter_samples()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
List of arrays. The alpha array has dimension (\code{num_components}, \code{num_samples}) and is simply a vector if \code{num_components = 1}.
The xi and beta arrays have dimension (\code{num_components}, \code{num_groups}, \code{num_samples}) and are simply matrices if \code{num_components = 1}.
The sigma array has dimension (\code{num_components}, \code{num_samples}) and is simply a vector if \code{num_components = 1}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-delete_sample"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-delete_sample}{}}}
\subsection{Method \code{delete_sample()}}{
Modify the \code{RandomEffectsSamples} object by removing the parameter samples index by \code{sample_num}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$delete_sample(sample_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{sample_num}}{Index of the RFX sample to be removed}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-extract_label_mapping"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-extract_label_mapping}{}}}
\subsection{Method \code{extract_label_mapping()}}{
Convert the mapping of group IDs to random effect components indices from C++ to R native format
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$extract_label_mapping()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
List mapping group ID to random effect components.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-num_samples"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-num_samples}{}}}
\subsection{Method \code{num_samples()}}{
Query the number of samples in the \code{RandomEffectsSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$num_samples()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Integer number of samples
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-num_components"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-num_components}{}}}
\subsection{Method \code{num_components()}}{
Query the number of components in the \code{RandomEffectsSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$num_components()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Integer number of components
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectSamples-num_groups"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectSamples-num_groups}{}}}
\subsection{Method \code{num_groups()}}{
Query the number of groups in the \code{RandomEffectsSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectSamples$num_groups()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Integer number of groups
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior_transformation.R
\name{sampleBARTPosteriorPredictive}
\alias{sampleBARTPosteriorPredictive}
\title{Sample BART Posterior Predictive}
\usage{
sampleBARTPosteriorPredictive(
  model_object,
  X = NULL,
  leaf_basis = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL,
  num_draws_per_sample = NULL
)
}
\arguments{
\item{model_object}{A fitted BART model object of class \code{bartmodel}.}

\item{X}{A matrix or data frame of covariates. Required if the BART model depends on covariates (e.g., contains a mean or variance forest).}

\item{leaf_basis}{A matrix of bases for mean forest models with regression defined in the leaves. Required for "leaf regression" models.}

\item{rfx_group_ids}{A vector of group IDs for random effects model. Required if the BART model includes random effects.}

\item{rfx_basis}{A matrix of bases for random effects model. Required if the BART model includes random effects.}

\item{num_draws_per_sample}{The number of posterior predictive samples to draw for each posterior sample. Defaults to a heuristic based on the number of samples in a BART model (i.e. if the BART model has >1000 draws, we use 1 draw from the likelihood per sample, otherwise we upsample to ensure intervals are based on at least 1000 posterior predictive draws).}
}
\value{
Array of posterior predictive samples with dimensions (num_observations, num_posterior_samples, num_draws_per_sample) if num_draws_per_sample > 1, otherwise (num_observations, num_posterior_samples).
}
\description{
Sample from the posterior predictive distribution for outcomes modeled by BART
}
\examples{
n <- 100
p <- 5
X <- matrix(rnorm(n * p), nrow = n, ncol = p)
y <- 2 * X[,1] + rnorm(n)
bart_model <- bart(y_train = y, X_train = X)
ppd_samples <- sampleBARTPosteriorPredictive(
  model_object = bart_model, X = X
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{getRandomEffectSamples}
\alias{getRandomEffectSamples}
\title{Extract Random Effect Samples Generic Function}
\usage{
getRandomEffectSamples(object, ...)
}
\arguments{
\item{object}{Fitted model object from which to extract random effects}

\item{...}{Other parameters to be used in random effects extraction}
}
\value{
List of random effect samples
}
\description{
Generic function for extracting random effect samples from a model object (BCF, BART, etc...)
}
\examples{
n <- 100
p <- 10
X <- matrix(runif(n*p), ncol = p)
rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
rfx_basis <- rep(1.0, n)
y <- (-5 + 10*(X[,1] > 0.5)) + (-2*(rfx_group_ids==1)+2*(rfx_group_ids==2)) + rnorm(n)
bart_model <- bart(X_train=X, y_train=y, rfx_group_ids_train=rfx_group_ids,
                   rfx_basis_train = rfx_basis, num_gfr=0, num_mcmc=10)
rfx_samples <- getRandomEffectSamples(bart_model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior_transformation.R
\name{computeBARTPosteriorInterval}
\alias{computeBARTPosteriorInterval}
\title{Compute BART Posterior Credible Intervals}
\usage{
computeBARTPosteriorInterval(
  model_object,
  terms,
  level = 0.95,
  scale = "linear",
  X = NULL,
  leaf_basis = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL
)
}
\arguments{
\item{model_object}{A fitted BART or BCF model object of class \code{bartmodel}.}

\item{terms}{A character string specifying the model term(s) for which to compute intervals. Options for BART models are \code{"mean_forest"}, \code{"variance_forest"}, \code{"rfx"}, or \code{"y_hat"}.}

\item{level}{A numeric value between 0 and 1 specifying the credible interval level (default is 0.95 for a 95\% credible interval).}

\item{scale}{(Optional) Scale of mean function predictions. Options are "linear", which returns predictions on the original scale of the mean forest / RFX terms, and "probability". \code{scale = "probability"} is only valid for models fit with a probit / cloglog links on binary or ordinal outcomes. For binary outcome models, \code{scale = "probability"} will return an interval over the probability that \code{y == 1}. For ordinal outcome models, \code{scale = "probability"} will return intervals over the "survival function" \code{P(y > k)} for \verb{k = 1, 2, ..., K-1} where \code{K} is the total number of categories. Default: "linear".}

\item{X}{A matrix or data frame of covariates at which to compute the intervals. Required if the requested term depends on covariates (e.g., mean forest, variance forest, or overall predictions).}

\item{leaf_basis}{An optional matrix of basis function evaluations for mean forest models with regression defined in the leaves. Required for "leaf regression" models.}

\item{rfx_group_ids}{An optional vector of group IDs for random effects. Required if the requested term includes random effects.}

\item{rfx_basis}{An optional matrix of basis function evaluations for random effects. Required if the requested term includes random effects.}
}
\value{
A list containing the lower and upper bounds of the credible interval for the specified term. If multiple terms are requested, a named list with intervals for each term is returned.
}
\description{
Compute posterior credible intervals for specified terms from a fitted BART model. Supports intervals for mean functions, variance functions, random effects, and overall outcome predictions.
}
\examples{
n <- 100
p <- 5
X <- matrix(rnorm(n * p), nrow = n, ncol = p)
y <- 2 * X[,1] + rnorm(n)
bart_model <- bart(y_train = y, X_train = X)
intervals <- computeBARTPosteriorInterval(
 model_object = bart_model,
 terms = c("mean_forest", "y_hat"),
 X = X,
 level = 0.90
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variance.R
\name{sampleLeafVarianceOneIteration}
\alias{sampleLeafVarianceOneIteration}
\title{Sample Leaf Scale}
\usage{
sampleLeafVarianceOneIteration(forest, rng, a, b)
}
\arguments{
\item{forest}{C++ forest}

\item{rng}{C++ random number generator}

\item{a}{Leaf variance shape parameter}

\item{b}{Leaf variance scale parameter}
}
\value{
None
}
\description{
Sample one iteration of the leaf parameter variance model (only for univariate basis and constant leaf!)

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
num_trees <- 100
leaf_dimension <- 1
is_leaf_constant <- TRUE
is_exponentiated <- FALSE
active_forest <- createForest(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
rng <- createCppRNG(1234)
a <- 1.0
b <- 1.0
tau <- sampleLeafVarianceOneIteration(active_forest, rng, a, b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{CppRNG}
\alias{CppRNG}
\title{Random Number Generator C++ Wrapper}
\description{
Wrapper around a C++ random number generator object (for reproducibility).
The class persists a C++ random number generator throughout an R session to
ensure a given seed generates the same outputs (on the same OS). If no seed is provided,
the C++ random number generator is initialized using \code{std::random_device}.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{rng_ptr}}{External pointer to a C++ std::mt19937 class}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CppRNG-new}{\code{CppRNG$new()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CppRNG-new"></a>}}
\if{latex}{\out{\hypertarget{method-CppRNG-new}{}}}
\subsection{Method \code{new()}}{
Create a new CppRNG object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CppRNG$new(random_seed = -1)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{random_seed}}{(Optional) random seed for sampling}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{CppRNG} object.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forest.R
\name{createForestSamples}
\alias{createForestSamples}
\title{Create ForestSamples Object}
\usage{
createForestSamples(
  num_trees,
  leaf_dimension = 1,
  is_leaf_constant = FALSE,
  is_exponentiated = FALSE
)
}
\arguments{
\item{num_trees}{Number of trees}

\item{leaf_dimension}{Dimensionality of the outcome model}

\item{is_leaf_constant}{Whether leaf is constant}

\item{is_exponentiated}{Whether forest predictions should be exponentiated before being returned}
}
\value{
\code{ForestSamples} object
}
\description{
Create a container of forest samples

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
num_trees <- 100
leaf_dimension <- 2
is_leaf_constant <- FALSE
is_exponentiated <- FALSE
forest_samples <- createForestSamples(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forest.R
\name{ForestSamples}
\alias{ForestSamples}
\title{Forest Container C++ Wrapper}
\description{
Wrapper around a C++ class that stores draws from an random ensemble of decision trees.

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{forest_container_ptr}}{External pointer to a C++ ForestContainer class}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ForestSamples-new}{\code{ForestSamples$new()}}
\item \href{#method-ForestSamples-collapse}{\code{ForestSamples$collapse()}}
\item \href{#method-ForestSamples-combine_forests}{\code{ForestSamples$combine_forests()}}
\item \href{#method-ForestSamples-add_to_forest}{\code{ForestSamples$add_to_forest()}}
\item \href{#method-ForestSamples-multiply_forest}{\code{ForestSamples$multiply_forest()}}
\item \href{#method-ForestSamples-load_from_json}{\code{ForestSamples$load_from_json()}}
\item \href{#method-ForestSamples-append_from_json}{\code{ForestSamples$append_from_json()}}
\item \href{#method-ForestSamples-load_from_json_string}{\code{ForestSamples$load_from_json_string()}}
\item \href{#method-ForestSamples-append_from_json_string}{\code{ForestSamples$append_from_json_string()}}
\item \href{#method-ForestSamples-predict}{\code{ForestSamples$predict()}}
\item \href{#method-ForestSamples-predict_raw}{\code{ForestSamples$predict_raw()}}
\item \href{#method-ForestSamples-predict_raw_single_forest}{\code{ForestSamples$predict_raw_single_forest()}}
\item \href{#method-ForestSamples-predict_raw_single_tree}{\code{ForestSamples$predict_raw_single_tree()}}
\item \href{#method-ForestSamples-set_root_leaves}{\code{ForestSamples$set_root_leaves()}}
\item \href{#method-ForestSamples-prepare_for_sampler}{\code{ForestSamples$prepare_for_sampler()}}
\item \href{#method-ForestSamples-adjust_residual}{\code{ForestSamples$adjust_residual()}}
\item \href{#method-ForestSamples-save_json}{\code{ForestSamples$save_json()}}
\item \href{#method-ForestSamples-load_json}{\code{ForestSamples$load_json()}}
\item \href{#method-ForestSamples-num_samples}{\code{ForestSamples$num_samples()}}
\item \href{#method-ForestSamples-num_trees}{\code{ForestSamples$num_trees()}}
\item \href{#method-ForestSamples-leaf_dimension}{\code{ForestSamples$leaf_dimension()}}
\item \href{#method-ForestSamples-is_leaf_constant}{\code{ForestSamples$is_leaf_constant()}}
\item \href{#method-ForestSamples-is_exponentiated}{\code{ForestSamples$is_exponentiated()}}
\item \href{#method-ForestSamples-add_forest_with_constant_leaves}{\code{ForestSamples$add_forest_with_constant_leaves()}}
\item \href{#method-ForestSamples-add_numeric_split_tree}{\code{ForestSamples$add_numeric_split_tree()}}
\item \href{#method-ForestSamples-get_tree_leaves}{\code{ForestSamples$get_tree_leaves()}}
\item \href{#method-ForestSamples-get_tree_split_counts}{\code{ForestSamples$get_tree_split_counts()}}
\item \href{#method-ForestSamples-get_forest_split_counts}{\code{ForestSamples$get_forest_split_counts()}}
\item \href{#method-ForestSamples-get_aggregate_split_counts}{\code{ForestSamples$get_aggregate_split_counts()}}
\item \href{#method-ForestSamples-get_granular_split_counts}{\code{ForestSamples$get_granular_split_counts()}}
\item \href{#method-ForestSamples-ensemble_tree_max_depth}{\code{ForestSamples$ensemble_tree_max_depth()}}
\item \href{#method-ForestSamples-average_ensemble_max_depth}{\code{ForestSamples$average_ensemble_max_depth()}}
\item \href{#method-ForestSamples-average_max_depth}{\code{ForestSamples$average_max_depth()}}
\item \href{#method-ForestSamples-num_forest_leaves}{\code{ForestSamples$num_forest_leaves()}}
\item \href{#method-ForestSamples-sum_leaves_squared}{\code{ForestSamples$sum_leaves_squared()}}
\item \href{#method-ForestSamples-is_leaf_node}{\code{ForestSamples$is_leaf_node()}}
\item \href{#method-ForestSamples-is_numeric_split_node}{\code{ForestSamples$is_numeric_split_node()}}
\item \href{#method-ForestSamples-is_categorical_split_node}{\code{ForestSamples$is_categorical_split_node()}}
\item \href{#method-ForestSamples-parent_node}{\code{ForestSamples$parent_node()}}
\item \href{#method-ForestSamples-left_child_node}{\code{ForestSamples$left_child_node()}}
\item \href{#method-ForestSamples-right_child_node}{\code{ForestSamples$right_child_node()}}
\item \href{#method-ForestSamples-node_depth}{\code{ForestSamples$node_depth()}}
\item \href{#method-ForestSamples-node_split_index}{\code{ForestSamples$node_split_index()}}
\item \href{#method-ForestSamples-node_split_threshold}{\code{ForestSamples$node_split_threshold()}}
\item \href{#method-ForestSamples-node_split_categories}{\code{ForestSamples$node_split_categories()}}
\item \href{#method-ForestSamples-node_leaf_values}{\code{ForestSamples$node_leaf_values()}}
\item \href{#method-ForestSamples-num_nodes}{\code{ForestSamples$num_nodes()}}
\item \href{#method-ForestSamples-num_leaves}{\code{ForestSamples$num_leaves()}}
\item \href{#method-ForestSamples-num_leaf_parents}{\code{ForestSamples$num_leaf_parents()}}
\item \href{#method-ForestSamples-num_split_nodes}{\code{ForestSamples$num_split_nodes()}}
\item \href{#method-ForestSamples-nodes}{\code{ForestSamples$nodes()}}
\item \href{#method-ForestSamples-leaves}{\code{ForestSamples$leaves()}}
\item \href{#method-ForestSamples-delete_sample}{\code{ForestSamples$delete_sample()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-new"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-new}{}}}
\subsection{Method \code{new()}}{
Create a new ForestContainer object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$new(
  num_trees,
  leaf_dimension = 1,
  is_leaf_constant = FALSE,
  is_exponentiated = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_trees}}{Number of trees}

\item{\code{leaf_dimension}}{Dimensionality of the outcome model}

\item{\code{is_leaf_constant}}{Whether leaf is constant}

\item{\code{is_exponentiated}}{Whether forest predictions should be exponentiated before being returned}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{ForestContainer} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-collapse"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-collapse}{}}}
\subsection{Method \code{collapse()}}{
Collapse forests in this container by a pre-specified batch size.
For example, if we have a container of twenty 10-tree forests, and we
specify a \code{batch_size} of 5, then this method will yield four 50-tree
forests. "Excess" forests remaining after the size of a forest container
is divided by \code{batch_size} will be pruned from the beginning of the
container (i.e. earlier sampled forests will be deleted). This method
has no effect if \code{batch_size} is larger than the number of forests
in a container.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$collapse(batch_size)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{batch_size}}{Number of forests to be collapsed into a single forest}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-combine_forests"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-combine_forests}{}}}
\subsection{Method \code{combine_forests()}}{
Merge specified forests into a single forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$combine_forests(forest_inds)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_inds}}{Indices of forests to be combined (0-indexed)}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-add_to_forest"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-add_to_forest}{}}}
\subsection{Method \code{add_to_forest()}}{
Add a constant value to every leaf of every tree of a given forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$add_to_forest(forest_index, constant_value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_index}}{Index of forest whose leaves will be modified (0-indexed)}

\item{\code{constant_value}}{Value to add to every leaf of every tree of the forest at \code{forest_index}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-multiply_forest"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-multiply_forest}{}}}
\subsection{Method \code{multiply_forest()}}{
Multiply every leaf of every tree of a given forest by constant value
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$multiply_forest(forest_index, constant_multiple)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_index}}{Index of forest whose leaves will be modified (0-indexed)}

\item{\code{constant_multiple}}{Value to multiply through by every leaf of every tree of the forest at \code{forest_index}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-load_from_json"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-load_from_json}{}}}
\subsection{Method \code{load_from_json()}}{
Create a new \code{ForestContainer} object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$load_from_json(json_object, json_forest_label)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_object}}{Object of class \code{CppJson}}

\item{\code{json_forest_label}}{Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{ForestContainer} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-append_from_json"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-append_from_json}{}}}
\subsection{Method \code{append_from_json()}}{
Append to a \code{ForestContainer} object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$append_from_json(json_object, json_forest_label)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_object}}{Object of class \code{CppJson}}

\item{\code{json_forest_label}}{Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-load_from_json_string"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-load_from_json_string}{}}}
\subsection{Method \code{load_from_json_string()}}{
Create a new \code{ForestContainer} object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$load_from_json_string(json_string, json_forest_label)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_string}}{JSON string which parses into object of class \code{CppJson}}

\item{\code{json_forest_label}}{Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{ForestContainer} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-append_from_json_string"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-append_from_json_string}{}}}
\subsection{Method \code{append_from_json_string()}}{
Append to a \code{ForestContainer} object from a json object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$append_from_json_string(json_string, json_forest_label)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_string}}{JSON string which parses into object of class \code{CppJson}}

\item{\code{json_forest_label}}{Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-predict"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-predict}{}}}
\subsection{Method \code{predict()}}{
Predict every tree ensemble on every sample in \code{forest_dataset}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$predict(forest_dataset)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{\code{ForestDataset} R class}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
matrix of predictions with as many rows as in forest_dataset
and as many columns as samples in the \code{ForestContainer}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-predict_raw"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-predict_raw}{}}}
\subsection{Method \code{predict_raw()}}{
Predict "raw" leaf values (without being multiplied by basis) for every tree ensemble on every sample in \code{forest_dataset}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$predict_raw(forest_dataset)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{\code{ForestDataset} R class}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Array of predictions for each observation in \code{forest_dataset} and
each sample in the \code{ForestSamples} class with each prediction having the
dimensionality of the forests' leaf model. In the case of a constant leaf model
or univariate leaf regression, this array is two-dimensional (number of observations,
number of forest samples). In the case of a multivariate leaf regression,
this array is three-dimension (number of observations, leaf model dimension,
number of samples).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-predict_raw_single_forest"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-predict_raw_single_forest}{}}}
\subsection{Method \code{predict_raw_single_forest()}}{
Predict "raw" leaf values (without being multiplied by basis) for a specific forest on every sample in \code{forest_dataset}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$predict_raw_single_forest(forest_dataset, forest_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{\code{ForestDataset} R class}

\item{\code{forest_num}}{Index of the forest sample within the container}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
matrix of predictions with as many rows as in forest_dataset
and as many columns as dimensions in the leaves of trees in \code{ForestContainer}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-predict_raw_single_tree"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-predict_raw_single_tree}{}}}
\subsection{Method \code{predict_raw_single_tree()}}{
Predict "raw" leaf values (without being multiplied by basis) for a specific tree in a specific forest on every observation in \code{forest_dataset}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$predict_raw_single_tree(forest_dataset, forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{\code{ForestDataset} R class}

\item{\code{forest_num}}{Index of the forest sample within the container}

\item{\code{tree_num}}{Index of the tree to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
matrix of predictions with as many rows as in \code{forest_dataset}
and as many columns as dimensions in the leaves of trees in \code{ForestContainer}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-set_root_leaves"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-set_root_leaves}{}}}
\subsection{Method \code{set_root_leaves()}}{
Set a constant predicted value for every tree in the ensemble.
Stops program if any tree is more than a root node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$set_root_leaves(forest_num, leaf_value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest sample within the container.}

\item{\code{leaf_value}}{Constant leaf value(s) to be fixed for each tree in the ensemble indexed by \code{forest_num}. Can be either a single number or a vector, depending on the forest's leaf dimension.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-prepare_for_sampler"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-prepare_for_sampler}{}}}
\subsection{Method \code{prepare_for_sampler()}}{
Set a constant predicted value for every tree in the ensemble.
Stops program if any tree is more than a root node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$prepare_for_sampler(
  dataset,
  outcome,
  forest_model,
  leaf_model_int,
  leaf_value
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dataset}}{\code{ForestDataset} Dataset class (covariates, basis, etc...)}

\item{\code{outcome}}{\code{Outcome} Outcome class (residual / partial residual)}

\item{\code{forest_model}}{\code{ForestModel} object storing tracking structures used in training / sampling}

\item{\code{leaf_model_int}}{Integer value encoding the leaf model type (0 = constant gaussian, 1 = univariate gaussian, 2 = multivariate gaussian, 3 = log linear variance).}

\item{\code{leaf_value}}{Constant leaf value(s) to be fixed for each tree in the ensemble indexed by \code{forest_num}. Can be either a single number or a vector, depending on the forest's leaf dimension.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-adjust_residual"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-adjust_residual}{}}}
\subsection{Method \code{adjust_residual()}}{
Adjusts residual based on the predictions of a forest

This is typically run just once at the beginning of a forest sampling algorithm.
After trees are initialized with constant root node predictions, their root predictions are subtracted out of the residual.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$adjust_residual(
  dataset,
  outcome,
  forest_model,
  requires_basis,
  forest_num,
  add
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dataset}}{\code{ForestDataset} object storing the covariates and bases for a given forest}

\item{\code{outcome}}{\code{Outcome} object storing the residuals to be updated based on forest predictions}

\item{\code{forest_model}}{\code{ForestModel} object storing tracking structures used in training / sampling}

\item{\code{requires_basis}}{Whether or not a forest requires a basis for prediction}

\item{\code{forest_num}}{Index of forest used to update residuals}

\item{\code{add}}{Whether forest predictions should be added to or subtracted from residuals}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-save_json"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-save_json}{}}}
\subsection{Method \code{save_json()}}{
Store the trees and metadata of \code{ForestDataset} class in a json file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$save_json(json_filename)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_filename}}{Name of output json file (must end in ".json")}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-load_json"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-load_json}{}}}
\subsection{Method \code{load_json()}}{
Load trees and metadata for an ensemble from a json file. Note that
any trees and metadata already present in \code{ForestDataset} class will
be overwritten.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$load_json(json_filename)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{json_filename}}{Name of model input json file (must end in ".json")}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-num_samples"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-num_samples}{}}}
\subsection{Method \code{num_samples()}}{
Return number of samples in a \code{ForestContainer} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$num_samples()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Sample count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-num_trees"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-num_trees}{}}}
\subsection{Method \code{num_trees()}}{
Return number of trees in each ensemble of a \code{ForestContainer} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$num_trees()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Tree count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-leaf_dimension"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-leaf_dimension}{}}}
\subsection{Method \code{leaf_dimension()}}{
Return output dimension of trees in a \code{ForestContainer} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$leaf_dimension()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Leaf node parameter size
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-is_leaf_constant"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-is_leaf_constant}{}}}
\subsection{Method \code{is_leaf_constant()}}{
Return constant leaf status of trees in a \code{ForestContainer} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$is_leaf_constant()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if leaves are constant, \code{FALSE} otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-is_exponentiated"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-is_exponentiated}{}}}
\subsection{Method \code{is_exponentiated()}}{
Return exponentiation status of trees in a \code{ForestContainer} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$is_exponentiated()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if leaf predictions must be exponentiated, \code{FALSE} otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-add_forest_with_constant_leaves"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-add_forest_with_constant_leaves}{}}}
\subsection{Method \code{add_forest_with_constant_leaves()}}{
Add a new all-root ensemble to the container, with all of the leaves
set to the value / vector provided
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$add_forest_with_constant_leaves(leaf_value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{leaf_value}}{Value (or vector of values) to initialize root nodes in tree}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-add_numeric_split_tree"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-add_numeric_split_tree}{}}}
\subsection{Method \code{add_numeric_split_tree()}}{
Add a numeric (i.e. \code{X[,i] <= c}) split to a given tree in the ensemble
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$add_numeric_split_tree(
  forest_num,
  tree_num,
  leaf_num,
  feature_num,
  split_threshold,
  left_leaf_value,
  right_leaf_value
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest which contains the tree to be split}

\item{\code{tree_num}}{Index of the tree to be split}

\item{\code{leaf_num}}{Leaf to be split}

\item{\code{feature_num}}{Feature that defines the new split}

\item{\code{split_threshold}}{Value that defines the cutoff of the new split}

\item{\code{left_leaf_value}}{Value (or vector of values) to assign to the newly created left node}

\item{\code{right_leaf_value}}{Value (or vector of values) to assign to the newly created right node}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-get_tree_leaves"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-get_tree_leaves}{}}}
\subsection{Method \code{get_tree_leaves()}}{
Retrieve a vector of indices of leaf nodes for a given tree in a given forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$get_tree_leaves(forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest which contains tree \code{tree_num}}

\item{\code{tree_num}}{Index of the tree for which leaf indices will be retrieved}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-get_tree_split_counts"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-get_tree_split_counts}{}}}
\subsection{Method \code{get_tree_split_counts()}}{
Retrieve a vector of split counts for every training set variable in a given tree in a given forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$get_tree_split_counts(forest_num, tree_num, num_features)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest which contains tree \code{tree_num}}

\item{\code{tree_num}}{Index of the tree for which split counts will be retrieved}

\item{\code{num_features}}{Total number of features in the training set}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-get_forest_split_counts"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-get_forest_split_counts}{}}}
\subsection{Method \code{get_forest_split_counts()}}{
Retrieve a vector of split counts for every training set variable in a given forest
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$get_forest_split_counts(forest_num, num_features)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest for which split counts will be retrieved}

\item{\code{num_features}}{Total number of features in the training set}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-get_aggregate_split_counts"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-get_aggregate_split_counts}{}}}
\subsection{Method \code{get_aggregate_split_counts()}}{
Retrieve a vector of split counts for every training set variable in a given forest, aggregated across ensembles and trees
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$get_aggregate_split_counts(num_features)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_features}}{Total number of features in the training set}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-get_granular_split_counts"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-get_granular_split_counts}{}}}
\subsection{Method \code{get_granular_split_counts()}}{
Retrieve a vector of split counts for every training set variable in a given forest, reported separately for each ensemble and tree
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$get_granular_split_counts(num_features)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_features}}{Total number of features in the training set}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-ensemble_tree_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-ensemble_tree_max_depth}{}}}
\subsection{Method \code{ensemble_tree_max_depth()}}{
Maximum depth of a specific tree in a specific ensemble in a \code{ForestSamples} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$ensemble_tree_max_depth(ensemble_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ensemble_num}}{Ensemble number}

\item{\code{tree_num}}{Tree index within ensemble \code{ensemble_num}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Maximum leaf depth
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-average_ensemble_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-average_ensemble_max_depth}{}}}
\subsection{Method \code{average_ensemble_max_depth()}}{
Average the maximum depth of each tree in a given ensemble in a \code{ForestSamples} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$average_ensemble_max_depth(ensemble_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ensemble_num}}{Ensemble number}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Average maximum depth
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-average_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-average_max_depth}{}}}
\subsection{Method \code{average_max_depth()}}{
Average the maximum depth of each tree in each ensemble in a \code{ForestContainer} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$average_max_depth()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Average maximum depth
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-num_forest_leaves"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-num_forest_leaves}{}}}
\subsection{Method \code{num_forest_leaves()}}{
Number of leaves in a given ensemble in a \code{ForestSamples} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$num_forest_leaves(forest_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the ensemble to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Count of leaves in the ensemble stored at \code{forest_num}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-sum_leaves_squared"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-sum_leaves_squared}{}}}
\subsection{Method \code{sum_leaves_squared()}}{
Sum of squared (raw) leaf values in a given ensemble in a \code{ForestSamples} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$sum_leaves_squared(forest_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the ensemble to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Average maximum depth
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-is_leaf_node"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-is_leaf_node}{}}}
\subsection{Method \code{is_leaf_node()}}{
Whether or not a given node of a given tree in a given forest in the \code{ForestSamples} is a leaf
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$is_leaf_node(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{TRUE} if node is a leaf, \code{FALSE} otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-is_numeric_split_node"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-is_numeric_split_node}{}}}
\subsection{Method \code{is_numeric_split_node()}}{
Whether or not a given node of a given tree in a given forest in the \code{ForestSamples} is a numeric split node
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$is_numeric_split_node(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{TRUE} if node is a numeric split node, \code{FALSE} otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-is_categorical_split_node"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-is_categorical_split_node}{}}}
\subsection{Method \code{is_categorical_split_node()}}{
Whether or not a given node of a given tree in a given forest in the \code{ForestSamples} is a categorical split node
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$is_categorical_split_node(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{TRUE} if node is a categorical split node, \code{FALSE} otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-parent_node"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-parent_node}{}}}
\subsection{Method \code{parent_node()}}{
Parent node of given node of a given tree in a given forest in a \code{ForestSamples} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$parent_node(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer ID of the parent node
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-left_child_node"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-left_child_node}{}}}
\subsection{Method \code{left_child_node()}}{
Left child node of given node of a given tree in a given forest in a \code{ForestSamples} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$left_child_node(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer ID of the left child node
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-right_child_node"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-right_child_node}{}}}
\subsection{Method \code{right_child_node()}}{
Right child node of given node of a given tree in a given forest in a \code{ForestSamples} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$right_child_node(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer ID of the right child node
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-node_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-node_depth}{}}}
\subsection{Method \code{node_depth()}}{
Depth of given node of a given tree in a given forest in a \code{ForestSamples} object, with 0 depth for the root node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$node_depth(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer valued depth of the node
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-node_split_index"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-node_split_index}{}}}
\subsection{Method \code{node_split_index()}}{
Split index of given node of a given tree in a given forest in a \code{ForestSamples} object. Returns \code{-1} is node is a leaf.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$node_split_index(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer valued depth of the node
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-node_split_threshold"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-node_split_threshold}{}}}
\subsection{Method \code{node_split_threshold()}}{
Threshold that defines a numeric split for a given node of a given tree in a given forest in a \code{ForestSamples} object.
Returns \code{Inf} if the node is a leaf or a categorical split node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$node_split_threshold(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Threshold defining a split for the node
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-node_split_categories"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-node_split_categories}{}}}
\subsection{Method \code{node_split_categories()}}{
Array of category indices that define a categorical split for a given node of a given tree in a given forest in a \code{ForestSamples} object.
Returns \code{c(Inf)} if the node is a leaf or a numeric split node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$node_split_categories(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Categories defining a split for the node
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-node_leaf_values"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-node_leaf_values}{}}}
\subsection{Method \code{node_leaf_values()}}{
Leaf node value(s) for a given node of a given tree in a given forest in a \code{ForestSamples} object.
Values are stale if the node is a split node.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$node_leaf_values(forest_num, tree_num, node_id)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}

\item{\code{node_id}}{Index of the node to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Vector (often univariate) of leaf values
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-num_nodes"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-num_nodes}{}}}
\subsection{Method \code{num_nodes()}}{
Number of nodes in a given tree in a given forest in a \code{ForestSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$num_nodes(forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Count of total tree nodes
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-num_leaves"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-num_leaves}{}}}
\subsection{Method \code{num_leaves()}}{
Number of leaves in a given tree in a given forest in a \code{ForestSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$num_leaves(forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Count of total tree leaves
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-num_leaf_parents"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-num_leaf_parents}{}}}
\subsection{Method \code{num_leaf_parents()}}{
Number of leaf parents (split nodes with two leaves as children) in a given tree in a given forest in a \code{ForestSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$num_leaf_parents(forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Count of total tree leaf parents
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-num_split_nodes"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-num_split_nodes}{}}}
\subsection{Method \code{num_split_nodes()}}{
Number of split nodes in a given tree in a given forest in a \code{ForestSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$num_split_nodes(forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Count of total tree split nodes
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-nodes"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-nodes}{}}}
\subsection{Method \code{nodes()}}{
Array of node indices in a given tree in a given forest in a \code{ForestSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$nodes(forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Indices of tree nodes
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-leaves"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-leaves}{}}}
\subsection{Method \code{leaves()}}{
Array of leaf indices in a given tree in a given forest in a \code{ForestSamples} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$leaves(forest_num, tree_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be queried}

\item{\code{tree_num}}{Index of the tree to be queried}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Indices of leaf nodes
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestSamples-delete_sample"></a>}}
\if{latex}{\out{\hypertarget{method-ForestSamples-delete_sample}{}}}
\subsection{Method \code{delete_sample()}}{
Modify the \code{ForestSamples} object by removing the forest sample indexed by `forest_num
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestSamples$delete_sample(forest_num)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_num}}{Index of the forest to be removed}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{ForestModelConfig}
\alias{ForestModelConfig}
\title{Forest Model Configuration Object}
\value{
Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)

Vector of (0-indexed) indices of trees to update in a sweep

Vector specifying sampling probability for all p covariates in ForestDataset

Number of trees in a forest

Number of features in a forest model training set

Number of observations in a forest model training set

Root node split probability in tree prior

Depth prior penalty in tree prior

Minimum number of samples in a tree leaf

Maximum depth of any tree in the ensemble in the model

Integer coded leaf model type

Scale parameter used in Gaussian leaf models

Shape parameter for IG leaf models

Scale parameter for IG leaf models

Shape parameter for conditional gamma component of cloglog leaf models

Rate parameter for conditional gamma component of cloglog leaf models

Number of unique cutpoints to consider

Number of features to subsample for the GFR algorithm
}
\description{
Object used to get / set parameters and other model configuration options
for a forest model in the "low-level" stochtree interface.
The "low-level" stochtree interface enables a high degreee of sampler
customization, in which users employ R wrappers around C++ objects
like ForestDataset, Outcome, CppRng, and ForestModel to run the
Gibbs sampler of a BART model with custom modifications.
ForestModelConfig allows users to specify / query the parameters of a
forest model they wish to run.

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{feature_types}}{Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)}

\item{\code{sweep_update_indices}}{Vector of trees to update in a sweep}

\item{\code{num_trees}}{Number of trees in the forest being sampled}

\item{\code{num_features}}{Number of features in training dataset}

\item{\code{num_observations}}{Number of observations in training dataset}

\item{\code{leaf_dimension}}{Dimension of the leaf model}

\item{\code{alpha}}{Root node split probability in tree prior}

\item{\code{beta}}{Depth prior penalty in tree prior}

\item{\code{min_samples_leaf}}{Minimum number of samples in a tree leaf}

\item{\code{max_depth}}{Maximum depth of any tree in the ensemble in the model. Setting to \code{-1} does not enforce any depth limits on trees.}

\item{\code{leaf_model_type}}{Integer specifying the leaf model type (0 = constant leaf, 1 = univariate leaf regression, 2 = multivariate leaf regression)}

\item{\code{leaf_model_scale}}{Scale parameter used in Gaussian leaf models}

\item{\code{variable_weights}}{Vector specifying sampling probability for all p covariates in ForestDataset}

\item{\code{variance_forest_shape}}{Shape parameter for IG leaf models (applicable when \code{leaf_model_type = 3})}

\item{\code{variance_forest_scale}}{Scale parameter for IG leaf models (applicable when \code{leaf_model_type = 3})}

\item{\code{cloglog_forest_shape}}{Shape parameter for conditional gamma component of cloglog leaf models (applicable when \code{leaf_model_type = 4})}

\item{\code{cloglog_forest_rate}}{Rate parameter for conditional gamma component of cloglog leaf models (applicable when \code{leaf_model_type = 4})}

\item{\code{cutpoint_grid_size}}{Number of unique cutpoints to consider}

\item{\code{num_features_subsample}}{Number of features to subsample for the GFR algorithm
Create a new ForestModelConfig object.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ForestModelConfig-new}{\code{ForestModelConfig$new()}}
\item \href{#method-ForestModelConfig-update_feature_types}{\code{ForestModelConfig$update_feature_types()}}
\item \href{#method-ForestModelConfig-update_sweep_indices}{\code{ForestModelConfig$update_sweep_indices()}}
\item \href{#method-ForestModelConfig-update_variable_weights}{\code{ForestModelConfig$update_variable_weights()}}
\item \href{#method-ForestModelConfig-update_alpha}{\code{ForestModelConfig$update_alpha()}}
\item \href{#method-ForestModelConfig-update_beta}{\code{ForestModelConfig$update_beta()}}
\item \href{#method-ForestModelConfig-update_min_samples_leaf}{\code{ForestModelConfig$update_min_samples_leaf()}}
\item \href{#method-ForestModelConfig-update_max_depth}{\code{ForestModelConfig$update_max_depth()}}
\item \href{#method-ForestModelConfig-update_leaf_model_scale}{\code{ForestModelConfig$update_leaf_model_scale()}}
\item \href{#method-ForestModelConfig-update_variance_forest_shape}{\code{ForestModelConfig$update_variance_forest_shape()}}
\item \href{#method-ForestModelConfig-update_variance_forest_scale}{\code{ForestModelConfig$update_variance_forest_scale()}}
\item \href{#method-ForestModelConfig-update_cloglog_forest_shape}{\code{ForestModelConfig$update_cloglog_forest_shape()}}
\item \href{#method-ForestModelConfig-update_cloglog_forest_rate}{\code{ForestModelConfig$update_cloglog_forest_rate()}}
\item \href{#method-ForestModelConfig-update_cutpoint_grid_size}{\code{ForestModelConfig$update_cutpoint_grid_size()}}
\item \href{#method-ForestModelConfig-update_num_features_subsample}{\code{ForestModelConfig$update_num_features_subsample()}}
\item \href{#method-ForestModelConfig-get_feature_types}{\code{ForestModelConfig$get_feature_types()}}
\item \href{#method-ForestModelConfig-get_sweep_indices}{\code{ForestModelConfig$get_sweep_indices()}}
\item \href{#method-ForestModelConfig-get_variable_weights}{\code{ForestModelConfig$get_variable_weights()}}
\item \href{#method-ForestModelConfig-get_num_trees}{\code{ForestModelConfig$get_num_trees()}}
\item \href{#method-ForestModelConfig-get_num_features}{\code{ForestModelConfig$get_num_features()}}
\item \href{#method-ForestModelConfig-get_num_observations}{\code{ForestModelConfig$get_num_observations()}}
\item \href{#method-ForestModelConfig-get_alpha}{\code{ForestModelConfig$get_alpha()}}
\item \href{#method-ForestModelConfig-get_beta}{\code{ForestModelConfig$get_beta()}}
\item \href{#method-ForestModelConfig-get_min_samples_leaf}{\code{ForestModelConfig$get_min_samples_leaf()}}
\item \href{#method-ForestModelConfig-get_max_depth}{\code{ForestModelConfig$get_max_depth()}}
\item \href{#method-ForestModelConfig-get_leaf_model_type}{\code{ForestModelConfig$get_leaf_model_type()}}
\item \href{#method-ForestModelConfig-get_leaf_model_scale}{\code{ForestModelConfig$get_leaf_model_scale()}}
\item \href{#method-ForestModelConfig-get_variance_forest_shape}{\code{ForestModelConfig$get_variance_forest_shape()}}
\item \href{#method-ForestModelConfig-get_variance_forest_scale}{\code{ForestModelConfig$get_variance_forest_scale()}}
\item \href{#method-ForestModelConfig-get_cloglog_forest_shape}{\code{ForestModelConfig$get_cloglog_forest_shape()}}
\item \href{#method-ForestModelConfig-get_cloglog_forest_rate}{\code{ForestModelConfig$get_cloglog_forest_rate()}}
\item \href{#method-ForestModelConfig-get_cutpoint_grid_size}{\code{ForestModelConfig$get_cutpoint_grid_size()}}
\item \href{#method-ForestModelConfig-get_num_features_subsample}{\code{ForestModelConfig$get_num_features_subsample()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-new"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$new(
  feature_types = NULL,
  sweep_update_indices = NULL,
  num_trees = NULL,
  num_features = NULL,
  num_observations = NULL,
  variable_weights = NULL,
  leaf_dimension = 1,
  alpha = 0.95,
  beta = 2,
  min_samples_leaf = 5,
  max_depth = -1,
  leaf_model_type = 1,
  leaf_model_scale = NULL,
  variance_forest_shape = 1,
  variance_forest_scale = 1,
  cloglog_forest_shape = 2,
  cloglog_forest_rate = 2,
  cutpoint_grid_size = 100,
  num_features_subsample = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{feature_types}}{Vector of integer-coded feature types (where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)}

\item{\code{sweep_update_indices}}{Vector of (0-indexed) indices of trees to update in a sweep}

\item{\code{num_trees}}{Number of trees in the forest being sampled}

\item{\code{num_features}}{Number of features in training dataset}

\item{\code{num_observations}}{Number of observations in training dataset}

\item{\code{variable_weights}}{Vector specifying sampling probability for all p covariates in ForestDataset}

\item{\code{leaf_dimension}}{Dimension of the leaf model (default: \code{1})}

\item{\code{alpha}}{Root node split probability in tree prior (default: \code{0.95})}

\item{\code{beta}}{Depth prior penalty in tree prior (default: \code{2.0})}

\item{\code{min_samples_leaf}}{Minimum number of samples in a tree leaf (default: \code{5})}

\item{\code{max_depth}}{Maximum depth of any tree in the ensemble in the model. Setting to \code{-1} does not enforce any depth limits on trees. Default: \code{-1}.}

\item{\code{leaf_model_type}}{Integer specifying the leaf model type (0 = constant leaf, 1 = univariate leaf regression, 2 = multivariate leaf regression). Default: \code{0}.}

\item{\code{leaf_model_scale}}{Scale parameter used in Gaussian leaf models (can either be a scalar or a q x q matrix, where q is the dimensionality of the basis and is only >1 when \code{leaf_model_int = 2}). Calibrated internally as \code{1/num_trees}, propagated along diagonal if needed for multivariate leaf models.}

\item{\code{variance_forest_shape}}{Shape parameter for IG leaf models (applicable when \code{leaf_model_type = 3}). Default: \code{1}.}

\item{\code{variance_forest_scale}}{Scale parameter for IG leaf models (applicable when \code{leaf_model_type = 3}). Default: \code{1}.}

\item{\code{cloglog_forest_shape}}{Shape parameter for conditional gamma component of cloglog leaf models (applicable when \code{leaf_model_type = 4}). Default: \code{1}.}

\item{\code{cloglog_forest_rate}}{Rate parameter for conditional gamma component of cloglog leaf models (applicable when \code{leaf_model_type = 4}). Default: \code{1}.}

\item{\code{cutpoint_grid_size}}{Number of unique cutpoints to consider (default: \code{100})}

\item{\code{num_features_subsample}}{Number of features to subsample for the GFR algorithm}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new ForestModelConfig object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_feature_types"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_feature_types}{}}}
\subsection{Method \code{update_feature_types()}}{
Update feature types
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_feature_types(feature_types)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{feature_types}}{Vector of integer-coded feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_sweep_indices"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_sweep_indices}{}}}
\subsection{Method \code{update_sweep_indices()}}{
Update sweep update indices
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_sweep_indices(sweep_update_indices)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{sweep_update_indices}}{Vector of (0-indexed) indices of trees to update in a sweep}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_variable_weights"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_variable_weights}{}}}
\subsection{Method \code{update_variable_weights()}}{
Update variable weights
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_variable_weights(variable_weights)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{variable_weights}}{Vector specifying sampling probability for all p covariates in ForestDataset}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_alpha"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_alpha}{}}}
\subsection{Method \code{update_alpha()}}{
Update root node split probability in tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_alpha(alpha)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{alpha}}{Root node split probability in tree prior}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_beta"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_beta}{}}}
\subsection{Method \code{update_beta()}}{
Update depth prior penalty in tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_beta(beta)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{beta}}{Depth prior penalty in tree prior}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_min_samples_leaf"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_min_samples_leaf}{}}}
\subsection{Method \code{update_min_samples_leaf()}}{
Update minimum number of samples per leaf node in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_min_samples_leaf(min_samples_leaf)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{min_samples_leaf}}{Minimum number of samples in a tree leaf}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_max_depth}{}}}
\subsection{Method \code{update_max_depth()}}{
Update max depth in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_max_depth(max_depth)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{max_depth}}{Maximum depth of any tree in the ensemble in the model}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_leaf_model_scale"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_leaf_model_scale}{}}}
\subsection{Method \code{update_leaf_model_scale()}}{
Update scale parameter used in Gaussian leaf models
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_leaf_model_scale(leaf_model_scale)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{leaf_model_scale}}{Scale parameter used in Gaussian leaf models}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_variance_forest_shape"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_variance_forest_shape}{}}}
\subsection{Method \code{update_variance_forest_shape()}}{
Update shape parameter for IG leaf models
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_variance_forest_shape(variance_forest_shape)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{variance_forest_shape}}{Shape parameter for IG leaf models}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_variance_forest_scale"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_variance_forest_scale}{}}}
\subsection{Method \code{update_variance_forest_scale()}}{
Update scale parameter for IG leaf models
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_variance_forest_scale(variance_forest_scale)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{variance_forest_scale}}{Scale parameter for IG leaf models}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_cloglog_forest_shape"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_cloglog_forest_shape}{}}}
\subsection{Method \code{update_cloglog_forest_shape()}}{
Update shape parameter for conditional gamma component of cloglog leaf models
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_cloglog_forest_shape(cloglog_forest_shape)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cloglog_forest_shape}}{Shape parameter for conditional gamma component of cloglog leaf models}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_cloglog_forest_rate"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_cloglog_forest_rate}{}}}
\subsection{Method \code{update_cloglog_forest_rate()}}{
Update rate parameter for conditional gamma component of cloglog leaf models
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_cloglog_forest_rate(cloglog_forest_rate)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cloglog_forest_rate}}{Rate parameter for conditional gamma component of cloglog leaf models}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_cutpoint_grid_size"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_cutpoint_grid_size}{}}}
\subsection{Method \code{update_cutpoint_grid_size()}}{
Update number of unique cutpoints to consider
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_cutpoint_grid_size(cutpoint_grid_size)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cutpoint_grid_size}}{Number of unique cutpoints to consider}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-update_num_features_subsample"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-update_num_features_subsample}{}}}
\subsection{Method \code{update_num_features_subsample()}}{
Update number of features to subsample for the GFR algorithm
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$update_num_features_subsample(num_features_subsample)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_features_subsample}}{Number of features to subsample for the GFR algorithm}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_feature_types"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_feature_types}{}}}
\subsection{Method \code{get_feature_types()}}{
Query feature types for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_feature_types()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_sweep_indices"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_sweep_indices}{}}}
\subsection{Method \code{get_sweep_indices()}}{
Query sweep update indices for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_sweep_indices()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_variable_weights"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_variable_weights}{}}}
\subsection{Method \code{get_variable_weights()}}{
Query variable weights for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_variable_weights()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_num_trees"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_num_trees}{}}}
\subsection{Method \code{get_num_trees()}}{
Query number of trees
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_num_trees()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_num_features"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_num_features}{}}}
\subsection{Method \code{get_num_features()}}{
Query number of features
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_num_features()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_num_observations"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_num_observations}{}}}
\subsection{Method \code{get_num_observations()}}{
Query number of observations
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_num_observations()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_alpha"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_alpha}{}}}
\subsection{Method \code{get_alpha()}}{
Query root node split probability in tree prior for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_alpha()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_beta"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_beta}{}}}
\subsection{Method \code{get_beta()}}{
Query depth prior penalty in tree prior for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_beta()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_min_samples_leaf"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_min_samples_leaf}{}}}
\subsection{Method \code{get_min_samples_leaf()}}{
Query root node split probability in tree prior for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_min_samples_leaf()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_max_depth}{}}}
\subsection{Method \code{get_max_depth()}}{
Query root node split probability in tree prior for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_max_depth()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_leaf_model_type"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_leaf_model_type}{}}}
\subsection{Method \code{get_leaf_model_type()}}{
Query (integer-coded) type of leaf model
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_leaf_model_type()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_leaf_model_scale"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_leaf_model_scale}{}}}
\subsection{Method \code{get_leaf_model_scale()}}{
Query scale parameter used in Gaussian leaf models for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_leaf_model_scale()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_variance_forest_shape"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_variance_forest_shape}{}}}
\subsection{Method \code{get_variance_forest_shape()}}{
Query shape parameter for IG leaf models for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_variance_forest_shape()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_variance_forest_scale"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_variance_forest_scale}{}}}
\subsection{Method \code{get_variance_forest_scale()}}{
Query scale parameter for IG leaf models for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_variance_forest_scale()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_cloglog_forest_shape"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_cloglog_forest_shape}{}}}
\subsection{Method \code{get_cloglog_forest_shape()}}{
Query shape parameter for conditional gamma component of cloglog leaf models for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_cloglog_forest_shape()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_cloglog_forest_rate"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_cloglog_forest_rate}{}}}
\subsection{Method \code{get_cloglog_forest_rate()}}{
Query rate parameter for conditional gamma component of cloglog leaf models for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_cloglog_forest_rate()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_cutpoint_grid_size"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_cutpoint_grid_size}{}}}
\subsection{Method \code{get_cutpoint_grid_size()}}{
Query number of unique cutpoints to consider for this ForestModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_cutpoint_grid_size()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModelConfig-get_num_features_subsample"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModelConfig-get_num_features_subsample}{}}}
\subsection{Method \code{get_num_features_subsample()}}{
Query number of features to subsample for the GFR algorithm
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModelConfig$get_num_features_subsample()}\if{html}{\out{</div>}}
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_effects.R
\name{RandomEffectsTracker}
\alias{RandomEffectsTracker}
\title{Random Effects Tracker C++ Wrapper}
\description{
Class that defines a "tracker" for random effects models, most notably
storing the data indices available in each group for quicker posterior
computation and sampling of random effects terms.
The class stores a mapping from every observation to its group index, a mapping
from group indices to the training sample observations available in that
group, and predictions for each observation.

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{rfx_tracker_ptr}}{External pointer to a C++ StochTree::RandomEffectsTracker class}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-RandomEffectsTracker-new}{\code{RandomEffectsTracker$new()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsTracker-new"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsTracker-new}{}}}
\subsection{Method \code{new()}}{
Create a new RandomEffectsTracker object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsTracker$new(rfx_group_indices)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{rfx_group_indices}}{Integer indices indicating groups used to define random effects}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{RandomEffectsTracker} object.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forest.R
\name{ForestStateManagement}
\alias{ForestStateManagement}
\alias{resetActiveForest}
\alias{resetForestModel}
\title{Forest State Management Routines}
\usage{
resetActiveForest(active_forest, forest_samples = NULL, forest_num = NULL)

resetForestModel(forest_model, forest, dataset, residual, is_mean_model)
}
\arguments{
\item{active_forest}{Current active forest}

\item{forest_samples}{(Optional) Container of forest samples from which to re-initialize active forest. If not provided, active forest will be reset to an ensemble of single-node (i.e. root) trees.}

\item{forest_num}{(Optional) Index of forest samples from which to initialize active forest. If not provided, active forest will be reset to an ensemble of single-node (i.e. root) trees.}

\item{forest_model}{Forest model with tracking data structures}

\item{forest}{Forest from which to re-initialize forest model}

\item{dataset}{Training dataset object}

\item{residual}{Residual which will also be updated}

\item{is_mean_model}{Whether the model being updated is a conditional mean model}
}
\value{
Both functions have no return type and operate in-place on the relevant \link{Forest} or \link{ForestModel} objects
}
\description{
A forest sampler features two types of state: ephemeral and persistent.
Persistent state includes objects like \link{ForestSamples} and \link{RandomEffectSamples}
which constitute part of the final sampled model.
Ephemeral state supports the sampling computations, but is not retained after the sampler finishes.

The two primary forest-based bits of ephemeral state are the \link{Forest} and \link{ForestModel} classes,
which represent the current state of a forest and its corresponding tracking data structures.

In a linear sampling loop, this ephemeral state is updated with each iteration of the sampler
and any retained forests are copied to a \link{ForestSamples} object. However, in multi-chain settings,
the state of a forest must typically be "reset" at the beginning of a new chain. These function enable
this process by synchronizing the state of a \link{Forest} and \link{ForestModel} with a corresponding element
of a \link{ForestSamples} object, or by resetting both to their default (root) state.

\code{resetActiveForest} resets a \link{Forest} object, either from a specific forest in a \link{ForestSamples}
object or to an ensemble of single-node (i.e. root) trees.
\code{resetForestModel} re-initializes a forest model (tracking data structures) from a specific forest in a \link{ForestSamples} object.

These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
n <- 100
p <- 10
num_trees <- 100
leaf_dimension <- 1
is_leaf_constant <- TRUE
is_exponentiated <- FALSE
alpha <- 0.95
beta <- 2.0
min_samples_leaf <- 2
max_depth <- 10
feature_types <- as.integer(rep(0, p))
leaf_model <- 0
sigma2 <- 1.0
leaf_scale <- as.matrix(1.0)
variable_weights <- rep(1/p, p)
a_forest <- 1
b_forest <- 1
cutpoint_grid_size <- 100
X <- matrix(runif(n*p), ncol = p)
forest_dataset <- createForestDataset(X)
y <- -5 + 10*(X[,1] > 0.5) + rnorm(n)
outcome <- createOutcome(y)
rng <- createCppRNG(1234)
global_model_config <- createGlobalModelConfig(global_error_variance=sigma2)
forest_model_config <- createForestModelConfig(feature_types=feature_types,
                                               num_trees=num_trees, num_observations=n,
                                               num_features=p, alpha=alpha, beta=beta,
                                               min_samples_leaf=min_samples_leaf,
                                               max_depth=max_depth,
                                               variable_weights=variable_weights,
                                               cutpoint_grid_size=cutpoint_grid_size,
                                               leaf_model_type=leaf_model,
                                               leaf_model_scale=leaf_scale)
forest_model <- createForestModel(forest_dataset, forest_model_config, global_model_config)
active_forest <- createForest(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
forest_samples <- createForestSamples(num_trees, leaf_dimension,
                                      is_leaf_constant, is_exponentiated)
active_forest$prepare_for_sampler(forest_dataset, outcome, forest_model, 0, 0.)
forest_model$sample_one_iteration(
    forest_dataset, outcome, forest_samples, active_forest,
    rng, forest_model_config, global_model_config,
    keep_forest = TRUE, gfr = FALSE
)
resetActiveForest(active_forest, forest_samples, 0)
resetForestModel(forest_model, active_forest, forest_dataset, outcome, TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_effects.R
\name{createRandomEffectSamples}
\alias{createRandomEffectSamples}
\title{Create RandomEffectSamples Object}
\usage{
createRandomEffectSamples(num_components, num_groups, random_effects_tracker)
}
\arguments{
\item{num_components}{Number of "components" or bases defining the random effects regression}

\item{num_groups}{Number of random effects groups}

\item{random_effects_tracker}{Object of type \code{RandomEffectsTracker}}
}
\value{
\code{RandomEffectSamples} object
}
\description{
Create a \code{RandomEffectSamples} object

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
n <- 100
rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
rfx_basis <- matrix(rep(1.0, n), ncol=1)
num_groups <- length(unique(rfx_group_ids))
num_components <- ncol(rfx_basis)
rfx_tracker <- createRandomEffectsTracker(rfx_group_ids)
rfx_samples <- createRandomEffectSamples(num_components, num_groups, rfx_tracker)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{ForestModel}
\alias{ForestModel}
\title{Forest Model C++ Wrapper}
\description{
Wraps the C++ data structures needed to sample an ensemble of decision
trees and exposes functionality to run a forest sampler
(using either MCMC or the grow-from-root algorithm).

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{tracker_ptr}}{External pointer to a C++ ForestTracker class}

\item{\code{tree_prior_ptr}}{External pointer to a C++ TreePrior class}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ForestModel-new}{\code{ForestModel$new()}}
\item \href{#method-ForestModel-sample_one_iteration}{\code{ForestModel$sample_one_iteration()}}
\item \href{#method-ForestModel-get_cached_forest_predictions}{\code{ForestModel$get_cached_forest_predictions()}}
\item \href{#method-ForestModel-propagate_basis_update}{\code{ForestModel$propagate_basis_update()}}
\item \href{#method-ForestModel-propagate_residual_update}{\code{ForestModel$propagate_residual_update()}}
\item \href{#method-ForestModel-update_alpha}{\code{ForestModel$update_alpha()}}
\item \href{#method-ForestModel-update_beta}{\code{ForestModel$update_beta()}}
\item \href{#method-ForestModel-update_min_samples_leaf}{\code{ForestModel$update_min_samples_leaf()}}
\item \href{#method-ForestModel-update_max_depth}{\code{ForestModel$update_max_depth()}}
\item \href{#method-ForestModel-get_alpha}{\code{ForestModel$get_alpha()}}
\item \href{#method-ForestModel-get_beta}{\code{ForestModel$get_beta()}}
\item \href{#method-ForestModel-get_min_samples_leaf}{\code{ForestModel$get_min_samples_leaf()}}
\item \href{#method-ForestModel-get_max_depth}{\code{ForestModel$get_max_depth()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-new"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-new}{}}}
\subsection{Method \code{new()}}{
Create a new ForestModel object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$new(
  forest_dataset,
  feature_types,
  num_trees,
  n,
  alpha,
  beta,
  min_samples_leaf,
  max_depth = -1
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{\code{ForestDataset} object, used to initialize forest sampling data structures}

\item{\code{feature_types}}{Feature types (integers where 0 = numeric, 1 = ordered categorical, 2 = unordered categorical)}

\item{\code{num_trees}}{Number of trees in the forest being sampled}

\item{\code{n}}{Number of observations in \code{forest_dataset}}

\item{\code{alpha}}{Root node split probability in tree prior}

\item{\code{beta}}{Depth prior penalty in tree prior}

\item{\code{min_samples_leaf}}{Minimum number of samples in a tree leaf}

\item{\code{max_depth}}{Maximum depth that any tree can reach}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{ForestModel} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-sample_one_iteration"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-sample_one_iteration}{}}}
\subsection{Method \code{sample_one_iteration()}}{
Run a single iteration of the forest sampling algorithm (MCMC or GFR)
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$sample_one_iteration(
  forest_dataset,
  residual,
  forest_samples,
  active_forest,
  rng,
  forest_model_config,
  global_model_config,
  num_threads = -1,
  keep_forest = TRUE,
  gfr = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{forest_dataset}}{Dataset used to sample the forest}

\item{\code{residual}}{Outcome used to sample the forest}

\item{\code{forest_samples}}{Container of forest samples}

\item{\code{active_forest}}{"Active" forest updated by the sampler in each iteration}

\item{\code{rng}}{Wrapper around C++ random number generator}

\item{\code{forest_model_config}}{ForestModelConfig object containing forest model parameters and settings}

\item{\code{global_model_config}}{GlobalModelConfig object containing global model parameters and settings}

\item{\code{num_threads}}{Number of threads to use in the GFR and MCMC algorithms, as well as prediction. If OpenMP is not available on a user's system, this will default to \code{1}, otherwise to the maximum number of available threads.}

\item{\code{keep_forest}}{(Optional) Whether the updated forest sample should be saved to \code{forest_samples}. Default: \code{TRUE}.}

\item{\code{gfr}}{(Optional) Whether or not the forest should be sampled using the "grow-from-root" (GFR) algorithm. Default: \code{TRUE}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-get_cached_forest_predictions"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-get_cached_forest_predictions}{}}}
\subsection{Method \code{get_cached_forest_predictions()}}{
Extract an internally-cached prediction of a forest on the training dataset in a sampler.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$get_cached_forest_predictions()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Vector with as many elements as observations in the training dataset
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-propagate_basis_update"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-propagate_basis_update}{}}}
\subsection{Method \code{propagate_basis_update()}}{
Propagates basis update through to the (full/partial) residual by iteratively
(a) adding back in the previous prediction of each tree, (b) recomputing predictions
for each tree (caching on the C++ side), (c) subtracting the new predictions from the residual.

This is useful in cases where a basis (for e.g. leaf regression) is updated outside
of a tree sampler (as with e.g. adaptive coding for binary treatment BCF).
Once a basis has been updated, the overall "function" represented by a tree model has
changed and this should be reflected through to the residual before the next sampling loop is run.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$propagate_basis_update(dataset, outcome, active_forest)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dataset}}{\code{ForestDataset} object storing the covariates and bases for a given forest}

\item{\code{outcome}}{\code{Outcome} object storing the residuals to be updated based on forest predictions}

\item{\code{active_forest}}{"Active" forest updated by the sampler in each iteration}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-propagate_residual_update"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-propagate_residual_update}{}}}
\subsection{Method \code{propagate_residual_update()}}{
Update the current state of the outcome (i.e. partial residual) data by subtracting the current predictions of each tree.
This function is run after the \code{Outcome} class's \code{update_data} method, which overwrites the partial residual with an entirely new stream of outcome data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$propagate_residual_update(residual)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{residual}}{Outcome used to sample the forest}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-update_alpha"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-update_alpha}{}}}
\subsection{Method \code{update_alpha()}}{
Update alpha in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$update_alpha(alpha)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{alpha}}{New value of alpha to be used}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-update_beta"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-update_beta}{}}}
\subsection{Method \code{update_beta()}}{
Update beta in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$update_beta(beta)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{beta}}{New value of beta to be used}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-update_min_samples_leaf"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-update_min_samples_leaf}{}}}
\subsection{Method \code{update_min_samples_leaf()}}{
Update min_samples_leaf in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$update_min_samples_leaf(min_samples_leaf)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{min_samples_leaf}}{New value of min_samples_leaf to be used}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-update_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-update_max_depth}{}}}
\subsection{Method \code{update_max_depth()}}{
Update max_depth in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$update_max_depth(max_depth)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{max_depth}}{New value of max_depth to be used}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-get_alpha"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-get_alpha}{}}}
\subsection{Method \code{get_alpha()}}{
Update alpha in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$get_alpha()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Value of alpha in the tree prior
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-get_beta"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-get_beta}{}}}
\subsection{Method \code{get_beta()}}{
Update beta in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$get_beta()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Value of beta in the tree prior
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-get_min_samples_leaf"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-get_min_samples_leaf}{}}}
\subsection{Method \code{get_min_samples_leaf()}}{
Query min_samples_leaf in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$get_min_samples_leaf()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Value of min_samples_leaf in the tree prior
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestModel-get_max_depth"></a>}}
\if{latex}{\out{\hypertarget{method-ForestModel-get_max_depth}{}}}
\subsection{Method \code{get_max_depth()}}{
Query max_depth in the tree prior
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestModel$get_max_depth()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Value of max_depth in the tree prior
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{extractParameter}
\alias{extractParameter}
\title{Extract Parameter Samples Generic Function}
\usage{
extractParameter(object, term)
}
\arguments{
\item{object}{Fitted model object from which to extract parameter samples}

\item{term}{Name of the parameter to extract (e.g., \code{"sigma2"}, \code{"y_hat_train"}, etc.)}
}
\value{
Parameter sample array
}
\description{
Generic function for extracting parameter samples from a model object (BCF, BART, etc...)
}
\examples{
n <- 100
p <- 10
X <- matrix(runif(n*p), ncol = p)
rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
rfx_basis <- rep(1.0, n)
y <- (-5 + 10*(X[,1] > 0.5)) + (-2*(rfx_group_ids==1)+2*(rfx_group_ids==2)) + rnorm(n)
bart_model <- bart(X_train=X, y_train=y, rfx_group_ids_train=rfx_group_ids,
                   rfx_basis_train = rfx_basis, num_gfr=0, num_mcmc=10)
sigma2_samples <- extractParameter(bart_model, "sigma2")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{getRandomEffectSamples.bartmodel}
\alias{getRandomEffectSamples.bartmodel}
\title{Extract Random Effects Samples from BART Model}
\usage{
\method{getRandomEffectSamples}{bartmodel}(object, ...)
}
\arguments{
\item{object}{Object of type \code{bartmodel} containing draws of a BART model and associated sampling outputs.}

\item{...}{Other parameters to be used in random effects extraction}
}
\value{
List of arrays. The alpha array has dimension (\code{num_components}, \code{num_samples}) and is simply a vector if \code{num_components = 1}.
The xi and beta arrays have dimension (\code{num_components}, \code{num_groups}, \code{num_samples}) and is simply a matrix if \code{num_components = 1}.
The sigma array has dimension (\code{num_components}, \code{num_samples}) and is simply a vector if \code{num_components = 1}.
}
\description{
Extract raw sample values for each of the random effect parameter terms.
}
\examples{
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
f_XW <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
snr <- 3
group_ids <- rep(c(1,2), n \%/\% 2)
rfx_coefs <- matrix(c(-1, -1, 1, 1), nrow=2, byrow=TRUE)
rfx_basis <- cbind(1, runif(n, -1, 1))
rfx_term <- rowSums(rfx_coefs[group_ids,] * rfx_basis)
E_y <- f_XW + rfx_term
y <- E_y + rnorm(n, 0, 1)*(sd(E_y)/snr)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
y_test <- y[test_inds]
y_train <- y[train_inds]
rfx_group_ids_test <- group_ids[test_inds]
rfx_group_ids_train <- group_ids[train_inds]
rfx_basis_test <- rfx_basis[test_inds,]
rfx_basis_train <- rfx_basis[train_inds,]
rfx_term_test <- rfx_term[test_inds]
rfx_term_train <- rfx_term[train_inds]
bart_model <- bart(X_train = X_train, y_train = y_train, X_test = X_test,
                   rfx_group_ids_train = rfx_group_ids_train,
                   rfx_group_ids_test = rfx_group_ids_test,
                   rfx_basis_train = rfx_basis_train,
                   rfx_basis_test = rfx_basis_test,
                   num_gfr = 10, num_burnin = 0, num_mcmc = 10)
rfx_samples <- getRandomEffectSamples(bart_model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior_transformation.R
\name{computeContrastBCFModel}
\alias{computeContrastBCFModel}
\title{Compute Contrast for BCF Model}
\usage{
computeContrastBCFModel(
  object,
  X_0,
  X_1,
  Z_0,
  Z_1,
  propensity_0 = NULL,
  propensity_1 = NULL,
  rfx_group_ids_0 = NULL,
  rfx_group_ids_1 = NULL,
  rfx_basis_0 = NULL,
  rfx_basis_1 = NULL,
  type = "posterior",
  scale = "linear"
)
}
\arguments{
\item{object}{Object of type \code{bcfmodel} containing draws of a Bayesian causal forest model and associated sampling outputs.}

\item{X_0}{Covariates used for prediction in the "control" case. Must be a matrix or dataframe.}

\item{X_1}{Covariates used for prediction in the "treatment" case. Must be a matrix or dataframe.}

\item{Z_0}{Treatments used for prediction in the "control" case. Must be a matrix or vector.}

\item{Z_1}{Treatments used for prediction in the "treatment" case. Must be a matrix or vector.}

\item{propensity_0}{(Optional) Propensities used for prediction in the "control" case. Must be a matrix or vector.}

\item{propensity_1}{(Optional) Propensities used for prediction in the "treatment" case. Must be a matrix or vector.}

\item{rfx_group_ids_0}{(Optional) Test set group labels used for prediction from an additive random effects
model in the "control" case. We do not currently support (but plan to in the near future), test set evaluation
for group labels that were not in the training set. Must be a vector.}

\item{rfx_group_ids_1}{(Optional) Test set group labels used for prediction from an additive random effects
model in the "treatment" case. We do not currently support (but plan to in the near future), test set evaluation
for group labels that were not in the training set. Must be a vector.}

\item{rfx_basis_0}{(Optional) Test set basis for used for prediction from an additive random effects model in the "control" case.  Must be a matrix or vector.}

\item{rfx_basis_1}{(Optional) Test set basis for used for prediction from an additive random effects model in the "treatment" case. Must be a matrix or vector.}

\item{type}{(Optional) Aggregation level of the contrast. Options are "mean", which averages the contrast evaluations over every draw of a BCF model, and "posterior", which returns the entire matrix of posterior contrast estimates. Default: "posterior".}

\item{scale}{(Optional) Scale of the contrast. Options are "linear", which returns a contrast on the original scale of the mean forest / RFX terms, and "probability", which transforms each contrast term into a probability of observing \code{y == 1} before taking their difference. "probability" is only valid for models fit with a probit outcome model. Default: "linear".}
}
\value{
List of prediction matrices or single prediction matrix / vector, depending on the terms requested.
}
\description{
Compute a contrast using a BCF model by making two sets of outcome predictions and taking their difference.
For simple BCF models with binary treatment, this will yield the same prediction as requesting \code{terms = "cate"}
in the \code{predict.bcfmodel} function. For more general models, such as models with continuous / multivariate treatments or
an additive random effects term with a coefficient on the treatment, this function provides the flexibility to compute a
any contrast of interest by specifying covariates, treatment, and random effects bases and IDs for both sides of a two term
contrast. For simplicity, we refer to the subtrahend of the contrast as the "control" or \code{Y0} term and the minuend of the
contrast as the \code{Y1} term, though the requested contrast need not match the "control vs treatment" terminology of a classic
two-arm experiment. We mirror the function calls and terminology of the \code{predict.bcfmodel} function, labeling each prediction
data term with a \code{1} to denote its contribution to the treatment prediction of a contrast and \code{0} to denote inclusion in the
control prediction.
}
\examples{
n <- 500
p <- 5
X <- matrix(runif(n*p), ncol = p)
mu_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
pi_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (0.2) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (0.4) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (0.6) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (0.8)
)
tau_x <- (
    ((0 <= X[,2]) & (0.25 > X[,2])) * (0.5) +
    ((0.25 <= X[,2]) & (0.5 > X[,2])) * (1.0) +
    ((0.5 <= X[,2]) & (0.75 > X[,2])) * (1.5) +
    ((0.75 <= X[,2]) & (1 > X[,2])) * (2.0)
)
Z <- rbinom(n, 1, pi_x)
noise_sd <- 1
y <- mu_x + tau_x*Z + rnorm(n, 0, noise_sd)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
pi_test <- pi_x[test_inds]
pi_train <- pi_x[train_inds]
Z_test <- Z[test_inds]
Z_train <- Z[train_inds]
y_test <- y[test_inds]
y_train <- y[train_inds]
mu_test <- mu_x[test_inds]
mu_train <- mu_x[train_inds]
tau_test <- tau_x[test_inds]
tau_train <- tau_x[train_inds]
bcf_model <- bcf(X_train = X_train, Z_train = Z_train, y_train = y_train,
                 propensity_train = pi_train, num_gfr = 10,
                 num_burnin = 0, num_mcmc = 10)
tau_hat_test <- computeContrastBCFModel(
    bcf_model, X_0=X_test, X_1=X_test, Z_0=rep(0, n_test), Z_1=rep(1, n_test),
    propensity_0 = pi_test, propensity_1 = pi_test
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior_transformation.R
\name{computeContrastBARTModel}
\alias{computeContrastBARTModel}
\title{Compute Contrast for BART Model}
\usage{
computeContrastBARTModel(
  object,
  X_0,
  X_1,
  leaf_basis_0 = NULL,
  leaf_basis_1 = NULL,
  rfx_group_ids_0 = NULL,
  rfx_group_ids_1 = NULL,
  rfx_basis_0 = NULL,
  rfx_basis_1 = NULL,
  type = "posterior",
  scale = "linear"
)
}
\arguments{
\item{object}{Object of type \code{bart} containing draws of a regression forest and associated sampling outputs.}

\item{X_0}{Covariates used for prediction in the "control" case. Must be a matrix or dataframe.}

\item{X_1}{Covariates used for prediction in the "treatment" case. Must be a matrix or dataframe.}

\item{leaf_basis_0}{(Optional) Bases used for prediction in the "control" case (by e.g. dot product with leaf values). Default: \code{NULL}.}

\item{leaf_basis_1}{(Optional) Bases used for prediction in the "treatment" case (by e.g. dot product with leaf values). Default: \code{NULL}.}

\item{rfx_group_ids_0}{(Optional) Test set group labels used for prediction from an additive random effects
model in the "control" case. We do not currently support (but plan to in the near future), test set evaluation
for group labels that were not in the training set. Must be a vector.}

\item{rfx_group_ids_1}{(Optional) Test set group labels used for prediction from an additive random effects
model in the "treatment" case. We do not currently support (but plan to in the near future), test set evaluation
for group labels that were not in the training set. Must be a vector.}

\item{rfx_basis_0}{(Optional) Test set basis for used for prediction from an additive random effects model in the "control" case.  Must be a matrix or vector.}

\item{rfx_basis_1}{(Optional) Test set basis for used for prediction from an additive random effects model in the "treatment" case. Must be a matrix or vector.}

\item{type}{(Optional) Aggregation level of the contrast. Options are "mean", which averages the contrast evaluations over every draw of a BART model, and "posterior", which returns the entire matrix of posterior contrast estimates. Default: "posterior".}

\item{scale}{(Optional) Scale of the contrast. Options are "linear", which returns contrast of predictions on the original scale of the mean forest / RFX terms, and "probability". \code{scale = "probability"} is only valid for models fit with a probit / cloglog links on binary or ordinal outcomes. For binary outcome models, \code{scale = "probability"} will return contrasts of the probability that \code{y == 1}. For ordinal outcome models, \code{scale = "probability"} will return contrasts over the "survival function" \code{P(y > k)} for \verb{k = 1, 2, ..., K-1} where \code{K} is the total number of categories. Default: "linear".}
}
\value{
Contrast matrix or vector, depending on whether type = "mean" or "posterior".
}
\description{
Compute a contrast using a BART model by making two sets of outcome predictions and taking their difference.
This function provides the flexibility to compute any contrast of interest by specifying covariates, leaf basis, and random effects
bases / IDs for both sides of a two term contrast. For simplicity, we refer to the subtrahend of the contrast as the "control" or
\code{Y0} term and the minuend of the contrast as the \code{Y1} term, though the requested contrast need not match the "control vs treatment"
terminology of a classic two-treatment causal inference problem. We mirror the function calls and terminology of the \code{predict.bartmodel}
function, labeling each prediction data term with a \code{1} to denote its contribution to the treatment prediction of a contrast and
\code{0} to denote inclusion in the control prediction.

Only valid when there is either a mean forest or a random effects term in the BART model.
}
\examples{
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
W <- matrix(runif(n*1), ncol = 1)
f_XW <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5*W[,1]) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5*W[,1]) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5*W[,1]) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5*W[,1])
)
noise_sd <- 1
y <- f_XW + rnorm(n, 0, noise_sd)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
W_test <- W[test_inds,]
W_train <- W[train_inds,]
y_test <- y[test_inds]
y_train <- y[train_inds]
bart_model <- bart(X_train = X_train, leaf_basis_train = W_train, y_train = y_train,
                   num_gfr = 10, num_burnin = 0, num_mcmc = 10)
contrast_test <- computeContrastBARTModel(
    bart_model,
    X_0 = X_test,
    X_1 = X_test,
    leaf_basis_0 = matrix(0, nrow = n_test, ncol = 1),
    leaf_basis_1 = matrix(1, nrow = n_test, ncol = 1),
    type = "posterior",
    scale = "linear"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_effects.R
\name{createRandomEffectsModel}
\alias{createRandomEffectsModel}
\title{Create RandomEffectsModel Object}
\usage{
createRandomEffectsModel(num_components, num_groups)
}
\arguments{
\item{num_components}{Number of "components" or bases defining the random effects regression}

\item{num_groups}{Number of random effects groups}
}
\value{
\code{RandomEffectsModel} object
}
\description{
Create a \code{RandomEffectsModel} object

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
n <- 100
rfx_group_ids <- sample(1:2, size = n, replace = TRUE)
rfx_basis <- matrix(rep(1.0, n), ncol=1)
num_groups <- length(unique(rfx_group_ids))
num_components <- ncol(rfx_basis)
rfx_model <- createRandomEffectsModel(num_components, num_groups)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior_transformation.R
\name{sampleBCFPosteriorPredictive}
\alias{sampleBCFPosteriorPredictive}
\title{Sample BCF Posterior Predictive}
\usage{
sampleBCFPosteriorPredictive(
  model_object,
  X = NULL,
  Z = NULL,
  propensity = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL,
  num_draws_per_sample = NULL
)
}
\arguments{
\item{model_object}{A fitted BCF model object of class \code{bcfmodel}.}

\item{X}{A matrix or data frame of covariates.}

\item{Z}{A vector or matrix of treatment assignments.}

\item{propensity}{(Optional) A vector or matrix of propensity scores. Required if the underlying model depends on user-provided propensities.}

\item{rfx_group_ids}{(Optional) A vector of group IDs for random effects model. Required if the BCF model includes random effects.}

\item{rfx_basis}{(Optional) A matrix of bases for random effects model. Required if the BCF model includes random effects.}

\item{num_draws_per_sample}{(Optional) The number of samples to draw from the likelihood for each draw of the posterior. Defaults to a heuristic based on the number of samples in a BCF model (i.e. if the BCF model has >1000 draws, we use 1 draw from the likelihood per sample, otherwise we upsample to ensure at least 1000 posterior predictive draws).}
}
\value{
Array of posterior predictive samples with dimensions (num_observations, num_posterior_samples, num_draws_per_sample) if num_draws_per_sample > 1, otherwise (num_observations, num_posterior_samples).
}
\description{
Sample from the posterior predictive distribution for outcomes modeled by BCF
}
\examples{
n <- 100
p <- 5
X <- matrix(rnorm(n * p), nrow = n, ncol = p)
pi_X <- pnorm(X[,1] / 2)
Z <- rbinom(n, 1, pi_X)
y <- 2 * X[,2] + 0.5 * X[,2] * Z + rnorm(n)
bcf_model <- bcf(X_train = X, Z_train = Z, y_train = y, propensity_train = pi_X)
ppd_samples <- sampleBCFPosteriorPredictive(
  model_object = bcf_model, X = X,
  Z = Z, propensity = pi_X
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forest.R
\name{print.ForestSamples}
\alias{print.ForestSamples}
\title{Summarize a ForestSamples Object}
\usage{
\method{print}{ForestSamples}(x, ...)
}
\arguments{
\item{x}{ForestSamples object}

\item{...}{Additional arguments}
}
\value{
ForestSamples object unchanged after printing summary
}
\description{
Prints a summary of the ForestSamples object, including number of forests and the underlying model of each forest.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{summary.bartmodel}
\alias{summary.bartmodel}
\title{Summarize BART Model Fit and Parameters}
\usage{
\method{summary}{bartmodel}(object, ...)
}
\arguments{
\item{object}{The BART model object}

\item{...}{Additional arguments}
}
\value{
BART model object unchanged after summarizing
}
\description{
Summarize a BART fit with a description of the model that was fit and numeric summaries of any sampled quantities.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{OutcomeModel}
\alias{OutcomeModel}
\title{Create a New Outcome Model Object}
\usage{
OutcomeModel(outcome = "continuous", link = NULL)
}
\arguments{
\item{outcome}{Character string specifying the outcome type.}

\item{link}{Character string specifying the link function.}
}
\value{
An object of class \code{outcome_model}.
}
\description{
Create a New Outcome Model Object
}
\examples{
my_model <- OutcomeModel(outcome = "continuous", link = "identity")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stochtree-package.R
\docType{package}
\name{stochtree-package}
\alias{stochtree}
\alias{stochtree-package}
\title{stochtree: Stochastic Tree Ensembles (XBART and BART) for Supervised Learning and Causal Inference}
\description{
Flexible stochastic tree ensemble software. Robust implementations of Bayesian Additive Regression Trees (BART) (Chipman, George, McCulloch (2010) \doi{10.1214/09-AOAS285}) for supervised learning and Bayesian Causal Forests (BCF) (Hahn, Murray, Carvalho (2020) \doi{10.1214/19-BA1195}) for causal inference. Enables model serialization and parallel sampling and provides a low-level interface for custom stochastic forest samplers. Includes the grow-from-root algorithm for accelerated forest sampling (He and Hahn (2021) \doi{10.1080/01621459.2021.1942012}), a log-linear leaf model for forest-based heteroskedasticity (Murray (2020) \doi{10.1080/01621459.2020.1813587}), and the cloglog BART model of Alam and Linero (2025) \doi{10.48550/arXiv.2502.00606} for ordinal outcomes.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://stochtree.ai/}
  \item \url{https://github.com/StochasticTree/stochtree}
  \item Report bugs at \url{https://github.com/StochasticTree/stochtree/issues}
}

}
\author{
\strong{Maintainer}: Drew Herren \email{drewherrenopensource@gmail.com} (\href{https://orcid.org/0000-0003-4109-6611}{ORCID})

Authors:
\itemize{
  \item Richard Hahn
  \item Jared Murray
  \item Carlos Carvalho
  \item Jingyu He
}

Other contributors:
\itemize{
  \item Pedro Lima [contributor]
  \item Entejar Alam [contributor]
  \item stochtree contributors [copyright holder]
  \item Eigen contributors (C++ source uses the Eigen library for matrix operations, see inst/COPYRIGHTS) [copyright holder]
  \item xgboost contributors (C++ tree code and related operations include or are inspired by code from the xgboost library, see inst/COPYRIGHTS) [copyright holder]
  \item treelite contributors (C++ tree code and related operations include or are inspired by code from the treelite library, see inst/COPYRIGHTS) [copyright holder]
  \item Microsoft Corporation (C++ I/O and various project structure code include or are inspired by code from the LightGBM library, which is a copyright of Microsoft, see inst/COPYRIGHTS) [copyright holder]
  \item Niels Lohmann (C++ source uses the JSON for Modern C++ library for JSON operations, see inst/COPYRIGHTS) [copyright holder]
  \item Daniel Lemire (C++ source uses the fast_double_parser library internally, see inst/COPYRIGHTS) [copyright holder]
  \item Victor Zverovich (C++ source uses the fmt library internally, see inst/COPYRIGHTS) [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialization.R
\name{ForestSamplesSerialization}
\alias{ForestSamplesSerialization}
\alias{loadForestContainerJson}
\alias{loadForestContainerCombinedJson}
\alias{loadForestContainerCombinedJsonString}
\title{Forest Container Serialization Routines}
\usage{
loadForestContainerJson(json_object, json_forest_label)

loadForestContainerCombinedJson(json_object_list, json_forest_label)

loadForestContainerCombinedJsonString(json_string_list, json_forest_label)
}
\arguments{
\item{json_object}{Object of class \code{CppJson}}

\item{json_forest_label}{Label referring to a particular forest (i.e. "forest_0") in the overall json hierarchy (must exist in every json object in a list if a list is provided)}

\item{json_object_list}{List of objects of class \code{CppJson}}

\item{json_string_list}{List of strings that parse into objects of type \code{CppJson}}
}
\value{
Each of the functions in this group returns a \code{ForestSamples} object.
}
\description{
While the \link{BARTSerialization} and \link{BCFSerialization} topics focus on JSON serialization / deserialization for
entire \code{bartmodel} and \code{bcfmodel} objects, this function group provides an interface for a more focused use case:
loading a single \link{ForestSamples} container from a broader BART / BCF model (which may include multiple forests and other parametric terms).

\code{loadForestContainerJson} converts a \link{CppJson} object representing a BART or BCF model into a \link{ForestSamples} container
by extracting the JSON indexed by a forest label (i.e. \code{"forest_0"}) and deserializing it into a \link{ForestSamples} object.

Both \code{loadForestContainerJson} and \code{loadForestContainerCombinedJson} operate similarly, but on a list of \link{CppJson} or JSON string
representations of BART / BCF models with the same structure.

These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
X <- matrix(runif(10*100), ncol = 10)
y <- -5 + 10*(X[,1] > 0.5) + rnorm(100)
bart_model <- bart(X, y, num_gfr=0, num_mcmc=10)
bart_json <- saveBARTModelToJson(bart_model)
bart_json_string <- saveBARTModelToJsonString(bart_model)
bart_json_list <- list(bart_json)
bart_json_string_list <- list(bart_json_string)
mean_forest <- loadForestContainerJson(bart_json, "forest_0")
mean_forest <- loadForestContainerCombinedJson(bart_json_list, "forest_0")
mean_forest <- loadForestContainerCombinedJsonString(bart_json_string_list, "forest_0")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{extractParameter.bartmodel}
\alias{extractParameter.bartmodel}
\title{Extract BART Parameter Samples}
\usage{
\method{extractParameter}{bartmodel}(object, term)
}
\arguments{
\item{object}{Object of type \code{bartmodel} containing draws of a BART model and associated sampling outputs.}

\item{term}{Name of the parameter to extract (e.g., \code{"sigma2"}, \code{"y_hat_train"}, etc.)}
}
\value{
Array of parameter samples. If the underlying parameter is a scalar, this will be a vector of length \code{num_samples}.
If the underlying parameter is vector-valued, this will be (\code{parameter_dimension} x \code{num_samples}) matrix, and if the underlying
parameter is multidimensional, this will be an array of dimension (\code{parameter_dimension_1} x \code{parameter_dimension_2} x ... x \code{num_samples}).
}
\description{
Extract a vector, matrix or array of parameter samples from a BART model by name.
Random effects are handled by a separate \code{getRandomEffectSamples} function due to the complexity of the random effects parameters.
If the requested model term is not found, an error is thrown.
The following conventions are used for parameter names:
\itemize{
\item Global error variance: \code{"sigma2"}, \code{"global_error_scale"}, \code{"sigma2_global"}
\item Leaf scale: \code{"sigma2_leaf"}, \code{"leaf_scale"}
\item In-sample mean function predictions: \code{"y_hat_train"}
\item Test set mean function predictions: \code{"y_hat_test"}
\item In-sample variance forest predictions: \code{"sigma2_x_train"}, \code{"var_x_train"}
\item Test set variance forest predictions: \code{"sigma2_x_test"}, \code{"var_x_test"}
\item Ordinal model cutpoints (valid only for ordinal cloglog models): \code{"cloglog_cutpoints"}, \code{"cutpoints"}
}
}
\examples{
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
f_XW <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
snr <- 3
group_ids <- rep(c(1,2), n \%/\% 2)
rfx_coefs <- matrix(c(-1, -1, 1, 1), nrow=2, byrow=TRUE)
rfx_basis <- cbind(1, runif(n, -1, 1))
rfx_term <- rowSums(rfx_coefs[group_ids,] * rfx_basis)
E_y <- f_XW + rfx_term
y <- E_y + rnorm(n, 0, 1)*(sd(E_y)/snr)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
y_test <- y[test_inds]
y_train <- y[train_inds]
rfx_group_ids_test <- group_ids[test_inds]
rfx_group_ids_train <- group_ids[train_inds]
rfx_basis_test <- rfx_basis[test_inds,]
rfx_basis_train <- rfx_basis[train_inds,]
rfx_term_test <- rfx_term[test_inds]
rfx_term_train <- rfx_term[train_inds]
bart_model <- bart(X_train = X_train, y_train = y_train, X_test = X_test,
                   rfx_group_ids_train = rfx_group_ids_train,
                   rfx_group_ids_test = rfx_group_ids_test,
                   rfx_basis_train = rfx_basis_train,
                   rfx_basis_test = rfx_basis_test,
                   num_gfr = 10, num_burnin = 0, num_mcmc = 10)
sigma2_samples <- extractParameter(bart_model, "sigma2")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{Outcome}
\alias{Outcome}
\title{Outcome Data C++ Wrapper}
\description{
Outcome / partial residual used to sample an additive model.
The outcome class is a wrapper around a vector of (mutable)
outcomes for ML tasks (supervised learning, causal inference).
When an additive tree ensemble is sampled, the outcome used to
sample a specific model term is the "partial residual" consisting
of the outcome minus the predictions of every other model term
(trees, group random effects, etc...).

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{data_ptr}}{External pointer to a C++ \code{Outcome} class}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Outcome-new}{\code{Outcome$new()}}
\item \href{#method-Outcome-get_data}{\code{Outcome$get_data()}}
\item \href{#method-Outcome-add_vector}{\code{Outcome$add_vector()}}
\item \href{#method-Outcome-subtract_vector}{\code{Outcome$subtract_vector()}}
\item \href{#method-Outcome-update_data}{\code{Outcome$update_data()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Outcome-new"></a>}}
\if{latex}{\out{\hypertarget{method-Outcome-new}{}}}
\subsection{Method \code{new()}}{
Create a new Outcome object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Outcome$new(outcome)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{outcome}}{Vector of outcome values}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{Outcome} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Outcome-get_data"></a>}}
\if{latex}{\out{\hypertarget{method-Outcome-get_data}{}}}
\subsection{Method \code{get_data()}}{
Extract raw data in R from the underlying C++ object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Outcome$get_data()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
R vector containing (copy of) the values in \code{Outcome} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Outcome-add_vector"></a>}}
\if{latex}{\out{\hypertarget{method-Outcome-add_vector}{}}}
\subsection{Method \code{add_vector()}}{
Update the current state of the outcome (i.e. partial residual) data by adding the values of \code{update_vector}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Outcome$add_vector(update_vector)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{update_vector}}{Vector to be added to outcome}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Outcome-subtract_vector"></a>}}
\if{latex}{\out{\hypertarget{method-Outcome-subtract_vector}{}}}
\subsection{Method \code{subtract_vector()}}{
Update the current state of the outcome (i.e. partial residual) data by subtracting the values of \code{update_vector}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Outcome$subtract_vector(update_vector)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{update_vector}}{Vector to be subtracted from outcome}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Outcome-update_data"></a>}}
\if{latex}{\out{\hypertarget{method-Outcome-update_data}{}}}
\subsection{Method \code{update_data()}}{
Update the current state of the outcome (i.e. partial residual) data by replacing each element with the elements of \code{new_vector}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Outcome$update_data(new_vector)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{new_vector}}{Vector from which to overwrite the current data}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_effects.R
\name{RandomEffectsModel}
\alias{RandomEffectsModel}
\title{Random Effects Model C++ Wrapper}
\description{
The core "model" class for sampling random effects.
Stores current model state, prior parameters, and procedures for
sampling from the conditional posterior of each parameter.

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{rfx_model_ptr}}{External pointer to a C++ StochTree::RandomEffectsModel class}

\item{\code{num_groups}}{Number of groups in the random effects model}

\item{\code{num_components}}{Number of components (i.e. dimension of basis) in the random effects model}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-RandomEffectsModel-new}{\code{RandomEffectsModel$new()}}
\item \href{#method-RandomEffectsModel-sample_random_effect}{\code{RandomEffectsModel$sample_random_effect()}}
\item \href{#method-RandomEffectsModel-predict}{\code{RandomEffectsModel$predict()}}
\item \href{#method-RandomEffectsModel-set_working_parameter}{\code{RandomEffectsModel$set_working_parameter()}}
\item \href{#method-RandomEffectsModel-set_group_parameters}{\code{RandomEffectsModel$set_group_parameters()}}
\item \href{#method-RandomEffectsModel-set_working_parameter_cov}{\code{RandomEffectsModel$set_working_parameter_cov()}}
\item \href{#method-RandomEffectsModel-set_group_parameter_cov}{\code{RandomEffectsModel$set_group_parameter_cov()}}
\item \href{#method-RandomEffectsModel-set_variance_prior_shape}{\code{RandomEffectsModel$set_variance_prior_shape()}}
\item \href{#method-RandomEffectsModel-set_variance_prior_scale}{\code{RandomEffectsModel$set_variance_prior_scale()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-new"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-new}{}}}
\subsection{Method \code{new()}}{
Create a new RandomEffectsModel object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$new(num_components, num_groups)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{num_components}}{Number of "components" or bases defining the random effects regression}

\item{\code{num_groups}}{Number of random effects groups}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{RandomEffectsModel} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-sample_random_effect"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-sample_random_effect}{}}}
\subsection{Method \code{sample_random_effect()}}{
Sample from random effects model.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$sample_random_effect(
  rfx_dataset,
  residual,
  rfx_tracker,
  rfx_samples,
  keep_sample,
  global_variance,
  rng
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{rfx_dataset}}{Object of type \code{RandomEffectsDataset}}

\item{\code{residual}}{Object of type \code{Outcome}}

\item{\code{rfx_tracker}}{Object of type \code{RandomEffectsTracker}}

\item{\code{rfx_samples}}{Object of type \code{RandomEffectSamples}}

\item{\code{keep_sample}}{Whether sample should be retained in \code{rfx_samples}. If \code{FALSE}, the state of \code{rfx_tracker} will be updated, but the parameter values will not be added to the sample container. Samples are commonly discarded due to burn-in or thinning.}

\item{\code{global_variance}}{Scalar global variance parameter}

\item{\code{rng}}{Object of type \code{CppRNG}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-predict"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-predict}{}}}
\subsection{Method \code{predict()}}{
Predict from (a single sample of a) random effects model.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$predict(rfx_dataset, rfx_tracker)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{rfx_dataset}}{Object of type \code{RandomEffectsDataset}}

\item{\code{rfx_tracker}}{Object of type \code{RandomEffectsTracker}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Vector of predictions with size matching number of observations in rfx_dataset
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-set_working_parameter"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-set_working_parameter}{}}}
\subsection{Method \code{set_working_parameter()}}{
Set value for the "working parameter." This is typically
used for initialization, but could also be used to interrupt
or override the sampler.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$set_working_parameter(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{Parameter input}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-set_group_parameters"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-set_group_parameters}{}}}
\subsection{Method \code{set_group_parameters()}}{
Set value for the "group parameters." This is typically
used for initialization, but could also be used to interrupt
or override the sampler.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$set_group_parameters(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{Parameter input}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-set_working_parameter_cov"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-set_working_parameter_cov}{}}}
\subsection{Method \code{set_working_parameter_cov()}}{
Set value for the working parameter covariance. This is typically
used for initialization, but could also be used to interrupt
or override the sampler.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$set_working_parameter_cov(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{Parameter input}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-set_group_parameter_cov"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-set_group_parameter_cov}{}}}
\subsection{Method \code{set_group_parameter_cov()}}{
Set value for the group parameter covariance. This is typically
used for initialization, but could also be used to interrupt
or override the sampler.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$set_group_parameter_cov(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{Parameter input}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-set_variance_prior_shape"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-set_variance_prior_shape}{}}}
\subsection{Method \code{set_variance_prior_shape()}}{
Set shape parameter for the group parameter variance prior.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$set_variance_prior_shape(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{Parameter input}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RandomEffectsModel-set_variance_prior_scale"></a>}}
\if{latex}{\out{\hypertarget{method-RandomEffectsModel-set_variance_prior_scale}{}}}
\subsection{Method \code{set_variance_prior_scale()}}{
Set shape parameter for the group parameter variance prior.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RandomEffectsModel$set_variance_prior_scale(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{Parameter input}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{createCppRNG}
\alias{createCppRNG}
\title{Create CppRNG Object}
\usage{
createCppRNG(random_seed = -1)
}
\arguments{
\item{random_seed}{(Optional) random seed for sampling}
}
\value{
\code{CppRng} object
}
\description{
Create an R class that wraps a C++ random number generator
}
\examples{
rng <- createCppRNG(1234)
rng <- createCppRNG()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{ForestDataset}
\alias{ForestDataset}
\title{Forest Dataset C++ Wrapper}
\description{
Wrapper around a C++ dataset class used to sample a forest.
A dataset consists of three matrices / vectors: covariates,
bases, and variance weights. Both the basis vector and variance
weights are optional.

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{data_ptr}}{External pointer to a C++ ForestDataset class}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ForestDataset-new}{\code{ForestDataset$new()}}
\item \href{#method-ForestDataset-update_basis}{\code{ForestDataset$update_basis()}}
\item \href{#method-ForestDataset-update_variance_weights}{\code{ForestDataset$update_variance_weights()}}
\item \href{#method-ForestDataset-num_observations}{\code{ForestDataset$num_observations()}}
\item \href{#method-ForestDataset-num_covariates}{\code{ForestDataset$num_covariates()}}
\item \href{#method-ForestDataset-num_basis}{\code{ForestDataset$num_basis()}}
\item \href{#method-ForestDataset-get_covariates}{\code{ForestDataset$get_covariates()}}
\item \href{#method-ForestDataset-get_basis}{\code{ForestDataset$get_basis()}}
\item \href{#method-ForestDataset-get_variance_weights}{\code{ForestDataset$get_variance_weights()}}
\item \href{#method-ForestDataset-has_basis}{\code{ForestDataset$has_basis()}}
\item \href{#method-ForestDataset-has_variance_weights}{\code{ForestDataset$has_variance_weights()}}
\item \href{#method-ForestDataset-has_auxiliary_dimension}{\code{ForestDataset$has_auxiliary_dimension()}}
\item \href{#method-ForestDataset-add_auxiliary_dimension}{\code{ForestDataset$add_auxiliary_dimension()}}
\item \href{#method-ForestDataset-get_auxiliary_data_value}{\code{ForestDataset$get_auxiliary_data_value()}}
\item \href{#method-ForestDataset-set_auxiliary_data_value}{\code{ForestDataset$set_auxiliary_data_value()}}
\item \href{#method-ForestDataset-get_auxiliary_data_vector}{\code{ForestDataset$get_auxiliary_data_vector()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-new"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-new}{}}}
\subsection{Method \code{new()}}{
Create a new ForestDataset object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$new(covariates, basis = NULL, variance_weights = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{covariates}}{Matrix of covariates}

\item{\code{basis}}{(Optional) Matrix of bases used to define a leaf regression}

\item{\code{variance_weights}}{(Optional) Vector of observation-specific variance weights}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{ForestDataset} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-update_basis"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-update_basis}{}}}
\subsection{Method \code{update_basis()}}{
Update basis matrix in a dataset
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$update_basis(basis)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{basis}}{Updated matrix of bases used to define a leaf regression}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-update_variance_weights"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-update_variance_weights}{}}}
\subsection{Method \code{update_variance_weights()}}{
Update variance_weights in a dataset
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$update_variance_weights(variance_weights, exponentiate = F)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{variance_weights}}{Updated vector of variance weights used to define individual variance / case weights}

\item{\code{exponentiate}}{Whether or not input vector should be exponentiated before being written to the Dataset's variance weights. Default: F.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-num_observations"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-num_observations}{}}}
\subsection{Method \code{num_observations()}}{
Return number of observations in a \code{ForestDataset} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$num_observations()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Observation count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-num_covariates"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-num_covariates}{}}}
\subsection{Method \code{num_covariates()}}{
Return number of covariates in a \code{ForestDataset} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$num_covariates()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Covariate count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-num_basis"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-num_basis}{}}}
\subsection{Method \code{num_basis()}}{
Return number of bases in a \code{ForestDataset} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$num_basis()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Basis count
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-get_covariates"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-get_covariates}{}}}
\subsection{Method \code{get_covariates()}}{
Return covariates as an R matrix
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$get_covariates()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Covariate data
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-get_basis"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-get_basis}{}}}
\subsection{Method \code{get_basis()}}{
Return bases as an R matrix
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$get_basis()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Basis data
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-get_variance_weights"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-get_variance_weights}{}}}
\subsection{Method \code{get_variance_weights()}}{
Return variance weights as an R vector
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$get_variance_weights()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Variance weight data
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-has_basis"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-has_basis}{}}}
\subsection{Method \code{has_basis()}}{
Whether or not a dataset has a basis matrix
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$has_basis()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
True if basis matrix is loaded, false otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-has_variance_weights"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-has_variance_weights}{}}}
\subsection{Method \code{has_variance_weights()}}{
Whether or not a dataset has variance weights
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$has_variance_weights()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
True if variance weights are loaded, false otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-has_auxiliary_dimension"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-has_auxiliary_dimension}{}}}
\subsection{Method \code{has_auxiliary_dimension()}}{
Whether or not a dataset has auxiliary data stored at the dimension indicated
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$has_auxiliary_dimension(dim_idx)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_idx}}{Dimension of auxiliary data}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
True if auxiliary data has been allocated for \code{dim_idx} False otherwise
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-add_auxiliary_dimension"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-add_auxiliary_dimension}{}}}
\subsection{Method \code{add_auxiliary_dimension()}}{
Initialize a new dimension / lane of auxiliary data and allocate data in its place
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$add_auxiliary_dimension(dim_size)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_size}}{Size of the new vector of data to allocate}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-get_auxiliary_data_value"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-get_auxiliary_data_value}{}}}
\subsection{Method \code{get_auxiliary_data_value()}}{
Retrieve auxiliary data value
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$get_auxiliary_data_value(dim_idx, element_idx)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_idx}}{Dimension from which data value to be retrieved}

\item{\code{element_idx}}{Element to retrieve from dimension \code{dim_idx}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Floating point value stored in the requested auxiliary data space
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-set_auxiliary_data_value"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-set_auxiliary_data_value}{}}}
\subsection{Method \code{set_auxiliary_data_value()}}{
Set auxiliary data value
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$set_auxiliary_data_value(dim_idx, element_idx, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_idx}}{Dimension in which data value to be set}

\item{\code{element_idx}}{Element to set within dimension \code{dim_idx}}

\item{\code{value}}{Data value to set at auxiliary data dimension \code{dim_idx} and element \code{element_idx}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
None
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ForestDataset-get_auxiliary_data_vector"></a>}}
\if{latex}{\out{\hypertarget{method-ForestDataset-get_auxiliary_data_vector}{}}}
\subsection{Method \code{get_auxiliary_data_vector()}}{
Retrieve entire auxiliary data vector
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ForestDataset$get_auxiliary_data_vector(dim_idx)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_idx}}{Dimension to retrieve}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Vector of all of the auxiliary data stored at dimension \code{dim_idx}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{GlobalModelConfig}
\alias{GlobalModelConfig}
\title{Global Model Configuration Object}
\value{
Global error variance parameter
}
\description{
Object used to get / set global parameters and other global model
configuration options in the "low-level" stochtree interface.
The "low-level" stochtree interface enables a high degreee of sampler
customization, in which users employ R wrappers around C++ objects
like ForestDataset, Outcome, CppRng, and ForestModel to run the
Gibbs sampler of a BART model with custom modifications.
GlobalModelConfig allows users to specify / query the global parameters
of a model they wish to run.

This class is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{global_error_variance}}{Global error variance parameter
Create a new GlobalModelConfig object.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-GlobalModelConfig-new}{\code{GlobalModelConfig$new()}}
\item \href{#method-GlobalModelConfig-update_global_error_variance}{\code{GlobalModelConfig$update_global_error_variance()}}
\item \href{#method-GlobalModelConfig-get_global_error_variance}{\code{GlobalModelConfig$get_global_error_variance()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GlobalModelConfig-new"></a>}}
\if{latex}{\out{\hypertarget{method-GlobalModelConfig-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GlobalModelConfig$new(global_error_variance = 1)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{global_error_variance}}{Global error variance parameter (default: \code{1.0})}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new GlobalModelConfig object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GlobalModelConfig-update_global_error_variance"></a>}}
\if{latex}{\out{\hypertarget{method-GlobalModelConfig-update_global_error_variance}{}}}
\subsection{Method \code{update_global_error_variance()}}{
Update global error variance parameter
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GlobalModelConfig$update_global_error_variance(global_error_variance)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{global_error_variance}}{Global error variance parameter}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GlobalModelConfig-get_global_error_variance"></a>}}
\if{latex}{\out{\hypertarget{method-GlobalModelConfig-get_global_error_variance}{}}}
\subsection{Method \code{get_global_error_variance()}}{
Query global error variance parameter for this GlobalModelConfig object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GlobalModelConfig$get_global_error_variance()}\if{html}{\out{</div>}}
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcf.R
\name{extractParameter.bcfmodel}
\alias{extractParameter.bcfmodel}
\title{Extract BCF Parameter Samples}
\usage{
\method{extractParameter}{bcfmodel}(object, term)
}
\arguments{
\item{object}{Object of type \code{bcfmodel} containing draws of a BCF model and associated sampling outputs.}

\item{term}{Name of the parameter to extract (e.g., \code{"sigma2"}, \code{"y_hat_train"}, etc.)}
}
\value{
Array of parameter samples. If the underlying parameter is a scalar, this will be a vector of length \code{num_samples}.
If the underlying parameter is vector-valued, this will be (\code{parameter_dimension} x \code{num_samples}) matrix, and if the underlying
parameter is multidimensional, this will be an array of dimension (\code{parameter_dimension_1} x \code{parameter_dimension_2} x ... x \code{num_samples}).
}
\description{
Extract a vector, matrix or array of parameter samples from a BCF model by name.
Random effects are handled by a separate \code{getRandomEffectSamples} function due to the complexity of the random effects parameters.
If the requested model term is not found, an error is thrown.
The following conventions are used for parameter names:
\itemize{
\item Global error variance: \code{"sigma2"}, \code{"global_error_scale"}, \code{"sigma2_global"}
\item Prognostic forest leaf scale: \code{"sigma2_leaf_mu"}, \code{"leaf_scale_mu"}, \code{"mu_leaf_scale"}
\item Treatment effect forest leaf scale: \code{"sigma2_leaf_tau"}, \code{"leaf_scale_tau"}, \code{"tau_leaf_scale"}
\item Adaptive coding parameters: \code{"adaptive_coding"} (returns both the control and treated parameters jointly, with control in the first row and treated in the second row)
\item In-sample mean function predictions: \code{"y_hat_train"}
\item Test set mean function predictions: \code{"y_hat_test"}
\item In-sample treatment effect forest predictions: \code{"tau_hat_train"}
\item Test set treatment effect forest predictions: \code{"tau_hat_test"}
\item Treatment effect intercept: \code{"tau_0"}, \code{"treatment_intercept"}, \code{"tau_intercept"}
\item In-sample variance forest predictions: \code{"sigma2_x_train"}, \code{"var_x_train"}
\item Test set variance forest predictions: \code{"sigma2_x_test"}, \code{"var_x_test"}
}
}
\examples{
n <- 500
p <- 5
X <- matrix(runif(n*p), ncol = p)
mu_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
pi_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (0.2) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (0.4) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (0.6) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (0.8)
)
tau_x <- (
    ((0 <= X[,2]) & (0.25 > X[,2])) * (0.5) +
    ((0.25 <= X[,2]) & (0.5 > X[,2])) * (1.0) +
    ((0.5 <= X[,2]) & (0.75 > X[,2])) * (1.5) +
    ((0.75 <= X[,2]) & (1 > X[,2])) * (2.0)
)
Z <- rbinom(n, 1, pi_x)
E_XZ <- mu_x + Z*tau_x
snr <- 3
rfx_group_ids <- rep(c(1,2), n \%/\% 2)
rfx_coefs <- matrix(c(-1, -1, 1, 1), nrow=2, byrow=TRUE)
rfx_basis <- cbind(1, runif(n, -1, 1))
rfx_term <- rowSums(rfx_coefs[rfx_group_ids,] * rfx_basis)
y <- E_XZ + rfx_term + rnorm(n, 0, 1)*(sd(E_XZ)/snr)
bcf_model <- bcf(X_train = X, y_train = y, Z_train = Z,
                 rfx_group_ids_train = rfx_group_ids,
                 rfx_basis_train = rfx_basis,
                 num_gfr = 10, num_burnin = 0, num_mcmc = 10)
sigma2_samples <- extractParameter(bcf_model, "sigma2")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{bart}
\alias{bart}
\title{Run BART for Supervised Learning}
\usage{
bart(
  X_train,
  y_train,
  leaf_basis_train = NULL,
  rfx_group_ids_train = NULL,
  rfx_basis_train = NULL,
  X_test = NULL,
  leaf_basis_test = NULL,
  rfx_group_ids_test = NULL,
  rfx_basis_test = NULL,
  observation_weights = NULL,
  num_gfr = 5,
  num_burnin = 0,
  num_mcmc = 100,
  previous_model_json = NULL,
  previous_model_warmstart_sample_num = NULL,
  general_params = list(),
  mean_forest_params = list(),
  variance_forest_params = list(),
  random_effects_params = list()
)
}
\arguments{
\item{X_train}{Covariates used to split trees in the ensemble. May be provided either as a dataframe or a matrix.
Matrix covariates will be assumed to be all numeric. Covariates passed as a dataframe will be
preprocessed based on the variable types (e.g. categorical columns stored as unordered factors will be one-hot encoded,
categorical columns stored as ordered factors will passed as integers to the core algorithm, along with the metadata
that the column is ordered categorical).}

\item{y_train}{Outcome to be modeled by the ensemble.}

\item{leaf_basis_train}{(Optional) Bases used to define a regression model \code{y ~ W} in
each leaf of each regression tree. By default, BART assumes constant leaf node
parameters, implicitly regressing on a constant basis of ones (i.e. \code{y ~ 1}).}

\item{rfx_group_ids_train}{(Optional) Group labels used for an additive random effects model.}

\item{rfx_basis_train}{(Optional) Basis for "random-slope" regression in an additive random effects model.
If \code{rfx_group_ids_train} is provided with a regression basis, an intercept-only random effects model
will be estimated.}

\item{X_test}{(Optional) Test set of covariates used to define "out of sample" evaluation data.
May be provided either as a dataframe or a matrix, but the format of \code{X_test} must be consistent with
that of \code{X_train}.}

\item{leaf_basis_test}{(Optional) Test set of bases used to define "out of sample" evaluation data.
While a test set is optional, the structure of any provided test set must match that
of the training set (i.e. if both \code{X_train} and \code{leaf_basis_train} are provided, then a test set must
consist of \code{X_test} and \code{leaf_basis_test} with the same number of columns).}

\item{rfx_group_ids_test}{(Optional) Test set group labels used for an additive random effects model.
We do not currently support (but plan to in the near future), test set evaluation for group labels
that were not in the training set.}

\item{rfx_basis_test}{(Optional) Test set basis for "random-slope" regression in additive random effects model.}

\item{observation_weights}{(Optional) Numeric vector of observation weights of length \code{nrow(X_train)}. Weights are
applied as \code{y_i | - ~ N(mu(X_i), sigma^2 / w_i)}, so larger weights increase an observation's influence on the fit.
All weights must be non-negative. Default: \code{NULL} (all observations equally weighted). Compatible with Gaussian
(continuous/identity) and probit outcome models; not compatible with cloglog link functions. Note: these are
referred to internally in the C++ layer as "variance weights" (\code{var_weights}), since they scale the residual variance.}

\item{num_gfr}{Number of "warm-start" iterations run using the grow-from-root algorithm (He and Hahn, 2021). Default: 5.}

\item{num_burnin}{Number of "burn-in" iterations of the MCMC sampler. Default: 0.}

\item{num_mcmc}{Number of "retained" iterations of the MCMC sampler. Default: 100.}

\item{previous_model_json}{(Optional) JSON string containing a previous BART model. This can be used to "continue" a sampler interactively after inspecting the samples or to run parallel chains "warm-started" from existing forest samples. Default: \code{NULL}.}

\item{previous_model_warmstart_sample_num}{(Optional) Sample number from \code{previous_model_json} that will be used to warmstart this BART sampler. One-indexed (so that the first sample is used for warm-start by setting \code{previous_model_warmstart_sample_num = 1}). Default: \code{NULL}.  If \code{num_chains} in the \code{general_params} list is > 1, then each successive chain will be initialized from a different sample, counting backwards from \code{previous_model_warmstart_sample_num}. That is, if \code{previous_model_warmstart_sample_num = 10} and \code{num_chains = 4}, then chain 1 will be initialized from sample 10, chain 2 from sample 9, chain 3 from sample 8, and chain 4 from sample 7. If \code{previous_model_json} is provided but \code{previous_model_warmstart_sample_num} is NULL, the last sample in the previous model will be used to initialize the first chain, counting backwards as noted before. If more chains are requested than there are samples in \code{previous_model_json}, a warning will be raised and only the last sample will be used.}

\item{general_params}{(Optional) A list of general (non-forest-specific) model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{cutpoint_grid_size} Maximum size of the "grid" of potential cutpoints to consider in the GFR algorithm. Default: \code{100}.
\item \code{standardize} Whether or not to standardize the outcome (and store the offset / scale in the model object). Default: \code{TRUE}.
\item \code{sample_sigma2_global} Whether or not to update the \code{sigma^2} global error variance parameter based on \code{IG(sigma2_global_shape, sigma2_global_scale)}. Default: \code{TRUE}.
\item \code{sigma2_global_init} Starting value of global error variance parameter. Calibrated internally as \code{1.0*var(y_train)}, where \code{y_train} is the possibly standardized outcome, if not set.
\item \code{sigma2_global_shape} Shape parameter in the \code{IG(sigma2_global_shape, sigma2_global_scale)} global error variance model. Default: \code{0}.
\item \code{sigma2_global_scale} Scale parameter in the \code{IG(sigma2_global_shape, sigma2_global_scale)} global error variance model. Default: \code{0}.
\item \code{variable_weights} Numeric weights reflecting the relative probability of splitting on each variable. Does not need to sum to 1 but cannot be negative. Defaults to \code{rep(1/ncol(X_train), ncol(X_train))} if not set here. Note that if the propensity score is included as a covariate in either forest, its weight will default to \code{1/ncol(X_train)}.
\item \code{random_seed} Integer parameterizing the C++ random number generator. If not specified, the C++ random number generator is seeded according to \code{std::random_device}.
\item \code{keep_burnin} Whether or not "burnin" samples should be included in the stored samples of forests and other parameters. Default \code{FALSE}. Ignored if \code{num_mcmc = 0}.
\item \code{keep_gfr} Whether or not "grow-from-root" samples should be included in the stored samples of forests and other parameters. Default \code{FALSE}. Ignored if \code{num_mcmc = 0}.
\item \code{keep_every} How many iterations of the burned-in MCMC sampler should be run before forests and parameters are retained. Default \code{1}. Setting \code{keep_every <- k} for some \code{k > 1} will "thin" the MCMC samples by retaining every \code{k}-th sample, rather than simply every sample. This can reduce the autocorrelation of the MCMC samples.
\item \code{num_chains} How many independent MCMC chains should be sampled. If \code{num_mcmc = 0}, this is ignored. If \code{num_gfr = 0}, then each chain is run from root for \code{num_mcmc * keep_every + num_burnin} iterations, with \code{num_mcmc} samples retained. If \code{num_gfr > 0}, each MCMC chain will be initialized from a separate GFR ensemble, with the requirement that \code{num_gfr >= num_chains}. Default: \code{1}. Note that if \code{num_chains > 1}, the returned model object will contain samples from all chains, stored consecutively. That is, if there are 4 chains with 100 samples each, the first 100 samples will be from chain 1, the next 100 samples will be from chain 2, etc... For more detail on working with multi-chain BART models, see \href{https://stochtree.ai/vignettes/multi-chain.html}{the multi chain vignette}.
\item \code{verbose} Whether or not to print progress during the sampling loops. Default: \code{FALSE}.
\item \code{outcome_model} A structured \code{OutcomeModel} object that specifies the outcome type and desired link function. This argument pre-empts the legacy (deprecated) \code{probit_outcome_model} option. Default: \code{OutcomeModel(outcome='continuous', link='identity')}.
\item \code{probit_outcome_model} Deprecated in favor of \code{outcome_model}. Whether or not the outcome should be modeled as explicitly binary via a probit link. If \code{TRUE}, \code{y} must only contain the values \code{0} and \code{1}. Default: \code{FALSE}.
\item \code{num_threads} Number of threads to use in the GFR and MCMC algorithms, as well as prediction. If OpenMP is not available on a user's setup, this will default to \code{1}, otherwise to the maximum number of available threads.
}}

\item{mean_forest_params}{(Optional) A list of mean forest model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{num_trees} Number of trees in the ensemble for the conditional mean model. Default: \code{200}. If \code{num_trees = 0}, the conditional mean will not be modeled using a forest, and the function will only proceed if \code{num_trees > 0} for the variance forest.
\item \code{alpha} Prior probability of splitting for a tree of depth 0 in the mean model. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{0.95}.
\item \code{beta} Exponent that decreases split probabilities for nodes of depth > 0 in the mean model. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{2}.
\item \code{min_samples_leaf} Minimum allowable size of a leaf, in terms of training samples, in the mean model. Default: \code{5}.
\item \code{max_depth} Maximum depth of any tree in the ensemble in the mean model. Default: \code{10}. Can be overridden with \code{-1} which does not enforce any depth limits on trees.
\item \code{sample_sigma2_leaf} Whether or not to update the leaf scale variance parameter based on \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)}. Cannot (currently) be set to true if \code{ncol(leaf_basis_train)>1}. Default: \code{FALSE}.
\item \code{sigma2_leaf_init} Starting value of leaf node scale parameter. Calibrated internally as \code{1/num_trees} if not set here.
\item \code{sigma2_leaf_shape} Shape parameter in the \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)} leaf node parameter variance model. Default: \code{3}.
\item \code{sigma2_leaf_scale} Scale parameter in the \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)} leaf node parameter variance model. Calibrated internally as \code{0.5/num_trees} if not set here.
\item \code{keep_vars} Vector of variable names or column indices denoting variables that should be included in the forest. Default: \code{NULL}.
\item \code{drop_vars} Vector of variable names or column indices denoting variables that should be excluded from the forest. Default: \code{NULL}. If both \code{drop_vars} and \code{keep_vars} are set, \code{drop_vars} will be ignored.
\item \code{num_features_subsample} How many features to subsample when growing each tree for the GFR algorithm. Defaults to the number of features in the training dataset.
}}

\item{variance_forest_params}{(Optional) A list of variance forest model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{num_trees} Number of trees in the ensemble for the conditional variance model. Default: \code{0}. Variance is only modeled using a tree / forest if \code{num_trees > 0}.
\item \code{alpha} Prior probability of splitting for a tree of depth 0 in the variance model. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{0.95}.
\item \code{beta} Exponent that decreases split probabilities for nodes of depth > 0 in the variance model. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{2}.
\item \code{min_samples_leaf} Minimum allowable size of a leaf, in terms of training samples, in the variance model. Default: \code{5}.
\item \code{max_depth} Maximum depth of any tree in the ensemble in the variance model. Default: \code{10}. Can be overridden with \code{-1} which does not enforce any depth limits on trees.
\item \code{leaf_prior_calibration_param} Hyperparameter used to calibrate the \code{IG(var_forest_prior_shape, var_forest_prior_scale)} conditional error variance model. If \code{var_forest_prior_shape} and \code{var_forest_prior_scale} are not set below, this calibration parameter is used to set these values to \code{num_trees / leaf_prior_calibration_param^2 + 0.5} and \code{num_trees / leaf_prior_calibration_param^2}, respectively. Default: \code{1.5}.
\item \code{var_forest_leaf_init} Starting value of root forest prediction in conditional (heteroskedastic) error variance model. Calibrated internally as \code{log(0.6*var(y_train))/num_trees}, where \code{y_train} is the possibly standardized outcome, if not set.
\item \code{var_forest_prior_shape} Shape parameter in the \code{IG(var_forest_prior_shape, var_forest_prior_scale)} conditional error variance model (which is only sampled if \code{num_trees > 0}). Calibrated internally as \code{num_trees / leaf_prior_calibration_param^2 + 0.5} if not set.
\item \code{var_forest_prior_scale} Scale parameter in the \code{IG(var_forest_prior_shape, var_forest_prior_scale)} conditional error variance model (which is only sampled if \code{num_trees > 0}). Calibrated internally as \code{num_trees / leaf_prior_calibration_param^2} if not set.
\item \code{keep_vars} Vector of variable names or column indices denoting variables that should be included in the forest. Default: \code{NULL}.
\item \code{drop_vars} Vector of variable names or column indices denoting variables that should be excluded from the forest. Default: \code{NULL}. If both \code{drop_vars} and \code{keep_vars} are set, \code{drop_vars} will be ignored.
\item \code{num_features_subsample} How many features to subsample when growing each tree for the GFR algorithm. Defaults to the number of features in the training dataset.
}}

\item{random_effects_params}{(Optional) A list of random effects model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{model_spec} Specification of the random effects model. Options are "custom" and "intercept_only". If "custom" is specified, then a user-provided basis must be passed through \code{rfx_basis_train}. If "intercept_only" is specified, a random effects basis of all ones will be dispatched internally at sampling and prediction time. If "intercept_plus_treatment" is specified, a random effects basis that combines an "intercept" basis of all ones with the treatment variable (\code{Z_train}) will be dispatched internally at sampling and prediction time. Default: "custom". If "intercept_only" is specified, \code{rfx_basis_train} and \code{rfx_basis_test} (if provided) will be ignored.
\item \code{working_parameter_prior_mean} Prior mean for the random effects "working parameter". Default: \code{NULL}. Must be a vector whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a vector.
\item \code{group_parameters_prior_mean} Prior mean for the random effects "group parameters." Default: \code{NULL}. Must be a vector whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a vector.
\item \code{working_parameter_prior_cov} Prior covariance matrix for the random effects "working parameter." Default: \code{NULL}. Must be a square matrix whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a diagonal matrix.
\item \code{group_parameter_prior_cov} Prior covariance matrix for the random effects "group parameters." Default: \code{NULL}. Must be a square matrix whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a diagonal matrix.
\item \code{variance_prior_shape} Shape parameter for the inverse gamma prior on the variance of the random effects "group parameter." Default: \code{1}.
\item \code{variance_prior_scale} Scale parameter for the inverse gamma prior on the variance of the random effects "group parameter." Default: \code{1}.
}}
}
\value{
List of sampling outputs and a wrapper around the sampled forests (which can be used for in-memory prediction on new data, or serialized to JSON on disk).
}
\description{
Run the BART algorithm for supervised learning.
}
\examples{
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
f_XW <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
noise_sd <- 1
y <- f_XW + rnorm(n, 0, noise_sd)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
y_test <- y[test_inds]
y_train <- y[train_inds]

bart_model <- bart(X_train = X_train, y_train = y_train, X_test = X_test,
                   num_gfr = 10, num_burnin = 0, num_mcmc = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forest.R
\name{createForest}
\alias{createForest}
\title{Create Forest Object}
\usage{
createForest(
  num_trees,
  leaf_dimension = 1,
  is_leaf_constant = FALSE,
  is_exponentiated = FALSE
)
}
\arguments{
\item{num_trees}{Number of trees in the forest}

\item{leaf_dimension}{Dimensionality of the outcome model}

\item{is_leaf_constant}{Whether leaf is constant}

\item{is_exponentiated}{Whether forest predictions should be exponentiated before being returned}
}
\value{
\code{Forest} object
}
\description{
Create a forest

This function is intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
num_trees <- 100
leaf_dimension <- 2
is_leaf_constant <- FALSE
is_exponentiated <- FALSE
forest <- createForest(num_trees, leaf_dimension, is_leaf_constant, is_exponentiated)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcf.R
\name{bcf}
\alias{bcf}
\title{Run BCF for Causal Effect Estimation}
\usage{
bcf(
  X_train,
  Z_train,
  y_train,
  propensity_train = NULL,
  rfx_group_ids_train = NULL,
  rfx_basis_train = NULL,
  X_test = NULL,
  Z_test = NULL,
  propensity_test = NULL,
  rfx_group_ids_test = NULL,
  rfx_basis_test = NULL,
  observation_weights = NULL,
  num_gfr = 5,
  num_burnin = 0,
  num_mcmc = 100,
  previous_model_json = NULL,
  previous_model_warmstart_sample_num = NULL,
  general_params = list(),
  prognostic_forest_params = list(),
  treatment_effect_forest_params = list(),
  variance_forest_params = list(),
  random_effects_params = list()
)
}
\arguments{
\item{X_train}{Covariates used to split trees in the ensemble. May be provided either as a dataframe or a matrix.
Matrix covariates will be assumed to be all numeric. Covariates passed as a dataframe will be
preprocessed based on the variable types (e.g. categorical columns stored as unordered factors will be one-hot encoded,
categorical columns stored as ordered factors will passed as integers to the core algorithm, along with the metadata
that the column is ordered categorical).}

\item{Z_train}{Vector of (continuous or binary) treatment assignments.}

\item{y_train}{Outcome to be modeled by the ensemble.}

\item{propensity_train}{(Optional) Vector of propensity scores. If not provided, this will be estimated from the data.
If \code{NULL} and \code{previous_model_json} is provided with an internally estimated propensity model, that model's
propensity estimates are re-used rather than re-fitted.}

\item{rfx_group_ids_train}{(Optional) Group labels used for an additive random effects model.}

\item{rfx_basis_train}{(Optional) Basis for "random-slope" regression in an additive random effects model.
If \code{rfx_group_ids_train} is provided with a regression basis, an intercept-only random effects model
will be estimated.}

\item{X_test}{(Optional) Test set of covariates used to define "out of sample" evaluation data.
May be provided either as a dataframe or a matrix, but the format of \code{X_test} must be consistent with
that of \code{X_train}.}

\item{Z_test}{(Optional) Test set of (continuous or binary) treatment assignments.}

\item{propensity_test}{(Optional) Vector of propensity scores. If not provided, this will be estimated from the data.}

\item{rfx_group_ids_test}{(Optional) Test set group labels used for an additive random effects model.
We do not currently support (but plan to in the near future), test set evaluation for group labels
that were not in the training set.}

\item{rfx_basis_test}{(Optional) Test set basis for "random-slope" regression in additive random effects model.}

\item{observation_weights}{(Optional) Numeric vector of observation weights of length \code{nrow(X_train)}. Weights are
applied as \code{y_i | - ~ N(mu(X_i), sigma^2 / w_i)}, so larger weights increase an observation's influence on the fit.
All weights must be non-negative. Default: \code{NULL} (all observations equally weighted). Applied to both the
prognostic and treatment effect forests. Compatible with Gaussian (continuous/identity) and probit outcome models;
not compatible with cloglog link functions.}

\item{num_gfr}{Number of "warm-start" iterations run using the grow-from-root algorithm (He and Hahn, 2021). Default: 5.}

\item{num_burnin}{Number of "burn-in" iterations of the MCMC sampler. Default: 0.}

\item{num_mcmc}{Number of "retained" iterations of the MCMC sampler. Default: 100.}

\item{previous_model_json}{(Optional) JSON string containing a previous BCF model. This can be used to "continue" a
sampler interactively after inspecting the samples or to run parallel chains "warm-started" from existing forest
samples. If the previous model used an internally estimated propensity score (i.e. \code{propensity_train} was not
supplied to that run), the fitted propensity model is carried forward and re-used rather than being re-estimated.
This ensures that multi-chain warm-starts remain consistent with the propensity scores used in the initial run.
Default: \code{NULL}.}

\item{previous_model_warmstart_sample_num}{(Optional) Sample number from \code{previous_model_json} that will be used to warmstart this BCF sampler. One-indexed (so that the first sample is used for warm-start by setting \code{previous_model_warmstart_sample_num = 1}). Default: \code{NULL}. If \code{num_chains} in the \code{general_params} list is > 1, then each successive chain will be initialized from a different sample, counting backwards from \code{previous_model_warmstart_sample_num}. That is, if \code{previous_model_warmstart_sample_num = 10} and \code{num_chains = 4}, then chain 1 will be initialized from sample 10, chain 2 from sample 9, chain 3 from sample 8, and chain 4 from sample 7. If \code{previous_model_json} is provided but \code{previous_model_warmstart_sample_num} is NULL, the last sample in the previous model will be used to initialize the first chain, counting backwards as noted before. If more chains are requested than there are samples in \code{previous_model_json}, a warning will be raised and only the last sample will be used.}

\item{general_params}{(Optional) A list of general (non-forest-specific) model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{cutpoint_grid_size} Maximum size of the "grid" of potential cutpoints to consider in the GFR algorithm. Default: \code{100}.
\item \code{standardize} Whether or not to standardize the outcome (and store the offset / scale in the model object). Default: \code{TRUE}.
\item \code{sample_sigma2_global} Whether or not to update the \code{sigma^2} global error variance parameter based on \code{IG(sigma2_global_shape, sigma2_global_scale)}. Default: \code{TRUE}.
\item \code{sigma2_global_init} Starting value of global error variance parameter. Calibrated internally as \code{1.0*var((y_train-mean(y_train))/sd(y_train))} if not set.
\item \code{sigma2_global_shape} Shape parameter in the \code{IG(sigma2_global_shape, sigma2_global_scale)} global error variance model. Default: \code{0}.
\item \code{sigma2_global_scale} Scale parameter in the \code{IG(sigma2_global_shape, sigma2_global_scale)} global error variance model. Default: \code{0}.
\item \code{variable_weights} Numeric weights reflecting the relative probability of splitting on each variable. Does not need to sum to 1 but cannot be negative. Defaults to \code{rep(1/ncol(X_train), ncol(X_train))} if not set here. Note that if the propensity score is included as a covariate in either forest, its weight will default to \code{1/ncol(X_train)}. A workaround if you wish to provide a custom weight for the propensity score is to include it as a column in \code{X_train} and then set \code{propensity_covariate} to \code{'none'} adjust \code{keep_vars} accordingly for the \code{prognostic} or \code{treatment_effect} forests.
\item \code{propensity_covariate} Whether to include the propensity score as a covariate in either or both of the forests. Enter \code{"none"} for neither, \code{"prognostic"} for the prognostic forest, \code{"treatment_effect"} for the treatment forest, and \code{"both"} for both forests. If this is not \code{"none"} and a propensity score is not provided, it will be estimated from (\code{X_train}, \code{Z_train}) using \code{stochtree::bart()}. Default: \code{"mu"}.
\item \code{adaptive_coding} Whether or not to use an "adaptive coding" scheme in which a binary treatment variable is not coded manually as (0,1) or (-1,1) but learned via parameters \code{b_0} and \code{b_1} that attach to the outcome model \verb{[b_0 (1-Z) + b_1 Z] tau(X)}. This is ignored when Z is not binary. Default: \code{FALSE}.
\item \code{control_coding_init} Initial value of the "control" group coding parameter. This is ignored when Z is not binary. Default: \code{-0.5}.
\item \code{treated_coding_init} Initial value of the "treatment" group coding parameter. This is ignored when Z is not binary. Default: \code{0.5}.
\item \code{rfx_prior_var} Prior on the (diagonals of the) covariance of the additive group-level random regression coefficients. Must be a vector of length \code{ncol(rfx_basis_train)}. Default: \code{rep(1, ncol(rfx_basis_train))}
\item \code{random_seed} Integer parameterizing the C++ random number generator. If not specified, the C++ random number generator is seeded according to \code{std::random_device}.
\item \code{keep_burnin} Whether or not "burnin" samples should be included in the stored samples of forests and other parameters. Default \code{FALSE}. Ignored if \code{num_mcmc = 0}.
\item \code{keep_gfr} Whether or not "grow-from-root" samples should be included in the stored samples of forests and other parameters. Default \code{FALSE}. Ignored if \code{num_mcmc = 0}.
\item \code{keep_every} How many iterations of the burned-in MCMC sampler should be run before forests and parameters are retained. Default \code{1}. Setting \code{keep_every <- k} for some \code{k > 1} will "thin" the MCMC samples by retaining every \code{k}-th sample, rather than simply every sample. This can reduce the autocorrelation of the MCMC samples.
\item \code{num_chains} How many independent MCMC chains should be sampled. If \code{num_mcmc = 0}, this is ignored. If \code{num_gfr = 0}, then each chain is run from root for \code{num_mcmc * keep_every + num_burnin} iterations, with \code{num_mcmc} samples retained. If \code{num_gfr > 0}, each MCMC chain will be initialized from a separate GFR ensemble, with the requirement that \code{num_gfr >= num_chains}. Default: \code{1}. Note that if \code{num_chains > 1}, the returned model object will contain samples from all chains, stored consecutively. That is, if there are 4 chains with 100 samples each, the first 100 samples will be from chain 1, the next 100 samples will be from chain 2, etc... For more detail on working with multi-chain BCF models, see \href{https://stochtree.ai/vignettes/multi-chain.html}{the multi chain vignette}.
\item \code{verbose} Whether or not to print progress during the sampling loops. Default: \code{FALSE}.
\item \code{outcome_model} A structured \code{OutcomeModel} object that specifies the outcome type and desired link function. This argument pre-empts the legacy (deprecated) \code{probit_outcome_model} option. Default: \code{OutcomeModel(outcome='continuous', link='identity')}.
\item \code{probit_outcome_model} Deprecated in favor of \code{outcome_model}. Whether or not the outcome should be modeled as explicitly binary via a probit link. If \code{TRUE}, \code{y} must only contain the values \code{0} and \code{1}. Default: \code{FALSE}.
\item \code{num_threads} Number of threads to use in the GFR and MCMC algorithms, as well as prediction. If OpenMP is not available on a user's setup, this will default to \code{1}, otherwise to the maximum number of available threads.
}}

\item{prognostic_forest_params}{(Optional) A list of prognostic forest model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{num_trees} Number of trees in the ensemble for the prognostic forest. Default: \code{250}. Must be a positive integer.
\item \code{alpha} Prior probability of splitting for a tree of depth 0 in the prognostic forest. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{0.95}.
\item \code{beta} Exponent that decreases split probabilities for nodes of depth > 0 in the prognostic forest. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{2}.
\item \code{min_samples_leaf} Minimum allowable size of a leaf, in terms of training samples, in the prognostic forest. Default: \code{5}.
\item \code{max_depth} Maximum depth of any tree in the ensemble in the prognostic forest. Default: \code{10}. Can be overridden with \code{-1} which does not enforce any depth limits on trees.
\item \code{variable_weights} Numeric weights reflecting the relative probability of splitting on each variable in the prognostic forest. Does not need to sum to 1 but cannot be negative. Defaults to \code{rep(1/ncol(X_train), ncol(X_train))} if not set here.
\item \code{sample_sigma2_leaf} Whether or not to update the leaf scale variance parameter based on \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)}.
\item \code{sigma2_leaf_init} Starting value of leaf node scale parameter. Calibrated internally as \code{1/num_trees} if not set here.
\item \code{sigma2_leaf_shape} Shape parameter in the \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)} leaf node parameter variance model. Default: \code{3}.
\item \code{sigma2_leaf_scale} Scale parameter in the \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)} leaf node parameter variance model. Calibrated internally as \code{0.5/num_trees} if not set here.
\item \code{keep_vars} Vector of variable names or column indices denoting variables that should be included in the forest. Default: \code{NULL}.
\item \code{drop_vars} Vector of variable names or column indices denoting variables that should be excluded from the forest. Default: \code{NULL}. If both \code{drop_vars} and \code{keep_vars} are set, \code{drop_vars} will be ignored.
\item \code{num_features_subsample} How many features to subsample when growing each tree for the GFR algorithm. Defaults to the number of features in the training dataset.
}}

\item{treatment_effect_forest_params}{(Optional) A list of treatment effect forest model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{num_trees} Number of trees in the ensemble for the treatment effect forest. Default: \code{100}. Must be a positive integer.
\item \code{alpha} Prior probability of splitting for a tree of depth 0 in the treatment effect forest. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{0.25}.
\item \code{beta} Exponent that decreases split probabilities for nodes of depth > 0 in the treatment effect forest. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{3}.
\item \code{min_samples_leaf} Minimum allowable size of a leaf, in terms of training samples, in the treatment effect forest. Default: \code{5}.
\item \code{max_depth} Maximum depth of any tree in the ensemble in the treatment effect forest. Default: \code{5}. Can be overridden with \code{-1} which does not enforce any depth limits on trees.
\item \code{variable_weights} Numeric weights reflecting the relative probability of splitting on each variable in the treatment effect forest. Does not need to sum to 1 but cannot be negative. Defaults to \code{rep(1/ncol(X_train), ncol(X_train))} if not set here.
\item \code{sample_sigma2_leaf} Whether or not to update the leaf scale variance parameter based on \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)}. Cannot (currently) be set to true if \code{ncol(Z_train)>1}. Default: \code{FALSE}.
\item \code{sigma2_leaf_init} Starting value of leaf node scale parameter. Calibrated internally as \code{0.5 * var(y)/num_trees} if not set here (\code{0.5 / num_trees} if \code{y} is continuous and \code{standardize = TRUE} in the \code{general_params} list).
\item \code{sigma2_leaf_shape} Shape parameter in the \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)} leaf node parameter variance model. Default: \code{3}.
\item \code{sigma2_leaf_scale} Scale parameter in the \code{IG(sigma2_leaf_shape, sigma2_leaf_scale)} leaf node parameter variance model. Calibrated internally as \code{0.5/num_trees} if not set here.
\item \code{delta_max} Maximum plausible conditional distributional treatment effect (i.e. P(Y(1) = 1 | X) - P(Y(0) = 1 | X)) when the outcome is binary. Only used when the outcome is specified as a probit model in \code{general_params}. Must be > 0 and < 1. Default: \code{0.9}. Ignored if \code{sigma2_leaf_init} is set directly, as this parameter is used to calibrate \code{sigma2_leaf_init}.
\item \code{keep_vars} Vector of variable names or column indices denoting variables that should be included in the forest. Default: \code{NULL}.
\item \code{drop_vars} Vector of variable names or column indices denoting variables that should be excluded from the forest. Default: \code{NULL}. If both \code{drop_vars} and \code{keep_vars} are set, \code{drop_vars} will be ignored.
\item \code{num_features_subsample} How many features to subsample when growing each tree for the GFR algorithm. Defaults to the number of features in the training dataset.
\item \code{sample_intercept} Whether to sample a global treatment effect intercept \code{tau_0} so the full CATE is \code{tau_0 + tau(X)}. Default: \code{TRUE}. Compatible with \code{adaptive_coding = TRUE}, in which case the recoded treatment basis is used.
\item \code{tau_0_prior_var} Variance of the normal prior on \code{tau_0} (a scalar applied to each treatment dimension independently). Auto-calibrated to outcome variance when \code{NULL} and outcome is continuous. Only used when \code{sample_intercept = TRUE}.
}}

\item{variance_forest_params}{(Optional) A list of variance forest model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{num_trees} Number of trees in the ensemble for the conditional variance model. Default: \code{0}. Variance is only modeled using a tree / forest if \code{num_trees > 0}.
\item \code{alpha} Prior probability of splitting for a tree of depth 0 in the variance model. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{0.95}.
\item \code{beta} Exponent that decreases split probabilities for nodes of depth > 0 in the variance model. Tree split prior combines \code{alpha} and \code{beta} via \code{alpha*(1+node_depth)^-beta}. Default: \code{2}.
\item \code{min_samples_leaf} Minimum allowable size of a leaf, in terms of training samples, in the variance model. Default: \code{5}.
\item \code{max_depth} Maximum depth of any tree in the ensemble in the variance model. Default: \code{10}. Can be overridden with \code{-1} which does not enforce any depth limits on trees.
\item \code{leaf_prior_calibration_param} Hyperparameter used to calibrate the \code{IG(var_forest_prior_shape, var_forest_prior_scale)} conditional error variance model. If \code{var_forest_prior_shape} and \code{var_forest_prior_scale} are not set below, this calibration parameter is used to set these values to \code{num_trees / leaf_prior_calibration_param^2 + 0.5} and \code{num_trees / leaf_prior_calibration_param^2}, respectively. Default: \code{1.5}.
\item \code{variance_forest_init} Starting value of root forest prediction in conditional (heteroskedastic) error variance model. Calibrated internally as \code{log(0.6*var((y_train-mean(y_train))/sd(y_train)))/num_trees} if not set.
\item \code{var_forest_prior_shape} Shape parameter in the \code{IG(var_forest_prior_shape, var_forest_prior_scale)} conditional error variance model (which is only sampled if \code{num_trees > 0}). Calibrated internally as \code{num_trees / 1.5^2 + 0.5} if not set.
\item \code{var_forest_prior_scale} Scale parameter in the \code{IG(var_forest_prior_shape, var_forest_prior_scale)} conditional error variance model (which is only sampled if \code{num_trees > 0}). Calibrated internally as \code{num_trees / 1.5^2} if not set.
\item \code{keep_vars} Vector of variable names or column indices denoting variables that should be included in the forest. Default: \code{NULL}.
\item \code{drop_vars} Vector of variable names or column indices denoting variables that should be excluded from the forest. Default: \code{NULL}. If both \code{drop_vars} and \code{keep_vars} are set, \code{drop_vars} will be ignored.
\item \code{num_features_subsample} How many features to subsample when growing each tree for the GFR algorithm. Defaults to the number of features in the training dataset.
}}

\item{random_effects_params}{(Optional) A list of random effects model parameters, each of which has a default value processed internally, so this argument list is optional.
\itemize{
\item \code{model_spec} Specification of the random effects model. Options are "custom", "intercept_only", and "intercept_plus_treatment". If "custom" is specified, then a user-provided basis must be passed through \code{rfx_basis_train}. If "intercept_only" is specified, a random effects basis of all ones will be dispatched internally at sampling and prediction time. If "intercept_plus_treatment" is specified, a random effects basis that combines an "intercept" basis of all ones with the treatment variable (\code{Z_train}) will be dispatched internally at sampling and prediction time. Default: "custom". If either "intercept_only" or "intercept_plus_treatment" is specified, \code{rfx_basis_train} and \code{rfx_basis_test} (if provided) will be ignored.
\item \code{working_parameter_prior_mean} Prior mean for the random effects "working parameter". Default: \code{NULL}. Must be a vector whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a vector.
\item \code{group_parameters_prior_mean} Prior mean for the random effects "group parameters." Default: \code{NULL}. Must be a vector whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a vector.
\item \code{working_parameter_prior_cov} Prior covariance matrix for the random effects "working parameter." Default: \code{NULL}. Must be a square matrix whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a diagonal matrix.
\item \code{group_parameter_prior_cov} Prior covariance matrix for the random effects "group parameters." Default: \code{NULL}. Must be a square matrix whose dimension matches the number of random effects bases, or a scalar value that will be expanded to a diagonal matrix.
\item \code{variance_prior_shape} Shape parameter for the inverse gamma prior on the variance of the random effects "group parameter." Default: \code{1}.
\item \code{variance_prior_scale} Scale parameter for the inverse gamma prior on the variance of the random effects "group parameter." Default: \code{1}.
}}
}
\value{
List of sampling outputs and a wrapper around the sampled forests (which can be used for in-memory prediction on new data, or serialized to JSON on disk).
}
\description{
Run the Bayesian Causal Forest (BCF) algorithm for regularized causal effect estimation.
}
\examples{
n <- 500
p <- 5
X <- matrix(runif(n*p), ncol = p)
mu_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
pi_x <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (0.2) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (0.4) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (0.6) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (0.8)
)
tau_x <- (
    ((0 <= X[,2]) & (0.25 > X[,2])) * (0.5) +
    ((0.25 <= X[,2]) & (0.5 > X[,2])) * (1.0) +
    ((0.5 <= X[,2]) & (0.75 > X[,2])) * (1.5) +
    ((0.75 <= X[,2]) & (1 > X[,2])) * (2.0)
)
Z <- rbinom(n, 1, pi_x)
noise_sd <- 1
y <- mu_x + tau_x*Z + rnorm(n, 0, noise_sd)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
pi_test <- pi_x[test_inds]
pi_train <- pi_x[train_inds]
Z_test <- Z[test_inds]
Z_train <- Z[train_inds]
y_test <- y[test_inds]
y_train <- y[train_inds]
mu_test <- mu_x[test_inds]
mu_train <- mu_x[train_inds]
tau_test <- tau_x[test_inds]
tau_train <- tau_x[train_inds]
bcf_model <- bcf(X_train = X_train, Z_train = Z_train, y_train = y_train,
                 propensity_train = pi_train, X_test = X_test, Z_test = Z_test,
                 propensity_test = pi_test, num_gfr = 10,
                 num_burnin = 0, num_mcmc = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel.R
\name{ForestKernelComputation}
\alias{ForestKernelComputation}
\alias{computeForestLeafIndices}
\alias{computeForestLeafVariances}
\alias{computeForestMaxLeafIndex}
\title{Forest Kernel Computation Routines}
\usage{
computeForestLeafIndices(
  model_object,
  covariates,
  forest_type = NULL,
  propensity = NULL,
  forest_inds = NULL
)

computeForestLeafVariances(model_object, forest_type, forest_inds = NULL)

computeForestMaxLeafIndex(model_object, forest_type = NULL, forest_inds = NULL)
}
\arguments{
\item{model_object}{Object of type \code{bartmodel}, \code{bcfmodel}, or \code{ForestSamples} corresponding to a BART / BCF model with at least one forest sample, or a low-level \code{ForestSamples} object.}

\item{covariates}{Covariates to use for prediction. Must have the same dimensions / column types as the data used to train a forest.}

\item{forest_type}{Which forest to use from \code{model_object}.
Valid inputs depend on the model type, and whether or not a

\strong{1. BART}
\itemize{
\item \code{'mean'}: Extracts leaf indices for the mean forest
\item \code{'variance'}: Extracts leaf indices for the variance forest
}

\strong{2. BCF}
\itemize{
\item \code{'prognostic'}: Extracts leaf indices for the prognostic forest
\item \code{'treatment'}: Extracts leaf indices for the treatment effect forest
\item \code{'variance'}: Extracts leaf indices for the variance forest
}

\strong{3. ForestSamples}
\itemize{
\item \code{NULL}: It is not necessary to disambiguate when this function is called directly on a \code{ForestSamples} object. This is the default value of this
}}

\item{propensity}{(Optional) Propensities used for prediction (BCF-only).}

\item{forest_inds}{(Optional) Indices of the forest sample(s) for which to compute max leaf indices. If not provided,
this function will return max leaf indices for every sample of a forest.
This function uses 0-indexing, so the first forest sample corresponds to \code{forest_num = 0}, and so on.}
}
\value{
\code{computeForestLeafIndices} returns a vector of size \code{num_obs * num_trees}, where \code{num_obs = nrow(covariates)}
and \code{num_trees} is the number of trees in the relevant forest of \code{model_object}.

\code{computeForestLeafVariances} returns a vector of size \code{length(forest_inds)} with the leaf scale parameter for each requested forest.

\code{computeForestMaxLeafIndex} returns a vector containing the largest possible leaf index computable by \code{computeForestLeafIndices} for the forests in a designated forest sample container.
}
\description{
Decision tree ensembles can be represented in part by a "kernel" function whose distance metric
is based on the extent to which two observations are mapped to the same leaf nodes.
This function group offers utilities for evaluating this kernel.

\code{computeForestLeafIndices} computes and return a vector representation of a forest's
leaf predictions for every observation in a dataset.
The resulting vector has a "row-major" format that can be easily re-represented as
as a CSR sparse matrix: elements are organized so that the first \code{n} elements
correspond to leaf predictions for all \code{n} observations in a dataset for the
first tree in an ensemble, the next \code{n} elements correspond to predictions for
the second tree and so on. The "data" for each element corresponds to a uniquely
mapped column index that corresponds to a single leaf of a single tree (i.e.
if tree 1 has 3 leaves, its column indices range from 0 to 2, and then tree 2's
leaf indices begin at 3, etc...).

\code{computeForestLeafVariances} returns each forest's leaf node scale parameters.
If leaf scale is not sampled for the forest in question, the function throws an error that the
leaf model does not have a stochastic scale parameter.

\code{computeForestMaxLeafIndex} computes and returns the largest possible leaf index computable by \code{computeForestLeafIndices} for the forests in a designated forest sample container.

These functions are intended for advanced use cases in which users require detailed control of sampling algorithms and data structures.
Minimal input validation and error checks are performed -- users are responsible for providing the correct inputs.
For tutorials on the "proper" usage of the stochtree's advanced workflow, we provide several vignettes at \url{https://stochtree.ai/}
}
\examples{
X <- matrix(runif(10*100), ncol = 10)
y <- -5 + 10*(X[,1] > 0.5) + rnorm(100)
bart_model <- bart(X, y, num_gfr=0, num_mcmc=10)
leaf_indices <- computeForestLeafIndices(bart_model, X, "mean")
leaf_indices <- computeForestLeafIndices(bart_model, X, "mean", 0)
leaf_indices <- computeForestLeafIndices(bart_model, X, "mean", c(1,3,9))
leaf_variances <- computeForestLeafVariances(bart_model, "mean")
leaf_variances <- computeForestLeafVariances(bart_model, "mean", 0)
leaf_variances <- computeForestLeafVariances(bart_model, "mean", c(1,3,5))
max_leaf_index <- computeForestMaxLeafIndex(bart_model, "mean")
max_leaf_index <- computeForestMaxLeafIndex(bart_model, "mean", 0)
max_leaf_index <- computeForestMaxLeafIndex(bart_model, "mean", c(1,3,9))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcf.R
\name{print.bcfmodel}
\alias{print.bcfmodel}
\title{Print Summary of BCF Model}
\usage{
\method{print}{bcfmodel}(x, ...)
}
\arguments{
\item{x}{The BCF model object}

\item{...}{Additional arguments (currently unused)}
}
\value{
BCF model object unchanged after printing summary
}
\description{
Prints a summary of the BCF model, including the model terms and their specifications.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bart.R
\name{predict.bartmodel}
\alias{predict.bartmodel}
\title{Predict from a BART Model}
\usage{
\method{predict}{bartmodel}(
  object,
  X,
  leaf_basis = NULL,
  rfx_group_ids = NULL,
  rfx_basis = NULL,
  type = "posterior",
  terms = "all",
  scale = "linear",
  ...
)
}
\arguments{
\item{object}{Object of type \code{bart} containing draws of a regression forest and associated sampling outputs.}

\item{X}{Covariates used to determine tree leaf predictions for each observation. Must be passed as a matrix or dataframe.}

\item{leaf_basis}{(Optional) Bases used for prediction (by e.g. dot product with leaf values). Default: \code{NULL}.}

\item{rfx_group_ids}{(Optional) Test set group labels used for an additive random effects model.
We do not currently support (but plan to in the near future), test set evaluation for group labels
that were not in the training set.}

\item{rfx_basis}{(Optional) Test set basis for "random-slope" regression in additive random effects model.}

\item{type}{(Optional) Type of prediction to return. Options are "mean", which averages the predictions from every draw of a BART model, and "posterior", which returns the entire matrix of posterior predictions. Default: "posterior".}

\item{terms}{(Optional) Which model terms to include in the prediction. This can be a single term or a list of model terms. Options include "y_hat", "mean_forest", "rfx", "variance_forest", or "all". If a model doesn't have mean forest, random effects, or variance forest predictions, but one of those terms is request, the request will simply be ignored. If none of the requested terms are present in a model, this function will return \code{NULL} along with a warning. Default: "all".}

\item{scale}{(Optional) Scale of mean function predictions. Options are "linear", which returns predictions on the original scale of the mean forest / RFX terms, "probability", which transforms predictions into class probabilities for models with discrete outcomes, and "class", which returns predicted outcome categories for discrete outcome models. "probability" is only valid for outcome models with \code{outcome == 'binary'} or \code{outcome == 'ordinal'}. For binary outcomes, this will return the probability that \code{y == 1}, and for ordinal outcomes, this will return probabilities for each outcome label. Default: "linear".}

\item{...}{(Optional) Other prediction parameters.}
}
\value{
List of prediction matrices or single prediction matrix / vector, depending on the terms requested.
}
\description{
Predict from a sampled BART model on new data
}
\examples{
n <- 100
p <- 5
X <- matrix(runif(n*p), ncol = p)
f_XW <- (
    ((0 <= X[,1]) & (0.25 > X[,1])) * (-7.5) +
    ((0.25 <= X[,1]) & (0.5 > X[,1])) * (-2.5) +
    ((0.5 <= X[,1]) & (0.75 > X[,1])) * (2.5) +
    ((0.75 <= X[,1]) & (1 > X[,1])) * (7.5)
)
noise_sd <- 1
y <- f_XW + rnorm(n, 0, noise_sd)
test_set_pct <- 0.2
n_test <- round(test_set_pct*n)
n_train <- n - n_test
test_inds <- sort(sample(1:n, n_test, replace = FALSE))
train_inds <- (1:n)[!((1:n) \%in\% test_inds)]
X_test <- X[test_inds,]
X_train <- X[train_inds,]
y_test <- y[test_inds]
y_train <- y[train_inds]
bart_model <- bart(X_train = X_train, y_train = y_train,
                   num_gfr = 10, num_burnin = 0, num_mcmc = 10)
y_hat_test <- predict(bart_model, X=X_test)$y_hat
}
